//#####################################################################
// Copyright 2013, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_INTERSECTION__
#define __CODIMENSIONAL_INTERSECTION__

#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Arrays/ARRAY.h>

#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"

namespace PhysBAM
{

template<class T>
class CYLINDER;
template<class T>
class SEGMENT_3D;
template <class T>
class TETRAHEDRON;
template<class T>
class TRIANGLE_3D;
    
template<class T>
class CODIMENSIONAL_INTERSECTION
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<T,1> TV1;
    typedef VECTOR<T,2> TV2;
    typedef VECTOR<T,3> TV3;
    typedef VECTOR<T,4> TV4;
    typedef VECTOR<int,1> TV_INT1;
    typedef VECTOR<int,2> TV_INT2;
    typedef VECTOR<int,3> TV_INT3;
    typedef VECTOR<int,4> TV_INT4;
public:

    //Intersection tests
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,int simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT2 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT3 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT4 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}

    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,int simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT2 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT4 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}

    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,int simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT2 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT3 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT4 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}

    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,int simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT2 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT3 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}
    static bool Geometries_Intersect(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT4 simplex2){return Distance_Between_Geometries(X1,X2,t1,t2,simplex1,simplex2)<=(T)0;}

    //Other cases
    static bool Cylinder_Triangle_Thickness_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2);
    static bool Cylinder_Tetrahedron_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT4 simplex2);
    static bool Capsule_Triangle_Thickness_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2,const TV& expansion_normal)
    {return Distance_Capsule_To_Triangle(X1,X2,t1,t2,simplex1,simplex2,expansion_normal)<=(T)0;}
    static bool Segment_Capsule_Intersection(const TV& s1,const TV& s2,const ARRAY_VIEW<TV>& X,const ARRAY_VIEW<T>& t,TV_INT2 simplex);
    static bool Prism_Prism_Intersection(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT3 simplex2);

private:
    //Distance tests
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,int simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT2 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT3 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int simplex1,TV_INT4 simplex2);

    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,int simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT2 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT3 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 simplex1,TV_INT4 simplex2);

    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,int simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT2 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT3 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 simplex1,TV_INT4 simplex2);

    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,int simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT2 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT3 simplex2);
    static T Distance_Between_Geometries(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 simplex1,TV_INT4 simplex2);

public:
    //Point to...
    static T Distance_Point_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element1,int point_element2);
    static T Distance_Point_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT2 segment_element);
    static T Distance_Point_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT3 triangle_element);
    static T Distance_Point_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point_element,TV_INT4 tetrahedron_element);

    //Segment to...
    static T Distance_Capsule_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,int point_element);
    static T Distance_Capsule_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element1,TV_INT2 segment_element2);
    static T Distance_Capsule_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT3 triangle_element);
    static T Distance_Capsule_To_Triangle(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT3 triangle_element,const TV& expansion_normal);
    static T Distance_Capsule_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT4 tetrahedron_element);

    //Triangle to...
    static T Distance_Triangle_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,int point_element);
    static T Distance_Triangle_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,TV_INT2 segment_element);
    static T Distance_Triangle_To_Triangle_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element1,TV_INT3 triangle_element2);
    static T Distance_Triangle_To_Tetrahedron_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT3 triangle_element,TV_INT4 tetrahedron_element);

    //Tetrahedra to...
    static T Distance_Tetrahedron_To_Point(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,int point_element);
    static T Distance_Tetrahedron_To_Capsule(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,TV_INT2 segment_element);
    static T Distance_Tetrahedron_To_Triangle_Thickness(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element,TV_INT3 triangle_element);
    static T Distance_Tetrahedron_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT4 tetrahedron_element1,TV_INT4 tetrahedron_element2);

    //Support functions
    static TV Closest_Point_On_Line_Segment(const TV& point,const TV& point_a,const TV& point_b,T& t);
    static TV Closest_Point_On_Plane(const TV& point,const TV& point_on_plane,const TV& plane_normal);
    static T Distance_Segment_To_Segment(const TV& p1,const TV& q1,const TV& p2,const TV& q2);
    static T Distance_Segment_To_Triangle(const SEGMENT_3D<T>& segment,const TRIANGLE_3D<T>& triangle);
    static T Distance_Segment_To_Tetrahedron(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,TV_INT2 segment_element,TV_INT4 tetrahedron_element);
    static T Distance_Triangle_To_Triangle(const TRIANGLE_3D<T>& tri1,const TRIANGLE_3D<T>& tri2);

    //Misc
    static T Fraction_Of_Segment_In_Sphere(const ARRAY_VIEW<TV>& X1,const ARRAY_VIEW<TV>& X2,const ARRAY_VIEW<T>& t1,const ARRAY_VIEW<T>& t2,int point,TV_INT2 segment);
};

}


#endif //__CODIMENSIONAL_INTERSECTION__
