//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <fstream>
using std::ifstream;
using std::string;

//#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>

#include "CODIMENSIONAL_INTERSECTION_DRIVER.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_INTERSECTION_DRIVER<TV>::CODIMENSIONAL_INTERSECTION_DRIVER()
    :test(0),default_thickness((T)0),input_filename("")
{
    Initialize_Particles();
    //Initialize_Read_Write_General_Structures();//?
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INTERSECTION_DRIVER<TV>::Initialize()
{
    if(input_filename.empty()){
        Initialize_Data();

        example.deformable_geometry_collection.Add_Structure(example.codimensional_mesh_object.Get_Mesh_Object_1D());
        example.deformable_geometry_collection.structures.Last()->Update_Number_Nodes();
        example.deformable_geometry_collection.Add_Structure(example.codimensional_mesh_object.Get_Mesh_Object_2D());
        example.deformable_geometry_collection.structures.Last()->Update_Number_Nodes();
        if(TV::dimension==3)example.deformable_geometry_collection.Add_Structure(example.codimensional_mesh_object.Get_Mesh_Object_3D());
        example.deformable_geometry_collection.structures.Last()->Update_Number_Nodes();

        LOG::cout<<"number of tetrahedra in example: "<<example.tetrahedron_mesh.elements.m<<std::endl;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INTERSECTION_DRIVER<TV>::Initialize_Data()
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    ARRAY<TV> X;
    ARRAY<TV> V;
    ARRAY<T> thickness;
    switch(test){
    case 0:////point-point test
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 10:////point-segment test
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)1;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        example.segment_mesh.elements.Append(VECTOR<int,2>(2,3));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 11:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)1;
        x.x=(T)-1;
        X.Append(x);
        x.y=(T)0;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        example.segment_mesh.elements.Append(VECTOR<int,2>(2,3));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 12:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)1;
        x.x=(T)1;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        example.segment_mesh.elements.Append(VECTOR<int,2>(2,3));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 13:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.x=(T)4;
        X.Append(x);
        x.x=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        example.segment_mesh.elements.Append(VECTOR<int,2>(2,3));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 14:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.x=(T)-4;
        x.y=(T)1;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        example.segment_mesh.elements.Append(VECTOR<int,2>(2,3));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 20:////point-triangle test
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)1;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        x(3)=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        x.x=(T)0;
        x(3)=(T)1;
        X.Append(x);
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,3,4));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 21:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)-1;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        x(3)=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        x.x=(T)0;
        x(3)=(T)1;
        X.Append(x);
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,3,4));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 22:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)-1;
        x(3)=5;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        x(3)=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        x.x=(T)0;
        x(3)=(T)1;
        X.Append(x);
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,3,4));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 23:
        {PHYSBAM_ASSERT(TV::m==3);
        TV x=TV();
        x.y=(T)1;
        x(3)=5;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        x(3)=(T)-1;
        X.Append(x);
        x.x=(T)1;
        X.Append(x);
        x.x=(T)0;
        x(3)=(T)1;
        X.Append(x);
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,3,4));
        for(int i=1;i<=X.m;i++)thickness.Append(default_thickness);
        }break;
    case 100: ////two tetrahedra colliding
        {PHYSBAM_ASSERT(TV::m==3);
        T initial_height=(T)4.5;
        TV x=TV();
        x.y=(T)-1;
        X.Append(x);
        x.y+=initial_height;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)-1;
        x(3)=(T)-1;
        X.Append(x);
        x.y+=initial_height;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)1;
        X.Append(x);
        x.y+=initial_height;
        X.Append(x);
        x.y=(T)0;
        x.x=(T)0;
        x(3)=(T)1;
        X.Append(x);
        x.y+=initial_height;
        X.Append(x);
        for(int i=1;i<=8;i++){
            if(i%2==0)V.Append(TV_Compatible<T,TV::m>((T)0,(T)-1,(T)0));
            else V.Append(TV());
        }
        example.tetrahedron_mesh.elements.Append(VECTOR<int,4>(1,3,5,7));//does order matter?
        example.tetrahedron_mesh.elements.Append(VECTOR<int,4>(2,4,6,8));//does order matter?
        //use_mesh_merge=true;
        }break;
    case 101: ////four triangles
        {PHYSBAM_ASSERT(TV::m==3);
        X.Append(TV_Compatible<T,TV::m>((T)-1,(T)0,(T)0));//1->0
        X.Append(TV_Compatible<T,TV::m>((T)1,(T)0,(T)0));//2->2
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)-0.5,(T)1));//3->1
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)-0.5,(T)-1));//4->3
        X.Append(TV_Compatible<T,TV::m>((T)-1,(T)1,(T)0));//5->4
        X.Append(TV_Compatible<T,TV::m>((T)1,(T)1,(T)0));//6->6
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)1.5,(T)1));//7->7
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)1.5,(T)-1));//8->5
        example.triangle_mesh.elements.Append(VECTOR<int,3>(1,2,3));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(1,4,2));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(5,7,6));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(5,6,8));
        }break;
    case 102: ////same points as case 101, meshed together
        {PHYSBAM_ASSERT(TV::m==3);
        X.Append(TV_Compatible<T,TV::m>((T)-1,(T)0,(T)0));//1->0
        X.Append(TV_Compatible<T,TV::m>((T)1,(T)0,(T)0));//2->2
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)-0.5,(T)1));//3->1
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)-0.5,(T)-1));//4->3
        X.Append(TV_Compatible<T,TV::m>((T)-1,(T)1,(T)0));//5->4
        X.Append(TV_Compatible<T,TV::m>((T)1,(T)1,(T)0));//6->6
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)1.5,(T)1));//7->7
        X.Append(TV_Compatible<T,TV::m>((T)0,(T)1.5,(T)-1));//8->5
        example.triangle_mesh.elements.Append(VECTOR<int,3>(1,5,3));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(3,5,7));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(3,7,2));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,7,6));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(2,6,4));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(4,6,8));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(4,8,1));
        example.triangle_mesh.elements.Append(VECTOR<int,3>(1,8,5));
        }break;
    default:{PHYSBAM_FATAL_ERROR("Invalid test id in CODIMENSIONAL_INTERSECTION_DRIVER");break;}
    }

    MPI_CODIMENSIONAL<TV>* null_ptr=0;
    example.codimensional_mesh.Update_States(X.m,null_ptr);

    while(thickness.m<X.m)thickness.Append((T)1.0);
    while(V.m<X.m)V.Append(TV());
    for(int i=1;i<=X.m;i++){
        int index=particles.Add_Particle();
        particles.X(index)=X(i);
        particles.thickness(index)=thickness(i);
        particles.V(index)=V(i);
        //particles.dim(index)=3;
    }
    switch(test){
    case 0:////point-point test
        {
        //example.Run_Distance_Test(1,true,2,true);
        }break;
    case 10:////point-segment test
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 11:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 12:
        //example.Run_Distance_Test(1,true,1,false);
    break;
    case 13:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 14:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 20:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 21:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 22:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    case 23:
        //example.Run_Distance_Test(1,true,1,false);
        break;
    default:break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INTERSECTION_DRIVER<TV>::Run_Input_Tests()
{
    ifstream fin(input_filename.c_str());
    if(fin.fail()){
        LOG::cout<<"error - failed to open input file "<<input_filename<<"."<<std::endl;
        return;
    }
    string input;
    int test_no;
    T x,y,z,thickness;
    int index;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    ARRAY<bool> is_particle;
    ARRAY<int> particle_indices;
    //ARRAY<int> simplex_indices;
    ARRAY<int> segment_indices;
    ARRAY<int> triangle_indices;
    ARRAY<int> tetrahedron_indices;
    T distance;
    bool verbose;
    fin>>input>>input;//read verbose
    verbose=(input=="true"?true:false);
    MPI_CODIMENSIONAL<TV>* null_ptr=0;
    while(fin>>input){//read 'test'
        particles.array_collection->Delete_All_Elements();
        example.codimensional_mesh.segment_mesh.elements.Remove_All();
        example.codimensional_mesh.triangle_mesh.elements.Remove_All();
        example.codimensional_mesh.tetrahedron_mesh.elements.Remove_All();

        is_particle.Remove_All();
        particle_indices.Remove_All();
        segment_indices.Remove_All();
        triangle_indices.Remove_All();
        tetrahedron_indices.Remove_All();
        fin>>test_no>>input;
        distance=(T)-1;
        LOG::cout<<"Test "<<test_no<<":"<<std::endl;
        while(input!="end"){
            if(input=="particle"){
                fin>>x>>y>>z>>thickness;
                index=particles.Add_Particle();
                particles.X(index)=TV(x,y,z);
                particles.thickness(index)=thickness;
                is_particle.Append(true);
                particle_indices.Append(index);
                segment_indices.Append(-1);
                triangle_indices.Append(-1);
                tetrahedron_indices.Append(-1);
            }
            else if(input=="segment"){
                for(int i=1;i<=2;i++){
                    fin>>x>>y>>z>>thickness;
                    index=particles.Add_Particle();
                    particles.X(index)=TV(x,y,z);
                    particles.thickness(index)=thickness;
                }
                is_particle.Append(false);
                particle_indices.Append(-1);
                example.segment_mesh.elements.Append(VECTOR<int,2>(index-1,index));
                example.codimensional_mesh.Update_States(particles.X.m,null_ptr);
                segment_indices.Append(example.segment_mesh.elements.m);
                triangle_indices.Append(-1);
                tetrahedron_indices.Append(-1);

            }
            else if(input=="triangle"){
                for(int i=1;i<=3;i++){
                    fin>>x>>y>>z>>thickness;
                    index=particles.Add_Particle();
                    particles.X(index)=TV(x,y,z);
                    particles.thickness(index)=thickness;
                }
                is_particle.Append(false);
                particle_indices.Append(-1);
                example.triangle_mesh.elements.Append(VECTOR<int,3>(index-2,index-1,index));
                example.codimensional_mesh.Update_States(particles.X.m,null_ptr);
                segment_indices.Append(-1);
                triangle_indices.Append(example.triangle_mesh.elements.m);
                tetrahedron_indices.Append(-1);
            }
            else if(input=="tetrahedron"){
                for(int i=1;i<=4;i++){
                    fin>>x>>y>>z>>thickness;
                    index=particles.Add_Particle();
                    particles.X(index)=TV(x,y,z);
                    particles.thickness(index)=thickness;
                }
                is_particle.Append(false);
                particle_indices.Append(-1);
                example.tetrahedron_mesh.elements.Append(VECTOR<int,4>(index-3,index-2,index-1,index));
                example.codimensional_mesh.Update_States(particles.X.m,null_ptr);
                segment_indices.Append(-1);
                triangle_indices.Append(-1);
                tetrahedron_indices.Append(example.tetrahedron_mesh.elements.m);

            }
            else if(input=="distance"){
                fin>>distance;
            }
            else if(input=="run"){
                int id1,id2;
                fin>>id1>>id2;

                TV_INT4 arg1(0,0,0,0),arg2(0,0,0,0);
                if(particle_indices(id1)!=-1)arg1(1)=particle_indices(id1);
                else if(segment_indices(id1)!=-1){for(int i=1;i<=2;i++)arg1(i)=example.segment_mesh.elements(segment_indices(id1))(i);}
                else if(triangle_indices(id1)!=-1){for(int i=1;i<=3;i++)arg1(i)=example.triangle_mesh.elements(triangle_indices(id1))(i);}
                else if(tetrahedron_indices(id1)!=-1){for(int i=1;i<=4;i++)arg1(i)=example.tetrahedron_mesh.elements(tetrahedron_indices(id1))(i);}

                if(particle_indices(id2)!=-1)arg2(1)=particle_indices(id2);
                else if(segment_indices(id2)!=-1){for(int i=1;i<=2;i++)arg2(i)=example.segment_mesh.elements(segment_indices(id2))(i);}
                else if(triangle_indices(id2)!=-1){for(int i=1;i<=3;i++)arg2(i)=example.triangle_mesh.elements(triangle_indices(id2))(i);}
                else if(tetrahedron_indices(id2)!=-1){for(int i=1;i<=4;i++)arg2(i)=example.tetrahedron_mesh.elements(tetrahedron_indices(id2))(i);}

                /*
                T solution=example.Run_Distance_Test(arg1,arg2,verbose);

                if(distance!=(T)-1){
                    const T epsilon=(T)1e-4;
                    LOG::cout<<"suggested solution: "<<distance<<", calculated distance: "<<solution<<std::endl;
                    LOG::cout<<"test "<<test_no<<(abs(solution-distance)<epsilon?" PASSED.":" FAILED.")<<std::endl;
                }
                else{
                    LOG::cout<<"calculated distance: "<<solution<<std::endl;
                }
                */

                bool intersection=example.Run_Intersection_Test(arg1,arg2);
                LOG::cout<<"Intersection: "<<(intersection?"yes":"no")<<std::endl;
            }
            else{
                getline(fin,input);//ignore line
            }
            fin>>input;
        }
    }
    fin.close();
}


template<class TV>
void PhysBAM::CODIMENSIONAL_INTERSECTION_DRIVER<TV>::Execute_Main_Program()
{
    Run_Input_Tests();
}

//template class CODIMENSIONAL_INTERSECTION_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_INTERSECTION_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_INTERSECTION_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_INTERSECTION_DRIVER<VECTOR<double,3> >;
#endif
