//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>

#include "CODIMENSIONAL_INTERSECTION_EXAMPLE.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_INTERSECTION_EXAMPLE<TV>::CODIMENSIONAL_INTERSECTION_EXAMPLE()
    :
    codimensional_mesh(&segment_mesh,&triangle_mesh,&tetrahedron_mesh),codimensional_mesh_object(codimensional_mesh,particles),//codimensional_intersection(codimensional_mesh_object),
    deformable_geometry_collection(particles,false)
{
    particles.Initialize_Array_Collection();
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_INTERSECTION_EXAMPLE<TV>::Run_Distance_Test(TV_INT4 arg1,TV_INT4 arg2,bool verbose)
{

    //distance functions are now private...
    /*
    bool tet1=(arg1(4)!=0);
    bool tri1=(arg1(3)!=0) && !tet1;
    bool seg1=(arg1(2)!=0) && !tri1 && !tet1;
    bool pnt1=(arg1(1)!=0) && !seg1 && !tri1 && !tet1;

    bool tet2=(arg2(4)!=0);
    bool tri2=(arg2(3)!=0) && !tet2;
    bool seg2=(arg2(2)!=0) && !tri2 && !tet2;
    bool pnt2=(arg2(1)!=0) && !seg2 && !tri2 && !tet2;

    int p1=arg1(1);
    int p2=arg2(1);

    TV_INT2 s1(arg1(1),arg1(2));
    TV_INT2 s2(arg2(1),arg2(2));

    TV_INT3 tr1(arg1(1),arg1(2),arg1(3));
    TV_INT3 tr2(arg2(1),arg2(2),arg2(3));

    TV_INT4 te1(arg1(1),arg1(2),arg1(3),arg1(4));
    TV_INT4 te2(arg2(1),arg2(2),arg2(3),arg2(4));

    T distance=(T)0;

    if(pnt1){
        if(pnt2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,p1,p2);
        else if(seg2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,p1,s2);
        else if(tri2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,p1,tr2);
        else distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,p1,te2); //tet2
    }
    else if(seg1){
        if(pnt2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,s1,p2);
        else if(seg2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,s1,s2);
        else if(tri2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,s1,tr2);
        else distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,s1,te2); //tet2
    }
    else if(tri1){
        if(pnt2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,tr1,p2);
        else if(seg2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,tr1,s2);
        else if(tri2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,tr1,tr2);
        else distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,tr1,te2); //tet2
    }
    else{//tet1
        if(pnt2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,te1,p2);
        else if(seg2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,te1,s2);
        else if(tri2) distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,te1,tr2);
        else distance=CODIMENSIONAL_INTERSECTION<T>::Distance_Between_Geometries(particles.X,particles.X,particles.thickness,particles.thickness,te1,te2); //tet2
    }

    if(verbose)LOG::cout<<"distance between elements: "<<distance<<std::endl;
    return distance;
    */
    return (T)0;
}

template<class TV>
bool PhysBAM::CODIMENSIONAL_INTERSECTION_EXAMPLE<TV>::Run_Intersection_Test(TV_INT4 arg1,TV_INT4 arg2,bool verbose)
{
    bool tet1=(arg1(4)!=0);
    bool tri1=(arg1(3)!=0) && !tet1;
    bool seg1=(arg1(2)!=0) && !tri1 && !tet1;
    bool pnt1=(arg1(1)!=0) && !seg1 && !tri1 && !tet1;

    bool tet2=(arg2(4)!=0);
    bool tri2=(arg2(3)!=0) && !tet2;
    bool seg2=(arg2(2)!=0) && !tri2 && !tet2;
    bool pnt2=(arg2(1)!=0) && !seg2 && !tri2 && !tet2;

    int p1=arg1(1);
    int p2=arg2(1);

    TV_INT2 s1(arg1(1),arg1(2));
    TV_INT2 s2(arg2(1),arg2(2));

    TV_INT3 tr1(arg1(1),arg1(2),arg1(3));
    TV_INT3 tr2(arg2(1),arg2(2),arg2(3));

    TV_INT4 te1(arg1(1),arg1(2),arg1(3),arg1(4));
    TV_INT4 te2(arg2(1),arg2(2),arg2(3),arg2(4));

    bool intersection=false;

    if(pnt1){
        if(pnt2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,p1,p2);
        else if(seg2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,p1,s2);
        else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,p1,tr2);
        else intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,p1,te2); //tet2
    }
    else if(seg1){
        if(pnt2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,s1,p2);
        else if(seg2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,s1,s2);
        else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,s1,tr2);
        //else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Cylinder_Triangle_Thickness_Intersection(particles.X,particles.X,particles.thickness,particles.thickness,s1,tr2); //testing cylinder-triangle intersection
        else intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,s1,te2); //tet2
        //else intersection=CODIMENSIONAL_INTERSECTION<T>::Cylinder_Tetrahedron_Intersection(particles.X,particles.X,particles.thickness,particles.thickness,s1,te2);
    }
    else if(tri1){
        if(pnt2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,tr1,p2);
        else if(seg2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,tr1,s2);
        //else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,tr1,tr2);
        else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Prism_Prism_Intersection(particles.X,particles.X,particles.thickness,particles.thickness,tr1,tr2);
        else intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,tr1,te2); //tet2
    }
    else{//tet1
        if(pnt2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,te1,p2);
        else if(seg2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,te1,s2);
        else if(tri2) intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,te1,tr2);
        else intersection=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,te1,te2); //tet2
    }

    if(verbose)LOG::cout<<"Intersection: "<<(intersection?"true":"false")<<std::endl;
    return intersection;
}

//template class CODIMENSIONAL_INTERSECTION_EXAMPLE<VECTOR<float,2> >;
template class CODIMENSIONAL_INTERSECTION_EXAMPLE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_INTERSECTION_EXAMPLE<VECTOR<double,2> >;
template class CODIMENSIONAL_INTERSECTION_EXAMPLE<VECTOR<double,3> >;
#endif
