//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Jets/CODIMENSIONAL_JETS_DRIVER.h"

using namespace PhysBAM;

enum CODIMENSIONAL_JETS_TESTS_ID
{
    test_filament_retraction=1,
    test_planar_sheet_retraction=2,
    test_circular_sheet_retraction=3,
    test_filament_pinchoff=51,
    test_fishbone=100,
    test_fluidpolygon=200,
    test_sheet_speed=201,

    test_infinite_planar_sheet_retraction=302
};

template<class TV>
PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::CODIMENSIONAL_JETS_DRIVER()
    :BASE(),infinite_planar_sheet_boundary(example.geometrical_codimensional_mesh,example.particles,example,length_scale)
{
}

template<class TV>
void PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::Initialize()
{
    BASE::Initialize();

    if(test==test_infinite_planar_sheet_retraction){
        infinite_planar_sheet_boundary.has_incompressibility=has_incompressibility;
        infinite_planar_sheet_boundary.has_viscosity=has_viscosity;
        infinite_planar_sheet_boundary.has_elasticity=has_elasticity;
        infinite_planar_sheet_boundary.Initialize();
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::Initialize_Test_Parameters()
{
    switch(test){
    case test_planar_sheet_retraction:
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;rim_thickness_scale=thickness_scale*(T)10;
            int m=5;int n=5;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,TV(),length_scale);

            frame_rate=50000;
        }break;
    case test_infinite_planar_sheet_retraction:
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;rim_thickness_scale=thickness_scale*(T)10;
            int m=30;int n=30;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,TV(),length_scale);

            frame_rate=50000;

            infinite_planar_sheet_boundary.length_scale=length_scale;
        }break;
    default:{PHYSBAM_FATAL_ERROR("invalid test");}break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::Initialize_Solver()
{
    BASE::Initialize_Solver();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::Update_Kinematic_Positions(const T dt,const T time)
{
    switch(test){
    case test_infinite_planar_sheet_retraction:
        {
            infinite_planar_sheet_boundary.Apply_Boundary_Conditions();
        }break;
    default:break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_JETS_DRIVER<TV>::Advance_One_Timestep(const T dt,const T time)
{
    BASE::Advance_One_Timestep(dt,time);
    infinite_planar_sheet_boundary.Update_Boundary_Region();
}

//template class CODIMENSIONAL_JETS_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_JETS_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_JETS_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_JETS_DRIVER<VECTOR<double,3> >;
#endif