//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <fstream>

#include <PhysBAM_Tools/Read_Write/Utilities/FILE_UTILITIES.h>
#include <PhysBAM_Tools/Read_Write/Arrays/READ_WRITE_ARRAY.h>
#include <PhysBAM_Tools/Read_Write/Data_Structures/READ_WRITE_HASHTABLE.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>


#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/PAIR.h>

#include "../Levelset/CODIMENSIONAL_LEVELSET_DRIVER.h"
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::CODIMENSIONAL_LEVELSET_DRIVER()
    :BASE(),read_phis_from_first_frame(false),update_slow_interface_phis(true),levelset_region_number(4),levelset(example.geometrical_codimensional_mesh,example.particles),ignore_tetrahedron_interior(false),only_update_interface(false),passive_update(false),
    use_narrow_band(false),narrow_band_width(1),has_eulerian_particles(false),eulerian_index(1)
{
    //no-op
    random.Set_Seed(120121);
}

template<class TV>
PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::~CODIMENSIONAL_LEVELSET_DRIVER()
{
    //no-op
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Initialize_Test_Parameters()
{
    example.particles.Store_Phis(levelset_region_number);
    prev_particles.Store_Phis(levelset_region_number);//necessary?

    example.particles.Store_Parents(8);//TODO: don't use magic number
    prev_particles.Store_Parents(8);//TODO: don't use magic number

    example.mesher.use_particle_parents=true;

    example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_1D());
    example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_2D());
    if(TV::m==3)example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_3D());
    if(example.deformable_geometry_collection.structures.m>0)example.deformable_geometry_collection.structures.Last()->Update_Number_Nodes();

    levelset.ignore_tetrahedron_interior=ignore_tetrahedron_interior;
    levelset.only_update_interface=only_update_interface;
    levelset.use_narrow_band=use_narrow_band;
    levelset.narrow_band_width=narrow_band_width;

    Read_Output_Files(current_frame);

    if(read_phis_from_first_frame)return;
    switch(test){
    case 1:{ ////paint splash tests
        PHYSBAM_ASSERT(levelset_region_number==4);
        ARRAY<T> heights;
        heights.Append((T)0.031);
        heights.Append((T)0.061);
        heights.Append((T)0.091);
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=2;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=heights(1);
        levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=heights(2);
        levelset_regions(3).max_corner(axis)=levelset_regions(4).min_corner(axis)=heights(3);
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 2:{ ////paint splash tests
        PHYSBAM_ASSERT(levelset_region_number==4);
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=3;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)-.281;
        levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=(T)-.021;
        levelset_regions(3).max_corner(axis)=levelset_regions(4).min_corner(axis)=(T).281;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 10:{ ////'si' in SIGGRAPH
        PHYSBAM_ASSERT(levelset_region_number==2);
        TV center((T)0,(T).006,(T)-.008);
        T radius=(T).005;
        for(int c=1;c<=levelset_region_number;c++){
            T sign=(c==1?(T)1:(T)-1);
            for(int i=1;i<=example.particles.X.m;i++){
                TV v=example.particles.X(i)-center;
                if(v.Magnitude()>radius)example.particles.phis(c)(i)=sign*(T)1e10;
                else example.particles.phis(c)(i)=(T)-1*sign*(T)1e10;
                //example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 11:{ ////'sig,' restarting from 'si'
        PHYSBAM_ASSERT(levelset_region_number==3);
        T offset_x=(T)-.011;
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        //initial_particles.Store_Phis(levelset_region_number-1);
        //initial_particles.Store_Parents();
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_si",250);
        for(int i=1;i<=initial_particles.X.m;i++){initial_particles.X(i)(1)+=offset_x;}
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 12:{ ////'sigg,' restarting from 'sig'
        PHYSBAM_ASSERT(levelset_region_number==4);
        T offset_x=(T)-.013;
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_sig",400);
        for(int i=1;i<=initial_particles.X.m;i++){initial_particles.X(i)(1)+=offset_x;}
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 13:{ ////'siggraph,' restarting from 'sigg'
        PHYSBAM_ASSERT(levelset_region_number==5);
        //T offset_x=(T)-.013;
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_sigg",450);
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 14:{////trace 'sigg',' starting from 'sigg'
        PHYSBAM_ASSERT(levelset_region_number==4);
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_sigg",450);
        Assign_Initial_Phis_By_Position(initial_particles);
    }break;
    case 15:{////'r'
        PHYSBAM_ASSERT(levelset_region_number==5);
        T large_number=1e10;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                T sign=(c==levelset_region_number?(T)-1:(T)1);
                example.particles.phis(c)(i)=sign*large_number;
            }
        }
    }break;
    case 16:{////'ra'
        PHYSBAM_ASSERT(levelset_region_number==6);
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_r",360);
        T offset_x=(T)-.025;
        for(int i=1;i<=initial_particles.X.m;i++){initial_particles.X(i)(1)+=offset_x;}
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 17:{////'rap'
        PHYSBAM_ASSERT(levelset_region_number==7);
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_ra",430);
        T offset_x=(T)-.007;
        for(int i=1;i<=initial_particles.X.m;i++){initial_particles.X(i)(1)+=offset_x;}
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 18:{////'raph'
        PHYSBAM_ASSERT(levelset_region_number==8);
        CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"low_rap",400);
        T offset_x=(T)-.013;
        for(int i=1;i<=initial_particles.X.m;i++){initial_particles.X(i)(1)+=offset_x;}
        Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(initial_particles);
    }break;
    case 41:{ ////corresponds to paint test 41
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=1;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)-2e-3;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 42:{ ////corresponds to paint test 41
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=3;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)-2e-4;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 100:{ ////setup for narrow band test
        PHYSBAM_ASSERT(levelset_region_number==2);
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector((T)-1000),TV::Constant_Vector((T)1000));
        int axis=3;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)1e-5;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }

    }break;
    case 200:{ ////paint fountain test
        PHYSBAM_ASSERT(levelset_region_number==2);
        has_eulerian_particles=true;
        std::string filename=output_directory+"/1/tube_ep_1";//just hard-code frame 1 for now...
        Read_Eulerian_Particles_From_File(filename);
        T large_number=(T)1e10;
        for(int c=1;c<=levelset_region_number;c++){
            T sign=(c==1?(T)-1:(T)1);
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=sign*large_number;
            }
        }
        //set up eulerian particle variations over time
        {
            int default_index=1;

            transition_times.Append(Time_At_Frame(20));
            eulerian_phis.Append(HASHTABLE<int,int>());
            T angle_max=(T)2*(T)pi;
            T angle_min=(T)pi;
            int color_index=2;
            for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
                HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
                T angle=iter.Data().angle;
                if(angle>=angle_min&&angle<angle_max)phis.Insert(iter.Key(),color_index);
                else phis.Insert(iter.Key(),default_index); 
            }

            transition_times.Append(Time_At_Frame(40));
            eulerian_phis.Append(HASHTABLE<int,int>());
            angle_max=(T)0.5*(T)pi;
            angle_min=(T)0;
            color_index=2;
            for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
                HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
                T angle=iter.Data().angle;
                if(angle>=angle_min&&angle<angle_max)phis.Insert(iter.Key(),color_index);
                else phis.Insert(iter.Key(),default_index);
            }
        }
    }break;
    case 201:{ ////paint fountain test
        PHYSBAM_ASSERT(levelset_region_number==5);
        has_eulerian_particles=true;
        std::string filename=output_directory+"/1/tube_ep_1";//just hard-code frame 1 for now...
        Read_Eulerian_Particles_From_File(filename);
        T large_number=(T)1e10;
        for(int c=1;c<=levelset_region_number;c++){
            T sign=(c==1?(T)-1:(T)1);
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=sign*large_number;
            }
        }
        //set up eulerian particle variations over time
        {
            int default_index=1;

            int phase_count=levelset_region_number;
            int frames_per_phase=60;
            for(int i=1;i<=phase_count;i++){
                transition_times.Append(Time_At_Frame((i-1)*frames_per_phase));
                eulerian_phis.Append(HASHTABLE<int,int>());
                HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
                for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
                    phis.Insert(iter.Key(),default_index);
                }
                for(int j=2;j<=i;j++){
                    T angle_min=(j-2)*(T)one_fourth_pi;
                    T angle_max=(j-1)*(T)one_fourth_pi;

                    for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
                        T angle=iter.Data().angle;
                        if(angle>=angle_min&&angle<angle_max)phis.Set(iter.Key(),j);
                    }
                }
            }
        }
    }break;
    case 202:{
        PHYSBAM_ASSERT(levelset_region_number==4);
        has_eulerian_particles=true;
        std::string filename=output_directory+"/1/tube_ep_1";//just hard-code frame 1 for now...
        Read_Eulerian_Particles_From_File(filename);

        ////set initial phis
        T large_number=(T)1e10;
        for(int c=1;c<=levelset_region_number;c++){
            //T sign=(c==1?(T)-1:(T)1);
            for(int i=1;i<=example.particles.X.m;i++){
                //example.particles.phis(c)(i)=sign*large_number;
                example.particles.phis(c)(i)=large_number;
            }
        }
        for(int i=1;i<=example.particles.X.m;i++){
            TV dir=(example.particles.X(i)-TV());
            dir(2)=(T)0;dir.Normalize();
            T dot=TV::Dot_Product(dir,TV::Axis_Vector(1));
            T sign=TV::Dot_Product(dir,(T)-1*TV::Axis_Vector(3));
            T angle=(T)acos(dot);
            if(sign<(T)0){
                angle+=(T)pi;
            }
            if(angle>=(T)0&&angle<(T)pi/(T)3)example.particles.phis(2)(i)=(T)-1*large_number;
            else if(angle>=(T)2*(T)pi/(T)3&&angle<(T)3*(T)pi/(T)3)example.particles.phis(3)(i)=(T)-1*large_number;
            else if(angle>=(T)4*pi/(T)3&&angle<(T)5*(T)pi/(T)3)example.particles.phis(4)(i)=(T)-1*large_number;
            else example.particles.phis(1)(i)=(T)-1*large_number;
        }
        
        ////set phis for eulerian particles
        int default_index=1;
        transition_times.Append(Time_At_Frame(0));
        eulerian_phis.Append(HASHTABLE<int,int>());
        HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
        for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
            phis.Insert(iter.Key(),default_index);

            T angle=iter.Data().angle;
            if(angle>=(T)0&&angle<(T)pi/(T)3)phis.Set(iter.Key(),2);
            else if(angle>=(T)2*(T)pi/(T)3&&angle<(T)3*(T)pi/(T)3)phis.Set(iter.Key(),3);
            else if(angle>=(T)4*pi/(T)3&&angle<(T)5*(T)pi/(T)3)phis.Set(iter.Key(),4);
        }
    }break;
    case 203:{
        PHYSBAM_ASSERT(levelset_region_number==4);
        has_eulerian_particles=true;
        std::string filename=output_directory+"/1/tube_ep_1";//just hard-code frame 1 for now...
        Read_Eulerian_Particles_From_File(filename);

        ////set initial phis
        T large_number=(T)1e10;
        //for(int c=1;c<=levelset_region_number;c++){
        //    //T sign=(c==1?(T)-1:(T)1);
        //    for(int i=1;i<=example.particles.X.m;i++){
        //        //example.particles.phis(c)(i)=sign*large_number;
        //        example.particles.phis(c)(i)=large_number;
        //    }
        //}
        //for(int i=1;i<=example.particles.X.m;i++){
        //    TV dir=(example.particles.X(i)-TV());
        //    dir(2)=(T)0;dir.Normalize();
        //    T dot=TV::Dot_Product(dir,TV::Axis_Vector(1));
        //    T sign=TV::Dot_Product(dir,(T)-1*TV::Axis_Vector(3));
        //    T angle=(T)acos(dot);
        //    if(sign<(T)0){
        //        angle+=(T)pi;
        //    }
        //    if(angle>=(T)0&&angle<(T)pi/(T)3)example.particles.phis(2)(i)=(T)-1*large_number;
        //    else if(angle>=(T)2*(T)pi/(T)3&&angle<(T)3*(T)pi/(T)3)example.particles.phis(3)(i)=(T)-1*large_number;
        //    else if(angle>=(T)4*pi/(T)3&&angle<(T)5*(T)pi/(T)3)example.particles.phis(4)(i)=(T)-1*large_number;
        //    else example.particles.phis(1)(i)=(T)-1*large_number;
        //}

        ////set phis for eulerian particles
        ARRAY<PAIR<T,int> > phis_map;
        //phis_map.Append(PAIR<T,int>((T)0,1));
        //phis_map.Append(PAIR<T,int>((T).1,2));
        //phis_map.Append(PAIR<T,int>((T).2,1));
        //phis_map.Append(PAIR<T,int>((T).5,3));
        //phis_map.Append(PAIR<T,int>((T).6,4));
        //phis_map.Append(PAIR<T,int>((T).7,1));
        //phis_map.Append(PAIR<T,int>((T)1,1));

        int n=8;
        T part=(T)1/(T)n;
        phis_map.Append(PAIR<T,int>((T)0,1));
        for(int i=1;i<n;i++){
            T frac=(T)(i-1)/(T)n+random.Get_Uniform_Number((T).3,(T)1)*part;
            //int rgn=random.Get_Uniform_Number((T)0,(T)1)>(T).3?1:0;
            //if(rgn==0){
            //    rgn=random.Get_Uniform_Integer(2,4);
            //}
            //int rgn=random.Get_Uniform_Integer(1,4);
            int rgn=i%4+1;
            phis_map.Append(PAIR<T,int>(frac,rgn));
            LOG::cout<<"phi color: "<<frac<<", "<<rgn<<std::endl;
        }
        phis_map.Append(PAIR<T,int>((T)1,1));

        int default_index=1;
        transition_times.Append(Time_At_Frame(0));
        eulerian_phis.Append(HASHTABLE<int,int>());

        //HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
        //for(HASHTABLE_ITERATOR<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
        //    phis.Insert(iter.Key(),default_index);

        //    T angle=iter.Data().angle;
        //    //if(angle>=(T)0&&angle<(T)pi/(T)3)phis.Set(iter.Key(),2);
        //    //else if(angle>=(T)2*(T)pi/(T)3&&angle<(T)3*(T)pi/(T)3)phis.Set(iter.Key(),3);
        //    //else if(angle>=(T)4*pi/(T)3&&angle<(T)5*(T)pi/(T)3)phis.Set(iter.Key(),4);

        //    //T split=(T)5;T splitted_angle=(T)two_pi/split;
        //    //int n=(int)floor(angle/splitted_angle);
        //    //if(angle-n*splitted_angle<(T).3*splitted_angle){
        //    //    int k=random.Get_Uniform_Integer(2,4);
        //    //    phis.Set(iter.Key(),k);
        //    //}

        //    T frac=angle/(T)two_pi;frac=clamp(frac,(T)0,(T)1);
        //    for(int j=1;j<=phis_map.m-1;j++){
        //        T f1=phis_map(j).x;T f2=phis_map(j+1).x;
        //        if(frac>=f1&&frac<f2){
        //            int idx=phis_map(j).y;
        //            phis.Set(iter.Key(),idx);break;
        //            //LOG::cout<<"key "<<iter.Key()<<", "<<idx<<std::endl;
        //        }
        //    }
        //}

        HASHTABLE<int,int>& phis=eulerian_phis(eulerian_phis.m);
        for(int i=1;i<=example.particles.X.m;i++){
            int guid=example.particles.guid(i);TV p0=TV();TV dir=TV::Axis_Vector(2);
            TV d1=TV::Axis_Vector(1);TV d2=example.particles.X(i)-p0;d2.y=(T)0;d2=d2.Normalized();
            T cos_theta=TV::Dot_Product(d1,d2);
            T sign=VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(d1,d2),dir)>(T)0?(T)1:(T)-1;
            T angle=sign*acos(cos_theta)+(T)pi;
            T frac=angle/(T)two_pi;frac=clamp(frac,(T)0,(T)1);
            int phi_idx=1;
            for(int j=1;j<=phis_map.m-1;j++){
                T f1=phis_map(j).x;T f2=phis_map(j+1).x;
                if(frac>=f1&&frac<f2){phi_idx=phis_map(j).y;break;}
            }

            //if(angle>=(T)0&&angle<(T)pi/(T)3)phi_idx=1;
            //else if(angle>=(T)2*(T)pi/(T)3&&angle<(T)3*(T)pi/(T)3)phi_idx=3;
            //else if(angle>=(T)4*pi/(T)3&&angle<(T)5*(T)pi/(T)3)phi_idx=4;
            //else phi_idx=1;

            if(eulerian_particles.Contains(guid)){
                PAINT_TUBE<TV>::EULERIAN_PARTICLE& ep=eulerian_particles.Get(guid);
                T ep_angle=ep.angle;
                LOG::cout<<"guid "<<guid<<", angle "<<angle<<", "<<ep_angle<<", rgn: "<<phi_idx<<std::endl;

                phis.Set(guid,phi_idx);
            }

            if(eulerian_particles.Contains(guid)){
                for(int j=1;j<=example.particles.phis.m;j++)example.particles.phis(j)(i)=large_number;
                example.particles.phis(phi_idx)(i)=(T)-1*large_number;
            }
            else{
                for(int j=1;j<=example.particles.phis.m;j++)example.particles.phis(j)(i)=(T)1;
                example.particles.phis(phi_idx)(i)=(T)-1;
            }
        }

    }break;
    case 300:{ ////assign phis by region for paint spray example
        Assign_Phis_By_Region();
    }break;
    case 401:{ ////'si' in SIGGRAPH
        PHYSBAM_ASSERT(levelset_region_number==8);

        T large_number=(T)1e10;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=large_number;
            }
        }

        TV center((T)0,(T).006,(T)-.008);
        T radius=(T).005;
        for(int i=1;i<=example.particles.X.m;i++){
            TV v=example.particles.X(i)-center;
            if(v.Magnitude()>radius)example.particles.phis(1)(i)*=(T)-1;
            else example.particles.phis(2)(i)*=(T)-1;
        }
    }break;
    case 402:{////trace 'gg',' starting from 'g'
        PHYSBAM_ASSERT(levelset_region_number==8);

        //CODIMENSIONAL_FLUID_PARTICLES<TV> initial_particles;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(initial_particles,"/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/paint_letters_2015_1_6/new_g/",600);
        //Assign_Initial_Phis_By_Position(initial_particles);

        T large_number=(T)1e10;
        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=large_number;
            }
        }

        TV center((T)0.01,(T)0.00701942,(T)-0.00980388);
        T radius=(T).005;
        for(int i=1;i<=example.particles.X.m;i++){
            TV v=example.particles.X(i)-center;
            if(v.Magnitude()>radius)example.particles.phis(3)(i)*=(T)-1;
            else example.particles.phis(4)(i)*=(T)-1;
        }
    }break;
    case 500:{////paint spray
        PHYSBAM_ASSERT(levelset_region_number==5);
        Assign_Phis_By_Region();
    }break;
    case 501:{////paint spray part 2, low res
        PHYSBAM_ASSERT(levelset_region_number==5);
        std::string path="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/spray_part1_frame0-1200/";
        int frame=1080;
        Copy_Phis_From_Other_Data(path,frame);

    }break;
    case 502:{////simple paint trace
        PHYSBAM_ASSERT(levelset_region_number==2);
        std::string path="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/paint_simple_2/";
        int frame=500;
        Copy_Phis_From_Other_Data(path,frame);
    }break;
    case 503:{////new ap
        PHYSBAM_ASSERT(levelset_region_number==2);
        std::string path="/data/equigley/PhysBAM/Private_Projects/codimensional_fluids/new_a/";
        int frame=600;
        Copy_Phis_From_Other_Data(path,frame);
    }break;
    case 504:{ ////'ap'
        PHYSBAM_ASSERT(levelset_region_number==2);
        TV center((T)0,(T).0066,(T).02);
        T radius=(T).003;
        for(int c=1;c<=levelset_region_number;c++){
            T sign=(c==1?(T)1:(T)-1);
            for(int i=1;i<=example.particles.X.m;i++){
                TV v=example.particles.X(i)-center;
                if(v.Magnitude()>radius)example.particles.phis(c)(i)=sign*(T)1e10;
                else example.particles.phis(c)(i)=(T)-1*sign*(T)1e10;
                //example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 505:{ ////'aph'
        PHYSBAM_ASSERT(levelset_region_number==3);
        //std::string path="/data/equigley/PhysBAM/Private_Projects/codimensional_fluids/new_ap/";
        std::string path="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_fluids/new_ap_narrowband/";
        int frame=540;

        STREAM_TYPE stream_type((T()));
        std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
        std::string file_name=path+f+"/codimensional_fluid_particles";
        if(FILE_UTILITIES::File_Exists(file_name)){
            FILE_UTILITIES::Read_From_File(stream_type,file_name,prev_particles);
        }

        TV offset((T)-.013,(T)0,(T)0);
        for(int i=1;i<=prev_particles.X.m;i++)prev_particles.X(i)+=offset;

        ////assign phis by position, instead of guid
        HASHTABLE<int,int> current_to_prev;
        T epsilon=(T)1e-6;
        T large_number=(T)1e10;

        for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=prev_particles.X.m;j++){
                T dist=(example.particles.X(i)-prev_particles.X(j)).Magnitude();
                if(dist<epsilon){
                    current_to_prev.Insert(i,j);
                    break;
                }
            }
        }

        LOG::cout<<"size of current_to_prev: "<<current_to_prev.Size()<<std::endl;

        ARRAY<ARRAY_VIEW<T> >& phis=example.particles.phis;

        //set phis to large number
        for(int i=1;i<=phis.m;i++){
            for(int j=1;j<=example.particles.X.m;j++){
                phis(i)(j)=large_number;
            }
        }

        for(int i=1;i<=phis.m;i++){
            for(int j=1;j<=example.particles.X.m;j++){
                if(current_to_prev.Contains(j)){
                    phis(i)(j)=prev_particles.phis(i)(current_to_prev.Get(j));
                }
                else if(i==levelset_region_number){
                    phis(i)(j)*=(T)-1;
                }
            }
        }


    }break;
    case 506:{ ////'sigg' for trace
        PHYSBAM_ASSERT(levelset_region_number==8);
        std::string path_si="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/paint_letters_2015_1_6/new_si/";
        std::string path_gg="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/paint_letters_2015_1_6/new_gg/";
        //std::string path_r="/phoenix/equigley/PhysBAM/Private_Projects/codimensional_mesh_postprocess/paint_letters_2015_1_6/new_r/";
        int frame_si=320;
        int frame_gg=500;
        //int frame_r=480;
        TV offset_si((T)-.024,(T)0,(T)0);
        TV offset_gg((T)0.005,(T)0,(T)0);

        //set phis to large number
        ARRAY<ARRAY_VIEW<T> >& phis=example.particles.phis;
        T large_number=(T)1e10;
        for(int i=1;i<=phis.m;i++){
            for(int j=1;j<=example.particles.X.m;j++){
                phis(i)(j)=large_number;
            }
        }

        for(int i=1;i<=2;i++){
            const std::string& path=(i==1?path_si:path_gg);
            const int& frame=(i==1?frame_si:frame_gg);

            STREAM_TYPE stream_type((T()));
            std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
            std::string file_name=path+f+"/codimensional_fluid_particles";
            if(FILE_UTILITIES::File_Exists(file_name)){
                FILE_UTILITIES::Read_From_File(stream_type,file_name,prev_particles);
            }

            const TV& offset=(i==1?offset_si:offset_gg);
            for(int i=1;i<=prev_particles.X.m;i++)prev_particles.X(i)+=offset;

            ////assign phis by position, instead of guid
            HASHTABLE<int,int> current_to_prev;
            T epsilon=(T)1e-15;

            for(int i=1;i<=example.particles.X.m;i++){
                for(int j=1;j<=prev_particles.X.m;j++){
                    T dist=(example.particles.X(i)-prev_particles.X(j)).Magnitude();
                    if(dist<epsilon){
                        current_to_prev.Insert(i,j);
                        break;
                    }
                }
            }

            LOG::cout<<"size of current_to_prev: "<<current_to_prev.Size()<<std::endl;


            for(int i=1;i<=phis.m;i++){
                for(int j=1;j<=example.particles.X.m;j++){
                    if(current_to_prev.Contains(j)){
                        phis(i)(j)=prev_particles.phis(i)(current_to_prev.Get(j));
                    }
                    //else if(i==levelset_region_number){
                    //    phis(i)(j)*=(T)-1;
                    //}
                }
            }

        }

    }break;
    default:{
        PHYSBAM_FATAL_ERROR("invalid test");
    }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Execute_Main_Program()
{
    for(current_frame=first_frame;current_frame<=last_frame;current_frame++){
        if(current_frame!=first_frame){
            Read_Output_Files(current_frame);
            Read_Previous_Output_Files(current_frame-1);
            Initialize_Phis_Using_Parents();
        }
        if(has_eulerian_particles)Set_Eulerian_Phis();

        Smooth_Phis_On_The_Rim();
        Reset_Phis_For_Source_Regions();

        if(!passive_update)levelset.Reinitialize(update_slow_interface_phis);
        Write_Output_Files(current_frame);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Write_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    if(example.particles.X.m>0){
        FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/codimensional_fluid_particles",example.particles);
        example.deformable_geometry_collection.Write(stream_type,output_directory,frame,frame,true);
    }
    LOG::cout<<"Finished frame "<<frame<<"."<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Read_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string file_name=output_directory+"/"+f+"/codimensional_fluid_particles";
    if(FILE_UTILITIES::File_Exists(file_name)){
        FILE_UTILITIES::Read_From_File(stream_type,file_name,example.particles);

        example.deformable_geometry_collection.Read(stream_type,output_directory,frame,frame,true);
        example.geometrical_codimensional_mesh.Update_States(example.particles.X.m,example.mpi_codim);

        //FILE_UTILITIES::Read_From_File(stream_type,output_directory+"/"+f+"/particle_parents",example.particles.parents_hashtable);

        LOG::cout<<"Read frame "<<frame<<"."<<std::endl;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Read_Previous_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string file_name=output_directory+"/"+f+"/codimensional_fluid_particles";
    if(FILE_UTILITIES::File_Exists(file_name)){
        FILE_UTILITIES::Read_From_File(stream_type,file_name,prev_particles);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Initialize_Phis_Using_Parents()
{
    HASHTABLE<int,int> guid_to_prev_index;
    for(int i=1;i<=prev_particles.X.m;i++){
        guid_to_prev_index.Insert(prev_particles.guid(i),i);
    }

    ARRAY<ARRAY_VIEW<T> >& phis=example.particles.phis;
    const ARRAY<ARRAY_VIEW<int> >& parents_guid=example.particles.parents_guid;
    const ARRAY<ARRAY_VIEW<T> >& parents_weight=example.particles.parents_weight;

    //zero out phis
    for(int i=1;i<=phis.m;i++){
        for(int j=1;j<=example.particles.X.m;j++){
            phis(i)(j)=(T)0;
        }
    }

    for(int i=1;i<=example.particles.X.m;i++){
        int parent_count=example.particles.parents_number(i);

        if(parent_count==0){
            LOG::cout<<"warning: particle "<<i<<" has no parents..."<<std::endl;
        }

        for(int j=1;j<=parent_count;j++){
            for(int k=1;k<=phis.m;k++){
                if(!guid_to_prev_index.Contains(parents_guid(j)(i))){
                    LOG::cout<<"about to fail, parent guid: "<<parents_guid(j)(i)<<std::endl;
                    //continue;
                }
                int parent_idx=guid_to_prev_index.Get(parents_guid(j)(i));
                //phis(k)(i)+=prev_particles.phis(k)(parent_idx)*parents_weight(j)(i);
                
                ////big phi!
                T prev_phi=prev_particles.phis(k)(parent_idx);
                T prev_w=parents_weight(j)(i);
                //if(prev_phi>(T)0)prev_phi=(T)100;
                //else if(prev_phi<(T)0)prev_phi=(T)-100;
                phis(k)(i)+=prev_phi*prev_w;
            }
        }
    }

    /*
    T epsilon=(T)1e-10;
    for(int i=1;i<=phis.m;i++){
        for(int j=1;j<=example.particles.X.m;j++){
            if(phis(i)(j)==(T)0){phis(i)(j)=epsilon;}
        }
    }
    */
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Initialize_UVs_Using_Parents()
{
    HASHTABLE<int,int> guid_to_prev_index;
    for(int i=1;i<=prev_particles.X.m;i++){
        guid_to_prev_index.Insert(prev_particles.guid(i),i);
    }

    ARRAY_VIEW<TV>& uvw=example.particles.texture_coordinate;
    const ARRAY<ARRAY_VIEW<int> >& parents_guid=example.particles.parents_guid;
    const ARRAY<ARRAY_VIEW<T> >& parents_weight=example.particles.parents_weight;

    for(int i=1;i<=example.particles.X.m;i++){
        uvw(i)=TV();
    }

    for(int i=1;i<=example.particles.X.m;i++){
        int parent_count=example.particles.parents_number(i);

        if(parent_count==0){
            LOG::cout<<"warning: particle "<<i<<" has no parents..."<<std::endl;
        }

        for(int j=1;j<=parent_count;j++){
            if(!guid_to_prev_index.Contains(parents_guid(j)(i))){
                LOG::cout<<"about to fail, parent guid: "<<parents_guid(j)(i)<<std::endl;
            }
            int parent_idx=guid_to_prev_index.Get(parents_guid(j)(i));
            uvw(i)+=prev_particles.texture_coordinate(i)(parent_idx)*parents_weight(j)(i);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Assign_Initial_Phis_By_Position(const CODIMENSIONAL_FLUID_PARTICLES<TV>& initial_particles)
{

    //for(int i=1;i<=phis.m;i++){array_collection->Add_Array((ATTRIBUTE_ID)(ATTRIBUTE_ID_PHI1+(i-1)),&phis(i));}

    ARRAY<ARRAY_VIEW<T>*> phis;

    for(int i=1;i<=levelset_region_number;i++){
        ARRAY_VIEW<T>* phi_attribute=initial_particles.array_collection->template Get_Array<T>(ATTRIBUTE_ID_PHI1+(i-1));
        if(phi_attribute!=0)phis.Append(phi_attribute);
        else LOG::cout<<"error, no attribute found for phi "<<i<<"..."<<std::endl;
    }

    T epsilon=(T)1e-10;
    T large_number=(T)1e10;
    ARRAY<bool> visited(example.particles.X.m);visited.Fill(false);
    ARRAY<bool> visited_initial(initial_particles.X.m);visited_initial.Fill(false);
    //T min_dist=(T)1000;
    for(int i=1;i<=example.particles.X.m;i++){
        for(int j=1;j<=initial_particles.X.m;j++){
            if(visited_initial(j))continue;
            T dist=(example.particles.X(i)-initial_particles.X(j)).Magnitude();
            //min_dist=min(min_dist,dist);LOG::cout<<"dist: "<<min_dist<<std::endl;
            if(dist<epsilon){
                visited(i)=true;
                visited_initial(j)=true;
                for(int k=1;k<=levelset_region_number;k++){
                    example.particles.phis(k)(i)=(*phis(k))(j);
                    //LOG::cout<<"phi is: "<<(*phis(k))(j)<<std::endl;
                    //if((*phis(k))(j)==(T)0)LOG::cout<<"zero phi index: "<<j<<std::endl;
                    //example.particles.phis(k)(i)=initial_particles.phis(k)(j);
                    //LOG::cout<<"phi value: "<<initial_particles.phis(k)(j)<<std::endl;
                }
            }
        }
    }

    for(int i=1;i<=example.particles.X.m;i++){
        if(visited(i))continue;
        for(int j=1;j<=levelset_region_number-1;j++){
            example.particles.phis(j)(i)=large_number;
        }
        example.particles.phis(levelset_region_number)(i)=(T)-1*large_number;
    }

    for(int i=1;i<=example.particles.X.m;i++){
        if(visited(i))continue;
        for(int j=1;j<=levelset_region_number;j++){
            T sign=(j==1?(T)-1:(T)1);
            example.particles.phis(j)(i)=sign*large_number;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(const CODIMENSIONAL_FLUID_PARTICLES<TV>& initial_particles)
{

    //for(int i=1;i<=phis.m;i++){array_collection->Add_Array((ATTRIBUTE_ID)(ATTRIBUTE_ID_PHI1+(i-1)),&phis(i));}

    ARRAY<ARRAY_VIEW<T>*> phis;

    for(int i=1;i<=levelset_region_number-1;i++){
        ARRAY_VIEW<T>* phi_attribute=initial_particles.array_collection->template Get_Array<T>(ATTRIBUTE_ID_PHI1+(i-1));
        if(phi_attribute!=0)phis.Append(phi_attribute);
        else LOG::cout<<"error, no attribute found for phi "<<i<<"..."<<std::endl;
    }

    T epsilon=(T)1e-10;
    T large_number=(T)1e10;
    ARRAY<bool> visited(example.particles.X.m);visited.Fill(false);
    ARRAY<bool> visited_initial(initial_particles.X.m);visited_initial.Fill(false);
    //T min_dist=(T)1000;
    for(int i=1;i<=example.particles.X.m;i++){
        for(int j=1;j<=initial_particles.X.m;j++){
            if(visited_initial(j))continue;
            T dist=(example.particles.X(i)-initial_particles.X(j)).Magnitude();
            //min_dist=min(min_dist,dist);LOG::cout<<"dist: "<<min_dist<<std::endl;
            if(dist<epsilon){
                visited(i)=true;
                visited_initial(j)=true;
                for(int k=1;k<=levelset_region_number-1;k++){
                    example.particles.phis(k)(i)=(*phis(k))(j);
                    //LOG::cout<<"phi is: "<<(*phis(k))(j)<<std::endl;
                    //if((*phis(k))(j)==(T)0)LOG::cout<<"zero phi index: "<<j<<std::endl;
                    //example.particles.phis(k)(i)=initial_particles.phis(k)(j);
                    //LOG::cout<<"phi value: "<<initial_particles.phis(k)(j)<<std::endl;
                }
                example.particles.phis(levelset_region_number)(i)=large_number;
            }
        }
    }

    for(int i=1;i<=example.particles.X.m;i++){
        if(visited(i))continue;
        for(int j=1;j<=levelset_region_number-1;j++){
            example.particles.phis(j)(i)=large_number;
        }
        example.particles.phis(levelset_region_number)(i)=(T)-1*large_number;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Read_Eulerian_Particles_From_File(const std::string& filename)
{
    std::ifstream fin(filename.c_str());
    if(fin.fail())LOG::cout<<"filename: "<<filename<<std::endl;
    if(fin.fail())PHYSBAM_FATAL_ERROR("failed to open eulerian particles file to read...");
    eulerian_particles.Remove_All();
    int size=0;fin>>size;
    for(int i=1;i<=size;i++){
        int guid;typename PAINT_TUBE<TV>::EULERIAN_PARTICLE ep;
        T thickness_scale;//unused

        fin>>guid>>ep.r>>ep.angle>>ep.object_space_pos.x>>ep.object_space_pos.y>>ep.object_space_pos.z>>
            ep.object_space_pos0.x>>ep.object_space_pos0.y>>ep.object_space_pos0.z>>ep.has_constant_thickness>>ep.thickness>>
            thickness_scale>>ep.texture_coordinates.x>>ep.texture_coordinates.y>>ep.texture_coordinates.z;
        eulerian_particles.Set(guid,ep);
    }
    fin.close();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Set_Eulerian_Phis()
{
    if(eulerian_index<transition_times.m&&Time_At_Frame(current_frame)>transition_times(eulerian_index))eulerian_index++;
    HASHTABLE<int,int>& phis=eulerian_phis(eulerian_index);
    T large_number=(T)1e10;

    for(int i=1;i<=example.particles.X.m;i++){
        if(phis.Contains(example.particles.guid(i))){
            int color_index=phis.Get(example.particles.guid(i));
            for(int c=1;c<=levelset_region_number;c++){
                T sign=(c==color_index?(T)-1:(T)1);
                example.particles.phis(c)(i)=sign*large_number;
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Assign_Phis_By_Region()
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
    mesh.Update_Neighbor_Nodes();
    ARRAY<int> region(particles.X.m);region.Fill(-1);
    int current_region=0;
    for(int i=1;i<=particles.X.m;i++){
        if(region(i)!=-1)continue;
        current_region++;
        region(i)=current_region;

        QUEUE<int> front(particles.X.m);
        front.Enqueue(i);

        while(!front.Empty()){
            int start_idx=front.Dequeue();
            const ARRAY<int>& neighbors=(*mesh.neighbor_nodes)(start_idx);
            for(int j=1;j<=neighbors.m;j++){
                if(region(neighbors(j))!=-1){
                    PHYSBAM_ASSERT(region(neighbors(j))==current_region);
                    continue;
                }
                region(neighbors(j))=current_region;
                front.Enqueue(neighbors(j));
            }
        }
    }

    PHYSBAM_ASSERT(current_region<=levelset_region_number);

    T large_number=(T)1e10;
    for(int c=1;c<=levelset_region_number;c++){
        for(int i=1;i<=particles.X.m;i++){
            particles.phis(c)(i)=large_number;
        }
    }

    for(int i=1;i<=particles.X.m;i++){
        int r=region(i);
        particles.phis(r)(i)=(T)-1*large_number;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Smooth_Phis_On_The_Rim()
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
    CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
    for(int i=1;i<=particles.X.m;i++){
        if(particles.on_boundary(i)(2)){
            for(int d=1;d<=particles.phis.m;d++){
                T avg_phi=(T)0;
                for(int j=1;j<=(*mesh.triangle_mesh.neighbor_nodes)(i).m;j++){
                    int nb_p=(*mesh.triangle_mesh.neighbor_nodes)(i)(j);
                    avg_phi+=particles.phis(d)(nb_p);
                }
                if((*mesh.triangle_mesh.neighbor_nodes)(i).m>0){
                    avg_phi/=(T)(*mesh.triangle_mesh.neighbor_nodes)(i).m;
                    particles.phis(d)(i)=avg_phi;
                    //LOG::cout<<"smooth phi "<<particles.guid(i)<<" "<<d<<": "<<particles.phis(d)(i)<<std::endl;
                }
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Reset_Phis_For_Source_Regions()
{
    T large_number=(T)1e10;
    switch(test){
    case 500:{
        LOG::cout<<"current frame: "<<current_frame<<std::endl;
        if(current_frame==244){
            CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
            CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
            mesh.Update_Neighbor_Nodes();
            int guid_of_interest=1;//wow such hack.
            int idx=-1;
            int phi_idx=3;
            for(int i=1;i<=particles.X.m;i++)if(particles.guid(i)==guid_of_interest){idx=i;break;}
            PHYSBAM_ASSERT(idx!=-1);

            QUEUE<int> front(particles.X.m);
            HASHTABLE<int> visited;
            front.Enqueue(idx);
            visited.Insert(idx);

            while(!front.Empty()){
                int start_idx=front.Dequeue();
                for(int c=1;c<=levelset_region_number;c++){
                    T sign=(c==phi_idx?(T)-1:(T)1);
                    particles.phis(c)(start_idx)=sign*large_number;
                }
                const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(start_idx);
                for(int i=1;i<=nbs.m;i++){
                    if(visited.Contains(nbs(i)))continue;
                    visited.Set(nbs(i));
                    front.Enqueue(nbs(i));
                }
            }
        }
        else if(current_frame==480){
            CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
            CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
            mesh.Update_Neighbor_Nodes();
            for(int g=1;g<=2;g++){
                //int guid_of_interest=(g==1?1:117);
                int guid_of_interest=(g==1?1:294);//for high res version
                int phi_idx=(g==1?4:5);
                int idx=-1;
                for(int i=1;i<=particles.X.m;i++)if(particles.guid(i)==guid_of_interest){idx=i;break;}
                PHYSBAM_ASSERT(idx!=-1);

                QUEUE<int> front(particles.X.m);
                HASHTABLE<int> visited;
                front.Enqueue(idx);
                visited.Insert(idx);

                while(!front.Empty()){
                    int start_idx=front.Dequeue();

                    for(int c=1;c<=levelset_region_number;c++){
                        T sign=(c==phi_idx?(T)-1:(T)1);
                        particles.phis(c)(start_idx)=sign*large_number;
                    }

                    const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(start_idx);
                    for(int i=1;i<=nbs.m;i++){
                        if(visited.Contains(nbs(i)))continue;
                        visited.Set(nbs(i));
                        front.Enqueue(nbs(i));
                    }
                }
            }
        }
        else if(current_frame==750){
            CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;
            CODIMENSIONAL_MESH& mesh=example.geometrical_codimensional_mesh;
            mesh.Update_Neighbor_Nodes();
            for(int g=1;g<=2;g++){
                //int guid_of_interest=(g==1?1:117);
                int guid_of_interest=(g==1?1:294);//for high res version
                int phi_idx=(g==1?3:4);
                int idx=-1;
                for(int i=1;i<=particles.X.m;i++)if(particles.guid(i)==guid_of_interest){idx=i;break;}
                PHYSBAM_ASSERT(idx!=-1);

                QUEUE<int> front(particles.X.m);
                HASHTABLE<int> visited;
                front.Enqueue(idx);
                visited.Insert(idx);

                while(!front.Empty()){
                    int start_idx=front.Dequeue();

                    for(int c=1;c<=levelset_region_number;c++){
                        T sign=(c==phi_idx?(T)-1:(T)1);
                        particles.phis(c)(start_idx)=sign*large_number;
                    }

                    const ARRAY<int>& nbs=(*mesh.neighbor_nodes)(start_idx);
                    for(int i=1;i<=nbs.m;i++){
                        if(visited.Contains(nbs(i)))continue;
                        visited.Set(nbs(i));
                        front.Enqueue(nbs(i));
                    }
                }
            }
        }
    }break;
    default:{
    }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_LEVELSET_DRIVER<TV>::Copy_Phis_From_Other_Data(const std::string& path,const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string file_name=path+f+"/codimensional_fluid_particles";
    if(FILE_UTILITIES::File_Exists(file_name)){
        FILE_UTILITIES::Read_From_File(stream_type,file_name,prev_particles);
    }

    switch(test){
    case 502:{
        TV offset((T)-.025,(T)0,(T).007);
        for(int i=1;i<=prev_particles.X.m;i++)prev_particles.X(i)+=offset;
    }break;
    case 503:{
        TV offset((T)-.007,(T)0,(T)0);
        for(int i=1;i<=prev_particles.X.m;i++)prev_particles.X(i)+=offset;
    }break;
    }

    ////assign phis by position, instead of guid
    HASHTABLE<int,int> current_to_prev;
    T epsilon=(T)1e-15;

    for(int i=1;i<=example.particles.X.m;i++){
        for(int j=1;j<=prev_particles.X.m;j++){
            T dist=(example.particles.X(i)-prev_particles.X(j)).Magnitude();
            if(dist<epsilon){
                current_to_prev.Insert(i,j);
                break;
            }
        }
    }

    LOG::cout<<"size of current_to_prev: "<<current_to_prev.Size()<<std::endl;

    ARRAY<ARRAY_VIEW<T> >& phis=example.particles.phis;
    //zero out phis
    for(int i=1;i<=phis.m;i++){
        for(int j=1;j<=example.particles.X.m;j++){
            phis(i)(j)=(T)0;
        }
    }

    for(int i=1;i<=phis.m;i++){
        for(int j=1;j<=example.particles.X.m;j++){
            if(!current_to_prev.Contains(j))continue;
            phis(i)(j)=prev_particles.phis(i)(current_to_prev.Get(j));
        }
    }
}

//template class CODIMENSIONAL_LEVELSET_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_LEVELSET_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_LEVELSET_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_LEVELSET_DRIVER<VECTOR<double,3> >;
#endif
