//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_LEVELSET_DRIVER__
#define __CODIMENSIONAL_LEVELSET_DRIVER__

#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include "../Levelset/LEVELSET_MULTIPLE_CODIMENSIONAL.h"
#include "../Solver/CODIMENSIONAL_FLUID_DRIVER.h"
#include "../Viscoelastic/PAINT_TUBE.h"

#include <string>

namespace PhysBAM{
template<class TV>
class CODIMENSIONAL_LEVELSET_DRIVER : public CODIMENSIONAL_FLUID_DRIVER<TV>
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef CODIMENSIONAL_FLUID_DRIVER<TV> BASE;
public:

    using BASE::test;using BASE::output_directory;using BASE::stream_type;
    using BASE::first_frame;using BASE::last_frame;using BASE::current_frame;using BASE::restart_frame;
    using BASE::example;
    using BASE::Time_At_Frame;

    bool read_phis_from_first_frame;
    bool update_slow_interface_phis;
    int levelset_region_number;
    LEVELSET_MULTIPLE_CODIMENSIONAL<TV> levelset;
    CODIMENSIONAL_FLUID_PARTICLES<TV> prev_particles;
    bool ignore_tetrahedron_interior;
    bool only_update_interface;
    bool passive_update;//don't do any reinitialization (do we still need to do projection in this case though?)
    bool use_narrow_band;
    int narrow_band_width;

    //for paint fountain example
    bool has_eulerian_particles;
    HASHTABLE<int,typename PAINT_TUBE<TV>::EULERIAN_PARTICLE> eulerian_particles; //map of guid to eulerian particle
    ARRAY<T> transition_times;
    int eulerian_index;
    ARRAY<HASHTABLE<int,int> > eulerian_phis; //for each transition time, for each eulerian particle guid, negative level set index

    ////random
    RANDOM_NUMBERS<T> random;

    CODIMENSIONAL_LEVELSET_DRIVER();
    virtual ~CODIMENSIONAL_LEVELSET_DRIVER();

    virtual void Initialize_Test_Parameters();
    virtual void Execute_Main_Program();
    virtual void Write_Output_Files(const int frame);
    virtual void Read_Output_Files(const int frame);
    void Read_Previous_Output_Files(const int frame);
    void Initialize_Phis_Using_Parents();
    void Initialize_UVs_Using_Parents();
    void Assign_Initial_Phis_By_Position(const CODIMENSIONAL_FLUID_PARTICLES<TV>& initial_particles);
    void Assign_Initial_Phis_By_Position_Using_One_Fewer_Phi(const CODIMENSIONAL_FLUID_PARTICLES<TV>& initial_particles);
    void Read_Eulerian_Particles_From_File(const std::string& filename);
    void Set_Eulerian_Phis();
    void Assign_Phis_By_Region();
    void Reset_Phis_For_Source_Regions();
    void Copy_Phis_From_Other_Data(const std::string& path,const int frame);
    void Smooth_Phis_On_The_Rim();
};
}
#endif //__CODIMENSIONAL_LEVELSET_DRIVER__
