//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __FAST_MARCHING_METHOD_CODIMENSIONAL__
#define __FAST_MARCHING_METHOD_CODIMENSIONAL__

#include <PhysBAM_Tools/Arrays/ARRAYS_FORWARD.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Solids_Geometry/RIGID_GEOMETRY.h>
#include <PhysBAM_Geometry/Level_Sets/FAST_MARCHING.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"

namespace PhysBAM{
template<class TV>
class FAST_MARCHING_METHOD_CODIMENSIONAL
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; typedef typename TV::SPIN T_SPIN;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    CODIMENSIONAL_MESH& mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    ARRAY<TRIANGLE_3D<T> > tet_interface_faces;
    ARRAY<TV_INT4> tet_interface_indices; //for debugging
    const T large_phi;
    bool update_slow_interface_particles;
    const T speed_threshold;
    bool use_non_drifting_near_interface;
    bool ignore_tetrahedron_interior;
    ARRAY<bool> active;//indices for which phi values should be updated
    bool only_update_interface;
    bool use_narrow_band;
    int narrow_band_width;
    ARRAY<int> distance_to_interface;

    FAST_MARCHING_METHOD_CODIMENSIONAL(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,bool update_slow_interface_particles_=true,bool ignore_tetrahedron_interior_=false,bool only_update_interface_=false,bool use_narrow_band_=false,int narrow_band_width_=1,T speed_threshold_=(T)1e-5);
    void Fast_Marching_Method(ARRAY_VIEW<T>& phi);
private:
    void Update_Or_Add_Neighbor(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,ARRAY<int>& close_k,ARRAY<int>& heap,int& heap_length,const int& neighbor,ARRAY<int>& codim_0x_particles);
    void Initialize_Interface(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,ARRAY<int>& close_k,ARRAY<int>& heap,int& heap_length,ARRAY<int>& codim_0x_particles);
    void Update_Close_Point(ARRAY_VIEW<T>& phi,ARRAY<bool>& done,const int& p,ARRAY<int>& codim_0x_particles);
    void Add_To_Initial(ARRAY<bool>& done,ARRAY<int>& close_k,const int& p);
    template<int N> int Find_Edge_Interface(const VECTOR<int,N>& v,ARRAY_VIEW<T>& phi,const T tolerance,ARRAY<TV>& added_points,ARRAY<TV_INT2>& interface_edges);
    void Initialize_Phi_Near_Interface(ARRAY_VIEW<T>& phi,ARRAY<bool>& done);
    void Set_Graph_Distance_To_Interface(const ARRAY<bool>& done);

    void Get_Tetrahedron_Interface_Faces(const ARRAY_VIEW<T>& phi);
    void Update_Close_Point_Using_Tetrahedra(const ARRAY_VIEW<TV>& X,ARRAY_VIEW<T>& phi,const TETRAHEDRON_MESH& tet_mesh,ARRAY<bool>& done,int index,ARRAY<int>& codim_0x_particles);
    //#####################################################################
};
}
#endif
