//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Levelset/LEVELSET_MULTIPLE_CODIMENSIONAL.h"
#include "../Levelset/FAST_MARCHING_METHOD_CODIMENSIONAL.h"

using namespace PhysBAM;
template<class TV>
PhysBAM::LEVELSET_MULTIPLE_CODIMENSIONAL<TV>::LEVELSET_MULTIPLE_CODIMENSIONAL(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input)
    :mesh(mesh_input),particles(particles_input),phis(particles.phis),ignore_tetrahedron_interior(false),only_update_interface(false),use_narrow_band(false),narrow_band_width(1)
{}

template<class TV>
void PhysBAM::LEVELSET_MULTIPLE_CODIMENSIONAL<TV>::Project_Levelset()
{
    for(int p=1;p<=particles.X.m;p++){
        int minimum_region,second_minimum_region;T minimum_phi,second_minimum_phi;
        Two_Minimum_Regions(p,minimum_region,second_minimum_region,minimum_phi,second_minimum_phi);
        T correction=(T).5*(minimum_phi+second_minimum_phi);
        for(int k=1;k<=phis.m;k++) phis(k)(p)-=correction;
    }
}

template<class TV>
void PhysBAM::LEVELSET_MULTIPLE_CODIMENSIONAL<TV>::Reinitialize(bool update_slow_interface_particles/*=false*/)
{
    if(ignore_tetrahedron_interior){
        mesh.tetrahedron_mesh.number_nodes=particles.X.m;
        mesh.tetrahedron_mesh.Initialize_Boundary_Mesh();
        mesh.tetrahedron_mesh.boundary_mesh->Initialize_Incident_Elements();
    }

    FAST_MARCHING_METHOD_CODIMENSIONAL<TV> ffm(mesh,particles,update_slow_interface_particles,ignore_tetrahedron_interior,only_update_interface,use_narrow_band,narrow_band_width);
    for(int i=1;i<=phis.m;i++)ffm.Fast_Marching_Method(phis(i));
    if(phis.m>1)Project_Levelset();
}

//template class LEVELSET_MULTIPLE_CODIMENSIONAL<VECTOR<float,2> >;
template class LEVELSET_MULTIPLE_CODIMENSIONAL<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class LEVELSET_MULTIPLE_CODIMENSIONAL<VECTOR<double,2> >;
template class LEVELSET_MULTIPLE_CODIMENSIONAL<VECTOR<double,3> >;
#endif
