//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __LEVELSET_MULTIPLE_CODIMENSIONAL__
#define __LEVELSET_MULTIPLE_CODIMENSIONAL__
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Geometry/Level_Sets/LEVELSET_UTILITIES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"

namespace PhysBAM
{
template<class TV>
class LEVELSET_MULTIPLE_CODIMENSIONAL
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; typedef typename TV::SPIN T_SPIN;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;

public:
    CODIMENSIONAL_MESH& mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    ARRAY<ARRAY_VIEW<T> >& phis;
    bool ignore_tetrahedron_interior;
    bool only_update_interface;
    bool use_narrow_band;
    int narrow_band_width;

    LEVELSET_MULTIPLE_CODIMENSIONAL(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input);
    T Phi(const int region,const int& p) const {return phis(region)(p);}
    template<int N> T Phi(const int region,const VECTOR<int,N>& v,const VECTOR<T,N>& w) const {return Phi(region,v,w,phis);}
    T Region_Value(const int p,const ARRAY<T>& value) const {return value(Inside_Region(p));}
    bool Interface(const int& p1,const int& p2) const {return Inside_Region(p1)!=Inside_Region(p2);}

    int Inside_Region(const int p) const {return Inside_Region(p,phis);} //assumes exactly one Phi<0 on a node
    int Inside_Region(const int p,T& phi) const {return Inside_Region(p,phi,phis);} // assumes exactly one Phi<0 on a node
    void Two_Minimum_Regions(const int& p,int& minimum_region,int& second_minimum_region,T& minimum_phi,T& second_minimum_phi) const 
    {Two_Minimum_Regions(p,minimum_region,second_minimum_region,minimum_phi,second_minimum_phi,phis);}

    void Reinitialize(bool update_slow_interface_particles=false);
    void Project_Levelset();

    static T Phi(const int region,const int& p,const ARRAY<ARRAY_VIEW<T> >& phis){return phis(region)(p);}
    template<int N> static T Phi(const int region,const VECTOR<int,N>& v,const VECTOR<T,N>& w,const ARRAY<ARRAY_VIEW<T> >& phis)
    {T phi=(T)0;for(int i=1;i<=N;i++)phi+=w(i)*phis(region)(v(i));return phi;}
    static T Sign(const int region,const int other_region)
    {assert(region!=other_region);if(region < other_region) return -1;else return 1;}
    static void Project_Levelset(T& minimum_phi,T& second_minimum_phi)
    {T correction=(T).5*(minimum_phi+second_minimum_phi);minimum_phi-=correction;second_minimum_phi-=correction;}
    static int Inside_Region(const int p,const ARRAY<ARRAY_VIEW<T> >& phis)// assumes exactly one Phi<0 on a node
    {for(int k=1;k<=phis.m;k++){if(phis(k)(p)<=(T)0) return k;}assert(phis(phis.m)(p)<=(T)0);return phis.m;}
    static int Inside_Region(const int p,T& phi,const ARRAY<ARRAY_VIEW<T> >& phis)
    {for(int k=1;k<=phis.m;k++){phi=phis(k)(p);if(phi<=(T)0)return k;}assert(phis(phis.m)(p)<=(T)0);return phis.m;}
    template<int N> static int Inside_Region(const VECTOR<int,N>& v,const VECTOR<T,N>& w,const ARRAY<ARRAY_VIEW<T> >& phis)
    {T minimum_phi=Phi(1,v,w,phis);int minimum_region=1;
    for(int k=2;k<=phis.m;k++){T candidate_phi=Phi(k,v,w,phis);if(candidate_phi<minimum_phi){minimum_phi=candidate_phi;minimum_region=k;}}return minimum_region;}
    template<int N> static int Inside_Region(const VECTOR<int,N>& v,const VECTOR<T,N>& w,T& phi,const ARRAY<ARRAY_VIEW<T> >& phis)
    {T minimum_phi=Phi(1,v,w,phis);int minimum_region=1;
    for(int k=2;k<=phis.m;k++){T candidate_phi=Phi(k,v,w,phis);if(candidate_phi<minimum_phi){minimum_phi=candidate_phi;minimum_region=k;}}phi=minimum_phi;return minimum_region;}

    static void Two_Minimum_Regions(const int& p,int& minimum_region,int& second_minimum_region,T& minimum_phi,T& second_minimum_phi,const ARRAY<ARRAY_VIEW<T> >& phis)
    {
        T phi1=phis(1)(p),phi2=phis(2)(p);
        if(phi1<phi2){minimum_phi=phi1;minimum_region=1;second_minimum_phi=phi2;second_minimum_region=2;}
        else{minimum_phi=phi2;minimum_region=2;second_minimum_phi=phi1;second_minimum_region=1;}
        for(int k=3;k<=phis.m;k++){T candidate_phi=phis(k)(p);
            if(candidate_phi<minimum_phi){second_minimum_phi=minimum_phi;second_minimum_region=minimum_region;minimum_phi=candidate_phi;minimum_region=k;}
            else if(candidate_phi<second_minimum_phi){second_minimum_phi=candidate_phi;second_minimum_region=k;}}
    }
    static void Minimum_Regions(const int& p1,const int& p2,int& region_1,int& region_2,T& phi_1,T& phi_2,const ARRAY<ARRAY_VIEW<T> >& phis)
    {region_1=Inside_Region(p1,phi_1,phis);region_2=Inside_Region(p2,phi_2,phis);}

    static T Convex_Average(const int & p1,const int & p2,const ARRAY<T>& value,const ARRAY<ARRAY_VIEW<T> >& phis)
    {int region_1,region_2;T phi_1=(T)0,phi_2=(T)0;Minimum_Regions(p1,p2,region_1,region_2,phi_1,phi_2,phis);
    return LEVELSET_UTILITIES<T>::Average(phi_1,value(region_1),-phi_2,value(region_2));}

    template<class T2> static T2 Convex_Average(const T phi_1,const T2& value_1,const T phi_2,const T2& value_2)
    {return (abs(phi_1)*value_2+abs(phi_2)*value_1)/(abs(phi_1)+abs(phi_2));}

    static void Tessellate_Triangle(const TV_INT3& v,ARRAY_VIEW<TV>& X,const ARRAY<ARRAY_VIEW<T> >& phis,ARRAY<TV>& added_points,ARRAY<TV_INT3>& triangles,ARRAY<int>& regions)
    {
        TV_INT3 vtx_region;TV3 vtx_phi;for(int d=1;d<=3;d++){vtx_region(d)=Inside_Region(v(d),vtx_phi(d),phis);}
        ARRAY<int> added_edge;
        if(vtx_region(1)!=vtx_region(2)){TV point=Convex_Average(vtx_phi(1),X(v(1)),vtx_phi(2),X(v(2)));added_points.Append(point);added_edge.Append(3);}
        if(vtx_region(2)!=vtx_region(3)){TV point=Convex_Average(vtx_phi(2),X(v(2)),vtx_phi(3),X(v(3)));added_points.Append(point);added_edge.Append(1);}
        if(vtx_region(3)!=vtx_region(1)){TV point=Convex_Average(vtx_phi(3),X(v(3)),vtx_phi(1),X(v(1)));added_points.Append(point);added_edge.Append(2);}

        if(added_points.m==0){triangles.Append(v);regions.Append(vtx_region(1));}
        else if(added_points.m==2){TV_INT3 flag;flag(added_edge(1))=flag(added_edge(2))=1;int c=(!flag(1)?1:(!flag(2)?2:3));int i1,i2,p1,p2,p3,r1,r2;
        switch(c){case 1:{i1=-1;i2=-2;p1=v(1);p2=v(2);p3=v(3);r1=1;r2=2;}break;
            case 2:{i1=-2;i2=-1;p1=v(2);p2=v(3);p3=v(1);r1=2;r2=3;}break;
            case 3:{i1=-2;i2=-1;p1=v(3);p2=v(1);p3=v(2);r1=3;r2=1;}break;}
            triangles.Append(TV_INT3(p1,i1,i2));regions.Append(vtx_region(r1));
            triangles.Append(TV_INT3(i1,p2,i2));regions.Append(vtx_region(r2));
            triangles.Append(TV_INT3(i2,p2,p3));regions.Append(vtx_region(r2));}
        else if(added_points.m==3){TV p=(T)one_third*(added_points(1)+added_points(2)+added_points(3));added_points.Append(p);
            triangles.Append(TV_INT3(v(1),-1,-4));regions.Append(vtx_region(1));
            triangles.Append(TV_INT3(v(1),-4,-3));regions.Append(vtx_region(1));
            triangles.Append(TV_INT3(v(2),-4,-1));regions.Append(vtx_region(2));
            triangles.Append(TV_INT3(v(2),-2,-4));regions.Append(vtx_region(2));
            triangles.Append(TV_INT3(v(3),-4,-2));regions.Append(vtx_region(3));
            triangles.Append(TV_INT3(v(3),-3,-4));regions.Append(vtx_region(3));}
        else PHYSBAM_FATAL_ERROR("add 1 point");
    }
};
}
#endif
