//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __BALL_PIVOTING_TRIANGULATION__
#define __BALL_PIVOTING_TRIANGULATION__

#include <PhysBAM_Geometry/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>

namespace PhysBAM
{
struct FRONT_EDGE
{
    int i,j,v;
    bool active;
    FRONT_EDGE* pre;
    FRONT_EDGE* next;

    FRONT_EDGE(const int i_input,const int j_input,const int v_input=0,FRONT_EDGE* pre_input=0,FRONT_EDGE* next_input=0)
        :i(i_input),j(j_input),v(v_input),active(true),pre(pre_input),next(next_input){}
};

template<class T>
class BALL_PIVOTING_TRIANGULATION
{
    typedef VECTOR<T,3> TV;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    GEOMETRY_PARTICLES<TV>& particles;
    TRIANGLE_MESH& triangle_mesh;
    T length_scale;
    T alpha_ball_radius;
    T max_segment_length;
    T epsilon;
    bool use_dbg;
    int dbg_number;

    BALL_PIVOTING_TRIANGULATION(GEOMETRY_PARTICLES<TV>& particles_input,TRIANGLE_MESH& triangle_mesh_input):particles(particles_input),triangle_mesh(triangle_mesh_input),use_dbg(false){}
    void Initialize_Length_Scale(const T length_scale_input);
    void Construct_Mesh();

protected:
    enum FRONT_PARTICLE_STATE{AVAILABLE=0,FRONT,USED,TERMINATED};
    ARRAY<int> triangulation_candidate_particles;
    ARRAY<int> front_particle_states;	////0 -- not used; 1 -- on front; 2 -- used, interior; 3 -- used, boundary
    ARRAY<int> front_particle_regions;	////0 -- isolated, 1 -- boundary

    HASHTABLE<TV_INT2,FRONT_EDGE*> front_edge_hashtable;
    HASHTABLE<TV_INT2 > interior_edge_hashtable;
    ARRAY<TV> pX;   ////perturbed particles.X
    ARRAY<TV> perturb;
    RANDOM_NUMBERS<T> random;

    BOX<TV> particle_bounding_box;
    GRID<TV> particle_grid;
    ARRAY<ARRAY<int>,TV_INT> particles_in_cell; 

    void Advancing_Front_Triangulation();
    VECTOR<int,3> Find_Seed_Triangle();
    int Find_Next_Triangle_Particle(const FRONT_EDGE* e);
    int Find_Next_Triangle_Particle_Ball_Pivoting(const FRONT_EDGE* e);
    FRONT_EDGE* Add_Triangle_Edges_To_Front(const VECTOR<int,3>& v);
    FRONT_EDGE* Find_Edge_From_Front();
    void Add_Particle_To_Front(FRONT_EDGE* & e,const int k);
    void Link_Edges(FRONT_EDGE* & pre,FRONT_EDGE* & next){pre->next=next;next->pre=pre;}
    void Check_Front(FRONT_EDGE* e);
    bool Circle_Circle_Intersection(const TV2& c1,const T r1,const TV2& c2,const T r2,TV2& p1,TV2& p2);

    BOX<TV> Get_Particle_Bounding_Box();
    GRID<TV> Get_Particle_Grid(const BOX<TV>& bounding_box,const T grid_dX);
    void Resize_Elements_In_Cell_With_Grid(ARRAY<ARRAY<int>,TV_INT>& p_in_cell,const GRID<TV>& grid);
};
}
#endif
