//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_HELPER__
#define __CODIMENSIONAL_HELPER__
#include <PhysBAM_Solids/PhysBAM_Rigids/Standard_Tests/RIGIDS_STANDARD_TESTS.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>
#include <PhysBAM_Tools/Vectors/VECTOR_ND.h>
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Math_Tools/pow.h>
#include <PhysBAM_Tools/Matrices/ROTATION.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_NXN.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <PhysBAM_Tools/Read_Write/Utilities/READ_WRITE.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Tools/Math_Tools/Is_NaN.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry_Computations/DELAUNAY_MESHING_WITH_ALPHA_SHAPE.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Dynamics/Particles/COMPRESSIBLE_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"

namespace PhysBAM
{
template<class T,class T_ARRAY,class ID> class ARRAY_BASE;

namespace Codimensional
{
////enumerations
enum TETRAHEDRON_TYPE{SLIVER=1,SPADE,CAP};

const int CONTACT_SOLID=0x1;
const int CONTACT_FIBER=0x2;
const int CONTACT_FLUID=0x4;
const int CONTACT_SOLID2=0x8; ////particle free side contact
const int CONTACT_BOUNDARY=0x10;
const int CONTACT_TANK=0x20;
const int CONTACT_SOLID3=0x40;  ////particle contact with fiber endpoints

inline void Set_Contact_With_Solid(int& contact){contact|=CONTACT_SOLID;}
inline void Set_Contact_With_Fiber(int& contact){contact|=CONTACT_FIBER;}
inline void Set_Contact_With_Fluid(int& contact){contact|=CONTACT_FLUID;}
inline void Set_Contact_With_Solid2(int& contact){contact|=CONTACT_SOLID2;}
inline void Set_Contact_With_Solid3(int& contact){contact|=CONTACT_SOLID3;}
inline void Set_Contact_Boundary(int& contact){contact|=CONTACT_BOUNDARY;}
inline void Set_Contact_With_Tank(int& contact){contact|=CONTACT_TANK;}
inline bool Is_Contact_With_Solid(const int contact){return (contact & CONTACT_SOLID)!=0;}
inline bool Is_Contact_With_Fiber(const int contact){return (contact & CONTACT_FIBER)!=0;}
inline bool Is_Contact_With_Fluid(const int contact){return (contact & CONTACT_FLUID)!=0;}
inline bool Is_Contact_With_Solid2(const int contact){return (contact & CONTACT_SOLID2)!=0;}
inline bool Is_Contact_With_Solid3(const int contact){return (contact & CONTACT_SOLID3)!=0;}
inline bool Is_Contact_Boundary(const int contact){return (contact & CONTACT_BOUNDARY)!=0;}
inline bool Is_Contact_With_Tank(const int contact){return (contact & CONTACT_TANK)!=0;}
inline void Clear_Contact_With_Solid(int& contact){contact&=(0xffff-CONTACT_SOLID);}
inline void Clear_Contact_With_Fiber(int& contact){contact&=(0xffff-CONTACT_FIBER);}
inline void Clear_Contact_With_Fluid(int& contact){contact&=(0xffff-CONTACT_FLUID);}
inline void Clear_Contact_With_Solid2(int& contact){contact&=(0xffff-CONTACT_SOLID2);}
inline void Clear_Contact_With_Solid3(int& contact){contact&=(0xffff-CONTACT_SOLID3);}
inline void Clear_Contact_Boundary(int& contact){contact&=(0xffff-CONTACT_BOUNDARY);}
inline void Clear_Contact_With_Tank(int& contact){contact&=(0xffff-CONTACT_TANK);}
inline void Set_Contact(int& contact,const int flag){contact|=flag;}
inline void Clear_Contact(int& contact,const int flag){contact&=(0xffff-flag);}

//////////////////////////////////////////////////////////////////////////
////index operations

inline VECTOR<int,1> Other_Nodes_In_Element(const VECTOR<int,2>& e,const int p){return e(1)==p?VECTOR<int,1>(e(2)):VECTOR<int,1>(e(1));}
inline VECTOR<int,2> Other_Nodes_In_Element(const VECTOR<int,3>& e,const int p){return e(1)==p?VECTOR<int,2>(e(2),e(3)):(e(2)==p?VECTOR<int,2>(e(3),e(1)):VECTOR<int,2>(e(1),e(2)));}
inline VECTOR<int,3> Other_Nodes_In_Element(const VECTOR<int,4>& e,const int p){return e(1)==p?VECTOR<int,3>(e(2),e(3),e(4)):(e(2)==p?VECTOR<int,3>(e(3),e(4),e(1)):(e(3)==p?VECTOR<int,3>(e(4),e(1),e(2)):VECTOR<int,3>(e(1),e(2),e(3))));}
inline int Other_Node_In_Element(const VECTOR<int,2>&e,const int p1){return e(1)==p1?e(2):e(1);}
inline int Other_Node_In_Element(const VECTOR<int,3>& e,const int p1,const int p2){return (e(1)!=p1&&e(1)!=p2)?e(1):(e(2)!=p1&&e(2)!=p2)?e(2):e(3);}
inline int Other_Node_In_Element(const VECTOR<int,4>& e,const int p1,const int p2,const int p3){for(int i=1;i<=4;i++)if(e(i)!=p1&&e(i)!=p2&&e(i)!=p3)return e(i);return 0;}
inline VECTOR<int,2> Other_Nodes_In_Element(const VECTOR<int,4>& e,const int p1,const int p2){VECTOR<int,2> nodes;int c=1;for(int d=1;d<=4;d++)if(e(d)!=p1&&e(d)!=p2){nodes(c++)=e(d);if(c>2)break;}return nodes;}
inline int Other_Node_In_Elements(const VECTOR<int,3>& e1,const VECTOR<int,3>& e2,const int a,const int b,const int c)
{for(int d=1;d<=3;d++){if(e1(d)!=a&&e1(d)!=b&&e1(d)!=c)return e1(d);}for(int d=1;d<=3;d++){if(e2(d)!=a&&e2(d)!=b&&e2(d)!=c)return e2(d);}return 0;}

inline VECTOR<int,2> Replaced(const VECTOR<int,2>& e,const int v0,const int v1){VECTOR<int,2> replaced_e=e;for(int i=1;i<=2;i++)if(replaced_e(i)==v0)replaced_e(i)=v1;return replaced_e;} ////replace v0 with v1
inline VECTOR<int,3> Replaced(const VECTOR<int,3>& e,const int v0,const int v1){VECTOR<int,3> replaced_e=e;for(int i=1;i<=3;i++)if(replaced_e(i)==v0)replaced_e(i)=v1;return replaced_e;} ////replace v0 with v1
inline VECTOR<int,4> Replaced(const VECTOR<int,4>& e,const int v0,const int v1){VECTOR<int,4> replaced_e=e;for(int i=1;i<=4;i++)if(replaced_e(i)==v0)replaced_e(i)=v1;return replaced_e;} ////replace v0 with v1
inline VECTOR<int,2> Replaced_Two(const VECTOR<int,2>& e,const int v0,const int v1,const int v2){VECTOR<int,2> replaced_e=e;for(int i=1;i<=2;i++)if(replaced_e(i)==v0||replaced_e(i)==v1)replaced_e(i)=v2;return replaced_e;} ////replace v0 and v1 with v2
inline VECTOR<int,3> Replaced_Two(const VECTOR<int,3>& e,const int v0,const int v1,const int v2){VECTOR<int,3> replaced_e=e;for(int i=1;i<=3;i++)if(replaced_e(i)==v0||replaced_e(i)==v1)replaced_e(i)=v2;return replaced_e;} ////replace v0 and v1 with v2
inline VECTOR<int,4> Replaced_Two(const VECTOR<int,4>& e,const int v0,const int v1,const int v2){VECTOR<int,4> replaced_e=e;for(int i=1;i<=4;i++)if(replaced_e(i)==v0||replaced_e(i)==v1)replaced_e(i)=v2;return replaced_e;} ////replace v0 and v1 with v2
inline bool Replaced_Two(const VECTOR<int,2>& e,VECTOR<int,2>& replaced_e,const int v0,const int v1,const int v2){replaced_e=e;bool replaced=false;for(int i=1;i<=2;i++)if(replaced_e(i)==v0||replaced_e(i)==v1){replaced_e(i)=v2;replaced=true;}return replaced;} ////replace v0 and v1 with v2
inline bool Replaced_Two(const VECTOR<int,3>& e,VECTOR<int,3>& replaced_e,const int v0,const int v1,const int v2){replaced_e=e;bool replaced=false;for(int i=1;i<=3;i++)if(replaced_e(i)==v0||replaced_e(i)==v1){replaced_e(i)=v2;replaced=true;}return replaced;} ////replace v0 and v1 with v2
inline bool Replaced_Two(const VECTOR<int,4>& e,VECTOR<int,4>& replaced_e,const int v0,const int v1,const int v2){replaced_e=e;bool replaced=false;for(int i=1;i<=4;i++)if(replaced_e(i)==v0||replaced_e(i)==v1){replaced_e(i)=v2;replaced=true;}return replaced;} ////replace v0 and v1 with v2
inline VECTOR<int,2> Reordered(const VECTOR<int,2>& e,const int v0){return v0==e(1)?VECTOR<int,2>(e(1),e(2)):VECTOR<int,2>(e(2),e(1));}
inline VECTOR<int,3> Reordered(const VECTOR<int,3>& e,const int v0){return v0==e(1)?VECTOR<int,3>(e(1),e(2),e(3)):v0==e(2)?VECTOR<int,3>(e(2),e(3),e(1)):VECTOR<int,3>(e(3),e(1),e(2));}
inline VECTOR<int,4> Reordered(const VECTOR<int,4>& e,const int v0){return v0==e(1)?VECTOR<int,4>(e(1),e(2),e(3),e(4)):v0==e(2)?VECTOR<int,4>(e(2),e(3),e(1),e(4)):v0==e(3)?VECTOR<int,4>(e(3),e(1),e(2),e(4)):VECTOR<int,4>(e(4),e(2),e(1),e(3));}

inline VECTOR<int,2> Reversed(const VECTOR<int,2>& e){return VECTOR<int,2>(e(2),e(1));}
inline VECTOR<int,3> Reversed(const VECTOR<int,3>& e){return VECTOR<int,3>(e(3),e(2),e(1));}
inline VECTOR<int,4> Reversed(const VECTOR<int,4>& e){return VECTOR<int,4>(e(1),e(2),e(4),e(3));}
inline bool Is_Same_Order(const VECTOR<int,3>& e1,const VECTOR<int,3>& e2){return e1==Reordered(e2,e1(1));}
inline bool Is_Same_Order(const VECTOR<int,2>& edge,const VECTOR<int,3>& tri){return Reordered(tri,edge(1))(2)==edge(2);}
inline bool Element_Contains(const VECTOR<int,3>& e,const VECTOR<int,2>& f){return e.Contains(f(1))&&e.Contains(f(2));}
inline bool Element_Contains(const VECTOR<int,4>& e,const VECTOR<int,3>& f){return e.Contains(f(1))&&e.Contains(f(2))&&e.Contains(f(3));}
inline bool Element_Contains_Not(const VECTOR<int,2>& e,const int f){return e(1)!=f||e(2)!=f;}
inline bool Element_Contains_Not(const VECTOR<int,3>& e,const int f){return e(1)!=f||e(2)!=f||e(3)!=f;}
inline bool Element_Contains_Not(const VECTOR<int,4>& e,const int f){return e(1)!=f||e(2)!=f||e(3)!=f||e(4)!=f;}

inline int Shared_Vertex(const VECTOR<int,3>& e1,const VECTOR<int,3>& e2,const VECTOR<int,3>& e3)
{return (e2.Contains(e1(1))&&e3.Contains(e1(1)))?e1(1):((e2.Contains(e1(2))&&e3.Contains(e1(2)))?e1(2):(e2.Contains(e1(3))&&e2.Contains(e1(3))?e1(3):-1));}
inline int Shared_Vertex(const VECTOR<int,2>& e1,const VECTOR<int,2>& e2)
{return e2.Contains(e1(1))?e1(1):(e2.Contains(e1(2))?e1(2):-1);}
inline VECTOR<int,2> Shared_Edge(const VECTOR<int,3>& e1,const VECTOR<int,3>& e2)
{if(!e2.Contains(e1(1)))return VECTOR<int,2>(e1(2),e1(3));else if(!e2.Contains(e1(2)))return VECTOR<int,2>(e1(3),e1(1));else return VECTOR<int,2>(e1(1),e1(2));}

inline bool Contains_Duplicate_Elements(const VECTOR<int,3>& e){return e(1)==e(2)||e(1)==e(3)||e(2)==e(3);}

inline VECTOR<int,2> Unique_Ordered(const VECTOR<int,2>& e){return e.Sorted();}
inline VECTOR<int,3> Unique_Ordered(const VECTOR<int,3>& e){return Reordered(e,e.Min());}
inline VECTOR<int,4> Unique_Ordered(const VECTOR<int,4>& e)
{VECTOR<int,4> v1=Reordered(e,e.Min());VECTOR<int,3> tri=Unique_Ordered(VECTOR<int,3>(v1(2),v1(3),v1(4)));v1(2)=tri(1);v1(3)=tri(2);v1(4)=tri(3);return v1;}

inline void Simplex_Edges(const VECTOR<int,4>& e,/*result*/ARRAY<VECTOR<int,2> >& edges)
{edges.Append(VECTOR<int,2>(e(1),e(2)));edges.Append(VECTOR<int,2>(e(3),e(4)));
edges.Append(VECTOR<int,2>(e(1),e(3)));edges.Append(VECTOR<int,2>(e(2),e(4)));
edges.Append(VECTOR<int,2>(e(1),e(4)));edges.Append(VECTOR<int,2>(e(2),e(3)));}
inline void Simplex_Edges(const VECTOR<int,3>& e,/*result*/ARRAY<VECTOR<int,2> >& edges){edges.Append(VECTOR<int,2>(e(1),e(2)));edges.Append(VECTOR<int,2>(e(2),e(3)));edges.Append(VECTOR<int,2>(e(3),e(1)));}
inline void Simplex_Edges(const VECTOR<int,2>& e,/*result*/ARRAY<VECTOR<int,2> >& edges){edges.Append(e);}

inline void Simplex_Faces(const VECTOR<int,3>& e,/*result*/ARRAY<VECTOR<int,2> >& faces)
{faces.Append(VECTOR<int,2>(e(1),e(2)));faces.Append(VECTOR<int,2>(e(2),e(3)));faces.Append(VECTOR<int,2>(e(3),e(1)));}
inline void Simplex_Faces(const VECTOR<int,4>& e,/*result*/ARRAY<VECTOR<int,3> >& faces)
{faces.Append(VECTOR<int,3>(e(1),e(3),e(2)));faces.Append(VECTOR<int,3>(e(1),e(2),e(4)));faces.Append(VECTOR<int,3>(e(1),e(4),e(3)));faces.Append(VECTOR<int,3>(e(2),e(3),e(4)));}

template<int dimension> struct SIMPLEX_NUMBER_HELPER;
template<> struct SIMPLEX_NUMBER_HELPER<2>{static int Edge_Number(){return 1;} static int Face_Number(){return 1;}};
template<> struct SIMPLEX_NUMBER_HELPER<3>{static int Edge_Number(){return 3;} static int Face_Number(){return 3;}};
template<> struct SIMPLEX_NUMBER_HELPER<4>{static int Edge_Number(){return 6;} static int Face_Number(){return 4;}};
template<int dimension> int Simplex_Edge_Number(){return SIMPLEX_NUMBER_HELPER<dimension>::Edge_Number();}
template<int dimension> int Simplex_Face_Number(){return SIMPLEX_NUMBER_HELPER<dimension>::Face_Number();}

inline VECTOR<int,3> Simplex_Opposite_Face(const VECTOR<int,4>& e,const int v)
{
    if(v==e(1))return VECTOR<int,3>(e(2),e(3),e(4));
    else if(v==e(2))return VECTOR<int,3>(e(1),e(4),e(3));
    else if(v==e(3))return VECTOR<int,3>(e(1),e(2),e(4));
    else return VECTOR<int,3>(e(1),e(3),e(2));
}

////split tet abcd to vbcd and vabd
inline void Two_Splitted_Tetrahedron_Vertices(const VECTOR<int,4>& tet_vtx,const int a,const int c,const int v,VECTOR<int,4>& vbcd,VECTOR<int,4>& vabd)
{   
    VECTOR<int,2> b_and_d=Other_Nodes_In_Element(tet_vtx,a,c);int b=b_and_d.x;int d=b_and_d.y;
    if(Unique_Ordered(tet_vtx)!=Unique_Ordered(VECTOR<int,4>(a,b,c,d))){int tmp=b;b=d;d=tmp;}
    vbcd=VECTOR<int,4>(v,b,c,d);vabd=VECTOR<int,4>(v,a,b,d);
}

////split tri acd to avd and vcd (or adv and vdc)
inline void Two_Splitted_Triangle_Vertices(const VECTOR<int,3>& tri_vtx,const int a,const int c,const int v,VECTOR<int,3>& avd,VECTOR<int,3>& vcd)
{
    int d=Other_Node_In_Element(tri_vtx,a,c);bool flip=(Unique_Ordered(tri_vtx)!=Unique_Ordered(VECTOR<int,3>(a,c,d)));
    if(!flip){avd=VECTOR<int,3>(a,v,d);vcd=VECTOR<int,3>(v,c,d);}else{avd=VECTOR<int,3>(a,d,v);vcd=VECTOR<int,3>(v,d,c);}
}

////split tet abcd 
inline void Three_Splitted_Tetrahedron_Vertices(const VECTOR<int,4>& tet,const VECTOR<int,3>& face,const int v,VECTOR<int,4>& adbv,VECTOR<int,4>& bdcv,VECTOR<int,4>& cdav)
{
    int a=face(1);int b=face(2);int c=face(3);int d=Other_Node_In_Element(tet,a,b,c);if(Unique_Ordered(tet)!=Unique_Ordered(VECTOR<int,4>(a,b,c,d))){int tmp=b;b=c;c=tmp;}
    adbv=VECTOR<int,4>(a,d,b,v);bdcv=VECTOR<int,4>(b,d,c,v);cdav=VECTOR<int,4>(c,d,a,v);
}

//////////////////////////////////////////////////////////////////////////
////array operations

template<class T> inline bool Replace_In_Array(ARRAY<T>& e,const T v0,const T v1){for(int i=1;i<=e.m;i++)if(e(i)==v0){e(i)=v1;return true;}return false;}
template<class T> inline bool Remove_From_Array(ARRAY<T>& e,const T v){int i=e.Find(v);if(i!=0){e.Remove_Index(i);return true;}return false;}
template<class T> inline int Shared_Element_Index(const ARRAY<T>& e1,const ARRAY<T>& e2){for(int i=1;i<=e1.m;i++)for(int j=1;j<=e2.m;j++)if(e1(i)==e2(j))return i;return 0;}

template<class TV> void VECTOR_ND_to_ARRAY_TV(const VECTOR_ND<typename TV::SCALAR>& vector,ARRAY<TV>& array)
{array.Resize(vector.n/TV::m);for(int i=1;i<=array.m;i++)for(int d=1;d<=TV::m;d++)array(i)(d)=vector((i-1)*TV::m+d);}

template<class TV,class TV_ARRAY,class ID> void ARRAY_TV_to_VECTOR_ND(const ARRAY_BASE<TV,TV_ARRAY,ID>& array,VECTOR_ND<typename TV::SCALAR>& vector)
{vector.Resize(array.Size()*TV::m);for(int i=1;i<=array.Size();i++)for(int d=1;d<=TV::m;d++)vector((i-1)*TV::m+d)=array(i)(d);}

template<class T,class T_ARRAY,class ID> void ARRAY_T_to_VECTOR_ND(const ARRAY_BASE<T,T_ARRAY,ID>& array,VECTOR_ND<T>& vector)
{vector.Resize(array.Size());for(int i=1;i<=array.Size();i++)vector(i)=array(i);}

template<int dimension> struct VECTOR_HELPER;
template<> struct VECTOR_HELPER<2>{template<class T> static VECTOR<T,2> V(const VECTOR<T,3>& v){return VECTOR<T,2>(v.x,v.y);}};
template<> struct VECTOR_HELPER<3>{template<class T> static VECTOR<T,3> V(const VECTOR<T,3>& v){return v;}};
template<class T,int dimension> VECTOR<T,dimension> TV_Compatible(const T x,const T y,const T z)	////make 3d vector compiled in 2d
{return VECTOR_HELPER<dimension>::V(VECTOR<T,3>(x,y,z));}
template<class T,int dimension> VECTOR<T,dimension> TV_Compatible(const VECTOR<T,3>& v)	////make 3d vector compiled in 2d
{return VECTOR_HELPER<dimension>::V(v);}

template<int dimension> struct CELL_NEIGHBOR_ITERATOR_HELPER;
template<> struct CELL_NEIGHBOR_ITERATOR_HELPER<2>{static VECTOR<int,2> Index(const int i){return VECTOR<int,2>((i-1)/3-1,(i-1)%3-1);}};
template<> struct CELL_NEIGHBOR_ITERATOR_HELPER<3>{static VECTOR<int,3> Index(const int i){return VECTOR<int,3>((i-1)/9-1,(i-1)%9/3-1,(i-1)%9%3-1);}};
template<int dimension> VECTOR<int,dimension> CELL_NEIGHBOR_ITERATOR(const int i)
{return CELL_NEIGHBOR_ITERATOR_HELPER<dimension>::Index(i);}

template<class T> inline VECTOR<T,1> Element(const ARRAY_VIEW<T>& values,const VECTOR<int,1>& index){return VECTOR<T,1>(values(index(1)));}
template<class T> inline VECTOR<T,2> Element(const ARRAY_VIEW<T>& values,const VECTOR<int,2>& index){return VECTOR<T,2>(values(index(1)),values(index(2)));}
template<class T> inline VECTOR<T,3> Element(const ARRAY_VIEW<T>& values,const VECTOR<int,3>& index){return VECTOR<T,3>(values(index(1)),values(index(2)),values(index(3)));}
template<class T> inline VECTOR<T,4> Element(const ARRAY_VIEW<T>& values,const VECTOR<int,4>& index){return VECTOR<T,4>(values(index(1)),values(index(2)),values(index(3)),values(index(4)));}
template<class T> inline VECTOR<T,1> Element(const ARRAY<T>& values,const VECTOR<int,1>& index){return VECTOR<T,1>(values(index(1)));}
template<class T> inline VECTOR<T,2> Element(const ARRAY<T>& values,const VECTOR<int,2>& index){return VECTOR<T,2>(values(index(1)),values(index(2)));}
template<class T> inline VECTOR<T,3> Element(const ARRAY<T>& values,const VECTOR<int,3>& index){return VECTOR<T,3>(values(index(1)),values(index(2)),values(index(3)));}
template<class T> inline VECTOR<T,4> Element(const ARRAY<T>& values,const VECTOR<int,4>& index){return VECTOR<T,4>(values(index(1)),values(index(2)),values(index(3)),values(index(4)));}

inline bool Element_Or(const ARRAY_VIEW<bool>& flags,const VECTOR<int,2>& index){return flags(index(1))||flags(index(2));}
inline bool Element_Or(const ARRAY_VIEW<bool>& flags,const VECTOR<int,3>& index){return flags(index(1))||flags(index(2))||flags(index(3));}
inline bool Element_Or(const ARRAY_VIEW<bool>& flags,const VECTOR<int,4>& index){return flags(index(1))||flags(index(2))||flags(index(3))||flags(index(4));}

template<class T> inline VECTOR<T,3> Normalize_With_Zero(const VECTOR<T,3>& v){T magnitude=v.Magnitude();if(magnitude>(T)0) return v*((T)1/magnitude);else return VECTOR<T,3>();}

//////////////////////////////////////////////////////////////////////////
////hashtable operations
template<class T_KEY> void Copy_Hashtable_Elements(HASHTABLE<T_KEY>& hashtable_from,HASHTABLE<T_KEY>& hashtable_to)
{for(HASHTABLE_ITERATOR<T_KEY> iter(hashtable_from);iter.Valid();iter.Next()){hashtable_to.Set(iter.Key());}}

template<class T_KEY,class T_DATA> void Copy_Hashtable_Elements(HASHTABLE<T_KEY,T_DATA>& hashtable_from,HASHTABLE<T_KEY,T_DATA>& hashtable_to)
{for(HASHTABLE_ITERATOR<T_KEY,T_DATA> iter(hashtable_from);iter.Valid();iter.Next()){hashtable_to.Set(iter.Key(),iter.Data());}}

//////////////////////////////////////////////////////////////////////////
////geometry operations
template<class T_MESH> void Find_Neighbor_Particles_On_Mesh(const T_MESH& mesh,const int p,ARRAY<int>& nbs,bool use_two_ring=false)
{
    nbs=(*mesh.neighbor_nodes)(p);if(nbs.m==0)return;
    if(use_two_ring){int one_ring_m=nbs.m;for(int i=1;i<=one_ring_m;i++){int nb_p=nbs(i);for(int j=1;j<=(*mesh.neighbor_nodes)(nb_p).m;j++){int p2=(*mesh.neighbor_nodes)(nb_p)(j);nbs.Append_Unique(p2);}}}
    else nbs.Append(p);
}

template<class T_MESH> void Find_Neighbor_Particles_On_Mesh(const T_MESH& mesh,const int p,ARRAY<int>& nbs,int number_of_rings=2)
{
    if((*mesh.neighbor_nodes)(p).m==0)return;
    HASHTABLE<int> nbs_hashtable;nbs_hashtable.Set(p);
    nbs.Append(p);int start=1;int end=nbs.m;
    for(int i=1;i<=number_of_rings;i++){int c=0;
        for(int j=start;j<=end;j++){int nb_p=nbs(j);
            for(int k=1;k<=(*mesh.neighbor_nodes)(nb_p).m;k++){
                int p2=(*mesh.neighbor_nodes)(nb_p)(k);
                if(!nbs_hashtable.Contains(p2)){
                    nbs_hashtable.Set(p2);nbs.Append(p2);c++;
                }
            }
        }
        start=end+1;end=end+c;
    }
}

template<class TV_INTN> void Find_Neighbor_Particles_On_Mesh(const ARRAY<TV_INTN>& elements,const HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const int p,ARRAY<int>& nbs,int number_of_rings=2)
{
    const ARRAY<int>* p_elements=vertex_element_hashtable.Get_Pointer(p);
    if(!p_elements||p_elements->m==0)return;

    HASHTABLE<int> nbs_hashtable;nbs_hashtable.Set(p);
    nbs.Append(p);int start=1;int end=nbs.m;
    for(int i=1;i<=number_of_rings;i++){int c=0;
        for(int j=start;j<=end;j++){int nb_p=nbs(j);
            const ARRAY<int>* v_elements=vertex_element_hashtable.Get_Pointer(nb_p);
            if(v_elements)for(int k=1;k<=(*v_elements).m;k++){
                const TV_INTN& vtx=elements((*v_elements)(k));
                for(int d=1;d<=TV_INTN::m;d++){
                    int p2=vtx(d);
                    if(!nbs_hashtable.Contains(p2)){
                        nbs_hashtable.Set(p2);nbs.Append(p2);c++;
                    }
                }
            }
        }
    start=end+1;end=end+c;
    }
}

template<class T> T Volume_Length_Measure(const VECTOR<T,3>& a,const VECTOR<T,3>& b,const VECTOR<T,3>& c,const VECTOR<T,3>& d)
{
    T edge_rms=sqrt(((a-b).Magnitude_Squared()+(a-c).Magnitude_Squared()+(a-d).Magnitude_Squared()+(b-c).Magnitude_Squared()+(b-d).Magnitude_Squared()+(c-d).Magnitude_Squared())/(T)6);
    return TETRAHEDRON<T>(a,b,c,d).Volume()/pow<3>(edge_rms)*(T)6*sqrt((T)2);
}

template<class T> T Area_Length_Measure(const VECTOR<T,3>& a,const VECTOR<T,3>& b,const VECTOR<T,3>& c)
{
    T edge_rms=sqrt(((a-b).Magnitude_Squared()+(a-c).Magnitude_Squared()+(b-c).Magnitude_Squared())*(T)one_third);
    return TRIANGLE_3D<T>::Area(a,b,c)/pow<2>(edge_rms)*(T)4/(T)3*sqrt((T)3);
}

template<class T> VECTOR<T,3> Projected_Point_On_Triangle_Plane(const TRIANGLE_3D<T>& tri,const VECTOR<T,3>& p)
{return p-tri.Signed_Distance(p)*tri.normal;}

template<class T> VECTOR<T,3> Projected_Barycentric_Coordinates(const TRIANGLE_3D<T>& tri,const VECTOR<T,3>& p)
{return tri.Barycentric_Coordinates(Projected_Point_On_Triangle_Plane(tri,p));}

template<class T> VECTOR<T,3> Closest_Point_On_Segment(const VECTOR<T,3>& point,const SEGMENT_3D<T>& segment,/*result*/T& t)
{                  
    VECTOR<T,3> v=segment.x2-segment.x1;
    T denominator=VECTOR<T,3>::Dot_Product(v,v);
    if(denominator==0){t=(T)0;return segment.x1;}
    else{
        t=VECTOR<T,3>::Dot_Product(point-segment.x1,v)/denominator;
        if(t<=0) return segment.x1;
        else if(t>=1) return segment.x2;
        else{v=segment.x1+(segment.x2-segment.x1)*t;return v;}
    }
}

template<class T> T Distance_From_Point_To_Segment(const VECTOR<T,3>& point,const SEGMENT_3D<T>& segment,/*result*/T& t)
{
    VECTOR<T,3> v=Closest_Point_On_Segment(point,segment,t),d=v-point;
    return d.Magnitude();
}

template<class T> T Nearest_Vertex_Edge_Pair_On_Tetrahedron(const TETRAHEDRON<T>& tet,/*result*/int& min_p,/*result*/VECTOR<int,2>& min_e,/*result*/T& min_t)
{
    T min_distance=(T)FLT_MAX;VECTOR<int,4> tet_idx(1,2,3,4);
    for(int i=1;i<=4;i++){
        int p=tet_idx(i);VECTOR<int,3> face=Simplex_Opposite_Face(tet_idx,p);
        ARRAY<VECTOR<int,2> > edges;Simplex_Edges(face,edges);
        for(int j=1;j<=3;j++){
            T t=(T)0;T distance=Distance_From_Point_To_Segment(tet.X(p),SEGMENT_3D<T>(tet.X(edges(j)(1)),tet.X(edges(j)(2))),t);
            LOG::cout<<"distance from v "<<p<<" to edge "<<edges(j)<<", t "<<t<<", d: "<<distance<<std::endl;
            if(distance<min_distance){min_p=p;min_e=edges(j);min_distance=distance;min_t=t;}
        }
    }
    return min_distance;
}

template<class T> T Nearest_Edge_Edge_Pair_On_Tetrahedron(const TETRAHEDRON<T>& tet,/*result*/VECTOR<int,2>& min_e1,/*result*/VECTOR<int,2>& min_e2,/*result*/T& weight1,/*result*/T& weight2)
{
    T min_distance=(T)FLT_MAX;VECTOR<int,4> tet_idx(1,2,3,4);
    ARRAY<VECTOR<int,2> > edges;Simplex_Edges(tet_idx,edges);VECTOR<T,2> weights;
    for(int i=1;i<=3;i++){
        const VECTOR<int,2>& e1=edges(i*2-1);const VECTOR<int,2>& e2=edges(i*2);    ////assuming Simplex_Edges return pair of oppositing edges in sequence
        SEGMENT_3D<T> s1(tet.X(e1(1)),tet.X(e1(2)));SEGMENT_3D<T> s2(tet.X(e2(1)),tet.X(e2(2)));
        VECTOR<T,3> shortest_vector=s1.Shortest_Vector_Between_Segments(s2,weights);T distance=shortest_vector.Magnitude();
        if(distance<min_distance){min_e1=e1;min_e2=e2;min_distance=distance;weight1=weights(1);weight2=weights(2);}
    }
    return min_distance;
}

template<class T> T Nearest_Vertex_Face_Pair_On_Tetrahedron(const TETRAHEDRON<T>& tet,/*result*/int& p,/*result*/VECTOR<int,3>& face,/*result*/VECTOR<T,3>& uvw)
{
    VECTOR<int,4> tet_idx(1,2,3,4);
    T min_altitude=tet.Minimum_Altitude(p);
    face=Simplex_Opposite_Face(tet_idx,p);
    uvw=Projected_Barycentric_Coordinates(TRIANGLE_3D<T>(tet.X(face(1)),tet.X(face(2)),tet.X(face(3))),tet.X(p));
    return min_altitude;
}

template<class T> bool Has_Short_Edge(const TETRAHEDRON<T>& tet,const T edge_length_threshold)
{
    VECTOR<int,4> tet_v(1,2,3,4);ARRAY<VECTOR<int,2> > edges;Simplex_Edges(tet_v,edges);
    for(int d=1;d<=6;d++)if((tet.X(edges(d)(1))-tet.X(edges(d)(2))).Magnitude()<edge_length_threshold){return true;}return false;
}

template<class T> class SKINNY_TETRAHEDRON
{typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    VECTOR<int,4> guid;
    ARRAY<int> vertices;
    ARRAY<TV_INT2> edges;
    ARRAY<TV_INT3> faces;
    ARRAY<T> edge_weights;
    ARRAY<TV3> face_uvw;

    ARRAY<int> edge_splitted_particle_guid;
    ARRAY<int> face_splitted_particle_guid;

    SKINNY_TETRAHEDRON(){}
    SKINNY_TETRAHEDRON(const SKINNY_TETRAHEDRON<T>& copy){
        guid=copy.guid;
        vertices=copy.vertices;
        edges=copy.edges;
        faces=copy.faces;
        edge_weights=copy.edge_weights;
        face_uvw=copy.face_uvw;
        edge_splitted_particle_guid=copy.edge_splitted_particle_guid;
        face_splitted_particle_guid=copy.face_splitted_particle_guid;
    }

    int Type() const {
        if(vertices.m==0&&edges.m==2&&faces.m==0)return SLIVER;
        else if(vertices.m==1&&edges.m==1&&faces.m==0)return SPADE;
        else if(vertices.m==1&&edges.m==0&&faces.m==1)return CAP;
        else return 0;
    }

    bool Edge_To_Collapse_Guid(TV_INT2& edge_to_collapse_guid) const
    {
        int type=Type();
        switch(type){
        case SLIVER:
            if(edge_splitted_particle_guid.m==2){
                edge_to_collapse_guid=TV_INT2(edge_splitted_particle_guid(1),edge_splitted_particle_guid(2));return true;
            }
            return false;
        case SPADE:
            if(edge_splitted_particle_guid.m==1&&vertices.m==1){
                edge_to_collapse_guid=TV_INT2(edge_splitted_particle_guid(1),guid(vertices(1)));return true;
            }
            return false;
        case CAP:
            if(face_splitted_particle_guid.m==1&&vertices.m==1){
                edge_to_collapse_guid=TV_INT2(face_splitted_particle_guid(1),guid(vertices(1)));return true;
            }
            return false;
        }
        return false;
    }

    void Add_Edge(const TV_INT2& edge,const T weight){edges.Append(edge);edge_weights.Append(weight);}
    void Add_Vertex(const int& v){vertices.Append(v);}
    void Add_Face(const TV_INT3& face,const TV3& uvw){faces.Append(face);face_uvw.Append(uvw);}
};

template<class T> int Split_Skinny_Tetrahedron(const TETRAHEDRON<T>& tet,const T distance_threshold,const T t_threshold,SKINNY_TETRAHEDRON<T>& skinny_tet)
{   
    typedef VECTOR<int,2> TV_INT2; typedef VECTOR<int,3> TV_INT3;typedef VECTOR<T,3> TV3;
    ARRAY<TV_INT2> edge_edge_e(2);ARRAY<T> weights(2);
    T min_edge_edge_distance=Nearest_Edge_Edge_Pair_On_Tetrahedron(tet,edge_edge_e(1),edge_edge_e(2),weights(1),weights(2));

    ////sliver or spade
    if(min_edge_edge_distance<distance_threshold){
        for(int e=1;e<=2;e++){
            if(weights(e)>=t_threshold&&weights(e)<=(T)1-t_threshold) skinny_tet.Add_Edge(edge_edge_e(e),weights(e));
            else if(weights(e)<t_threshold) skinny_tet.Add_Vertex(edge_edge_e(e)(1));
            else /*weights(e)>1-t_threshold*/ skinny_tet.Add_Vertex(edge_edge_e(e)(2));
        }
    }
    ////cap
    else{
        int vertex_face_p=0;TV_INT3 vertex_face_f;TV3 uvw;
        T min_vertex_face_distance=Nearest_Vertex_Face_Pair_On_Tetrahedron(tet,vertex_face_p,vertex_face_f,uvw);
        if(min_vertex_face_distance<distance_threshold){
            bool valid_uvw=true;for(int d=1;d<=3;d++)if(uvw(d)<(T)0||uvw(d)>(T)1){valid_uvw=false;break;}
            if(valid_uvw){
                skinny_tet.Add_Face(vertex_face_f,uvw);
                skinny_tet.Add_Vertex(vertex_face_p);
            }
        }
    }
    
    return skinny_tet.Type();
}

template<class T> int Split_Skinny_Tetrahedron(const TETRAHEDRON<T>& tet,const T distance_threshold,const T t_threshold,ARRAY<VECTOR<int,2> >& edges,ARRAY<T>& edge_t,VECTOR<int,3>& face,VECTOR<T,3>& face_uvw)
{   
    typedef VECTOR<int,2> TV_INT2; typedef VECTOR<int,3> TV_INT3;typedef VECTOR<T,3> TV3;
    ARRAY<TV_INT2> edge_edge_e(2);ARRAY<T> weights(2);
    T min_edge_edge_distance=Nearest_Edge_Edge_Pair_On_Tetrahedron(tet,edge_edge_e(1),edge_edge_e(2),weights(1),weights(2));
    if(min_edge_edge_distance<distance_threshold){
        for(int e=1;e<=2;e++)
            if(weights(e)>=t_threshold&&weights(e)<=(T)1-t_threshold){edges.Append(edge_edge_e(e));edge_t.Append(weights(e));}else{}
        if(edges.m==1)return SPADE;else if(edges.m==2)return SLIVER;
    }

    ////cap tet
    int vertex_face_p=0;TV_INT3 vertex_face_f;TV3 uvw;
    T min_vertex_face_distance=Nearest_Vertex_Face_Pair_On_Tetrahedron(tet,vertex_face_p,vertex_face_f,uvw);
    if(min_vertex_face_distance<distance_threshold){
        bool valid_uvw=true;for(int d=1;d<=3;d++)if(uvw(d)<(T)0||uvw(d)>(T)1){valid_uvw=false;break;}
        if(valid_uvw){face=vertex_face_f;face_uvw=uvw;
        //LOG::cout<<"find cap tet, uvw: "<<uvw<<", distance: "<<min_vertex_face_distance<<" < "<<distance_threshold<<std::endl;
        return CAP;}
    }

    return 0;
}

template<class T> inline T Cotangent(const VECTOR<T,3>& a,const VECTOR<T,3>& b,const VECTOR<T,3>& c)
{
    VECTOR<T,3> ab=b-a;VECTOR<T,3> ac=c-a;
    return VECTOR<T,3>::Dot_Product(ab,ac)/VECTOR<T,3>::Cross_Product(ab,ac).Magnitude();
}

template<class T> T Dihedral_Angle(const TETRAHEDRON<T>& tet,const VECTOR<int,2>& edge)
{
    int a=edge(1);int c=edge(2);
    VECTOR<int,4> indices(1,2,3,4);
    VECTOR<const TRIANGLE_3D<T>*,4> tris;tris(4)=&(tet.triangle1);tris(3)=&(tet.triangle2);tris(2)=&(tet.triangle3);tris(1)=&(tet.triangle4);
    VECTOR<int,2> b_and_d=Other_Nodes_In_Element(indices,a,c);int b=b_and_d(1);int d=b_and_d(2);
    const TRIANGLE_3D<T>* tri_acd=tris(b);const TRIANGLE_3D<T>* tri_acb=tris(d);
    T dihedral_angle=acos(-VECTOR<T,3>::Dot_Product(tri_acd->normal,tri_acb->normal));
    return dihedral_angle;
}

template<class T> T Dihedral_Angle(const VECTOR<T,3>& a,const VECTOR<T,3>& c,const VECTOR<T,3>& b,const VECTOR<T,3>& d)     ////dihedral angle between abc and acd
{
    T dihedral_angle=acos(-VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(a-c,b-c).Normalized(),VECTOR<T,3>::Cross_Product(d-c,a-c).Normalized()));
    return dihedral_angle;
}

template<class T> T Oriented_Dihedral_Angle(const VECTOR<T,3>& a,const VECTOR<T,3>& c,/*left*/const VECTOR<T,3>& b,/*right*/const VECTOR<T,3>& d)     ////dihedral angle between abc and acd, [0-2pi)
{
    VECTOR<T,3> n_abc=VECTOR<T,3>::Cross_Product(a-c,b-c).Normalized();
    VECTOR<T,3> n_acd=VECTOR<T,3>::Cross_Product(d-c,a-c).Normalized();
    VECTOR<T,3> n_ac=(c-a).Normalized();
    T cos_theta=-VECTOR<T,3>::Dot_Product(n_abc,n_acd);
    T sign=VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(n_abc,n_acd),n_ac)>(T)0?(T)1:(T)-1;
    T angle=sign*acos(cos_theta);//if(angle<(T)0)angle+=(T)two_pi;return angle;
    return angle;
}

inline void Oriented_Triangle_Pair_Vertices(const VECTOR<int,3>& v1,const VECTOR<int,3>& v2,int& a,int& c,int& b,int& d)
{
    VECTOR<int,2> ac=Shared_Edge(v1,v2);a=ac(1);c=ac(2);d=Other_Node_In_Element(v1,a,c);b=Other_Node_In_Element(v2,a,c);
}

template<class T> inline T Oriented_Dihedral_Angle(const ARRAY_VIEW<VECTOR<T,3> >& X,const VECTOR<int,3>& v1,const VECTOR<int,3>& v2)
{
    int a,c,b,d;Oriented_Triangle_Pair_Vertices(v1,v2,a,c,b,d);
    return Oriented_Dihedral_Angle(X(a),X(c),X(b),X(d));
}

template<class T> inline T Min_Cosine_Angle(const VECTOR<T,3>& a,const VECTOR<T,3>& b,const VECTOR<T,3>& c)
{
    VECTOR<T,3> ab=(b-a).Normalized();VECTOR<T,3> bc=(c-b).Normalized();VECTOR<T,3> ca=(a-c).Normalized();
    return min(VECTOR<T,3>::Dot_Product(ab,-ca),VECTOR<T,3>::Dot_Product(-ab,bc),VECTOR<T,3>::Dot_Product(ca,-bc));
}

template<class T> inline T Max_Cosine_Angle(const VECTOR<T,3>& a,const VECTOR<T,3>& b,const VECTOR<T,3>& c)
{
    VECTOR<T,3> ab=(b-a).Normalized();VECTOR<T,3> bc=(c-b).Normalized();VECTOR<T,3> ca=(a-c).Normalized();
    return max(VECTOR<T,3>::Dot_Product(ab,-ca),VECTOR<T,3>::Dot_Product(-ab,bc),VECTOR<T,3>::Dot_Product(ca,-bc));
}

template<class T> RIGID_BODY<VECTOR<T,2> >* Get_Cylinder_Rigid_Body(RIGIDS_STANDARD_TESTS<VECTOR<T,2> >& rigid_standard_tests,const VECTOR<T,2>& euler_angles,const VECTOR<T,2>& translation,T height,T radius)
{
    return 0;
}

template<class T> RIGID_BODY<VECTOR<T,3> >* Get_Cylinder_Rigid_Body(RIGIDS_STANDARD_TESTS<VECTOR<T,3> >& rigid_standard_tests,const VECTOR<T,3>& euler_angles,const VECTOR<T,3>& translation,T height,T radius)
{
    RIGID_BODY<VECTOR<T,3> >* cylinder=&rigid_standard_tests.Add_Analytic_Cylinder(height,radius);
    cylinder->Rotation()*=ROTATION<VECTOR<T,3> >::From_Euler_Angles(euler_angles(1),euler_angles(2),euler_angles(3));
    cylinder->X()+=translation;
    return cylinder;
}

////smoothing
template<class T> T Gaussian_Kernel(const T distance_squared){return exp(-distance_squared*(T)4);}

////accurate arithmetic 
template<class T> int Accurate_Tetrahedron_Orientation(const VECTOR<T,3>& p1,const VECTOR<T,3>& p2,const VECTOR<T,3>& p3,const VECTOR<T,3>& p4)
{double orient=ADAPTIVE_EXACT_PREDICATES::orient3d(VECTOR<double,3>(p1).begin(),VECTOR<double,3>(p2).begin(),VECTOR<double,3>(p3).begin(),VECTOR<double,3>(p4).begin());return orient<0.?1:(orient==0.?0:-1);}
template<class T> int Accurate_Tetrahedron_Orientation(ARRAY_VIEW<VECTOR<T,3> >& X,const VECTOR<int,4>& v){return Accurate_Tetrahedron_Orientation(X(v(1)),X(v(2)),X(v(3)),X(v(4)));}

template<class T> double Accurate_Tetrahedron_Volume(const VECTOR<T,3>& p1,const VECTOR<T,3>& p2,const VECTOR<T,3>& p3,const VECTOR<T,3>& p4)
{return -one_sixth*ADAPTIVE_EXACT_PREDICATES::orient3d(VECTOR<double,3>(p1).begin(),VECTOR<double,3>(p2).begin(),VECTOR<double,3>(p3).begin(),VECTOR<double,3>(p4).begin());}
template<class T> double Accurate_Tetrahedron_Volume(ARRAY_VIEW<VECTOR<T,3> >& X,const VECTOR<int,4>& v){return Accurate_Tetrahedron_Volume(X(v(1)),X(v(2)),X(v(3)),X(v(4)));}

template<class T> bool Is_Negative_Volume_Tetrahedron(const VECTOR<T,3>& p1,const VECTOR<T,3>& p2,const VECTOR<T,3>& p3,const VECTOR<T,3>& p4)
{return Accurate_Tetrahedron_Orientation(p1,p2,p3,p4)<0;}
template<class T> bool Is_Negative_Volume_Tetrahedron(ARRAY_VIEW<VECTOR<T,3> >& X,const VECTOR<int,4>& v){return Is_Negative_Volume_Tetrahedron(X(v(1)),X(v(2)),X(v(3)),X(v(4)));}

template<class T> bool Is_Non_Positive_Volume_Tetrahedron(const VECTOR<T,3>& p1,const VECTOR<T,3>& p2,const VECTOR<T,3>& p3,const VECTOR<T,3>& p4,const T epsilon=(T)1e-25)
{return Accurate_Tetrahedron_Orientation(p1,p2,p3,p4)<=epsilon;}
template<class T> bool Is_Non_Positive_Volume_Tetrahedron(ARRAY_VIEW<VECTOR<T,3> >& X,const VECTOR<int,4>& v,const T epsilon=(T)1e-25){return Is_Non_Positive_Volume_Tetrahedron(X(v(1)),X(v(2)),X(v(3)),X(v(4)),epsilon);}

template<class T> bool Has_Negative_Volume_Tetrahedron(const ARRAY<VECTOR<VECTOR<T,3>, 4> >& tet_X)
{for(int i=1;i<=tet_X.m;i++)if(Is_Negative_Volume_Tetrahedron(tet_X(i)(1),tet_X(i)(2),tet_X(i)(3),tet_X(i)(4)))return true;return false;}
template<class T> bool Has_Non_Positive_Volume_Tetrahedron(const ARRAY<VECTOR<VECTOR<T,3>, 4> >& tet_X,const T epsilon=(T)1e-25)
{for(int i=1;i<=tet_X.m;i++)if(Is_Non_Positive_Volume_Tetrahedron(tet_X(i)(1),tet_X(i)(2),tet_X(i)(3),tet_X(i)(4),epsilon))return true;return false;}
template<class T> bool Has_Non_Positive_Volume_Tetrahedron_And_Values(const ARRAY<VECTOR<VECTOR<T,3>, 4> >& tet_X,/*rst*/T& total_volume,const T epsilon=(T)1e-25)
{
    T vol_sum=(T)0;for(int i=1;i<=tet_X.m;i++){
        double volume=Accurate_Tetrahedron_Volume(tet_X(i)(1),tet_X(i)(2),tet_X(i)(3),tet_X(i)(4));
        if(volume<=epsilon){return true;}vol_sum+=(T)volume;}
    total_volume=vol_sum;return false;
}
//////////////////////////////////////////////////////////////////////////
////cubic equation
////solve equation a*x^3+b*x^2+c*x+d=0
////Reference: http://stackoverflow.com/questions/13328676/c-solving-cubic-equations
template<class T> inline T Cubic_Equation(T a,T b,T c,T d)
{
    b/=a;c/=a;d/=a;
    T q=((T)3*c-b*b)/(T)9;
    T r=(-(T)27*d+b*((T)9*c-(T)2*pow<2>(b)))/(T)54;
    T disc=pow<3>(q)+pow<2>(r);
    T term1=b*(T)one_third;
    
    T x1=(T)0;T x2=(T)0;T x3=(T)0;

    if(disc==(T)0){
        T r13=r>(T)0?pow(r,(T)one_third):-pow(-r,(T)one_third);
        x1=-term1+(T)2*r13;
    }
    else if(disc<(T)0){
        T dum1=acos(r/sqrt((T)-1*pow<3>(q)));
        T r13=(T)2*sqrt(-q);
        x1=-term1+r13*cos(dum1*(T)one_third);
        x2=-term1+r13*cos((dum1+(T)2*(T)pi)*(T)one_third);
        x3=-term1+r13*cos((dum1+(T)4*(T)pi)*(T)one_third);
    }
    else{// if(disc>(T)0){
        T s=r+sqrt(disc)>(T)0?pow(r+sqrt(disc),(T)one_third):-pow(-(r+sqrt(disc)),(T)one_third);
        T t=r-sqrt(disc)>(T)0?pow(r-sqrt(disc),(T)one_third):-pow(-(r-sqrt(disc)),(T)one_third);
        x1=-term1+s+t;
    }

    //LOG::cout<<"cubic eq: "<<x1<<", "<<a*pow<3>(x1)+b*pow<2>(x1)+c*x1+d<<std::endl;
    return x1;
}

//////////////////////////////////////////////////////////////////////////
////quadratic equation
////solve equation a*x^2+b*x+c=0
template<class T> inline bool Quadratic_Equation(T a,T b,T c,T& x1, T& x2)
{
    if(a==(T)0){x1=x2=-c/b;return true;}

    b/=a;c/=a;
    T disc=pow<2>(b)-(T)4*c;
    if(disc<(T)0)return false;
    x1=(T).5*(-b+sqrt(disc));x2=(T).5*(-b-sqrt(disc));
    return true;
}

//////////////////////////////////////////////////////////////////////////
////ranges
template<class TV,class TV_INT> bool Has_Vertex_Inside_Regions(ARRAY_VIEW<TV>& X,const TV_INT& vtx,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext)
{
    typedef typename TV::SCALAR T;
    for(int i=1;i<=TV_INT::m;i++){const TV& pos=X(vtx(i));for(int j=1;j<=ranges.m;j++){if(ranges(j).Inside_Half_Open(pos,(T)-ext)){return true;}}} return false;
}
template<class TV,class TV_INT> bool Has_Vertex_Inside_Region(ARRAY_VIEW<TV>& X,const TV_INT& vtx,const RANGE<TV>& range,const typename TV::SCALAR ext)
{
    typedef typename TV::SCALAR T;
    for(int i=1;i<=TV_INT::m;i++){const TV& pos=X(vtx(i));if(range.Inside_Half_Open(pos,(T)-ext)){return true;}} return false;
}
template<class TV> RANGE<TV> Merge_Regions(const ARRAY<RANGE<TV> >& regions)
{
    typedef typename TV::SCALAR T;
    RANGE<TV> merged(TV::Constant_Vector((T)FLT_MAX),TV::Constant_Vector((T)-FLT_MAX));
    for(int i=1;i<=regions.m;i++){merged.Enlarge_To_Include_Box(regions(i));}return merged;
}

//////////////////////////////////////////////////////////////////////////
////particles
template<class TV_INT> void Replace_Vertex_In_Elements(ARRAY<TV_INT>& elements,const ARRAY<int>& v_elements,const VECTOR<int,2> ac,const int v)
{
    for(int i=1;i<=v_elements.m;i++){
        int t=v_elements(i);
        TV_INT replaced_v;bool replaced=Replaced_Two(elements(t),replaced_v,ac(1),ac(2),v);
        if(replaced){elements(t)=replaced_v;}
    }
}

template<class TV_INT> void Remap_Elements(ARRAY<TV_INT>& elements,const ARRAY<int>& remap,bool remove_invalid_elements)
{
    ARRAY<int> elements_to_remove;
    for(int i=1;i<=elements.m;i++){
        const TV_INT vtx=elements(i);
        elements(i)=Element(remap,vtx);
        if(remove_invalid_elements&&elements(i).Contains(0)){
            elements_to_remove.Append(i);
        }
    }
    if(remove_invalid_elements)elements.Remove_Sorted_Indices(elements_to_remove);
}

template<class TV> void Remove_Sorted_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<int>& particles_to_remove,
    CODIMENSIONAL_MESH& mesh,bool remove_invalid_elements=false)
{
    ARRAY<bool> rmv_flag(particles.X.m);rmv_flag.Fill(false);
    for(int i=1;i<=particles_to_remove.m;i++){rmv_flag(particles_to_remove(i))=true;}
    ARRAY<int> remap(particles.X.m);remap.Fill(0);int c=1;
    for(int i=1;i<=particles.X.m;i++)if(!rmv_flag(i)){remap(i)=c++;}

    Remap_Elements(mesh.tetrahedron_mesh.elements,remap,remove_invalid_elements);
    Remap_Elements(mesh.triangle_mesh.elements,remap,remove_invalid_elements);
    Remap_Elements(mesh.segment_mesh.elements,remap,remove_invalid_elements);

    for(int i=1;i<=particles_to_remove.m;i++)particles.array_collection->Add_To_Deletion_List(particles_to_remove(i));
    particles.array_collection->Delete_Elements_On_Deletion_List(true);
}

template<class TV> void Debug_Output_Particle_Guid(const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const int guid,int step)
{
    for(int i=1;i<=particles.X.m;i++){if(particles.guid(i)==guid){
        LOG::cout<<"step "<<step<<" dbg particle "<<guid<<", V: "<<particles.V(i)<<", mass: "<<particles.mass(i)
            <<", dim_mass: "<<particles.dimensional_mass(i)<<", thickness: "<<particles.thickness(i)<<std::endl;}
    }
}

template<class TV> TV Center(ARRAY_VIEW<TV>& X,const VECTOR<int,2>&v){return (typename TV::SCALAR).5*(X(v(1))+X(v(2)));}
template<class TV> TV Center(ARRAY_VIEW<TV>& X,const VECTOR<int,3>&v){return (typename TV::SCALAR)one_third*(X(v(1))+X(v(2))+X(v(3)));}
template<class TV> TV Center(ARRAY_VIEW<TV>& X,const VECTOR<int,4>&v){return (typename TV::SCALAR).25*(X(v(1))+X(v(2))+X(v(3))+X(v(4)));}

template<class TV> TV Total_Momentum(ARRAY_VIEW<typename TV::SCALAR>& mass,ARRAY_VIEW<TV>& V){TV total_momentum=TV();for(int i=1;i<=mass.m;i++)total_momentum+=mass(i)*V(i);return total_momentum;}
template<class TV> TV Sum(ARRAY_VIEW<TV>& F){TV sum=TV();for(int i=1;i<=F.m;i++)sum+=F(i);return sum;}
template<class TV> TV Sum(const ARRAY<TV>& F){TV sum=TV();for(int i=1;i<=F.m;i++)sum+=F(i);return sum;}
template<class T> T Sum(const VECTOR_ND<T>& F){T sum=(T)0;for(int i=1;i<=F.n;i++)sum+=F(i);return sum;}

//////////////////////////////////////////////////////////////////////////
////bspline
template<class T> void Initialize_Bspline_Helper(ARRAY<VECTOR<T,2> >& bspline_time_value,BSPLINE<T,T>*& bspline,int order=1)
{
    ARRAY<T> bspline_time;ARRAY<T> bspline_value;
    for(int i=1;i<=bspline_time_value.m;i++){
        bspline_time.Append(bspline_time_value(i).x);
        bspline_value.Append(bspline_time_value(i).y);
    }
    bspline=new BSPLINE<T,T>(bspline_time,bspline_value,order);
}

template<class T> void Add_Bspline_Time_Value_Helper(const T time,const T value,ARRAY<VECTOR<T,2> >& bspline_time_value)
{
    int n=2;
    if(bspline_time_value.m==0)for(int j=1;j<=n;j++){bspline_time_value.Append(VECTOR<T,2>(time,value));}
    else for(int j=1;j<=n;j++){bspline_time_value.Pop();}////first node
    bspline_time_value.Append(VECTOR<T,2>(time,value));
    for(int j=1;j<=n;j++){bspline_time_value.Append(VECTOR<T,2>(time,value));}////last node
}

template<class T,class T2> void Initialize_Bspline_Helper(ARRAY<PAIR<T,T2> >& bspline_time_value,BSPLINE<T,T2>*& bspline,int order=1)
{
    ARRAY<T> bspline_time;ARRAY<T2> bspline_value;
    for(int i=1;i<=bspline_time_value.m;i++){
        bspline_time.Append(bspline_time_value(i).x);
        bspline_value.Append(bspline_time_value(i).y);
    }
    bspline=new BSPLINE<T,T2>(bspline_time,bspline_value,order);
}

template<class T,class T2> void Add_Bspline_Time_Value_Helper(const T time,const T2 value,ARRAY<PAIR<T,T2> >& bspline_time_value)
{
    int n=2;
    if(bspline_time_value.m==0)for(int j=1;j<=n;j++){bspline_time_value.Append(PAIR<T,T2>(time,value));}
    else for(int j=1;j<=n;j++){bspline_time_value.Pop();}////first node
    bspline_time_value.Append(PAIR<T,T2>(time,value));
    for(int j=1;j<=n;j++){bspline_time_value.Append(PAIR<T,T2>(time,value));}////last node
}

//////////////////////////////////////////////////////////////////////////
////matrix
template<class T> void Sparse_Flat_Matrix_MXN_To_NXN(const SPARSE_MATRIX_FLAT_MXN<T>& mtx_from,SPARSE_MATRIX_FLAT_NXN<T>& mtx_to)
{
    assert(mtx_from.m==mtx_from.n);
    mtx_to.n=mtx_from.n;mtx_to.offsets=mtx_from.offsets;mtx_to.A=mtx_from.A;mtx_to.Initialize_Diagonal_Index();delete mtx_to.C;mtx_to.C=0;
}

template<class T> void Sparse_Flat_Matrix_NXN_To_Dense_Matrix_MXN(const SPARSE_MATRIX_FLAT_NXN<T>& mtx_from,MATRIX_MXN<T>& mtx_to)
{
    assert(mtx_from.n==mtx_to.m&&mtx_from.n==mtx_to.n);
    for(int i=0;i<mtx_to.m*mtx_to.n;i++)mtx_to.x[i]=(T)0;
    for(int i=1;i<=mtx_from.n;i++){for(int c=mtx_from.offsets(i);c<mtx_from.offsets(i+1);c++){T entry=mtx_from.A(c).a;int j=mtx_from.A(c).j;mtx_to(i,j)=entry;}}
}

////matrix checker
template<class T> bool Check_Matrix_Invalid_Index(const SPARSE_MATRIX_FLAT_MXN<T>& G)
{
    bool has_invalid_index=false;
    for(int r=1;r<=G.m;r++){for(int j=G.offsets(r);j<G.offsets(r+1);j++){if(G.A(j).j==0){LOG::cout<<"invalid element in G, row: "<<r<<std::endl;has_invalid_index=true;}}}
    return has_invalid_index;
}

template<class T> void Ensure_Matrix_Symmetry(SPARSE_MATRIX_FLAT_NXN<T>& A)
{
    for(int i=1;i<=A.n;i++){
        for(int j=1;j<=A.n;j++){
            bool ij_present=A.Element_Present(i,j);bool ji_present=A.Element_Present(j,i);
            T A_ij=ij_present?A(i,j):(T)0;T A_ji=ji_present?A(j,i):(T)0;
            if(A_ij!=A_ji&&ij_present){A.Set_Element(i,j,A_ji);}
        }
    }
}

template<class T> bool Check_Matrix_Symmetry(const SPARSE_MATRIX_FLAT_NXN<T>& A)
{
    bool has_antisym_element=false;
    for(int i=1;i<=A.n;i++){
        for(int j=1;j<=A.n;j++){
            T A_ij=A.Element_Present(i,j)?A(i,j):(T)0;T A_ji=A.Element_Present(j,i)?A(j,i):(T)0;
            if(A_ij!=A_ji){LOG::cout<<"antisymmetric matrix element: "<<i<<", "<<j<<": "<<A(i,j)<<"!="<<A(j,i)<<std::endl;}
            has_antisym_element=true;
        }
    }
    return has_antisym_element;
}

template<class T> bool Check_Matrix_Diagonal_Dominance(const SPARSE_MATRIX_FLAT_NXN<T>& A)
{
    int number_of_non_diagonally_dominant_rows=0;
    for(int r=1;r<=A.n;r++){
        int row_first_element=A.offsets(r);
        int column_number=A.offsets(r+1)-row_first_element;
        T diagonal=(T)0;T off_diagonal=(T)0;
        for(int c=row_first_element;c<row_first_element+column_number;c++){
            if(A.A(c).j==r)diagonal+=abs(A.A(c).a);
            else off_diagonal+=abs(A.A(c).a);}
        if(off_diagonal>diagonal) number_of_non_diagonally_dominant_rows++;}
    if(number_of_non_diagonally_dominant_rows>0)
        LOG::cout<<"Non diagonally dominant rows: "<<number_of_non_diagonally_dominant_rows<<std::endl;
    return number_of_non_diagonally_dominant_rows==0;
}

template<class T> bool Check_Matrix_Conservativity(const SPARSE_MATRIX_FLAT_MXN<T>& A,bool check_row=false,const T zero=(T)1e-16)
{
    bool conservative=true;
    if(check_row){
        for(int i=1;i<=A.m;i++){
            T row_sum=(T)0;for(int j=1;j<=A.m;j++){T A_ij=A.Element_Present(i,j)?A(i,j):(T)0;row_sum+=A_ij;}
            if(abs(row_sum)>zero){LOG::cout<<"Non-conservative row "<<i<<": "<<row_sum<<std::endl;conservative=false;}
        }
    }
    else{
        for(int j=1;j<=A.n;j++){
            T col_sum=(T)0;for(int i=1;i<=A.m;i++){T A_ij=A.Element_Present(i,j)?A(i,j):(T)0;col_sum+=A_ij;}
            if(abs(col_sum)>zero){LOG::cout<<"Non-conservative column "<<j<<": "<<col_sum<<std::endl;conservative=false;}
        }
    }

    return conservative;
}
template<class T> bool Check_Matrix_Conservativity(const SPARSE_MATRIX_FLAT_NXN<T>& A,bool check_row=false,const T zero=(T)1e-16)
{SPARSE_MATRIX_FLAT_MXN<T> B=A;return Check_Matrix_Conservativity(B,check_row,zero);}

//////////////////////////////////////////////////////////////////////////
////floats
template<class T> inline bool Is_Inf(T scalar)
{
#ifdef WIN32
    return _finite(scalar)==0?true:false;
#elif defined(__linux__) || defined(__CYGWIN__) || defined(__APPLE__)
    return isinf(scalar);
#else
    return false;
#endif
}

template<class T> inline bool Is_Inf(const VECTOR<T,1>& v)
{return Is_Inf(v.x);}

template<class T> inline bool Is_Inf(const VECTOR<T,2>& v)
{return Is_Inf(v.x)||Is_Inf(v.y);}

template<class T> inline bool Is_Inf(const VECTOR<T,3>& v)
{return Is_Inf(v.x)||Is_Inf(v.y)||Is_Inf(v.z);}

template<class T> inline bool Is_Inf(const VECTOR<T,4>& v)
{return Is_Inf(v(1))||Is_Inf(v(2))||Is_Inf(v(3))||Is_Inf(v(4));}

template<class T> inline bool Is_Inf(const VECTOR<T,5>& v)
{return Is_Inf(v(1))||Is_Inf(v(2))||Is_Inf(v(3))||Is_Inf(v(4))||Is_Inf(v(5));}

template<class T> inline bool Is_Valid_Number(T scalar)
{return !Is_NaN(scalar)&&!Is_Inf(scalar);}

}//namespace Codimensional
}//namespace PhysBAM
#endif
