//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Arrays_Computations/SORT.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include <PhysBAM_Tools/Data_Structures/STACK.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"
#include "../Meshing/CODIMENSIONAL_MESH_HASHTABLE.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

using namespace PhysBAM;
using namespace Codimensional;

///////////////////////////////////////////////////////////////////////////////
////Hashtable_Comparator Helper Function
///////////////////////////////////////////////////////////////////////////////
bool Hashtable_Comparator(const HASHTABLE<int>& lhs,const HASHTABLE<int>& rhs)
{
    int lhs_min=-1,rhs_min=-1;
    for(HASHTABLE_ITERATOR<int> iter(lhs);iter.Valid();iter.Next()){//is this necessary? or is min element the first in the hashtable?
        if(lhs_min==-1||iter.Key()<lhs_min)lhs_min=iter.Key();
    }
    for(HASHTABLE_ITERATOR<int> iter(rhs);iter.Valid();iter.Next()){
        if(rhs_min==-1||iter.Key()<rhs_min)rhs_min=iter.Key();
    }
    return lhs_min<rhs_min;
}

PhysBAM::CODIMENSIONAL_MESH::CODIMENSIONAL_MESH(SEGMENT_MESH* segment_mesh_input/*=0*/,TRIANGLE_MESH* triangle_mesh_input/*=0*/,TETRAHEDRON_MESH* tetrahedron_mesh_input/*=0*/)
    :number_nodes(0),number_of_regions(0),number_of_regions_in_high_dim(0),segment_mesh(segment_mesh_input?*segment_mesh_input:*(new SEGMENT_MESH())),triangle_mesh(triangle_mesh_input?(*triangle_mesh_input):*(new TRIANGLE_MESH())),
    tetrahedron_mesh(tetrahedron_mesh_input?(*tetrahedron_mesh_input):*(new TETRAHEDRON_MESH())),neighbor_nodes(0)/*,incident_elements(0)*/{}

PhysBAM::CODIMENSIONAL_MESH::CODIMENSIONAL_MESH(const CODIMENSIONAL_MESH& copy)
    :number_nodes(0),number_of_regions(0),number_of_regions_in_high_dim(0),segment_mesh(*(new SEGMENT_MESH())),triangle_mesh(*(new TRIANGLE_MESH())),
    tetrahedron_mesh(*(new TETRAHEDRON_MESH())),neighbor_nodes(0)/*,incident_elements(0)*/{Initialize_Mesh(copy);}

void PhysBAM::CODIMENSIONAL_MESH::Initialize_Mesh(const CODIMENSIONAL_MESH& copy)
{
    number_nodes=copy.number_nodes;
    segment_mesh.Initialize_Mesh(copy.segment_mesh);
    triangle_mesh.Initialize_Mesh(copy.triangle_mesh);
    tetrahedron_mesh.Initialize_Mesh(copy.tetrahedron_mesh);
    rim_segment_mesh.Initialize_Mesh(copy.rim_segment_mesh);

    number_of_regions=copy.number_of_regions;
    node_connectivities=copy.node_connectivities;
    node_connectivities_in_high_dim=copy.node_connectivities_in_high_dim;
    node_types=copy.node_types;

    Update_Auxiliary_Structures();
}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH::Update_Node_Connectivities(MPI_CODIMENSIONAL<TV>* mpi_codim)
{
    //int region_num;
    //if(mpi_codim){
    //    region_num=Update_Connectivities_Across_Mpi_Nodes(mpi_codim);
    //}
    //else{ //no mpi
    //    node_connectivities.Resize(number_nodes);node_connectivities.Fill(-1);
    //    int number_checked=0;
    //    int region=1;
    //    for(int i=1;i<=number_nodes;i++)if((*neighbor_nodes)(i).m==0){node_connectivities(i)=0;number_checked++;}

    //    int first=0;for(int i=1;i<=number_nodes;i++)if(node_connectivities(i)==-1){first=i;break;}
    //    STACK<int> front;if(first!=0){front.Push(first);node_connectivities(first)=region;}

    //    while(!front.Empty()){
    //        int p=front.Pop();number_checked++;
    //        for(int i=1;i<=(*neighbor_nodes)(p).m;i++){
    //            int nb_p=(*neighbor_nodes)(p)(i);
    //            if(node_connectivities(nb_p)==-1){front.Push(nb_p);node_connectivities(nb_p)=region;}
    //        }
    //        if(front.Empty()){
    //            if(number_checked==number_nodes)break;
    //            int old_first=first;for(int i=old_first+1;i<=number_nodes;i++)if(node_connectivities(i)==-1){first=i;break;}
    //            if(first!=old_first){front.Push(first);region++;node_connectivities(first)=region;}
    //        }
    //    }
    //    region_num=region;
    //}

    //return region_num;

    ////assuming has a single region
    node_connectivities.Resize(number_nodes);node_connectivities.Fill(1);
    for(int i=1;i<=number_nodes;i++){if(Is_Isolated(node_types(i)))node_connectivities(i)=0;}
    return 1;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH::Update_Connectivities_Across_Mpi_Nodes(MPI_CODIMENSIONAL<TV>* mpi_codim)
{
    ////1. compute local regions
    ARRAY<int> regions;
    node_connectivities.Resize(number_nodes);node_connectivities.Fill(-1);
    int number_checked=0;
    int region=mpi_codim->first_region_number;
    for(int i=1;i<=number_nodes;i++)if((*neighbor_nodes)(i).m==0){node_connectivities(i)=0;number_checked++;}

    int first=0;for(int i=1;i<=number_nodes;i++)if(node_connectivities(i)==-1){first=i;break;}
    STACK<int> front;if(first!=0){front.Push(first);node_connectivities(first)=region;regions.Append(region);}

    while(!front.Empty()){
        int p=front.Pop();number_checked++;
        for(int i=1;i<=(*neighbor_nodes)(p).m;i++){
            int nb_p=(*neighbor_nodes)(p)(i);
            if(node_connectivities(nb_p)==-1){front.Push(nb_p);node_connectivities(nb_p)=region;}
        }
        if(front.Empty()){
            if(number_checked==number_nodes)break;
            int old_first=first;for(int i=old_first+1;i<=number_nodes;i++)if(node_connectivities(i)==-1){first=i;break;}
            if(first!=old_first){front.Push(first);region++;node_connectivities(first)=region;regions.Append(region);}
        }
    }

    ////2. exchange local regions
    ARRAY<ARRAY<int> > recv_regions;
    mpi_codim->Mpi_Broadcast_Arrays(regions,recv_regions);
    HASHTABLE<int> all_regions;
    for(int i=1;i<=regions.m;i++){PHYSBAM_ASSERT(!all_regions.Contains(regions(i)));all_regions.Insert(regions(i));}
    for(int i=1;i<=recv_regions.m;i++){
        for(int j=1;j<=recv_regions(i).m;j++){
            PHYSBAM_ASSERT(!all_regions.Contains(recv_regions(i)(j)));
            all_regions.Insert(recv_regions(i)(j));
        }
    }

    if(all_regions.Size()==0){//early exit if no regions--will this ever happen?
        return 0;
    }

    ////3. exchange info about boundary/ghost regions
    ARRAY_VIEW<int> node_connectivities_view(node_connectivities);
    mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(node_connectivities_view);

    ////4. find neighboring regions
    ARRAY<PAIR<int,int> > correspondences;
    Get_Region_Correspondences(correspondences);
    //flatten array of corresponding pairs
    ARRAY<int> pairs;
    for(int i=1;i<=correspondences.m;i++){pairs.Append(correspondences(i).x);pairs.Append(correspondences(i).y);}

    ////5. broadcast cross-node region connectivities
    ARRAY<ARRAY<int> > recv_pairs;
    mpi_codim->Mpi_Broadcast_Arrays(pairs,recv_pairs);

    ////6. concatenate received pairs and local pairs
    for(int i=1;i<=recv_pairs.m;i++){
        for(int j=1;j<=recv_pairs(i).m;j++){
            pairs.Append(recv_pairs(i)(j));
        }
    }

    ////7. union all these sets together
    ARRAY<HASHTABLE<int> > sets;
    //insert all regions and pairs
    for(HASHTABLE_ITERATOR<int> iter(all_regions);iter.Valid();iter.Next()){
        sets.Append(HASHTABLE<int>());
        sets(sets.m).Insert(iter.Key());
    }
    for(int i=1;i<=pairs.m;i+=2){
        sets.Append(HASHTABLE<int>());
        sets(sets.m).Insert(pairs(i));
        sets(sets.m).Insert(pairs(i+1));
    }

    //union
    int idx=1;
    while(idx<=sets.m){
        bool merge=false;
        for(int j=sets.m;j>idx;j--){
            if(Common_Hashtable_Entry(sets(idx),sets(j))){
                for(HASHTABLE_ITERATOR<int> iter(sets(j));iter.Valid();iter.Next()){
                    if(!sets(idx).Contains(iter.Key()))sets(idx).Insert(iter.Key());
                }
                sets.Remove_Index(j);
                merge=true;
            }
        }
        if(!merge)idx++;
    }

    ////8. re-compute region numbers
    Sort(sets,Hashtable_Comparator);

    HASHTABLE<int,int> local_region_to_final_region;
    for(int i=1;i<=sets.m;i++){
        for(HASHTABLE_ITERATOR<int> iter(sets(i));iter.Valid();iter.Next()){
            PHYSBAM_ASSERT(!local_region_to_final_region.Contains(iter.Key()));
            local_region_to_final_region.Insert(iter.Key(),i);
        }
    }
    for(int i=1;i<=node_connectivities.m;i++){
        if(node_connectivities(i)>0){
            node_connectivities(i)=local_region_to_final_region.Get(node_connectivities(i));
        }
    }
    return sets.m;
}


bool PhysBAM::CODIMENSIONAL_MESH::Common_Hashtable_Entry(const HASHTABLE<int>& lhs,const HASHTABLE<int>& rhs)
{
    bool common_entry=false;
    for(HASHTABLE_ITERATOR<int> iter(rhs);iter.Valid();iter.Next()){
        if(lhs.Contains(iter.Key())){common_entry=true;break;}
    }
    return common_entry;
}

void PhysBAM::CODIMENSIONAL_MESH::Get_Region_Correspondences(ARRAY<PAIR<int,int> >& correspondences)
{
    ARRAY<bool> visited(node_connectivities.m);visited.Fill(false);
    int count=0;
    int first=1;
    while(count!=node_connectivities.m){
        for(int i=first;i<=node_connectivities.m;i++)if(!visited(i)){first=i;break;}
        visited(first)=true;
        STACK<int> front;
        front.Push(first);
        while(!front.Empty()){
            int p=front.Pop();count++;
            for(int i=1;i<=(*neighbor_nodes)(p).m;i++){
                int nb_p=(*neighbor_nodes)(p)(i);
                if(!visited(nb_p)){
                    visited(nb_p)=true;
                    front.Push(nb_p);
                    if(node_connectivities(nb_p)!=node_connectivities(p)){
                        int lower=min(node_connectivities(nb_p),node_connectivities(p)),upper=max(node_connectivities(nb_p),node_connectivities(p));
                        correspondences.Append_Unique(PAIR<int,int>(lower,upper));
                    }
                }
            }
        }
    }

}

void PhysBAM::CODIMENSIONAL_MESH::Update_Segment_Boundary_Mesh()
{
    segment_mesh.Initialize_Incident_Elements();
    segment_mesh.Initialize_Boundary_Mesh();
    (*segment_mesh.boundary_mesh).Initialize_Incident_Elements();
    (*segment_mesh.boundary_mesh).Initialize_Neighbor_Nodes();
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Triangle_Boundary_Mesh()
{
    triangle_mesh.Initialize_Incident_Elements();
    triangle_mesh.Initialize_Boundary_Mesh();
    (*triangle_mesh.boundary_mesh).Initialize_Incident_Elements();
    (*triangle_mesh.boundary_mesh).Initialize_Neighbor_Nodes();
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Tetrahedron_Boundary_Mesh()
{
    tetrahedron_mesh.Initialize_Incident_Elements();
    tetrahedron_mesh.Initialize_Boundary_Mesh();
    (*tetrahedron_mesh.boundary_mesh).Initialize_Incident_Elements();
    (*tetrahedron_mesh.boundary_mesh).Initialize_Neighbor_Nodes();
}

//void PhysBAM::CODIMENSIONAL_MESH::Update_Rim_Segment_Mesh()
//{
//    ////moved to mesher, now update incident elemetns only
//    //rim_segment_mesh.elements.Remove_All();
//    //for(int i=1;i<=(*triangle_mesh.boundary_mesh).elements.m;i++){
//    //    const TV_INT2& v=(*triangle_mesh.boundary_mesh).elements(i);
//    //    if(!Is_Boundary_Edge_On_Tetrahedron_Mesh(v)){rim_segment_mesh.elements.Append(v);}
//    //}
//
//    rim_segment_mesh.Initialize_Incident_Elements();
//}

void PhysBAM::CODIMENSIONAL_MESH::Update_Node_Types(ARRAY<int>& node_types)
{
    node_types.Resize(number_nodes);node_types.Fill(0);
    for(int i=1;i<=segment_mesh.elements.m;i++){
        for(int j=1;j<=2;j++)Set_On_Segment_Mesh(node_types(segment_mesh.elements(i)(j)));
    }
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        for(int j=1;j<=3;j++)Set_On_Triangle_Mesh(node_types(triangle_mesh.elements(i)(j)));
    }
    for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
        for(int j=1;j<=4;j++){Set_On_Tetrahedron_Mesh(node_types(tetrahedron_mesh.elements(i)(j)));}
    }

    for(int i=1;i<=(*segment_mesh.boundary_mesh).elements.m;i++){
        Set_On_Segment_Mesh_Boundary(node_types((*segment_mesh.boundary_mesh).elements(i)(1)));
    }
	
	for(int i=1;i<=(*triangle_mesh.boundary_mesh).elements.m;i++){
		for(int j=1;j<=2;j++)Set_On_Triangle_Mesh_Boundary(node_types((*triangle_mesh.boundary_mesh).elements(i)(j)));
	}

	for(int i=1;i<=(*tetrahedron_mesh.boundary_mesh).elements.m;i++){
		for(int j=1;j<=3;j++)Set_On_Tetrahedron_Mesh_Boundary(node_types((*tetrahedron_mesh.boundary_mesh).elements(i)(j)));
	}
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Number_Nodes(int number_nodes_input)
{
    number_nodes=number_nodes_input;
    segment_mesh.number_nodes=number_nodes;
    triangle_mesh.number_nodes=number_nodes;
    tetrahedron_mesh.number_nodes=number_nodes;
    rim_segment_mesh.number_nodes=number_nodes;
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Auxiliary_Structures()
{
    Update_Segment_Auxiliary_Structures();
    Update_Triangle_Auxiliary_Structures();
    Update_Tetrahedron_Auxiliary_Structures();
    Update_Segment_Boundary_Mesh();
    Update_Triangle_Boundary_Mesh();
    Update_Tetrahedron_Boundary_Mesh();
    Update_Neighbor_Nodes();
    //Update_Rim_Segment_Mesh();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH::Update_States(int number_nodes_input,MPI_CODIMENSIONAL<TV>* mpi_codim/*=0*/)
{
    Update_Number_Nodes(number_nodes_input);
    Update_Auxiliary_Structures();
    Update_Node_Types(node_types);
    number_of_regions=Update_Node_Connectivities(mpi_codim); ////needs node_types
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Segment_Auxiliary_Structures()
{
    segment_mesh.Initialize_Neighbor_Nodes();
    segment_mesh.Initialize_Incident_Elements();
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Triangle_Auxiliary_Structures()
{
    triangle_mesh.Initialize_Neighbor_Nodes();
    triangle_mesh.Initialize_Incident_Elements();
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Tetrahedron_Auxiliary_Structures()
{
    tetrahedron_mesh.Initialize_Neighbor_Nodes();
    tetrahedron_mesh.Initialize_Incident_Elements();
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Edge_On_Triangle_Mesh(const VECTOR<int,2>& v) const 
{
    for(int i=1;i<=(*triangle_mesh.boundary_mesh->neighbor_nodes)(v(1)).m;i++){
        if((*triangle_mesh.boundary_mesh->neighbor_nodes)(v(1))(i)==v(2))return true;
    }
    return false;
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Edge_On_Tetrahedron_Mesh(const VECTOR<int,2>& v) const 
{
    for(int i=1;i<=(*tetrahedron_mesh.boundary_mesh->neighbor_nodes)(v(1)).m;i++){
        int nb_v=(*tetrahedron_mesh.boundary_mesh->neighbor_nodes)(v(1))(i);
        if(nb_v==v(2))return true;
    }
    return false;
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Face_On_Tetrahedron_Mesh(const TV_INT3& v) const
{
    VECTOR<int,2> v2v3=v.Remove_Index(1);
    for(int i=1;i<=(*tetrahedron_mesh.boundary_mesh->incident_elements)(v(1)).m;i++){
        VECTOR<int,3> nb_v=tetrahedron_mesh.boundary_mesh->elements((*tetrahedron_mesh.boundary_mesh->incident_elements)(v(1))(i));
        VECTOR<int,2> nb_v2v3=nb_v.Remove_Index(nb_v.Find(v(1)));
        if(nb_v2v3.Contains(v2v3(1))&&nb_v2v3.Contains(v2v3(2)))return true;
    }
    return false;
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Particle_On_Triangle_Mesh(const int p) const 
{
    //return Is_On_Triangle_Mesh_Boundary(node_types(p))&&!Is_On_Tetrahedron_Mesh_Boundary(node_types(p));
    return (rim_segment_mesh.incident_elements&&(*rim_segment_mesh.incident_elements)(p).m>0);
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Rim_Particle(const int p) const 
{
    return Is_Boundary_Particle_On_Triangle_Mesh(p)||(Is_On_Segment_Mesh(node_types(p))&&!Is_On_Triangle_Mesh(node_types(p))&&!Is_On_Tetrahedron_Mesh(node_types(p)));
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Particle_On_Segment_Mesh(const int p) const 
{
    return Is_On_Segment_Mesh_Boundary(node_types(p))&&!Is_On_Triangle_Mesh(node_types(p))&&!Is_On_Tetrahedron_Mesh(node_types(p));
}

bool PhysBAM::CODIMENSIONAL_MESH::Is_Boundary_Particle_On_Tetrahedron_Mesh(const int p) const
{
    return Is_On_Tetrahedron_Mesh_Boundary(node_types(p))&&!Is_On_Triangle_Mesh(node_types(p))&&!Is_On_Segment_Mesh(node_types(p));
}

void PhysBAM::CODIMENSIONAL_MESH::Update_Neighbor_Nodes()
{
    if(neighbor_nodes) delete neighbor_nodes;neighbor_nodes=new ARRAY<ARRAY<int> >(number_nodes);
    for(int t=1;t<=segment_mesh.elements.m;t++){
        const TV_INT2& e=segment_mesh.elements(t);
        for(int i=1;i<=2;i++)for(int j=1;j<=2;j++)if(i!=j)
            (*neighbor_nodes)(e(i)).Append_Unique(e(j));
    }
    for(int t=1;t<=triangle_mesh.elements.m;t++){
        const TV_INT3& e=triangle_mesh.elements(t);
        for(int i=1;i<=3;i++)for(int j=1;j<=3;j++)if(i!=j)
            (*neighbor_nodes)(e(i)).Append_Unique(e(j));
    }
    for(int t=1;t<=tetrahedron_mesh.elements.m;t++){
        const TV_INT4& e=tetrahedron_mesh.elements(t);
        for(int i=1;i<=4;i++)for(int j=1;j<=4;j++)if(i!=j)
            (*neighbor_nodes)(e(i)).Append_Unique(e(j));
    }
    for(int p=1;p<=number_nodes;p++) (*neighbor_nodes)(p).Compact();
}

#define INSTANTIATION_HELPER(T,d)                                        \
    template int PhysBAM::CODIMENSIONAL_MESH::Update_Node_Connectivities<VECTOR<T,d> >(MPI_CODIMENSIONAL<VECTOR<T,d> >*); \
    template void PhysBAM::CODIMENSIONAL_MESH::Update_States<VECTOR<T,d> >(int,MPI_CODIMENSIONAL<VECTOR<T,d> >*);

INSTANTIATION_HELPER(float,3)
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
INSTANTIATION_HELPER(double,3)
#endif
