//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESH_CHECKER.h"
#include "../Meshing/CODIMENSIONAL_MESH_HASHTABLE.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D_DEFINITIONS.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D_ATOMIC_OPERATION_DEFINITIONS.h"
#include "../Intersection/CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION.h"
#include "../Intersection/CODIMENSIONAL_INTERSECTION.h"
#include "../Viscoelastic/SOLID_FLUID_CONTACT_CODIMENSIONAL.h"
#include "../Viscoelastic/FIBER.h"

using namespace PhysBAM;
using namespace Codimensional;

template<class T_input>
PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::CODIMENSIONAL_MESHING_3D(CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :BASE(mesh_object_input,mpi_codim_input)
{
    Initialize_Mpi(mpi_codim_input);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Initialize_Sub_Regions()
{
    sub_regions.Remove_All();
    if(mpi_codim){
        //sub_regions.Resize(1);
        //sub_regions(1).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,0,0)));
        //LOG::cout<<"init sub region here "<<sub_regions(1).m<<": "<<sub_regions(1)<<std::endl;

        sub_regions.Resize(8);
        sub_regions(1).Append(mpi_codim->Boundary_Shrink_Region(TV_INT()));

        ARRAY<RANGE<TV> > main_regions;
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT()));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,0,0)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,1,0)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,0,1)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,1,1)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,0,1)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,1,1)));
        main_regions.Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,1,1)));
        sub_regions(1).Append(Merge_Regions(main_regions));

        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,0,0)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,-1,0)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,0,-1)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,-1,-1)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,0,-1)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,-1,0)));
        sub_regions(2).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,-1,-1)));

        sub_regions(3).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,0,1)));
        sub_regions(3).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,1,0)));
        sub_regions(3).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,1,1)));

        sub_regions(4).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,1,-1)));
        sub_regions(4).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,0,-1)));
        sub_regions(4).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,1,-1)));

        sub_regions(5).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,-1,0)));
        sub_regions(5).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(0,-1,1)));
        sub_regions(5).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,-1,1)));

        sub_regions(6).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,-1,1)));
        sub_regions(7).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(-1,1,-1)));
        sub_regions(8).Append(mpi_codim->Boundary_Shrink_Region(TV_INT(1,-1,-1)));
    }
    else{
        sub_regions.Resize(1);
        sub_regions(1).Append(RANGE<TV>(TV::Constant_Vector((T)-FLT_MAX),TV::Constant_Vector((T)FLT_MAX)));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Mesh()
{
    if(use_topology_repairing) Update_Mesh_Using_Topology_Repair();
    else PHYSBAM_FATAL_ERROR("CODIMENSIONAL_MESHING_3D::Update_Mesh() not implemented");
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Mesh_Using_Topology_Repair()
{
    Update_Mesh_Topology();
    Update_Mesh_States();
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Mesh_Topology()
{
    //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
    //LOG::cout<<"checked mesh before, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    if(preserve_thickness_in_remeshing){
        for(int i=1;i<=particles.X.m;i++){particles.thickness(i)=Discrete_Particle_Thickness(i,has_rim);}
        if(mpi_codim){mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.thickness);}
    }

    if(use_boundary_vertex_snap){
        Boundary_Vertex_Snap();
        Check_Mass();
        //LOG::cout<<"checked mesh boundary vtx snap, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    }
    if(use_codim_vertex_snap){
        Codimensional_Vertex_Snap();
        Check_Mass();
        //LOG::cout<<"checked mesh codim vtx snap, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    }

    if(use_tetrahedron_mesh){
        Tetrahedron_Mesh_Edge_Flip();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh edge flip, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Tetrahedron_Mesh_Face_Flip();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh face flip, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Tetrahedron_Mesh_Edge_Split();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh edge split, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        //Tetrahedron_Mesh_Face_Split();    
        //Tetrahedron_Mesh_Edge_Collapse();
        for(int i=1;i<=2;i++)Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh tet clps, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Tetrahedron_Mesh_Edge_Flip();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh edge flip2, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Tetrahedron_Mesh_Face_Flip();
        //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
        //Check_Mass();
        //LOG::cout<<"checked mesh face flip2, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

        //if(use_crumple_merge)Tetrahedron_Mesh_Boundary_Crumple_Face_Merge();
        //Tetrahedron_Mesh_Boundary_Thin_Element_Remove();
    }

    if(use_triangle_mesh){
        if(use_contact_aware_triangle_remeshing)Triangle_Mesh_Contact_Edge_Split();
        Triangle_Mesh_Edge_Split();
        //Check_Mass();
        //LOG::cout<<"checked mesh tri edge split, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Triangle_Mesh_Edge_Flip();
        //Check_Mass();
        //LOG::cout<<"checked mesh tri edge flip, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Triangle_Mesh_Edge_Collapse();
        //Check_Mass();
        //LOG::cout<<"checked mesh tri edge collapse, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        Triangle_Mesh_Edge_Flip();
        //Check_Mass();
        //LOG::cout<<"checked mesh tri edge flip, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

        //if(use_crumple_based_dimension_lifting&&use_tetrahedron_mesh){
        //    Triangle_Mesh_Dimension_Lift_Based_On_Crumple();
        //}
    }
    else triangle_mesh.elements.Remove_All();

    if(use_segment_mesh){
        Segment_Mesh_Edge_Split();
        Segment_Mesh_Edge_Collapse();
        if(use_mesh_edge_connect){Segment_Mesh_Edge_Connect();}
    }
    else segment_mesh.elements.Remove_All();

    if(use_remove_codim_elements_in_the_air){
        Remove_Codim_Elements_In_The_Air();
        //Check_Mass();
        //LOG::cout<<"checked mesh rmv codim ele, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    }
    
    if(use_triangle_mesh&&use_break_thin_triangles){
        if(use_constrained_thin_triangle_break)Break_Thin_Triangles_Constrained_Boundary();
        else Break_Thin_Triangles();
        //Check_Mass();
        //LOG::cout<<"checked mesh break thin tri, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    }
    if(use_segment_mesh&&use_break_thin_segments){
        if(use_constrained_thin_segment_break)Break_Thin_Segments_Constrained_Boundary();
        else Break_Thin_Segments();
        //Check_Mass();
        //LOG::cout<<"checked mesh break thin seg, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    }

    if(use_remove_invalid_elements){Remove_Invalid_Elements();}
    //LOG::cout<<"checked mesh rmv invalid ele, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

    //if(use_remove_low_dim_elements_on_tank_bottom){Remove_Low_Dim_Elements_On_Tank_Bottom_Threaded();}
    //if(use_elements_merging){Elements_Merge();}
    //Check_Invalid_Vertex(tetrahedron_mesh.elements,1,particles.X.m);
    //Check_Duplicate_Element(tetrahedron_mesh.elements);
    //Check_Dupliciate_Vertex(tetrahedron_mesh.elements);
    //Check_Nonmanifold_Face(tetrahedron_mesh.elements);

    //Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
    //LOG::cout<<"checked mesh after, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Mesh_States()
{
    ////particle positions and indices would not change after this
    if(mpi_codim)mpi_codim->Update_Boundary_And_Ghost_Particle_Indices();
    mesh.Update_States(particles.X.m,mpi_codim);
    Update_Rim_Segment_Mesh();
    //LOG::cout<<"checked mesh update_mesh_states, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

    Update_Particle_Attributes_After_Having_Mesh_Topology();

    //LOG::cout<<"checked mesh update p attri aft having topo, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

    //bool valid_mesh_after=Check_Tetrahedron_Mesh()&&Check_Triangle_Mesh()&&Check_Segment_Mesh();
    //LOG::cout<<"checked mesh after, N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Check_Tetrahedron_Mesh()
{
    bool check_mesh=true;
    bool check=Check_Duplicate_Element(tetrahedron_mesh.elements,true,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Invalid_Vertex(tetrahedron_mesh.elements,1,particles.X.m,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Dupliciate_Vertex(tetrahedron_mesh.elements,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Nonmanifold_Face(tetrahedron_mesh.elements,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Non_Positive_Volume_Tetrahedron(tetrahedron_mesh.elements,particles.X,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    return check_mesh;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Check_Triangle_Mesh()
{
    bool check_mesh=true;
    bool check=Check_Duplicate_Element(triangle_mesh.elements,true,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Invalid_Vertex(triangle_mesh.elements,1,particles.X.m,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Dupliciate_Vertex(triangle_mesh.elements,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    return check_mesh;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Check_Segment_Mesh()
{
    bool check_mesh=true;
    bool check=Check_Duplicate_Element(segment_mesh.elements,false,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Invalid_Vertex(segment_mesh.elements,1,particles.X.m,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    check=Check_Dupliciate_Vertex(segment_mesh.elements,&particles.guid,(mpi_codim?mpi_codim->node_id:0));check_mesh=check_mesh&&check;
    return check_mesh;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Split()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Edge_Split_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Face_Split()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Face_Split_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Flip()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Edge_Flip_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Face_Flip()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Face_Flip_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Collapse()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Edge_Collapse_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse()
{
    for(int i=1;i<=sub_regions.m;i++){
        Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Split()
{
    for(int i=1;i<=sub_regions.m;i++){
        Triangle_Mesh_Edge_Split_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Flip()
{
    for(int i=1;i<=sub_regions.m;i++){
        Triangle_Mesh_Edge_Flip_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Collapse()
{
    for(int i=1;i<=sub_regions.m;i++){
        Triangle_Mesh_Edge_Collapse_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Split()
{
    for(int i=1;i<=sub_regions.m;i++){
        Segment_Mesh_Edge_Split_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Collapse()
{
    for(int i=1;i<=sub_regions.m;i++){
        Segment_Mesh_Edge_Collapse_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Boundary_Vertex_Snap()
{
    for(int i=1;i<=sub_regions.m;i++){
        Boundary_Vertex_Snap_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Codimensional_Vertex_Snap()
{
    for(int i=1;i<=sub_regions.m;i++){
        Codimensional_Vertex_Snap_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Connect()
{
    for(int i=1;i<=sub_regions.m;i++){
        Segment_Mesh_Edge_Connect_Subregion(sub_regions(i));
    }
}
template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Contact_Edge_Split()
{
    for(int i=1;i<=sub_regions.m;i++){
        Triangle_Mesh_Edge_Split_Subregion(sub_regions(i),/*split contact boundary edge*/true);
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Remove_Particles_And_Elements(ARRAY<int>* p_to_rmv,ARRAY<int>* tet_to_rmv,ARRAY<int>* tri_to_rmv,ARRAY<int>* seg_to_rmv)
{
    if(tet_to_rmv){Sort(*tet_to_rmv);tetrahedron_mesh.elements.Remove_Sorted_Indices(*tet_to_rmv);}
    if(tri_to_rmv){Sort(*tri_to_rmv);triangle_mesh.elements.Remove_Sorted_Indices(*tri_to_rmv);}
    if(seg_to_rmv){Sort(*seg_to_rmv);segment_mesh.elements.Remove_Sorted_Indices(*seg_to_rmv);}
    if(p_to_rmv){Sort(*p_to_rmv);Remove_Sorted_Particles(particles,*p_to_rmv,mesh,false);}
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Length_Scale(const TV_INT4& v)
{
    return use_adaptive_tetrahedron_remeshing?(T).25*(particles.length_scale(v(1))+particles.length_scale(v(2))+particles.length_scale(v(3))+particles.length_scale(v(4))):length_scale;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Edge_Length_Scale(const int a,const int c)
{
    return use_adaptive_tetrahedron_remeshing?(T).5*(particles.length_scale(a)+particles.length_scale(c)):length_scale;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Edge_Length_Scale(const int a,const int c)
{
    T particle_length_scale=use_adaptive_triangle_remeshing?(T).5*(particles.length_scale(a)+particles.length_scale(c)):length_scale;
    if(use_anisotropic_triangle_remeshing&&(particles.meshing_direction(a)!=TV()&&particles.meshing_direction(c)!=TV())){
        T length_scale_coef=clamp((particle_length_scale-adaptive_length_scale_min)/((T).3*length_scale-adaptive_length_scale_min),(T)0,(T)1);
        T min_coef=(T).25;T anisotropic_coef=min_coef+((T)1-min_coef)*length_scale_coef;
        TV edge_n=(particles.X(a)-particles.X(c)).Normalized();
        TV edge_meshing_n=(particles.meshing_direction(a)+particles.meshing_direction(c)).Normalized();
        T projected=anisotropic_coef+((T)1-anisotropic_coef)*pow<2>(abs(TV::Dot_Product(edge_n,edge_meshing_n)));
        particle_length_scale*=projected;
        //LOG::cout<<"anisotropic: "<<projected<<", "<<particle_length_scale<<std::endl;
    }
    return particle_length_scale;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Edge_Length_Scale(const int a,const int c)
{
    return use_adaptive_segment_remeshing?(T).5*(particles.length_scale(a)+particles.length_scale(c)):length_scale;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Incident_Tetrahedron_Volume(const int v,const HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable)
{
    T total_tet_vol=(T)0;
    const ARRAY<int>* v_tets=vertex_tetrahedron_hashtable.Get_Pointer(v);
    if(v_tets)for(int i=1;i<=v_tets->m;i++){
        const TV_INT4& vtx=tetrahedron_mesh.elements((*v_tets)(i));
        total_tet_vol+=TETRAHEDRON<T>::Volume(particles.X(vtx(1)),particles.X(vtx(2)),particles.X(vtx(3)),particles.X(vtx(4)));
    }
    return total_tet_vol;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Incident_Triangle_Area(const int v,const HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable)
{
    T total_tri_area=(T)0;
    const ARRAY<int>* v_tris=vertex_triangle_hashtable.Get_Pointer(v);
    if(v_tris)for(int i=1;i<=v_tris->m;i++){
        const TV_INT3& vtx=triangle_mesh.elements((*v_tris)(i));
        total_tri_area+=TRIANGLE_3D<T>::Area(particles.X(vtx(1)),particles.X(vtx(2)),particles.X(vtx(3)));
    }
    return total_tri_area;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Incident_Triangle_Area(const TV_INT2& ac,const int v,const ARRAY<int>& tris)
{
    T total_tri_area=(T)0;
    for(int i=1;i<=tris.m;i++){
        const TV_INT3& vtx=triangle_mesh.elements(tris(i));
        TV_INT3 replaced_v;Replaced_Two(vtx,replaced_v,ac(1),ac(2),v);
        total_tri_area+=TRIANGLE_3D<T>::Area(particles.X(replaced_v(1)),particles.X(replaced_v(2)),particles.X(replaced_v(3)));
    }
    return total_tri_area;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable=mesh.edge_tetrahedron_hashtable;edge_tetrahedron_hashtable.Remove_All();
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable=mesh.edge_triangle_hashtable;edge_triangle_hashtable.Remove_All();
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    Update_Edge_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_tetrahedron_hashtable);
    Update_Edge_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_triangle_hashtable);
    Update_Particle_Dimensional_Size(tetrahedron_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);
    Update_Particle_Dimensional_Size(triangle_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);

    HASHTABLE<TV_INT2,T> edges_to_split_hashtable;  ////edge index, split frac
    ARRAY<int> tetrahedra_to_remove;ARRAY<int> tetrahedra_added;
    ARRAY<int> triangles_to_remove;ARRAY<int> triangles_added;
    ARRAY<int> particles_added;HASHTABLE<int> particles_affected_hashtable;

    ////split long edges
    for(HASHTABLE_ITERATOR<TV_INT2,ARRAY<int> > iter(edge_tetrahedron_hashtable);iter.Valid();iter.Next()){
        TV_INT2 edge_key=iter.Key();int a=edge_key.x;int c=edge_key.y;
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,edge_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,edge_key))continue;   ////for inactive particles

        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        if(!particles.on_boundary(a)(TV::m)&&!particles.on_boundary(c)(TV::m)){   ////interior edge split
            bool split=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()>length_to_split_coef*particle_length_scale;
            if(split){edges_to_split_hashtable.Set(edge_key,(T).5);}
        }
        else if(particles.on_boundary(a)(TV::m)&&particles.on_boundary(c)(TV::m)){    
            bool split=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()>length_to_split_coef*particle_length_scale;
            if(!split){ ////check if it needs split for incident triangles
                if(edge_triangle_hashtable.Contains(edge_key)){
                    T large_angle_cosine=cos((T)(.7*pi));
                    const ARRAY<int>& e_tris=edge_triangle_hashtable.Get(edge_key);
                    for(int i=1;i<=e_tris.m;i++){
                        int t=e_tris(i);
                        const TV_INT3& v=triangle_mesh.elements(t);
                        int b=Other_Node_In_Element(v,a,c);
                        T cos_abc=TV::Dot_Product((particles.X(a)-particles.X(b)).Normalized(),(particles.X(c)-particles.X(b)).Normalized());
                        if(cos_abc<(T)large_angle_cosine){split=true;break;}
                    }
                }
            }
            if(split){edges_to_split_hashtable.Set(edge_key,(T).5);}
        }
        else{   ////buffer edge split
            bool split=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()>length_to_split_coef*particle_length_scale;
            if(split){edges_to_split_hashtable.Set(edge_key,(T).5);}
        }
    }

    ////////split edges on skinny tets (spade and sliver) and negative volume tets
    //T short_edge_frac=(T).2;T short_distance_frac=(T).1;T t_threshold=(T).01;
    //for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
    //    const TV_INT4& v=tetrahedron_mesh.elements(i);
    //    if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
    //    if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

    //    TETRAHEDRON<T> tet(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));

    //    ////skip if it is not a skinny tet
    //    T particle_length_scale=use_adaptive_tetrahedron_remeshing?(T).25*(particles.length_scale(v(1))+particles.length_scale(v(2))+particles.length_scale(v(3))+particles.length_scale(v(4))):length_scale;     
    //    T short_edge_threshold=particle_length_scale*(T)short_edge_frac;if(Has_Short_Edge(tet,short_edge_threshold))continue;
    //    T distance_threshold=particle_length_scale*short_distance_frac;
    //    SKINNY_TETRAHEDRON<T> skinny_tet;int skinny_tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,skinny_tet);
    //    if(skinny_tet_type==SLIVER||skinny_tet_type==SPADE){
    //        for(int j=1;j<=skinny_tet.edges.m;j++){
    //            TV_INT2 e=TV_INT2(v(skinny_tet.edges(j)(1)),v(skinny_tet.edges(j)(2)));
    //            TV_INT2 e_key=e.Sorted();T t=skinny_tet.edge_weights(j);if(e_key!=e){t=(T)1-t;}
    //            if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;
    //            if(Element_Or(particles.inactive,e_key))continue;

    //            //////adjust t
    //            //T edge_length=(particles.X(e(1))-particles.X(e(2))).Magnitude();
    //            //T edge_particle_length_scale=(T).5*(particles.length_scale(e(1))+particles.length_scale(e(2)));
    //            //T min_edge_length=length_to_collapse_coef*edge_particle_length_scale;
    //            //if(edge_length<min_edge_length*(T)2)t=(T).5;
    //            //else t=clamp(t,min_edge_length/edge_length,(T)1-min_edge_length/edge_length);

    //            edges_to_split_hashtable.Set(e_key,t);
    //        }
    //    }
    //}

    QUEUE<TV_INT2> edges_to_split_queue((int)((T)edges_to_split_hashtable.Size()*(T)1.1));
    for(HASHTABLE_ITERATOR<TV_INT2,T> iter(edges_to_split_hashtable);iter.Valid();iter.Next()){
        TV_INT2 edge_key=iter.Key();
        edges_to_split_queue.Enqueue(edge_key);
    }

    HASHTABLE<TV_INT2,int> edge_split_particle_hashtable;
    int number_of_splitted_edge=0;
    while(!edges_to_split_queue.Empty()){
        const TV_INT2& e_key=edges_to_split_queue.Dequeue();
        bool splitted=Split_Tetrahedron_Edge(e_key,edge_tetrahedron_hashtable,edge_triangle_hashtable,particles_added,particles_affected_hashtable,triangles_to_remove,triangles_added,tetrahedra_to_remove,tetrahedra_added);
        if(splitted){
            number_of_splitted_edge++;
            TV_INT2 e_guid_key=Element(particles.guid,e_key).Sorted();
            int splitted_particle_guid=particles.guid(particles.guid.m);
            edge_split_particle_hashtable.Set(e_guid_key,splitted_particle_guid);
        }
    }

    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());
    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,0,&particles_affected,&tetrahedra_added,&tetrahedra_to_remove,&triangles_added,&triangles_to_remove,0,0);
    else Remove_Particles_And_Elements(0,&tetrahedra_to_remove,&triangles_to_remove,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Face_Split_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable=mesh.face_tetrahedron_hashtable;face_tetrahedron_hashtable.Remove_All();
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    Update_Face_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,face_tetrahedron_hashtable);
    Update_Particle_Dimensional_Size(tetrahedron_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);

    HASHTABLE<TV_INT3> face_to_split_hashtable;
    ARRAY<int> tetrahedra_to_remove;
    ARRAY<int> tetrahedra_added;
    ARRAY<int> particles_added;HASHTABLE<int> particles_affected_hashtable;

    //////find cap tets
    //T uvw_epsilon=(T)1e-8;
    //T min_altitude_threshold=length_scale*(T)1e-2;
    //for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
    //    const TV_INT4& v=tetrahedron_mesh.elements(i);
    //    TETRAHEDRON<T> tet(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));

    //    int min_idx;T min_altitude=tet.Minimum_Altitude(min_idx);int a=v(min_idx);
    //    T particle_length_scale=use_adaptive_tetrahedron_remeshing?(T).25*(particles.length_scale(v(1))+particles.length_scale(v(2))+particles.length_scale(v(3))+particles.length_scale(v(4))):length_scale;
    //    if(min_altitude<min_altitude_threshold){
    //        //LOG::cout<<"try tet face split "<<v<<std::endl;
    //        ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);bool has_short_edge=false;
    //        for(int d=1;d<=6;d++)if((particles.X(edges(d)(1))-particles.X(edges(d)(2))).Magnitude()<(T).2*length_to_collapse_coef*particle_length_scale){
    //            //LOG::cout<<"has short edge: "<<(particles.X(edges(d)(1))-particles.X(edges(d)(2))).Magnitude()<<" < "<<(T).2*length_to_collapse_coef*particle_length_scale<<std::endl;
    //            has_short_edge=true;break;}
    //        if(has_short_edge){continue;}   ////skip the tet with short edge

    //        TV_INT3 face=Simplex_Opposite_Face(v,a);
    //        TV_INT3 face_key=face.Sorted();if(!face_tetrahedron_hashtable.Contains(face_key))continue;
    //        TV uvw=Projected_Barycentric_Coordinates(TRIANGLE_3D<T>(particles.X(face(1)),particles.X(face(2)),particles.X(face(3))),particles.X(a));
    //        if(uvw(1)>uvw_epsilon && uvw(1)<(T)1 && uvw(2)>uvw_epsilon && uvw(2)<(T)1 && uvw(3)>uvw_epsilon && uvw(3)<(T)1){
    //            face_to_split_hashtable.Set(face_key);
    //        }
    //        else{LOG::cout<<"invalid uvw of cap tet: "<<uvw<<std::endl;}
    //    }
    //}

    ////split face on cap tets
    T short_edge_frac=(T).2;T short_distance_frac=(T).1;T t_threshold=(T).01;
    for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
        const TV_INT4& v=tetrahedron_mesh.elements(i);
        TETRAHEDRON<T> tet(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));

        ////skip if it is not a skinny tet
        T particle_length_scale=Tetrahedron_Length_Scale(v);     
        T short_edge_threshold=particle_length_scale*(T)short_edge_frac;if(Has_Short_Edge(tet,short_edge_threshold))continue;

        T distance_threshold=particle_length_scale*short_distance_frac;
        SKINNY_TETRAHEDRON<T> skinny_tet;
        int skinny_tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,skinny_tet);
        if(skinny_tet_type==CAP){
            TV_INT3 face(v(skinny_tet.faces(1)(1)),v(skinny_tet.faces(1)(2)),v(skinny_tet.faces(1)(3)));TV_INT3 face_key=face.Sorted();
            //LOG::cout<<"split cap tet: "<<v<<" guid "<<Element(particles.guid,v)<<std::endl;
            if(face_tetrahedron_hashtable.Contains(face_key))face_to_split_hashtable.Set(face_key);
        }

        //ARRAY<TV_INT2> edges;ARRAY<T> edge_t;TV_INT3 face_local_v;TV3 face_uvw; ////face index is 1-4!
        //T distance_threshold=particle_length_scale*short_distance_frac;
        //int tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,edges,edge_t,face_local_v,face_uvw);
        //if(tet_type==CAP){
        //    TV_INT3 face(v(face_local_v(1)),v(face_local_v(2)),v(face_local_v(3)));
        //    TV_INT3 face_key=face.Sorted();
        //    //LOG::cout<<"split cap tet: "<<v<<" guid "<<Element(particles.guid,v)<<std::endl;
        //    if(face_tetrahedron_hashtable.Contains(face_key))face_to_split_hashtable.Set(face_key);
        //}
    }

    int number_of_splitted_face=0;
    for(HASHTABLE_ITERATOR<TV_INT> iter(face_to_split_hashtable);iter.Valid();iter.Next()){
        TV_INT3 face_key=iter.Key();
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,face_key,regions,(T)0))continue;  ////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,face_key))continue;   ////for inactive particles

        bool splitted=Split_Tetrahedron_Face(face_key,face_tetrahedron_hashtable,particles_added,particles_affected_hashtable,tetrahedra_to_remove,tetrahedra_added);
        if(splitted)number_of_splitted_face++;
    }

    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());
    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,0,&particles_affected,&tetrahedra_added,&tetrahedra_to_remove,0,0,0,0);
    else Remove_Particles_And_Elements(0,&tetrahedra_to_remove,0,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Flip_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable=mesh.edge_tetrahedron_hashtable;edge_tetrahedron_hashtable.Remove_All();
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable=mesh.edge_triangle_hashtable;edge_triangle_hashtable.Remove_All();
    HASHTABLE<TV_INT2>& edge_on_tet_mesh_boundary_hashtable=mesh.edge_on_tet_mesh_boundary_hashtable;edge_on_tet_mesh_boundary_hashtable.Remove_All();

    //Update_Edge_On_Tetrahedron_Mesh_Boundary_Hashtable_Threaded(codimensional_mesh_object,mesher_tid,edge_on_tet_mesh_boundary_hashtable);
    Update_Edge_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,(T)0,edge_tetrahedron_hashtable);
    Update_Edge_Element_Hashtable(triangle_mesh.elements,particles.X,regions,(T)0,edge_triangle_hashtable);
    Update_Boundary_Edge_Hashtable(tetrahedron_mesh.elements,particles.X,regions,(T)0,edge_on_tet_mesh_boundary_hashtable);
    Update_Particle_Dimensional_Size(tetrahedron_mesh.elements,particles.X,regions,(T)0,particles.dimensional_size);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;

    ////TODO: add edge_tet_boundary_hashtable
    QUEUE<TV_INT2> edges_to_flip_queue(max(tetrahedron_mesh.elements.m*6,200));
    HASHTABLE<TV_INT2> edges_to_flip_hashtable;
    ARRAY<int> tetrahedra_to_remove;
    ARRAY<int> tetrahedra_added;
    HASHTABLE<int> particles_affected_hashtable;

    for(HASHTABLE_ITERATOR<TV_INT2,ARRAY<int> > iter(edge_tetrahedron_hashtable);iter.Valid();iter.Next()){
        TV_INT2 edge_key=iter.Key();
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,edge_key,regions,(T)0)){continue;}    ////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,edge_key))continue;   ////for inactive particles

        edges_to_flip_hashtable.Set(edge_key);
        edges_to_flip_queue.Enqueue(edge_key);
    }

    int number_of_flipped_edge=0;
    while(!edges_to_flip_queue.Empty()){
        TV_INT2 edge_key=edges_to_flip_queue.Dequeue();
        edges_to_flip_hashtable.Delete(edge_key);

        bool flipped=false;ARRAY<TV_INT2> affected_edges;
        if(edge_on_tet_mesh_boundary_hashtable.Contains(edge_key)){
            flipped=Flip_Tetrahedron_Boundary_Edge(edge_key,edge_tetrahedron_hashtable,edge_on_tet_mesh_boundary_hashtable,edge_triangle_hashtable,
                particles_affected_hashtable,tetrahedra_to_remove,tetrahedra_added,affected_edges);
        }
        else{
            flipped=Flip_Tetrahedron_Edge(edge_key,edge_tetrahedron_hashtable,particles_affected_hashtable,tetrahedra_to_remove,tetrahedra_added,affected_edges);
        }
        if(flipped){number_of_flipped_edge++;}
    }

    Sort(tetrahedra_to_remove);
    tetrahedron_mesh.elements.Remove_Sorted_Indices(tetrahedra_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,&particles_affected,&tets_added,&tets_to_rmv,0,0,0,0);
    else Remove_Particles_And_Elements(0,&tets_to_rmv,0,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Face_Flip_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable=mesh.face_tetrahedron_hashtable;face_tetrahedron_hashtable.Remove_All();

    Update_Face_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,(T)0,face_tetrahedron_hashtable);
    Update_Particle_Dimensional_Size(tetrahedron_mesh.elements,particles.X,regions,(T)0,particles.dimensional_size);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;

    QUEUE<TV_INT3> faces_to_flip_queue(max(tetrahedron_mesh.elements.m*4,200));
    HASHTABLE<TV_INT3> faces_to_flip_hashtable;

    ARRAY<int> tetrahedra_to_remove;
    ARRAY<int> tetrahedra_added;
    HASHTABLE<int> particles_affected_hashtable;

    int number_of_flipped_tetrahedron_face=0;
    for(HASHTABLE_ITERATOR<TV_INT3,ARRAY<int> > iter(face_tetrahedron_hashtable);iter.Valid();iter.Next()){
        TV_INT3 face_key=iter.Key();
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,face_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,face_key))continue;   ////for inactive particles

        faces_to_flip_queue.Enqueue(face_key);
        faces_to_flip_hashtable.Set(face_key);
    }

    while(!faces_to_flip_queue.Empty()){
        TV_INT3 face_key=faces_to_flip_queue.Dequeue();
        faces_to_flip_hashtable.Delete(face_key);

        ARRAY<TV_INT3> affected_faces;
        bool flipped=Flip_Tetrahedron_Face(face_key,face_tetrahedron_hashtable,particles_affected_hashtable,tetrahedra_to_remove,tetrahedra_added,affected_faces);
        if(flipped){number_of_flipped_tetrahedron_face++;}
    }

    Sort(tetrahedra_to_remove);
    tetrahedron_mesh.elements.Remove_Sorted_Indices(tetrahedra_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,&particles_affected,&tets_added,&tets_to_rmv,0,0,0,0);
    else Remove_Particles_And_Elements(0,&tets_to_rmv,0,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);
    HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable=mesh.face_tetrahedron_hashtable;face_tetrahedron_hashtable.Remove_All();
    Update_Face_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,face_tetrahedron_hashtable);
    HASHTABLE<TV_INT2>& edge_on_tet_mesh_hashtable=mesh.edge_on_tet_mesh_hashtable;edge_on_tet_mesh_hashtable.Remove_All();
    Update_Edge_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_on_tet_mesh_hashtable);
    Update_Particle_Dimensional_Size(tetrahedron_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    //QUEUE<PAIR<TV_INT2,int> > edges_to_collapse_queue(max(tetrahedron_mesh.elements.m*2,200));
    HASHTABLE<TV_INT2,int> edges_to_collapse_hashtable; ////int for edge types: 0-interior, 1-diagonal, 2-boundary, 3-buffer, 4-negvol
    HASHTABLE<TV_INT2> negative_volume_tet_edge_hashtable;

    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    ARRAY<int> triangle_indices_to_remove;
    ARRAY<int> tetrahedron_indices_to_remove;

    for(HASHTABLE_ITERATOR<TV_INT2> iter(edge_on_tet_mesh_hashtable);iter.Valid();iter.Next()){
        TV_INT2 edge_key=iter.Key();int a=edge_key.x;int c=edge_key.y;
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,edge_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,edge_key))continue;   ////for inactive particles

        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        if(!particles.on_boundary(a)(TV::m)&&!particles.on_boundary(c)(TV::m)){   ////interior edge collapse
            bool collapse=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()<length_to_collapse_coef*particle_length_scale;
            if(collapse)edges_to_collapse_hashtable.Set(edge_key,0);
        }
        else if(particles.on_boundary(a)(TV::m)&&particles.on_boundary(c)(TV::m)){    
            //////diagonal edge collapse
            //if(!mesh.Is_Boundary_Edge_On_Tetrahedron_Mesh(edge_key)){
            //    bool collapse=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()<diagonal_min_segment_length;
            //    if(collapse)edges_to_collapse_hashtable.Set(edge_key,1);
            //}
            //else
            {////boundary edge collapse
                bool collapse=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()<length_to_collapse_coef*particle_length_scale;
                if(collapse)edges_to_collapse_hashtable.Set(edge_key,2);
            }
        }
        else{   ////buffer edge collapse
            bool collapse=(particles.X(edge_key(1))-particles.X(edge_key(2))).Magnitude()<length_to_collapse_coef*particle_length_scale;
            if(collapse)edges_to_collapse_hashtable.Set(edge_key,3);
        }
    }

    ////check negative volume tet
    for(HASHTABLE_ITERATOR<TV_INT3,ARRAY<int> > iter(face_tetrahedron_hashtable);iter.Valid();iter.Next()){
        const ARRAY<int>& f_tets=iter.Data();
        for(int i=1;i<=f_tets.m;i++){
            const TV_INT4& tet_v=tetrahedron_mesh.elements(f_tets(i));
            if(Is_Negative_Volume_Tetrahedron(particles.X,tet_v)){
                ARRAY<TV_INT2> edges;Simplex_Edges(tet_v,edges);
                T shortest_edge_length=(T)FLT_MAX;int shortest_edge_idx=0;
                for(int j=1;j<=6;j++){const TV_INT2& e=edges(j);
                    if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e,regions,(T)0))continue;
                    if(Element_Or(particles.inactive,e))continue;   ////for inactive particles

                    T edge_length=(particles.X(e(1))-particles.X(e(2))).Magnitude();if(edge_length<shortest_edge_length){shortest_edge_length=edge_length;shortest_edge_idx=j;}
                }
                TV_INT2 shortest_edge_key=edges(shortest_edge_idx).Sorted();
                negative_volume_tet_edge_hashtable.Set(shortest_edge_key);
                edges_to_collapse_hashtable.Set(shortest_edge_key,4);
                //LOG::cout<<"set neg tet edge to collapse: "<<tet_v<<", edge: "<<shortest_edge_key<<" guid "<<Element(particles.guid,shortest_edge_key)<<std::endl;
            }
        }
    }

    int number_of_collapsed_edge=0;
    while(edges_to_collapse_hashtable.Size()>0){
        ////pick edges on neg tet first?

        ////pick the shortest edge then
        T shortest_edge_length=(T)FLT_MAX;TV_INT2 e_key;int e_type=0;
        for(HASHTABLE_ITERATOR<TV_INT2,int> iter(edges_to_collapse_hashtable);iter.Valid();iter.Next()){
            const TV_INT2& v=iter.Key();T edge_length=(particles.X(v(1))-particles.X(v(2))).Magnitude();
            if(edge_length<shortest_edge_length){shortest_edge_length=edge_length;e_key=v;e_type=iter.Data();}
        }
        edges_to_collapse_hashtable.Delete(e_key);

        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles

        int a=e_key(1);int c=e_key(2);

        //////dbg
        //TV_INT2 e_guid=Element(particles.guid,e_key);
        //if(e_guid.Contains(461)&&e_guid.Contains(1004276)){break;}

        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        T min_edge_length=length_to_collapse_coef*particle_length_scale;
        if(e_type!=4&&(particles.X(a)-particles.X(c)).Magnitude()>min_edge_length){continue;}    ////collapse fail: no longer short and is not neg tet edge

        ARRAY<int> ear_particles_to_remove;
        bool collapsed=Collapse_Tetrahedron_Edge(e_key,e_type,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,face_tetrahedron_hashtable,
            particles_to_remove,ear_particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,tetrahedron_indices_to_remove);

        if(collapsed){
            number_of_collapsed_edge++;
            ////replace a and c with v in edge queue
            int v=particles_added(particles_added.m);

            ARRAY<TV_INT2> keys_to_delete;ARRAY<TV_INT2> keys_to_add;ARRAY<int> data_to_add;
            for(HASHTABLE_ITERATOR<TV_INT2,int> iter(edges_to_collapse_hashtable);iter.Valid();iter.Next()){
                TV_INT2 old_e_key=iter.Key();TV_INT2 new_e_key=Replaced_Two(old_e_key,a,c,v).Sorted();int e_type=iter.Data();
                if(new_e_key!=old_e_key){
                    keys_to_delete.Append(old_e_key);keys_to_add.Append(new_e_key);data_to_add.Append(e_type);
                }
            }
            for(int i=1;i<=keys_to_delete.m;i++){edges_to_collapse_hashtable.Delete(keys_to_delete(i));}
            for(int i=1;i<=keys_to_add.m;i++){edges_to_collapse_hashtable.Set(keys_to_add(i),data_to_add(i));}
        }

        //if(ear_particles_to_remove.m>0){
        //    for(int k=1;k<=ear_particles_to_remove.m;k++){
        //        int p=ear_particles_to_remove(k);
        //        if(particles_to_remove.Contains(p))continue;
        //        for(int i=1;i<=edges_to_collapse_queue.Size();i++){
        //            if(edges_to_collapse_queue(i).x.Contains(p)){   ////set both vertices to be p to force it to terminate in the collapse_tet_edge function
        //                edges_to_collapse_queue(i).x(1)=p;edges_to_collapse_queue(i).x(2)=p;
        //            }
        //        }
        //        particles_to_remove.Append(p);
        //    }
        //}
    }

    Sort(tetrahedron_indices_to_remove);
    tetrahedron_mesh.elements.Remove_Sorted_Indices(tetrahedron_indices_to_remove);
    Sort(triangle_indices_to_remove);
    triangle_mesh.elements.Remove_Sorted_Indices(triangle_indices_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,&particles_affected,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}


template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Predict_Atomic_Cap_Tetrahedron_Collapse(const int tet_idx,const TV_INT3& face,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable)
{
    const TV_INT4& vtx=tetrahedron_mesh.elements(tet_idx);int d=Other_Node_In_Element(vtx,face(1),face(2),face(3));
    int v=0;int guid_v=0;int on_boundary_v=particles.on_boundary(face(1))(TV::m)&&particles.on_boundary(face(2))(TV::m)&&particles.on_boundary(face(3))(TV::m);
    int contact_v=particles.contact(face(1))&particles.contact(face(2))&particles.contact(face(3));
    TV X_v=(T)one_third*(particles.X(face(1))+particles.X(face(2))+particles.X(face(3)));
    
    ARRAY<TV_INT4> new_one_ring_tet_vtx;
    ARRAY<int>* d_tets=vertex_tetrahedron_hashtable.Get_Pointer(d);
    if(d_tets)for(int i=1;i<=(*d_tets).m;i++){int t=(*d_tets)(i);new_one_ring_tet_vtx.Append(tetrahedron_mesh.elements(t));}
    TV_INT3 f_key=face.Sorted();ARRAY<int>* f_tets=face_tetrahedron_hashtable.Get_Pointer(f_key);
    if(f_tets)for(int i=1;i<=(*f_tets).m;i++){
        int t=(*f_tets)(i);if(t==tet_idx)continue;
        const TV_INT4& abcd=tetrahedron_mesh.elements(t);
        TV_INT4 adbv;TV_INT4 bdcv;TV_INT4 cdav;Three_Splitted_Tetrahedron_Vertices(abcd,f_key,v,adbv,bdcv,cdav);
        new_one_ring_tet_vtx.Append(adbv);new_one_ring_tet_vtx.Append(bdcv);new_one_ring_tet_vtx.Append(cdav);
    }
    TV v_pos;return Valid_Tetrahedron_Edge_Collapse(v,d,guid_v,particles.guid(d),on_boundary_v,particles.on_boundary(d)(TV::m),contact_v,particles.contact(d),X_v,particles.X(d),new_one_ring_tet_vtx,true,v_pos);
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Predict_Atomic_Spade_Tetrahedron_Collapse(const TV_INT2& edge,const int d,const T w,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable)
{
    int a=edge(1);int c=edge(2);int v=0;int guid_v=0;int on_boundary_v=particles.on_boundary(edge(1))(TV::m)&&particles.on_boundary(edge(2))(TV::m);
    int contact_v=particles.contact(edge(1))&particles.contact(edge(2));TV X_v=((T)1-w)*particles.X(edge(1))+w*particles.X(edge(2));

    ARRAY<TV_INT4> new_one_ring_tet_vtx;
    ARRAY<int>* d_tets=vertex_tetrahedron_hashtable.Get_Pointer(d);
    if(d_tets)for(int i=1;i<=(*d_tets).m;i++){
        int t=(*d_tets)(i);const TV_INT4& vtx=tetrahedron_mesh.elements(t);
        if(vtx.Contains(a)&&vtx.Contains(c))continue;
        new_one_ring_tet_vtx.Append(vtx);
    }

    ARRAY<int>* ac_tets=edge_tetrahedron_hashtable.Get_Pointer(edge.Sorted());
    if(ac_tets)for(int i=1;i<=(*ac_tets).m;i++){
        int t=(*ac_tets)(i);const TV_INT4& vtx=tetrahedron_mesh.elements(t);if(vtx.Contains(d))continue;
        TV_INT4 vbcd;TV_INT4 vabd;Two_Splitted_Tetrahedron_Vertices(vtx,a,c,v,vbcd,vabd);
        new_one_ring_tet_vtx.Append(vbcd);new_one_ring_tet_vtx.Append(vabd);
    }
    TV v_pos;return Valid_Tetrahedron_Edge_Collapse(v,d,guid_v,particles.guid(d),on_boundary_v,particles.on_boundary(d)(TV::m),contact_v,particles.contact(d),X_v,particles.X(d),new_one_ring_tet_vtx,true,v_pos);
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Predict_Atomic_Sliver_Tetrahedron_Collapse(const TV_INT2& e1,const TV_INT2& e2,const T w1,const T w2,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable)
{
    int v1=0;int guid_v1=0;int on_boundary_v1=particles.on_boundary(e1(1))(TV::m)&&particles.on_boundary(e1(2))(TV::m);
    int contact_v1=particles.contact(e1(1))&particles.contact(e1(2));TV X_v1=((T)1-w1)*particles.X(e1(1))+w1*particles.X(e1(2));
    int v2=-1;int guid_v2=-1;int on_boundary_v2=particles.on_boundary(e2(1))(TV::m)&&particles.on_boundary(e2(2))(TV::m);
    int contact_v2=particles.contact(e2(1))&particles.contact(e2(2));TV X_v2=((T)1-w2)*particles.X(e2(1))+w2*particles.X(e2(2));

    ARRAY<TV_INT4> new_one_ring_tet_vtx;ARRAY<TV_INT2> es;es.Append(e1);es.Append(e2);
    for(int j=1;j<=2;j++){
        TV_INT2 ac=es(j);TV_INT2 bd=(j==1?es(2):es(1));int v=(j==1?v1:v2);
        ARRAY<int>* ac_tets=edge_tetrahedron_hashtable.Get_Pointer(ac.Sorted());
        if(ac_tets)for(int i=1;i<=(*ac_tets).m;i++){
            int t=(*ac_tets)(i);const TV_INT4& vtx=tetrahedron_mesh.elements(t);if(vtx.Contains(bd(1))&&vtx.Contains(bd(2)))continue;
            TV_INT4 vbcd;TV_INT4 vabd;Two_Splitted_Tetrahedron_Vertices(vtx,ac(1),ac(2),v,vbcd,vabd);
            new_one_ring_tet_vtx.Append(vbcd);new_one_ring_tet_vtx.Append(vabd);
        }
    }
    TV v_pos;return Valid_Tetrahedron_Edge_Collapse(v1,v2,guid_v1,guid_v2,on_boundary_v1,on_boundary_v2,contact_v1,contact_v2,X_v1,X_v2,new_one_ring_tet_vtx,true,v_pos);
}


template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Has_Short_Edge(const TV_INT4& vtx,TV_INT2& shortest_edge)
{
    ARRAY<TV_INT2> edges;Simplex_Edges(vtx,edges);T shortest_edge_length=(T)FLT_MAX;bool has_short_edge=false;
    for(int d=1;d<=6;d++){
        int a=edges(d)(1);int c=edges(d)(2);
        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        T edge_length=(particles.X(a)-particles.X(c)).Magnitude();
        if(edge_length<length_to_collapse_coef*particle_length_scale&&edge_length<shortest_edge_length){
            shortest_edge_length=edge_length;shortest_edge=TV_INT2(a,c);has_short_edge=true;
        }
    }
    return has_short_edge;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Has_Short_Edge_With_Boundary(const TV_INT4& vtx,ARRAY_VIEW<TV_INT>& on_boundary,TV_INT2& shortest_edge)
{
    ARRAY<TV_INT2> edges;Simplex_Edges(vtx,edges);T shortest_edge_length=(T)FLT_MAX;bool has_short_edge=false;
    for(int d=1;d<=6;d++){
        int a=edges(d)(1);int c=edges(d)(2);
        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        bool edge_on_boundary=on_boundary(a)(TV::m)&&on_boundary(c)(TV::m);
        T boundary_coef=edge_on_boundary?(T).5:(T)1;
        T edge_length=(particles.X(a)-particles.X(c)).Magnitude();
        if(edge_length<length_to_collapse_coef*boundary_coef*particle_length_scale&&edge_length<shortest_edge_length){
            shortest_edge_length=edge_length;shortest_edge=TV_INT2(a,c);has_short_edge=true;
        }
    }
    return has_short_edge;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Has_Short_Edge_With_Diagonal(const TV_INT4& vtx,ARRAY_VIEW<TV_INT>& on_boundary,const HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,TV_INT2& shortest_edge)
{
    ARRAY<TV_INT2> edges;Simplex_Edges(vtx,edges);T shortest_edge_length=(T)FLT_MAX;bool has_short_edge=false;
    for(int d=1;d<=6;d++){
        //int a=edges(d)(1);int c=edges(d)(2);
        TV_INT2 e_key=edges(d).Sorted();int a=e_key(1);int c=e_key(2);
        T particle_length_scale=Tetrahedron_Edge_Length_Scale(a,c);
        bool edge_on_boundary=on_boundary(a)(TV::m)&&on_boundary(c)(TV::m);
        bool is_diagonal_edge=false;if(edge_on_boundary){
            const ARRAY<int>& e_tets=edge_tetrahedron_hashtable.Get(e_key);
            is_diagonal_edge=Is_Diagonal_Edge(e_key,e_tets);
        }

        T diagonal_coef=is_diagonal_edge?(T).5:(T)1;
        T edge_length=(particles.X(a)-particles.X(c)).Magnitude();
        if(edge_length<length_to_collapse_coef*diagonal_coef*particle_length_scale&&edge_length<shortest_edge_length){
            shortest_edge_length=edge_length;shortest_edge=TV_INT2(a,c);has_short_edge=true;
        }
    }
    return has_short_edge;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Diagonal_Edge(const TV_INT2& e,const ARRAY<int>& e_tets)
{
    T dihedral_angle_sum=(T)0;
    for(int i=1;i<=e_tets.m;i++){
        const TV_INT4& v=tetrahedron_mesh.elements(e_tets(i));
        int a_idx=v.Find(e(1));int c_idx=v.Find(e(2));
        TETRAHEDRON<T> tet(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));
        dihedral_angle_sum+=Dihedral_Angle(tet,TV_INT2(a_idx,c_idx));
    }
    return dihedral_angle_sum>(T).9*(T)two_pi;////diagonal edge
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable=mesh.edge_tetrahedron_hashtable;edge_tetrahedron_hashtable.Remove_All();
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable=mesh.edge_triangle_hashtable;edge_triangle_hashtable.Remove_All();
    HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable=mesh.face_tetrahedron_hashtable;face_tetrahedron_hashtable.Remove_All();
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);
    Update_Edge_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_tetrahedron_hashtable);
    Update_Edge_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_triangle_hashtable);
    Update_Face_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,face_tetrahedron_hashtable);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    ARRAY<int> triangle_indices_to_remove;ARRAY<int> triangle_indices_added;
    ARRAY<int> tetrahedron_indices_to_remove;ARRAY<int> tetrahedron_indices_added;

    HASHTABLE<int> skinny_tet_hashtable;
    HASHTABLE<int> short_edge_tet_hashtable;
    //T short_edge_frac=(T).2;
    T short_distance_frac=(T).1;T t_threshold=(T).01;
    for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
        const TV_INT4& vtx=tetrahedron_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,vtx,regions,(T)0))continue;  ////for mpi boundaries, at least one vertex is inside the region
        TV_INT2 short_edge;bool has_short_edge=false;
        if(use_thin_sheet_aware_tetrahedron_remeshing) has_short_edge=Tetrahedron_Has_Short_Edge_With_Diagonal(vtx,particles.on_boundary,edge_tetrahedron_hashtable,short_edge);
        else has_short_edge=Tetrahedron_Has_Short_Edge(vtx,short_edge);
        
        if(has_short_edge){
            if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,short_edge,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
            short_edge_tet_hashtable.Set(i);
        }
        else{
            TETRAHEDRON<T> tet(particles.X(vtx(1)),particles.X(vtx(2)),particles.X(vtx(3)),particles.X(vtx(4)));
            T particle_length_scale=Tetrahedron_Length_Scale(vtx);
            T distance_threshold=particle_length_scale*short_distance_frac;SKINNY_TETRAHEDRON<T> skinny_tet;
            int skinny_tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,skinny_tet);
            if(skinny_tet_type!=0){skinny_tet_hashtable.Set(i);}
        }
    }

    HASHTABLE<int> visited_tet_hashtable;
    //int c=0;
    while(skinny_tet_hashtable.Size()>0||short_edge_tet_hashtable.Size()>0){
        ARRAY<int> affected_v;int t=0;
        //LOG::cout<<"hashtables: "<<skinny_tet_hashtable.Size()<<", "<<short_edge_tet_hashtable.Size()<<std::endl;
        //if(c==3001)break;

        ////check short edge first
        if(short_edge_tet_hashtable.Size()>0){
            HASHTABLE_ITERATOR<int> iter(short_edge_tet_hashtable);
            t=iter.Key();short_edge_tet_hashtable.Delete(t);
            visited_tet_hashtable.Set(t);
            if(tetrahedron_indices_to_remove.Contains(t))continue;

            ////check again to find its current shortest edge
            const TV_INT4 vtx=tetrahedron_mesh.elements(t);
            TV_INT2 short_edge;
            bool has_short_edge=false;
            if(use_thin_sheet_aware_tetrahedron_remeshing)Tetrahedron_Has_Short_Edge_With_Diagonal(vtx,particles.on_boundary,edge_tetrahedron_hashtable,short_edge);
            else has_short_edge=Tetrahedron_Has_Short_Edge(vtx,short_edge);
            bool collapsed=false;

            //LOG::cout<<"check "<<t<<": "<<vtx<<std::endl;
            ////collapse short edge
            if(has_short_edge){
                if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,short_edge,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions

                ARRAY<int> ear_particles_to_remove; //unused
                collapsed=Collapse_Tetrahedron_Edge(short_edge,0,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,face_tetrahedron_hashtable,
                    particles_to_remove,ear_particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,tetrahedron_indices_to_remove,
                    false,&edge_tetrahedron_hashtable,&edge_triangle_hashtable);
                skinny_tet_hashtable.Delete_If_Present(t);
            }
            if(collapsed)affected_v.Append(particles.X.m);
        }
        ///then check skinny tet
        else if(skinny_tet_hashtable.Size()>0){
            HASHTABLE_ITERATOR<int> iter(skinny_tet_hashtable);
            t=iter.Key();skinny_tet_hashtable.Delete(t);
            visited_tet_hashtable.Set(t);            
            if(tetrahedron_indices_to_remove.Contains(t))continue;

            const TV_INT4 vtx=tetrahedron_mesh.elements(t);
            TETRAHEDRON<T> tet(particles.X(vtx(1)),particles.X(vtx(2)),particles.X(vtx(3)),particles.X(vtx(4)));

            ////check again to see if it's still skinny tet
            if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,vtx,regions,(T)0))continue;  ////for mpi boundaries, at least one vertex is inside the region
            T particle_length_scale=Tetrahedron_Length_Scale(vtx);    
            //T short_edge_threshold=particle_length_scale*(T)short_edge_frac;if(Has_Short_Edge(tet,short_edge_threshold))continue;

            T distance_threshold=particle_length_scale*short_distance_frac;
            SKINNY_TETRAHEDRON<T> skinny_tet;
            int skinny_tet_type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,skinny_tet);
            bool force_split=true;

            if(use_dbg&&skinny_tet_type!=0)LOG::cout<<"try to collapse skinny tet "<<vtx<<" guid "<<Element(particles.guid,vtx)<<std::endl;

            switch(skinny_tet_type){
            case CAP:{
                if(use_dbg)LOG::cout<<"start collapse cap"<<std::endl;
                TV_INT3 face=TV_INT3(vtx(skinny_tet.faces(1)(1)),vtx(skinny_tet.faces(1)(2)),vtx(skinny_tet.faces(1)(3)));/*TV3 split_weight=TV::Constant_Vector((T)one_third);*/TV3 split_weight=skinny_tet.face_uvw(1);
                TV_INT3 face_key=face.Sorted();if(face_key!=face){TV3 tmp;for(int i=1;i<=3;i++){for(int j=1;j<=3;j++){if(face_key(i)==face(j)){tmp(i)=split_weight(j);break;}}}split_weight=tmp;}

                int d=Other_Node_In_Element(vtx,face_key(1),face_key(2),face_key(3));
                bool splitted=false;bool collapsed=false;

                ////predict if it is atomic 
                if(!Predict_Atomic_Cap_Tetrahedron_Collapse(t,face,vertex_tetrahedron_hashtable,face_tetrahedron_hashtable))break;

                splitted=Split_Tetrahedron_Face(face_key,face_tetrahedron_hashtable,particles_added,particles_affected_hashtable,tetrahedron_indices_to_remove,tetrahedron_indices_added,
                    split_weight,force_split,&vertex_tetrahedron_hashtable,&vertex_triangle_hashtable,&vertex_segment_hashtable,&edge_tetrahedron_hashtable,&edge_triangle_hashtable);
                int v=particles.X.m;
                TV_INT2 e_key=TV_INT2(v,d).Sorted();

                ARRAY<int> ear_particles_to_remove; //unused
                if(splitted)collapsed=Collapse_Tetrahedron_Edge(e_key,0,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,face_tetrahedron_hashtable,
                    particles_to_remove,ear_particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,tetrahedron_indices_to_remove,
                    force_split,&edge_tetrahedron_hashtable,&edge_triangle_hashtable);

                if(collapsed)affected_v.Append(particles.X.m);
                //else if(splitted)affected_v.Append(v);

                if(use_dbg){if(collapsed)LOG::cout<<"succeed collapse cap"<<std::endl;
                else LOG::cout<<"************************** fail to collapse cap **************************"<<std::endl;}
            }break;
            case SPADE:{
                if(use_dbg)LOG::cout<<"start collapse spade"<<std::endl;

                bool splitted=false;bool collapsed=false;
                TV_INT2 e1(vtx(skinny_tet.edges(1)(1)),vtx(skinny_tet.edges(1)(2)));/*T weight1=(T).5;*/T weight1=skinny_tet.edge_weights(1);
                TV_INT2 e1_key=e1.Sorted();if(e1_key!=e1)weight1=(T)1-weight1;int d=vtx(skinny_tet.vertices(1));

                ////predict if the operation is atomic
                if(!Predict_Atomic_Spade_Tetrahedron_Collapse(e1_key,d,weight1,vertex_tetrahedron_hashtable,edge_tetrahedron_hashtable))break;

                splitted=Split_Tetrahedron_Edge(e1_key,edge_tetrahedron_hashtable,edge_triangle_hashtable,particles_added,particles_affected_hashtable,triangle_indices_to_remove,triangle_indices_added,tetrahedron_indices_to_remove,tetrahedron_indices_added,
                    weight1,force_split,&vertex_tetrahedron_hashtable,&vertex_triangle_hashtable,&vertex_segment_hashtable,&face_tetrahedron_hashtable);
                int v=particles.X.m;
                TV_INT2 e3_key=TV_INT2(v,d).Sorted();

                ARRAY<int> ear_particles_to_remove; //unused
                if(splitted)collapsed=Collapse_Tetrahedron_Edge(e3_key,0,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,face_tetrahedron_hashtable,
                    particles_to_remove,ear_particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,tetrahedron_indices_to_remove,
                    force_split,&edge_tetrahedron_hashtable,&edge_triangle_hashtable);

                if(collapsed)affected_v.Append(particles.X.m);
                //else if(splitted)affected_v.Append(v);

                if(use_dbg){if(collapsed)LOG::cout<<"succeed collapse spade"<<std::endl;
                else LOG::cout<<"************************** fail to collapse spade **************************"<<std::endl;}
            }break;
            case SLIVER:{
                if(use_dbg)LOG::cout<<"start collapse sliver"<<std::endl;

                TV_INT2 e1(vtx(skinny_tet.edges(1)(1)),vtx(skinny_tet.edges(1)(2)));/*T weight1=(T).5;*/T weight1=skinny_tet.edge_weights(1);
                TV_INT2 e1_key=e1.Sorted();if(e1_key!=e1)weight1=(T)1-weight1;
                TV_INT2 e2(vtx(skinny_tet.edges(2)(1)),vtx(skinny_tet.edges(2)(2)));/*T weight2=(T).5;*/T weight2=skinny_tet.edge_weights(2);
                TV_INT2 e2_key=e2.Sorted();if(e2_key!=e2)weight2=(T)1-weight2;
                bool splitted_1=false;bool splitted_2=false;bool collapsed=false;
                //LOG::cout<<"weight1: "<<e1<<" guid "<<Element(particles.guid,e1)<<": "<<weight1<<std::endl;
                //LOG::cout<<"weight2: "<<e2<<" guid "<<Element(particles.guid,e2)<<": "<<weight2<<std::endl;

                ////predict if the operation is atomic
                if(!Predict_Atomic_Sliver_Tetrahedron_Collapse(e1,e2,weight1,weight2,vertex_tetrahedron_hashtable,edge_tetrahedron_hashtable))break;

                splitted_1=Split_Tetrahedron_Edge(e1_key,edge_tetrahedron_hashtable,edge_triangle_hashtable,particles_added,particles_affected_hashtable,triangle_indices_to_remove,triangle_indices_added,tetrahedron_indices_to_remove,tetrahedron_indices_added,
                    weight1,force_split,&vertex_tetrahedron_hashtable,&vertex_triangle_hashtable,&vertex_segment_hashtable,&face_tetrahedron_hashtable);
                int v1=particles.X.m;
                splitted_2=Split_Tetrahedron_Edge(e2_key,edge_tetrahedron_hashtable,edge_triangle_hashtable,particles_added,particles_affected_hashtable,triangle_indices_to_remove,triangle_indices_added,tetrahedron_indices_to_remove,tetrahedron_indices_added,
                    weight2,force_split,&vertex_tetrahedron_hashtable,&vertex_triangle_hashtable,&vertex_segment_hashtable,&face_tetrahedron_hashtable);
                int v2=particles.X.m;
                TV_INT2 e3_key=TV_INT2(v1,v2).Sorted();

                ARRAY<int> ear_particles_to_remove; //unused
                if(splitted_1&&splitted_2)collapsed=Collapse_Tetrahedron_Edge(e3_key,0,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,face_tetrahedron_hashtable,
                    particles_to_remove,ear_particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,tetrahedron_indices_to_remove,
                    force_split,&edge_tetrahedron_hashtable,&edge_triangle_hashtable);

                if(collapsed){affected_v.Append(particles.X.m);}
                //else{if(splitted_1)affected_v.Append(v1);if(splitted_2)affected_v.Append(v2);}

                if(use_dbg){if(collapsed)LOG::cout<<"succeed collapse sliver"<<std::endl;
                else LOG::cout<<"************************** fail to collapse sliver **************************"<<std::endl;}
            }break;
            }
        }

        ARRAY<int> affected_tets;
        for(int i=1;i<=affected_v.m;i++){
            int v=affected_v(i);ARRAY<int>* v_tets=vertex_tetrahedron_hashtable.Get_Pointer(v);
            for(int i=1;i<=(*v_tets).m;i++){
                int affected_t=(*v_tets)(i);
                if(affected_t==t)continue;  ////skip itself
                affected_tets.Append_Unique(affected_t);
            }
        }
        for(int i=1;i<=affected_tets.m;i++){
            short_edge_tet_hashtable.Set(affected_tets(i));
            //skinny_tet_hashtable.Set(affected_tets(i));   ////TODO: fix the infinite loop for neighboring skinny tets
        }
    }

    Sort(tetrahedron_indices_to_remove);
    tetrahedron_mesh.elements.Remove_Sorted_Indices(tetrahedron_indices_to_remove);
    Sort(triangle_indices_to_remove);
    triangle_mesh.elements.Remove_Sorted_Indices(triangle_indices_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,&particles_affected,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions,bool split_contact_boundary_edge/*=false*/)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<TV_INT2>& edge_on_tet_mesh_hashtable=codimensional_mesh_object.mesh.edge_on_tet_mesh_hashtable;edge_on_tet_mesh_hashtable.Remove_All();
    Update_Edge_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_on_tet_mesh_hashtable);
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable=codimensional_mesh_object.mesh.edge_triangle_hashtable;edge_triangle_hashtable.Remove_All();
    Update_Edge_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_triangle_hashtable);
    Update_Particle_Dimensional_Size(triangle_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);

    QUEUE<TV_INT2> edges_to_split_queue(max(triangle_mesh.elements.m*2,200));
    HASHTABLE<TV_INT2> edges_to_split_hashtable;
    ARRAY<int> triangles_to_remove;
    ARRAY<int> triangles_added;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    HASHTABLE<TV_INT2,T> contact_boundary_edge_hashtable;
    ////find the edges to split
    for(HASHTABLE_ITERATOR<TV_INT2,ARRAY<int> > iter(edge_triangle_hashtable);iter.Valid();iter.Next()){
        TV_INT2 edge_key=iter.Key();int a=edge_key(1);int c=edge_key(2);
        ARRAY<int>& e_tris=iter.Data();
        
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,edge_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,edge_key))continue;   ////for inactive particles

        ////do not split if the edge in on tetrahedron mesh (since long edge should be already splitted if it is on the tet mesh)
        //bool is_ac_tet_edge=false;
        if(edge_on_tet_mesh_hashtable.Contains(edge_key)){continue;}

        
        if(split_contact_boundary_edge){
            if(use_contact_aware_triangle_remeshing){   ////split if the edge is on contact boundary
                bool contact_a=Is_Contact_With_Solid2(particles.contact(a));
                bool contact_c=Is_Contact_With_Solid2(particles.contact(c));
                //bool contact_boundary_a=Is_Contact_Boundary(particles.contact(a));
                //bool contact_boundary_c=Is_Contact_Boundary(particles.contact(c));
                if(/*!contact_boundary_a&&!contact_boundary_c&&*/contact_a!=contact_c){
                    int contact_idx=contact_a?a:c;int id=0;T weight=(T)-1;T tolerance=(T)1e-3;

                    if(kinematic_rigid_bodies.m>0){ ////contact with kinematic rigid bodies
                        bool intersect=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(contact_idx,particles,kinematic_rigid_bodies,id);
                        if(intersect)weight=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Edge_Intersection_Fraction_With_Solid(edge_key,particles,*kinematic_rigid_bodies(id));
                        if(weight>tolerance&&weight<(T)1-tolerance){edges_to_split_hashtable.Set(edge_key);contact_boundary_edge_hashtable.Set(edge_key,weight);continue;}
                        else if(weight<=tolerance&&weight!=(T)-1){Set_Contact_Boundary(particles.contact(a));}
                        else if(weight>=(T)1-tolerance){Set_Contact_Boundary(particles.contact(c));}
                        else LOG::cout<<"cannot split "<<edge_key<<": "<<weight<<", "<<intersect<<std::endl;
                    }
                    else if(fiber!=0){
                        bool intersect=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(contact_idx,particles,*fiber,id);
                        if(intersect)weight=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Edge_Intersection_Fraction_With_Solid(edge_key,particles,*fiber);
                        if(weight>tolerance&&weight<(T)1-tolerance){edges_to_split_hashtable.Set(edge_key);contact_boundary_edge_hashtable.Set(edge_key,weight);continue;}
                        else if(weight<=tolerance&&weight!=(T)-1){Set_Contact_Boundary(particles.contact(a));}
                        else if(weight>=(T)1-tolerance){Set_Contact_Boundary(particles.contact(c));}
                        else LOG::cout<<"cannot split "<<edge_key<<": "<<weight<<", "<<intersect<<std::endl;
                    }
                }
            }
        }
        else{
            ////split if it is too long
            T particle_length_scale=Triangle_Edge_Length_Scale(a,c);
            T edge_length=(particles.X(a)-particles.X(c)).Magnitude();
            if(edge_length>length_to_split_coef*particle_length_scale){
                edges_to_split_hashtable.Set(edge_key);continue;
            }

            ////split if the opposite angle is large
            T large_angle_cosine=cos((T)(.7*pi));
            for(int i=1;i<=e_tris.m;i++){
                int t=e_tris(i);
                const TV_INT3& v=triangle_mesh.elements(t);
                int b=Other_Node_In_Element(v,a,c);
                T cos_abc=TV::Dot_Product((particles.X(a)-particles.X(b)).Normalized(),(particles.X(c)-particles.X(b)).Normalized());
                if(cos_abc<large_angle_cosine){
                    if(edge_length<(T).2*length_to_split_coef*particle_length_scale)break; ////if edge is too short, do not split, ignore large angle
                    edges_to_split_hashtable.Set(edge_key);break;
                }
            }
        }
    }

    for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_split_hashtable);iter.Valid();iter.Next()){TV_INT2 e=iter.Key();edges_to_split_queue.Enqueue(e);}

    int number_of_splitted_edges=0;
    while(!edges_to_split_queue.Empty()){
        const TV_INT2& e_key=edges_to_split_queue.Dequeue();
        T weight=(T).5;if(use_contact_aware_triangle_remeshing){T* w=contact_boundary_edge_hashtable.Get_Pointer(e_key);if(w)weight=*w;}
        bool splitted=Split_Triangle_Edge(e_key,edge_triangle_hashtable,particles_added,particles_affected_hashtable,triangles_to_remove,triangles_added,weight);
        if(splitted&&split_contact_boundary_edge&&use_contact_aware_triangle_remeshing&&contact_boundary_edge_hashtable.Contains(e_key)){int v=particles.X.m;Set_Contact_Boundary(particles.contact(v));}
        if(splitted){number_of_splitted_edges++;}
    }

    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,0,&particles_affected,0,0,&triangles_added,&triangles_to_remove,0,0);
    else Remove_Particles_And_Elements(0,0,&triangles_to_remove,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Flip_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable=codimensional_mesh_object.mesh.edge_triangle_hashtable;edge_triangle_hashtable.Remove_All();
    Update_Edge_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_triangle_hashtable);
    HASHTABLE<TV_INT2>& edge_on_tet_mesh_hashtable=codimensional_mesh_object.mesh.edge_on_tet_mesh_hashtable;edge_on_tet_mesh_hashtable.Remove_All();
    Update_Edge_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_on_tet_mesh_hashtable);
    //HASHTABLE<int,ARRAY<int> >& rim_vertex_vertex_hashtable=codimensional_mesh_object.mesh.rim_vertex_vertex_hashtable;rim_vertex_vertex_hashtable.Remove_All();
    HASHTABLE<int,ARRAY<int> > rim_vertex_vertex_hashtable;
    ////TODO: implement rim_vtx_vtx_hashtable update
    //Update_Rim_Vertex_Vertex_Hashtable_Threaded(codimensional_mesh_object,mesher_tid,rim_vertex_vertex_hashtable);
    Update_Particle_Dimensional_Size(triangle_mesh.elements,particles.X,regions,domain_ext,particles.dimensional_size);
    
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;

    HASHTABLE<TV_INT2> edges_to_flip_hashtable;
    QUEUE<TV_INT2> edges_to_flip_queue(max(triangle_mesh.elements.m*2,200));

    ARRAY<int> triangles_to_remove;    
    ARRAY<int> triangles_added;
    HASHTABLE<int> particles_affected_hashtable;

    for(HASHTABLE_ITERATOR<TV_INT2,ARRAY<int> > iter(edge_triangle_hashtable);iter.Valid();iter.Next()){
        TV_INT2 e_key=iter.Key();
        ARRAY<int>& e_tri=iter.Data();
        if(e_tri.m!=2)continue; ////only flip manifold

        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles

        int a=e_key.x;int c=e_key.y;
        int i_abc=e_tri(1);int i_acd=e_tri(2);
        TV_INT3 reordered_vtx=Reordered(triangle_mesh.elements(e_tri(1)),a);
        if(reordered_vtx(2)==c){i_abc=e_tri(2);i_acd=e_tri(1);}
        int b=Other_Node_In_Element(triangle_mesh.elements(i_abc),a,c);
        int d=Other_Node_In_Element(triangle_mesh.elements(i_acd),a,c);  

        bool flip_to_improve_triangle_quality=false;
        ARRAY<TV_INT3> old_tris;old_tris.Append(TV_INT3(a,b,c));old_tris.Append(TV_INT3(a,c,d));
        ARRAY<TV_INT3> new_tris;new_tris.Append(TV_INT3(a,b,d));new_tris.Append(TV_INT3(b,c,d));
        T old_measure=Measure_Triangles(old_tris);
        T new_measure=Measure_Triangles(new_tris);
        if(old_measure<new_measure){flip_to_improve_triangle_quality=true;}

        bool flip_to_improve_triangle_thickness=false;bool not_flip_to_preserve_triangle_thickness=false;
        if(use_thickness_aware_triangle_remeshing){
            T ac_thickness_difference=abs(particles.thickness(a)-particles.thickness(c));
            T bd_thickness_difference=abs(particles.thickness(b)-particles.thickness(d));
            T max_factor=(T)3;T min_factor=(T)1/max_factor;
            if(bd_thickness_difference<ac_thickness_difference*(T)min_factor){flip_to_improve_triangle_thickness=true;}
            if(bd_thickness_difference>ac_thickness_difference*(T)max_factor){not_flip_to_preserve_triangle_thickness=true;}
        }

        if(flip_to_improve_triangle_thickness||(flip_to_improve_triangle_quality&&!not_flip_to_preserve_triangle_thickness)){
            TV_INT2 ac_key=TV_INT2(a,c).Sorted();
            if(!edges_to_flip_hashtable.Contains(ac_key)){
                edges_to_flip_queue.Enqueue(ac_key);
                edges_to_flip_hashtable.Insert(ac_key);
            }
        }
    }

    while(!edges_to_flip_queue.Empty()){
        TV_INT2 e_key=edges_to_flip_queue.Dequeue();
        edges_to_flip_hashtable.Delete(e_key);

        ARRAY<TV_INT2> new_edges_to_flip;
        /*bool flipped=*/Flip_Triangle_Edge(e_key,edge_triangle_hashtable,edge_on_tet_mesh_hashtable,rim_vertex_vertex_hashtable,particles_affected_hashtable,triangles_to_remove,triangles_added,new_edges_to_flip);

        for(int i=1;i<=new_edges_to_flip.m;i++){
            TV_INT2 ac_key=new_edges_to_flip(i).Sorted();
            if(!edges_to_flip_hashtable.Contains(ac_key)){
                if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,ac_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
                if(Element_Or(particles.inactive,ac_key))continue;   ////for inactive particles

                edges_to_flip_queue.Enqueue(ac_key);
                edges_to_flip_hashtable.Insert(ac_key);
            }
        }
    }

    Sort(triangles_to_remove);
    triangle_mesh.elements.Remove_Sorted_Indices(triangles_to_remove);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,&particles_affected,0,0,&tris_added,&tris_to_rmv,0,0);
    else Remove_Particles_And_Elements(0,0,&tris_to_rmv,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Triangle_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=codimensional_mesh_object.mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);
    HASHTABLE<TV_INT2>& edge_on_tri_mesh_hashtable=codimensional_mesh_object.mesh.edge_on_tri_mesh_hashtable;edge_on_tri_mesh_hashtable.Remove_All();
    Update_Edge_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_on_tri_mesh_hashtable);
    HASHTABLE<int,ARRAY<int> > rim_vertex_vertex_hashtable;     ////TODO: implement rim_vtx_vtx_hashtable update
    
    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    //QUEUE<TV_INT2> edges_to_collapse_queue(max(triangle_mesh.elements.m*2,200));
    HASHTABLE<TV_INT2> edges_to_collapse_hashtable;

    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    ARRAY<int> triangle_indices_to_remove;
    ARRAY<int> segment_indices_to_remove;

    ////initialize vertex_triangle_hashtable and find edges to collapse

    for(HASHTABLE_ITERATOR<TV_INT2> iter(edge_on_tri_mesh_hashtable);iter.Valid();iter.Next()){
        const TV_INT2& e_key=iter.Key();int a=e_key(1);int c=e_key(2);
        //skip collapse if ac is an edge on the tetrahedron mesh
        bool is_ac_tet_edge=false;
        if(vertex_tetrahedron_hashtable.Contains(a)){
            ARRAY<int>& a_tet=vertex_tetrahedron_hashtable.Get(a);
            for(int j=1;j<=a_tet.m;j++){
                const TV_INT4& v=tetrahedron_mesh.elements(a_tet(j));
                if(v.Contains(c)){is_ac_tet_edge=true;break;}
            }
            if(is_ac_tet_edge)continue;
        }

        ////add to collapse edge queue if it is short
        T particle_length_scale=Triangle_Edge_Length_Scale(a,c);
        if((particles.X(a)-particles.X(c)).Magnitude()<(T)length_to_collapse_coef*particle_length_scale){
            edges_to_collapse_hashtable.Set(e_key);
            //edges_to_collapse_queue.Enqueue(e_key);
        }
    }

    int number_of_collapsed_edge=0;
    while(edges_to_collapse_hashtable.Size()>0){
        ////pick the shortest edge
        T shortest_edge_length=(T)FLT_MAX;TV_INT2 e_key;
        for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_collapse_hashtable);iter.Valid();iter.Next()){
            const TV_INT2& v=iter.Key();T edge_length=(particles.X(v(1))-particles.X(v(2))).Magnitude();
            if(edge_length<shortest_edge_length){shortest_edge_length=edge_length;e_key=v;}
        }
        edges_to_collapse_hashtable.Delete(e_key);

        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles

        bool collapsed=Collapse_Triangle_Edge(e_key,vertex_triangle_hashtable,vertex_tetrahedron_hashtable,vertex_segment_hashtable,rim_vertex_vertex_hashtable,
            particles_to_remove,particles_added,particles_affected_hashtable,triangle_indices_to_remove,segment_indices_to_remove);

        if(collapsed){
            number_of_collapsed_edge++;
            ////replace a and c with v in edge queue
            int a=e_key(1);int c=e_key(2);int v=particles_added(particles_added.m);
            ARRAY<TV_INT2> keys_to_delete;ARRAY<TV_INT2> keys_to_add;
            for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_collapse_hashtable);iter.Valid();iter.Next()){
                TV_INT2 old_e_key=iter.Key();TV_INT2 new_e_key=Replaced_Two(old_e_key,a,c,v).Sorted();
                if(new_e_key!=old_e_key){
                    //edges_to_collapse_hashtable.Delete(old_e_key);edges_to_collapse_hashtable.Set(new_e_key);
                    keys_to_delete.Append(old_e_key);keys_to_add.Append(new_e_key);
                }
            }

            //////check new short edges
            //ARRAY<int>* v_tris=vertex_triangle_hashtable.Get_Pointer(v);
            //if(v_tris){
            //    for(int i=1;i<=(*v_tris).m;i++){
            //        int t=(*v_tris)(i);const TV_INT3& vtx=triangle_mesh.elements(t);
            //        ARRAY<TV_INT2> edges;Simplex_Edges(vtx,edges);
            //        for(int j=1;j<=3;j++){const TV_INT2& e=edges(j);
            //            T particle_length_scale=use_adaptive_triangle_remeshing?(particles.length_scale(e(1))+particles.length_scale(e(2))):length_scale;
            //            if((particles.X(e(1))-particles.X(e(2))).Magnitude()<(T)length_to_collapse_coef*particle_length_scale){keys_to_add.Append(e.Sorted());}
            //        }

            //    }
            //}

            for(int i=1;i<=keys_to_delete.m;i++){edges_to_collapse_hashtable.Delete(keys_to_delete(i));}
            for(int i=1;i<=keys_to_add.m;i++){edges_to_collapse_hashtable.Set(keys_to_add(i));}
        }
    }

    Sort(triangle_indices_to_remove);
    triangle_mesh.elements.Remove_Sorted_Indices(triangle_indices_to_remove);
    Sort(segment_indices_to_remove);
    segment_mesh.elements.Remove_Sorted_Indices(segment_indices_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);
    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    //////dbg
    //if(segs_to_rmv.m>0||segs_added.m>0){
    //    LOG::cout<<"N"<<mpi_codim->node_id<<" col_tri_edge seg_to_rmv: "<<segs_to_rmv<<", seg_added: "<<segs_added<<", p_added: "<<particles_added<<", p_to_rmv: "<<particles_to_remove<<std::endl;
    //    for(int i=1;i<=segs_to_rmv.m;i++){
    //        const TV_INT2& v=segment_mesh.elements(segs_to_rmv(i));
    //        //LOG::cout<<"s_to_rmv: "<<v<<" guid "<<Element(particles.guid,v)<<std::endl;
    //    }
    //    for(int i=1;i<=segs_added.m;i++){
    //        const TV_INT2& v=segment_mesh.elements(segs_added(i));
    //        LOG::cout<<"s_added: "<<v<<" guid "<<Element(particles.guid,v)<<std::endl;
    //    }
    //}
    
    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,&particles_affected,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    ARRAY<int> segs_added;
    ARRAY<int> segs_to_rmv;

    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& v=segment_mesh.elements(i);int a=v(1);int c=v(2);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles
        if(has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))))continue;     ////split fail: should not collapse any eulerian particle
        if(use_mesh_edge_connect){TV_INT2 guid_key=Element(particles.guid,v).Sorted();if(mesh.connected_segment_hashtable.Contains(guid_key))continue;} ////connect fail, has connected segment

        T segment_length=(particles.X(a)-particles.X(c)).Magnitude();
        T particle_length_scale=Segment_Edge_Length_Scale(a,c);
        if(segment_length>(T)length_to_split_coef*particle_length_scale){
            int v=Split_Particles(a,c,TV_INT(1,0,0),(T).5*(particles.X(a)+particles.X(c)),(T).5);
            particles.thickness(v)=sqrt((T).5*(pow<2>(particles.thickness(a))+pow<2>(particles.thickness(c))));
            particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);
            particles.on_boundary(v)(1)=0; ////fix the on_boundary for segment particles

            int s1=segment_mesh.elements.Append(TV_INT2(a,v));
            int s2=segment_mesh.elements.Append(TV_INT2(c,v));
            
            particles_added.Append(v);segs_to_rmv.Append(i);
            segs_added.Append(s1);segs_added.Append(s2);
        }
    }

    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,0,&particles_affected,0,0,0,0,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(0,0,0,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    QUEUE<int> segments_to_collapse_queue(max(segment_mesh.elements.m*2,200));
    HASHTABLE<int> segments_to_collapse_hashtable;
    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    ARRAY<int> segment_indices_to_remove;

    for(int i=1;i<=segment_mesh.elements.m;i++){
        TV_INT2 e_key=segment_mesh.elements(i).Sorted();int a=e_key(1);int c=e_key(2);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles

        T segment_length=(particles.X(a)-particles.X(c)).Magnitude();
        T particle_length_scale=Segment_Edge_Length_Scale(a,c);
        if(segment_length<(T)length_to_collapse_coef*particle_length_scale){
            segments_to_collapse_hashtable.Set(i);
            segments_to_collapse_queue.Enqueue(i);
        }
    }

    while(!segments_to_collapse_queue.Empty()){
        int s_i=segments_to_collapse_queue.Dequeue();
        Collapse_Segment_Edge(s_i,vertex_triangle_hashtable,vertex_tetrahedron_hashtable,vertex_segment_hashtable,particles_to_remove,particles_added,segment_indices_to_remove);
    }

    ARRAY<int> triangles_added;
    ARRAY<int> segments_added;

    Sort(segment_indices_to_remove);
    segment_mesh.elements.Remove_Sorted_Indices(segment_indices_to_remove);

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);

    //////dbg
    //LOG::cout<<" tets_to_rmv N"<<mpi_codim->node_id<<": "<<tets_to_rmv<<std::endl;
    //if(tets_to_rmv.m>0)for(int i=1;i<=tets_to_rmv.m;i++)LOG::cout<<"t_2_rmv N"<<mpi_codim->node_id<<": "<<tetrahedron_mesh.elements(tets_to_rmv(i))<<", guid "<<Element(particles.guid,old_tets(tets_to_rmv(i)));LOG::cout<<std::endl;
    //LOG::cout<<" tets_added N"<<mpi_codim->node_id<<": "<<tets_added<<std::endl;
    //if(tets_added.m>0)for(int i=1;i<=tets_added.m;i++)LOG::cout<<"t_added N"<<mpi_codim->node_id<<": "<<tetrahedron_mesh.elements(tets_added(i))<<", guid "<<Element(particles.guid,old_tets(tets_added(i)));LOG::cout<<std::endl;

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,0,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Triangles()
{
    for(int i=1;i<=sub_regions.m;i++){
        Break_Thin_Triangles_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Triangles_Constrained_Boundary()
{
    for(int i=1;i<=sub_regions.m;i++){
        Break_Thin_Triangles_Constrained_Boundary_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Segments()
{
    for(int i=1;i<=sub_regions.m;i++){
        Break_Thin_Segments_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Segments_Constrained_Boundary()
{
    for(int i=1;i<=sub_regions.m;i++){
        Break_Thin_Segments_Constrained_Boundary_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Rim_Segment_Mesh()
{
    rim_segment_mesh.elements.Remove_All();
    for(int i=1;i<=sub_regions.m;i++){
        Update_Rim_Segment_Mesh_Subregion(sub_regions(i));
    }
    rim_segment_mesh.Initialize_Incident_Elements();
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Remove_Codim_Elements_In_The_Air()
{
    for(int i=1;i<=sub_regions.m;i++){
        Remove_Codim_Elements_In_The_Air_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Remove_Invalid_Elements()
{
    for(int i=1;i<=sub_regions.m;i++){
        Remove_Invalid_Elements_Subregion(sub_regions(i));
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Remove_Invalid_Elements_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T y_threshold=(T)2*length_scale;

    ARRAY<int> invalid_tets;
    HASHTABLE<TV_INT4,int> tet_hashtable;
    HASHTABLE<TV_INT3,ARRAY<int> > face_tet_hashtable;
    for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
        const TV_INT4& v=tetrahedron_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

        ////check negtive volume tet
        bool is_negative_volume=Is_Non_Positive_Volume_Tetrahedron(particles.X,v);
        if(is_negative_volume){invalid_tets.Append(i);continue;}

        ////check duplicated tet
        const TV_INT4& t_key=Unique_Ordered(v);
        const TV_INT4& reverted_t_key=Unique_Ordered(Reversed(v));
        if(!tet_hashtable.Contains(t_key)&&!tet_hashtable.Contains(reverted_t_key)){tet_hashtable.Insert(t_key,i);}
        else{invalid_tets.Append(i);continue;}

        ////check non-manifold face
        ARRAY<TV_INT3> faces;Simplex_Faces(v,faces);
        bool has_nonmanifold_face=false;
        for(int j=1;j<=Simplex_Face_Number<4>();j++){
            TV_INT3 f_key=faces(j).Sorted();TV_INT3 f_key_guid=Element(particles.guid,f_key);
            ARRAY<int>& f_e=face_tet_hashtable.Get_Or_Insert(f_key,ARRAY<int>());
            if(f_e.m>=2){has_nonmanifold_face=true;/*LOG::cout<<"nonmanifold face: "<<f_key_guid<<std::endl;*/break;}
        }

        if(has_nonmanifold_face){invalid_tets.Append(i);continue;}
        else{
            for(int j=1;j<=Simplex_Face_Number<4>();j++){
                TV_INT3 f_key=faces(j).Sorted();TV_INT3 f_key_guid=Element(particles.guid,f_key);
                ARRAY<int>& f_e=face_tet_hashtable.Get_Or_Insert(f_key,ARRAY<int>());f_e.Append(i);
            }
        }
    }

    ARRAY<int> invalid_tris;
    HASHTABLE<TV_INT3,int> tri_hashtable;
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        const TV_INT3& v=triangle_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

        ////check duplicated tri
        const TV_INT3& t_key=Unique_Ordered(v);
        const TV_INT3& reverted_t_key=Unique_Ordered(Reversed(v));
        if(!tri_hashtable.Contains(t_key)&&!tri_hashtable.Contains(reverted_t_key)){tri_hashtable.Insert(t_key,i);}
        else{invalid_tris.Append(i);continue;}
    }

    ARRAY<int> invalid_segments;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        TV_INT2 e_key=segment_mesh.elements(i).Sorted();
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles
        if(e_key(1)==e_key(2)){invalid_segments.Append(i);}
    }

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,&invalid_tets,0,&invalid_tris,0,&invalid_segments);
    else Remove_Particles_And_Elements(0,&invalid_tets,&invalid_tris,&invalid_segments);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Remove_Codim_Elements_In_The_Air_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    //T y_threshold=(T)2*length_scale;

    ARRAY<int> triangles_in_the_air;
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        const TV_INT3& v=triangle_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles
        if(particles.X(v(1)).y>y_in_the_air_min&&particles.X(v(2)).y>y_in_the_air_min&&particles.X(v(3)).y>y_in_the_air_min){triangles_in_the_air.Append(i);}
    }

    ARRAY<int> segments_in_the_air;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        TV_INT2 e_key=segment_mesh.elements(i).Sorted();//int a=e_key(1);int c=e_key(2);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles
        if(particles.X(e_key(1)).y>y_in_the_air_min&&particles.X(e_key(2)).y>y_in_the_air_min){segments_in_the_air.Append(i);}
    }

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,0,0,&triangles_in_the_air,0,&segments_in_the_air);
    else Remove_Particles_And_Elements(0,0,&triangles_in_the_air,&segments_in_the_air);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Triangles_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    ARRAY<int> thin_triangles;
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        const TV_INT3& v=triangle_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

        T triangle_thickness=(T)0;
        if(preserve_thickness_in_remeshing)triangle_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
        else triangle_thickness=(T)one_third*(Discrete_Particle_Thickness(v(1),has_rim)+Discrete_Particle_Thickness(v(2),has_rim)+Discrete_Particle_Thickness(v(3),has_rim));

        if(triangle_thickness<min_thickness){thin_triangles.Append(i);continue;}
    }

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,0,0,&thin_triangles,0,0);
    else Remove_Particles_And_Elements(0,0,&thin_triangles,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Segments_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);

    ARRAY<int> thin_segments;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        TV_INT2 e_key=segment_mesh.elements(i).Sorted();int a=e_key(1);int c=e_key(2);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles

        T segment_thickness=(T)0;
        if(preserve_thickness_in_remeshing)segment_thickness=(T).5*(particles.thickness(a)+particles.thickness(c));
        else segment_thickness=(T).5*(Discrete_Particle_Thickness(a,has_rim)+Discrete_Particle_Thickness(c,has_rim));     ////TODO: may have problem for mpi

        if(segment_thickness<min_rim_thickness){thin_segments.Append(i);}
    }

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,0,0,0,0,&thin_segments);
    else Remove_Particles_And_Elements(0,0,0,&thin_segments);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Rim_Segment_Mesh_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<TV_INT2,int> edge_triangle_number_hashtable;
    Update_Edge_Element_Number_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,edge_triangle_number_hashtable);
    HASHTABLE<TV_INT2> edge_on_tetrahedron_hashtable;
    Update_Edge_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,edge_on_tetrahedron_hashtable);

    ARRAY<int> rim_segs_on_boundary_region;
    for(HASHTABLE_ITERATOR<TV_INT2,int> iter(edge_triangle_number_hashtable);iter.Valid();iter.Next()){
        TV_INT2 e_key=iter.Key();
        if(iter.Data()<2){
            if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
            if(edge_on_tetrahedron_hashtable.Contains(e_key))continue;

            int s=rim_segment_mesh.elements.Append(e_key);
            
            if(mpi_codim&&(mpi_codim->Inside_Domain_Boundary(particles.X(e_key(1)))||mpi_codim->Inside_Domain_Boundary(particles.X(e_key(2))))&&
                (!mpi_codim->Inside_Domain_Shrink(particles.X(e_key(1)))&&!mpi_codim->Inside_Domain_Shrink(particles.X(e_key(2))))){
                rim_segs_on_boundary_region.Append(s);
            }
        }
    }

    if(mpi_codim){
        ARRAY<ARRAY<TV_INT2> > send_segs;ARRAY<int> send_seg_numbers;ARRAY<ARRAY<TV_INT2> > recv_segs;ARRAY<int> recv_seg_numbers;
        mpi_codim->Add_Elements(rim_segment_mesh.elements,rim_segs_on_boundary_region,send_segs,send_seg_numbers,recv_segs,recv_seg_numbers);
    }
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Neighbor_Particle(const int p1,const int p2,const HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,const HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,const HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable)
{
    const ARRAY<int>* nb_tets=vertex_tetrahedron_hashtable.Get_Pointer(p1);
    if(nb_tets)for(int i=1;i<=nb_tets->m;i++){
        const TV_INT4& t=tetrahedron_mesh.elements((*nb_tets)(i));
        if(t.Contains(p2))return true;
    }
    const ARRAY<int>* nb_tris=vertex_triangle_hashtable.Get_Pointer(p1);
    if(nb_tris)for(int i=1;i<=nb_tris->m;i++){
        const TV_INT3& t=triangle_mesh.elements((*nb_tris)(i));
        if(t.Contains(p2))return true;
    }
    const ARRAY<int>* nb_segs=vertex_segment_hashtable.Get_Pointer(p1);
    if(nb_segs)for(int i=1;i<=nb_segs->m;i++){
        const TV_INT2& s=segment_mesh.elements((*nb_segs)(i));
        if(s.Contains(p2))return true;
    }
    return false;
}
template<class T2> class ELEMENT_WITH_THICKNESS{
public:T2 thickness;int index;ELEMENT_WITH_THICKNESS<T2>(T2 thickness_input=(T2)0,int index_input=0):thickness(thickness_input),index(index_input){}
       bool operator < (const ELEMENT_WITH_THICKNESS<T2>& t){return thickness<t.thickness;}
       bool operator < (const ELEMENT_WITH_THICKNESS<T2>& t)const {return thickness<t.thickness;}};

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Triangles_Constrained_Boundary_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);

    ARRAY<ELEMENT_WITH_THICKNESS<T> > thin_triangles;
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        const TV_INT3& v=triangle_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

        T triangle_thickness=(T)0;
        if(preserve_thickness_in_remeshing)triangle_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
        else triangle_thickness=(T)one_third*(Discrete_Particle_Thickness(v(1),has_rim)+Discrete_Particle_Thickness(v(2),has_rim)+Discrete_Particle_Thickness(v(3),has_rim));

        if(triangle_thickness<min_thickness){thin_triangles.Append(ELEMENT_WITH_THICKNESS<T>(triangle_thickness,i));}
    }
    Sort(thin_triangles);

    int number_of_rings=3;
    ARRAY<int> tris_to_rmv;
    ARRAY<bool> flag(particles.X.m);flag.Fill(false);
    for(int i=1;i<=particles.X.m;i++){
        if(particles.on_boundary(i)(2)!=0){
            ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(triangle_mesh.elements,vertex_triangle_hashtable,i,nbs,number_of_rings);
            for(int k=1;k<=nbs.m;k++)flag(nbs(k))=true;
        }
    }

    for(int i=1;i<=thin_triangles.m;i++){int t=thin_triangles(i).index;
        const TV_INT3& v=triangle_mesh.elements(t);
        VECTOR<bool,3> v_flag=Element(flag,v);

        if(Is_Contact_With_Solid(particles.contact(v(1)))&&Is_Contact_With_Solid(particles.contact(v(2)))&&Is_Contact_With_Solid(particles.contact(v(3)))){continue;} ////break fail if tri contact with solid

        bool remove_tri=!v_flag.Contains(true);

        if(use_enforce_thin_tri_break_constrained&&!remove_tri){    ////remove tri if it's too thin
            T triangle_thickness=(T)0;
            if(preserve_thickness_in_remeshing)triangle_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
            else triangle_thickness=(T)one_third*(Discrete_Particle_Thickness(v(1),has_rim)+Discrete_Particle_Thickness(v(2),has_rim)+Discrete_Particle_Thickness(v(3),has_rim));

            T epsilon=(T)1e-3;
            if(triangle_thickness<min_thickness*(T).5*((T)1+epsilon))remove_tri=true;
        }

        if(remove_tri){
            //TV_INT3 v=triangle_mesh.elements(t);
            //TV_INT3 guid=Element(particles.guid,v);
            //TV tk=Element(particles.thickness,v);
            //LOG::cout<<"rmv thin tri: "<<guid<<", thk: "<<tk<<", min_thk: "<<min_thickness<<", thk_scale: "<<thickness_scale<<std::endl;

            tris_to_rmv.Append(t);
            for(int j=1;j<=3;j++){
                ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(triangle_mesh.elements,vertex_triangle_hashtable,v(j),nbs,number_of_rings);
                for(int k=1;k<=nbs.m;k++)flag(nbs(k))=true;
            }
        }
    }

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,0,0,&tris_to_rmv,0,0);
    else Remove_Particles_And_Elements(0,0,&tris_to_rmv,0);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Break_Thin_Segments_Constrained_Boundary_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);

    ARRAY<ELEMENT_WITH_THICKNESS<T> > thin_segments;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& v=segment_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(Element_Or(particles.inactive,v))continue;   ////for inactive particles

        T segment_thickness=(T)0;
        if(preserve_thickness_in_remeshing)segment_thickness=(T).5*(particles.thickness(v(1))+particles.thickness(v(2)));
        else segment_thickness=(T).5*(Discrete_Particle_Thickness(v(1),has_rim)+Discrete_Particle_Thickness(v(2),has_rim));     ////TODO: may have problem for mpi

        if(segment_thickness<min_rim_thickness){
            thin_segments.Append(ELEMENT_WITH_THICKNESS<T>(segment_thickness,i));
            TV_INT2 guid=Element(particles.guid,v);
        }
    }
    Sort(thin_segments);

    int number_of_rings=3;
    ARRAY<int> segs_to_rmv;
    ARRAY<bool> flag(particles.X.m);flag.Fill(false);
    for(int i=1;i<=particles.X.m;i++){      
        if(particles.on_boundary(i)(1)!=0){
            ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(segment_mesh.elements,vertex_segment_hashtable,i,nbs,number_of_rings);
            for(int k=1;k<=nbs.m;k++)flag(nbs(k))=true;
        }
    }

    for(int i=1;i<=thin_segments.m;i++){int t=thin_segments(i).index;
        const TV_INT2& v=segment_mesh.elements(t);
        VECTOR<bool,2> v_flag=Element(flag,v);

        bool remove_seg=!v_flag.Contains(true);
        if(!remove_seg){        ////remove tri if it's too thin
            T segment_thickness=(T)0;
            if(preserve_thickness_in_remeshing)segment_thickness=(T).5*(particles.thickness(v(1))+particles.thickness(v(2)));
            else segment_thickness=(T).5*(Discrete_Particle_Thickness(v(1),has_rim)+Discrete_Particle_Thickness(v(2),has_rim));     ////TODO: may have problem for mpi

            T epsilon=(T)1e-3;
            if(segment_thickness<min_rim_thickness*(T).5*((T)1+epsilon))remove_seg=true;
        }

        if(remove_seg){
            if(use_allowed_break_seg_num){
                if(break_seg_num>=allowed_break_seg_num_max)break;
                break_seg_num++;
            }

            segs_to_rmv.Append(t);
            for(int j=1;j<=2;j++){
                ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(segment_mesh.elements,vertex_segment_hashtable,v(j),nbs,number_of_rings);
                for(int k=1;k<=nbs.m;k++)flag(nbs(k))=true;
            }
        }
    }

    if(use_allowed_break_seg_num){
        LOG::cout<<"break thin seg "<<break_seg_num<<" in "<<allowed_break_seg_num_max<<std::endl;
    }
    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(0,0,0,0,0,0,0,0,&segs_to_rmv);
    else Remove_Particles_And_Elements(0,0,0,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Boundary_Vertex_Snap_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;

    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV> spatial_partition;
    spatial_partition.Set_Voxel_Size(length_scale*(T)1.5);
    int object_id=1;
    HASHTABLE<int> boundary_vertex_hashtable;
    for(int i=1;i<=particles.X.m;i++){
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,VECTOR<int,1>(i),regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        if(particles.on_boundary(i)!=TV_INT3()){boundary_vertex_hashtable.Set(i);}
    }
    for(HASHTABLE_ITERATOR<int> iter(boundary_vertex_hashtable);iter.Valid();iter.Next()){
        int p=iter.Key();RANGE<TV_INT> voxel_range;spatial_partition.Insert_Into_Hashtable(particles.X,TV_INT1(p),p,object_id,voxel_range,(T)0);
    }

    HASHTABLE<TV_INT2> edges_to_snap_hashtable;
    for(HASHTABLE_ITERATOR<int> iter(boundary_vertex_hashtable);iter.Valid();iter.Next()){
        const int p=iter.Key();
        ARRAY<CODIMENSIONAL_ELEMENT_INDEX> element_indices;
        spatial_partition.Get_From_Cell_With_One_Ring_Neighbors(particles.X(p),element_indices);
        for(int i=1;i<=element_indices.m;i++){
            int nb=element_indices(i).index;
            if(nb!=p&&!Is_Neighbor_Particle(p,nb,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable)){
                T distance=(particles.X(p)-particles.X(nb)).Magnitude();
                T particle_length_scale=Triangle_Edge_Length_Scale(p,nb);
                if(distance<particle_length_scale){edges_to_snap_hashtable.Set(TV_INT2(p,nb).Sorted());}
            }
        }
    }

    int number_of_snapped_edge=0;
    while(edges_to_snap_hashtable.Size()>0){
        ////pick the shortest edge
        T shortest_edge_length=(T)FLT_MAX;TV_INT2 e_key;
        for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_snap_hashtable);iter.Valid();iter.Next()){
            const TV_INT2& v=iter.Key();T edge_length=(particles.X(v(1))-particles.X(v(2))).Magnitude();
            if(edge_length<shortest_edge_length){shortest_edge_length=edge_length;e_key=v;}
        }
        edges_to_snap_hashtable.Delete(e_key);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles
        if(Is_Neighbor_Particle(e_key(1),e_key(2),vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable)){continue;}

        bool snapped=Snap_Boundary_Vertex(e_key(1),e_key(2),vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,particles_to_remove,particles_added,particles_affected_hashtable);
        if(snapped){
            number_of_snapped_edge++;
            ////replace a and c with v in edge queue
            int a=e_key(1);int c=e_key(2);int v=particles_added(particles_added.m);
            ARRAY<TV_INT2> keys_to_delete;ARRAY<TV_INT2> keys_to_add;
            for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_snap_hashtable);iter.Valid();iter.Next()){
                TV_INT2 old_e_key=iter.Key();TV_INT2 new_e_key=Replaced_Two(old_e_key,a,c,v).Sorted();
                if(new_e_key!=old_e_key){keys_to_delete.Append(old_e_key);keys_to_add.Append(new_e_key);}
            }
            for(int i=1;i<=keys_to_delete.m;i++){edges_to_snap_hashtable.Delete(keys_to_delete(i));}
            for(int i=1;i<=keys_to_add.m;i++){edges_to_snap_hashtable.Set(keys_to_add(i));}

            ////delete a and c and and v in spatial partition
            for(int i=1;i<=2;i++)spatial_partition.Remove_From_Cell(spatial_partition.Voxel(particles.X(e_key(i))),CODIMENSIONAL_ELEMENT_INDEX(e_key(i),(SIMPLEX_TYPE)1,object_id));
            RANGE<TV_INT> voxel_range;spatial_partition.Insert_Into_Hashtable(particles.X,TV_INT1(v),v,object_id,voxel_range,(T)0);
        }
    }

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,0,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Codimensional_Vertex_Snap_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    T domain_ext=mpi_codim?(T)2*mpi_codim->dX:(T)0;
    HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable=mesh.vertex_tetrahedron_hashtable;vertex_tetrahedron_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(tetrahedron_mesh.elements,particles.X,regions,domain_ext,vertex_tetrahedron_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable=mesh.vertex_triangle_hashtable;vertex_triangle_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(triangle_mesh.elements,particles.X,regions,domain_ext,vertex_triangle_hashtable);
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable=mesh.vertex_segment_hashtable;vertex_segment_hashtable.Remove_All();
    Update_Vertex_Element_Hashtable(segment_mesh.elements,particles.X,regions,domain_ext,vertex_segment_hashtable);

    ARRAY<TV_INT4> old_tets=tetrahedron_mesh.elements;
    ARRAY<TV_INT3> old_tris=triangle_mesh.elements;
    ARRAY<TV_INT2> old_segs=segment_mesh.elements;

    ARRAY<int> particles_to_remove;
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;

    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV> spatial_partition;
    spatial_partition.Set_Voxel_Size(length_scale*(T)1.5);
    int object_id=1;

    HASHTABLE<int> boundary_vertex_hashtable;
    for(int i=1;i<=particles.X.m;i++){
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,VECTOR<int,1>(i),regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        bool on_tet_surface=(particles.on_boundary(i)(3)!=0);
        bool on_tri=vertex_triangle_hashtable.Contains(i);
        bool on_seg=vertex_segment_hashtable.Contains(i);
        if(on_tet_surface||on_tri||on_seg){boundary_vertex_hashtable.Set(i);}
    }
    for(HASHTABLE_ITERATOR<int> iter(boundary_vertex_hashtable);iter.Valid();iter.Next()){
        int p=iter.Key();RANGE<TV_INT> voxel_range;spatial_partition.Insert_Into_Hashtable(particles.X,TV_INT1(p),p,object_id,voxel_range,(T)0);
    }

    HASHTABLE<TV_INT2> edges_to_snap_hashtable;
    for(HASHTABLE_ITERATOR<int> iter(boundary_vertex_hashtable);iter.Valid();iter.Next()){
        const int p=iter.Key();if(!vertex_triangle_hashtable.Contains(p))continue;

        ARRAY<CODIMENSIONAL_ELEMENT_INDEX> element_indices;
        spatial_partition.Get_From_Cell_With_One_Ring_Neighbors(particles.X(p),element_indices);
        for(int i=1;i<=element_indices.m;i++){
            int nb=element_indices(i).index;
            bool valid_to_snap=vertex_triangle_hashtable.Contains(p)||vertex_tetrahedron_hashtable.Contains(p);
            if(!valid_to_snap)continue;

            if(nb!=p&&!Is_Neighbor_Particle(p,nb,vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable)){
                T distance=(particles.X(p)-particles.X(nb)).Magnitude();
                T particle_length_scale=Triangle_Edge_Length_Scale(p,nb);
                if(distance<particle_length_scale){edges_to_snap_hashtable.Set(TV_INT2(p,nb).Sorted());}
            }
        }
    }

    int number_of_snapped_edge=0;
    while(edges_to_snap_hashtable.Size()>0){
        ////pick the shortest edge
        T shortest_edge_length=(T)FLT_MAX;TV_INT2 e_key;
        for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_snap_hashtable);iter.Valid();iter.Next()){
            const TV_INT2& v=iter.Key();T edge_length=(particles.X(v(1))-particles.X(v(2))).Magnitude();
            if(edge_length<shortest_edge_length){shortest_edge_length=edge_length;e_key=v;}
        }
        edges_to_snap_hashtable.Delete(e_key);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,e_key,regions,(T)0))continue;   ////for mpi, make sure has at least one vertex inside regions
        if(Element_Or(particles.inactive,e_key))continue;   ////for inactive particles
        if(Is_Neighbor_Particle(e_key(1),e_key(2),vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable)){continue;}

        bool snapped=Snap_Boundary_Vertex(e_key(1),e_key(2),vertex_tetrahedron_hashtable,vertex_triangle_hashtable,vertex_segment_hashtable,particles_to_remove,particles_added,particles_affected_hashtable);
        if(snapped){
            number_of_snapped_edge++;
            ////replace a and c with v in edge queue
            int a=e_key(1);int c=e_key(2);int v=particles_added(particles_added.m);
            ARRAY<TV_INT2> keys_to_delete;ARRAY<TV_INT2> keys_to_add;
            for(HASHTABLE_ITERATOR<TV_INT2> iter(edges_to_snap_hashtable);iter.Valid();iter.Next()){
                TV_INT2 old_e_key=iter.Key();TV_INT2 new_e_key=Replaced_Two(old_e_key,a,c,v).Sorted();
                if(new_e_key!=old_e_key){keys_to_delete.Append(old_e_key);keys_to_add.Append(new_e_key);}
            }
            for(int i=1;i<=keys_to_delete.m;i++){edges_to_snap_hashtable.Delete(keys_to_delete(i));}
            for(int i=1;i<=keys_to_add.m;i++){edges_to_snap_hashtable.Set(keys_to_add(i));}

            ////delete a and c and and v in spatial partition
            for(int i=1;i<=2;i++)spatial_partition.Remove_From_Cell(spatial_partition.Voxel(particles.X(e_key(i))),CODIMENSIONAL_ELEMENT_INDEX(e_key(i),(SIMPLEX_TYPE)1,object_id));
            RANGE<TV_INT> voxel_range;spatial_partition.Insert_Into_Hashtable(particles.X,TV_INT1(v),v,object_id,voxel_range,(T)0);
        }
    }

    ARRAY<int> tets_to_rmv;ARRAY<int> tets_added;
    Update_Added_And_To_Remove_Element_Indices(old_tets,tetrahedron_mesh.elements,tets_to_rmv,tets_added);
    ARRAY<int> tris_to_rmv;ARRAY<int> tris_added;
    Update_Added_And_To_Remove_Element_Indices(old_tris,triangle_mesh.elements,tris_to_rmv,tris_added);
    ARRAY<int> segs_to_rmv;ARRAY<int> segs_added;
    Update_Added_And_To_Remove_Element_Indices(old_segs,segment_mesh.elements,segs_to_rmv,segs_added);

    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,&particles_to_remove,0,&tets_added,&tets_to_rmv,&tris_added,&tris_to_rmv,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(&particles_to_remove,&tets_to_rmv,&tris_to_rmv,&segs_to_rmv);
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Connected_By_Segment(const int s1,const int s2,const HASHTABLE<TV_INT2>& segment_hashtable,/*rst*/TV_INT2& free_pair)
{
    const TV_INT2& vtx1=segment_mesh.elements(s1);const TV_INT2& vtx2=segment_mesh.elements(s2);
    ARRAY<int> contact_1;for(int i=1;i<=2;i++)if(Is_Contact_With_Fluid(particles.contact(vtx1(i))))contact_1.Append(vtx1(i));
    ARRAY<int> contact_2;for(int i=1;i<=2;i++)if(Is_Contact_With_Fluid(particles.contact(vtx2(i))))contact_2.Append(vtx2(i));
    for(int i=1;i<=contact_1.m;i++){int p1=contact_1(i);
        for(int j=1;j<=contact_2.m;j++){int p2=contact_2(j);
            TV_INT2 key=TV_INT2(p1,p2).Sorted();
            if(segment_hashtable.Contains(key)){
                int free_p1=p1==vtx1(1)?vtx1(2):vtx1(1);int free_p2=p2==vtx2(1)?vtx2(2):vtx2(1);
                TV_INT2 free_key=TV_INT2(free_p1,free_p2).Sorted();
                if(free_key(1)!=free_key(2)&&!Is_Contact_With_Fluid(particles.contact(free_key(1)))
                    &&!Is_Contact_With_Fluid(particles.contact(free_key(2)))&&!segment_hashtable.Contains(free_key))free_pair=TV_INT2(free_p1,free_p2);
                return true;
            }
        }
    }
    return false;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Segment_Intersect(const TV_INT2& pair_key,/*rst*/TV2& pair_weight)
{
    int s1=pair_key(1);int s2=pair_key(2);
    const TV_INT2& vtx_1=segment_mesh.elements(s1);const TV_INT2& vtx_2=segment_mesh.elements(s2);
    SEGMENT_3D<T> seg_1(particles.X(vtx_1(1)),particles.X(vtx_1(2)));SEGMENT_3D<T> seg_2(particles.X(vtx_2(1)),particles.X(vtx_2(2)));
    TV2 seg_weights;TV3 shortest_vector=seg_1.Shortest_Vector_Between_Segments(seg_2,seg_weights);T distance=shortest_vector.Magnitude();
    if(seg_weights(1)<(T)0||seg_weights(1)>(T)1||seg_weights(2)<(T)0||seg_weights(2)>(T)1)return false;
    T epsilon=(T)1e-2;if((seg_weights(1)<=epsilon||seg_weights(1)>=(T)1-epsilon)&&(seg_weights(2)<=epsilon||seg_weights(2)>=(T)1-epsilon))return false;
    T thickness_1=(T).5*(particles.thickness(vtx_1(1))+particles.thickness(vtx_1(2)));T thickness_2=(T).5*(particles.thickness(vtx_2(1))+particles.thickness(vtx_2(2)));
    if(distance<(T).5*(thickness_1+thickness_2)){pair_weight=seg_weights;return true;}
    return false;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Vertex_Segment_Intersect(const int p,const int s1,const int s2,/*rst*/T& weight)
{
    const TV_INT2& vtx_1=segment_mesh.elements(s1);const TV_INT2& vtx_2=segment_mesh.elements(s2);
    T w2=SEGMENT_3D<T>::Interpolation_Fraction(particles.X(p),particles.X(vtx_2(1)),particles.X(vtx_2(2)));if(w2<(T)0||w2>(T)1)return false;
    TV pos_on_seg=((T)1-w2)*particles.X(vtx_2(1))+w2*particles.X(vtx_2(2));
    T distance=(particles.X(p)-pos_on_seg).Magnitude();
    T thickness_1=(T).5*(particles.thickness(vtx_1(1))+particles.thickness(vtx_1(2)));T thickness_2=(T).5*(particles.thickness(vtx_2(1))+particles.thickness(vtx_2(2)));
    //LOG::cout<<"p: "<<p<<", s1: "<<s1<<", vtx1: "<<vtx_1<<", s2: "<<s2<<", vtx2: "<<vtx_2<<", w2: "<<w2<<", dis: "<<distance<<", "<<(T).5*(thickness_1+thickness_2)<<std::endl;
    if(distance<(T).5*(thickness_1+thickness_2)){weight=w2;return true;}
    return false;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Segment_Mesh_Edge_Connect_Subregion(const ARRAY<RANGE<TV> >& regions)
{
    ARRAY<int> particles_added;
    HASHTABLE<int> particles_affected_hashtable;
    ARRAY<int> segs_added;
    ARRAY<int> segs_to_rmv;
    HASHTABLE<TV_INT2> segment_hashtable;
    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV> spatial_partition;
    HASHTABLE<TV_INT2>& connected_segment_hashtable=mesh.connected_segment_hashtable;

    spatial_partition.Set_Voxel_Size(length_scale*(T)1.5);int object_id=1;
    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& v=segment_mesh.elements(i);
        if(mpi_codim&&!Has_Vertex_Inside_Regions(particles.X,v,regions,(T)0))continue;////for mpi boundaries, at least one vertex is inside the region
        T seg_thickness=(T).5*(particles.thickness(v(1))+particles.thickness(v(2)));
        RANGE<TV_INT> voxel_range;spatial_partition.Insert_Into_Hashtable(particles.X,v,i,object_id,voxel_range,seg_thickness);
    }

    ////find all contacting edge-edge pairs and vertex-edge pairs
    HASHTABLE<TV_INT2,TV2> contacting_segment_pairs_hashtable;TV2 invalid_weight=TV2::Constant_Vector((T)-100000);
    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& seg_v=segment_mesh.elements(i);
        segment_hashtable.Set(seg_v.Sorted());
        T seg_thickness=(T).5*(particles.thickness(seg_v(1))+particles.thickness(seg_v(2)));
        const RANGE<TV_INT>& voxel_range=spatial_partition.Voxel(RANGE<TV>::Bounding_Box(particles.X(seg_v(1)),particles.X(seg_v(2))).Thickened(seg_thickness));
        ARRAY<CODIMENSIONAL_ELEMENT_INDEX> potential_collision_elements;
        spatial_partition.Get_From_Cell(voxel_range,1,potential_collision_elements);
        for(int j=1;j<=potential_collision_elements.m;j++){
            const CODIMENSIONAL_ELEMENT_INDEX& element=potential_collision_elements(j);
            int s=element.index;if(s<=i)continue;
            const TV_INT2& nb_seg_v=segment_mesh.elements(s);
            if(seg_v.Contains(nb_seg_v(1))||seg_v.Contains(nb_seg_v(2)))continue;
            //bool intersect=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(particles.X,particles.X,particles.thickness,particles.thickness,seg_v,nb_seg_v);
            //if(intersect){contacting_segment_pairs.Append(TV_INT2(i,s).Sorted());/*LOG::cout<<"contact seg pair "<<i<<", "<<s<<std::endl;*/}
            //SEGMENT_3D<T> seg_1(particles.X(seg_v(1)),particles.X(seg_v(2)));SEGMENT_3D<T> seg_2(particles.X(nb_seg_v(1)),particles.X(nb_seg_v(2)));
            //TV2 seg_weights;TV3 shortest_vector=seg_1.Shortest_Vector_Between_Segments(seg_2,seg_weights);T distance=shortest_vector.Magnitude();
            //T epsilon=(T)1e-10;if((seg_weights(1)<=epsilon||seg_weights(1)>=(T)1-epsilon)&&(seg_weights(2)<=epsilon||seg_weights(2)>=(T)1-epsilon))continue;
            //T thickness_1=(T).5*(particles.thickness(seg_v(1))+particles.thickness(seg_v(2)));T thickness_2=(T).5*(particles.thickness(nb_seg_v(1))+particles.thickness(nb_seg_v(2)));
            //if(distance<(T).5*(thickness_1+thickness_2)){
            //    TV_INT2 pair_key=TV_INT2(i,s).Sorted();
            //    if(pair_key!=TV_INT2(i,s)){T tmp=seg_weights(1);seg_weights(1)=seg_weights(2);seg_weights(2)=tmp;}
            //    contacting_segment_pairs_hashtable.Set(pair_key,seg_weights);/*LOG::cout<<"contact seg pair "<<i<<", "<<s<<std::endl;*/
            //}
            TV_INT2 pair_key=TV_INT2(i,s).Sorted();TV2 pair_weights;bool intersect=Segment_Segment_Intersect(pair_key,pair_weights);
            if(intersect){
                T y_threshold=(T).001;  ////for cheese falling example only
                if(particles.X(seg_v(1)).y>y_threshold||particles.X(seg_v(2)).y>y_threshold||particles.X(nb_seg_v(1)).y>y_threshold||particles.X(nb_seg_v(2)).y>y_threshold)continue;

                contacting_segment_pairs_hashtable.Set(pair_key,pair_weights);
            }
        }
    }

    //for(int i=1;i<=contacting_segment_pairs.m;i++){
    //for(HASHTABLE_ITERATOR<TV_INT2,TV2> iter(contacting_segment_pairs_hashtable);iter.Valid();iter.Next()){
    while(contacting_segment_pairs_hashtable.Size()>0){
        HASHTABLE_ITERATOR<TV_INT2,TV2> iter(contacting_segment_pairs_hashtable);
        TV_INT2 pair_key=iter.Key();TV2 pair_weights=iter.Data();TV_INT2 connect_p;
        VECTOR<bool,2> segment_splitted(false,false);VECTOR<TV_INT2,2> segment_pair_added;
        contacting_segment_pairs_hashtable.Delete(pair_key);

        const TV_INT2& seg_vtx_1=segment_mesh.elements(pair_key(1));
        const TV_INT2& seg_vtx_2=segment_mesh.elements(pair_key(2));
        TV_INT2 seg_guid_key_1=Element(particles.guid,seg_vtx_1).Sorted();
        TV_INT2 seg_guid_key_2=Element(particles.guid,seg_vtx_2).Sorted();
        if(connected_segment_hashtable.Contains(seg_guid_key_1)||connected_segment_hashtable.Contains(seg_guid_key_2))continue; ////connect fail, has connected segment
        if(!segment_hashtable.Contains(seg_vtx_1.Sorted())||!segment_hashtable.Contains(seg_vtx_2.Sorted()))continue;   ////connect fail, has invalid segment
        
        bool has_short_seg=false;       ////connect fail: short segments
        for(int d=1;d<=2;d++){
            const TV_INT2& seg_vtx=segment_mesh.elements(pair_key(d));
            T particle_length_scale=use_adaptive_segment_remeshing?(T).5*(particles.length_scale(seg_vtx(1))+particles.length_scale(seg_vtx(2))):length_scale;
            T seg_length=(particles.X(seg_vtx(1))-particles.X(seg_vtx(2))).Magnitude();
            if(seg_length<=length_to_collapse_coef*particle_length_scale){has_short_seg=true;break;}
        }
        if(has_short_seg)continue;

        TV_INT2 free_pair;bool connected=Connected_By_Segment(pair_key(1),pair_key(2),segment_hashtable,free_pair);
        if(connected){
            continue;   ////............
            if(free_pair==TV_INT2())continue;    ////connect fail, no free particle
            else{////TODO: handle incident segments
                if(use_dbg)LOG::cout<<"free pair: "<<free_pair<<" guid "<<Element(particles.guid,free_pair)<<std::endl;
                pair_weights(1)=free_pair(1)==seg_vtx_1(1)?(T)0:(T)1;
                pair_weights(2)=free_pair(2)==seg_vtx_2(1)?(T)0:(T)1;

                bool p1_intersect_s2=false;bool p2_intersect_s1=false;
                p1_intersect_s2=Vertex_Segment_Intersect(free_pair(1),pair_key(1),pair_key(2),pair_weights(2));
                if(!p1_intersect_s2)p2_intersect_s1=Vertex_Segment_Intersect(free_pair(2),pair_key(2),pair_key(1),pair_weights(1));
                if(!p1_intersect_s2&&!p2_intersect_s1)continue; ////connect fail: vertex does not intersect segment

                if(use_dbg)LOG::cout<<"free_pair intersect: "<<p1_intersect_s2<<", "<<p2_intersect_s1<<std::endl;
                //T frac_1=SEGMENT_3D<T>::Interpolation_Fraction(particles.X(free_pair(1)),particles.X(seg_vtx_2(1)),particles.X(seg_vtx_2(2)));
                //if(frac_1>(T)0&&frac_1<(T)1){   ////free vtx 1 intersects seg2
                //    T distance=
                //    pair_weights(2)=frac_1;
                //}
                //else{
                //    T frac_2=SEGMENT_3D<T>::Interpolation_Fraction(particles.X(free_pair(2)),particles.X(seg_vtx_1(1)),particles.X(seg_vtx_1(2)));
                //    if(frac_2>(T)0&&frac_2<(T)1)pair_weights(1)=frac_2;
                //}
            }
        }
        else{
            //if(segs_to_rmv.Contains(pair_key(1))||segs_to_rmv.Contains(pair_key(2)))continue;

            //bool has_connnected_particle=false;for(int j=1;j<=2;j++){
            //    int s=pair_key(j);int a=segment_mesh.elements(s)(1);int c=segment_mesh.elements(s)(2);
            //    if(Is_Contact_With_Fluid(particles.contact(a))||Is_Contact_With_Fluid(particles.contact(c))){has_connnected_particle=true;break;}
            //}
            //if(has_connnected_particle)continue;

            //for(int j=1;j<=2;j++){
            //    int s=pair_key(j);int a=segment_mesh.elements(s)(1);int c=segment_mesh.elements(s)(2);
            //    LOG::cout<<"contact pair "<<j<<": "<<a<<", "<<c<<std::endl;
            //}

            ////recompute pair weights
            if(pair_weights==invalid_weight){bool intersect=Segment_Segment_Intersect(pair_key,pair_weights);if(!intersect)continue;}
        }

        ////split segments
        T tolerance=(T).1;
        for(int j=1;j<=2;j++){
            int s=pair_key(j);int a=segment_mesh.elements(s)(1);int c=segment_mesh.elements(s)(2);
            if(pair_weights(j)>=tolerance&&pair_weights(j)<=(T)1-tolerance){
                TV v_pos=((T)1-pair_weights(j))*particles.X(a)+pair_weights(j)*particles.X(c);
                connect_p(j)=Split_Particles(a,c,TV_INT(1,0,0),v_pos,pair_weights(j));
                particles.thickness(connect_p(j))=sqrt((T).5*(pow<2>(particles.thickness(a))+pow<2>(particles.thickness(c))));
                particles.V(connect_p(j))=(T).5*(particles.V(a)+particles.V(c));
                particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);
                particles.on_boundary(connect_p(j))(1)=0; ////fix the on_boundary for segment particles

                int s1=segment_mesh.elements.Append(TV_INT2(a,connect_p(j)));segment_hashtable.Set(TV_INT2(a,connect_p(j)).Sorted());
                int s2=segment_mesh.elements.Append(TV_INT2(c,connect_p(j)));segment_hashtable.Set(TV_INT2(c,connect_p(j)).Sorted());
                particles_added.Append(connect_p(j));segs_to_rmv.Append(s);segment_hashtable.Delete(TV_INT2(a,c).Sorted());
                segs_added.Append(s1);segs_added.Append(s2);segment_splitted(j)=true;segment_pair_added(j)=TV_INT2(s1,s2);
            }
            else if(pair_weights(j)<tolerance){connect_p(j)=a;}
            else if(pair_weights(j)>(T)1-tolerance){connect_p(j)=c;}

            Set_Contact_With_Fluid(particles.contact(connect_p(j)));
        }

        if(use_dbg)LOG::cout<<"pair_key: "<<pair_key<<", weight: "<<pair_weights<<", free_pair: "<<free_pair<<", conn: "<<connected<<std::endl;
        ////add new connecting segment
        int s_v=segment_mesh.elements.Append(TV_INT2(connect_p(1),connect_p(2)));
        segment_hashtable.Set(TV_INT2(connect_p(1),connect_p(2)).Sorted());
        connected_segment_hashtable.Set(TV_INT2(particles.guid(connect_p(1)),particles.guid(connect_p(2))).Sorted());
        segs_added.Append(s_v);
        if(use_dbg)LOG::cout<<"add new connecting seg: "<<TV_INT2(connect_p(1),connect_p(2))<<" guid "<<TV_INT2(particles.guid(connect_p(1)),particles.guid(connect_p(2)))<<std::endl;

        //////add splitted segments to hashtable
        //if(segment_splitted(1)&&segment_splitted(2)){
        //    int s1=segment_pair_added(1)(1);int s2=segment_pair_added(1)(2);
        //    int s3=segment_pair_added(2)(1);int s4=segment_pair_added(2)(2);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s1,s3).Sorted(),invalid_weight);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s1,s4).Sorted(),invalid_weight);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s2,s3).Sorted(),invalid_weight);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s2,s4).Sorted(),invalid_weight);
        //}
        //else if(segment_splitted(1)||segment_splitted(2)){
        //    int splitted_j=segment_splitted(1)?1:2;int other_j=splitted_j==1?2:1;
        //    int s1=segment_pair_added(splitted_j)(1);int s2=segment_pair_added(splitted_j)(2);int s3=pair_key(other_j);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s1,s3).Sorted(),invalid_weight);
        //    contacting_segment_pairs_hashtable.Set(TV_INT2(s2,s3).Sorted(),invalid_weight);
        //}

        ////replace segments in hashtable
        ARRAY<TV_INT2> keys_to_rmv;ARRAY<TV_INT2> keys_to_add;
        for(HASHTABLE_ITERATOR<TV_INT2,TV2> iter(contacting_segment_pairs_hashtable);iter.Valid();iter.Next()){
            TV_INT2 key2=iter.Key();int splitted_j=0;if(segment_splitted(1)&&key2.Contains(pair_key(1)))splitted_j=1;else if(segment_splitted(2)&&key2.Contains(pair_key(2)))splitted_j=2;
            if(splitted_j!=0){
                keys_to_rmv.Append(key2);
                int splitted_j=key2.Contains(pair_key(1))?1:2;int other_j=splitted_j==1?2:1;
                int s1=segment_pair_added(splitted_j)(1);int s2=segment_pair_added(splitted_j)(2);int s3=key2(other_j);
                keys_to_add.Append(TV_INT2(s1,s3).Sorted());
                keys_to_add.Append(TV_INT2(s2,s3).Sorted());
            }
        }
        for(int i=1;i<=keys_to_add.m;i++){contacting_segment_pairs_hashtable.Set(keys_to_add(i),invalid_weight);/*LOG::cout<<"hashtable set: "<<keys_to_add(i)<<std::endl;*/}
        for(int i=1;i<=keys_to_rmv.m;i++){contacting_segment_pairs_hashtable.Delete(keys_to_rmv(i));}
    }

    ARRAY<int> particles_affected;for(HASHTABLE_ITERATOR<int> iter(particles_affected_hashtable);iter.Valid();iter.Next())particles_affected.Append(iter.Key());
    if(mpi_codim)mpi_codim->Exchange_Particles_And_Elements(&particles_added,0,&particles_affected,0,0,0,0,&segs_added,&segs_to_rmv);
    else Remove_Particles_And_Elements(0,0,0,&segs_to_rmv);
}

template<class TV> template<class TV_INTN>
void PhysBAM::CODIMENSIONAL_MESHING_3D<TV>::Update_Added_And_To_Remove_Element_Indices(const ARRAY<TV_INTN>& old_e,ARRAY<TV_INTN>& new_e,ARRAY<int>& e_to_rmv,ARRAY<int>& e_added)
{
    HASHTABLE<TV_INTN> old_e_hashtable;
    HASHTABLE<TV_INTN> new_e_hashtable;

    for(int i=1;i<=new_e.m;i++){new_e_hashtable.Set(Unique_Ordered(new_e(i)));}

    ////traverse old_e, if not in new_e, remove
    ARRAY<TV_INTN> e_vtx_to_rmv;
    for(int i=1;i<=old_e.m;i++){
        TV_INTN e=Unique_Ordered(old_e(i));
        old_e_hashtable.Set(e);
        if(!new_e_hashtable.Contains(e)){
            e_vtx_to_rmv.Append(e);
        }
    }

    ////traverse new_e, if not in old_e, add
    for(int i=1;i<=new_e.m;i++){
        TV_INTN e=Unique_Ordered(new_e(i));
        if(!old_e_hashtable.Contains(e)){
            e_added.Append(i);
        }
    }

    ////add e_to_rmv to new_e
    for(int i=1;i<=e_vtx_to_rmv.m;i++){
        int idx=new_e.Append(e_vtx_to_rmv(i));
        e_to_rmv.Append(idx);
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Particle_Distance_To_Rim()
{
    ARRAY<bool> rim_particle;
    ARRAY<bool> visited;
    QUEUE<int> queue(particles.X.m);
    TRIANGLE_MESH triangles;//the union of the triangle mesh and the tetrahedron surface mesh

    for(int i=1;i<=triangle_mesh.elements.m;i++)triangles.elements.Append(triangle_mesh.elements(i));
    for(int i=1;i<=tetrahedron_mesh.boundary_mesh->elements.m;i++)triangles.elements.Append(tetrahedron_mesh.boundary_mesh->elements(i));
    for(int i=1;i<=particles.X.m;i++){particles.distance_to_rim(i)=-1;visited.Append(false);}

    triangles.number_nodes=particles.X.m;
    triangles.Initialize_Neighbor_Nodes();

    for(int p=1;p<=particles.X.m;p++){
        rim_particle.Append(codimensional_mesh_object.mesh.Is_Rim_Particle(p));
    }

    //if using mpi, discard false positives for rim particles
    if(mpi_codim){
        ARRAY_VIEW<bool> rim_particle_array_view(rim_particle);
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(rim_particle_array_view);
    }

    for(int p=1;p<=particles.X.m;p++){
        if(!rim_particle(p))continue;
        visited(p)=true;
        particles.distance_to_rim(p)=0;
        queue.Enqueue(p);
    }

    while(!queue.Empty()){
        int start_index=queue.Dequeue();
        ARRAY<int>& neighbors=(*triangles.neighbor_nodes)(start_index);

        for(int n=1;n<=neighbors.m;n++){
            int neighbor_index=neighbors(n);
            if(visited(neighbor_index))continue;
            visited(neighbor_index)=true;
            particles.distance_to_rim(neighbor_index)=particles.distance_to_rim(start_index)+1;
            queue.Enqueue(neighbor_index);
        }
    }

    //get shared particles that may have smaller distance_to_rim from other meshes
    if(mpi_codim){
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.distance_to_rim);

        //second pass
        for(int i=1;i<=particles.X.m;i++)visited(i)=false;

        for(int p=1;p<=particles.X.m;p++){
            if(!rim_particle(p))continue;
            visited(p)=true;
            queue.Enqueue(p);
        }

        while(!queue.Empty()){
            int start_index=queue.Dequeue();
            ARRAY<int>& neighbors=(*triangles.neighbor_nodes)(start_index);

            for(int n=1;n<=neighbors.m;n++){
                int neighbor_index=neighbors(n);
                if(visited(neighbor_index))continue;
                visited(neighbor_index)=true;
                particles.distance_to_rim(neighbor_index)=min(particles.distance_to_rim(neighbor_index),particles.distance_to_rim(start_index)+1);
                queue.Enqueue(neighbor_index);
            }
        }
    }//if mpi_codim
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Particle_Area_Weighted_Normal()
{
    for(int i=1;i<=particles.X.m;i++){
        particles.area_weighted_normal(i)=codimensional_mesh_object.Nodal_Area_Weighted_Normal(i);
    }

    if(mpi_codim){mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(particles.area_weighted_normal);}
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Particle_Normal()
{
    for(int i=1;i<=particles.X.m;i++){
        particles.normal(i)=codimensional_mesh_object.Nodal_Normal(i);
    }
    ////not need mpi sync?
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Edge_On_Mesh(const TV_INT2& edge,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable)
{
    int a=edge(1);int c=edge(2);
    if(!vertex_triangle_hashtable.Contains(a)||!vertex_triangle_hashtable.Contains(c))return false;
    ARRAY<int>& a_tris=vertex_triangle_hashtable.Get(a);ARRAY<int>& c_tris=vertex_triangle_hashtable.Get(c);
    if(Shared_Element_Index(a_tris,c_tris)!=0)return true;
    if(!vertex_tetrahedron_hashtable.Contains(a)||!vertex_tetrahedron_hashtable.Contains(c))return false;
    ARRAY<int>& a_tets=vertex_tetrahedron_hashtable.Get(a);ARRAY<int>& c_tets=vertex_tetrahedron_hashtable.Get(c);
    if(Shared_Element_Index(a_tets,c_tets)!=0)return true;
    return false;
}

template class CODIMENSIONAL_MESHING_3D<float>;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class CODIMENSIONAL_MESHING_3D<double>;
#endif
