//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESHING_3D__
#define __CODIMENSIONAL_MESHING_3D__

#include <PhysBAM_Geometry/Topology_Based_Geometry_Computations/DELAUNAY_MESHING_WITH_ALPHA_SHAPE.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include <PhysBAM_Tools/Parallel_Computation/THREADED_UNIFORM_GRID.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include "../Meshing/CODIMENSIONAL_MESHING_BASE.h"
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Misc/WATER_TANK.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
using namespace Codimensional;

template<class T_input>
class CODIMENSIONAL_MESHING_3D : public CODIMENSIONAL_MESHING_BASE<VECTOR<T_input,3> >
{
    typedef T_input T;typedef VECTOR<T,3> TV;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,1> TV_INT1;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,SEGMENTED_CURVE_2D<T>,TRIANGULATED_SURFACE<T> >::TYPE T_SURFACE_MESH_OBJECT;
    typedef typename IF<TV::m==2,SEGMENT_MESH,TRIANGLE_MESH>::TYPE SURFACE_MESH;
    typedef typename IF<TV::m==2,TRIANGULATED_AREA<T>,TETRAHEDRALIZED_VOLUME<T> >::TYPE T_VOLUME_MESH_OBJECT;
    typedef typename IF<TV::m==2,TRIANGLE_MESH,TETRAHEDRON_MESH>::TYPE VOLUME_MESH;
    typedef CODIMENSIONAL_MESHING_BASE<TV> BASE;
public:
    using BASE::codimensional_mesh_object;using BASE::particles;using BASE::mesh;using BASE::segment_mesh;using BASE::triangle_mesh;using BASE::tetrahedron_mesh;using BASE::rim_segment_mesh;using BASE::length_scale;
    using BASE::adaptive_length_scale_min;using BASE::thickness_scale;using BASE::length_to_split_coef;using BASE::length_to_collapse_coef;
    using BASE::density;using BASE::max_particle_radius;using BASE::min_particle_radius;using BASE::max_particle_volume;using BASE::min_particle_volume;using BASE::max_particle_mass;using BASE::min_particle_mass;using BASE::incompressible_segment_length;using BASE::incompressible_triangle_area;using BASE::max_thickness;
    using BASE::min_thickness;using BASE::max_rim_thickness;using BASE::min_rim_thickness;using BASE::max_thin_simplex_volume;using BASE::epsilon;using BASE::smooth_rim_thickness_weight;using BASE::smooth_interior_thickness_weight;using BASE::smooth_rim_thickness_mode;using BASE::rim_interior_mass_transfer_coefficient;
    using BASE::particle_bounding_box;using BASE::particle_grid;using BASE::particles_in_cell; using BASE::segments_in_cell;using BASE::triangles_in_cell;using BASE::tetrahedra_in_cell;
    using BASE::use_low_dimensional_incompressibility;using BASE::use_thickness_based_low_dimensional_incompressibility;using BASE::use_distance_to_rim_for_incompressibility;using BASE::min_distance_to_rim_for_incompressibility;using BASE::segment_incompressibilities;using BASE::triangle_incompressibilities;
    using BASE::use_thickness_based_dimension_lifting;using BASE::use_crumple_based_dimension_lifting;using BASE::use_crumple_merge;using BASE::use_reseed;using BASE::use_tetrahedron_mesh;using BASE::use_triangle_mesh;using BASE::use_segment_mesh;using BASE::initialize_tetrahedron_mesh;using BASE::initialize_triangle_mesh;using BASE::initialize_segment_mesh;using BASE::use_break_thin_segments;using BASE::use_break_thin_triangles;using BASE::use_remove_bad_tetrahedra;
    using BASE::use_topology_repairing;using BASE::has_tank;using BASE::has_solid;using BASE::has_eulerian_particles;using BASE::use_rim_adaptivity;using BASE::use_smooth_rim_thickness;using BASE::use_smooth_interior_thickness;using BASE::use_align_triangle_normals;
    using BASE::use_rim_mass_redistribution;using BASE::use_break_elements_in_solids;using BASE::use_high_res_volume_mesh;using BASE::use_high_res_segment_mesh;using BASE::use_elements_merging;using BASE::use_elements_merging_to_triangle_mesh;using BASE::use_elements_merging_to_tetrahedron_mesh;
    using BASE::use_remove_low_dim_elements_on_tank_bottom;using BASE::use_continuous_meshing_operations;using BASE::has_rim;using BASE::use_adaptive_tetrahedron_remeshing;using BASE::use_adaptive_triangle_remeshing;using BASE::use_adaptive_segment_remeshing;
    using BASE::preserve_thickness_in_remeshing;using BASE::use_thickness_aware_triangle_remeshing;using BASE::use_boundary_vertex_snap;using BASE::use_codim_vertex_snap;using BASE::use_contact_aware_triangle_remeshing;using BASE::use_anisotropic_triangle_remeshing;using BASE::use_mesh_edge_connect;using BASE::use_thin_sheet_aware_tetrahedron_remeshing;
    using BASE::use_constrained_thin_triangle_break;using BASE::use_constrained_thin_segment_break;using BASE::use_enforce_thin_tri_break_constrained;using BASE::use_particle_parents;using BASE::use_fix_poping;using BASE::use_split_eulerian_particles;using BASE::use_remove_codim_elements_in_the_air;using BASE::use_remove_invalid_elements;using BASE::mpi_codim;using BASE::sub_regions;using BASE::random;
    using BASE::rigid_bodies;using BASE::kinematic_rigid_bodies;using BASE::eulerian_particles_hashtable;using BASE::added_eulerian_particle_parents_hashtable;using BASE::fiber;using BASE::water_tank;using BASE::use_dbg;using BASE::dbg_number;using BASE::timer;//using BASE::solid_contacting_particles_hashtable;
    using BASE::Update_Particle_Attributes_After_Having_Mesh_Topology;using BASE::Update_Particle_Geometry_Attributes;
    using BASE::Initialize_Length_Scale;using BASE::Initialize_Particle_Mass_Based_On_Geometry;using BASE::Update_Mesh_Incompressibilities;using BASE::Update_Particle_Incompressibilities;using BASE::Update_Segment_Incompressibilities;using BASE::Update_Triangle_Incompressibilities;
    using BASE::Update_Incompressibility_Mesh;using BASE::Update_Viscosity_Mesh;
    using BASE::Update_Particle_Mass_Based_On_Geometry;using BASE::Update_Particle_Distance_To_Rim;using BASE::Update_Particle_Thickness;using BASE::Break_Thin_Segments;using BASE::Smooth_Rim_Thickness;using BASE::Smooth_Interior_Thickness;using BASE::Add_Solid;using BASE::Initialize_Tank;
    using BASE::Add_Eulerian_Particles;using BASE::Clear_Eulerian_Particles;
    using BASE::Get_Particle_Bounding_Box;using BASE::Get_Particle_Grid;using BASE::Resize_Elements_In_Cell_With_Grid;
    using BASE::Node_Triangle_Area_Sum;using BASE::Node_Segment_Length_Sum;
    using BASE::Check_Mass;

    CODIMENSIONAL_MESHING_3D(CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    virtual void Update_Mesh_Using_Topology_Repair();
    virtual void Update_Mesh_Topology();    ////topology repair part 1
    virtual void Update_Mesh_States();      ////topology repair part 2
    virtual void Update_Mesh();
    virtual void Initialize_Sub_Regions();
    virtual void Break_Thin_Segments();
    virtual void Break_Thin_Segments_Constrained_Boundary();
    virtual void Break_Thin_Triangles();
    virtual void Break_Thin_Triangles_Constrained_Boundary();
    void Update_Rim_Segment_Mesh();

    void Tetrahedron_Mesh_Edge_Split();
    void Tetrahedron_Mesh_Face_Split();
    void Tetrahedron_Mesh_Edge_Flip();
    void Tetrahedron_Mesh_Face_Flip();
    void Tetrahedron_Mesh_Edge_Collapse();
    void Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse();
    void Triangle_Mesh_Edge_Split();
    void Triangle_Mesh_Edge_Flip();
    void Triangle_Mesh_Edge_Collapse();
    void Segment_Mesh_Edge_Split();
    void Segment_Mesh_Edge_Collapse();
    void Boundary_Vertex_Snap();
    void Segment_Mesh_Edge_Connect();
    void Triangle_Mesh_Contact_Edge_Split();
    void Codimensional_Vertex_Snap();
    void Remove_Codim_Elements_In_The_Air();    ////for paint trace example only
    void Remove_Invalid_Elements();

    bool Check_Tetrahedron_Mesh();
    bool Check_Triangle_Mesh();
    bool Check_Segment_Mesh();
    void Reset_Particle_Parents();
    void Update_Particle_Parents_Hashtable();

protected:
    HASHTABLE<TV_INT2> tet_edge_guid_to_collapse_hashtable;

    void Tetrahedron_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Tetrahedron_Mesh_Face_Split_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Tetrahedron_Mesh_Edge_Flip_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Tetrahedron_Mesh_Face_Flip_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Tetrahedron_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Tetrahedron_Mesh_Skinny_Tetrahedron_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Triangle_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions,bool split_contact_boundary_edge=false);
    void Triangle_Mesh_Edge_Flip_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Triangle_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Segment_Mesh_Edge_Split_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Segment_Mesh_Edge_Collapse_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Boundary_Vertex_Snap_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Segment_Mesh_Edge_Connect_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Codimensional_Vertex_Snap_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Remove_Codim_Elements_In_The_Air_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Remove_Invalid_Elements_Subregion(const ARRAY<RANGE<TV> >& regions);

    void Break_Thin_Triangles_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Break_Thin_Triangles_Constrained_Boundary_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Break_Thin_Segments_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Break_Thin_Segments_Constrained_Boundary_Subregion(const ARRAY<RANGE<TV> >& regions);
    void Update_Rim_Segment_Mesh_Subregion(const ARRAY<RANGE<TV> >& regions);
    
    template<class TV_INTN> void Update_Added_And_To_Remove_Element_Indices(const ARRAY<TV_INTN>& old_e,ARRAY<TV_INTN>& new_e,ARRAY<int>& e_to_rmv,ARRAY<int>& e_added);
    void Remove_Particles_And_Elements(ARRAY<int>* p_to_rmv,ARRAY<int>* tet_to_rmv,ARRAY<int>* tri_to_rmv,ARRAY<int>* seg_to_rmv);
    //void Remove_Sorted_Particles(const ARRAY<int>& particles_to_remove,bool remove_invalid_elements=false);

    ////topology repairing atomic operations
    bool Split_Tetrahedron_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,
        ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,
        const T split_weight=(T).5,bool force_split=false,HASHTABLE<int,ARRAY<int> >* vertex_tetrahedron_hashtable=0,HASHTABLE<int,ARRAY<int> >* vertex_triangle_hashtable=0,HASHTABLE<int,ARRAY<int> >* vertex_segment_hashtable=0,
        HASHTABLE<TV_INT3,ARRAY<int> >* face_tetrahedron_hashtable=0);
    bool Split_Tetrahedron_Face(const TV_INT3& face,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,
        ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,
        const TV3& split_weight=TV3((T)one_third,(T)one_third,(T)one_third),bool force_split=false,HASHTABLE<int,ARRAY<int> >* vertex_tetrahedron_hashtable=0,HASHTABLE<int,ARRAY<int> >* vertex_triangle_hashtable=0,HASHTABLE<int,ARRAY<int> >* vertex_segment_hashtable=0,
        HASHTABLE<TV_INT2,ARRAY<int> >* edge_tetrahedron_hashtable=0,HASHTABLE<TV_INT2,ARRAY<int> >* edge_triangle_hashtable=0);
    bool Flip_Tetrahedron_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,
        HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT2>& affected_edges);
    bool Flip_Tetrahedron_Boundary_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,HASHTABLE<TV_INT2>& boundary_edge_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,
        HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT2>& affected_edges);
    bool Flip_Tetrahedron_Face(const TV_INT3& face,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,
        HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT3>& affected_faces);
    bool Collapse_Tetrahedron_Edge(const TV_INT2& edge,const int edge_type,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,
        ARRAY<int>& particles_to_remove,ARRAY<int>& ear_particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangle_indices_to_remove,ARRAY<int>& tetrahedron_indices_to_remove,
        bool force_split=false,HASHTABLE<TV_INT2,ARRAY<int> >* edge_tetrahedron_hashtable=0,HASHTABLE<TV_INT2,ARRAY<int> >* edge_triangle_hashtable=0);
    bool Split_Triangle_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,
        ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,
        const T split_weight=(T).5);
    bool Flip_Triangle_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,const HASHTABLE<TV_INT2>& edge_on_tet_mesh_hashtable,HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable,
        HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,/*result*/ARRAY<TV_INT2>& new_edges_to_flip);
    bool Collapse_Triangle_Edge(const TV_INT2& edge,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable,
        ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangle_indices_to_remove,ARRAY<int>& segment_indices_to_remove);
    bool Collapse_Segment_Edge(const int segment_index,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,
        ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,ARRAY<int>& segment_indices_to_remove);
    bool Snap_Boundary_Vertex(const int a,const int c,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,
        ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable);
    
    bool Predict_Atomic_Cap_Tetrahedron_Collapse(const int tet_idx,const TV_INT3& face,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable);
    bool Predict_Atomic_Spade_Tetrahedron_Collapse(const TV_INT2& edge,const int d,const T w,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable);
    bool Predict_Atomic_Sliver_Tetrahedron_Collapse(const TV_INT2& e1,const TV_INT2& e2,const T w1,const T w2,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable);
    bool Valid_Tetrahedron_Edge_Collapse(const int a,const int c,const int guid_a,const int guid_c,const int on_boundary_a,const int on_boundary_c,
        const int contact_a,const int contact_c,const TV& X_a,const TV& X_c,const ARRAY<TV_INT4>& new_one_ring_tets,bool force_split,/*rst*/TV& v_pos,const T old_vol=(T)-1,const ARRAY<int>* ac_tets=0);
    
    ////mass and momentum redistribution
    void Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Split(const int a,const int c,const int v,const ARRAY<int>& e_tets,const ARRAY<int>& e_tris);
    void Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Split(const int a,const int b,const int c,const int v,const ARRAY<int>& e_tets);
    void Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Flip(const int a,const int c,const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& ring_v,const ARRAY<TV_INT4>& flipped_tets_v);
    void Redistribute_Mass_And_Momentum_For_Tetrahedron_Boundary_Edge_Flip(const int a,const int c,const int b,const int p,const int q,const ARRAY<TV_INT4>& tets_v,const ARRAY<TV_INT4>& flipped_tets_v);
    void Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Flip(const ARRAY<TV_INT4>& tets_v,const ARRAY<TV_INT4>& flipped_tets_v);
    void Redistribute_Mass_And_Momentum_For_Triangle_Edge_Split(const int a,const int c,const int v,const ARRAY<int>& e_tris);
    void Redistribute_Mass_And_Momentum_For_Triangle_Edge_Flip(const int a,const int c,const int b,const int d);

    void Simple_Redistribute_Mass_And_Momentum(const int a,const int c,const int v);
    void Simple_Redistribute_Mass_And_Momentum(const int a,const int b,const int c,const int v);

    ////particle splitting and merging operations
    int Split_Particles(const int a,const int c,const TV_INT& dim,const TV& pos,const T weight,bool change_mass_and_v=true);
    int Split_Particles(const int a,const int b,const int c,const TV_INT& dim,const TV& pos,const TV& weights,bool change_mass_and_v=true);
    int Merge_Particles(const int a,const int c,const TV_INT& dim,const TV& pos,const T weight,bool change_mass_and_v=true);
    ////tracking parent particles
    void Split_Particle_Parents(const int a,const int c,const int v,const T weight);
    void Split_Particle_Parents(const int a,const int b,const int c,const int v,const TV& weights);
    void Merge_Particle_Parents(const int a,const int c,const int v,const T weight);
    void Add_Particle_Parent_Helper(const int a,const int v,const T weight);
    void Delete_Particle_Parent_Helper(const int a);

    ////topological repair helper functions
    bool Is_Convex_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& v_ring,const int a,const int c);
    bool Is_Convex_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,/*result*/ARRAY<TV_INT2>& coplanar_edges);
    bool Is_Convex_Pair_Of_Boundary_Tetrahedra(const TV_INT2& boundary_edge,const int& b,const int& p,const int& q);
    bool Get_One_Ring_Vertices_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const int a,const int c,/*result*/ARRAY<int>& vtx_ring);
    bool Get_Flipped_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& v_ring,const int a,const int c,/*result*/ARRAY<TV_INT4>& flipped_v_tets);
    bool Get_23_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT3>& boundary_faces);
    bool Get_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,const TV_INT2& boundary_edge,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT3>& boundary_faces);
    bool Get_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT2& boundary_edge,/*result*/int& b,/*result*/int& p,/*result*/int& q,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT2>& boundary_edges);
    T Measure_Tetrahedra(const ARRAY<TV_INT4>& tets_v);
    T Measure_Flipped_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& v_ring,const int a,const int c);
    T Measure_23_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face);
    T Measure_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,const TV_INT2& boundary_edge);
    T Measure_Triangles(const ARRAY<TV_INT3>& tris_v);
    bool Is_Edge_On_Mesh(const TV_INT2& edge,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable);
    bool Is_Neighbor_Particle(const int p1,const int p2,const HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,const HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,const HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable);
    ////other helper functions
    bool Tetrahedron_Has_Short_Edge(const TV_INT4& vtx,TV_INT2& shortest_edge);
    bool Tetrahedron_Has_Short_Edge_With_Boundary(const TV_INT4& vtx,ARRAY_VIEW<TV_INT>& on_boundary,TV_INT2& shortest_edge);
    bool Tetrahedron_Has_Short_Edge_With_Diagonal(const TV_INT4& vtx,ARRAY_VIEW<TV_INT>& on_boundary,const HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,TV_INT2& shortest_edge);
    bool Is_Diagonal_Edge(const TV_INT2& e,const ARRAY<int>& e_tets);
    T Incident_Tetrahedron_Volume(const int v,const HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable);
    T Incident_Triangle_Area(const int v,const HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable);
    T Incident_Triangle_Area(const TV_INT2& ac,const int v,const ARRAY<int>& tris);     ////replace a, c with v

    //void Align_Triangle_Normals();
    //bool Same_Vertex_Order(int index1,int index2);
    //static bool Triangle_Sort_Greater_Than(const PAIR<int,TV>& lhs,const PAIR<int,TV>& rhs);
    //bool Get_Rim_Neighbor_Particles(HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable,const int p,/*result*/ARRAY<PAIR<int,T> >& nb_p,const int radius=3);
    
    virtual void Update_Particle_Distance_To_Rim();
    virtual void Update_Particle_Area_Weighted_Normal();
    virtual void Update_Particle_Normal();

    T Tetrahedron_Length_Scale(const TV_INT4& v);
    T Tetrahedron_Edge_Length_Scale(const int a,const int c);
    T Triangle_Edge_Length_Scale(const int a,const int c);
    T Segment_Edge_Length_Scale(const int a,const int c);

    bool Connected_By_Segment(const int s1,const int s2,const HASHTABLE<TV_INT2>& segment_hashtable,/*rst*/TV_INT2& free_pair);
    bool Segment_Segment_Intersect(const TV_INT2& pair_key,/*rst*/TV2& pair_weight);
    bool Vertex_Segment_Intersect(const int p,const int s1,const int s2,/*rst*/T& weight); ////p is vertex of segment s1

    T Vertex_Segment_Length_Sum(const int p,const HASHTABLE<int,ARRAY<int> >& vtx_seg_hashtable);
};
}
#endif
