//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

using namespace PhysBAM;
using namespace Codimensional;

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Split tetrahedron edge
////Indices: none(?)
////Particles: added
////Elements: tetrahedron added and deleted, triangle added and deleted
////Topology: one ring of incident tetrahedra of the splitted edge
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Tetrahedron_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,
    HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,
    const T split_weight/*=(T).5*/,bool force_split/*=false*/,HASHTABLE<int,ARRAY<int> >* vertex_tetrahedron_hashtable/*=0*/,HASHTABLE<int,ARRAY<int> >* vertex_triangle_hashtable/*=0*/,HASHTABLE<int,ARRAY<int> >* vertex_segment_hashtable/*=0*/,HASHTABLE<TV_INT3,ARRAY<int> >* face_tetrahedron_hashtable/*=0*/)
{
    TV_INT2 e_key=edge.Sorted();
    ARRAY<int> e_tets=edge_tetrahedron_hashtable.Get(e_key);
    ARRAY<int> e_tris=edge_triangle_hashtable.Contains(e_key)?edge_triangle_hashtable.Get(e_key):ARRAY<int>();
    int a=e_key.x;int c=e_key.y;

    if(a==c)return false;   ////split fail: duplicated edges in the queue
    if(!use_split_eulerian_particles&&has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     ////split fail: should not collapse any eulerian particle

    //////////////////////////////////////////////////////////////////////////
    ////start to split

    ////update particle attributes
    TV v_pos=((T)1-split_weight)*particles.X(a)+split_weight*particles.X(c);
    int v=Split_Particles(a,c,TV_INT(0,1,1),v_pos,split_weight,/*don't change mass and v*/false);
    if(use_split_eulerian_particles&&eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))){
        eulerian_particles_hashtable.Set(particles.guid(v));
        added_eulerian_particle_parents_hashtable.Set(particles.guid(v),PAIR<TV_INT2,T>(TV_INT2(particles.guid(a),particles.guid(c)),split_weight));
    }

    if(use_dbg)LOG::cout<<"split tet edge "<<a<<", "<<c<<" -> "<<v<<" guid "<<particles.guid(a)<<", "<<particles.guid(c)<<" -> "<<particles.guid(v)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    particles_added.Append(v);

    ////fix the on_boundary for the diagonal edge split
    if(particles.on_boundary(v)(TV::m)){////check the sum of dihedral angles
        //T dihedral_angle_sum=(T)0;
        //for(int i=1;i<=e_tets.m;i++){
        //    const TV_INT4& v=tetrahedron_mesh.elements(e_tets(i));
        //    int a_idx=v.Find(a);int c_idx=v.Find(c);
        //    TETRAHEDRON<T> tet(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));
        //    dihedral_angle_sum+=Dihedral_Angle(tet,TV_INT2(a_idx,c_idx));
        //}
        //if(dihedral_angle_sum>(T).9*(T)two_pi){////diagonal edge
        //    particles.on_boundary(v)(TV::m)=0;
        //}
        if(Is_Diagonal_Edge(edge,e_tets))particles.on_boundary(v)(TV::m)=0;
    }

    //////Temporally use averaged v
    if(!preserve_thickness_in_remeshing)/*Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Split(a,c,v,e_tets,e_tris)*/; ////code removed
    else Simple_Redistribute_Mass_And_Momentum(a,c,v);

    particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);

    ////update tet's topology
    for(int i=1;i<=e_tets.m;i++){
        int i_abcd=e_tets(i);TV_INT4 abcd=tetrahedron_mesh.elements(i_abcd);
        TV_INT4 vbcd;TV_INT4 vabd;Two_Splitted_Tetrahedron_Vertices(abcd,a,c,v,vbcd,vabd);
        int i_vbcd=tetrahedron_mesh.elements.Append(vbcd);tetrahedra_added.Append(i_vbcd);
        int i_vabd=tetrahedron_mesh.elements.Append(vabd);tetrahedra_added.Append(i_vabd);

        tetrahedra_to_remove.Append(i_abcd);
        
        //////dbg
        //TV_INT4 guid=Element(particles.guid,vbcd);
        ////if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2018266)){
        //if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2016233)){

        //    LOG::cout<<"find it in split !!!!!!!!!!!!!!!!!!!!!!!!!!!!!: "<<i_vbcd<<" ac: "<<particles.guid(a)<<", "<<particles.guid(c)<<", v: "<<particles.guid(v)<<std::endl;
        //}
        //guid=Element(particles.guid,vabd);
        ////if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2018266)){
        //if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2018233)){

        //    LOG::cout<<"find it in split !!!!!!!!!!!!!!!!!!!!!!!!!!!!!: "<<i_vabd<<" ac: "<<particles.guid(a)<<", "<<particles.guid(c)<<", v: "<<particles.guid(v)<<std::endl;
        //}

        ////update edge_tet_hashtable
        Remove_Element(edge_tetrahedron_hashtable,abcd,i_abcd);
        Add_Element(edge_tetrahedron_hashtable,vbcd,i_vbcd);
        Add_Element(edge_tetrahedron_hashtable,vabd,i_vabd);

        ////update extra hashtables
        if(vertex_tetrahedron_hashtable){
            Remove_Element(*vertex_tetrahedron_hashtable,abcd,i_abcd);
            Add_Element(*vertex_tetrahedron_hashtable,vbcd,i_vbcd);
            Add_Element(*vertex_tetrahedron_hashtable,vabd,i_vabd);
        }
        if(face_tetrahedron_hashtable){
            Remove_Element(*face_tetrahedron_hashtable,abcd,i_abcd);
            Add_Element(*face_tetrahedron_hashtable,vbcd,i_vbcd);
            Add_Element(*face_tetrahedron_hashtable,vabd,i_vabd);
        }
    }
    Remove_Key(edge_tetrahedron_hashtable,e_key);

    ////updates tri's topology
    for(int i=1;i<=e_tris.m;i++){
        int i_acd=e_tris(i);const TV_INT3 acd=triangle_mesh.elements(i_acd);
        TV_INT3 avd;TV_INT3 vcd;Two_Splitted_Triangle_Vertices(acd,a,c,v,avd,vcd);
        int i_avd=triangle_mesh.elements.Append(avd);triangles_added.Append(i_avd);
        int i_vcd=triangle_mesh.elements.Append(vcd);triangles_added.Append(i_vcd);
        triangles_to_remove.Append(i_acd);

        ////update edge_tri_hashtable
        Remove_Element(edge_triangle_hashtable,acd,i_acd);
        Add_Element(edge_triangle_hashtable,avd,i_avd);
        Add_Element(edge_triangle_hashtable,vcd,i_vcd);

        ////update extra hashtables
        if(vertex_triangle_hashtable){
            Remove_Element(*vertex_triangle_hashtable,acd,i_acd);
            Add_Element(*vertex_triangle_hashtable,avd,i_avd);
            Add_Element(*vertex_triangle_hashtable,vcd,i_vcd);
        }
    }
    Remove_Key(edge_triangle_hashtable,e_key);

    return true;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Simple_Redistribute_Mass_And_Momentum(const int a,const int c,const int v)
{
    particles.thickness(v)=(T).5*(particles.thickness(a)+particles.thickness(c));
    particles.mass(v)=(T).5*(particles.mass(a)+particles.mass(c));
    particles.dimensional_mass(v)=(T).5*(particles.dimensional_mass(a)+particles.dimensional_mass(c));
    particles.V(v)=(particles.V(a)*particles.mass(a)+particles.V(c)*particles.mass(c))/(particles.mass(a)+particles.mass(c));
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Simple_Redistribute_Mass_And_Momentum(const int a,const int b,const int c,const int v)
{
    particles.thickness(v)=(T)one_third*(particles.thickness(a)+particles.thickness(b)+particles.thickness(c));
    particles.mass(v)=(T)one_third*(particles.mass(a)+particles.mass(b)+particles.mass(c));
    particles.dimensional_mass(v)=(T)one_third*(particles.dimensional_mass(a)+particles.dimensional_mass(b)+particles.dimensional_mass(c));
    particles.V(v)=(particles.V(a)*particles.mass(a)+particles.V(b)*particles.mass(b)+particles.V(c)*particles.mass(c))/(particles.mass(a)+particles.mass(b)+particles.mass(c));
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Split(const int a,const int c,const int v,const ARRAY<int>& e_tets,const ARRAY<int>& e_tris)
{
    TV_INT2 e_key=TV_INT2(a,c);
    T delta_volume_a=(T)0;T delta_volume_c=(T)0;
    T delta_mass_a_3d=(T)0;T delta_mass_c_3d=(T)0;
    T delta_area_a=(T)0;T delta_area_c=(T)0;
    T delta_mass_a_2d=(T)0;T delta_mass_c_2d=(T)0;
    ////3d
    for(int i=1;i<=e_tets.m;i++){
        int t=e_tets(i);
        const TV_INT4& tet_vtx=tetrahedron_mesh.elements(t);
        T node_vol=(T).25*TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        delta_volume_a+=(T).5*node_vol;delta_volume_c+=(T).5*node_vol;
    }
    delta_mass_a_3d=delta_volume_a/particles.dimensional_size(a)(3)*particles.dimensional_mass(a)(TV::m);
    delta_mass_c_3d=delta_volume_c/particles.dimensional_size(c)(3)*particles.dimensional_mass(c)(TV::m);
    particles.dimensional_mass(v)(TV::m)=(delta_mass_a_3d+delta_mass_c_3d);
    particles.dimensional_mass(a)(TV::m)-=delta_mass_a_3d;
    particles.dimensional_mass(c)(TV::m)-=delta_mass_c_3d;
    particles.dimensional_size(v)(3)=(delta_volume_a+delta_volume_c);
    particles.dimensional_size(a)(3)-=delta_volume_a;
    particles.dimensional_size(c)(3)-=delta_volume_c;
    PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m)>(T)0);
    ////2d
    if(e_tris.m>0){
        PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>(T)0);
        for(int i=1;i<=e_tris.m;i++){
            int b=Other_Node_In_Element(triangle_mesh.elements(e_tris(i)),a,c);
            T area_abc=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(c));
            delta_area_a+=(T)(1./6.)*area_abc;
            delta_area_c+=(T)(1./6.)*area_abc;
            T delta_mass_2d_vtx_a=delta_area_a/particles.dimensional_size(a)(2)*particles.dimensional_mass(a)(TV::m-1);
            T delta_mass_2d_vtx_c=delta_area_c/particles.dimensional_size(c)(2)*particles.dimensional_mass(c)(TV::m-1);
            delta_mass_a_2d+=delta_mass_2d_vtx_a;
            delta_mass_c_2d+=delta_mass_2d_vtx_c;
        }

        particles.dimensional_mass(v)(TV::m-1)=(delta_mass_a_2d+delta_mass_c_2d);
        particles.dimensional_mass(a)(TV::m-1)-=delta_mass_a_2d;
        particles.dimensional_mass(c)(TV::m-1)-=delta_mass_c_2d;
        PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>=(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>=(T)0);
        particles.dimensional_size(v)(2)=(delta_area_a+delta_area_c);
        particles.dimensional_size(a)(2)-=delta_area_a;
        particles.dimensional_size(c)(2)-=delta_area_c;
    }
    particles.mass(v)=particles.dimensional_mass(v).Sum();
    particles.mass(a)=particles.dimensional_mass(a).Sum();
    particles.mass(c)=particles.dimensional_mass(c).Sum();
    T delta_mass_a=delta_mass_a_3d+delta_mass_a_2d;
    T delta_mass_c=delta_mass_c_3d+delta_mass_c_2d;
    PHYSBAM_ASSERT(delta_mass_a+delta_mass_c>(T)0);
    particles.V(v)=(delta_mass_a*particles.V(a)+delta_mass_c*particles.V(c))/(delta_mass_a+delta_mass_c);

    PHYSBAM_ASSERT(particles.dimensional_mass(v)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(v)(TV::m-1)>=(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>=(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>=(T)0);
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Flip tetrahedron edge
////Indices: none
////Particles: add
////Elements: tetrahedron added (3) and deleted (1)
////Topology: two incident tetrahedra of the splitted face
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Tetrahedron_Face(const TV_INT3& face,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,
    const TV3& split_weight/*=TV3((T)one_third,(T)one_third,(T)one_third)*/,bool force_split/*=false*/,HASHTABLE<int,ARRAY<int> >* vertex_tetrahedron_hashtable/*=0*/,HASHTABLE<int,ARRAY<int> >* vertex_triangle_hashtable/*=0*/,HASHTABLE<int,ARRAY<int> >* vertex_segment_hashtable/*=0*/,HASHTABLE<TV_INT2,ARRAY<int> >* edge_tetrahedron_hashtable/*=0*/,HASHTABLE<TV_INT2,ARRAY<int> >* edge_triangle_hashtable/*=0*/)
{
    TV_INT3 face_key=face.Sorted();
    int a=face_key(1);int b=face_key(2);int c=face_key(3);

    ARRAY<int> f_tets=face_tetrahedron_hashtable.Get(face_key);

    TV v_pos=split_weight(1)*particles.X(a)+split_weight(2)*particles.X(b)+split_weight(3)*particles.X(c);
    int v=Split_Particles(a,b,c,TV_INT(0,0,1),v_pos,split_weight,/*don't change mass and v*/false);
    if(use_dbg)LOG::cout<<"split tet face "<<face_key<<" guid "<<Element(particles.guid,face_key)<<" to "<<v<<" guid "<<particles.guid(v)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
    particles_added.Append(v);

    //////Temporally use averaged v
    //particles.V(v)=(particles.V(a)+particles.V(b)+particles.V(c))*(T)one_third;
    if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Split(a,b,c,v,f_tets)*/;}
    else{Simple_Redistribute_Mass_And_Momentum(a,b,c,v);}

    particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(b);particles_affected_hashtable.Set(c);

    for(int i=1;i<=f_tets.m;i++){
        int i_abcd=f_tets(i);const TV_INT4 abcd=tetrahedron_mesh.elements(i_abcd);
        TV_INT4 adbv;TV_INT4 bdcv;TV_INT4 cdav;Three_Splitted_Tetrahedron_Vertices(abcd,face_key,v,adbv,bdcv,cdav);
        int i_adbv=tetrahedron_mesh.elements.Append(adbv);tetrahedra_added.Append(i_adbv);
        int i_bdcv=tetrahedron_mesh.elements.Append(bdcv);tetrahedra_added.Append(i_bdcv);
        int i_cdav=tetrahedron_mesh.elements.Append(cdav);tetrahedra_added.Append(i_cdav);
        tetrahedra_to_remove.Append(i_abcd);

        ////update face_tet_hashtable
        Remove_Element(face_tetrahedron_hashtable,abcd,i_abcd);
        Add_Element(face_tetrahedron_hashtable,adbv,i_adbv);
        Add_Element(face_tetrahedron_hashtable,bdcv,i_bdcv);
        Add_Element(face_tetrahedron_hashtable,cdav,i_cdav);

        ////update extra hashtables
        if(vertex_tetrahedron_hashtable){
            Remove_Element(*vertex_tetrahedron_hashtable,abcd,i_abcd);
            Add_Element(*vertex_tetrahedron_hashtable,adbv,i_adbv);
            Add_Element(*vertex_tetrahedron_hashtable,bdcv,i_bdcv);
            Add_Element(*vertex_tetrahedron_hashtable,cdav,i_cdav);
        }
        if(edge_tetrahedron_hashtable){
            Remove_Element(*edge_tetrahedron_hashtable,abcd,i_abcd);
            Add_Element(*edge_tetrahedron_hashtable,adbv,i_adbv);
            Add_Element(*edge_tetrahedron_hashtable,bdcv,i_bdcv);
            Add_Element(*edge_tetrahedron_hashtable,cdav,i_cdav);
        }
    }
    Remove_Key(face_tetrahedron_hashtable,face_key);

    return true;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Split(const int a,const int b,const int c,const int v,const ARRAY<int>& e_tets)
{
    ////update particle mass and velocity
    T delta_volume_a=(T)0;T delta_volume_b=(T)0;T delta_volume_c=(T)0;
    T delta_mass_a=(T)0;T delta_mass_b=(T)0;T delta_mass_c=(T)0;
    for(int i=1;i<=e_tets.m;i++){
        int t=e_tets(i);
        const TV_INT4& tet_vtx=tetrahedron_mesh.elements(t);
        T node_vol=(T).25*TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        delta_volume_a+=(T)one_third*node_vol;delta_volume_b+=(T)one_third*node_vol;delta_volume_c+=(T)one_third*node_vol;
    }
    delta_mass_a=delta_volume_a/particles.dimensional_size(a)(3)*particles.dimensional_mass(a)(TV::m);
    delta_mass_b=delta_volume_b/particles.dimensional_size(b)(3)*particles.dimensional_mass(b)(TV::m);
    delta_mass_c=delta_volume_c/particles.dimensional_size(c)(3)*particles.dimensional_mass(c)(TV::m);
    particles.dimensional_mass(v)(TV::m)=(delta_mass_a+delta_mass_b+delta_mass_c);
    particles.dimensional_mass(a)(TV::m)-=delta_mass_a;
    particles.dimensional_mass(b)(TV::m)-=delta_mass_b;
    particles.dimensional_mass(c)(TV::m)-=delta_mass_c;
    particles.mass(v)=particles.dimensional_mass(v).Sum();
    particles.mass(a)=particles.dimensional_mass(a).Sum();
    particles.mass(b)=particles.dimensional_mass(b).Sum();
    particles.mass(c)=particles.dimensional_mass(c).Sum();
    PHYSBAM_ASSERT(delta_mass_a+delta_mass_b+delta_mass_c>(T)0);
    particles.V(v)=(delta_mass_a*particles.V(a)+delta_mass_b*particles.V(b)+delta_mass_c*particles.V(c))/(delta_mass_a+delta_mass_b+delta_mass_c);
    particles.dimensional_size(a)(3)-=delta_volume_a;
    particles.dimensional_size(b)(3)-=delta_volume_b;
    particles.dimensional_size(c)(3)-=delta_volume_c;
    particles.dimensional_size(v)(3)=(delta_volume_a+delta_volume_b+delta_volume_c);

    PHYSBAM_ASSERT(particles.dimensional_mass(v)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(b)(TV::m)>(T)0);
    PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m)>(T)0);
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Flip tetrahedron edge
////Indices: none
////Particles: none
////Elements: triangle checked, tetrahedron added and deleted 
////Topology: one ring of incident tetrahedra of the flipped edge
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Flip_Tetrahedron_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable, 
    HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT2>& affected_edges)
{
    TV_INT2 edge_key=edge.Sorted();int a=edge_key.x;int c=edge_key.y;
    ARRAY<int> e_tets=edge_tetrahedron_hashtable.Get(edge_key);
    if(e_tets.m!=3&&e_tets.m!=4){return false;}  ////flip fail: only allow 3-2 and 4-4 flip
    ARRAY<TV_INT4> tets_v;for(int i=1;i<=e_tets.m;i++){int t=e_tets(i);tets_v.Append(tetrahedron_mesh.elements(t));}
    //for(int i=1;i<=tets_v.m;i++){if(Is_Negative_Volume_Tetrahedron(tets_v(i)))return false;}    ////flip fail: has negative volume tet
    ARRAY<int> ring_v;if(!Get_One_Ring_Vertices_Of_Tetrahedra(tets_v,a,c,ring_v)){return false;}    ////flip fail: not manifold tet ring
    if(!Is_Convex_Ring_Of_Tetrahedra(tets_v,ring_v,a,c)){return false;}    ////flip fail: not convex tet ring
    T old_measure=Measure_Tetrahedra(tets_v);
    T new_measure=Measure_Flipped_Ring_Of_Tetrahedra(tets_v,ring_v,a,c);
    T measure_coef=(T).95;
    if(old_measure<measure_coef*new_measure){    ////flip
        if(use_dbg)LOG::cout<<"flip tet edge "<<edge_key<<" guid "<<Element(particles.guid,edge_key)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
        ARRAY<TV_INT4> flipped_tets_v;Get_Flipped_Ring_Of_Tetrahedra(tets_v,ring_v,a,c,flipped_tets_v);
        
        if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Flip(a,c,tets_v,ring_v,flipped_tets_v)*/;}
        else{}
        particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);for(int i=1;i<=ring_v.m;i++)particles_affected_hashtable.Set(ring_v(i));

        //////////////////////////////////////////////////////////////////////////
        //// start to flip
        ////delete old tets incident to e
        for(int i=1;i<=e_tets.m;i++){
            int t=e_tets(i);const TV_INT4& tet_vtx=tets_v(i);tetrahedra_to_remove.Append(t);
            Remove_Element(edge_tetrahedron_hashtable,tet_vtx,t);
        }
        Remove_Key(edge_tetrahedron_hashtable,edge_key);

        ////add new flipped tets
        for(int i=1;i<=flipped_tets_v.m;i++){
            const TV_INT4& tet_vtx=flipped_tets_v(i);int t=tetrahedron_mesh.elements.Append(tet_vtx);tetrahedra_added.Append(t);
            Add_Element(edge_tetrahedron_hashtable,tet_vtx,t);

            ////dbg
            TV_INT4 vtx_guid=Element(particles.guid,tet_vtx);
            if(vtx_guid.Contains(1027)&&vtx_guid.Contains(1484)&&vtx_guid.Contains(2611)){
                LOG::cout<<"add flip tet "<<vtx_guid<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
            }
        }
        return true;
    }
    return false;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Tetrahedron_Edge_Flip(const int a,const int c,const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& ring_v,const ARRAY<TV_INT4>& flipped_tets_v)
{
    ////update particle mass and velocity
    ARRAY<int> convex_v;
    ARRAY<T> convex_v_partial_volume;
    ARRAY<T> convex_v_flipped_partial_volume;
    ARRAY<T> convex_v_partial_mass;
    ARRAY<T> convex_v_flipped_partial_mass;

    convex_v.Append(a);convex_v.Append(c);
    for(int i=1;i<=ring_v.m;i++){convex_v.Append(ring_v(i));}
    convex_v_partial_volume.Resize(convex_v.Size());convex_v_partial_volume.Fill((T)0);
    convex_v_flipped_partial_volume.Resize(convex_v.Size());convex_v_flipped_partial_volume.Fill((T)0);
    convex_v_partial_mass.Resize(convex_v.Size());convex_v_partial_mass.Fill((T)0);
    convex_v_flipped_partial_mass.Resize(convex_v.Size());convex_v_flipped_partial_mass.Fill((T)0);

    std::map<int,int> convex_v_map;
    for(int i=1;i<=convex_v.m;i++){
        convex_v_map.insert(std::pair<int,int>(convex_v(i),i));
    }

    T convex_vol=(T)0;
    for(int i=1;i<=tets_v.m;i++){
        const TV_INT4& tet_vtx=tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_flipped_vol=(T)0;
    for(int i=1;i<=flipped_tets_v.m;i++){
        const TV_INT4& tet_vtx=flipped_tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_flipped_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_flipped_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_mass=(T)0;
    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        T node_partial_mass=convex_v_partial_volume(i)/particles.dimensional_size(p)(3)*particles.dimensional_mass(p)(TV::m);
        convex_v_partial_mass(i)=node_partial_mass;
        convex_mass+=node_partial_mass;
    }
    for(int i=1;i<=convex_v.m;i++){
        T node_partial_mass=convex_v_flipped_partial_volume(i)/convex_flipped_vol*convex_mass;
        convex_v_flipped_partial_mass(i)=node_partial_mass;
    }

    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        particles.dimensional_mass(p)(TV::m)+=(-convex_v_partial_mass(i)+convex_v_flipped_partial_mass(i));
        PHYSBAM_ASSERT(particles.dimensional_mass(p)(TV::m)>(T)0);
        particles.mass(p)=particles.dimensional_mass(p).Sum();
        particles.dimensional_size(p)(3)+=(-convex_v_partial_volume(i)+convex_v_flipped_partial_volume(i));
    }
    ////didn't change v here
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Flip_Tetrahedron_Boundary_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_tetrahedron_hashtable,HASHTABLE<TV_INT2>& boundary_edge_hashtable,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable, 
    HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT2>& affected_edges)
{
    TV_INT2 edge_key=edge.Sorted();//int a=edge_key.x;int c=edge_key.y;
    ARRAY<int> e_tets=edge_tetrahedron_hashtable.Get(edge_key);
    if(e_tets.m!=2)return false;    ////flip fail: only allow 2-2 flip

    if(edge_triangle_hashtable.Contains(edge_key)&&edge_triangle_hashtable.Get(edge_key).m>0){return false;}////flip fail: incident to a triangle edge

    ARRAY<TV_INT4> tets_v;for(int i=1;i<=e_tets.m;i++){int t=e_tets(i);tets_v.Append(tetrahedron_mesh.elements(t));}
    ARRAY<TV_INT4> flipped_tets_v;int b;int p;int q;
    if(!Get_22_Flipped_Pair_Of_Tetrahedra(tets_v,edge_key,b,p,q,flipped_tets_v,affected_edges))return false;
    if(!Is_Convex_Pair_Of_Boundary_Tetrahedra(edge_key,b,p,q))return false;

    T old_measure=Measure_Tetrahedra(tets_v);
    T new_measure=Measure_Tetrahedra(flipped_tets_v);
    T measure_coef=(T).95;
    if(old_measure<measure_coef*new_measure){
        T old_volume=TETRAHEDRON<T>::Volume(particles.X(tets_v(1)(1)),particles.X(tets_v(1)(2)),particles.X(tets_v(1)(3)),particles.X(tets_v(1)(4)))+
            TETRAHEDRON<T>::Volume(particles.X(tets_v(2)(1)),particles.X(tets_v(2)(2)),particles.X(tets_v(2)(3)),particles.X(tets_v(2)(4)));
        T new_volume=TETRAHEDRON<T>::Volume(particles.X(flipped_tets_v(1)(1)),particles.X(flipped_tets_v(1)(2)),particles.X(flipped_tets_v(1)(3)),particles.X(flipped_tets_v(1)(4)))+
            TETRAHEDRON<T>::Volume(particles.X(flipped_tets_v(2)(1)),particles.X(flipped_tets_v(2)(2)),particles.X(flipped_tets_v(2)(3)),particles.X(flipped_tets_v(2)(4)));
        T volume_change_threshold=(T).1*old_volume;
        if(abs(new_volume-old_volume)<volume_change_threshold){
            //////////////////////////////////////////////////////////////////////////
            //// start to flip
            if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Tetrahedron_Boundary_Edge_Flip(edge_key(1),edge_key(2),b,p,q,tets_v,flipped_tets_v);*/}
            else{}

            particles_affected_hashtable.Set(edge_key(1));particles_affected_hashtable.Set(edge_key(2));particles_affected_hashtable.Set(b);particles_affected_hashtable.Set(p);particles_affected_hashtable.Set(q);
            if(use_dbg)LOG::cout<<"flip tet boundary edge "<<edge_key<<" guid "<<Element(particles.guid,edge_key)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

            ////delete old tets incident to e
            for(int i=1;i<=e_tets.m;i++){
                int t=e_tets(i);const TV_INT4& tet_vtx=tets_v(i);tetrahedra_to_remove.Append(t);
                Remove_Element(edge_tetrahedron_hashtable,tet_vtx,t);
            }
            Remove_Key(edge_tetrahedron_hashtable,edge_key);

            ////add new flipped tets
            for(int i=1;i<=flipped_tets_v.m;i++){
                const TV_INT4& tet_vtx=flipped_tets_v(i);int t=tetrahedron_mesh.elements.Append(tet_vtx);tetrahedra_added.Append(t);
                Add_Element(edge_tetrahedron_hashtable,tet_vtx,t);
            }

            boundary_edge_hashtable.Delete(edge_key);
            boundary_edge_hashtable.Insert(TV_INT2(p,q).Sorted());
            return true;
        }
    }
    return false;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Tetrahedron_Boundary_Edge_Flip(const int a,const int c,
    const int b,const int p,const int q,const ARRAY<TV_INT4>& tets_v,const ARRAY<TV_INT4>& flipped_tets_v)
{
    TV_INT2 edge_key=TV_INT2(a,c);
    ////update particle mass and velocity
    ARRAY<int> convex_v;
    ARRAY<T> convex_v_partial_volume;
    ARRAY<T> convex_v_flipped_partial_volume;
    ARRAY<T> convex_v_partial_mass;
    ARRAY<T> convex_v_flipped_partial_mass;

    convex_v.Append(edge_key(1));convex_v.Append(edge_key(2));convex_v.Append(b);convex_v.Append(p);convex_v.Append(q);
    convex_v_partial_volume.Resize(convex_v.Size());convex_v_partial_volume.Fill((T)0);
    convex_v_flipped_partial_volume.Resize(convex_v.Size());convex_v_flipped_partial_volume.Fill((T)0);
    convex_v_partial_mass.Resize(convex_v.Size());convex_v_partial_mass.Fill((T)0);
    convex_v_flipped_partial_mass.Resize(convex_v.Size());convex_v_flipped_partial_mass.Fill((T)0);

    std::map<int,int> convex_v_map;
    for(int i=1;i<=convex_v.m;i++){
        convex_v_map.insert(std::pair<int,int>(convex_v(i),i));
    }

    T convex_vol=(T)0;
    for(int i=1;i<=tets_v.m;i++){
        const TV_INT4& tet_vtx=tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_flipped_vol=(T)0;
    for(int i=1;i<=flipped_tets_v.m;i++){
        const TV_INT4& tet_vtx=flipped_tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_flipped_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_flipped_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_mass=(T)0;
    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        T node_partial_mass=convex_v_partial_volume(i)/particles.dimensional_size(p)(3)*particles.dimensional_mass(p)(TV::m);
        convex_v_partial_mass(i)=node_partial_mass;
        convex_mass+=node_partial_mass;
    }
    for(int i=1;i<=convex_v.m;i++){
        //int p=convex_v(i);
        T node_partial_mass=convex_v_flipped_partial_volume(i)/convex_flipped_vol*convex_mass;
        convex_v_flipped_partial_mass(i)=node_partial_mass;
    }

    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        particles.dimensional_mass(p)(TV::m)+=(-convex_v_partial_mass(i)+convex_v_flipped_partial_mass(i));
        PHYSBAM_ASSERT(particles.dimensional_mass(p)(TV::m)>(T)0);
        particles.mass(p)=particles.dimensional_mass(p).Sum();
        particles.dimensional_size(p)(3)+=(-convex_v_partial_volume(i)+convex_v_flipped_partial_volume(i));
    }
    ////didn't change v here
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Flip tetrahedron face
////Indices: none
////Particles: none
////Elements: tetrahedra added and deleted 
////Topology: two incident tetrahedra of the flipped face
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Flip_Tetrahedron_Face(const TV_INT3& face,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,
    HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& tetrahedra_to_remove,ARRAY<int>& tetrahedra_added,ARRAY<TV_INT3>& affected_faces)
{
    TV_INT3 face_key=face.Sorted();
    ARRAY<int> f_tets=face_tetrahedron_hashtable.Get(face_key);
    ARRAY<TV_INT4> tets_v;for(int i=1;i<=f_tets.m;i++){int t=f_tets(i);tets_v.Append(tetrahedron_mesh.elements(t));}
    if(f_tets.m!=2){return false;}  ////flip fail: does not have two tets 
    ARRAY<TV_INT2> coplanar_edges;
    if(!Is_Convex_Pair_Of_Tetrahedra(tets_v,face_key,coplanar_edges)){/*LOG::cout<<"non-convex face "<<face_key<<std::endl;*/return false;}   ////flip fail: not convex tet pair

    bool is_22_flip=false;TV_INT2 boundary_edge;
    ////TODO: fix this, should not use Is_Boundary_Edge_On_Tetrahedron_Mesh in meshing operations
    //if(coplanar_edges.m>0){
    //    for(int i=1;i<=coplanar_edges.m;i++){
    //        TV_INT2 edge_key=coplanar_edges(i).Sorted();
    //        if(codimensional_mesh_object.mesh.Is_Boundary_Edge_On_Tetrahedron_Mesh(edge_key)){boundary_edge=edge_key;is_22_flip=true;break;}
    //    }
    //}

    T old_measure=Measure_Tetrahedra(tets_v);
    T new_measure=is_22_flip?Measure_22_Flipped_Pair_Of_Tetrahedra(tets_v,face,boundary_edge):Measure_23_Flipped_Pair_Of_Tetrahedra(tets_v,face_key);

    //LOG::cout<<"measure face: "<<face_key<<": "<<old_measure<<", "<<new_measure<<std::endl;
    T measure_coef=(T).95;
    if(old_measure<measure_coef*new_measure){
        ARRAY<TV_INT4> flipped_tets_v;
        if(is_22_flip){Get_22_Flipped_Pair_Of_Tetrahedra(tets_v,face_key,boundary_edge,flipped_tets_v,affected_faces);}
        else{Get_23_Flipped_Pair_Of_Tetrahedra(tets_v,face_key,flipped_tets_v,affected_faces);}

        //////////////////////////////////////////////////////////////////////////
        //// start to flip
        int a=Other_Node_In_Element(tets_v(1),face_key(1),face_key(2),face_key(3));int c=Other_Node_In_Element(tets_v(2),face_key(1),face_key(2),face_key(3));TV_INT2 ac=TV_INT2(a,c);  ////dbg
        if(use_dbg)LOG::cout<<"flip tet face "<<face_key<<" guid "<<Element(particles.guid,face_key)<<" to ["<<a<<", "<<c<<"], guid "<<Element(particles.guid,ac)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;

        if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Flip(tets_v,flipped_tets_v);*/}
        else{}

        for(int i=1;i<=tets_v.m;i++)for(int d=1;d<=4;d++){particles_affected_hashtable.Set(tets_v(i)(d));}

        ////delete old tets incident to face
        for(int i=1;i<=f_tets.m;i++){
            int t=f_tets(i);const TV_INT4& tet_vtx=tets_v(i);tetrahedra_to_remove.Append(t);
            Remove_Element(face_tetrahedron_hashtable,tet_vtx,t);
        }
        Remove_Key(face_tetrahedron_hashtable,face_key);

        ////add new flipped tets
        for(int i=1;i<=flipped_tets_v.m;i++){
            const TV_INT4& tet_vtx=flipped_tets_v(i);int t=tetrahedron_mesh.elements.Append(tet_vtx);tetrahedra_added.Append(t);
            Add_Element(face_tetrahedron_hashtable,tet_vtx,t);
        }

        return true;
    }
    return false;
}


template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Tetrahedron_Face_Flip(const ARRAY<TV_INT4>& tets_v,const ARRAY<TV_INT4>& flipped_tets_v)
{
    ////update particle mass and velocity
    ARRAY<int> convex_v;
    ARRAY<T> convex_v_partial_volume;
    ARRAY<T> convex_v_flipped_partial_volume;
    ARRAY<T> convex_v_partial_mass;
    ARRAY<T> convex_v_flipped_partial_mass;

    for(int i=1;i<=tets_v.m;i++){
        const TV_INT4& vtx=tets_v(i);for(int d=1;d<=4;d++)convex_v.Append_Unique(vtx(d));
    }
    convex_v_partial_volume.Resize(convex_v.Size());convex_v_partial_volume.Fill((T)0);
    convex_v_flipped_partial_volume.Resize(convex_v.Size());convex_v_flipped_partial_volume.Fill((T)0);
    convex_v_partial_mass.Resize(convex_v.Size());convex_v_partial_mass.Fill((T)0);
    convex_v_flipped_partial_mass.Resize(convex_v.Size());convex_v_flipped_partial_mass.Fill((T)0);

    std::map<int,int> convex_v_map;
    for(int i=1;i<=convex_v.m;i++){
        convex_v_map.insert(std::pair<int,int>(convex_v(i),i));
    }

    T convex_vol=(T)0;
    for(int i=1;i<=tets_v.m;i++){
        const TV_INT4& tet_vtx=tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_flipped_vol=(T)0;
    for(int i=1;i<=flipped_tets_v.m;i++){
        const TV_INT4& tet_vtx=flipped_tets_v(i);
        T tet_vol=TETRAHEDRON<T>::Volume(particles.X(tet_vtx(1)),particles.X(tet_vtx(2)),particles.X(tet_vtx(3)),particles.X(tet_vtx(4)));
        convex_flipped_vol+=tet_vol;
        T node_vol=(T).25*tet_vol;
        for(int d=1;d<=4;d++){
            int node_idx=convex_v_map.at(tet_vtx(d));
            convex_v_flipped_partial_volume(node_idx)+=node_vol;
        }
    }

    T convex_mass=(T)0;
    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        T node_partial_mass=convex_v_partial_volume(i)/particles.dimensional_size(p)(3)*particles.dimensional_mass(p)(TV::m);
        convex_v_partial_mass(i)=node_partial_mass;
        convex_mass+=node_partial_mass;
    }
    for(int i=1;i<=convex_v.m;i++){
        T node_partial_mass=convex_v_flipped_partial_volume(i)/convex_flipped_vol*convex_mass;
        convex_v_flipped_partial_mass(i)=node_partial_mass;
    }

    for(int i=1;i<=convex_v.m;i++){
        int p=convex_v(i);
        particles.dimensional_mass(p)(TV::m)+=(-convex_v_partial_mass(i)+convex_v_flipped_partial_mass(i));
        PHYSBAM_ASSERT(particles.dimensional_mass(p)(TV::m)>(T)0);
        particles.mass(p)=particles.dimensional_mass(p).Sum();
        particles.dimensional_size(p)(3)+=(-convex_v_partial_volume(i)+convex_v_flipped_partial_volume(i));
    }
    ////didn't change v here
}


template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Valid_Tetrahedron_Edge_Collapse(const int a,const int c,
    const int guid_a,const int guid_c,const int on_boundary_a,const int on_boundary_c,
    const int contact_a,const int contact_c,const TV& X_a,const TV& X_c,const ARRAY<TV_INT4>& new_one_ring_tets,bool force_split,/*rst*/TV& v_pos,const T old_vol/*=(T)-1*/,const ARRAY<int>* ac_tets/*=0*/)
{
    if(a==c)return false;   ////collapse fail: duplicated edges in the queue
    if(!force_split&&has_eulerian_particles){
        //if(use_split_eulerian_particles&&eulerian_particles_hashtable.Contains(guid_a)&&eulerian_particles_hashtable.Contains(guid_c))return false;     ////collapse fail: should not collapse any eulerian particle
        if(!use_split_eulerian_particles&&(eulerian_particles_hashtable.Contains(guid_a)||eulerian_particles_hashtable.Contains(guid_c)))return false;
    }

    //////collapse fail: solid and not inverted particle, TODO: find a better way to handle solid contact particles
    //if(edge_type!=4&&particles.contact(a)&&particles.contact(c)){
    //    LOG::cout<<"solid particle collapse fail: "<<particles.guid(a)<<", "<<particles.guid(c)<<std::endl;return false;}

    ////find candidate merged particle position
    TV candidate_v_pos=TV();int candidate_v_index=0;int contact_v_index=0;int eulerian_v_index=0;    ////a-1,midpoint-2,c-3
    ARRAY<TV> candidate_v_positions;candidate_v_positions.Append(X_a);candidate_v_positions.Append((T).5*(X_a+X_c));candidate_v_positions.Append(X_c);
    ARRAY<T> candidate_v_tet_vol;candidate_v_tet_vol.Resize(3);candidate_v_tet_vol.Fill((T)0);ARRAY<T> vol_dif;vol_dif.Resize(3);vol_dif.Fill((T)0);
    ARRAY<bool> candidate_v_has_neg_tet;candidate_v_has_neg_tet.Resize(3);candidate_v_has_neg_tet.Fill((T)0);
    ARRAY<int> candidate_v_on_boundary;candidate_v_on_boundary.Append(on_boundary_a);candidate_v_on_boundary.Append(on_boundary_a&&on_boundary_c);candidate_v_on_boundary.Append(on_boundary_c);
    int only_boundary_idx=0;

    for(int j=1;j<=3;j++){
        ARRAY<VECTOR<TV,4> > tets_X;for(int i=1;i<=new_one_ring_tets.m;i++){const TV_INT4& vtx=new_one_ring_tets(i);
            VECTOR<TV,4> tet_X;for(int d=1;d<=4;d++){tet_X(d)=((vtx(d)==a||vtx(d)==c)?candidate_v_positions(j):particles.X(vtx(d)));}tets_X.Append(tet_X);}
        if(old_vol>(T)0){
            candidate_v_has_neg_tet(j)=Has_Non_Positive_Volume_Tetrahedron_And_Values(tets_X,candidate_v_tet_vol(j));
            vol_dif(j)=abs(candidate_v_tet_vol(j)-old_vol);}
        else candidate_v_has_neg_tet(j)=Has_Non_Positive_Volume_Tetrahedron(tets_X);
    }


    if(eulerian_particles_hashtable.Contains(guid_a))eulerian_v_index=1;
    else if(eulerian_particles_hashtable.Contains(guid_c))eulerian_v_index=3;
    
    if(!on_boundary_a&&!on_boundary_c){candidate_v_index=2;}    ////no boundary particle
    else if(on_boundary_a&&on_boundary_c){    ////two boundary particles
        ////if one of the two boundary particles contacts with solid, set candidate_v_pos to be on solid
        if(Is_Contact_With_Solid(contact_a)&&!Is_Contact_With_Solid(contact_c)){candidate_v_index=1;contact_v_index=1;} //candidate_positions=particles.X(a);
        else if(Is_Contact_With_Solid(contact_c)&&!Is_Contact_With_Solid(contact_a)){candidate_v_index=3;contact_v_index=3;} //candidate_v_pos=particles.X(c);
        else{   
            if(old_vol>(T)0){////pick the one that has least volume change
                int min_idx=1;for(int j=2;j<=3;j++)if(vol_dif(j)<vol_dif(min_idx))min_idx=j;
                candidate_v_index=min_idx;

                //////check volume changes
                //bool is_diagonal_edge=ac_tets?Is_Diagonal_Edge(TV_INT2(a,c),*ac_tets):false;
                //if(!is_diagonal_edge){
                //    T tol_frac=(T).15;
                //    if(vol_dif(min_idx)>old_vol*tol_frac)return false; ////collapse fail: do not allow volume change too much
                //    else LOG::cout<<"col vol frac: "<<guid_a<<", "<<guid_c<<": "<<vol_dif(min_idx)/old_vol<<std::endl;
                //}
            }
            else candidate_v_index=2;//candidate_v_pos=(T).5*(particles.X(a)+particles.X(c));
        }

        ////handle tank walls
        if(has_tank){
            TV_INT corner_a=water_tank->Tank_Corner(X_a);
            TV_INT corner_c=water_tank->Tank_Corner(X_c);
            TV_INT difference=abs(corner_a)-abs(corner_c);
            if(difference!=TV_INT()){
                bool positive=false;bool negative=false;
                for(int d=1;d<=TV::m;d++)if(difference(d)>0)positive=true;else if(difference(d)<0)negative=true;
                if(positive&&negative){return false;}    ////collapse fail: particles on different walls
                else if(positive){candidate_v_index=1;} //candidate_v_pos=particles.X(a);
                else if(negative){candidate_v_index=3;} //candidate_v_pos=particles.X(c);
            }
        }
    }
    else{   ////one boundary particle
        only_boundary_idx=on_boundary_a?1:3;candidate_v_index=only_boundary_idx;   //candidate_v_pos=particles.X(boundary_idx);
    }

    ////collapse fail: make sure all new tet volumes > 0
    if(/*force_split*/false){
        candidate_v_pos=(T).5*(X_a+X_c);
        ARRAY<VECTOR<TV,4> > tets_X;
        for(int i=1;i<=new_one_ring_tets.m;i++){const TV_INT4& vtx=new_one_ring_tets(i);
            VECTOR<TV,4> tet_X;for(int d=1;d<=4;d++){tet_X(d)=((vtx(d)==a||vtx(d)==c)?candidate_v_pos:particles.X(vtx(d)));}tets_X.Append(tet_X);
        }
        bool has_non_positive_tet=Has_Non_Positive_Volume_Tetrahedron(tets_X);
        if(has_non_positive_tet){if(use_dbg)LOG::cout<<"inverted particle in tet edge collapse, non-positive tet "<<a<<", "<<c<<" guid "<<guid_a<<", "<<guid_c<<std::endl;return false;}
    }
    else{
        ARRAY<int> candidate_v_indices;candidate_v_indices.Append(candidate_v_index);
        for(int i=1;i<=3;i++){if(i!=candidate_v_index)candidate_v_indices.Append(i);}
        if((candidate_v_positions(candidate_v_indices(1))-candidate_v_positions(candidate_v_indices(2))).Magnitude_Squared()>(candidate_v_positions(candidate_v_indices(1))-candidate_v_positions(candidate_v_indices(3))).Magnitude_Squared()){int tmp=candidate_v_indices(2);candidate_v_indices(2)=candidate_v_indices(3);candidate_v_indices(3)=tmp;}
        bool has_valid_candidate=false;
        for(int j=1;j<=3;j++){
            //ARRAY<VECTOR<TV,4> > tets_X;
            //for(int i=1;i<=new_one_ring_tets.m;i++){const TV_INT4& vtx=new_one_ring_tets(i);
            //    VECTOR<TV,4> tet_X;for(int d=1;d<=4;d++){tet_X(d)=((vtx(d)==a||vtx(d)==c)?candidate_v_positions(candidate_v_indices(j)):particles.X(vtx(d)));}tets_X.Append(tet_X);
            //}
            //bool has_non_positive_tet=Has_Non_Positive_Volume_Tetrahedron(tets_X);
            bool has_non_positive_tet=candidate_v_has_neg_tet(candidate_v_indices(j));
            if(has_non_positive_tet){continue;}
            else if(contact_v_index!=0&&candidate_v_indices(j)!=contact_v_index){continue;}  ////do not allow a contact particle merged to a non-contact particle
            else if(only_boundary_idx!=0&&candidate_v_indices(j)!=only_boundary_idx){continue;} ////do not allow a boundary particle merged to a interior particle
            else if(eulerian_v_index!=0&&candidate_v_indices(j)!=eulerian_v_index){continue;}   ////do not allow a eulerian particle merged to a non-eulerian particle
            else{has_valid_candidate=true;candidate_v_pos=candidate_v_positions(candidate_v_indices(j));break;}
        }
        if(!has_valid_candidate){if(use_dbg)LOG::cout<<"inverted particle in tet edge collapse, non-positive tet "<<", "<<guid_a<<", "<<guid_c<<std::endl;return false;}
    }

    v_pos=candidate_v_pos;
    return true;
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Flip tetrahedron face
////Indices: replace the indices of collapsed vertices
////Particles: particles added and deleted
////Elements: tetrahedron added and deleted 
////Topology: one ring of tetrahedra of collapsed vertices
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Collapse_Tetrahedron_Edge(const TV_INT2& edge,const int edge_type,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,HASHTABLE<TV_INT3,ARRAY<int> >& face_tetrahedron_hashtable,ARRAY<int>& particles_to_remove,ARRAY<int>& ear_particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,
    ARRAY<int>& triangle_indices_to_remove,ARRAY<int>& tetrahedron_indices_to_remove,
    bool force_split/*=false*/,HASHTABLE<TV_INT2,ARRAY<int> >* edge_tetrahedron_hashtable/*=0*/,HASHTABLE<TV_INT2,ARRAY<int> >* edge_triangle_hashtable/*=0*/)
{
    TV_INT2 e_key=edge.Sorted();int a=e_key.x;int c=e_key.y;  
    ////tet neighbors
    ARRAY<int> a_and_c;ARRAY<int> old_two_ring_tets;ARRAY<int> new_one_ring_tets;ARRAY<int> ac_tets;a_and_c.Append(a);a_and_c.Append(c);
    for(int i=1;i<=a_and_c.m;i++){  ////old_two_ring_tets is the union of ac_tets and new_one_ring_tets
        int a_or_c=a_and_c(i);
        ARRAY<int>& a_or_c_tets=vertex_tetrahedron_hashtable.Get(a_or_c);
        for(int j=1;j<=a_or_c_tets.m;j++){
            int t=a_or_c_tets(j);
            const VECTOR<int,4>& vtx=tetrahedron_mesh.elements(t);
            if(vtx.Contains(a)&&vtx.Contains(c)) ac_tets.Append_Unique(t);
            else new_one_ring_tets.Append_Unique(t);
            old_two_ring_tets.Append_Unique(t);
        }
    }

    T old_vol=(T)0;
    for(int i=1;i<=old_two_ring_tets.m;i++){
        const int t=old_two_ring_tets(i);const TV_INT4& v=tetrahedron_mesh.elements(t);
        T vol=TETRAHEDRON<T>::Volume(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4)));old_vol+=vol;
    }
    
    ////check validation and compute candidate v position
    ARRAY<TV_INT4> new_one_ring_tet_vtx;for(int i=1;i<=new_one_ring_tets.m;i++){new_one_ring_tet_vtx.Append(tetrahedron_mesh.elements(new_one_ring_tets(i)));}
    TV candidate_v_pos;bool valid_for_collapse=Valid_Tetrahedron_Edge_Collapse(a,c,particles.guid(a),particles.guid(c),particles.on_boundary(a)(TV::m),particles.on_boundary(c)(TV::m),
        particles.contact(a),particles.contact(c),particles.X(a),particles.X(c),new_one_ring_tet_vtx,force_split,candidate_v_pos,old_vol,&ac_tets);
    if(!valid_for_collapse)return false;

    ////tri neighbors
    ARRAY<int> old_two_ring_tris;ARRAY<int> new_one_ring_tris;ARRAY<int> ac_tris;
    for(int i=1;i<=a_and_c.m;i++){
        int a_or_c=a_and_c(i);
        if(vertex_triangle_hashtable.Contains(a_or_c)){
            ARRAY<int>& a_or_c_tris=vertex_triangle_hashtable.Get(a_or_c);
            for(int j=1;j<=a_or_c_tris.m;j++){
                int t=a_or_c_tris(j);
                VECTOR<int,3> vtx=triangle_mesh.elements(t);
                if(vtx.Contains(a)&&vtx.Contains(c)) ac_tris.Append_Unique(t);
                else new_one_ring_tris.Append_Unique(t);
                old_two_ring_tris.Append_Unique(t);
            }
        }
    }

    HASHTABLE<int,T> vtx_vol;
    for(int i=1;i<=ac_tris.m;i++){int t=ac_tris(i);
        const TV_INT3& vtx=triangle_mesh.elements(t);
        for(int j=1;j<=3;j++){
            int p=vtx(j);
            if(!vtx_vol.Contains(p)){
                T tet_area_sum=Nodal_Tetrahedron_Surface_Area_Sum(p,vertex_tetrahedron_hashtable);
                T tri_area_sum=Node_Triangle_Area_Sum(p,vertex_triangle_hashtable);
                T seg_length_sum=Node_Segment_Length_Sum(p,vertex_segment_hashtable);
                T vol=(tri_area_sum+(T).5*tet_area_sum)*particles.thickness(p)+seg_length_sum*(T).25*(T)pi*pow<2>(particles.thickness(p));
                vtx_vol.Set(p,vol);
            }
        }
    }

    //////////////////////////////////////////////////////////////////////////
    ////start to collapse

    ////add new particle v
    T weight=(T).5;if(candidate_v_pos==particles.X(a))weight=(T)0;else if(candidate_v_pos==particles.X(c))weight=(T)1;
    int v=Merge_Particles(a,c,TV_INT(1,1,1),candidate_v_pos,weight,/*don't update mass and v*/false);particles_added.Append(v);
    if(use_dbg)LOG::cout<<"collapse tet edge N"<<(mpi_codim?mpi_codim->node_id:0)<<": "<<a<<", "<<c<<"-> "<<v<<", guid "<<particles.guid(a)<<", "<<particles.guid(c)<<" -> "<<particles.guid(v)<<std::endl;

    bool a_or_c_is_eulerian=has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c)));
    if(a_or_c_is_eulerian&&use_split_eulerian_particles){
        eulerian_particles_hashtable.Delete_If_Present(particles.guid(a));eulerian_particles_hashtable.Delete_If_Present(particles.guid(c));
        eulerian_particles_hashtable.Set(particles.guid(v));
        added_eulerian_particle_parents_hashtable.Set(particles.guid(v),PAIR<TV_INT2,T>(TV_INT2(particles.guid(a),particles.guid(c)),weight));
    }

    if(!preserve_thickness_in_remeshing){}  ////code removed
    else{Simple_Redistribute_Mass_And_Momentum(a,c,v);}

    if(!preserve_thickness_in_remeshing){} ////redistribute mass and momentum step 1, code removed

    ////delete a and c from and add v to the vertex_tet_hashtable
    for(int i=1;i<=a_and_c.m;i++){int a_or_c=a_and_c(i);vertex_tetrahedron_hashtable.Delete(a_or_c);}
    vertex_tetrahedron_hashtable.Insert(v,new_one_ring_tets);

    ////delete incident tets of ac from the vtx_tet hashtable
    for(int i=1;i<=ac_tets.m;i++){
        int t_ac=ac_tets(i);VECTOR<int,4> vtx=tetrahedron_mesh.elements(t_ac);TV_INT2 b_and_d=Other_Nodes_In_Element(vtx,a,c);
        for(int j=1;j<=b_and_d.m;j++){int b=b_and_d(j);ARRAY<int>& b_tets=vertex_tetrahedron_hashtable.Get(b);Remove_From_Array(b_tets,t_ac);}
    }

    ////update extra hashtable
    if(edge_tetrahedron_hashtable){
        for(int i=1;i<=ac_tets.m;i++){  ////remove ac incident tets from edge_tet_hashtable
            int t=ac_tets(i);const TV_INT4& vtx=tetrahedron_mesh.elements(t);
            Remove_Element(*edge_tetrahedron_hashtable,vtx,t);
        }
        for(int i=1;i<=old_two_ring_tets.m;i++){    ////replace a,c with v in edge_tet_hashtable
            int t=old_two_ring_tets(i);const TV_INT4& vtx=tetrahedron_mesh.elements(t);
            Replace_Element(*edge_tetrahedron_hashtable,vtx,edge,v);
        }
    }

    ////delete incident tris of ac from the vtx_tri hashtable
    HASHTABLE<int> p_thickness_needs_update_hashtable;
    for(int i=1;i<=ac_tris.m;i++){
        int t_ac=ac_tris(i);
        VECTOR<int,3> vtx=triangle_mesh.elements(t_ac);
        int b=Other_Node_In_Element(vtx,a,c);
        //LOG::cout<<"ac "<<a<<", "<<c<<", b: "<<b<<" guid "<<particles.guid(a)<<", "<<particles.guid(c)<<", "<<particles.guid(v)<<": "<<vertex_triangle_hashtable.Contains(b)<<std::endl;

        if(vertex_triangle_hashtable.Contains(b)){
            ARRAY<int>& b_tris=vertex_triangle_hashtable.Get(b);
            Remove_From_Array(b_tris,t_ac);
            bool b_incident_to_a_or_c=false;
            for(int i=1;i<=b_tris.m;i++){
                const int& t=b_tris(i);const TV_INT3& vtx=triangle_mesh.elements(t);
                if(vtx.Contains(a)||vtx.Contains(c)){b_incident_to_a_or_c=true;break;}
            }

            if(!b_incident_to_a_or_c){    ////b is not on triangle mesh, triangle degenerate to segment
                //bool is_on_seg_mesh=vertex_segment_hashtable.Contains(b)&&vertex_segment_hashtable.Get(b).m>0;
                //bool is_on_tet_mesh=vertex_tetrahedron_hashtable.Contains(b)&&vertex_tetrahedron_hashtable.Get(b).m>0;
                bool tet_mesh_has_edge_ab=false;ARRAY<int>& v_tets=vertex_tetrahedron_hashtable.Get(v);
                for(int j=1;j<=v_tets.m;j++){const TV_INT4& v=tetrahedron_mesh.elements(v_tets(j));if(v.Contains(b)){tet_mesh_has_edge_ab=true;break;}}

                if(!tet_mesh_has_edge_ab){
                    int s=segment_mesh.elements.Append(TV_INT2(b,v));
                    if(use_dbg)LOG::cout<<"tri deg to seg in tet clps: "<<TV_INT2(b,v)<<" guid "<<Element(particles.guid,TV_INT2(b,v))<<std::endl;
                    ARRAY<int>& b_segs=vertex_segment_hashtable.Get_Or_Insert(b,ARRAY<int>());b_segs.Append_Unique(s);
                    ARRAY<int>& v_segs=vertex_segment_hashtable.Get_Or_Insert(v,ARRAY<int>());v_segs.Append_Unique(s);
                } 
                
                if(!preserve_thickness_in_remeshing){}  ////code removed
                else{
                    //T A_abc=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(c));
                    //T Vol_b=(T)one_third*A_abc*particles.thickness(b);
                    //T Vol_ac=(T)one_third*A_abc*(particles.thickness(a)+particles.thickness(c));
                    //T seg_length=(particles.X(v)-particles.X(b)).Magnitude();
                    //particles.thickness(b)=sqrt((T)2*Vol_b/((T)pi*(T).5*seg_length));
                    //particles.thickness(v)=sqrt((T)2*Vol_ac/((T)pi*(T).5*seg_length));
                    p_thickness_needs_update_hashtable.Set(v);
                    p_thickness_needs_update_hashtable.Set(b);
                }
            }
        }
    }

    //delete key a and c from and add v to the vertex_tri_hashtable (if there is triangle)
    for(int i=1;i<=a_and_c.m;i++){int a_or_c=a_and_c(i);if(vertex_triangle_hashtable.Contains(a_or_c)){vertex_triangle_hashtable.Delete(a_or_c);}}

    if(new_one_ring_tris.m>0)vertex_triangle_hashtable.Insert(v,new_one_ring_tris);
    else{   ////v does not have any incident tris, then remove the 2-dim mass, ////TODO: may lose 2-dim mass, needs to be added to the target mass
        ////TODO: handle the tri degenerate seg case
        if(!preserve_thickness_in_remeshing){}  ////code removed
    }

    ////update extra hashtable
    if(edge_tetrahedron_hashtable){
        for(int i=1;i<=ac_tris.m;i++){  ////remove ac incident tets from edge_tet_hashtable
            int t=ac_tris(i);const TV_INT3& vtx=triangle_mesh.elements(t);
            Remove_Element(*edge_triangle_hashtable,vtx,t);
        }
        for(int i=1;i<=old_two_ring_tris.m;i++){    ////replace a,c with v in edge_tri_hashtable
            int t=old_two_ring_tris(i);const TV_INT3& vtx=triangle_mesh.elements(t);
            Replace_Element(*edge_triangle_hashtable,vtx,edge,v);
        }
    }

    ////delete incident tets of ac from the face_tet hashtable
    ARRAY<TV_INT3> v_fan_faces;
    for(int i=1;i<=ac_tets.m;i++){
        int t_ac=ac_tets(i);
        TV_INT4 vtx=tetrahedron_mesh.elements(t_ac);
        TV_INT2 b_and_d=Other_Nodes_In_Element(vtx,a,c);int b=b_and_d(1);int d=b_and_d(2);

        ////delete triangle acb, acd
        face_tetrahedron_hashtable.Delete_If_Present(TV_INT3(a,c,b).Sorted());
        face_tetrahedron_hashtable.Delete_If_Present(TV_INT3(a,c,d).Sorted());

        ////delete tet from triangle abd,cbd
        ARRAY<int>& abd_tets=face_tetrahedron_hashtable.Get(TV_INT3(a,b,d).Sorted());
        Remove_From_Array(abd_tets,t_ac);
        ARRAY<int>& cbd_tets=face_tetrahedron_hashtable.Get(TV_INT3(c,b,d).Sorted());
        Remove_From_Array(cbd_tets,t_ac);
        v_fan_faces.Append(TV_INT3(v,b,d).Sorted());
    }

    ////replace a and c with v in face_tet_hashtable     ////TODO: fix the redundant check here
    //LOG::cout<<"ac_tets: "<<std::endl;
    //for(int i=1;i<=ac_tets.m;i++)LOG::cout<<tetrahedron_mesh.elements(ac_tets(i))<<" ";LOG::cout<<std::endl;
    //LOG::cout<<"new_one_ring_tets: "<<std::endl;
    //for(int i=1;i<=new_one_ring_tets.m;i++)LOG::cout<<tetrahedron_mesh.elements(new_one_ring_tets(i))<<" ";LOG::cout<<std::endl;
    for(int i=1;i<=old_two_ring_tets.m;i++){
        int t=old_two_ring_tets(i);
        ARRAY<TV_INT3> tet_faces;Simplex_Faces(tetrahedron_mesh.elements(t),tet_faces);
        for(int j=1;j<=4;j++){
            TV_INT3 old_face_key=tet_faces(j).Sorted();
            if(old_face_key.Contains(a)&&old_face_key.Contains(c))continue; ////skip face incident to ac
            TV_INT3 new_face_key=Replaced_Two(old_face_key,a,c,v).Sorted();
            if(old_face_key==new_face_key)continue; ////skip face that does not contain a or c
            ARRAY<int> old_face_tets;
            
            if(face_tetrahedron_hashtable.Get(old_face_key,old_face_tets)){
                ARRAY<int>& new_face_tets=face_tetrahedron_hashtable.Get_Or_Insert(new_face_key,ARRAY<int>());
                
                for(int k=1;k<=old_face_tets.m;k++){
                    new_face_tets.Append_Unique(old_face_tets(k));
                }

                if(use_dbg&&new_face_tets.m>2){
                    LOG::cout<<"faces with >2 incident tets"<<std::endl;
                    LOG::cout<<"old face key: "<<old_face_key<<" guid "<<Element(particles.guid,old_face_key)<<", new face key: "<<new_face_key<<" guid "<<Element(particles.guid,new_face_key)<<", N"<<(mpi_codim?mpi_codim->node_id:0)<<std::endl;
                    //LOG::cout<<"old tet: "<<t<<", "<<tetrahedron_mesh.elements(t)<<" guid "<<Element(particles.guid,tetrahedron_mesh.elements(t))<<std::endl;
                    //LOG::cout<<"ac: "<<a<<", "<<c<<" guid ["<<particles.guid(a)<<", "<<particles.guid(c)<<"] "<<" -> v: "<<v<<" guid "<<particles.guid(v)<<std::endl;
                    //LOG::cout<<"new_face_tets: "<<new_face_tets.m<<std::endl;
                    //for(int k=1;k<=new_face_tets.m;k++){
                    //    LOG::cout<<"new "<<k<<": "<<new_face_tets(k)<<", v "<<tetrahedron_mesh.elements(new_face_tets(k))<<", guid "<<Element(particles.guid,tetrahedron_mesh.elements(new_face_tets(k)))
                    //        <<", is_neg_vol: "<<Is_Negative_Volume_Tetrahedron(particles.X,tetrahedron_mesh.elements(new_face_tets(k)))<<std::endl;
                    //}LOG::cout<<std::endl;
                    //LOG::cout<<"old_face_tets: "<<old_face_tets.m<<std::endl;
                    //for(int k=1;k<=old_face_tets.m;k++){
                    //    LOG::cout<<"old "<<k<<", v "<<tetrahedron_mesh.elements(old_face_tets(k))<<" guid "<<Element(particles.guid,tetrahedron_mesh.elements(old_face_tets(k)))
                    //        <<", is_neg_vol: "<<Is_Negative_Volume_Tetrahedron(particles.X,tetrahedron_mesh.elements(old_face_tets(k)))<<std::endl;
                    //}LOG::cout<<std::endl;
                    //LOG::cout<<"ac_tets: "<<ac_tets.m<<std::endl;
                    //for(int k=1;k<=ac_tets.m;k++){
                    //    LOG::cout<<"act "<<k<<", v "<<tetrahedron_mesh.elements(ac_tets(k))<<" guid "<<Element(particles.guid,tetrahedron_mesh.elements(ac_tets(k)))
                    //        <<", is_neg_vol: "<<Is_Negative_Volume_Tetrahedron(particles.X,tetrahedron_mesh.elements(ac_tets(k)))<<std::endl;
                    //}LOG::cout<<std::endl;
                    //for(int k=1;k<=tetrahedron_mesh.elements.m;k++){
                    //    const TV_INT4& v=tetrahedron_mesh.elements(k);
                    //    const TV_INT4 guid=Element(particles.guid,v);
                    //    if(guid.Contains(particles.guid(a))&&guid.Contains(particles.guid(c))){
                    //        LOG::cout<<"actet "<<v<<", "<<guid<<std::endl;
                    //    }
                    //}
                }
                face_tetrahedron_hashtable.Delete(old_face_key);
            } 
        }
    }

    ////tetrahedron degenerate to triangle
    T lost_tet_vol=(T)0;T added_tri_area=(T)0;
    for(int i=1;i<=v_fan_faces.m;i++){
        const TV_INT3& face_key=v_fan_faces(i);
        ARRAY<int>& face_tets=face_tetrahedron_hashtable.Get(face_key);
        if(face_tets.m==0){      
            ////do not allow eulerian particle to degenerate
            if(has_eulerian_particles&&!use_split_eulerian_particles){
                bool face_contains_eulerian_particle=eulerian_particles_hashtable.Contains(particles.guid(face_key(1)))
                    ||eulerian_particles_hashtable.Contains(particles.guid(face_key(2)))||eulerian_particles_hashtable.Contains(particles.guid(face_key(3)));
                if(face_contains_eulerian_particle)continue;
            }

            int t=triangle_mesh.elements.Append(face_key);

            ////update vtx_tri hashtable
            for(int j=1;j<=3;j++){int vtx=face_key(j);ARRAY<int>& vtx_tris=vertex_triangle_hashtable.Get_Or_Insert(vtx,ARRAY<int>());vtx_tris.Append(t);}

            ////update extra hashtable
            if(edge_triangle_hashtable){
                Add_Element(*edge_triangle_hashtable,face_key,t);
            }

            if(!preserve_thickness_in_remeshing){}   ////redistribute mass and momentum step 2, code removed
            else{
                //particles.thickness(v)=(particles.X(a)-particles.X(c)).Magnitude();
                
                TV_INT2 b_and_d=Other_Nodes_In_Element(face_key,v);
                T tet_vol=TETRAHEDRON<T>::Volume(particles.X(a),particles.X(c),particles.X(b_and_d(1)),particles.X(b_and_d(2)));
                lost_tet_vol+=tet_vol;
                T tri_area=TRIANGLE_3D<T>::Area(particles.X(v),particles.X(b_and_d(1)),particles.X(b_and_d(2)));
                added_tri_area+=tri_area;

                for(int j=1;j<=2;j++){int b=b_and_d(j);
                    //ARRAY<int>& b_tets=vertex_tetrahedron_hashtable.Get(b);
                    //if(b_tets.m==0)particles.thickness(b)=(T).5*particles.thickness(v);
                    //if(b_tets.m==0){    ////if no incident tet
                    //    T min_triangle_rim_thickness=min_thickness*(T)2;
                    //    particles.thickness(b)=max(min_triangle_rim_thickness,particles.thickness(b));
                    //}
                    particles.thickness(b)=tet_vol/tri_area;
                    particles.thickness(b)=clamp_min(particles.thickness(b),thickness_scale*(T).2); ////TODO: fix this
                }
            }

            if(use_dbg){
                TV_INT3 face_guid=Element(particles.guid,face_key);
                LOG::cout<<"ac: "<<particles.guid(a)<<", "<<particles.guid(c)<<", "<<particles.guid(v)<<std::endl;
                LOG::cout<<"tet degenerate to tri: "<<face_key<<" guid "<<face_guid<<", thickness: "<<particles.thickness(v)<<", tet_vol: "<<lost_tet_vol<<", tri_area: "<<added_tri_area<<std::endl;
            }
        }
    }

    if(!preserve_thickness_in_remeshing){}
    else{
        if(lost_tet_vol>0&&added_tri_area>(T)0){
            particles.thickness(v)=(T)3*lost_tet_vol/added_tri_area;
            particles.thickness(v)=clamp_min(particles.thickness(v),thickness_scale*(T).2); ////TODO: fix this
            //LOG::cout<<"deg thickness: "<<v<<" guid "<<particles.guid(v)<<": "<<particles.thickness(v)<<", "<<lost_tet_vol<<", "<<added_tri_area<<std::endl;
        }
    }

    //if(abs(convex_mass_new)<epsilon)convex_mass_new=(T)0;   ////fix numerical errors
    //PHYSBAM_ASSERT(convex_mass_new>=(T)0);

    ////replace a and c with v in tet elements
    for(int i=1;i<=new_one_ring_tets.m;i++){int t=new_one_ring_tets(i);
        TV_INT4 replaced_v;bool replaced=Replaced_Two(tetrahedron_mesh.elements(t),replaced_v,a,c,v);
        if(replaced){
            //TV_INT4 old_v=tetrahedron_mesh.elements(t);
            tetrahedron_mesh.elements(t)=replaced_v;

            //////dbg
            //TV_INT4 guid_old=Element(particles.guid,old_v);
            //TV_INT4 guid=Element(particles.guid,replaced_v);
            ////if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2018266)){
            //if(guid.Contains(2017378)&&guid.Contains(3525)&&guid.Contains(2009280)&&guid.Contains(2016233)){
            //    for(int j=1;j<=tetrahedron_mesh.elements.m;j++){
            //        TV_INT4 v=tetrahedron_mesh.elements(j);
            //        TV_INT4 guid_v=Element(particles.guid,v);
            //        if(guid_v.Contains(3525)&&guid_v.Contains(2009280)&&guid.Contains(2016233)){
            //            LOG::cout<<"face "<<j<<": "<<guid_v<<", new: "<<new_one_ring_tets.Contains(j)<<", old: "<<old_two_ring_tets.Contains(j)<<", ac: "<<ac_tets.Contains(j)<<", del: "<<tetrahedron_indices_to_remove.Contains(j)<<std::endl;
            //        }
            //    }
            //    bool non_posi_bef=Is_Non_Positive_Volume_Tetrahedron(particles.X(old_v(1)),particles.X(old_v(2)),particles.X(old_v(3)),particles.X(old_v(4)));
            //    bool non_posi_aft=Is_Non_Positive_Volume_Tetrahedron(particles.X(replaced_v(1)),particles.X(replaced_v(2)),particles.X(replaced_v(3)),particles.X(replaced_v(4)));
            //    double vol_bef=Accurate_Tetrahedron_Volume(particles.X(old_v(1)),particles.X(old_v(2)),particles.X(old_v(3)),particles.X(old_v(4)));
            //    double vol_aft=Accurate_Tetrahedron_Volume(particles.X(replaced_v(1)),particles.X(replaced_v(2)),particles.X(replaced_v(3)),particles.X(replaced_v(4)));
            //    LOG::cout<<"find int !!!!!!!!!!!!!!!!!!!!!!!!!!!!!: "<<t<<" ac: "<<particles.guid(a)<<", "<<particles.guid(c)<<", v: "<<particles.guid(v)<<", old: "<<guid_old<<", non_pos: "<<non_posi_bef<<", "<<non_posi_aft
            //        <<", vol: "<<vol_bef<<", "<<vol_aft<<", X:"<<particles.X(a)<<", "<<particles.X(c)<<", "<<particles.X(v)<<std::endl;
            //}
        }
    }
    ////replace a and c with v in tri elements
    for(int i=1;i<=new_one_ring_tris.m;i++){int t=new_one_ring_tris(i);
        TV_INT3 replaced_v;bool replaced=Replaced_Two(triangle_mesh.elements(t),replaced_v,a,c,v);
        if(replaced){triangle_mesh.elements(t)=replaced_v;}
    }

    ////delete a and c from and add v to the vertex_segment_hashtable
    ARRAY<int> ac_segs;
    for(int i=1;i<=a_and_c.m;i++){int a_or_c=a_and_c(i);
        if(vertex_segment_hashtable.Contains(a_or_c)){
            ARRAY<int>& a_or_c_seg=vertex_segment_hashtable.Get(a_or_c);
            for(int j=1;j<=a_or_c_seg.m;j++){ac_segs.Append(a_or_c_seg(j));}
            vertex_segment_hashtable.Delete(a_or_c);
        }
    }
    if(ac_segs.m>0){
        ARRAY<int>& v_segs=vertex_segment_hashtable.Get_Or_Insert(v);
        for(int i=1;i<=ac_segs.m;i++){v_segs.Append_Unique(ac_segs(i));}
    }
    ////replace a and c with v in segment elements
    for(int i=1;i<=ac_segs.m;i++){int s=ac_segs(i);
        TV_INT2 replaced_v;bool replaced=Replaced_Two(segment_mesh.elements(s),replaced_v,a,c,v);
        if(replaced){segment_mesh.elements(s)=replaced_v;}
    }

    if(!preserve_thickness_in_remeshing){}   ////redistributes mass and momentum step 3, code removed

    ////remove incident tets of edge ac 
    for(int i=1;i<=ac_tets.m;i++){tetrahedron_indices_to_remove.Append(ac_tets(i));}
    ////remove incident tris of edge ac
    for(int i=1;i<=ac_tris.m;i++){int t=ac_tris(i);triangle_indices_to_remove.Append(t);}
    ////remove vertices a and c
    particles_to_remove.Append(a);particles_to_remove.Append(c);

    if(p_thickness_needs_update_hashtable.Size()>0){
        for(HASHTABLE_ITERATOR<int> iter(p_thickness_needs_update_hashtable);iter.Valid();iter.Next()){
            int p=iter.Key();T old_vol=(T)0;
            if(p==v){old_vol=vtx_vol.Get(a)+vtx_vol.Get(c);}
            else old_vol=vtx_vol.Get(p);
            T A=Node_Triangle_Area_Sum(p,vertex_triangle_hashtable)+(T).5*Nodal_Tetrahedron_Surface_Area_Sum(p,vertex_tetrahedron_hashtable);
            T L=Node_Segment_Length_Sum(p,vertex_segment_hashtable);
            if(A!=0||L!=0){
                T t=Joint_Thickness(old_vol,A,L);
                //PHYSBAM_ASSERT(t>=(T)0);
                if(!(t>=(T)0))t=thickness_scale;
                particles.thickness(p)=t;
            }
            else particles.thickness(p)=(T)0;
        }
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Split triangle edge
////Indices: none
////Particles: particles added 
////Elements: triangles added and deleted 
////Topology: one ring of incident triangles of the splitted edge
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Triangle_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,
    const T split_weight/*=(T).5*/)
{
    TV_INT2 e_key=edge.Sorted();int a=e_key.x;int c=e_key.y;
    if(a==c)return false;   ////split fail: duplicated edges in the queue

    if(!use_split_eulerian_particles&&has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     ////split fail: should not collapse any eulerian particle

    ARRAY<int> e_tris;edge_triangle_hashtable.Get(e_key,e_tris); 

    ////add new particle v
    TV v_pos=((T)1-split_weight)*particles.X(a)+split_weight*particles.X(c);
    int v=Split_Particles(a,c,TV_INT(0,1,0),v_pos,split_weight,/*don't change mass and v*/false);particles_added.Append(v);
    if(use_dbg)LOG::cout<<"split tri edge "<<e_key<<" to "<<v<<" guid "<<Element(particles.guid,e_key)<<" to "<<particles.guid(v)<<std::endl;
    if(use_split_eulerian_particles&&eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))){
        eulerian_particles_hashtable.Set(particles.guid(v));
        added_eulerian_particle_parents_hashtable.Set(particles.guid(v),PAIR<TV_INT2,T>(TV_INT2(particles.guid(a),particles.guid(c)),split_weight));    
    }

    particles.on_boundary(v)(2)=e_tris.m<2;     ////fix the on_boundary for diagonal edge

    if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Triangle_Edge_Split(a,c,v,e_tris);*/} ////code removed
    else Simple_Redistribute_Mass_And_Momentum(a,c,v);
    
    if(has_rim){
        int a_on_boundary=particles.on_boundary(a)(TV::m-1);int c_on_boundary=particles.on_boundary(c)(TV::m-1);
        if(a_on_boundary!=c_on_boundary&&!eulerian_particles_hashtable.Contains(particles.guid(a))&&!eulerian_particles_hashtable.Contains(particles.guid(c))){
            particles.thickness(v)=a_on_boundary?particles.thickness(c):particles.thickness(a);
        }
    }

    particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);

    ////update topology
    for(int i=1;i<=e_tris.m;i++){
        int i_acd=e_tris(i);const TV_INT3 acd=triangle_mesh.elements(i_acd);
        TV_INT3 avd;TV_INT3 vcd;Two_Splitted_Triangle_Vertices(acd,a,c,v,avd,vcd);
        int i_avd=triangle_mesh.elements.Append(avd);triangles_added.Append(i_avd);
        int i_vcd=triangle_mesh.elements.Append(vcd);triangles_added.Append(i_vcd);
        triangles_to_remove.Append(i_acd);

        ////update edge_tri_hashtable
        Remove_Element(edge_triangle_hashtable,acd,i_acd);
        Add_Element(edge_triangle_hashtable,avd,i_avd);
        Add_Element(edge_triangle_hashtable,vcd,i_vcd);
    }
    Remove_Key(edge_triangle_hashtable,e_key);

    return true;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Triangle_Edge_Split(const int a,const int c,const int v,const ARRAY<int>& e_tris)
{
    int rim_p_num=(particles.distance_to_rim(a)==0?1:0)+(particles.distance_to_rim(c)==0?1:0);

    for(int i=1;i<=e_tris.m;i++){
        int b=Other_Node_In_Element(triangle_mesh.elements(e_tris(i)),a,c);
        T area_abc=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(c));
        T area_abv=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(v));
        T area_bcv=TRIANGLE_3D<T>::Area(particles.X(b),particles.X(c),particles.X(v));
        particles.dimensional_size(a)(2)+=((T)two_thirds*area_abv-(T)one_third*area_abc);
        particles.dimensional_size(b)(2)+=((T)one_third*(area_abv+area_bcv)-(T)one_third*area_abc);
        particles.dimensional_size(c)(2)+=((T)two_thirds*area_bcv-(T)one_third*area_abc);
    }

    T delta_mass_a=(T)0;T delta_mass_c=(T)0;T delta_area_a=(T)0;T delta_area_c=(T)0;
    for(int i=1;i<=e_tris.m;i++){
        int b=Other_Node_In_Element(triangle_mesh.elements(e_tris(i)),a,c);
        T area_abv=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(v));
        T area_bcv=TRIANGLE_3D<T>::Area(particles.X(b),particles.X(c),particles.X(v));
        delta_area_a+=(T)one_third*area_abv;
        delta_area_c+=(T)one_third*area_bcv;
        T delta_mass_vtx_a=(T)one_third*area_abv/particles.dimensional_size(a)(2)*particles.dimensional_mass(a)(TV::m-1);
        T delta_mass_vtx_c=(T)one_third*area_bcv/particles.dimensional_size(c)(2)*particles.dimensional_mass(c)(TV::m-1);
        delta_mass_a+=delta_mass_vtx_a;
        delta_mass_c+=delta_mass_vtx_c;
        //LOG::cout<<"a: "<<a<<"("<<Guid(a)<<")"<<delta_mass_vtx_a<<", "<<particles.dimensional_size(a)(2)<<", "<<particles.dimensional_mass(a)(TV::m-1)<<std::endl;
        //LOG::cout<<"c: "<<c<<"("<<Guid(c)<<")"<<delta_mass_vtx_c<<", "<<particles.dimensional_size(c)(2)<<", "<<particles.dimensional_mass(c)(TV::m-1)<<std::endl;
    }

    particles.dimensional_size(v)(2)=(delta_area_a+delta_area_c);
    particles.dimensional_size(a)(2)-=delta_area_a;
    particles.dimensional_size(c)(2)-=delta_area_c;

    ////update particle mass and velocity
    if(!use_rim_mass_redistribution)rim_p_num=0;

    if(rim_p_num==0){   ////both particles are interior
        particles.dimensional_mass(v)(TV::m-1)=(delta_mass_a+delta_mass_c);
        ////TODO: fix this
        particles.dimensional_mass(a)(TV::m-1)-=delta_mass_a;
        particles.dimensional_mass(c)(TV::m-1)-=delta_mass_c;
        PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>(T)0);
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.mass(a)=particles.dimensional_mass(a).Sum();
        particles.mass(c)=particles.dimensional_mass(c).Sum();
        PHYSBAM_ASSERT(delta_mass_a+delta_mass_c>(T)0);
        particles.V(v)=(delta_mass_a*particles.V(a)+delta_mass_c*particles.V(c))/(delta_mass_a+delta_mass_c);
    }
    else if(rim_p_num==1){  ////1 rim particle, 1 interior particle: only use the mass of interior particle
        int rim_p=(particles.distance_to_rim(a)==0?a:c);
        int interior_p=(rim_p==a?c:a);
        T delta_area_rim=(rim_p==a?delta_area_a:delta_area_c);
        T delta_area_interior=(rim_p==a?delta_area_c:delta_area_a);
        T old_delta_mass_rim=(rim_p==a?delta_mass_a:delta_mass_c);
        T delta_mass_interior=(rim_p==a?delta_mass_c:delta_mass_a);

        T delta_mass_rim=rim_interior_mass_transfer_coefficient*delta_area_rim/delta_area_interior*delta_mass_interior;    ////get mass from the rim with the same thickness as the interior particle
        if(delta_mass_rim>old_delta_mass_rim)delta_mass_rim=old_delta_mass_rim;

        particles.dimensional_mass(v)(TV::m-1)=(delta_mass_rim+delta_mass_interior);
        particles.dimensional_mass(rim_p)(TV::m-1)-=delta_mass_rim;
        particles.dimensional_mass(interior_p)(TV::m-1)-=delta_mass_interior;
        PHYSBAM_ASSERT(particles.dimensional_mass(rim_p)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(interior_p)(TV::m-1)>(T)0);
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.mass(rim_p)=particles.dimensional_mass(rim_p).Sum();
        particles.mass(interior_p)=particles.dimensional_mass(interior_p).Sum();
        particles.V(v)=(delta_mass_rim*particles.V(rim_p)+delta_mass_interior*particles.V(interior_p))/(delta_mass_rim+delta_mass_interior);

        //Thread_Mutex_Lock();
        //LOG::cout<<"1-rim p: "<<rim_p<<", "<<interior_p<<": "<<particles.mass(rim_p)<<", "<<particles.mass(interior_p)<<std::endl;
        //Thread_Mutex_Unlock();
    }
    else if(rim_p_num==2){  ////2 rim particles, averaging two rim particles
        delta_mass_a=(T)one_third*particles.dimensional_mass(a)(2);
        delta_mass_c=(T)one_third*particles.dimensional_mass(c)(2);

        particles.dimensional_mass(v)(TV::m-1)=(delta_mass_a+delta_mass_c);
        particles.dimensional_mass(a)(TV::m-1)-=delta_mass_a;
        particles.dimensional_mass(c)(TV::m-1)-=delta_mass_c;
        PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>(T)0);
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.mass(a)=particles.dimensional_mass(a).Sum();
        particles.mass(c)=particles.dimensional_mass(c).Sum();
        PHYSBAM_ASSERT(delta_mass_a+delta_mass_c>(T)0);
        particles.V(v)=(delta_mass_a*particles.V(a)+delta_mass_c*particles.V(c))/(delta_mass_a+delta_mass_c);
    }
}

//////////////////////////////////////////////////////////////////////////
////Topology Atomic Operation: Flip triangle edge
////Indices: none
////Particles: none
////Elements: triangles added and deleted 
////Topology: one ring of incident triangles of the splitted edge
////Degeneration: none
template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Flip_Triangle_Edge(const TV_INT2& edge,HASHTABLE<TV_INT2,ARRAY<int> >& edge_triangle_hashtable,const HASHTABLE<TV_INT2>& edge_on_tet_mesh_hashtable,HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable, 
    HASHTABLE<int>& particles_affected_hashtable,ARRAY<int>& triangles_to_remove,ARRAY<int>& triangles_added,/*result*/ARRAY<TV_INT2>& new_edges_to_flip)
{
    TV_INT2 e_key=edge.Sorted();
    ARRAY<int>& e_tri=edge_triangle_hashtable.Get(e_key);
    int a=e_key.x;int c=e_key.y;
    int i_abc=e_tri(1);int i_acd=e_tri(2);
    TV_INT3 reordered_vtx=Reordered(triangle_mesh.elements(e_tri(1)),a);
    if(reordered_vtx(2)==c){i_abc=e_tri(2);i_acd=e_tri(1);}
    int b=Other_Node_In_Element(triangle_mesh.elements(i_abc),a,c);
    int d=Other_Node_In_Element(triangle_mesh.elements(i_acd),a,c);  

    ////flip fail: should not collapse any eulerian particle
    if(has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))&&eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     

    ////flip fail: contact line
    if(use_contact_aware_triangle_remeshing){
        bool a_and_c_contact=Is_Contact_Boundary(particles.contact(a))&&Is_Contact_Boundary(particles.contact(c));
        bool b_and_d_contact=Is_Contact_Boundary(particles.contact(b))&&Is_Contact_Boundary(particles.contact(d));
        if(a_and_c_contact&&!b_and_d_contact)return false;
    }

    ////flip fail: non-manifold mesh
    if(edge_triangle_hashtable.Contains(TV_INT2(b,d).Sorted())){return false;}
    ////flip fail: edge on tet mesh
    if(edge_on_tet_mesh_hashtable.Contains(e_key)){return false;}

    //if(rim_vtx_vtx_hashtable.Contains(a)||rim_vtx_vtx_hashtable.Contains(c)){return false;}

    ////flip fail
    bool flip_to_improve_triangle_quality=false;
    ARRAY<TV_INT3> old_tris;old_tris.Append(TV_INT3(a,b,c));old_tris.Append(TV_INT3(a,c,d));
    ARRAY<TV_INT3> new_tris;new_tris.Append(TV_INT3(a,b,d));new_tris.Append(TV_INT3(b,c,d));
    T old_measure=Measure_Triangles(old_tris);T new_measure=Measure_Triangles(new_tris);T measure_coef=(T).95;
    if(old_measure<measure_coef*new_measure){flip_to_improve_triangle_quality=true;}

    bool flip_to_improve_triangle_thickness=false;bool not_flip_to_preserve_triangle_thickness=false;
    if(use_thickness_aware_triangle_remeshing){
        T ac_thickness_difference=abs(particles.thickness(a)-particles.thickness(c));
        T bd_thickness_difference=abs(particles.thickness(b)-particles.thickness(d));
        T max_factor=(T)3;T min_factor=(T)1/max_factor;
        if(bd_thickness_difference<ac_thickness_difference*(T)min_factor){flip_to_improve_triangle_thickness=true;}
        if(bd_thickness_difference>ac_thickness_difference*(T)max_factor){not_flip_to_preserve_triangle_thickness=true;}
    }

    if(!(flip_to_improve_triangle_thickness||(flip_to_improve_triangle_quality&&!not_flip_to_preserve_triangle_thickness))){return false;}

    ////flip fail: bad shape triangle
    T min_cosine_angle=cos((T)(.95*pi));
    T max_cosine_angle=cos((T)(.05*pi));
    if(Max_Cosine_Angle(particles.X(a),particles.X(b),particles.X(d))>max_cosine_angle||
        Max_Cosine_Angle(particles.X(b),particles.X(c),particles.X(d))>max_cosine_angle||
        Min_Cosine_Angle(particles.X(a),particles.X(b),particles.X(d))<min_cosine_angle||
        Min_Cosine_Angle(particles.X(b),particles.X(c),particles.X(d))<min_cosine_angle){return false;}
    ////flip fail: flipped triangles have large dihedral angle
    TV n_abd=(TV::Cross_Product(particles.X(b)-particles.X(a),particles.X(d)-particles.X(a))).Normalized();
    TV n_bcd=(TV::Cross_Product(particles.X(c)-particles.X(b),particles.X(d)-particles.X(b))).Normalized();
    if(TV::Angle_Between(n_abd,n_bcd)>(T).25*(T)pi){return false;}

    //////////////////////////////////////////////////////////////////////////
    ////start to flip
    if(!preserve_thickness_in_remeshing){/*Redistribute_Mass_And_Momentum_For_Triangle_Edge_Flip(a,c,b,d);*/} ////code removed
    else{}

    particles_affected_hashtable.Set(a);particles_affected_hashtable.Set(c);particles_affected_hashtable.Set(b);particles_affected_hashtable.Set(d);
    
    TV_INT3 vtx_acd(a,c,d);triangles_to_remove.Append(i_acd);Remove_Element(edge_triangle_hashtable,vtx_acd,i_acd);
    TV_INT3 vtx_abc(a,b,c);triangles_to_remove.Append(i_abc);Remove_Element(edge_triangle_hashtable,vtx_abc,i_abc);
    Remove_Key(edge_triangle_hashtable,e_key);

    TV_INT3 vtx_abd(a,b,d);int i_abd=triangle_mesh.elements.Append(vtx_abd);triangles_added.Append(i_abd);Add_Element(edge_triangle_hashtable,vtx_abd,i_abd);
    TV_INT3 vtx_bcd(b,c,d);int i_bcd=triangle_mesh.elements.Append(vtx_bcd);triangles_added.Append(i_bcd);Add_Element(edge_triangle_hashtable,vtx_bcd,i_bcd);

    ////update checklist
    ARRAY<TV_INT2 > edges_to_check;
    edges_to_check.Append(TV_INT2(b,c).Sorted());
    edges_to_check.Append(TV_INT2(c,d).Sorted());
    edges_to_check.Append(TV_INT2(d,a).Sorted());
    edges_to_check.Append(TV_INT2(a,b).Sorted());
    for(int i=1;i<=edges_to_check.m;i++){
        TV_INT2 e_key=edges_to_check(i);ARRAY<int>& e_tri=edge_triangle_hashtable.Get(e_key);if(e_tri.m!=2)continue;

        int a=e_key.x;int c=e_key.y;int i_abc=e_tri(1);int i_acd=e_tri(2);
        TV_INT3 reordered_vtx=Reordered(triangle_mesh.elements(e_tri(1)),a);
        if(reordered_vtx(2)==c){i_abc=e_tri(2);i_acd=e_tri(1);}
        int b=Other_Node_In_Element(triangle_mesh.elements(i_abc),a,c);
        int d=Other_Node_In_Element(triangle_mesh.elements(i_acd),a,c);  

        ARRAY<TV_INT3> old_tris;old_tris.Append(TV_INT3(a,b,c));old_tris.Append(TV_INT3(a,c,d));
        ARRAY<TV_INT3> new_tris;new_tris.Append(TV_INT3(a,b,d));new_tris.Append(TV_INT3(b,c,d));
        T old_measure=Measure_Triangles(old_tris);
        T new_measure=Measure_Triangles(new_tris);
        if(old_measure<measure_coef*new_measure){new_edges_to_flip.Append(TV_INT2(a,c));}
    }
    return true;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Redistribute_Mass_And_Momentum_For_Triangle_Edge_Flip(const int a,const int c,const int b,const int d)
{
    //////update particle mass and thickness
    //bool has_rim=(particles.distance_to_rim(a)==0||particles.distance_to_rim(b)==0||particles.distance_to_rim(c)==0||particles.distance_to_rim(d)==0);
    //bool has_rim=(rim_vtx_vtx_hashtable.Contains(a)||rim_vtx_vtx_hashtable.Contains(b)||rim_vtx_vtx_hashtable.Contains(c)||rim_vtx_vtx_hashtable.Contains(d));
    //if(!use_rim_mass_redistribution)has_rim=false;
    //has_rim=false;

    T area_acd=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(c),particles.X(d));
    T area_abc=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(c));
    T area_abd=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(d));
    T area_bcd=TRIANGLE_3D<T>::Area(particles.X(b),particles.X(c),particles.X(d));
    T old_area_part_a=(T)one_third*(area_abc+area_acd);
    T old_area_part_b=(T)one_third*area_abc;
    T old_area_part_c=(T)one_third*(area_abc+area_acd);
    T old_area_part_d=(T)one_third*area_acd;
    T new_area_part_a=(T)one_third*area_abd;
    T new_area_part_b=(T)one_third*(area_abd+area_bcd);
    T new_area_part_c=(T)one_third*area_bcd;
    T new_area_part_d=(T)one_third*(area_abd+area_bcd);
    T new_area_abcd=new_area_part_a+new_area_part_b+new_area_part_c+new_area_part_d;

    T old_mass_part_a=old_area_part_a/particles.dimensional_size(a)(2)*particles.dimensional_mass(a)(TV::m-1);
    T old_mass_part_b=old_area_part_b/particles.dimensional_size(b)(2)*particles.dimensional_mass(b)(TV::m-1);
    T old_mass_part_c=old_area_part_c/particles.dimensional_size(c)(2)*particles.dimensional_mass(c)(TV::m-1);
    T old_mass_part_d=old_area_part_d/particles.dimensional_size(d)(2)*particles.dimensional_mass(d)(TV::m-1);
    T mass_abcd=old_mass_part_a+old_mass_part_b+old_mass_part_c+old_mass_part_d;

    T new_mass_part_a=new_area_part_a/new_area_abcd*mass_abcd;
    T new_mass_part_b=new_area_part_b/new_area_abcd*mass_abcd;
    T new_mass_part_c=new_area_part_c/new_area_abcd*mass_abcd;
    T new_mass_part_d=new_area_part_d/new_area_abcd*mass_abcd;

    //if(!has_rim){
        particles.dimensional_mass(a)(TV::m-1)+=(new_mass_part_a-old_mass_part_a);
        particles.dimensional_mass(b)(TV::m-1)+=(new_mass_part_b-old_mass_part_b);
        particles.dimensional_mass(c)(TV::m-1)+=(new_mass_part_c-old_mass_part_c);
        particles.dimensional_mass(d)(TV::m-1)+=(new_mass_part_d-old_mass_part_d);
        PHYSBAM_ASSERT(particles.dimensional_mass(a)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(b)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(c)(TV::m-1)>(T)0);
        PHYSBAM_ASSERT(particles.dimensional_mass(d)(TV::m-1)>(T)0);
        particles.mass(a)=particles.dimensional_mass(a).Sum();
        particles.mass(b)=particles.dimensional_mass(b).Sum();
        particles.mass(c)=particles.dimensional_mass(c).Sum();
        particles.mass(d)=particles.dimensional_mass(d).Sum();
    //}

    //particles.V(a)+=(-delta_momentum_a)/particles.mass(a);
    //particles.V(b)+=(-delta_momentum_b)/particles.mass(b);
    //particles.V(c)+=(-delta_momentum_c)/particles.mass(c);
    //particles.V(d)+=(-delta_momentum_d)/particles.mass(d);

    particles.dimensional_size(a)(2)+=(new_area_part_a-old_area_part_a);
    particles.dimensional_size(b)(2)+=(new_area_part_b-old_area_part_b);
    particles.dimensional_size(c)(2)+=(new_area_part_c-old_area_part_c);
    particles.dimensional_size(d)(2)+=(new_area_part_d-old_area_part_d);
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Collapse_Triangle_Edge(const TV_INT2& edge,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,
    HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable, ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable,
    ARRAY<int>& triangle_indices_to_remove,ARRAY<int>& segment_indices_to_remove)
{
    TV_INT2 e_key=edge.Sorted();
    int a=e_key.x;int c=e_key.y;  
    TV_INT2 a_and_c=e_key; ////vertex a, c

    if(a==c)return false;   ////collapse fail: duplicated edges in the queue
    if(has_eulerian_particles&&!use_split_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     ////collapse fail: should not collapse any eulerian particle

    T particle_length_scale=Triangle_Edge_Length_Scale(a,c);
    T edge_length=(particles.X(a)-particles.X(c)).Magnitude();
    if(edge_length>=length_to_collapse_coef*particle_length_scale){return false;}    ////collapse fail: no longer short
    //ARRAY<int>& a_tri=vertex_triangle_hashtable.Get(a);
    //ARRAY<int>& c_tri=vertex_triangle_hashtable.Get(c);

    //TV old_a_dimensional_mass=particles.dimensional_mass(a);
    //TV old_c_dimensional_mass=particles.dimensional_mass(c);

    //if(a_tri.m<=1||c_tri.m<=1)return false;     ////???

    int rim_p_num=(rim_vtx_vtx_hashtable.Contains(a)?1:0)+(rim_vtx_vtx_hashtable.Contains(c)?1:0);

    ARRAY<int> old_two_ring_tri;
    ARRAY<int> new_one_ring_tri;
    ARRAY<int> one_ring_nodes;  ////one ring neighboring nodes of v
    ARRAY<int> ac_tri;      ////incident triangles of edge ac (if exists)
    for(int i=1;i<=a_and_c.m;i++){
        int a_or_c=a_and_c(i);
        ARRAY<int>& v_tri=vertex_triangle_hashtable.Get(a_or_c);
        for(int j=1;j<=v_tri.m;j++){
            int t=v_tri(j);VECTOR<int,3> vtx=triangle_mesh.elements(t);
            if(vtx.Contains(a)&&vtx.Contains(c)){ac_tri.Append_Unique(t);}
            else{new_one_ring_tri.Append(t);}

            old_two_ring_tri.Append_Unique(t);
            for(int d=1;d<=3;d++){if(vtx(d)!=a&&vtx(d)!=c){one_ring_nodes.Append_Unique(vtx(d));}}
        }
    }

    if(ac_tri.m==0)return false;

    HASHTABLE<int,T> vtx_vol;
    for(int i=1;i<=ac_tri.m;i++){int t=ac_tri(i);
        const TV_INT3& vtx=triangle_mesh.elements(t);
        for(int j=1;j<=3;j++){
            int p=vtx(j);
            if(!vtx_vol.Contains(p)){
                T tet_area_sum=Nodal_Tetrahedron_Surface_Area_Sum(p,vertex_tetrahedron_hashtable);
                T tri_area_sum=Node_Triangle_Area_Sum(p,vertex_triangle_hashtable);
                T seg_length_sum=Node_Segment_Length_Sum(p,vertex_segment_hashtable);
                T vol=(tri_area_sum+(T).5*tet_area_sum)*particles.thickness(p)+seg_length_sum*(T).25*(T)pi*pow<2>(particles.thickness(p));
                vtx_vol.Set(p,vol);
            }
        }
    }

    ////compute the candidate position
    TV pos_v;

    if(particles.on_boundary(a)(TV::m)&&particles.on_boundary(c)(TV::m)){
        return false;   ////collapse fail: both particles are on the volume boundary 
    }
    ////TODO: may need to handle this case for two volumes colliding
    else if(particles.on_boundary(a)(TV::m)||particles.on_boundary(c)(TV::m)){  ////one of the particles is on tet mesh boundary
        int volume_boundary_idx=particles.on_boundary(a)(TV::m)?a:c;
        
        ////collapse fail: one particle on tet mesh boundary, the other on tri mesh boundary, use shorter length scale
        if(use_fix_poping){
            int other_idx=volume_boundary_idx==a?c:a;
            if(particles.on_boundary(other_idx)(TV::m-1)){
                //if(edge_length>=(T).25*length_to_collapse_coef*particle_length_scale)return false; 
                return false;
            }
        }

        pos_v=particles.X(volume_boundary_idx); ////collapse to the volume boundary
    }
    else{
        ////one of the particles is on the triangle mesh boundary
        if(particles.on_boundary(a)(TV::m-1)!=particles.on_boundary(c)(TV::m-1)){
            int boundary_idx=particles.on_boundary(a)(TV::m-1)?a:c;pos_v=particles.X(boundary_idx);}
        else pos_v=(T).5*(particles.X(a)+particles.X(c));
    }

    if(use_contact_aware_triangle_remeshing){
        bool contact_a=Is_Contact_Boundary(particles.contact(a));bool contact_c=Is_Contact_Boundary(particles.contact(c));
        if(contact_a!=contact_c){int contact_idx=contact_a?a:c;pos_v=particles.X(contact_idx);}
    }

    ////collapse fail: triangle is inverted or almost inverted
    bool has_inverted_triangle=false;
    bool has_large_angle=false;
    T cos_epsilon=(T)1e-10;
    T max_cos=(T)1-cos_epsilon;
    for(int i=1;i<=new_one_ring_tri.m;i++){
        int t=new_one_ring_tri(i);
        TV_INT3 vtx=triangle_mesh.elements(t);
        VECTOR<TV,3> vtx_pos;for(int d=1;d<=3;d++)vtx_pos(d)=particles.X(vtx(d));
        TV old_normal=TV::Cross_Product(vtx_pos(2)-vtx_pos(1),vtx_pos(3)-vtx_pos(1)).Normalized();
        for(int d=1;d<=3;d++){if(vtx(d)==a||vtx(d)==c)vtx_pos(d)=pos_v;}
        TV new_normal=TV::Cross_Product(vtx_pos(2)-vtx_pos(1),vtx_pos(3)-vtx_pos(1)).Normalized();
        if(TV::Dot_Product(old_normal,new_normal)<cos_epsilon){has_inverted_triangle=true;break;}
        if(Max_Cosine_Angle(vtx_pos(1),vtx_pos(2),vtx_pos(3))>max_cos){has_large_angle=true;/*break;*/}
    }
    if(has_inverted_triangle){return false;}
    if(has_large_angle){
        if(edge_length>(T).2*length_to_collapse_coef*particle_length_scale){return false;} ////if edge is too short, ignore large angle
    }

    //////////////////////////////////////////////////////////////////////////
    //// start to collapse

    ////add new particle v and update particle attributes
    T weight=(T).5;if(pos_v==particles.X(a))weight=(T)0;else if(pos_v==particles.X(c))weight=(T)1;
    int v=Merge_Particles(a,c,TV_INT(1,1,1),pos_v,weight,false);particles_added.Append(v);

    if(has_eulerian_particles&&use_split_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c)))){
        eulerian_particles_hashtable.Delete_If_Present(particles.guid(a));eulerian_particles_hashtable.Delete_If_Present(particles.guid(c));
        eulerian_particles_hashtable.Set(particles.guid(v));
        added_eulerian_particle_parents_hashtable.Set(particles.guid(v),PAIR<TV_INT2,T>(TV_INT2(particles.guid(a),particles.guid(c)),weight));
    }

    if(!preserve_thickness_in_remeshing){}  ////code removed 
    else{
        Simple_Redistribute_Mass_And_Momentum(a,c,v);
        
        //bool is_a_rim=particles.on_boundary(a)(TV::m-1)&&!vertex_tetrahedron_hashtable.Contains(a);
        //bool is_c_rim=particles.on_boundary(c)(TV::m-1)&&!vertex_tetrahedron_hashtable.Contains(c);
        //if(is_a_rim!=is_c_rim){LOG::cout<<"double v: "<<v<<", "<<particles.guid(v)<<std::endl;} ////dbg!!!

        //////correct thickness based on incident area
        //T tri_area_a=Incident_Triangle_Area(a,vertex_triangle_hashtable);
        //T tri_area_c=Incident_Triangle_Area(c,vertex_triangle_hashtable);
        ////T tri_area_v=Incident_Triangle_Area(edge,v,new_one_ring_tri);
        //if(tri_area_a+tri_area_c>(T)0)particles.thickness(v)=(particles.thickness(a)*tri_area_a+particles.thickness(c)*tri_area_c)/(tri_area_a+tri_area_c);
    }

    if(use_dbg)LOG::cout<<"collapse tri edge "<<a<<", "<<c<<" to "<<v<<" guid "<<particles.guid(a)<<", "<<particles.guid(c)<<" to "<<particles.guid(v)<<std::endl;

    ////delete triangles containing ac from the vertex_triangle_hashtable
    HASHTABLE<int> p_thickness_needs_update_hashtable;
    for(int i=1;i<=ac_tri.m;i++){
        int t_abc=ac_tri(i);
        TV_INT3 vtx=triangle_mesh.elements(t_abc);
        int b=Other_Node_In_Element(vtx,a,c);
        ARRAY<int>& b_tris=vertex_triangle_hashtable.Get(b);
        Remove_From_Array(b_tris,t_abc);
        ARRAY<int>& a_tris=vertex_triangle_hashtable.Get(a);
        Remove_From_Array(a_tris,t_abc);
        ARRAY<int>& c_tris=vertex_triangle_hashtable.Get(c);
        Remove_From_Array(c_tris,t_abc);

        ////triangle degenerate to segment
        bool is_ab_on_mesh=Is_Edge_On_Mesh(TV_INT2(a,b),vertex_triangle_hashtable,vertex_tetrahedron_hashtable);
        bool is_cb_on_mesh=Is_Edge_On_Mesh(TV_INT2(c,b),vertex_triangle_hashtable,vertex_tetrahedron_hashtable);
        if(!is_ab_on_mesh&&!is_cb_on_mesh){   
            if(use_dbg)LOG::cout<<"triangle deg to segment: "<<vtx<<" to "<<TV_INT2(b,v)<<" guid "<<Element(particles.guid,TV_INT2(b,v))<<std::endl;
            int s=segment_mesh.elements.Append(TV_INT2(b,v));
            if(!preserve_thickness_in_remeshing){}   ////transfer mass from dimension 2 to 1, code removed
            else{
                //T A_abc=TRIANGLE_3D<T>::Area(particles.X(a),particles.X(b),particles.X(c));
                //T Vol_b=(T)one_third*A_abc*particles.thickness(b);
                //T Vol_ac=(T)one_third*A_abc*(particles.thickness(a)+particles.thickness(c));
                //T seg_length=(particles.X(v)-particles.X(b)).Magnitude();
                //particles.thickness(b)=sqrt((T)2*Vol_b/((T)pi*(T).5*seg_length));
                //particles.thickness(v)=sqrt((T)2*Vol_ac/((T)pi*(T).5*seg_length));
                p_thickness_needs_update_hashtable.Set(v);
                p_thickness_needs_update_hashtable.Set(b);
            }

            particles_affected_hashtable.Set(b);
            ARRAY<int>& b_seg=vertex_segment_hashtable.Get_Or_Insert(b);b_seg.Append(s);
            ARRAY<int>& v_seg=vertex_segment_hashtable.Get_Or_Insert(v);v_seg.Append(s);
        }
    }

    ////delete a and c from and add v to the vertex_triangle_hashtable
    for(int i=1;i<=a_and_c.m;i++){
        int a_or_c=a_and_c(i);
        vertex_triangle_hashtable.Delete(a_or_c);
    }
    vertex_triangle_hashtable.Insert(v,new_one_ring_tri);

    ////delete a and c from and add v to the vertex_tetrahedra_hashtable
    Merge_Keys(vertex_tetrahedron_hashtable,edge,v);
    

    ////delete a and c from and add v to the vertex_segment_hashtable
    Merge_Keys(vertex_segment_hashtable,edge,v);

    if(!preserve_thickness_in_remeshing){}  ////no tri on v anymore, transfer all 2d mass to 1d (if has segment) or 3d, code removed

    ////replace a and c with v in triangle elements and check (and remove) redundant triangles
    ////e.g., tri bda, bdc -> bdv, (redundant) bdv 
    ARRAY<TV_INT3> one_ring_tris_checking_redundancy;
    ARRAY<int> redundant_tris;
    for(int i=1;i<=old_two_ring_tri.m;i++){
        int t=old_two_ring_tri(i);
        TV_INT3 vtx=triangle_mesh.elements(t);

        if(vtx.Contains(a)&&vtx.Contains(c)){   ////remove incident triangles of edge ac
            redundant_tris.Append(t);
        }
        else{
            TV_INT3 replaced_v;bool replaced=Replaced_Two(triangle_mesh.elements(t),replaced_v,a,c,v);
            if(replaced){
                TV_INT3 uni_old_v=Unique_Ordered(triangle_mesh.elements(t));
                TV_INT3 uni_new_v=Unique_Ordered(replaced_v);
                triangle_mesh.elements(t)=replaced_v;
            }

            TV_INT3 t_key=replaced_v.Sorted();
            if(!one_ring_tris_checking_redundancy.Contains(t_key)){
                one_ring_tris_checking_redundancy.Append(t_key);
            }
            else{   ////remove redundant triangle
                redundant_tris.Append(t);
            }
        }
    }
    for(int i=1;i<=redundant_tris.m;i++){
        int t=redundant_tris(i);triangle_indices_to_remove.Append(t);
    }
    ////delete redundant triangles from its corresponding vertices in the hashtable
    for(int i=1;i<=redundant_tris.m;i++){
        VECTOR<int,3> vtx=triangle_mesh.elements(redundant_tris(i));
        for(int d=1;d<=3;d++){
            int v=vtx(d);
            if(v!=a&&v!=c){
                ARRAY<int>& v_tri=vertex_triangle_hashtable.Get(v);
                Remove_From_Array(v_tri,redundant_tris(i));
            }
        }
    }

    ////replace a and c with v in tetrahedron elements
    ARRAY<int>* v_tets=vertex_tetrahedron_hashtable.Get_Pointer(v);
    if(v_tets)Replace_Vertex_In_Elements(tetrahedron_mesh.elements,*v_tets,edge,v);

    ////replace a and c with v in segment elements
    ARRAY<int>* v_segs=vertex_segment_hashtable.Get_Pointer(v);
    if(v_segs)Replace_Vertex_In_Elements(segment_mesh.elements,*v_segs,edge,v);

    ////update segments to remove
    ARRAY<int>& v_tris=vertex_triangle_hashtable.Get(v);
    if(v_segs)for(int i=1;i<=(*v_segs).m;i++){
        int s=(*v_segs)(i);
        TV_INT2 vtx=segment_mesh.elements(s);
        int other_v=Other_Node_In_Element(vtx,v);
        bool seg_on_tri=false;
        for(int j=1;j<=v_tris.m;j++){
            const TV_INT3& tri_v=triangle_mesh.elements(v_tris(j));
            if(tri_v.Contains(other_v)){
                seg_on_tri=true;break;
            }
        }
        if(seg_on_tri){segment_indices_to_remove.Append_Unique(s);}
    }

    ////remove particles a c 
    particles_to_remove.Append(a);
    particles_to_remove.Append(c);

    ////update rim vtx hashtable
    if(rim_p_num==2){
        bool has_pinching_off_point=false;
        ARRAY<int>& a_nbs=rim_vtx_vtx_hashtable.Get(a);
        if(!a_nbs.Contains(c))has_pinching_off_point=true;
        int degenerated_rim_p=0;
        if(has_pinching_off_point){
            ARRAY<int>& c_nbs=rim_vtx_vtx_hashtable.Get(c);
            for(int i=1;i<=a_nbs.m;i++){
                if(c_nbs.Contains(a_nbs(i))){
                    degenerated_rim_p=a_nbs(i);    ////find point shared by a and c
                    break;
                }
            }
        }

        ARRAY<int>& v_nbs=rim_vtx_vtx_hashtable.Get_Or_Insert(v,ARRAY<int>());
        for(int j=1;j<=a_and_c.m;j++){
            int a=a_and_c(j);int c=Other_Node_In_Element(a_and_c,a);
            if(rim_vtx_vtx_hashtable.Contains(a)){
                ARRAY<int>& a_nbs=rim_vtx_vtx_hashtable.Get(a);
                for(int i=1;i<=a_nbs.m;i++){
                    if(a_nbs(i)!=c)v_nbs.Append_Unique(a_nbs(i));
                }
                rim_vtx_vtx_hashtable.Delete(a);
            }
        }

        if(has_pinching_off_point&&degenerated_rim_p!=0){
            ARRAY<int>& v_nbs=rim_vtx_vtx_hashtable.Get(v);
            Remove_From_Array(v_nbs,degenerated_rim_p);
            rim_vtx_vtx_hashtable.Delete(degenerated_rim_p);
        }
    }
    else if(rim_p_num==1){
        ARRAY<int>& v_nbs=rim_vtx_vtx_hashtable.Get_Or_Insert(v,ARRAY<int>());
        int rp=rim_vtx_vtx_hashtable.Contains(a)?a:c;
        ARRAY<int>& rp_nbs=rim_vtx_vtx_hashtable.Get(rp);
        for(int i=1;i<=rp_nbs.m;i++){
            v_nbs.Append(rp_nbs(i));
        }
        rim_vtx_vtx_hashtable.Delete(rp);
    }

    if(p_thickness_needs_update_hashtable.Size()>0){
        for(HASHTABLE_ITERATOR<int> iter(p_thickness_needs_update_hashtable);iter.Valid();iter.Next()){
            int p=iter.Key();T old_vol=(T)0;
            if(p==v){old_vol=vtx_vol.Get(a)+vtx_vol.Get(c);}
            else old_vol=vtx_vol.Get(p);
            T A=Node_Triangle_Area_Sum(p,vertex_triangle_hashtable)+(T).5*Nodal_Tetrahedron_Surface_Area_Sum(p,vertex_tetrahedron_hashtable);
            T L=Node_Segment_Length_Sum(p,vertex_segment_hashtable);
            if(A!=0||L!=0){
                T t=Joint_Thickness(old_vol,A,L);
                //PHYSBAM_ASSERT(t>=(T)0);
                if(!(t>=(T)0))t=thickness_scale;
                particles.thickness(p)=t;
            }
            else particles.thickness(p)=(T)0;
        }
    }

    return true;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Vertex_Segment_Length_Sum(const int a,const HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable)
{
    if(!vertex_segment_hashtable.Contains(a))return (T)0;

    T vtx_seg_length_sum=(T)0;
    const ARRAY<int>& a_segs=vertex_segment_hashtable.Get(a);
    for(int j=1;j<=a_segs.m;j++){
        const TV_INT2& v=segment_mesh.elements(a_segs(j));
        T length=(T).5*(particles.X(v(1))-particles.X(v(2))).Magnitude();
        vtx_seg_length_sum+=length;
    }
    return vtx_seg_length_sum;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Collapse_Segment_Edge(const int segment_index,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable, ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,ARRAY<int>& segment_indices_to_remove)
{
    TV_INT2 e_key=segment_mesh.elements(segment_index).Sorted();int a=e_key(1);int c=e_key(2);
    if(a==c)return false;   ////collapse fail: identical vertices
    if(has_eulerian_particles&&!use_split_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     ////collapse fail: should not collapse any eulerian particle
    if(use_mesh_edge_connect&&(Is_Contact_With_Fluid(particles.contact(a))||Is_Contact_With_Fluid(particles.contact(c))))return false;  ////merge fail: do not collapse connect segment

    T segment_length=(particles.X(a)-particles.X(c)).Magnitude();
    T particle_length_scale=use_adaptive_segment_remeshing?(T).5*(particles.length_scale(a)+particles.length_scale(c)):length_scale;
    if(segment_length>length_to_collapse_coef*particle_length_scale)return false;  ////collapse fail: no longer short

    if(use_mesh_edge_connect){TV_INT2 guid_key=Element(particles.guid,e_key).Sorted();if(mesh.connected_segment_hashtable.Contains(guid_key))return false;} ////merge fail, has connected segment

    ////compute the candidate position
    TV pos_v;
    bool a_on_higher_dim_boundary=(vertex_triangle_hashtable.Contains(a)&&vertex_triangle_hashtable.Get(a).m>0)||(vertex_tetrahedron_hashtable.Contains(a)&&vertex_tetrahedron_hashtable.Get(a).m>0);
    bool c_on_higher_dim_boundary=(vertex_triangle_hashtable.Contains(c)&&vertex_triangle_hashtable.Get(c).m>0)||(vertex_tetrahedron_hashtable.Contains(c)&&vertex_tetrahedron_hashtable.Get(c).m>0);        ////also handle the vertex inside mesh case

    int boundary_idx=0;
    if(a_on_higher_dim_boundary && c_on_higher_dim_boundary)return false;    ////collapse fail: both particles are on tet or tri boundary
    else if(a_on_higher_dim_boundary || c_on_higher_dim_boundary){boundary_idx=a_on_higher_dim_boundary?a:c;pos_v=particles.X(boundary_idx);}
    else{
        bool a_on_s_boundary=particles.on_boundary(a)(1)!=0;
        bool c_on_s_boundary=particles.on_boundary(c)(1)!=0;
        if((a_on_s_boundary&&!c_on_s_boundary)||(c_on_s_boundary&&!a_on_s_boundary)){
            boundary_idx=a_on_s_boundary?a:c;pos_v=particles.X(boundary_idx);}
        else pos_v=(T).5*(particles.X(a)+particles.X(c));
    }

    //////////////////////////////////////////////////////////////////////////
    //// start to collapse

    T a_and_c_V_sum=(T)0;
    for(int d=1;d<=2;d++){
        int p=e_key(d);
        T vtx_seg_length_sum=Vertex_Segment_Length_Sum(p,vertex_segment_hashtable);
        a_and_c_V_sum+=vtx_seg_length_sum*pow<2>(particles.thickness(p));
        if(has_rim&&particles.on_boundary(p)(1)){
            T hemisphere_vol=(T).5*(T)two_thirds*pow<3>(particles.thickness(p));
            a_and_c_V_sum+=hemisphere_vol;
        }
    }

    ////add new particle v and update particle attributes
    T weight=(T).5;if(pos_v==particles.X(a))weight=(T)0;else if(pos_v==particles.X(c))weight=(T)1;
    int v=Merge_Particles(a,c,TV_INT(1,1,1),pos_v,weight);particles_added.Append(v);
    if(use_dbg)LOG::cout<<"collapse seg N"<<(mpi_codim?mpi_codim->node_id:0)<<": "<<a<<", "<<c<<" to "<<v<<" guid "<<particles.guid(a)<<", "<<particles.guid(c)<<" "<<particles.guid(v)<<std::endl;

    if(has_eulerian_particles&&use_split_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c)))){
        eulerian_particles_hashtable.Delete_If_Present(particles.guid(a));eulerian_particles_hashtable.Delete_If_Present(particles.guid(c));
        eulerian_particles_hashtable.Set(particles.guid(v));
        added_eulerian_particle_parents_hashtable.Set(particles.guid(v),PAIR<TV_INT2,T>(TV_INT2(particles.guid(a),particles.guid(c)),weight));
    }

    if(!preserve_thickness_in_remeshing){}
    else{particles.thickness(v)=sqrt(((T)1-weight)*pow<2>(particles.thickness(a))+weight*pow<2>(particles.thickness(c)));}

    ////delete a and c from and add v to the vertex_segment_hashtable
    ARRAY<int> v_seg;
    for(int i=1;i<=e_key.m;i++){
        int a_or_c=e_key(i);
        ARRAY<int>& a_or_c_seg=vertex_segment_hashtable.Get(a_or_c);
        Remove_From_Array(a_or_c_seg,segment_index);
        for(int j=1;j<=a_or_c_seg.m;j++){
            v_seg.Append(a_or_c_seg(j));
        }
        vertex_segment_hashtable.Delete(a_or_c);
    }
    if(v_seg.m>0){
        ////replace a and c with v in segment elements
        ////remove duplicated segments
        HASHTABLE<TV_INT2> v_segs_vertices;
        ARRAY<int> v_segs_to_rmv;
        for(int i=1;i<=v_seg.m;i++){
            int s=v_seg(i);
            segment_mesh.elements(s)=Replaced_Two(segment_mesh.elements(s),a,c,v);
            TV_INT2 v_key=segment_mesh.elements(s).Sorted();
            if(!v_segs_vertices.Contains(v_key)){v_segs_vertices.Set(v_key);}
            else{   ////duplicated segments
                v_segs_to_rmv.Append(s);
                segment_indices_to_remove.Append(s);
            }
        }
        for(int i=1;i<=v_segs_to_rmv.m;i++){
            int s_to_rmv=v_segs_to_rmv(i);
            Remove_From_Array(v_seg,s_to_rmv);
        }
        vertex_segment_hashtable.Insert(v,v_seg);
    }
    else{
        if(!a_on_higher_dim_boundary&&!c_on_higher_dim_boundary){
            T seg_vol=segment_length*(T).125*(T)pi*(pow<2>(particles.thickness(a))+pow<2>(particles.thickness(c)));
            particles.thickness(v)=std::pow((T)6*seg_vol/(T)pi,(T)one_third);
        }
    }   ////segment degenerate to particle, ////transfer 1-dim mass back to 2-dim mass ??, code removed

    ////delete a and c from and add v to the vertex_triangle_hashtable
    ARRAY<int> v_tris;
    if(boundary_idx!=0){
        int b=boundary_idx;
        if(vertex_triangle_hashtable.Contains(b)){
            ARRAY<int> b_tris=vertex_triangle_hashtable.Get(b);
            vertex_triangle_hashtable.Delete(b);
            vertex_triangle_hashtable.Insert(v,b_tris);
            for(int i=1;i<=b_tris.m;i++){
                int t=b_tris(i);
                triangle_mesh.elements(t)=Replaced(triangle_mesh.elements(t),b,v);
            }
        }
        if(vertex_tetrahedron_hashtable.Contains(b)){
            ARRAY<int> b_tets=vertex_tetrahedron_hashtable.Get(b);
            vertex_tetrahedron_hashtable.Delete(b);
            vertex_tetrahedron_hashtable.Insert(v,b_tets);
            for(int i=1;i<=b_tets.m;i++){
                int t=b_tets(i);TV_INT4 old_v=tetrahedron_mesh.elements(t);
                tetrahedron_mesh.elements(t)=Replaced(tetrahedron_mesh.elements(t),b,v);
                //LOG::cout<<"replace b N"<<(mpi_codim?mpi_codim->node_id:0)<<": "<<b<<" guid"<<particles.guid(b)<<" tet: "<<old_v<<" -> "<<tetrahedron_mesh.elements(t)<<std::endl;
            }
        }
    }


    segment_indices_to_remove.Append(segment_index);
    particles_to_remove.Append(a);
    particles_to_remove.Append(c);
    
    ////update v thickness
    T v_seg_length_sum=Vertex_Segment_Length_Sum(v,vertex_segment_hashtable);
    if(v_seg_length_sum>0)particles.thickness(v)=sqrt(a_and_c_V_sum/v_seg_length_sum);
    else{particles.thickness(v)=std::pow((T)1.5*a_and_c_V_sum,(T)one_third);}

    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Snap_Boundary_Vertex(const int a,const int c,HASHTABLE<int,ARRAY<int> >& vertex_tetrahedron_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_triangle_hashtable,HASHTABLE<int,ARRAY<int> >& vertex_segment_hashtable,
    ARRAY<int>& particles_to_remove,ARRAY<int>& particles_added,HASHTABLE<int>& particles_affected_hashtable)
{
    TV_INT2 ac(a,c);
    if(a==c)return false;   ////collapse fail: duplicated edges in the queue
    if(has_eulerian_particles&&(eulerian_particles_hashtable.Contains(particles.guid(a))||eulerian_particles_hashtable.Contains(particles.guid(c))))return false;     ////collapse fail: should not collapse any eulerian particle
    T particle_length_scale=Triangle_Edge_Length_Scale(a,c);
    if((particles.X(a)-particles.X(c)).Magnitude()>=(T).5*particle_length_scale)return false;    ////collapse fail: no longer short

    TV pos_v=(T).5*(particles.X(a)+particles.X(c));
    ////TODO: collapse fail: inverted tetrahedron

    //////collapse fail: triangle is inverted or almost inverted
    //bool has_inverted_triangle=false;
    //bool has_large_angle=false;
    //T cos_epsilon=(T)1e-10;
    //T max_cos=(T)1-cos_epsilon;
    //for(int i=1;i<=new_one_ring_tri.m;i++){
    //    int t=new_one_ring_tri(i);
    //    TV_INT3 vtx=triangle_mesh.elements(t);
    //    VECTOR<TV,3> vtx_pos;for(int d=1;d<=3;d++)vtx_pos(d)=particles.X(vtx(d));
    //    TV old_normal=TV::Cross_Product(vtx_pos(2)-vtx_pos(1),vtx_pos(3)-vtx_pos(1)).Normalized();
    //    for(int d=1;d<=3;d++){if(vtx(d)==a||vtx(d)==c)vtx_pos(d)=pos_v;}
    //    TV new_normal=TV::Cross_Product(vtx_pos(2)-vtx_pos(1),vtx_pos(3)-vtx_pos(1)).Normalized();
    //    if(TV::Dot_Product(old_normal,new_normal)<cos_epsilon){has_inverted_triangle=true;break;}
    //    if(Max_Cosine_Angle(vtx_pos(1),vtx_pos(2),vtx_pos(3))>max_cos){has_large_angle=true;break;}
    //}
    //if(has_inverted_triangle||has_large_angle){return false;}

    HASHTABLE<int,T> vtx_vol;
    for(int i=1;i<=2;i++){
        int p=ac(i);
        if(!vtx_vol.Contains(p)){
            T tet_area_sum=Nodal_Tetrahedron_Surface_Area_Sum(p,vertex_tetrahedron_hashtable);
            T tri_area_sum=Node_Triangle_Area_Sum(p,vertex_triangle_hashtable);
            T seg_length_sum=Node_Segment_Length_Sum(p,vertex_segment_hashtable);
            T vol=(tri_area_sum+(T).5*tet_area_sum)*particles.thickness(p)+seg_length_sum*(T).25*(T)pi*pow<2>(particles.thickness(p));
            vtx_vol.Set(p,vol);
        }
    }

    //////////////////////////////////////////////////////////////////////////
    //// start to snap

    ////add new particle v and update particle attributes
    T weight=(T).5;if(pos_v==particles.X(a))weight=(T)0;else if(pos_v==particles.X(c))weight=(T)1;
    int v=Merge_Particles(a,c,TV_INT(1,1,1),pos_v,weight,false);
    particles_added.Append(v);
    particles_to_remove.Append(a);particles_to_remove.Append(c);
    //LOG::cout<<"snap "<<a<<", "<<c<<" to "<<v<<" guid "<<particles.guid(a)<<", "<<particles.guid(c)<<" to "<<particles.guid(v)<<std::endl;

    ////update hashtables
    Merge_Keys(vertex_tetrahedron_hashtable,ac,v);
    Merge_Keys(vertex_triangle_hashtable,ac,v);
    Merge_Keys(vertex_segment_hashtable,ac,v);
    
    ARRAY<int>* v_tets=vertex_tetrahedron_hashtable.Get_Pointer(v);
    if(v_tets){Replace_Vertex_In_Elements(tetrahedron_mesh.elements,*v_tets,ac,v);}
    ARRAY<int>* v_tris=vertex_triangle_hashtable.Get_Pointer(v);
    if(v_tris)Replace_Vertex_In_Elements(triangle_mesh.elements,*v_tris,ac,v);
    ARRAY<int>* v_segs=vertex_segment_hashtable.Get_Pointer(v);
    if(v_segs)Replace_Vertex_In_Elements(segment_mesh.elements,*v_segs,ac,v);

    if(!preserve_thickness_in_remeshing){
        particles.dimensional_mass(v)(1)=particles.dimensional_mass(a)(1)+particles.dimensional_mass(c)(1);
        particles.dimensional_mass(v)(TV::m-1)=particles.dimensional_mass(a)(TV::m-1)+particles.dimensional_mass(c)(TV::m-1);
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.V(v)=(particles.V(a)*particles.mass(a)+particles.V(c)*particles.mass(c))/particles.mass(v);    
    } 
    else{
        Simple_Redistribute_Mass_And_Momentum(a,c,v);

        T old_vol=vtx_vol.Get(a)+vtx_vol.Get(c);
        T A=Node_Triangle_Area_Sum(v,vertex_triangle_hashtable)+(T).5*Nodal_Tetrahedron_Surface_Area_Sum(v,vertex_tetrahedron_hashtable);
        T L=Node_Segment_Length_Sum(v,vertex_segment_hashtable);
        if(A!=0||L!=0){
            T t=Joint_Thickness(old_vol,A,L);
            //PHYSBAM_ASSERT(t>=(T)0);
            if(!(t>=(T)0))t=thickness_scale;
            particles.thickness(v)=t;
        }
        else particles.thickness(v)=(T)0;
    }

    ////TODO: fix on_boundary for incident tets of v
    if(particles.on_boundary(v)(TV::m)){}

    return true;
}
