//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

using namespace PhysBAM;
using namespace Codimensional;

template<class T_input>
int PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Particles(const int a,const int c,const TV_INT& dim,const TV& pos,const T weight,bool change_mass_and_v/*=true*/)
{
    int v=particles.Add_Particle();
    particles.X(v)=pos;
    if(change_mass_and_v){  ////TODO: fix for weight
        for(int d=1;d<=TV::m;d++){
            if(dim(d)!=0){
                particles.dimensional_mass(v)(d)=(T)one_third*(particles.dimensional_mass(a)(d)+particles.dimensional_mass(c)(d));
                particles.dimensional_mass(a)(d)=(T)two_thirds*particles.dimensional_mass(a)(d);
                particles.dimensional_mass(c)(d)=(T)two_thirds*particles.dimensional_mass(c)(d);
            }
        }
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.mass(a)=particles.dimensional_mass(a).Sum();
        particles.mass(c)=particles.dimensional_mass(c).Sum();
        particles.V(v)=(particles.mass(a)*particles.V(a)+particles.mass(c)*particles.V(c))/(particles.mass(a)+particles.mass(c));
    }

    for(int d=1;d<=3;d++){if(dim(d)!=0){particles.on_boundary(v)(d)=particles.on_boundary(a)(d)||particles.on_boundary(c)(d);}} ////TODO: or or and?
    particles.contact(v)=particles.contact(a)&particles.contact(c);
    particles.id(v)=particles.id(a);
    particles.thickness(v)=((T)1-weight)*particles.thickness(a)+weight*particles.thickness(c);
    particles.length_scale(v)=((T)1-weight)*particles.length_scale(a)+weight*particles.length_scale(c);
    if(particles.store_deformation_gradient)particles.deformation_gradient(v)=((T)1-weight)*particles.deformation_gradient(a)+weight*particles.deformation_gradient(c);
    if(particles.store_elastic_deformation_gradient)particles.elastic_deformation_gradient(v)=((T)1-weight)*particles.elastic_deformation_gradient(a)+weight*particles.elastic_deformation_gradient(c);
    if(particles.store_plastic_deformation_gradient)particles.plastic_deformation_gradient(v)=((T)1-weight)*particles.plastic_deformation_gradient(a)+weight*particles.plastic_deformation_gradient(c);
    if(particles.store_shear_rate)particles.shear_rate(v)=((T)1-weight)*particles.shear_rate(a)+weight*particles.shear_rate(c);
    if(particles.store_color)particles.color(v)=((T)1-weight)*particles.color(a)+weight*particles.color(c);
    if(particles.phis_number>0)for(int i=1;i<=particles.phis.m;i++)particles.phis(i)(v)=((T)1-weight)*particles.phis(i)(a)+weight*particles.phis(i)(c);
    if(particles.store_texture_coordinate)particles.texture_coordinate(v)=((T)1-weight)*particles.texture_coordinate(a)+weight*particles.texture_coordinate(c);
    if(particles.store_meshing_direction)particles.meshing_direction(v)=((T)1-weight)*particles.meshing_direction(a)+weight*particles.meshing_direction(c);
    if(particles.store_temperature)particles.temperature(v)=((T)1-weight)*particles.temperature(a)+weight*particles.temperature(c);
    if(use_particle_parents){Split_Particle_Parents(a,c,v,weight);}
    return v;
}

template<class T_input>
int PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Particles(const int a,const int b,const int c,const TV_INT& dim,const TV& pos,const TV& weights,bool change_mass_and_v/*=true*/)
{
    int v=particles.Add_Particle();
    particles.X(v)=pos;
    //if(change_mass_and_v){
    //    for(int d=1;d<=TV::m;d++){
    //        if(dim(d)!=0){
    //            particles.dimensional_mass(v)(d)=(T)one_third*(particles.dimensional_mass(a)(d)+particles.dimensional_mass(b)(d)+particles.dimensional_mass(c)(d));
    //            particles.dimensional_mass(a)(d)=(T)two_thirds*particles.dimensional_mass(a)(d);
    //            particles.dimensional_mass(b)(d)=(T)two_thirds*particles.dimensional_mass(b)(d);
    //            particles.dimensional_mass(c)(d)=(T)two_thirds*particles.dimensional_mass(c)(d);
    //        }
    //    }
    //    particles.mass(v)=particles.dimensional_mass(v).Sum();
    //    particles.mass(a)=particles.dimensional_mass(a).Sum();
    //    particles.mass(b)=particles.dimensional_mass(b).Sum();
    //    particles.mass(c)=particles.dimensional_mass(c).Sum();

    //    particles.V(v)=(particles.mass(a)*particles.V(a)+particles.mass(b)*particles.V(b)+particles.mass(c)*particles.V(c))/(particles.mass(a)+particles.mass(b)+particles.mass(c));
    //}

    for(int d=1;d<=TV::m;d++){if(dim(d)!=0){particles.on_boundary(v)(d)=particles.on_boundary(a)(d)||particles.on_boundary(b)(d)||particles.on_boundary(c)(d);}}

    particles.contact(v)=particles.contact(a)&particles.contact(b)&particles.contact(c);
    particles.id(v)=particles.id(a);
    particles.thickness(v)=weights(1)*particles.thickness(a)+weights(2)*particles.thickness(b)+weights(3)*particles.thickness(c);
    particles.length_scale(v)=weights(1)*particles.length_scale(a)+weights(2)*particles.length_scale(b)+weights(3)*particles.length_scale(c);
    if(particles.store_deformation_gradient)particles.deformation_gradient(v)=weights(1)*particles.deformation_gradient(a)+weights(2)*particles.deformation_gradient(b)+weights(3)*particles.deformation_gradient(c);
    if(particles.store_elastic_deformation_gradient)particles.elastic_deformation_gradient(v)=weights(1)*particles.elastic_deformation_gradient(a)+weights(2)*particles.elastic_deformation_gradient(b)+weights(3)*particles.elastic_deformation_gradient(c);
    if(particles.store_plastic_deformation_gradient)particles.plastic_deformation_gradient(v)=weights(1)*particles.plastic_deformation_gradient(a)+weights(2)*particles.plastic_deformation_gradient(b)+weights(3)*particles.plastic_deformation_gradient(c);
    if(particles.store_shear_rate)particles.shear_rate(v)=weights(1)*particles.shear_rate(a)+weights(2)*particles.shear_rate(b)+weights(3)*particles.shear_rate(c);
    if(particles.store_color)particles.color(v)=weights(1)*particles.color(a)+weights(2)*particles.color(b)+weights(3)*particles.color(c);    
    if(particles.phis_number>0)for(int i=1;i<=particles.phis.m;i++) particles.phis(i)(v)=weights(1)*particles.phis(i)(a)+weights(2)*particles.phis(i)(b)+weights(3)*particles.phis(i)(c);
    if(particles.store_texture_coordinate)particles.texture_coordinate(v)=weights(1)*particles.texture_coordinate(a)+weights(2)*particles.texture_coordinate(b)+weights(3)*particles.texture_coordinate(c);
    if(particles.store_meshing_direction)particles.meshing_direction(v)=weights(1)*particles.meshing_direction(a)+weights(2)*particles.meshing_direction(b)+weights(3)*particles.meshing_direction(c);
    if(particles.store_temperature)particles.temperature(v)=weights(1)*particles.temperature(a)+weights(2)*particles.temperature(b)+weights(3)*particles.temperature(c);
    if(use_particle_parents){Split_Particle_Parents(a,b,c,v,weights);}
    return v;
}

template<class T_input>
int PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Merge_Particles(const int a,const int c,const TV_INT& dim,const TV& pos,const T weight,bool change_mass_and_v/*=true*/)
{
    int v=particles.Add_Particle();
    particles.X(v)=pos;
    if(change_mass_and_v){  ////just update v, does not subtract from a and c
        for(int d=1;d<=3;d++){if(dim(d)!=0)particles.dimensional_mass(v)(d)=particles.dimensional_mass(a)(d)+particles.dimensional_mass(c)(d);}
        particles.mass(v)=particles.dimensional_mass(v).Sum();
        particles.V(v)=(particles.mass(a)*particles.V(a)+particles.mass(c)*particles.V(c))/particles.mass(v);
    }

    for(int d=1;d<=3;d++)if(dim(d)!=0)particles.on_boundary(v)(d)=(particles.on_boundary(a)(d)||particles.on_boundary(c)(d));   ////TODO: check this
    particles.contact(v)=particles.contact(a)|particles.contact(c);    
    particles.id(v)=particles.id(a);
    particles.thickness(v)=((T)1-weight)*particles.thickness(a)+weight*particles.thickness(c);
    particles.length_scale(v)=((T)1-weight)*particles.length_scale(a)+weight*particles.length_scale(c);
    if(particles.store_deformation_gradient)particles.deformation_gradient(v)=((T)1-weight)*particles.deformation_gradient(a)+weight*particles.deformation_gradient(c);
    if(particles.store_elastic_deformation_gradient)particles.elastic_deformation_gradient(v)=((T)1-weight)*particles.elastic_deformation_gradient(a)+weight*particles.elastic_deformation_gradient(c);
    if(particles.store_plastic_deformation_gradient)particles.plastic_deformation_gradient(v)=((T)1-weight)*particles.plastic_deformation_gradient(a)+weight*particles.plastic_deformation_gradient(c);
    if(particles.store_shear_rate)particles.shear_rate(v)=((T)1-weight)*particles.shear_rate(a)+weight*particles.shear_rate(c);
    if(particles.store_color)particles.color(v)=((T)1-weight)*particles.color(a)+weight*particles.color(c);
    if(particles.phis_number>0)for(int i=1;i<=particles.phis.m;i++)particles.phis(i)(v)=((T)1-weight)*particles.phis(i)(a)+weight*particles.phis(i)(c);
    if(particles.store_texture_coordinate)particles.texture_coordinate(v)=((T)1-weight)*particles.texture_coordinate(a)+weight*particles.texture_coordinate(c);
    if(particles.store_meshing_direction)particles.meshing_direction(v)=((T)1-weight)*particles.meshing_direction(a)+weight*particles.meshing_direction(c);
    if(particles.store_temperature)particles.temperature(v)=((T)1-weight)*particles.temperature(a)+weight*particles.temperature(c);
    if(use_particle_parents)Merge_Particle_Parents(a,c,v,weight);
    return v;
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Add_Particle_Parent_Helper(const int a,const int v,const T weight)
{
    if(particles.parents_number(a)>0){
        for(int i=1;i<=particles.parents_number(a);i++){
            int ap_guid=particles.parents_guid(i)(a);
            T ap_w=particles.parents_weight(i)(a);
            
            int p=0;for(int j=1;j<=particles.parents_number(v);j++){
                int vp_guid=particles.parents_guid(j)(v);
                if(vp_guid==ap_guid){p=j;break;}
            }

            if(p!=0){particles.parents_weight(p)(v)+=weight*ap_w;}
            else{
                int idx=particles.parents_number(v)+1;
                if(idx>particles.parents_size){ ////if the list is full, replace the smallest one
                    int smallest_idx=0;T smallest_w=weight*ap_w;
                    for(int j=1;j<=particles.parents_size;j++){
                        if(particles.parents_weight(j)(v)<smallest_w){
                            smallest_w=particles.parents_weight(j)(v);
                            smallest_idx=j;
                        }
                    }
                    if(smallest_idx!=0){
                        particles.parents_guid(smallest_idx)(v)=ap_guid;
                        particles.parents_weight(smallest_idx)(v)=weight*ap_w;
                    }
                }
                else{
                    particles.parents_guid(idx)(v)=ap_guid;
                    particles.parents_weight(idx)(v)=weight*ap_w;
                    particles.parents_number(v)=idx;
                }
            }
        }
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Delete_Particle_Parent_Helper(const int a)
{
    //particles.parents_hashtable.Delete(particles.guid(a));
    
    ////sync on particles
    particles.parents_number(a)=0;
    for(int i=1;i<=particles.parents_size;i++){particles.parents_guid(i)(a)=0;particles.parents_weight(i)(a)=(T)0;}
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Reset_Particle_Parents()
{
    for(int i=1;i<=particles.X.m;i++){
        particles.parents_number(i)=1;particles.parents_guid(1)(i)=particles.guid(i);particles.parents_weight(1)(i)=(T)1;
        for(int j=2;j<=particles.parents_size;j++){
            particles.parents_guid(j)(i)=0;particles.parents_weight(j)(i)=(T)0;
        }
    }
    //Update_Particle_Parents_Hashtable();
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Update_Particle_Parents_Hashtable()
{
    HASHTABLE<int,ARRAY<PAIR<int,T> > >& parents_hashtable=particles.parents_hashtable;parents_hashtable.Remove_All();
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);ARRAY<PAIR<int,T> > parents;
        for(int j=1;j<=particles.parents_number(i);j++){
            parents.Append(PAIR<int,T>(particles.parents_guid(j)(i),particles.parents_weight(j)(i)));
        }
        parents_hashtable.Set(guid,parents);
    }
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Particle_Parents(const int a,const int c,const int v,const T weight)
{
    Add_Particle_Parent_Helper(a,v,(T)1-weight);
    Add_Particle_Parent_Helper(c,v,weight);
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Split_Particle_Parents(const int a,const int b,const int c,const int v,const TV& weights)
{
    Add_Particle_Parent_Helper(a,v,weights(1));
    Add_Particle_Parent_Helper(b,v,weights(2));
    Add_Particle_Parent_Helper(c,v,weights(3));
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Merge_Particle_Parents(const int a,const int c,const int v,const T weight)
{
    Add_Particle_Parent_Helper(a,v,(T)1-weight);
    Add_Particle_Parent_Helper(c,v,weight);
    Delete_Particle_Parent_Helper(a);
    Delete_Particle_Parent_Helper(c);
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Convex_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& vtx_ring,const int a,const int c)
{
    TV_INT2 ac(a,c);
    for(int i=1;i<=vtx_ring.m;i++){
        int b=vtx_ring(i);int d=vtx_ring((i!=vtx_ring.m)?i+1:1);TV_INT2 bd(b,d);
        for(int j=1;j<=2;j++){
            int p=ac(j);int q=Other_Node_In_Element(ac,p);
            int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(b,d,p,q));
            for(int k=1;k<=vtx_ring.m;k++){
                int q2=vtx_ring(k);if(q2==b||q2==d)continue;
                int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(b,d,p,q2));
                if(sign*sign2<=0){return false;}
            }
        }
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Convex_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,/*result*/ARRAY<TV_INT2>& coplanar_edges)
{
    PHYSBAM_ASSERT(tets_v.m==2);
    TV_INT2 ac;for(int i=1;i<=2;i++){ac(i)=Other_Node_In_Element(tets_v(i),face(1),face(2),face(3));}
    for(int i=1;i<=2;i++){
        for(int j=1;j<=3;j++){
            int f1=face(j);int f2=face(j+1>3?1:j+1);
            int p=ac(i);int q=Other_Node_In_Element(ac,p);int f3=Other_Node_In_Element(face,f1,f2);
            int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(p,f1,f2,f3));
            int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(p,f1,f2,q));
            if(sign2==0){coplanar_edges.Append(TV_INT2(f1,f2));}
            if(sign*sign2<0){return false;}
        }
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Is_Convex_Pair_Of_Boundary_Tetrahedra(const TV_INT2& boundary_edge,const int& b,const int& p,const int& q)
{
    int a=boundary_edge(1);int c=boundary_edge(2);
    {   ////face abp
        int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(a,b,p,c));
        int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(a,b,p,q));
        if(sign*sign2<0)return false;
    }
    {   ////face cbp
        int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(c,b,p,a));
        int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(c,b,p,q));
        if(sign*sign2<0)return false;
    }
    {   ////face abq
        int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(a,b,q,p));
        int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(a,b,q,c));
        if(sign*sign2<0)return false;
    }
    {   ////face cbq
        int sign=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(c,b,q,p));
        int sign2=Accurate_Tetrahedron_Orientation(particles.X,TV_INT4(c,b,q,a));
        if(sign*sign2<0)return false;
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Get_One_Ring_Vertices_Of_Tetrahedra(const ARRAY<VECTOR<int,4> >& tets_v,const int a,const int c,/*result*/ARRAY<int>& vtx_ring)
{
    ARRAY<TV_INT2> edges;
    for(int i=1;i<=tets_v.m;i++){
        TV_INT2 edge=Other_Nodes_In_Element(tets_v(i),a,c);
        edges.Append(edge);
    }
    ARRAY<bool> flag;flag.Resize(edges.m);flag.Fill(true);
    vtx_ring.Remove_All();vtx_ring.Append(edges(1)(1));vtx_ring.Append(edges(1)(2));flag(1)=false;
    int head=vtx_ring(1);
    int next=vtx_ring(vtx_ring.m);
    while(true){
        bool find=false;
        for(int i=1;i<=edges.m;i++){
            if(!flag(i))continue;
            if(edges(i).Contains(next)){
                int next_next=edges(i)(1)==next?edges(i)(2):edges(i)(1);
                if(next_next==head){
                    return true;
                }
                vtx_ring.Append(next_next);
                next=vtx_ring(vtx_ring.m);
                flag(i)=false;find=true;
            }
        }
        if(!find){return false;}
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Get_Flipped_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& v_ring,const int a,const int c,/*result*/ARRAY<TV_INT4>& flipped_v_tets)
{
    TV_INT2 e(a,c);
    for(int i=2;i<=v_ring.m-1;i++){
        for(int j=1;j<=2;j++){
            TV_INT4 v=TV_INT4(v_ring(1),v_ring(i),v_ring(i+1),e(j));
            if(Is_Negative_Volume_Tetrahedron(particles.X,v)){int tmp=v(2);v(2)=v(3);v(3)=tmp;}
            flipped_v_tets.Append(v);
        }
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Get_23_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT3>& boundary_faces)
{
    int a=Other_Node_In_Element(tets_v(1),face(1),face(2),face(3));int c=Other_Node_In_Element(tets_v(2),face(1),face(2),face(3));
    for(int i=1;i<=3;i++){
        int b=face(i);int d=face(i+1>3?1:(i+1));
        TV_INT4 v(a,c,b,d);
        int sign=Accurate_Tetrahedron_Orientation(particles.X,v);
        if(sign==0){LOG::cout<<"error in Get_Flipped_23_Pair_Of_Tetrahedra"<<std::endl;return false;}  ////should not have 0-vol tet if it is 2-3 flip
        if(sign<0){int tmp=v(2);v(2)=v(3);v(3)=tmp;}
        flipped_v_tets.Append(v);
        boundary_faces.Append(TV_INT3(b,d,a));boundary_faces.Append(TV_INT3(b,d,c));
    }
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Get_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT2& boundary_edge,/*result*/int& b,/*result*/int& p,/*result*/int& q,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT2>& affected_edges)
{
    int a=boundary_edge(1);int c=boundary_edge(2);
    const TV_INT4& v1=tets_v(1);const TV_INT4& v2=tets_v(2);
    TV_INT2 b_and_p=Other_Nodes_In_Element(v1,a,c);TV_INT2 b_and_q=Other_Nodes_In_Element(v2,a,c);
    b=Shared_Vertex(b_and_p,b_and_q);if(b==-1)return false;
    p=Other_Node_In_Element(b_and_p,b);q=Other_Node_In_Element(b_and_q,b);
    TV_INT4 new_v1(b,p,q,a);TV_INT4 new_v2(b,p,q,c);
    int sign_1=Accurate_Tetrahedron_Orientation(particles.X,new_v1);int sign_2=Accurate_Tetrahedron_Orientation(particles.X,new_v2);
    if(sign_1*sign_2>=0)return false;
    if(sign_1<0){int tmp=new_v1(2);new_v1(2)=new_v1(3);new_v1(3)=tmp;}
    if(sign_2<0){int tmp=new_v2(2);new_v2(2)=new_v2(3);new_v2(3)=tmp;}
    flipped_v_tets.Append(new_v1);flipped_v_tets.Append(new_v2);
    //boundary_faces.Append(TV_INT3(p,q,a));boundary_faces.Append(TV_INT3(p,q,c));
    //boundary_faces.Append(TV_INT3(c,b,p));boundary_faces.Append(TV_INT3(c,b,q));
    //boundary_faces.Append(TV_INT3(a,b,p));boundary_faces.Append(TV_INT3(a,b,q));
    affected_edges.Append(TV_INT2(p,a));affected_edges.Append(TV_INT2(p,b));affected_edges.Append(TV_INT2(p,c));
    affected_edges.Append(TV_INT2(q,a));affected_edges.Append(TV_INT2(q,b));affected_edges.Append(TV_INT2(q,c));
    /*affected_edges.Append(TV_INT2(p,q));*/affected_edges.Append(TV_INT2(a,b));affected_edges.Append(TV_INT2(c,b));
    return true;
}

template<class T_input>
bool PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Get_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,const TV_INT2& boundary_edge,/*result*/ARRAY<TV_INT4>& flipped_v_tets,/*result*/ARRAY<TV_INT3>& boundary_faces)
{
    int a=Other_Node_In_Element(tets_v(1),face(1),face(2),face(3));int c=Other_Node_In_Element(tets_v(2),face(1),face(2),face(3));
    for(int i=1;i<=3;i++){
        int b=face(i);int d=face(i+1>3?1:(i+1));
        TV_INT2 edge(b,d);
        TV_INT4 v=TV_INT4(a,c,edge(1),edge(2));
        if(edge.Sorted()==boundary_edge.Sorted()){
            boundary_faces.Append(TV_INT3(a,c,b));boundary_faces.Append(TV_INT3(a,c,d));continue;
        }
        int sign=Accurate_Tetrahedron_Orientation(particles.X,v);
        if(sign==0){LOG::cout<<"error in Get_Flipped_22_Pair_Of_Tetrahedra"<<std::endl;return false;}  ////should not have 0-vol tet if it is not the boundary tet in 2-2 flip
        if(sign<0){int tmp=v(2);v(2)=v(3);v(3)=tmp;}
        flipped_v_tets.Append(v);
        boundary_faces.Append(TV_INT3(b,d,a));boundary_faces.Append(TV_INT3(b,d,c));
    }
    return true;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Measure_Tetrahedra(const ARRAY<TV_INT4>& tets_v)
{
    T measure=(T)FLT_MAX;
    for(int i=1;i<=tets_v.m;i++){
        measure=min(measure,Volume_Length_Measure(particles.X(tets_v(i)(1)),particles.X(tets_v(i)(2)),particles.X(tets_v(i)(3)),particles.X(tets_v(i)(4))));
    }
    return measure;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Measure_Triangles(const ARRAY<TV_INT3>& tris_v)
{
    T measure=(T)FLT_MAX;
    for(int i=1;i<=tris_v.m;i++){
        measure=min(measure,Area_Length_Measure(particles.X(tris_v(i)(1)),particles.X(tris_v(i)(2)),particles.X(tris_v(i)(3))));
    }
    return measure;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Measure_Flipped_Ring_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const ARRAY<int>& v_ring,const int a,const int c)
{
    T measure=(T)FLT_MAX;TV_INT2 e(a,c);
    for(int i=2;i<=v_ring.m-1;i++){
        for(int j=1;j<=2;j++){
            TV_INT4 v=TV_INT4(v_ring(1),v_ring(i),v_ring(i+1),e(j));
            measure=min(measure,Volume_Length_Measure(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)),particles.X(v(4))));
        }
    }
    return measure;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Measure_23_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face)
{
    PHYSBAM_ASSERT(tets_v.m==2);
    T measure=(T)FLT_MAX;int a=Other_Node_In_Element(tets_v(1),face(1),face(2),face(3));int c=Other_Node_In_Element(tets_v(2),face(1),face(2),face(3));
    for(int i=1;i<=3;i++){
        T tet_measure=Volume_Length_Measure(particles.X(a),particles.X(c),particles.X(face(i)),particles.X(face(i+1>3?1:(i+1))));
        measure=min(measure,tet_measure);
    }
    return measure;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESHING_3D<T_input>::Measure_22_Flipped_Pair_Of_Tetrahedra(const ARRAY<TV_INT4>& tets_v,const TV_INT3& face,const TV_INT2& boundary_edge)
{
    PHYSBAM_ASSERT(tets_v.m==2);
    T measure=(T)FLT_MAX;int a=Other_Node_In_Element(tets_v(1),face(1),face(2),face(3));int c=Other_Node_In_Element(tets_v(2),face(1),face(2),face(3));
    for(int i=1;i<=3;i++){
        TV_INT2 edge=TV_INT2(face(i),face(i+1>3?1:(i+1)));if(edge.Sorted()==boundary_edge.Sorted())continue;
        T tet_measure=Volume_Length_Measure(particles.X(a),particles.X(c),particles.X(edge(1)),particles.X(edge(2)));
        measure=min(measure,tet_measure);
    }
    return measure;
}
