//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Meshing/CODIMENSIONAL_MESHING_DRIVER.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_MESHING_DRIVER<TV>::CODIMENSIONAL_MESHING_DRIVER():BASE()
{
    has_viscosity=false;has_gravity=false;has_surface_tension=false;has_elasticity=false;has_plasticity=false;has_fiber_fluid_coupling_force=false;has_incompressibility=false;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_DRIVER<TV>::Execute_Main_Program()
{
    Test_Skinny_Tetrahedron();
    //BASE::Execute_Main_Program();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_DRIVER<TV>::Initialize_Test_Parameters()
{
    switch(test){
    case 1:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)3);box_edge_length.x*=(T)5;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        }break;
    case 2:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.1),TV::Constant_Vector((T).1));
            length_scale=(T)1e-3;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)10);
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            frame_rate=10000;
        }break;
    case 3:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.1),TV::Constant_Vector((T).1));
            length_scale=(T)1e-3;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)3);box_edge_length.x*=(T)5;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            frame_rate=10000;
        }break;
    case 4:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)3);
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            T height=(T).7*box_edge_length.y;box_min.y+=height;box_max.y+=height;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=true;
            has_surface_tension=true;
            kinematic_viscosity*=(T)1000;

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 5:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;int n=5;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)(n-1));
            TV box_min=-(T).5*box_edge_length;box_min.z=(T)0;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,n,n,box_min,length_scale);
        }break;
    case 6:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;int n=5;T filament_length=length_scale*(T)(n-1);
            TV box_min=-TV::Axis_Vector(1)*(T).5*filament_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,box_min,box_min+TV::Axis_Vector(1)*filament_length,length_scale);
        }break;
    case 7:////MPI region test
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Mpi_Region_Test_Shape(initial_mesh,initial_X,length_scale);
        }break;
    case 8:////MPI region test
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            //length_scale=(T)1e-2;
            length_scale=(T)2e-2;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV(),(T)10*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)-20*length_scale,(T)0,(T)-20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)-20*length_scale,(T)0,(T)20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)20*length_scale,(T)0,(T)20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)20*length_scale,(T)0,(T)-20*length_scale),(T)5*length_scale,length_scale);
        }break;
    case 9:////MPI region test
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            //length_scale=(T)1e-2;
            length_scale=(T)2e-2;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)-20*length_scale,(T)0,(T)-20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)-20*length_scale,(T)0,(T)20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)20*length_scale,(T)0,(T)20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV((T)20*length_scale,(T)0,(T)-20*length_scale),(T)5*length_scale,length_scale);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV(),(T)10*length_scale,length_scale);
        }break;
    case 10:////MPI region test
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            //length_scale=(T)1e-2;
            length_scale=(T)2e-2;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV(),(T)10*length_scale,length_scale);
        }break;
    case 11:////cylinder with rounded off endpoint
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),(T)7*length_scale*TV::Constant_Vector((T)1));
            T radius=(T)5*length_scale;
            T endpoint_frac=(T).8;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder_With_Rounded_End(initial_mesh,initial_X,TV(),TV::Axis_Vector(2),radius,length_scale,endpoint_frac);
        }break;
    case 12:////volume from tri file
        {
            //global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),(T)7*length_scale*TV::Constant_Vector((T)1));
            //std::string filename="../model/Rigid_Bodies/tube_body_double.tri.gz";
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Volume_From_Tri_File(initial_mesh,initial_X,10*length_scale,filename);
            length_scale=(T)0.25*(T)1e-3;
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),(T)7*length_scale*TV::Constant_Vector((T)1));
            std::string filename="../model/Rigid_Bodies/blob.tri.gz";
            CODIMENSIONAL_MESH_GENERATOR<TV>::Volume_From_Tri_File(initial_mesh,initial_X,length_scale,filename);

        }break;
    case 13:////another volume from tri file
        {
            //global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),(T)7*length_scale*TV::Constant_Vector((T)1));
            //std::string filename="../model/Rigid_Bodies/tube_body_double.tri.gz";
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Volume_From_Tri_File(initial_mesh,initial_X,10*length_scale,filename);
            //length_scale=(T)0.25*(T)1e-3;
            length_scale=(T)1e-3;
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),(T)7*length_scale*TV::Constant_Vector((T)1));
            std::string filename="../model/Rigid_Bodies/sauce_dbl.tri.gz";
            CODIMENSIONAL_MESH_GENERATOR<TV>::Volume_From_Tri_File(initial_mesh,initial_X,length_scale,filename);
        }break;
    case 14:////pizza box test
        {
            length_scale=(T)1e-3;
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            TV min;
            TV max=TV::Constant_Vector((T)50*length_scale);
            max(2)=(T)1*length_scale;

            T min_thickness=length_scale;
            T max_thickness=(T)3*length_scale;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume(initial_mesh,initial_X,min,max,length_scale,min_thickness,max_thickness);
        }break;
    case 15:////pizza cylinder test
        {
            length_scale=(T)1e-3;
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            T radius=25*length_scale;
            TV center;

            T min_thickness=length_scale;
            T max_thickness=(T)3*length_scale;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume(initial_mesh,initial_X,center,radius,length_scale,min_thickness,max_thickness,(T)400,TV::Constant_Vector((T)10));
        }break;
    default:
        {
            PHYSBAM_FATAL_ERROR("invalid test");
        }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_DRIVER<TV>::Initialize_Solver()
{
    BASE::Initialize_Solver();

    ////velocity
    switch(test){
    case 1:case 5:case 6:{
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.V(i)=(example.particles.X(i).x>(T)0?(T)1:(T)-1)*TV::Axis_Vector(1);
        }
    }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESHING_DRIVER<TV>::Test_Skinny_Tetrahedron()
{
    TV v1((T)0,(T)1,(T)0);TV v2((T)0,(T)-1,(T)0);TV v3((T)1,(T)0,(T)0);
    ARRAY<TV> v4s;
    v4s.Append(TV((T)-1,(T)0,(T).1));
    v4s.Append(TV((T)-.1,(T)0,(T).1));
    v4s.Append(TV((T).1,(T)0,(T).1));
    v4s.Append(TV((T).5,(T)0,(T).1));

    LOG::cout<<"test basic distance functions..."<<std::endl;
    for(int i=1;i<=v4s.m;i++){
        TETRAHEDRON<T> tet(v1,v2,v3,v4s(i));
        LOG::cout<<"tet "<<i<<std::endl;
        LOG::cout<<"X: "<<tet.X<<std::endl;
        int vertex_edge_p=0;TV_INT2 vertex_edge_e=TV_INT2();T min_t;
        T min_vertex_edge_distance=Nearest_Vertex_Edge_Pair_On_Tetrahedron(tet,vertex_edge_p,vertex_edge_e,min_t);
        LOG::cout<<"nearest vertex edge distance: "<<min_vertex_edge_distance<<", v: "<<vertex_edge_p<<", e: "<<vertex_edge_e<<", t: "<<min_t<<std::endl;
        TV_INT2 e1;TV_INT2 e2;T t1;T t2;
        T min_edge_edge_distance=Nearest_Edge_Edge_Pair_On_Tetrahedron(tet,e1,e2,t1,t2);
        LOG::cout<<"nearest edge edge distance: "<<min_edge_edge_distance<<", e1: "<<e1<<", e2: "<<e2<<", t1: "<<t1<<", t2: "<<t2<<std::endl;
    }

    LOG::cout<<"test Shortest_Vector_Between_Segments..."<<std::endl;
    SEGMENT_3D<T> s1(v1,v2);SEGMENT_3D<T> s2(v3,TV((T).1,(T)0,(T).1));TV2 weights;
    TV shortest_vector=s1.Shortest_Vector_Between_Segments(s2,weights);
    LOG::cout<<"shortest vector: "<<shortest_vector<<", w: "<<weights(1)<<", "<<weights(2)<<std::endl;

    LOG::cout<<"test Split_Skinny_Tetrahedron..."<<std::endl;
    TV start((T)-1,(T)0,(T)0);TV end((T).5,(T)0,(T)0);
    int steps=100;
    T distance_threshold=(T).1;T t_threshold=(T).05;
    for(int i=0;i<=steps;i++){
        T frac=(T)i/(T)steps;
        TV v4=((T)1-frac)*start+frac*end;
        TETRAHEDRON<T> tet(v1,v2,v3,v4);
        ARRAY<TV_INT2> edges;ARRAY<T> weights;TV_INT3 face;TV3 face_uvw;
        int type=Split_Skinny_Tetrahedron(tet,distance_threshold,t_threshold,edges,weights,face,face_uvw);
        LOG::cout<<"v :"<<v4<<", type: "<<type<<", edges: "<<edges<<", weights: "<<weights<<", face: "<<face<<", uvw: "<<face_uvw<<std::endl;
    }

    LOG::cout<<"test projected uvw..."<<std::endl;
}

//template class CODIMENSIONAL_MESHING_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_MESHING_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_MESHING_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_MESHING_DRIVER<VECTOR<double,3> >;
#endif
