//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Geometry/Geometry_Particles/GEOMETRY_PARTICLES_FORWARD.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry_Computations/DELAUNAY_MESHING_WITH_ALPHA_SHAPE.h>
#include <PhysBAM_Tools/Math_Tools/constants.h>
#include <PhysBAM_Tools/Random_Numbers/NOISE.h>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"

using namespace PhysBAM;

//template<class TV>
//void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int segment_num,const TV& min_corner,const T length_scale)
//{
//    ////particle positions
//    for(int i=0;i<=segment_num;i++){TV pos=min_corner+(T)i*length_scale*TV::Axis_Vector(2);X.Append(pos);}
//
//    ////mesh topology
//    mesh.segment_mesh.elements.Remove_All();
//    mesh.triangle_mesh.elements.Remove_All();
//    mesh.tetrahedron_mesh.elements.Remove_All();
//    for(int i=1;i<=segment_num;i++){
//        mesh.segment_mesh.elements.Append(TV_INT2(i,i+1));
//    }
//    mesh.Update_States(X.m);
//}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T length_scale)
{
    ////particle positions
    int c=X.m;T filament_length=(e2-e1).Magnitude();TV dir=(e2-e1).Normalized();int segment_num=(int)(filament_length/length_scale+(T).5);T segment_length=filament_length/(T)segment_num;
    for(int i=0;i<=segment_num;i++){TV pos=e1+(T)i*segment_length*dir;X.Append(pos);}

    ////mesh topology
    for(int i=1;i<=segment_num;i++){mesh.segment_mesh.elements.Append(TV_INT2(c+i,c+i+1));}
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Bended_Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV e1,const T filament_length,const T length_scale,BSPLINE<T,TV>& dir_bspline)
{
    ////particle positions
    int c=X.m;int segment_num=(int)(filament_length/length_scale+(T).5);
    TV start=e1;
    for(int i=0;i<segment_num;i++){
        X.Append(start);
        T frac=(T)i/(T)segment_num;
        TV dir=dir_bspline.Evaluate(frac);
        TV pos=start+length_scale*dir;
        start=pos;
    }
    X.Append(start);

    ////mesh topology
    for(int i=1;i<=segment_num;i++){mesh.segment_mesh.elements.Append(TV_INT2(c+i,c+i+1));}
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Branch(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int branch_n,const int branch_seg_n,const TV& center,const T length_scale)
{
    ////particle positions
    X.Append(center);
    T angle_per_branch=(T)two_pi/(T)branch_n;

    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();

    for(int a=1;a<=branch_n;a++){
        T angle=(T)(a-1)*angle_per_branch;
        TV dir;dir.x=cos(angle);dir.y=sin(angle);
        mesh.segment_mesh.elements.Append(TV_INT2(1,X.m+1));
        for(int i=1;i<=branch_seg_n;i++){
            TV pos=center+dir*(T)i*length_scale;X.Append(pos);
            if(i>=2){mesh.segment_mesh.elements.Append(TV_INT2(X.m-1,X.m));}
        }
    }

    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int m,const int n,const TV& min_corner,const T length_scale,const int axis/*=3*/)
{
    int axis1=0;int axis2=0;if(axis==1){axis1=2;axis2=3;}else if(axis==2){axis1=1;axis2=3;}else{axis1=1;axis2=2;}int c=X.m;
    ////particle positions
    for(int i=0;i<n;i++){
        for(int j=0;j<m;j++){
            TV pos=min_corner+TV::Axis_Vector(axis1)*(T)i*length_scale+TV::Axis_Vector(axis2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    ////mesh topology  
    TRIANGLE_MESH triangle_mesh;triangle_mesh.Initialize_Herring_Bone_Mesh(m,n);
    for(int i=1;i<=triangle_mesh.elements.m;i++){const TV_INT3& v=triangle_mesh.elements(i);mesh.triangle_mesh.elements.Append(TV_INT3(c+v(1),c+v(2),c+v(3)));}
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int m,const int n,const TV& min_corner,const T length_scale,const int axis/*=3*/)
{
    int axis1=0;int axis2=0;if(axis==1){axis1=2;axis2=3;}else if(axis==2){axis1=1;axis2=3;}else{axis1=1;axis2=2;}int c=X.m;
    ////particle positions
    for(int i=0;i<n;i++){
        for(int j=0;j<m;j++){
            TV pos=min_corner+TV::Axis_Vector(axis1)*(T)i*length_scale+TV::Axis_Vector(axis2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    ////mesh topology
    TRIANGLE_MESH triangle_mesh;triangle_mesh.Initialize_Herring_Bone_Mesh(m,n);   
    Randomize_Herring_Bone_Triangle_Mesh(triangle_mesh,m,n);
    for(int i=1;i<=triangle_mesh.elements.m;i++){const TV_INT3& v=triangle_mesh.elements(i);mesh.triangle_mesh.elements.Append(TV_INT3(c+v(1),c+v(2),c+v(3)));}
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,const TV& min_corner,const VECTOR<T,2>& size,const T length_scale,const int axis/*=3*/)
{
    int n=(int)(size.y/length_scale)+1;int m=(int)(size.x/length_scale)+1;
    Herring_Bone_Rectangle(mesh,X,m,n,min_corner,length_scale,axis);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min_corner,const VECTOR<T,2>& size,T length_scale,const int axis/*=3*/)
{
    int n=(int)(size.y/length_scale)+1;int m=(int)(size.x/length_scale)+1;
    Randomized_Herring_Bone_Rectangle(mesh,X,m,n,min_corner,length_scale,axis);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Centrosymmetric_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int m,const int n,const TV& min_corner,const TV& center,const T length_scale,const bool reverse/*=false*/,const int axis/*=3*/)
{
    int axis1=0;int axis2=0;if(axis==1){axis1=2;axis2=3;}else if(axis==2){axis1=1;axis2=3;}else{axis1=1;axis2=2;}
    ////particle positions
    for(int i=0;i<n;i++){
        for(int j=0;j<m;j++){
            TV pos=min_corner+TV::Axis_Vector(axis1)*(T)i*length_scale+TV::Axis_Vector(axis2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    ////mesh topology
    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();
    mesh.triangle_mesh.Initialize_Herring_Bone_Mesh(m,n);   
    Centrosymmetrize_Herring_Bone_Triangle_Mesh(mesh.triangle_mesh,m,n,X,center,reverse);

    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Square_Dumbbell(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int square_n,const int stick_n,const TV& min_corner,const T length_scale)
{
    TV start=min_corner;
    int start_i1=1;
    for(int i=0;i<square_n;i++){
        for(int j=0;j<square_n;j++){
            TV pos=start+TV::Axis_Vector(1)*(T)i*length_scale+TV::Axis_Vector(2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    start=min_corner+TV::Axis_Vector(1)*(T)(square_n-1)*length_scale+TV::Axis_Vector(2)*(T)(square_n-1)*length_scale;
    int start_i2=X.m;
    TV dir=(TV::Axis_Vector(1)+TV::Axis_Vector(2)).Normalized();
    for(int i=1;i<=stick_n;i++){
        TV pos=start+dir*length_scale*(T)i;
        X.Append(pos);
    }

    start=min_corner+TV::Axis_Vector(1)*((square_n-1)*length_scale+(stick_n+1)*length_scale*(T)(.5*sqrt(2.)))+
        TV::Axis_Vector(2)*((square_n-1)*length_scale+(stick_n+1)*length_scale*(T)(.5*sqrt(2.)));
    int start_i3=X.m+1;
    for(int i=0;i<square_n;i++){
        for(int j=0;j<square_n;j++){
            TV pos=start+TV::Axis_Vector(1)*(T)i*length_scale+TV::Axis_Vector(2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();

    TRIANGLE_MESH triangle_mesh_copy;triangle_mesh_copy.Initialize_Herring_Bone_Mesh(square_n,square_n);
    Randomize_Herring_Bone_Triangle_Mesh(triangle_mesh_copy,square_n,square_n);

    for(int i=1;i<=triangle_mesh_copy.elements.m;i++){
        TV_INT3& e=triangle_mesh_copy.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(start_i1-1+e(1),start_i1-1+e(2),start_i1-1+e(3)));
    }
    for(int i=1;i<=stick_n+1;i++){
        mesh.segment_mesh.elements.Append(TV_INT2(start_i2-1+i,start_i2-1+i+1));
    }
    for(int i=1;i<=triangle_mesh_copy.elements.m;i++){
        TV_INT3& e=triangle_mesh_copy.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(start_i3-1+e(1),start_i3-1+e(2),start_i3-1+e(3)));
    }
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Randomize_Herring_Bone_Triangle_Mesh(/*result*/TRIANGLE_MESH& triangle_mesh,const int m,const int n)
{
    RANDOM_NUMBERS<T> random(1234567);

    int nx=m-1;int ny=n-1;
    for(int i=1;i<=nx;i++){
        for(int j=1;j<=ny;j++){
            bool flip=random.Get_Uniform_Number((T)0,(T)1)>(T).5;
            if(!flip)continue;

            int t1=(j-1)*2*nx+(i-1)*2+1;
            int t2=t1+1;
            VECTOR<int,3> v1=triangle_mesh.elements(t1); VECTOR<bool,3> flag1=VECTOR<bool,3>(true,true,true);
            VECTOR<int,3> v2=triangle_mesh.elements(t2); VECTOR<bool,3> flag2=VECTOR<bool,3>(true,true,true);
            for(int d=1;d<=3;d++){
                for(int k=1;k<=3;k++){
                    if(v1(d)==v2(k)){flag1(d)=false;flag2(k)=false;break;}
                }
            }
            int i1=flag1(1)?1:(flag1(2)?2:3);int next_i1=(i1==3?1:i1+1);int next_next_i1=(next_i1==3?1:next_i1+1);
            int i2=flag2(1)?1:(flag2(2)?2:3);int next_i2=(i2==3?1:i2+1);int next_next_i2=(next_i2==3?1:next_i2+1);

            triangle_mesh.elements(t1)(next_next_i1)=v2(i2);
            triangle_mesh.elements(t2)(next_next_i2)=v1(i1);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Centrosymmetrize_Herring_Bone_Triangle_Mesh(/*result*/TRIANGLE_MESH& triangle_mesh,const int m,const int n,
    const ARRAY<TV>& X,const TV& center,const bool reverse)
{
    int nx=m-1;int ny=n-1;
    for(int i=1;i<=nx;i++){
        for(int j=1;j<=ny;j++){
            int t1=(j-1)*2*nx+(i-1)*2+1;
            int t2=t1+1;
            VECTOR<int,3> v1=triangle_mesh.elements(t1); VECTOR<bool,3> flag1=VECTOR<bool,3>(true,true,true);
            VECTOR<int,3> v2=triangle_mesh.elements(t2); VECTOR<bool,3> flag2=VECTOR<bool,3>(true,true,true);
            for(int d=1;d<=3;d++){
                for(int k=1;k<=3;k++){
                    if(v1(d)==v2(k)){flag1(d)=false;flag2(k)=false;break;}
                }
            }
            int i1=flag1(1)?1:(flag1(2)?2:3);int next_i1=(i1==3?1:i1+1);int next_next_i1=(next_i1==3?1:next_i1+1);
            int i2=flag2(1)?1:(flag2(2)?2:3);int next_i2=(i2==3?1:i2+1);int next_next_i2=(next_i2==3?1:next_i2+1);

            TV dir=X(v1(i1))-X(v2(i2));TV c=(T).5*(X(v1(i1))+X(v2(i2)));

            if((!reverse&&!((c.x*c.y>0&&dir.x*dir.y>0)||(c.x*c.y<0&&dir.x*dir.y<0)))||
                (reverse&&!((c.x*c.y>0&&dir.x*dir.y<0)||(c.x*c.y<0&&dir.x*dir.y>0)))){
                ////flip
                triangle_mesh.elements(t1)(next_next_i1)=v2(i2);
                triangle_mesh.elements(t2)(next_next_i2)=v1(i1);
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Pascal_Triangle(/*result*/TRIANGLE_MESH& triangle_mesh,/*result*/ARRAY<TV>& X,const int tri_n,const T x_dir,const TV& corner,const T length_scale)
{
    triangle_mesh.elements.Remove_All();

    int p0=0,p1=0,q0=0,q1=0;
    for(int i=1;i<=tri_n+1;i++){
        ////add particles
        TV start;start.x=x_dir*(T)(sqrt(3.)*.5)*(T)(i-1)*length_scale;start.y=(T)(-i+1)*(T).5*length_scale;
        TV dir=TV::Axis_Vector(2);
        q0=p0;q1=p1;p0=X.m+1;
        for(int j=1;j<=i;j++){
            TV pos=corner+start+dir*(T)(j-1)*length_scale;X.Append(pos);
        }
        p1=X.m;

        ////build triangle stripe
        for(int j=1;j<i;j++){
            triangle_mesh.elements.Append(TV_INT3(p0+(j-1),q0+(j-1),p0+j));
            if(j<i-1)triangle_mesh.elements.Append(TV_INT3(p0+j,q0+(j-1),q0+j));
        }
    }

    triangle_mesh.number_nodes=X.m;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Sandglass(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int tri_seg_n,const int stick_seg_n,const TV& corner,const T length_scale)
{
    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();

    ARRAY<TV> X_copy1;TRIANGLE_MESH triangle_mesh_copy1;Pascal_Triangle(triangle_mesh_copy1,X_copy1,tri_seg_n,(T)-1,corner,length_scale);
    for(int i=1;i<=X_copy1.m;i++){X.Append(X_copy1(i));}
    for(int i=1;i<=triangle_mesh_copy1.elements.m;i++){
        const TV_INT3& e=triangle_mesh_copy1.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(e(1),e(2),e(3)));
    }

    mesh.segment_mesh.elements.Append(TV_INT2(1,X.m+1));
    TV dir=TV::Axis_Vector(1);
    for(int i=1;i<=stick_seg_n-1;i++){
        TV pos=corner+(T)i*length_scale*dir;X.Append(pos);
        mesh.segment_mesh.elements.Append(TV_INT2(X.m,X.m+1));
    }

    TV start=corner+(T)stick_seg_n*length_scale*dir;int p=X.m;
    ARRAY<TV> X_copy2;TRIANGLE_MESH triangle_mesh_copy2;Pascal_Triangle(triangle_mesh_copy2,X_copy2,tri_seg_n,(T)1,start,length_scale);
    for(int i=1;i<=X_copy2.m;i++){X.Append(X_copy2(i));}
    for(int i=1;i<=triangle_mesh_copy2.elements.m;i++){
        const TV_INT3& e=triangle_mesh_copy2.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(p+e(1),p+e(2),p+e(3)));
    }

    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Square_Sticks(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int square_n,const int stick_n,const TV& min_corner,const T length_scale)
{
    TV start=min_corner;
    int start_i1=1;
    for(int i=0;i<square_n;i++){
        for(int j=0;j<square_n;j++){
            TV pos=start+TV::Axis_Vector(1)*(T)i*length_scale+TV::Axis_Vector(2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    start=min_corner+TV::Axis_Vector(1)*(T)(square_n-1)*length_scale+TV::Axis_Vector(2)*(T)(square_n-1)*length_scale;
    int start_i2=X.m;
    TV dir=(TV::Axis_Vector(1)+TV::Axis_Vector(2)).Normalized();
    for(int i=1;i<=stick_n;i++){
        TV pos=start+dir*length_scale*(T)i;
        X.Append(pos);
    }

    start=min_corner+TV::Axis_Vector(1)*((square_n-1)*length_scale+(stick_n+1)*length_scale*(T)(.5*sqrt(2.)))+
        TV::Axis_Vector(2)*((square_n-1)*length_scale+(stick_n+1)*length_scale*(T)(.5*sqrt(2.)));
    int start_i3=X.m+1;
    for(int i=0;i<square_n;i++){
        for(int j=0;j<square_n;j++){
            TV pos=start+TV::Axis_Vector(1)*(T)i*length_scale+TV::Axis_Vector(2)*(T)j*length_scale;
            X.Append(pos);
        }
    }

    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();

    TRIANGLE_MESH triangle_mesh_copy;triangle_mesh_copy.Initialize_Herring_Bone_Mesh(square_n,square_n);
    Randomize_Herring_Bone_Triangle_Mesh(triangle_mesh_copy,square_n,square_n);

    for(int i=1;i<=triangle_mesh_copy.elements.m;i++){
        TV_INT3& e=triangle_mesh_copy.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(start_i1-1+e(1),start_i1-1+e(2),start_i1-1+e(3)));
    }
    for(int i=1;i<=stick_n+1;i++){
        mesh.segment_mesh.elements.Append(TV_INT2(start_i2-1+i,start_i2-1+i+1));
    }
    for(int i=1;i<=triangle_mesh_copy.elements.m;i++){
        TV_INT3& e=triangle_mesh_copy.elements(i);
        mesh.triangle_mesh.elements.Append(TV_INT3(start_i3-1+e(1),start_i3-1+e(2),start_i3-1+e(3)));
    }
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Curve(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int seg_n,const TV& center,const T R,const T length_scale)
{
    ////TODO: implement this
}


template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Tetrahedron_Mesh(/*result*/CODIMENSIONAL_MESH& mesh,const ARRAY<TV>& X,const T length_scale)
{
    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();

    BOX<TV> bounding_box=RANGE<TV>::Bounding_Box(X);
    T alpha_circumradius=length_scale*(T)1.5;
    T dX=alpha_circumradius*(T)1.2;
    for(int d=1;d<=TV::m;d++)if(bounding_box.Edge_Lengths()(d)<dX)
        bounding_box.Change_Size(TV::Axis_Vector(d)*(dX-bounding_box.Edge_Lengths()(d)+(T)2*length_scale));
    TV_INT counts=(TV_INT)(ceil(bounding_box.Edge_Lengths()/dX));
    BOX<TV> delaunay_meshing_grid_box;
    delaunay_meshing_grid_box.min_corner=bounding_box.min_corner;
    delaunay_meshing_grid_box.max_corner=bounding_box.min_corner+TV::Constant_Vector(dX)*(TV)counts;
    GRID<TV> delaunay_meshing_background_grid(counts,delaunay_meshing_grid_box,true);

    GEOMETRY_PARTICLES<TV> volume_particles;ARRAY<int> remap;
    for(int i=1;i<=X.m;i++){
        int p=volume_particles.array_collection->Add_Element();
        remap.Append(i);
        volume_particles.X(p)=X(i);
    }
    TETRAHEDRON_MESH volume_mesh;

    DELAUNAY_MESHING_WITH_ALPHA_SHAPE<TV>* delaunay_meshing=new DELAUNAY_MESHING_WITH_ALPHA_SHAPE<TV>(delaunay_meshing_background_grid,volume_particles);
    delaunay_meshing->Set_Grid_Radius(alpha_circumradius/delaunay_meshing_background_grid.dX.Max());
    delaunay_meshing->use_manifold_mesh=false;
    delaunay_meshing->Initialize_Grid();
    delaunay_meshing->Construct_Mesh(volume_mesh);

    for(int i=1;i<=volume_mesh.elements.m;i++){
        VECTOR<int,TV::m+1> vtx=volume_mesh.elements(i);
        for(int d=1;d<=TV::m+1;d++)vtx(d)=remap(vtx(d));
        mesh.tetrahedron_mesh.elements.Append(vtx);
    }

    mesh.Update_States<TV>(X.m);
    delete delaunay_meshing;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Multiple_Box_Particles(/*result*/ARRAY<TV>& positions,const ARRAY<BOX<TV> >& boxes,const ARRAY<ROTATION<TV> >& rotations,const T& length_scale)
{
    T threshold=pow<2>(length_scale*(T).3);
    for(int i=1;i<=boxes.m;i++){
        ARRAY<TV> box_positions;ARRAY<bool> box_boundary;const BOX<TV>& box=boxes(i);
        Box_Particles(box_positions,box_boundary,box.min_corner,box.max_corner,length_scale);
        ROTATION<TV> rot=rotations(i);
        for(int j=1;j<=box_positions.m;j++){box_positions(j)=rot.Rotate(box_positions(j));}
        
        for(int j=1;j<=box_positions.m;j++){
            bool add=true;
            for(int k=1;k<=positions.m;k++){
                if((box_positions(j)-positions(k)).Magnitude_Squared()<threshold){
                    add=false;break;
                }
            }
            if(add){positions.Append(box_positions(j));}
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Box_Particles(/*result*/ARRAY<TV>& positions,/*result*/ARRAY<bool>& boundary,const TV& min,const TV& max,const T& length_scale)
{
    TV edge_length=(max-min);
    VECTOR<int,3> edge_samples(ceil(VECTOR<T,3>(edge_length)/length_scale));
    for(int i=1;i<=TV::m;i++)if(edge_samples(i)%2==1)edge_samples(i)-=1;
    for(int i=1;i<=TV::m;i++)if(edge_samples(i)==0)edge_samples(i)=1;

    for(int y=0;y<=edge_samples.y;y++){
        T y_factor=((T)y)/edge_samples.y;
        for(int x=0;x<=edge_samples.x;x++){
            T x_factor=((T)x)/edge_samples.x;
            T offset=(x%2==1?(T)0.5*length_scale:(T)0);
            if(y==edge_samples.y)offset=(T)0;

            if(TV::m==3){
                for(int z=0;z<=edge_samples(3);z++){
                    T z_factor=((T)z)/edge_samples.z;
                    T offset2=(z%2==1?(T)0.5*length_scale:(T)0);
                    if(x==edge_samples.x)offset2=(T)0;

                    if(offset!=(T)0&&offset2!=(T)0&&y==0&&x==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),
                            min(2)+y_factor*edge_length(2),
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    else if(offset!=(T)0&&y==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1)+offset2,
                            min(2)+y_factor*edge_length(2),
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    else if(offset2!=(T)0&&x==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),
                            min(2)+y_factor*edge_length(2)+offset,
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1)+offset2,
                        min(2)+y_factor*edge_length(2)+offset,
                        min(3)+z_factor*edge_length(3)));
                    if(x==0||y==0||z==0||x==edge_samples(1)||y==edge_samples(2)||z==edge_samples(3))boundary.Append(true);
                    else boundary.Append(false);
                }//for z
            }//if
            else{
                if(offset!=(T)0&&y==0){
                    positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),min(2)+y_factor*edge_length(2),(T)0));
                    boundary.Append(true);
                }
                positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),min(2)+y_factor*edge_length(2)+offset,(T)0));
                if(x==0||y==0||x==edge_samples(1)||y==edge_samples(2))boundary.Append(true);
                else boundary.Append(false);
            }//else
        }//for x
    }//for y
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Smoothed_Box_Particles(/*result*/ARRAY<TV>& positions,/*result*/ARRAY<bool>& boundary,const TV& min,const TV& max,const T& length_scale)
{
    TV edge_length=(max-min);
    VECTOR<int,3> edge_samples(ceil(VECTOR<T,3>(edge_length)/length_scale));
    for(int i=1;i<=TV::m;i++)if(edge_samples(i)%2==1)edge_samples(i)-=1;
    for(int i=1;i<=TV::m;i++)if(edge_samples(i)==0)edge_samples(i)=1;

    for(int y=0;y<=edge_samples.y;y++){
        T y_factor=((T)y)/edge_samples.y;
        for(int x=0;x<=edge_samples.x;x++){
            T x_factor=((T)x)/edge_samples.x;
            T offset=(x%2==1?(T)0.5*length_scale:(T)0);
            if(y==edge_samples.y)offset=(T)0;

            if(TV::m==3){
                for(int z=0;z<=edge_samples(3);z++){
                    
                    if(z==0&&x==0)continue;
                    if(z==0&&x==edge_samples.x)continue;
                    if(z==edge_samples(3)&&x==0)continue;
                    if(z==edge_samples(3)&&x==edge_samples.x)continue;

                    T z_factor=((T)z)/edge_samples.z;
                    T offset2=(z%2==1?(T)0.5*length_scale:(T)0);
                    if(x==edge_samples.x)offset2=(T)0;

                    if(offset!=(T)0&&offset2!=(T)0&&y==0&&x==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),
                            min(2)+y_factor*edge_length(2),
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    else if(offset!=(T)0&&y==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1)+offset2,
                            min(2)+y_factor*edge_length(2),
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    else if(offset2!=(T)0&&x==0){
                        positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),
                            min(2)+y_factor*edge_length(2)+offset,
                            min(3)+z_factor*edge_length(3)));
                        boundary.Append(true);
                    }
                    positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1)+offset2,
                        min(2)+y_factor*edge_length(2)+offset,
                        min(3)+z_factor*edge_length(3)));
                    if(x==0||y==0||z==0||x==edge_samples(1)||y==edge_samples(2)||z==edge_samples(3))boundary.Append(true);
                    else boundary.Append(false);
                }//for z
            }//if
            else{
                if(offset!=(T)0&&y==0){
                    positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),min(2)+y_factor*edge_length(2),(T)0));
                    boundary.Append(true);
                }
                positions.Append(TV_Compatible<T,TV::m>(min(1)+x_factor*edge_length(1),min(2)+y_factor*edge_length(2)+offset,(T)0));
                if(x==0||y==0||x==edge_samples(1)||y==edge_samples(2))boundary.Append(true);
                else boundary.Append(false);
            }//else
        }//for x
    }//for y
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Box(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale)
{
    ARRAY<bool> boundary;
    Box_Particles(X,boundary,min,max,length_scale);
    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Multiple_Boxes(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const ARRAY<BOX<TV> >& boxes,const ARRAY<ROTATION<TV> >& rotations,const T& length_scale)
{
    Multiple_Box_Particles(X,boxes,rotations,length_scale);
    Tetrahedron_Mesh(mesh,X,length_scale);

    ////prune tets
    ARRAY<int> tets_to_rmv;
    for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){
        const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
        TV center=(T).25*(X(v(1))+X(v(2))+X(v(3))+X(v(4)));
        bool inside=false;
        for(int j=1;j<=boxes.m;j++){
            TV pos=rotations(j).Inverse_Rotate(center);
            if(boxes(j).Inside(pos,-length_scale*(T).1)){
                inside=true;break;
            }
        }
        if(!inside){
            tets_to_rmv.Append(i);
            LOG::cout<<"to rmv: "<<i<<": "<<center<<", boxes: "<<boxes<<std::endl;
        }
    }
    mesh.tetrahedron_mesh.elements.Remove_Sorted_Indices(tets_to_rmv);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Smoothed_Box(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale)
{
    ARRAY<bool> boundary;
    Smoothed_Box_Particles(X,boundary,min,max,length_scale);
    Tetrahedron_Mesh(mesh,X,length_scale);
    /*
    //laplacian smooth
    mesh.tetrahedron_mesh.Initialize_Neighbor_Nodes();
    ARRAY<TV> new_X(X.m);
    const T lambda=(T).5;
    int passes=2;
    for(int i=1;i<=passes;i++){
        ARRAY<TV> change(X.m);
        for(int j=1;j<=X.m;j++){
            const ARRAY<int>& neighbors=(*mesh.tetrahedron_mesh.neighbor_nodes)(j);
            if(neighbors.m==0)continue;
            for(int k=1;k<=neighbors.m;k++){
                change(j)+=X(neighbors(k))-X(j);
            }
            change(j)*=(T)1./(T)neighbors.m;
        }
        for(int j=1;j<=X.m;j++){
            X(j)+=lambda*change(j);
        }
    }
    */
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere_Particles(/*result*/ARRAY<TV>& positions,const TV& origin,const T radius,const T length_scale)
{
    int radial_samples,circular_samples,height_samples;
    height_samples=(int)ceil(radius/length_scale);
    TV direction=TV_Compatible<T,TV::m>((T)1,(T)0,(T)0);//any direction should work
    TV orthogonal=direction.Orthogonal_Vector().Normalized();

    for(int sign=0;sign<=1;sign++){
        if(sign==1)direction*=(T)-1;
        for(int h=sign;h<=(int)height_samples+2;h++){
            T h_factor=min(((T)h)/height_samples,(T)1);
            if(h==height_samples)h_factor=((T)height_samples-(T).4)/height_samples;
            else if(h==height_samples+1)h_factor=((T)height_samples-(T).2)/height_samples;

            T height=h_factor*radius;
            T radial_distance=(T)sqrt((radius*radius)-(height*height));
            radial_samples=(int)ceil(radial_distance/length_scale);
            TV center=origin+height*direction;

            positions.Append(center);

            for(int r=1;r<=radial_samples;r++){
                T r_factor=((T)r)/radial_samples;
                T subradius=r_factor*radial_distance;
                T circumference=(T)two_pi*subradius;
                //circular_samples=(int)ceil(circumference/length_scale);
                circular_samples=(int)(circumference/length_scale);

                for(int c=1;c<=circular_samples;c++){
                    //if(c==circular_samples&&c!=0)continue;
                    T circular_factor=((T)c)/circular_samples;
                    T offset=(h%2==0?((T)0.5/circular_samples):(T)0);
                    //T offset=(T)0;
                    T offset2=(r%2==0?((T)0.5/radial_samples):(T)0);
                    //T offset2=(T)0;
                    T theta=(T)two_pi*(circular_factor+offset+offset2);
                    ROTATION<TV> rotation(theta,TV(direction));
                    TV v=rotation.Rotate(TV(orthogonal)).Normalized();
                    TV new_position=TV(center)+subradius*v;
                    //if(r!=radial_samples){
                    //    new_position+=perturb_coef*random.Get_Uniform_Vector(TV3::Constant_Vector(length_scale),TV3::Constant_Vector(length_scale));
                    //}
                    positions.Append(TV_Compatible<T,TV::m>(new_position));
                }//circular samples
            }//radial samples
        }//height samples
    }//sign
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& origin,const T radius,const T length_scale)
{
    Sphere_Particles(X,origin,radius,length_scale);
    Tetrahedron_Mesh(mesh,X,length_scale*(T)2);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T radius,const T length_scale,const int predefined_circular_samples/*=0*/)
{
    Cylinder_Particles(X,e1,e2,radius,length_scale,predefined_circular_samples);
    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder_With_Rounded_End(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T radius,const T length_scale,const T endpoint_frac/*=0.8*/,const int predefined_circular_samples/*=0*/)
{
    Cylinder_With_Rounded_End_Particles(X,e1,e2,radius,length_scale,endpoint_frac,predefined_circular_samples);
    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Circle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& dir,const T radius,const T length_scale)
{
    Circle_Particles(X,e1,dir,radius,length_scale);
    TV e2=e1+dir*length_scale;ARRAY<TV> X2;
    Circle_Particles(X2,e2,dir,radius,length_scale);
    ARRAY<TV> merged_X;for(int i=1;i<=X.m;i++)merged_X.Append(X(i));for(int i=1;i<=X.m;i++)merged_X.Append(X2(i));
    CODIMENSIONAL_MESH merged_mesh;Tetrahedron_Mesh(merged_mesh,merged_X,length_scale);
    merged_mesh.tetrahedron_mesh.Initialize_Boundary_Mesh();
    for(int i=1;i<=merged_mesh.tetrahedron_mesh.boundary_mesh->elements.m;i++){
        const TV_INT3& v=merged_mesh.tetrahedron_mesh.boundary_mesh->elements(i);
        if(v(1)<=X.m&&v(2)<=X.m&&v(3)<=X.m){mesh.triangle_mesh.elements.Append(v);}
    }
    mesh.Update_States<TV>(X.m);    
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder_Particles(/*result*/ARRAY<TV>& positions,const TV& e1,const TV& e2,const T radius,const T length_scale,const int predefined_circular_samples/*=0*/)
{
    CYLINDER<T> primary_cylinder(e1,e2,radius);
    TV direction=(e2-e1).Normalized();

    int radial_samples,circular_samples,height_samples;
    radial_samples=(int)ceil(primary_cylinder.radius/length_scale);
    height_samples=(int)ceil(primary_cylinder.height/length_scale);
    TV orthogonal=direction.Orthogonal_Vector().Normalized();
    for(int h=0;h<=height_samples;h++){//includes zero
        T h_factor=((T)h)/height_samples;
        TV center=e1+h_factor*primary_cylinder.height*direction;
        positions.Append(center);

        for(int r=1;r<=radial_samples;r++){
            T r_factor=((T)r)/radial_samples;
            T subradius=r_factor*primary_cylinder.radius;
            T circumference=(T)two_pi*subradius;
            circular_samples=(int)ceil(circumference/length_scale);
            if(predefined_circular_samples>0) circular_samples=predefined_circular_samples;
            for(int c=1;c<=circular_samples;c++){
                T circular_factor=((T)c)/circular_samples;
                T offset=(h%2==0?((T)0.5/circular_samples):(T)0);
                T theta=(T)two_pi*(circular_factor-offset);
                ROTATION<TV> rotation(theta,TV(direction));
                TV v=rotation.Rotate(TV(orthogonal)).Normalized();
                TV new_position=TV(center)+subradius*v;
                //if(r!=radial_samples&&h!=0&&h!=height_samples){
                //    new_position+=perturb_coef*random.Get_Uniform_Vector(TV::Constant_Vector((T)-1*length_scale),TV::Constant_Vector(length_scale));
                //}
                positions.Append(new_position);
            }//circular_samples
        }//radial_samples
    }//height_samples
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder_With_Rounded_End_Particles(/*result*/ARRAY<TV>& positions,const TV& e1,const TV& e2,const T radius,const T length_scale,const T endpoint_frac/*=0.8*/,const int predefined_circular_samples/*=0*/)
{
    CYLINDER<T> primary_cylinder(e1,e2,radius);
    TV direction=(e2-e1).Normalized();

    int radial_samples,circular_samples,height_samples;
    radial_samples=(int)ceil(primary_cylinder.radius/length_scale);
    height_samples=(int)ceil(primary_cylinder.height/length_scale);
    TV orthogonal=direction.Orthogonal_Vector().Normalized();
    for(int h=0;h<=height_samples;h++){//includes zero
        T h_factor=((T)h)/height_samples;
        TV center=e1+h_factor*primary_cylinder.height*direction;
        positions.Append(center);

        for(int r=1;r<=radial_samples;r++){
            T r_factor=((T)r)/radial_samples;
            T subradius=r_factor*primary_cylinder.radius;
            T circumference=(T)two_pi*subradius;
            circular_samples=(int)ceil(circumference/length_scale);
            if(predefined_circular_samples>0) circular_samples=predefined_circular_samples;
            for(int c=1;c<=circular_samples;c++){
                T circular_factor=((T)c)/circular_samples;
                T offset=(h%2==0?((T)0.5/circular_samples):(T)0);
                T theta=(T)two_pi*(circular_factor-offset);
                ROTATION<TV> rotation(theta,TV(direction));
                TV v=rotation.Rotate(TV(orthogonal)).Normalized();
                TV new_position=TV(center)+subradius*v;
                //if(r!=radial_samples&&h!=0&&h!=height_samples){
                //    new_position+=perturb_coef*random.Get_Uniform_Vector(TV::Constant_Vector((T)-1*length_scale),TV::Constant_Vector(length_scale));
                //}
                positions.Append(new_position);
            }//circular_samples
        }//radial_samples
    }//height_samples

    //add partial hemisphere after endpoint e2
    T height=primary_cylinder.radius*endpoint_frac;
    height_samples=(int)ceil(height/length_scale);
    for(int h=1;h<=height_samples;h++){
        T h_factor=((T)h)/height_samples;
        TV center=e2+h_factor*height*direction;
        positions.Append(center);

        T radius=sqrt(sqr(primary_cylinder.radius)-sqr(h_factor*height));
        radial_samples=(int)ceil(radius/length_scale);

        for(int r=1;r<=radial_samples;r++){
            T r_factor=((T)r)/radial_samples;
            T subradius=r_factor*radius;
            T circumference=(T)two_pi*subradius;
            circular_samples=(int)ceil(circumference/length_scale);
            if(predefined_circular_samples>0) circular_samples=predefined_circular_samples;
            for(int c=1;c<=circular_samples;c++){
                T circular_factor=((T)c)/circular_samples;
                T offset=(h%2==0?((T)0.5/circular_samples):(T)0);
                T theta=(T)two_pi*(circular_factor-offset);
                ROTATION<TV> rotation(theta,TV(direction));
                TV v=rotation.Rotate(TV(orthogonal)).Normalized();
                TV new_position=TV(center)+subradius*v;
                positions.Append(new_position);
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Circle_Particles(/*result*/ARRAY<TV>& X,const TV& center,const TV& normal,const T radius,const T length_scale)
{
    RANDOM_NUMBERS<T> random(120102341);
    int radial_samples=(int)ceil(radius/length_scale);
    int circular_samples;
    TV orthogonal=normal.Orthogonal_Vector().Normalized();
    X.Append(center);
    for(int r=1;r<=radial_samples;r++){
        T r_factor=((T)r)/radial_samples;
        T subradius=r_factor*radius;
        T circumference=(T)two_pi*subradius;
        circular_samples=(int)ceil(circumference/length_scale);
        T perturb_theta=random.Get_Uniform_Number((T)-0.5*length_scale/subradius,(T)0.5*length_scale/subradius);

        for(int c=1;c<=circular_samples;c++){
            T circular_factor=((T)c)/circular_samples;
            T theta=(T)two_pi*(T)circular_factor;
            ROTATION<VECTOR<T,3> > rotation(theta+perturb_theta,VECTOR<T,3>(normal));
            VECTOR<T,3> v=rotation.Rotate(VECTOR<T,3>(orthogonal)).Normalized();
            VECTOR<T,3> new_position=VECTOR<T,3>(center)+subradius*v;
            X.Append(TV_Compatible<T,TV::m>(new_position));
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles(/*result*/ARRAY<TV>& positions,const int m,const int n,const TV& min,const T length_scale,const TV& axis_1,const TV& axis_2,const int p_in_cell/*=1*/,const T perturb/*=(T)0*/)
{
    RANDOM_NUMBERS<T> random(120102341);
    for(int p=1;p<=p_in_cell;p++)for(int i=0;i<m;i++){for(int j=0;j<n;j++){TV pos=min+axis_1*length_scale*(T)i+axis_2*length_scale*(T)j;if(perturb!=(T)0)pos+=random.Get_Uniform_Vector(-TV::Constant_Vector(perturb),TV::Constant_Vector(perturb));positions.Append(pos);}}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(/*result*/ARRAY<TV>& positions,const int m,const int n,const TV& center,const T length_scale,const TV& axis_1,const TV& axis_2,const int p_in_cell/*=1*/,const T perturb/*=(T)0*/)
{
    TV start=center-(T).5*(axis_1*length_scale*(T)(m-1)+axis_2*length_scale*(T)(n-1));Rectangle_Particles(positions,m,n,start,length_scale,axis_1,axis_2,p_in_cell,perturb);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(/*result*/ARRAY<TV>& positions,TRIANGULATED_SURFACE<T> &guide_geometry,const int subdivision_step)
{
    for(int i=1;i<=subdivision_step;++i) guide_geometry.Loop_Subdivide();
    for(int i=1;i<=guide_geometry.particles.X.m;++i) positions.Append(guide_geometry.particles.X(i));
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,/*result*/ARRAY<T>* thickness,/*result*/ARRAY<TV>* V,ARRAY<ARRAY<T> >* phis,const std::string& dir_name,const int frame/*=0*/,const int region_id/*=0*/,BOX<TV>* box/*=0*/)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>* particles_ptr=new CODIMENSIONAL_FLUID_PARTICLES<TV>();//memory leak?
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=*particles_ptr;
    DEFORMABLE_GEOMETRY_COLLECTION<TV> deformable_geometry_collection(particles,false);
    CODIMENSIONAL_MESH_OBJECT<TV> mesh_object(mesh,particles);
    
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string codim_particle_file_name=dir_name+"/"+f+"/codimensional_fluid_particles";
    deformable_geometry_collection.Add_Structure(mesh_object.Get_Mesh_Object_1D());
    deformable_geometry_collection.Add_Structure(mesh_object.Get_Mesh_Object_2D());
    if(TV::m==3)deformable_geometry_collection.Add_Structure(mesh_object.Get_Mesh_Object_3D());
    if(deformable_geometry_collection.structures.m>0)deformable_geometry_collection.structures.Last()->Update_Number_Nodes();

    STREAM_TYPE stream_type((T()));
    if(FILE_UTILITIES::File_Exists(codim_particle_file_name)){FILE_UTILITIES::Read_From_File(stream_type,codim_particle_file_name,particles);}
    deformable_geometry_collection.Read(stream_type,dir_name,frame,frame,true);

    //if(region_id!=0){
    //    ARRAY<int> tets_to_rmv;
    //    for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);TV_INT4 v_id=;if(Element_Contains_Not(Element(particles.id,v),region_id)){tets_to_rmv.Append(i);}}
    //    mesh.tetrahedron_mesh.elements.Remove_Sorted_Indices(tets_to_rmv);

    //    ARRAY<int> tris_to_rmv;
    //    for(int i=1;i<=mesh.triangle_mesh.elements.m;i++){const TV_INT3& v=mesh.triangle_mesh.elements(i);if(Element_Contains_Not(Element(particles.id,v),region_id)){tris_to_rmv.Append(i);}}
    //    mesh.triangle_mesh.elements.Remove_Sorted_Indices(tris_to_rmv);

    //    ARRAY<int> segs_to_rmv;
    //    for(int i=1;i<=mesh.segment_mesh.elements.m;i++){const TV_INT2& v=mesh.segment_mesh.elements(i);if(Element_Contains_Not(Element(particles.id,v),region_id)){segs_to_rmv.Append(i);}}
    //    mesh.segment_mesh.elements.Remove_Sorted_Indices(segs_to_rmv);
    //}

    ARRAY<int> p_to_rmv;
    for(int i=1;i<=particles.X.m;i++){
        if(particles.id(i)!=region_id||(box&&!box->Lazy_Inside(particles.X(i))))p_to_rmv.Append(i);
    }
    Remove_Sorted_Particles(particles,p_to_rmv,mesh,true);

    X.Resize(particles.X.m);for(int i=1;i<=X.m;i++)X(i)=particles.X(i);
    if(thickness!=0){thickness->Resize(particles.X.m);for(int i=1;i<=X.m;i++)(*thickness)(i)=particles.thickness(i);}
    if(V!=0){V->Resize(particles.X.m);for(int i=1;i<=X.m;i++)(*V)(i)=particles.V(i);}
    if(phis!=0){phis->Resize(particles.phis.m);for(int c=1;c<=phis->m;c++){(*phis)(c).Resize(particles.X.m);for(int i=1;i<=particles.X.m;i++){(*phis)(c)(i)=particles.phis(c)(i);}}}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Particle_Data_From_File(/*result*/CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const std::string& dir_name,const int frame/*=0*/)
{
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string codim_particle_file_name=dir_name+"/"+f+"/codimensional_fluid_particles";
    STREAM_TYPE stream_type((T()));
    if(FILE_UTILITIES::File_Exists(codim_particle_file_name)){FILE_UTILITIES::Read_From_File(stream_type,codim_particle_file_name,particles);}
    else{LOG::cout<<"error--unable to read particle data from "<<codim_particle_file_name<<"..."<<std::endl;}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Tetrahedralized_Volume_From_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const std::string& file_name,const T& length_scale)
{
    TETRAHEDRON_MESH tet_mesh;
    GEOMETRY_PARTICLES<TV> particles;
    TETRAHEDRALIZED_VOLUME<T> tet_volume(tet_mesh,particles);
    //STREAM_TYPE stream_type((T()));
    STREAM_TYPE stream_type(0.f);
    if(FILE_UTILITIES::File_Exists(file_name))FILE_UTILITIES::Read_From_File(stream_type,file_name,tet_volume);
    for(int i=1;i<=particles.X.m;i++){
        X.Append(length_scale*particles.X(i));
    }
    mesh.segment_mesh.elements.Remove_All();
    mesh.triangle_mesh.elements.Remove_All();
    mesh.tetrahedron_mesh.elements.Remove_All();
    for(int i=1;i<=tet_mesh.elements.m;i++){
        mesh.tetrahedron_mesh.elements.Append(tet_mesh.elements(i));
    }
    mesh.Update_States<TV>(X.m);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Mpi_Region_Test_Shape(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const T length_scale)
{
    Mpi_Region_Test_Shape_Particles(X,length_scale);
    Tetrahedron_Mesh(mesh,X,length_scale*(T)2);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Mpi_Region_Test_Shape_Particles(/*result*/ARRAY<TV>& positions,const T& length_scale)
{
    ARRAY<bool> boundary;
    TV minimum,maximum;
    minimum=length_scale*TV((T)-60,(T)-5,(T)-30);
    maximum=length_scale*TV((T)60,(T)5,(T)30);
    //minimum=length_scale*TV((T)-30,(T)-5,(T)-15);
    //maximum=length_scale*TV((T)30,(T)5,(T)15);

    TV length=maximum-minimum;

    ////
    TV local_min=minimum,local_max=maximum;
    local_max(1)-=(T)two_thirds*length(1);
    local_max(3)-=(T)two_thirds*length(3);
    Box_Particles(positions,boundary,local_min,local_max,length_scale);

    ////
    local_min=minimum;local_max=maximum;
    local_min(3)+=(T)two_thirds*length(3);
    local_max(1)-=(T)two_thirds*length(1);
    Box_Particles(positions,boundary,local_min,local_max,length_scale);

    ////
    local_min=minimum;local_max=maximum;
    local_min(1)+=(T)two_thirds*length(1);
    local_max(3)-=(T)two_thirds*length(3);
    Box_Particles(positions,boundary,local_min,local_max,length_scale);

    ////
    local_min=minimum;local_max=maximum;
    local_min(1)+=(T)two_thirds*length(1);
    local_min(3)+=(T)two_thirds*length(3);
    Box_Particles(positions,boundary,local_min,local_max,length_scale);

    ////
    local_min=minimum;local_max=maximum;
    local_min(1)+=(T)one_third*length(1);
    local_min+=TV::Constant_Vector((T)1e-4);
    local_max(1)-=(T)one_third*length(1);
    local_max-=TV::Constant_Vector((T)1e-4);
    Box_Particles(positions,boundary,local_min,local_max,length_scale);

}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Volume_From_Tri_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const T& length_scale, const std::string& filename)
{
    TRIANGLE_MESH tri_mesh;
    GEOMETRY_PARTICLES<TV> tri_particles;
    TRIANGULATED_SURFACE<T> tri_surface(tri_mesh,tri_particles);

    STREAM_TYPE stream_type((T()));
    FILE_UTILITIES::Read_From_File(stream_type,filename,tri_surface);

    if(tri_particles.X.m==0){LOG::cout<<"error--no particles in triangulated surface."<<std::endl;return;}

    tri_mesh.Initialize_Adjacent_Elements();
    tri_surface.Update_Triangle_List();
    tri_surface.Initialize_Hierarchy();
    tri_surface.Update_Bounding_Box();

    TV minimum=tri_particles.X(1),maximum=tri_particles.X(1);
    for(int i=1;i<=tri_particles.X.m;i++){
        //X.Append(tri_particles.X(i));

        for(int d=1;d<=3;d++){
            if(minimum(d)>tri_particles.X(i)(d))minimum(d)=tri_particles.X(i)(d);
            if(maximum(d)<tri_particles.X(i)(d))maximum(d)=tri_particles.X(i)(d);
        }
    }

    VECTOR<int,3> samples;
    for(int d=1;d<=3;d++)samples(d)=(int)((maximum(d)-minimum(d))/length_scale);

    for(int i=1;i<=samples(1);i++){
        for(int j=1;j<=samples(2);j++){
            T offset=(j%2==0?(T)0.5*length_scale:(T)0);
            for(int k=1;k<=samples(3);k++){
                TV pos=minimum;
                pos(1)+=length_scale*(T)i;
                pos(2)+=length_scale*(T)j;
                pos(3)+=offset+length_scale*(T)k;

                if(tri_surface.Inside(pos))X.Append(pos);
            }
        }
    }

    LOG::cout<<"finished adding particles, now generating tetrahedra..."<<std::endl;

    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Ring(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& origin,const TV& normal,const T radius,const int number_of_rings,const T length_scale)
{
    ////initialize position
    TV axis1=normal.Orthogonal_Vector().Normalized();
    TV axis2=TV::Cross_Product(normal,axis1).Normalized();
    int n=(int)ceil((T)two_pi*radius/length_scale); ////particle number per ring
    T theta=(T)two_pi/(T)n;
    for(int j=0;j<number_of_rings;j++){
        T r=radius+(T)j*length_scale;
        for(int i=0;i<n;i++){
            T angle=(T)(j%2)*(T).5*theta+(T)i*theta;
            TV dir=(axis1*cos(angle)+axis2*sin(angle)).Normalized();
            TV pos=origin+r*dir;
            X.Append(pos);
        }
    }

    ////initialize mesh
    for(int j=0;j<number_of_rings-1;j++){
        int offset1=j*n;
        int offset2=(j+1)*n;
        for(int i=1;i<n;i++){
            if(j%2==0){
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset1+i+1,offset2+i));
                mesh.triangle_mesh.elements.Append(TV_INT3(offset2+i,offset1+i+1,offset2+i+1));
            }
            else{
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset2+i+1,offset2+i));
                mesh.triangle_mesh.elements.Append(TV_INT3(offset1+i,offset1+i+1,offset2+i+1));
            }
        }
        if(j%2==0){
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset1+1,offset2+n));
            mesh.triangle_mesh.elements.Append(TV_INT3(offset2+n,offset1+1,offset2+1));
        }
        else{
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset2+1,offset2+n));
            mesh.triangle_mesh.elements.Append(TV_INT3(offset1+n,offset1+1,offset2+1));
        }
    }
    mesh.Update_States<TV>(X.m);  
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale,const T& min_thickness,const T& max_thickness)
{
    Noise_Box_Volume_Particles(X,min,max,length_scale,min_thickness,max_thickness);
    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume_Particles(/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale,const T& min_thickness,const T& max_thickness)
{
    int x_samples=(int)ceil((max(1)-min(1))/length_scale);
    int z_samples=(int)ceil((max(3)-min(3))/length_scale);

    for(int i=0;i<=x_samples;i++){
        T x=min(1)+i*length_scale;
        for(int j=0;j<=z_samples;j++){
            T z=min(3)+j*length_scale;

            X.Append(TV(x,min(2),z));
        }
    }

    T epsilon=(T).1*length_scale;

    ARRAY<TV> to_add;
    for(int i=1;i<=X.m;i++){
        TV pos=X(i);

        T scale=(T)400;
        TV offset=TV::Constant_Vector((T)10);
        T frac=pow(NOISE<T>::Noise1(pos*scale+offset,1),(T)1);
        T height=min_thickness+frac*(max_thickness-min_thickness);

        to_add.Append(pos+TV((T)0,min(52)+min_thickness,(T)0));

        if(height>min_thickness+epsilon){
            T current_height=min_thickness;

            while(current_height<height){
                current_height+=length_scale;
                if(current_height>height)current_height=height;
                to_add.Append(pos+TV((T)0,min(2)+current_height,(T)0));
            }
        }
    }
    for(int i=1;i<=to_add.m;i++)X.Append(to_add(i));
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& center,const T& radius,const T& length_scale,const T& min_thickness,const T& max_thickness,const T scale,const TV offset)
{
    Noise_Cylinder_Volume_Particles(X,center,radius,length_scale,min_thickness,max_thickness,scale,offset);
    Tetrahedron_Mesh(mesh,X,length_scale);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume_Particles(/*result*/ARRAY<TV>& X,const TV& center,const T& radius,const T& length_scale,const T& min_thickness,const T& max_thickness,const T scale,const TV offset)
{
    Circle_Particles(X,center,TV::Axis_Vector(2),radius,length_scale);

    T epsilon=(T).1*length_scale;

    ARRAY<TV> to_add;

    for(int i=1;i<=X.m;i++){
        TV pos=X(i);

        T frac=pow(NOISE<T>::Noise1(pos*scale+offset,1),(T)1);

        T height=min_thickness+frac*(max_thickness-min_thickness);

        to_add.Append(pos+TV((T)0,center(2)+min_thickness,(T)0));

        if(height>min_thickness+epsilon){
            T current_height=min_thickness;

            while(current_height<height){
                current_height+=length_scale;
                if(current_height>height)current_height=height;
                to_add.Append(pos+TV((T)0,center(2)+current_height,(T)0));
            }
        }
    }

    for(int i=1;i<=to_add.m;i++)X.Append(to_add(i));
}

//template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<float,2> >;
template class CODIMENSIONAL_MESH_GENERATOR<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<double,2> >;
template class CODIMENSIONAL_MESH_GENERATOR<VECTOR<double,3> >;
#endif
