//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH_GENERATOR__
#define __CODIMENSIONAL_MESH_GENERATOR__
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE.h>
#include <PhysBAM_Geometry/Solids_Geometry/DEFORMABLE_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include <PhysBAM_Geometry/Basic_Geometry/ELLIPSE.h>
#include <PhysBAM_Geometry/Basic_Geometry/ELLIPSOID.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"

#include <string>

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_MESH_GENERATOR
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,ELLIPSE<T>,ELLIPSOID<T> >::TYPE T_ELLIPSOID;
    typedef typename IF<TV::m==2,SEGMENT_MESH,TRIANGLE_MESH>::TYPE T_SURFACE_MESH;
    typedef typename IF<TV::m==2,SEGMENTED_CURVE_2D<T>,TRIANGULATED_SURFACE<T> >::TYPE T_SURFACE_MESH_OBJECT;

public:
    CODIMENSIONAL_MESH_GENERATOR(){}

    //static void Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& dir,const int segment_num,const TV& min_corner,const T length_scale);    ////in x direction
    static void Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T length_scale);    ////in x direction
    static void Bended_Filament(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV e1,const T length,const T length_scale,BSPLINE<T,TV>& dir_bspline);

    static void Branch(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int branch_n,const int branch_seg_n,const TV& center,const T length_scale);
    
    static void Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,/*y-axis*/const int m,/*x-axis*/const int n,const TV& min_corner,const T length_scale,const int axis=3); ////in x-y plane
    static void Randomized_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int m,const int n,const TV& min_corner,const T length_scale,const int axis=3); ////in x-y plane
    static void Centrosymmetric_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int m,const int n,
        const TV& min_corner,const TV& center,const T length_scale,const bool reverse=false,const int axis=3); ////in x-y plane
    static void Herring_Bone_Rectangle(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,const TV& min_corner,const VECTOR<T,2>& size,const T length_scale,const int axis=3);  
    static void Randomized_Herring_Bone_Rectangle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min_corner,const VECTOR<T,2>& size,T length_scale,const int axis=3);
    static void Ring(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& origin,const TV& normal,const T inner_radius,const int number_of_rings,const T length_scale);

    static void Square_Dumbbell(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int square_n,const int stick_n,const TV& min_corner,const T length_scale);  ////in x-y plane
    static void Sandglass(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int tri_seg_n,const int stick_seg_n,const TV& corner,const T length_scale);
    static void Square_Sticks(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int square_n,const int stick_n,const TV& min_corner,const T length_scale);    ////in x-y plane
    static void Curve(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const int seg_n,const TV& center,const T R,const T length_scale);
    
    static void Box(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale);
    static void Multiple_Boxes(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const ARRAY<BOX<TV> >& boxes,const ARRAY<ROTATION<TV> >& rotations,const T& length_scale);
    static void Smoothed_Box(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale);
    static void Tetrahedron_Mesh(/*result*/CODIMENSIONAL_MESH& mesh,const ARRAY<TV>& X,const T length_scale);
    static void Sphere(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& origin,const T radius,const T length_scale);
    static void Cylinder(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T radius,const T length_scale,const int predefined_circular_samples=0);
    static void Cylinder_With_Rounded_End(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& e2,const T radius,const T length_scale,const T endpoint_frac=(T)0.8,const int predefined_circular_samples=0);
    static void Circle(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& e1,const TV& dir,const T radius,const T length_scale);

    static void Mpi_Region_Test_Shape(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const T length_scale);
    
    static void Read_Mesh_And_Attributes_From_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,/*result*/ARRAY<T>* thickness,/*result*/ARRAY<TV>* V,/*result*/ARRAY<ARRAY<T> >* phis,const std::string& dir_name,const int frame=0,const int region_id=0,BOX<TV>* box=0);
    static void Read_Particle_Data_From_File(/*result*/CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const std::string& dir_name,const int frame=0);
    static void Read_Tetrahedralized_Volume_From_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const std::string& file_name,const T& length_scale);
    
    ////helper functions
    static void Pascal_Triangle(/*result*/TRIANGLE_MESH& mesh,/*result*/ARRAY<TV>& X,const int tri_n,const T x_dir,const TV& corner,const T length_scale);
    static void Randomize_Herring_Bone_Triangle_Mesh(/*result*/TRIANGLE_MESH& triangle_mesh,const int m,const int n);
    static void Centrosymmetrize_Herring_Bone_Triangle_Mesh(/*result*/TRIANGLE_MESH& triangle_mesh,const int m,const int n,const ARRAY<TV>& X,const TV& center,const bool reverse); ////in x-y plane

    static void Box_Particles(/*result*/ARRAY<TV>& positions,/*result*/ARRAY<bool>& boundary,const TV& min,const TV& max,const T& length_scale);
    static void Multiple_Box_Particles(/*result*/ARRAY<TV>& positions,const ARRAY<BOX<TV> >& boxes,const ARRAY<ROTATION<TV> >& rotations,const T& length_scale);    
    static void Smoothed_Box_Particles(/*result*/ARRAY<TV>& positions,/*result*/ARRAY<bool>& boundary,const TV& min,const TV& max,const T& length_scale);
    static void Sphere_Particles(/*result*/ARRAY<TV>& positions,const TV& origin,const T radius,const T length_scale);
    static void Cylinder_Particles(/*result*/ARRAY<TV>& positions,const TV& e1,const TV& e2,const T radius,const T length_scale,const int predefined_circular_samples=0);
    static void Cylinder_With_Rounded_End_Particles(/*result*/ARRAY<TV>& positions,const TV& e1,const TV& e2,const T radius,const T length_scale,const T endpoint_frac=(T)0.8,const int predefined_circular_samples=0);
    static void Circle_Particles(/*result*/ARRAY<TV>& positions,const TV& e1,const TV& dir,const T radius,const T length_scale);
    static void Rectangle_Particles(/*result*/ARRAY<TV>& positions,const int m,const int n,const TV& min,const T length_scale,const TV& axis_1,const TV& axis_2,const int p_in_cell=1,const T perturb=(T)0);
    static void Rectangle_Particles_Centered(/*result*/ARRAY<TV>& positions,const int m,const int n,const TV& center,const T length_scale,const TV& axis_1,const TV& axis_2,const int p_in_cell=1,const T perturb=(T)0);

    // minjae functions
    static void Brush_Particles(/*result*/ARRAY<TV>& positions,TRIANGULATED_SURFACE<T> &guide_geometry,const int subdivision_step);

    static void Mpi_Region_Test_Shape_Particles(/*result*/ARRAY<TV>& positions,const T& length_scale);
    static void Volume_From_Tri_File(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const T& length_scale, const std::string& filename);

    static void Noise_Box_Volume(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale,const T& min_thickness,const T& max_thickness);
    static void Noise_Box_Volume_Particles(/*result*/ARRAY<TV>& X,const TV& min,const TV& max,const T& length_scale,const T& min_thickness,const T& max_thickness);

    static void Noise_Cylinder_Volume(/*result*/CODIMENSIONAL_MESH& mesh,/*result*/ARRAY<TV>& X,const TV& center,const T& radius,const T& length_scale,const T& min_thickness,const T& max_thickness,const T scale,const TV offset);
    static void Noise_Cylinder_Volume_Particles(/*result*/ARRAY<TV>& X,const TV& center,const T& radius,const T& length_scale,const T& min_thickness,const T& max_thickness,const T scale,const TV offset);
};
}
#endif
