//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH_HASHTABLE__
#define __CODIMENSIONAL_MESH_HASHTABLE__
#include "../Meshing/CODIMENSIONAL_HELPER.h"

namespace PhysBAM
{
namespace Codimensional
{
typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;

template<class TV_INTE>
void Update_Vertex_Element_Hashtable(const ARRAY<TV_INTE>& elements,HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable)
{
    vertex_element_hashtable.Remove_All();
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        for(int j=1;j<=TV_INTE::m;j++){
            ARRAY<int>& v_elements=vertex_element_hashtable.Get_Or_Insert(v(j),ARRAY<int>());
            v_elements.Append(i);
        }
    }
}

template<class TV_INTE>
void Update_Edge_Element_Hashtable(const ARRAY<TV_INTE>& elements,HASHTABLE<TV_INT2,ARRAY<int> >& edge_element_hashtable)
{
    edge_element_hashtable.Remove_All();
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
        for(int j=1;j<=Simplex_Edge_Number<TV_INTE::m>();j++){
            TV_INT2 edge_key=edges(j).Sorted();
            ARRAY<int>& e_tets=edge_element_hashtable.Get_Or_Insert(edge_key,ARRAY<int>());
            e_tets.Append_Unique(i);
        }
    }
}

template<class TV_INTE,class TV_INTF>
void Update_Face_Element_Hashtable(const ARRAY<TV_INTE>& elements,HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable)
{
    //typedef VECTOR<int,TV_INTE::m-1> TV_INTF;
    face_element_hashtable.Remove_All();
    for(int i=1;i<=elements.m;i++){
        ARRAY<TV_INTF> faces;Simplex_Faces(elements(i),faces);
        for(int j=1;j<=Simplex_Face_Number<TV_INTE::m>();j++){
            TV_INTF face_key=faces(j).Sorted();
            ARRAY<int>& face_e=face_element_hashtable.Get_Or_Insert(face_key,ARRAY<int>());
            face_e.Append(i);
        }
    }
}

template<class TV_INTE>
void Update_Edge_Hashtable(const ARRAY<TV_INTE>& elements,HASHTABLE<TV_INT2>& edge_hashtable)
{
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
        for(int j=1;j<=Simplex_Edge_Number<TV_INTE::m>();j++){
            TV_INT2 edge_key=edges(j).Sorted();
            edge_hashtable.Set(edge_key);
        }
    }
}

template<class TV_INTE,class TV_INTF>
void Update_Boundary_Face_Hashtable(HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable,HASHTABLE<TV_INTF>& boundary_face_hashtable)
{
    boundary_face_hashtable.Remove_All();
    for(HASHTABLE_ITERATOR<TV_INTF,ARRAY<int> > iter(face_element_hashtable);iter.Valid();iter.Next()){
        if(iter.Data().m==1)boundary_face_hashtable.Set(iter.Key());
    }
}

template<class TV,class TV_INTE> void Update_Vertex_Element_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable)
{
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        if(Has_Vertex_Inside_Regions(X,v,ranges,ext)){
            for(int j=1;j<=TV_INTE::m;j++){
                ARRAY<int>& v_elements=vertex_element_hashtable.Get_Or_Insert(v(j),ARRAY<int>());
                v_elements.Append(i);
            }
        }
    }
}

template<class TV_INTE> void Add_Element(HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const TV_INTE& v,const int t)
{
    for(int i=1;i<=TV_INTE::m;i++){ARRAY<int>& v_elements=vertex_element_hashtable.Get_Or_Insert(v(i),ARRAY<int>());v_elements.Append(t);}
}

template<class TV_INTE> void Remove_Element(HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const TV_INTE& v,const int t)
{
    for(int i=1;i<=TV_INTE::m;i++){
        ARRAY<int>* v_elements=vertex_element_hashtable.Get_Pointer(v(i));
        if(v_elements)Remove_From_Array(*v_elements,t);
    }
}

inline void Remove_Key(HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const int v)
{vertex_element_hashtable.Delete_If_Present(v);}

inline void Merge_Keys(HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const VECTOR<int,2> ac,const int v)
{
    ARRAY<int> v_elements;if(vertex_element_hashtable.Contains(v))v_elements=vertex_element_hashtable.Get(v);
    for(int i=1;i<=2;i++){
        int a=ac(i);ARRAY<int>* a_elements=vertex_element_hashtable.Get_Pointer(a);
        if(a_elements)for(int j=1;j<=(*a_elements).m;j++){v_elements.Append_Unique((*a_elements)(j));}
        vertex_element_hashtable.Delete_If_Present(a);
    }
    if(v_elements.m>0)vertex_element_hashtable.Set(v,v_elements);
}

template<class TV_INTE> void Replace_Element(HASHTABLE<int,ARRAY<int> >& vertex_element_hashtable,const TV_INTE& v,const VECTOR<int,2>& ac,const int p)
{
    for(int i=1;i<=TV_INTE::m;i++){
        if(ac.Contains(v(i))){
            ARRAY<int>* v_elements=vertex_element_hashtable.Get_Pointer(v(i));
            if(v_elements){ ////TODO: didn't eliminate ac incident elements
                ARRAY<int> v_elements_copy=(*v_elements);
                ARRAY<int>& p_elements=vertex_element_hashtable.Get_Or_Insert(p,ARRAY<int>());
                for(int j=1;j<=v_elements_copy.m;j++){p_elements.Append_Unique(v_elements_copy(j));}
                Remove_Key(vertex_element_hashtable,v(i));
            }
        }
    }
}

template<class TV_INTE> void Add_Element(HASHTABLE<TV_INT2,ARRAY<int> >& edge_element_hashtable,const TV_INTE& v,const int t)
{
    typedef VECTOR<int,2> TV_INT2;ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
    for(int i=1;i<=Simplex_Edge_Number<TV_INTE::m>();i++){
        TV_INT2 edge_key=edges(i).Sorted();
        ARRAY<int>& e_elements=edge_element_hashtable.Get_Or_Insert(edge_key,ARRAY<int>());
        e_elements.Append(t);
    }
}

template<class TV_INTE> void Remove_Element(HASHTABLE<VECTOR<int,2>,ARRAY<int> >& edge_element_hashtable,const TV_INTE& v,const int t)
{
    typedef VECTOR<int,2> TV_INT2;ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
    for(int i=1;i<=Simplex_Edge_Number<TV_INTE::m>();i++){
        TV_INT2 edge_key=edges(i).Sorted();
        ARRAY<int>* e_elements=edge_element_hashtable.Get_Pointer(edge_key);
        if(e_elements)Remove_From_Array(*e_elements,t);
    }
}

inline void Remove_Key(HASHTABLE<VECTOR<int,2>,ARRAY<int> >& edge_element_hashtable,const VECTOR<int,2>& e_key)
{edge_element_hashtable.Delete_If_Present(e_key);}

template<class TV_INTE> void Replace_Element(HASHTABLE<VECTOR<int,2>,ARRAY<int> >& edge_element_hashtable,const TV_INTE& v,const VECTOR<int,2>& ac,const int p)
{
    typedef VECTOR<int,2> TV_INT2;ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
    for(int i=1;i<=Simplex_Edge_Number<TV_INTE::m>();i++){
        TV_INT2 edge_key=edges(i).Sorted();
        TV_INT2 replaced_edge_key=Replaced_Two(edge_key,ac(1),ac(2),p).Sorted();
        if(replaced_edge_key!=edge_key){
            ARRAY<int>* e_elements=edge_element_hashtable.Get_Pointer(edge_key);
            if(e_elements){
                if(replaced_edge_key(1)!=replaced_edge_key(2)){ ////check duplicate vertices in element
                    ARRAY<int> e_elements_copy=(*e_elements);
                    ARRAY<int>& replaced_e_elements=edge_element_hashtable.Get_Or_Insert(replaced_edge_key,ARRAY<int>());
                    for(int j=1;j<=e_elements_copy.m;j++){replaced_e_elements.Append_Unique((e_elements_copy)(j));}
                }
                Remove_Key(edge_element_hashtable,edge_key);
            }
        }
    }
}

template<class TV_INTE,class TV_INTF> void Add_Element(HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable,const TV_INTE& v,const int t)
{
    ARRAY<TV_INTF> faces;Simplex_Faces(v,faces);
    for(int i=1;i<=Simplex_Face_Number<TV_INTE::m>();i++){
        TV_INTF face_key=faces(i).Sorted();
        ARRAY<int>& f_elements=face_element_hashtable.Get_Or_Insert(face_key,ARRAY<int>());
        f_elements.Append(t);
    }
}

template<class TV_INTE,class TV_INTF> void Remove_Element(HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable,const TV_INTE& v,const int t)
{
    ARRAY<TV_INTF> faces;Simplex_Faces(v,faces);
    for(int i=1;i<=Simplex_Face_Number<TV_INTE::m>();i++){
        TV_INTF face_key=faces(i).Sorted();
        ARRAY<int>* f_elements=face_element_hashtable.Get_Pointer(face_key);
        if(f_elements)Remove_From_Array(*f_elements,t);
    }
}

template<class TV_INTF> void Remove_Key(HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable,const TV_INTF& f_key)
{face_element_hashtable.Delete_If_Present(f_key);}

template<class TV_INTE,class TV_INTF> void Replace_Element(HASHTABLE<VECTOR<int,3>,ARRAY<int> >& face_element_hashtable,const TV_INTE& v,const VECTOR<int,2>& ac,const int p)
{
    ARRAY<TV_INTF> faces;Simplex_Faces(v,faces);
    for(int i=1;i<=Simplex_Face_Number<TV_INTE::m>();i++){
        TV_INT3 face_key=faces(i).Sorted();
        TV_INT3 replaced_face_key=Replaced_Two(face_key,ac(1),ac(2),p).Sorted();
        if(replaced_face_key!=face_key){
            ARRAY<int>* f_elements=face_element_hashtable.Get_Pointer(face_key);
            if(f_elements){
                if(!Contains_Duplicate_Elements(replaced_face_key)){    ////check duplicate vertices in element
                    ARRAY<int> f_elements_copy=*f_elements;
                    ARRAY<int>& replaced_f_elements=face_element_hashtable.Get_Or_Insert(replaced_face_key,ARRAY<int>());
                    for(int j=1;j<=f_elements_copy.m;j++){replaced_f_elements.Append_Unique(f_elements_copy(j));}
                }
                Remove_Key(face_element_hashtable,face_key);
            }
        }
    }
}

template<class TV,class TV_INTE> void Update_Edge_Element_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INT2,ARRAY<int> >& edge_element_hashtable)
{
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        if(Has_Vertex_Inside_Regions(X,v,ranges,ext)){
            ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
            for(int j=1;j<=Simplex_Edge_Number<TV_INTE::m>();j++){
                TV_INT2 edge_key=edges(j).Sorted();
                ARRAY<int>& e_tets=edge_element_hashtable.Get_Or_Insert(edge_key,ARRAY<int>());
                e_tets.Append_Unique(i);
            }
        }
    }
}

template<class TV,class TV_INTE> void Update_Edge_Element_Number_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INT2,int>& edge_element_hashtable)
{
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        if(Has_Vertex_Inside_Regions(X,v,ranges,ext)){
            ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
            for(int j=1;j<=Simplex_Edge_Number<TV_INTE::m>();j++){
                TV_INT2 edge_key=edges(j).Sorted();
                int& e_tet_number=edge_element_hashtable.Get_Or_Insert(edge_key,0);
                e_tet_number++;
            }
        }
    }
}

template<class TV,class TV_INTE,class TV_INTF> void Update_Face_Element_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INTF,ARRAY<int> >& face_element_hashtable)
{
    ////typedef VECTOR<int,TV_INTE::m-1> TV_INTF;
    //for(int i=1;i<=elements.m;i++){
    //    ARRAY<TV_INTF> faces;Simplex_Faces(elements(i),faces);
    //    for(int j=1;j<=Simplex_Face_Number<TV_INTE::m>();j++){
    //        TV_INTF face_key=faces(j).Sorted();
    //        if(Has_Vertex_Inside_Regions(X,face_key,ranges,ext)){
    //            ARRAY<int>& face_e=face_element_hashtable.Get_Or_Insert(face_key,ARRAY<int>());
    //            face_e.Append(i);
    //        }
    //    }
    //}

    ////allow ghost face exist in hashtable, needs to check face region explicitly
    for(int i=1;i<=elements.m;i++){
        if(Has_Vertex_Inside_Regions(X,elements(i),ranges,ext)){
            ARRAY<TV_INTF> faces;Simplex_Faces(elements(i),faces);
            for(int j=1;j<=Simplex_Face_Number<TV_INTE::m>();j++){
                TV_INTF face_key=faces(j).Sorted();
                ARRAY<int>& face_e=face_element_hashtable.Get_Or_Insert(face_key,ARRAY<int>());
                face_e.Append(i);
            }
        }
    }
}

template<class TV,class TV_INTE,class TV_INTF> void Update_Face_Element_Number_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INTF,int>& face_element_number_hashtable)
{
    ////does not allow ghost face exist in hashtable, TODO: make it consistent with Update_Face_Element_Hashtable
    //typedef VECTOR<int,TV_INTE::m-1> TV_INTF;
    for(int i=1;i<=elements.m;i++){
        ARRAY<TV_INTF> faces;Simplex_Faces(elements(i),faces);
        for(int j=1;j<=Simplex_Face_Number<TV_INTE::m>();j++){
            TV_INTF face_key=faces(j).Sorted();
            if(Has_Vertex_Inside_Regions(X,face_key,ranges,ext)){
                if(face_element_number_hashtable.Contains(face_key)){
                    int n=face_element_number_hashtable.Get(face_key);
                    face_element_number_hashtable.Set(face_key,n+1);
                }
                else face_element_number_hashtable.Set(face_key,1);
            }
        }
    }
}

template<class TV,class TV_INTE,class TV_INTF> void Update_Boundary_Face_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INTF>& boundary_face_hashtable)
{
    HASHTABLE<TV_INTF,int> face_element_number_hashtable;
    Update_Face_Element_Number_Hashtable(elements,X,ranges,ext,face_element_number_hashtable);
    for(HASHTABLE_ITERATOR<TV_INTF,int> iter(face_element_number_hashtable);iter.Valid();iter.Next()){
        if(iter.Data()==1){boundary_face_hashtable.Set(iter.Key());}
    }
}

template<class TV,class TV_INTE> void Update_Boundary_Edge_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INT2>& boundary_edge_hashtable)
{
    typedef VECTOR<int,TV_INTE::m-1> TV_INTF;
    HASHTABLE<TV_INTF> boundary_face_hashtable;
    Update_Boundary_Face_Hashtable(elements,X,ranges,ext,boundary_face_hashtable);
    for(HASHTABLE_ITERATOR<TV_INTF> iter(boundary_face_hashtable);iter.Valid();iter.Next()){
        const TV_INTF& v=iter.Key();
        ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
        for(int j=1;j<=Simplex_Edge_Number<TV_INTF::m>();j++){
            TV_INT2 edge_key=edges(j).Sorted();
            boundary_edge_hashtable.Set(edge_key);
        }
    }
}

template<class TV,class TV_INTE> void Update_Edge_Hashtable(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,HASHTABLE<TV_INT2>& edge_hashtable)
{
    for(int i=1;i<=elements.m;i++){
        const TV_INTE& v=elements(i);
        if(Has_Vertex_Inside_Regions(X,v,ranges,ext)){
            ARRAY<TV_INT2> edges;Simplex_Edges(v,edges);
            for(int j=1;j<=Simplex_Edge_Number<TV_INTE::m>();j++){
                TV_INT2 edge_key=edges(j).Sorted();
                edge_hashtable.Set(edge_key);
            }
        }
    }
}

template<class TV,class TV_INTE> void Update_Particle_Dimensional_Size(const ARRAY<TV_INTE>& elements,ARRAY_VIEW<TV>& X,const ARRAY<RANGE<TV> >& ranges,const typename TV::SCALAR ext,ARRAY_VIEW<TV>& dimensional_size)
{
    typedef typename TV::SCALAR T;
    typedef typename IF<TV_INTE::m==2,SEGMENT_3D<T>, typename IF<TV_INTE::m==3,TRIANGLE_3D<T>,TETRAHEDRON<T> >::TYPE >::TYPE T_ELEMENT;

    int dim=TV_INTE::m-1;
    for(int i=1;i<=dimensional_size.m;i++)dimensional_size(i)(dim)=(T)0;

    for(int i=1;i<=elements.m;i++){
        TV_INTE v=elements(i);
        if(Has_Vertex_Inside_Regions(X,v,ranges,ext)){
            INDIRECT_ARRAY<ARRAY_VIEW<TV>,TV_INTE& > indir_X(X,v);
            T size=T_ELEMENT::Size(indir_X);
            for(int d=1;d<=TV_INTE::m;d++)dimensional_size(v(d))(dim)+=(T)1/(T)(TV_INTE::m)*size;
        }
    }
}

}
}

#endif
