//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT.h"
using namespace PhysBAM;

//////////////////////////////////////////////////////////////////////////
////Codimensional_Mesh_Object_Base
template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Segment_Length(const int e)
{
    return (particles.X(mesh.segment_mesh.elements(e)(1))-particles.X(mesh.segment_mesh.elements(e)(2))).Magnitude();
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Triangle_Area(const int e)
{
    return T_TRIANGLE::Area(particles.X(mesh.triangle_mesh.elements(e)(1)),
        particles.X(mesh.triangle_mesh.elements(e)(2)),particles.X(mesh.triangle_mesh.elements(e)(3)));
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum(const int p)
{
    T length_sum=(T)0;
    for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
        int s=(*mesh.segment_mesh.incident_elements)(p)(i);
        length_sum+=(T).5*Segment_Length(s);
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum_On_Triangle_Mesh_Boundary( const int p )
{
    T length_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.boundary_mesh->incident_elements)(p).m;i++){
        int s=(*mesh.triangle_mesh.boundary_mesh->incident_elements)(p)(i);
        VECTOR<int,2>& v=(*mesh.triangle_mesh.boundary_mesh).elements(s);
        length_sum+=(T).5*(particles.X(v(1))-particles.X(v(2))).Magnitude();
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum_On_Rim_Mesh(const int p)
{
    T length_sum=(T)0;
    for(int i=1;i<=(*mesh.rim_segment_mesh.incident_elements)(p).m;i++){
        int s=(*mesh.rim_segment_mesh.incident_elements)(p)(i);
        VECTOR<int,2>& v=mesh.rim_segment_mesh.elements(s);
        length_sum+=(T).5*(particles.X(v(1))-particles.X(v(2))).Magnitude();
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Average(const int p)
{
    int degree=Nodal_Segment_Degree(p);T length_sum=Nodal_Segment_Length_Sum(p);
    return degree>0?length_sum/(T)degree:(T)0;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Triangle_Area_Sum(const int p)
{
    T area_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        area_sum+=(T)one_third*Triangle_Area(t);
    }
    return area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Extra_Triangle_Area_Sum(const int p)
{
    if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(p))return (T)0;
    T extra_area_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        VECTOR<int,2> other_nodes=Other_Nodes_In_Element(mesh.triangle_mesh.elements(t),p);
        int rim_p_num=(mesh.Is_Boundary_Particle_On_Triangle_Mesh(other_nodes(1))?1:0)+(mesh.Is_Boundary_Particle_On_Triangle_Mesh(other_nodes(2))?1:0);
        switch(rim_p_num){
        case 1:extra_area_sum+=(T)one_sixth*Triangle_Area(t);break;
        case 2:extra_area_sum+=(T)two_thirds*Triangle_Area(t);break;
        }
    }

    return extra_area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum(ARRAY_VIEW<TV>& X,const int p,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable)
{
    ARRAY<int> nb_p;T length_sum=(T)0;
    if(rim_vtx_vtx_hashtable.Get(p,nb_p)){
        for(int i=1;i<=nb_p.m;i++){length_sum+=(T).5*(X(p)-X(nb_p(i))).Magnitude();}
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Fraction(ARRAY_VIEW<TV>& X,const int p,const int p2,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable)
{
    T length_sum=Nodal_Segment_Length_Sum(X,p,rim_vtx_vtx_hashtable);
    T seg_length=(T).5*(X(p)-X(p2)).Magnitude();
    return seg_length/length_sum;
}

//////////////////////////////////////////////////////////////////////////
////Codimensional_Mesh_Object 2D
template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::CODIMENSIONAL_MESH_OBJECT()
:BASE(*(new CODIMENSIONAL_MESH()),*(new CODIMENSIONAL_FLUID_PARTICLES<TV>())),volume_mesh(mesh.triangle_mesh),segmented_curve(mesh.segment_mesh,particles),triangulated_area(mesh.triangle_mesh,particles)
{}

template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::CODIMENSIONAL_MESH_OBJECT(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input)
:BASE(mesh_input,particles_input),volume_mesh(mesh.triangle_mesh),segmented_curve(mesh.segment_mesh,particles),triangulated_area(mesh.triangle_mesh,particles)
{}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Triangle_Degree(const int p)
{
    //return (*mesh.triangle_mesh.incident_elements)(p).m;
    return 0;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Degree(const int p)
{
    //return (*mesh.segment_mesh.incident_elements)(p).m;
    return 0;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Triangle_Area_Average(const int p)
{
    int degree=Nodal_Triangle_Degree(p);T area_sum=Nodal_Triangle_Area_Sum(p);
    return degree>0?area_sum/(T)degree:(T)0;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Point_Volume(const int p)
{   ////TODO: or mass/density?
    //return (T)pi*pow<2>(particles.radius(p));
    return particles.mass(p)/(T)1000;   ////TODO: specify density
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Segment_Volume(const int e)
{
    T thickness1=particles.thickness(mesh.segment_mesh.elements(e)(1));T thickness2=particles.thickness(mesh.segment_mesh.elements(e)(2));
    return Segment_Cross_Section_Area(thickness1)*(T).5*Segment_Length(e)+Segment_Cross_Section_Area(thickness2)*(T).5*Segment_Length(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Triangle_Volume(const int e)
{
    return Triangle_Area(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Volumetric_Element_Volume(const int e)
{
    return Triangle_Volume(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Segment_Volume_Fraction(const int p,const int s)
{
    int p2=(Other_Nodes_In_Element(mesh.segment_mesh.elements(s),p))(1);
    T s1=particles.thickness(p);T s2=particles.thickness(p2);
    return s1/(s1+s2);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Segment_Volume(const int p,const int s)
{
    return (T).5*Segment_Length(s)*particles.thickness(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Segment_Volume(const int p)
{
    return Nodal_Segment_Length_Sum(p)*Segment_Cross_Section_Area(particles.thickness(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Triangle_Volume_Fraction(const int p,const int t)
{
    return (T)one_third;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Triangle_Volume(const int p,const int t)
{
    return (T)one_third*Triangle_Area(t);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Triangle_Volume(const int p)
{
    return Nodal_Triangle_Area_Sum(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Volume(const int p)
{
    T volume=(T)0;
    //for(int i=1;i<=(*mesh.incident_elements)(p).m;i++){
    //    CODIMENSIONAL_ELEMENT &e=mesh.elements((*mesh.incident_elements)(p)(i));
    //    switch(e.type){
    //    case SIMPLEX_SEGMENT:
    //        volume+=Nodal_Segment_Volume(p,e.index);break;
    //    case SIMPLEX_TRIANGLE:
    //        volume+=Nodal_Triangle_Volume(p,e.index);break;
    //    default:break;
    //    }
    //}
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Volume_In_Highest_Dimension(const int p)
{
    T volume=(T)0;
    //for(int i=1;i<=(*mesh.incident_elements)(p).m;i++){
    //    CODIMENSIONAL_ELEMENT &e=mesh.elements((*mesh.incident_elements)(p)(i));
    //    switch(e.type){
    //    case SIMPLEX_TRIANGLE:
    //        volume+=Nodal_Triangle_Volume(p,e.index);break;
    //    default:break;
    //    }
    //}
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,2> >::Nodal_Volume_In_Highest_Dimension_Fraction(const int p)
{
    T nodal_volume=Nodal_Volume(p);if(nodal_volume==(T)0)return (T)1;
    T nodal_volume_highest_dim=Nodal_Volume_In_Highest_Dimension(p);
    return nodal_volume_highest_dim/nodal_volume;
}

//////////////////////////////////////////////////////////////////////////
////Codimensional_Mesh_Object 3D
template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::CODIMENSIONAL_MESH_OBJECT()
    :BASE(*(new CODIMENSIONAL_MESH()),*(new CODIMENSIONAL_FLUID_PARTICLES<TV>())),volume_mesh(mesh.tetrahedron_mesh),
    segmented_curve(mesh.segment_mesh,particles),triangulated_surface(mesh.triangle_mesh,particles),tetrahedralized_volume(mesh.tetrahedron_mesh,particles)
{}

template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::CODIMENSIONAL_MESH_OBJECT(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input)
    :BASE(mesh_input,particles_input),volume_mesh(mesh.tetrahedron_mesh),
    segmented_curve(mesh.segment_mesh,particles),triangulated_surface(mesh.triangle_mesh,particles),tetrahedralized_volume(mesh.tetrahedron_mesh,particles)
{}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Point_Volume(const int p)
{
    //return (T)4/(T)3*(T)pi*pow<3>(particles.radius(p));
    return particles.mass(p)/(T)1000;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Segment_Volume(const int e)
{
    T thickness1=particles.thickness(mesh.segment_mesh.elements(e)(1));T thickness2=particles.thickness(mesh.segment_mesh.elements(e)(2));
    return Segment_Cross_Section_Area(thickness1)*(T).5*Segment_Length(e)+Segment_Cross_Section_Area(thickness2)*(T).5*Segment_Length(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Triangle_Volume(const int e)
{
    T thickness=(T)one_third*(particles.thickness(mesh.triangle_mesh.elements(e)(1))+particles.thickness(mesh.triangle_mesh.elements(e)(2))+particles.thickness(mesh.triangle_mesh.elements(e)(3)));
    return thickness*Triangle_Area(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Tetrahedron_Volume(const int e)
{
    return TETRAHEDRON<T>::Volume(particles.X(mesh.tetrahedron_mesh.elements(e)(1)),
        particles.X(mesh.tetrahedron_mesh.elements(e)(2)),particles.X(mesh.tetrahedron_mesh.elements(e)(3)),particles.X(mesh.tetrahedron_mesh.elements(e)(4)));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Volumetric_Element_Volume(const int e)
{
    return Tetrahedron_Volume(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume_Fraction(const int p,const int s)
{
    int p2=(Other_Nodes_In_Element(mesh.segment_mesh.elements(s),p))(1);
    T s1=pow<2>(particles.thickness(p));T s2=pow<2>(particles.thickness(p2));
    return pow<2>(s1)/(pow<2>(s1)+pow<2>(s2));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume(const int p,const int s)
{
    return (T).5*Segment_Length(s)*Segment_Cross_Section_Area(particles.thickness(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume(const int p)
{
    return Nodal_Segment_Length_Sum(p)*Segment_Cross_Section_Area(particles.thickness(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume_Fraction(const int p,const int t)
{
    T thickness_sum=particles.thickness(mesh.triangle_mesh.elements(t)(1))+particles.thickness(mesh.triangle_mesh.elements(t)(2))+particles.thickness(mesh.triangle_mesh.elements(t)(3));
    PHYSBAM_ASSERT(thickness_sum!=(T)0);
    return particles.thickness(p)/thickness_sum;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume(const int p,const int t)
{
    return (T)one_third*Triangle_Area(t)*particles.thickness(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume(const int p)
{
    return Nodal_Triangle_Area_Sum(p)*particles.thickness(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Volume(const int p,const int t)
{
    return (T).25*Tetrahedron_Volume(t);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Volume(const int p)
{
    return Nodal_Tetrahedron_Volume_Sum(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Volume_Sum(const int p)
{
    T volume_sum=(T)0;
    //for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
    //    int t=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
    //    volume_sum+=(T).25*Tetrahedron_Volume(t);
    //}
    return volume_sum;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Area_Weighted_Normal(const int p)
{
    TV area_weighted_normal=TV();
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        VECTOR<int,3> v=mesh.triangle_mesh.elements(t);
        TV normal=TV::Cross_Product(particles.X(v(2))-particles.X(v(1)),particles.X(v(3))-particles.X(v(1))).Normalized();
        area_weighted_normal+=(T)one_third*Triangle_Area(t)*normal;
    }
    return area_weighted_normal;
}


template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume(const int p)
{
    T volume=(T)0;
    //for(int i=1;i<=(*mesh.incident_elements)(p).m;i++){
    //    CODIMENSIONAL_ELEMENT &e=mesh.elements((*mesh.incident_elements)(p)(i));
    //    switch(e.type){
    //    case SIMPLEX_SEGMENT:
    //        volume+=Nodal_Segment_Volume(p,e.index);break;
    //    case SIMPLEX_TRIANGLE:
    //        volume+=Nodal_Triangle_Volume(p,e.index);break;
    //    case SIMPLEX_TETRAHEDRON:
    //        volume+=Nodal_Tetrahedron_Volume(p,e.index);break;
    //    default:break;
    //    }
    //}
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_In_Highest_Dimension(const int p)
{
    T volume=(T)0;
    //for(int i=1;i<=(*mesh.incident_elements)(p).m;i++){
    //    CODIMENSIONAL_ELEMENT &e=mesh.elements((*mesh.incident_elements)(p)(i));
    //    switch(e.type){
    //    case SIMPLEX_TETRAHEDRON:
    //        volume+=(T).25*Tetrahedron_Volume(e.index);break;
    //    default:break;
    //    }
    //}
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_In_Highest_Dimension_Fraction(const int p)
{
    T nodal_volume=Nodal_Volume(p);if(nodal_volume==(T)0)return (T)1;
    T nodal_volume_highest_dim=Nodal_Volume_In_Highest_Dimension(p);
    return nodal_volume_highest_dim/nodal_volume;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Area_Weighted_Mean_Curvature_Normal(const int p)
{
    TV area_weighted_mean_curvature_normal=TV();
    //for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
    //    int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
    //    VECTOR<int,3> v=Reordered(mesh.triangle_mesh.elements(t),p);
    //    area_weighted_mean_curvature_normal+=Cotangent(particles.X(v(3)),particles.X(p),particles.X(v(2)))*(particles.X(p)-particles.X(v(2)));
    //    area_weighted_mean_curvature_normal+=Cotangent(particles.X(v(2)),particles.X(v(3)),particles.X(p))*(particles.X(p)-particles.X(v(3)));
    //}
    return (T).5*area_weighted_mean_curvature_normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Mean_Curvature_Normal(const int p)
{
    T area=Nodal_Triangle_Area_Sum(p);
    return area==(T)0?TV():(T)3*Nodal_Area_Weighted_Mean_Curvature_Normal(p)/area;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_Weighted_Normal(const int p)
{
    return TV();
}

template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<float,2> >;
template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<double,2> >;
template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<double,3> >;
#endif

template class CODIMENSIONAL_MESH_OBJECT<VECTOR<float,2> >;
template class CODIMENSIONAL_MESH_OBJECT<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class CODIMENSIONAL_MESH_OBJECT<VECTOR<double,2> >;
template class CODIMENSIONAL_MESH_OBJECT<VECTOR<double,3> >;
#endif