//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"   ////TODO: move the function in subsystem to helper

using namespace PhysBAM;

template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::CODIMENSIONAL_MESH_OBJECT()
    :BASE(*(new CODIMENSIONAL_MESH()),*(new CODIMENSIONAL_FLUID_PARTICLES<TV>())),volume_mesh(mesh.tetrahedron_mesh),
    segmented_curve(mesh.segment_mesh,particles),triangulated_surface(mesh.triangle_mesh,particles),tetrahedralized_volume(mesh.tetrahedron_mesh,particles)
{}

template<class T_input>
PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::CODIMENSIONAL_MESH_OBJECT(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input)
    :BASE(mesh_input,particles_input),volume_mesh(mesh.tetrahedron_mesh),
    segmented_curve(mesh.segment_mesh,particles),triangulated_surface(mesh.triangle_mesh,particles),tetrahedralized_volume(mesh.tetrahedron_mesh,particles)
{}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Point_Volume(const int p)
{
    //return (T)4/(T)3*(T)pi*pow<3>(particles.radius(p));
    return particles.mass(p)/(T)1000;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Segment_Volume(const int e)
{
    T thickness1=particles.thickness(mesh.segment_mesh.elements(e)(1));T thickness2=particles.thickness(mesh.segment_mesh.elements(e)(2));
    return Segment_Cross_Section_Area(thickness1)*(T).5*Segment_Length(e)+Segment_Cross_Section_Area(thickness2)*(T).5*Segment_Length(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Triangle_Volume(const int e)
{
    T thickness=(T)one_third*(particles.thickness(mesh.triangle_mesh.elements(e)(1))+particles.thickness(mesh.triangle_mesh.elements(e)(2))+particles.thickness(mesh.triangle_mesh.elements(e)(3)));
    return thickness*Triangle_Area(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Tetrahedron_Volume(const int e)
{
    return TETRAHEDRON<T>::Volume(particles.X(mesh.tetrahedron_mesh.elements(e)(1)),
        particles.X(mesh.tetrahedron_mesh.elements(e)(2)),particles.X(mesh.tetrahedron_mesh.elements(e)(3)),particles.X(mesh.tetrahedron_mesh.elements(e)(4)));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Volumetric_Element_Volume(const int e)
{
    return Tetrahedron_Volume(e);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Element_Volume(const int e,const int type)
{
    switch(type){
    case SIMPLEX_PARTICLE:return Nodal_Volume(e);
    case SIMPLEX_SEGMENT:return Segment_Volume(e);
    case SIMPLEX_TRIANGLE:return Triangle_Volume(e);
    case SIMPLEX_TETRAHEDRON:return Tetrahedron_Volume(e);
    default:PHYSBAM_FATAL_ERROR("Invalid type in Element_Volume");return (T)0;
    }
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Element_Volume_With_Thickness(const int e,const int type,const T thickness)
{
    switch(type){
    case SIMPLEX_PARTICLE:return Nodal_Volume_With_Thickness(e,thickness);
    case SIMPLEX_SEGMENT:return Segment_Length(e)*Segment_Cross_Section_Area(thickness);
    case SIMPLEX_TRIANGLE:return Triangle_Area(e)*thickness;
    case SIMPLEX_TETRAHEDRON:return Tetrahedron_Volume(e);
    default:PHYSBAM_FATAL_ERROR("Invalid type in Element_Volume");return (T)0;
    }
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Element_Partial_Volume(const int e,const int type,const HASHTABLE<int>& psi_D_hashtable)
{
    switch(type){
    case SIMPLEX_PARTICLE:return Nodal_Partial_Volume(e,psi_D_hashtable);
    case SIMPLEX_SEGMENT:return Segment_Volume(e);  ////TODO: may return 0 if element has all vertices in psi_D_hashtable
    case SIMPLEX_TRIANGLE:return Triangle_Volume(e);
    case SIMPLEX_TETRAHEDRON:return Tetrahedron_Volume(e);
    default:PHYSBAM_FATAL_ERROR("Invalid type in Element_Volume");return (T)0;
    }
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Segment_Center(const int e)
{const TV_INT2& v=mesh.segment_mesh.elements(e);return (T).5*(particles.X(v(1))+particles.X(v(2)));}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Triangle_Center(const int e)
{const TV_INT3& v=mesh.triangle_mesh.elements(e);return (T)one_third*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3)));}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Tetrahedron_Center(const int e)
{const TV_INT4& v=mesh.tetrahedron_mesh.elements(e);return (T).25*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3))+particles.X(v(4)));}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Element_Center(const int e,const int type)
{
    switch(type){
    case SIMPLEX_PARTICLE:return particles.X(e);
    case SIMPLEX_SEGMENT:return Segment_Center(e);
    case SIMPLEX_TRIANGLE:return Triangle_Center(e);
    case SIMPLEX_TETRAHEDRON:return Tetrahedron_Center(e);
    default:PHYSBAM_FATAL_ERROR("Invalid type in Element_Volume");return TV();
    }
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume_Fraction(const int p,const int s)
{
    int p2=(Other_Nodes_In_Element(mesh.segment_mesh.elements(s),p))(1);
    T s1=pow<2>(particles.thickness(p));T s2=pow<2>(particles.thickness(p2));
    return pow<2>(s1)/(pow<2>(s1)+pow<2>(s2));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume(const int p,const int s)
{
    return (T).5*Segment_Length(s)*Segment_Cross_Section_Area(particles.thickness(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Volume(const int p)
{
    return Nodal_Segment_Length(p)*Segment_Cross_Section_Area(particles.thickness(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume_Fraction(const int p,const int t)
{
    T thickness_sum=particles.thickness(mesh.triangle_mesh.elements(t)(1))+particles.thickness(mesh.triangle_mesh.elements(t)(2))+particles.thickness(mesh.triangle_mesh.elements(t)(3));
    PHYSBAM_ASSERT(thickness_sum!=(T)0);
    return particles.thickness(p)/thickness_sum;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume(const int p,const int t)
{
    return (T)one_third*Triangle_Area(t)*particles.thickness(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Triangle_Volume(const int p)
{
    return Nodal_Triangle_Area(p)*particles.thickness(p);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Volume(const int p,const int t)
{
    return (T).25*Tetrahedron_Volume(t);
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Volume(const int p)
{
    T volume_sum=(T)0;
    for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
        volume_sum+=(T).25*Tetrahedron_Volume(t);
    }
    return volume_sum;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume(const int p)
{
    return Nodal_Segment_Volume(p)+Nodal_Triangle_Volume(p)+Nodal_Tetrahedron_Volume(p);

    //T volume=(T)0;
    //for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
    //    int e=(*mesh.segment_mesh.incident_elements)(p)(i);
    //    volume+=Nodal_Segment_Volume(p,e);
    //}
    //for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
    //    int e=(*mesh.triangle_mesh.incident_elements)(p)(i);
    //    volume+=Nodal_Triangle_Volume(p,e);
    //}
    //for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
    //    int e=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
    //    volume+=Nodal_Tetrahedron_Volume(p,e);
    //}
    //return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Partial_Volume(const int p,const HASHTABLE<int>& psi_D_hashtable)
{
    T volume=(T)0;
    for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
        int e=(*mesh.segment_mesh.incident_elements)(p)(i);
        const TV_INT2& v=mesh.segment_mesh.elements(e);
        bool has_full_ring_psi_D=true;
        for(int i=1;i<=2;i++)if(!psi_D_hashtable.Contains(v(i))){has_full_ring_psi_D=false;break;}
        if(!has_full_ring_psi_D)volume+=Nodal_Segment_Volume(p,e);
    }
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int e=(*mesh.triangle_mesh.incident_elements)(p)(i);
        //const TV_INT3& v=mesh.triangle_mesh.elements(e);
        //bool has_full_ring_psi_D=true;
        //for(int i=1;i<=3;i++)if(!psi_D_hashtable.Contains(v(i))){has_full_ring_psi_D=false;break;}
        bool has_full_ring_psi_D=CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Full_Ring_Dirichlet_Boundary(mesh,psi_D_hashtable,e,SIMPLEX_TRIANGLE);
        if(!has_full_ring_psi_D)volume+=Nodal_Triangle_Volume(p,e);
    }
    for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        int e=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
        //const TV_INT4& v=mesh.tetrahedron_mesh.elements(e);
        //bool has_full_ring_psi_D=true;
        //for(int i=1;i<=4;i++)if(!psi_D_hashtable.Contains(v(i))){has_full_ring_psi_D=false;break;}
        bool has_full_ring_psi_D=CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Full_Ring_Dirichlet_Boundary(mesh,psi_D_hashtable,e,SIMPLEX_TETRAHEDRON);
        if(!has_full_ring_psi_D)volume+=Nodal_Tetrahedron_Volume(p,e);
    }
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_With_Thickness(const int p,const T thickness)
{
    return Nodal_Tetrahedron_Volume(p)+Nodal_Triangle_Area(p)*thickness+Nodal_Segment_Length(p)*Segment_Cross_Section_Area(thickness);
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Dimensional_Volume(const int p)
{
    return TV(Nodal_Segment_Volume(p),Nodal_Triangle_Volume(p),Nodal_Tetrahedron_Volume(p));
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Dimensional_Volume_With_Thickness(const int p,const T thickness)
{
    return TV(Nodal_Segment_Length(p)*Segment_Cross_Section_Area(thickness),Nodal_Triangle_Area(p)*thickness,Nodal_Tetrahedron_Volume(p));
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_In_Highest_Dimension(const int p)
{
    T volume=(T)0;
    for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        int e=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
        volume+=Nodal_Tetrahedron_Volume(p,e);
    }
    return volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Volume_In_Highest_Dimension_Fraction(const int p)
{
    T nodal_volume=Nodal_Volume(p);if(nodal_volume==(T)0)return (T)1;
    T nodal_volume_highest_dim=Nodal_Volume_In_Highest_Dimension(p);
    return nodal_volume_highest_dim/nodal_volume;
}

template<class T_input>
T_input PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Tetrahedron_Surface_Area(const int p)
{
    T total_area=(T)0;
    const ARRAY<int>& v_tris=(*mesh.tetrahedron_mesh.boundary_mesh->incident_elements)(p);
    for(int i=1;i<=v_tris.m;i++){
        const TV_INT3& vtx=(*mesh.tetrahedron_mesh.boundary_mesh).elements(v_tris(i));
        total_area+=(T)one_third*TRIANGLE_3D<T>::Area(particles.X(vtx(1)),particles.X(vtx(2)),particles.X(vtx(3)));
    }
    return total_area;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Area_Weighted_Mean_Curvature_Triangle_Normal(const int p)
{
    TV area_weighted_mean_curvature_normal=TV();
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        VECTOR<int,3> v=Reordered(mesh.triangle_mesh.elements(t),p);
        area_weighted_mean_curvature_normal+=Cotangent(particles.X(v(3)),particles.X(p),particles.X(v(2)))*(particles.X(p)-particles.X(v(2)));
        area_weighted_mean_curvature_normal+=Cotangent(particles.X(v(2)),particles.X(v(3)),particles.X(p))*(particles.X(p)-particles.X(v(3)));
    }
    return (T).5*area_weighted_mean_curvature_normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Mean_Curvature_Normal(const int p)
{
    T area=Nodal_Triangle_Area(p);
    return area==(T)0?TV():(T)3*Nodal_Area_Weighted_Mean_Curvature_Triangle_Normal(p)/area;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Area_Weighted_Tetrahedron_Normal(const int p)
{
    TV normal=TV();
    for(int i=1;i<=(*mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        const int t=(*mesh.tetrahedron_mesh.incident_elements)(p)(i);
        const TV_INT4& v=mesh.tetrahedron_mesh.elements(t);
        TV tet_normal=Face_Area_Weighted_Volumetric_Vertex_Normal(particles.X,v,p);
        normal+=tet_normal;
    }
    return normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Length_Weighted_Triangle_Normal(const int p)
{
    TV normal=TV();
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        const int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        const TV_INT3& v=mesh.triangle_mesh.elements(t);
        TV tri_normal=Length_Weighted_Triangle_Vertex_Normal(particles.X,v,p);
        normal+=tri_normal;
    }
    return normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Segment_Normal(const int p)
{
    TV normal=TV();
    for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
        const int s=(*mesh.segment_mesh.incident_elements)(p)(i);
        const TV_INT2& v=mesh.segment_mesh.elements(s);
        TV seg_normal=Segment_Vertex_Normal(particles.X,v,p);
        normal+=seg_normal;
    }
    return normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Area_Weighted_Normal(const int p)
{
    return Nodal_Area_Weighted_Tetrahedron_Normal(p)+particles.thickness(p)*Nodal_Length_Weighted_Triangle_Normal(p)+Segment_Cross_Section_Area(particles.thickness(p))*Nodal_Segment_Normal(p);
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Area_Weighted_Normal(const int p,const T thickness_scale,const T rim_thickness_scale)
{
    return Nodal_Area_Weighted_Tetrahedron_Normal(p)+thickness_scale*Nodal_Length_Weighted_Triangle_Normal(p)+Segment_Cross_Section_Area(rim_thickness_scale)*Nodal_Segment_Normal(p);
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Face_Area_Weighted_Volumetric_Vertex_Normal(const ARRAY_VIEW<TV>& X,const TV_INT4& element,const int v)
{
    int i,j,k,l;element.Get(i,j,k,l);int node_index=(i==v?1:(j==v?2:(k==v?3:4)));
    MATRIX<T,3> Ds=STRAIN_MEASURE<VECTOR<T,3>,3>::Ds(X,element);MATRIX<T,3> Bs=/*(T)1/FACTORIAL<4>::value**/(T)one_sixth*Ds.Cofactor_Matrix();  //// 1/3*Area
    switch(node_index){
    case 1:return -(Bs.Column(1)+Bs.Column(2)+Bs.Column(3));
    case 2:return Bs.Column(1);
    case 3:return Bs.Column(2);
    case 4:return Bs.Column(3);
    default:PHYSBAM_FATAL_ERROR("invalid node index");return VECTOR<T,3>();}
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const ARRAY_VIEW<TV>& X,const int a,const int b,const int c)
{return Length_Weighted_Triangle_Vertex_Normal(X(a),X(b),X(c));}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const ARRAY_VIEW<TV>& X,const TV_INT3& element,const int a)
{
    TV_INT3 abc=Reordered(element,a);
    return Length_Weighted_Triangle_Vertex_Normal(X,a,abc(2),abc(3));
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const TV& X_a,const TV& X_b,const TV& X_c)
{
    TV ab=X_b-X_a;TV bc=X_c-X_b;TV ca=X_a-X_c;
    typename TV::SPIN N=(TV::Cross_Product(ab,-ca)).Normalized();
    return /*(T)1/FACTORIAL<3>::value**/(T).5*TV::Cross_Product(N,bc);      //// .5*Area
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const ARRAY_VIEW<TV>& X,const ARRAY_VIEW<int>& contact,const int a,const int b,const int c)
{return Length_Weighted_Triangle_Vertex_Normal(Is_Contact_With_Solid(contact(a)),Is_Contact_With_Solid(contact(b)),Is_Contact_With_Solid(contact(c)),X(a),X(b),X(c));}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const bool contact_a,const bool contact_b,const bool contact_c,const TV& X_a,const TV& X_b,const TV& X_c)
{
    TV ab=X_b-X_a;TV bc=X_c-X_b;TV ca=X_a-X_c;
    typename TV::SPIN N=(TV::Cross_Product(ab,-ca)).Normalized();
    TV length_weighted_normal=TV();
    if(!contact_a||!contact_b)length_weighted_normal+=(T).5*(TV::Cross_Product(ab,N));
    if(!contact_a||!contact_c)length_weighted_normal+=(T).5*(TV::Cross_Product(ca,N));
    return length_weighted_normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Length_Weighted_Triangle_Vertex_Normal(const ARRAY_VIEW<TV>& X,const ARRAY_VIEW<int>& contact,const TV_INT3& element,const int a)
{
    TV_INT3 abc=Reordered(element,a);
    return Length_Weighted_Triangle_Vertex_Normal(Is_Contact_With_Solid(contact(a)),Is_Contact_With_Solid(contact(abc(2))),Is_Contact_With_Solid(contact(abc(3))),X(a),X(abc(2)),X(abc(3)));
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Face_Area_Weighted_Volumetric_Vertex_Normal(const ARRAY_VIEW<TV>& X,const ARRAY_VIEW<int>& contact,const TV_INT4& element,const int a)
{   
    TV_INT4 v=Reordered(element,a);
    MATRIX<T,3> Ds=STRAIN_MEASURE<VECTOR<T,3>,3>::Ds(X,v);MATRIX<T,3> Bs=(T)one_sixth*Ds.Cofactor_Matrix();  //// 1/3*Area
    TV area_weighted_normal=TV();
    VECTOR<bool,4> c;for(int i=1;i<=4;i++)c(i)=Is_Contact_With_Solid(contact(v(i)));
    if(!c(1)||!c(3)||!c(4)){area_weighted_normal+=-Bs.Column(1);}
    if(!c(1)||!c(2)||!c(4)){area_weighted_normal+=-Bs.Column(2);}
    if(!c(1)||!c(2)||!c(3)){area_weighted_normal+=-Bs.Column(3);}
    return area_weighted_normal;
}

template<class T_input>
VECTOR<T_input,3> PhysBAM::CODIMENSIONAL_MESH_OBJECT<VECTOR<T_input,3> >::Nodal_Normal(const int p,bool has_triangles/*=true*/)
{
    TV normal=TV();
    for(int i=1;i<=(*mesh.tetrahedron_mesh.boundary_mesh->incident_elements)(p).m;i++){
        int t=(*mesh.tetrahedron_mesh.boundary_mesh->incident_elements)(p)(i);
        const TV_INT3& v=(*mesh.tetrahedron_mesh.boundary_mesh).elements(t);
        TV tri_normal=TV::Cross_Product(particles.X(v(2))-particles.X(v(1)),particles.X(v(3))-particles.X(v(1)));
        normal+=tri_normal;
    }

    if(has_triangles)for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        const TV_INT3& v=mesh.triangle_mesh.elements(t);
        TV tri_normal=TV::Cross_Product(particles.X(v(2))-particles.X(v(1)),particles.X(v(3))-particles.X(v(1)));
        normal+=tri_normal;
    }

    return Normalize_With_Zero(normal);
}

//template class CODIMENSIONAL_MESH_OBJECT<VECTOR<float,2> >;
template class CODIMENSIONAL_MESH_OBJECT<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_MESH_OBJECT<VECTOR<double,2> >;
template class CODIMENSIONAL_MESH_OBJECT<VECTOR<double,3> >;
#endif
