//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_BASE.h"
using namespace PhysBAM;

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Segment_Length(const int e)
{
    return (particles.X(mesh.segment_mesh.elements(e)(1))-particles.X(mesh.segment_mesh.elements(e)(2))).Magnitude();
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Triangle_Area(const int e)
{
    return T_TRIANGLE::Area(particles.X(mesh.triangle_mesh.elements(e)(1)),
        particles.X(mesh.triangle_mesh.elements(e)(2)),particles.X(mesh.triangle_mesh.elements(e)(3)));
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length(const int p)
{
    T length_sum=(T)0;
    for(int i=1;i<=(*mesh.segment_mesh.incident_elements)(p).m;i++){
        int s=(*mesh.segment_mesh.incident_elements)(p)(i);
        length_sum+=(T).5*Segment_Length(s);
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum_On_Triangle_Mesh_Boundary(const int p)
{
    T length_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.boundary_mesh->incident_elements)(p).m;i++){
        int s=(*mesh.triangle_mesh.boundary_mesh->incident_elements)(p)(i);
        VECTOR<int,2>& v=(*mesh.triangle_mesh.boundary_mesh).elements(s);
        length_sum+=(T).5*(particles.X(v(1))-particles.X(v(2))).Magnitude();
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Sum_On_Rim_Mesh(const int p)
{
    T length_sum=(T)0;
    if(mesh.rim_segment_mesh.incident_elements&&(*mesh.rim_segment_mesh.incident_elements)(p).m>0)for(int i=1;i<=(*mesh.rim_segment_mesh.incident_elements)(p).m;i++){
        int s=(*mesh.rim_segment_mesh.incident_elements)(p)(i);
        VECTOR<int,2>& v=mesh.rim_segment_mesh.elements(s);
        length_sum+=(T).5*(particles.X(v(1))-particles.X(v(2))).Magnitude();
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Triangle_Area(const int p)
{
    T area_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        area_sum+=(T)one_third*Triangle_Area(t);
    }
    return area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Extra_Triangle_Area_Sum(const int p)
{
    if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(p))return (T)0;
    T extra_area_sum=(T)0;
    for(int i=1;i<=(*mesh.triangle_mesh.incident_elements)(p).m;i++){
        int t=(*mesh.triangle_mesh.incident_elements)(p)(i);
        VECTOR<int,2> other_nodes=Other_Nodes_In_Element(mesh.triangle_mesh.elements(t),p);
        int rim_p_num=(mesh.Is_Boundary_Particle_On_Triangle_Mesh(other_nodes(1))?1:0)+(mesh.Is_Boundary_Particle_On_Triangle_Mesh(other_nodes(2))?1:0);
        switch(rim_p_num){
        case 1:extra_area_sum+=(T)one_sixth*Triangle_Area(t);break;
        case 2:extra_area_sum+=(T)two_thirds*Triangle_Area(t);break;
        }
    }

    return extra_area_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length(ARRAY_VIEW<TV>& X,const int p,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable)
{
    ARRAY<int> nb_p;T length_sum=(T)0;
    if(rim_vtx_vtx_hashtable.Get(p,nb_p)){
        for(int i=1;i<=nb_p.m;i++){length_sum+=(T).5*(X(p)-X(nb_p(i))).Magnitude();}
    }
    return length_sum;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Length_Fraction(ARRAY_VIEW<TV>& X,const int p,const int p2,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable)
{
    T length_sum=Nodal_Segment_Length(X,p,rim_vtx_vtx_hashtable);
    T seg_length=(T).5*(X(p)-X(p2)).Magnitude();
    return seg_length/length_sum;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Triangle_Degree(const int p)
{
    return (*mesh.triangle_mesh.incident_elements)(p).m;
}

template<class TV>
int PhysBAM::CODIMENSIONAL_MESH_OBJECT_BASE<TV>::Nodal_Segment_Degree(const int p)
{
    return (*mesh.segment_mesh.incident_elements)(p).m;
}

//template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<float,2> >;
template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<double,2> >;
template class CODIMENSIONAL_MESH_OBJECT_BASE<VECTOR<double,3> >;
#endif