//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_MESH_OBJECT_BASE__
#define __CODIMENSIONAL_MESH_OBJECT_BASE__

#include <PhysBAM_Tools/Vectors/VECTOR.h>
#include <PhysBAM_Tools/Math_Tools/pow.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_HELPER.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_MESH_OBJECT_BASE
{
    typedef typename TV::SCALAR T;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TRIANGLE_3D<T> >::TYPE T_TRIANGLE;
public:
    CODIMENSIONAL_MESH& mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;

    CODIMENSIONAL_MESH_OBJECT_BASE():mesh(*(new CODIMENSIONAL_MESH())),particles(*(new CODIMENSIONAL_FLUID_PARTICLES<TV>())){}
    CODIMENSIONAL_MESH_OBJECT_BASE(CODIMENSIONAL_MESH& mesh_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input):mesh(mesh_input),particles(particles_input){}

    ////simplex computations
    T Segment_Length(const int e);
    T Triangle_Area(const int e);

    ////nodal simplex computations
    T Nodal_Segment_Length(const int p);
    T Nodal_Segment_Length_Sum_On_Triangle_Mesh_Boundary(const int p);      ////including segments on triangle mesh boundary only
    T Nodal_Segment_Length_Sum_On_Rim_Mesh(const int p);    ////including segments on both triangle mesh boundary and segment mesh
    int Nodal_Segment_Degree(const int p);

    T Nodal_Triangle_Area(const int p);
    int Nodal_Triangle_Degree(const int p);
    T Nodal_Extra_Triangle_Area_Sum(const int p);

    static T Nodal_Segment_Length(ARRAY_VIEW<TV>& X,const int p,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable);
    static T Nodal_Segment_Length_Fraction(ARRAY_VIEW<TV>& X,const int p,const int p2,const HASHTABLE<int,ARRAY<int> >& rim_vtx_vtx_hashtable);
};
}
   
#endif
