//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################

#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/QUEUE.h>

#include "../Misc/INFINITE_PLANAR_SHEET_BOUNDARY.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::INFINITE_PLANAR_SHEET_BOUNDARY(CODIMENSIONAL_MESH& codimensional_mesh_,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_,CODIMENSIONAL_FLUID_EXAMPLE<TV>& example_,const T& length_scale_,const int num_border_particles_/*=3*/)
    :codimensional_mesh(codimensional_mesh_),particles(particles_),example(example_),length_scale(length_scale),num_border_particles(num_border_particles_),
    has_incompressibility(false),has_viscosity(false),has_elasticity(false)
{
    //no-op
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Initialize()
{
    ARRAY<TV_INT> distance_to_min,distance_to_max;//graph distances

    Get_Distance_To_Edges(distance_to_min,distance_to_max);

    //set ranges--for now, assume the two sides are y_min and y_max
    TV side1_min,side1_max;

    bool min_first=true;
    bool max_first=true;
    for(int i=1;i<=particles.X.m;i++){
        if(distance_to_min(i)(2)==1){
            if(min_first){min_first=false;side1_min=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)side1_min(j)=min(side1_min(j),particles.X(i)(j));
        }
        if(distance_to_min(i)(2)==num_border_particles){
            if(max_first){max_first=false;side1_max=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)side1_max(j)=max(side1_max(j),particles.X(i)(j));
        }
    }

    TV side2_min,side2_max;

    min_first=true;
    max_first=true;
    for(int i=1;i<=particles.X.m;i++){
        if(distance_to_max(i)(2)==1){
            if(max_first){max_first=false;side2_max=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)side2_max(j)=max(side2_max(j),particles.X(i)(j));
        }
        if(distance_to_max(i)(2)==num_border_particles){
            if(min_first){min_first=false;side2_min=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)side2_min(j)=min(side2_min(j),particles.X(i)(j));
        }
    }

    TV back_min,back_max;

    min_first=true;
    max_first=true;
    for(int i=1;i<=particles.X.m;i++){
        if(distance_to_max(i)(1)==1){
            if(max_first){max_first=false;back_max=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)back_max(j)=max(back_max(j),particles.X(i)(j));
        }
        if(distance_to_max(i)(1)==num_border_particles){
            if(min_first){min_first=false;back_min=particles.X(i);}
            else for(int j=1;j<=TV::m;j++)back_min(j)=min(back_min(j),particles.X(i)(j));
        }
    }

    //some room for error
    side1_min(1)-=(T)0.1*length_scale;
    side1_min(2)-=(T)0.1*length_scale;
    side1_min(3)-=(T)0.1*length_scale;
    side1_max(1)+=(T)0.1*length_scale;
    side1_max(3)+=(T)0.1*length_scale;

    side2_min(1)-=(T)0.1*length_scale;
    side2_min(3)-=(T)0.1*length_scale;
    side2_max(1)+=(T)0.1*length_scale;
    side2_max(2)+=(T)0.1*length_scale;
    side2_max(3)+=(T)0.1*length_scale;

    //back_min(1)-=(T)0.1*length_scale;
    back_min(2)-=(T)0.1*length_scale;
    back_min(3)-=(T)0.1*length_scale;
    back_max(1)+=(T)0.1*length_scale;
    back_max(2)+=(T)0.1*length_scale;
    back_max(3)+=(T)0.1*length_scale;

    range_side1.min_corner=side1_min;
    range_side1.max_corner=side1_max;

    range_side2.min_corner=side2_min;
    range_side2.max_corner=side2_max;

    range_back.min_corner=back_min;
    range_back.max_corner=back_max;

    std::cout<<"range_side1: "<<range_side1.min_corner<<" to "<<range_side1.max_corner<<std::endl;
    std::cout<<"range_side2: "<<range_side2.min_corner<<" to "<<range_side2.max_corner<<std::endl;
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Get_Distance_To_Edges(/*result*/ARRAY<TV_INT>& distance_to_min,/*result*/ARRAY<TV_INT>& distance_to_max)
{
    TV min_position=particles.X(1),max_position=particles.X(1);
    for(int i=2;i<=particles.X.m;i++){
        for(int j=1;j<=TV::m;j++){
            min_position(j)=min(min_position(j),particles.X(i)(j));
            max_position(j)=max(max_position(j),particles.X(i)(j));
        }
    }

    T epsilon=(T)1e-6*length_scale;

    for(int i=1;i<=particles.X.m;i++){
        distance_to_min.Append(TV_INT::Constant_Vector(-1));
        distance_to_max.Append(TV_INT::Constant_Vector(-1));
    }
    for(int i=1;i<=particles.X.m;i++){
        for(int j=1;j<=TV::m;j++){
            if(abs(particles.X(i)(j)-min_position(j))<epsilon)distance_to_min(i)(j)=1;
            if(abs(particles.X(i)(j)-max_position(j))<epsilon)distance_to_max(i)(j)=1;
        }
    }

    TRIANGLE_MESH& triangle_mesh=codimensional_mesh.triangle_mesh;
    triangle_mesh.Initialize_Neighbor_Nodes();

    for(int outer=1;outer<=2;outer++){
        ARRAY<TV_INT>& distance = (outer==1?distance_to_min:distance_to_max);
        for(int d=1;d<=TV::m;d++){
            QUEUE<int> front(particles.X.m);
            ARRAY<bool> visited;
            for(int i=1;i<=particles.X.m;i++)visited.Append(false);
            for(int i=1;i<=particles.X.m;i++){
                if(distance(i)(d)==1){
                    front.Enqueue(i);visited(i)=true;
                }
            }

            while(!front.Empty()){
                int start=front.Dequeue();
                ARRAY<int>& neighbors = (*triangle_mesh.neighbor_nodes)(start);
                for(int i=1;i<=neighbors.m;i++){
                    int n=neighbors(i);
                    if(visited(n))continue;
                    visited(n)=true;
                    distance(n)(d)=distance(start)(d)+1;
                    front.Enqueue(n);
                }
            }
        }
    }
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Get_Distance_To_Edges_Along_Boundary(/*result*/HASHTABLE<int,TV_INT>& distance_to_min,/*result*/HASHTABLE<int,TV_INT>& distance_to_max)
{
    TRIANGLE_MESH& triangle_mesh=codimensional_mesh.triangle_mesh;
    triangle_mesh.Initialize_Boundary_Mesh();
    SEGMENT_MESH& boundary_mesh=*triangle_mesh.boundary_mesh;
    boundary_mesh.Initialize_Neighbor_Nodes();

    ARRAY<int> boundary_particles;
    for(int i=1;i<=boundary_mesh.elements.m;i++){
        boundary_particles.Append_Unique(boundary_mesh.elements(i)(1));
        boundary_particles.Append_Unique(boundary_mesh.elements(i)(2));
    }

    TV min_position=particles.X(boundary_particles(1)),max_position=particles.X(boundary_particles(1));
    for(int i=2;i<=boundary_particles.m;i++){
        for(int j=1;j<=TV::m;j++){
            min_position(j)=min(min_position(j),particles.X(boundary_particles(i))(j));
            max_position(j)=max(max_position(j),particles.X(boundary_particles(i))(j));
        }
    }

    T epsilon=(T)1e-6*length_scale;

    for(int i=1;i<=boundary_particles.m;i++){
        int p=boundary_particles(i);
        distance_to_min.Set(p,TV_INT::Constant_Vector(-1));
        distance_to_max.Set(p,TV_INT::Constant_Vector(-1));
        for(int j=1;j<=TV::m;j++){
            if(abs(particles.X(p)(j)-min_position(j))<epsilon){
                TV_INT& dist=distance_to_min.Get(p);
                dist(j)=1;
            }
            if(abs(particles.X(p)(j)-max_position(j))<epsilon){
                TV_INT& dist=distance_to_max.Get(p);
                dist(j)=1;
            }
        }
    }

    for(int i=1;i<=2;i++){
        HASHTABLE<int,TV_INT>& distance=(i==1?distance_to_min:distance_to_max);
        for(int d=1;d<=TV::m;d++){
            QUEUE<int> front(boundary_particles.m);
            HASHTABLE<int> visited;
            for(int j=1;j<=boundary_particles.m;j++){
                int p=boundary_particles(j);
                if(distance.Get(p)(d)==1){
                    front.Enqueue(p);
                    visited.Set(p);
                }
            }

            while(!front.Empty()){
                int start=front.Dequeue();
                ARRAY<int>& neighbors = (*boundary_mesh.neighbor_nodes)(start);
                for(int j=1;j<=neighbors.m;j++){
                    int n=neighbors(j);
                    if(visited.Contains(n))continue;
                    visited.Set(n);
                    distance.Get(n)(d)=distance.Get(start)(d)+1;
                    front.Enqueue(n);
                }
            }
        }
    }
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Apply_Boundary_Conditions()
{
    HASHTABLE<int,TV_INT> distance_to_min,distance_to_max;
    Get_Distance_To_Edges_Along_Boundary(distance_to_min,distance_to_max);
    for(int i=1;i<=particles.X.m;i++){
        if(distance_to_min.Contains(i)&&distance_to_min.Get(i)(2)<=num_border_particles){
            Clear_Forces(i);
        }
        else if(distance_to_max.Contains(i)&&distance_to_max.Get(i)(2)<=num_border_particles){
            Clear_Forces(i);
        }
        else if(range_back.Lazy_Inside(particles.X(i))){
            Clear_Forces(i);
        }
    }
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Clear_Forces(int i)
{
    particles.F(i)=TV();
    particles.pressure_force(i)=TV();
    particles.surface_tension_force(i)=TV();
    particles.viscosity_force(i)=TV();
    particles.elasticity_force(i)=TV();
    particles.body_force(i)=TV();
    particles.rim_surface_tension_force(i)=TV();
    particles.film_surface_tension_force(i)=TV();

    if(has_incompressibility){
        example.incompressible.psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(1),(T)0));
        example.incompressible.psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(2),(T)0));
        example.incompressible.psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(3),(T)0));
    }
}

template<class TV>
void PhysBAM::INFINITE_PLANAR_SHEET_BOUNDARY<TV>::Update_Boundary_Region()
{
    TRIANGLE_MESH& triangle_mesh=codimensional_mesh.triangle_mesh;

    HASHTABLE<int,TV_INT> distance_to_min,distance_to_max;
    Get_Distance_To_Edges_Along_Boundary(distance_to_min,distance_to_max);

    SEGMENT_MESH& boundary=*triangle_mesh.boundary_mesh;

    T average=(T)0;;
    int count=0;
    ARRAY<int> boundary_particles;
    for(int i=1;i<=boundary.elements.m;i++){
        boundary_particles.Append_Unique(boundary.elements(i)(1));
        boundary_particles.Append_Unique(boundary.elements(i)(2));
    }

    for(int i=1;i<=boundary_particles.m;i++){
        int p=boundary_particles(i);
        if(distance_to_min.Get(p)(2)<=num_border_particles)continue;
        if(distance_to_max.Get(p)(2)<=num_border_particles)continue;
        if(range_back.Lazy_Inside(particles.X(p)))continue;

        average+=particles.X(p)(1);
        count++;
    }
    average/=(T)count;

    int start_distance=num_border_particles+1;
    if(start_distance!=-1){
        int distance=start_distance;
        while(distance>1){
            for(int i=1;i<=boundary.elements.m;i++){
                const TV_INT2& segment=boundary.elements(i);

                if(distance_to_min.Get(segment(1))(2)==distance_to_min.Get(segment(2))(2))continue;
                if(range_back.Lazy_Inside(particles.X(segment(1)))||range_back.Lazy_Inside(particles.X(segment(2))))continue;


                int p1=(distance_to_min.Get(segment(1))(2)>distance_to_min.Get(segment(2))(2)?segment(1):segment(2));
                int p2=(distance_to_min.Get(segment(1))(2)>distance_to_min.Get(segment(2))(2)?segment(2):segment(1));

                if(distance_to_min.Get(p1)(2)==distance){
                    particles.X(p2)(1)=average;
                }
            }
            distance--;
        }
    }

    if(start_distance!=-1){
        int distance=start_distance;
        while(distance>1){
            for(int i=1;i<=boundary.elements.m;i++){
                const TV_INT2& segment=boundary.elements(i);

                if(distance_to_max.Get(segment(1))(2)==distance_to_max.Get(segment(2))(2))continue;
                if(range_back.Lazy_Inside(particles.X(segment(1)))||range_back.Lazy_Inside(particles.X(segment(2))))continue;

                int p1=(distance_to_max.Get(segment(1))(2)>distance_to_max.Get(segment(2))(2)?segment(1):segment(2));
                int p2=(distance_to_max.Get(segment(1))(2)>distance_to_max.Get(segment(2))(2)?segment(2):segment(1));

                if(distance_to_max.Get(p1)(2)==distance){
                    particles.X(p2)(1)=average;
                }
            }
            distance--;
        }
    }
}

//template class INFINITE_PLANAR_SHEET_BOUNDARY<VECTOR<float, 2> >;
template class INFINITE_PLANAR_SHEET_BOUNDARY<VECTOR<float, 3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class INFINITE_PLANAR_SHEET_BOUNDARY<VECTOR<double, 2> >;
template class INFINITE_PLANAR_SHEET_BOUNDARY<VECTOR<double, 3> >;
#endif
