//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __INFINITE_PLANAR_SHEET_BOUNDARY__
#define __INFINITE_PLANAR_SHEET_BOUNDARY__

#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
//#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
//#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Tools/Math_Tools/RANGE.h>

#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Solver/CODIMENSIONAL_FLUID_EXAMPLE.h"

namespace PhysBAM
{
template<class TV>
class INFINITE_PLANAR_SHEET_BOUNDARY
{
    typedef typename TV::SCALAR T;
    typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<int,2> TV_INT2;
    typedef VECTOR<int,3> TV_INT3;
public:
    //triangle mesh is assumed to be in x-y plane
    INFINITE_PLANAR_SHEET_BOUNDARY(CODIMENSIONAL_MESH& codimensional_mesh_,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_,CODIMENSIONAL_FLUID_EXAMPLE<TV>& example_,const T& length_scale_,const int num_border_particles_=3);
    void Initialize();
    void Get_Distance_To_Edges(/*result*/ARRAY<TV_INT>& distance_to_min,/*result*/ARRAY<TV_INT>& distance_to_max);
    void Get_Distance_To_Edges_Along_Boundary(/*result*/HASHTABLE<int,TV_INT>& distance_to_min,/*result*/HASHTABLE<int,TV_INT>& distance_to_max);
    void Apply_Boundary_Conditions();
    void Clear_Forces(int i);
    void Update_Boundary_Region();


    CODIMENSIONAL_MESH& codimensional_mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    CODIMENSIONAL_FLUID_EXAMPLE<TV>& example;
    T length_scale;
    int num_border_particles;

    RANGE<TV> range_side1,range_side2,range_back;

    bool has_incompressibility;
    bool has_viscosity;
    bool has_elasticity;
    
};
}

#endif //__INFINITE_PLANAR_SHEET_BOUNDARY__
