//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __WATER_ANNULAR_GAP_SOURCE__
#define __WATER_ANNULAR_GAP_SOURCE__

#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>

#include "../Misc/WATER_SOURCE.h"

namespace PhysBAM
{

template<class TV>
class WATER_ANNULAR_GAP_SOURCE: public WATER_SOURCE<TV>
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<T,3> TV3;
    typedef WATER_SOURCE<TV> BASE;
public:
    using BASE::random;
    using BASE::endpoint1;
    using BASE::endpoint2;
    using BASE::direction;
    using BASE::length_scale;
    using BASE::thickness;
    using BASE::perturb_coef;
    using BASE::initial_speed;
    using BASE::parity;
    using BASE::eulerian_particles;
    using BASE::Source_Velocity;

    T inner_radius;
    T outer_radius;

    CYLINDER<T> inner_cylinder;
    CYLINDER<T> outer_cylinder;
    CYLINDER<T> fixed_cylinder;

    WATER_ANNULAR_GAP_SOURCE(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const TV& closed_end,const TV& open_end,const T& inner_radius_,const T& outer_radius_,const T& length_scale_,const T& initial_speed_);

    virtual int Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    virtual int Reseed_Source_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    virtual bool Particle_Is_Inside_Source(const TV& location) const;
    virtual bool Particle_Is_Fixed(const TV& location) const;
    virtual bool Reseed_Necessary(const ARRAY_VIEW<TV>& positions) const;
    virtual RANGE<TV> Bounding_Box() const;
};
}


#endif //__WATER_ANNULAR_GAP_SOURCE__
