//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __WATER_CYLINDRICAL_SOURCE__
#define __WATER_CYLINDRICAL_SOURCE__

#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>

#include "../Misc/WATER_SOURCE.h"

namespace PhysBAM
{

template<class TV_input>
class WATER_CYLINDRICAL_SOURCE: public WATER_SOURCE<TV_input>//public CYLINDER<typename TV_input::SCALAR>
{
    typedef TV_input TV;
    typedef typename TV::SCALAR T;
    typedef VECTOR<T,3> TV3;
    typedef WATER_SOURCE<TV> BASE;
    typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef typename TV3::template REBIND<int>::TYPE TV3_INT;
public:

    using BASE::random;
    using BASE::endpoint1;
    using BASE::endpoint2;
    using BASE::direction;
    using BASE::length_scale;
    using BASE::thickness;
    using BASE::perturb_coef;
    using BASE::initial_speed;
    using BASE::parity;
    using BASE::eulerian_particles;
    using BASE::Source_Velocity;

    CYLINDER<T> primary_cylinder;
    CYLINDER<T> boundary_cylinder;
    CYLINDER<T> closed_end;
    CYLINDER<T> closed_end_strict;

    WATER_CYLINDRICAL_SOURCE(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const TV& closed_end_input,const TV& open_end_input,const T& radius_input,const T& length_scale_input,const T& initial_speed_input);

    virtual int Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    virtual int Reseed_Source_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);
    virtual bool Particle_Is_Inside_Source(const TV& location) const;
    virtual bool Particle_Is_Fixed(const TV& location) const;
    virtual bool Reseed_Necessary(const ARRAY_VIEW<TV>& positions) const;
    virtual RANGE<TV> Bounding_Box() const;
};
}

#endif //__WATER_CYLINDRICAL_SOURCE__
