//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __WATER_RING_SOURCE__
#define __WATER_RING_SOURCE__

#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Misc/WATER_SOURCE.h"

namespace PhysBAM
{
template<class TV_input>
class WATER_RING_SOURCE:public WATER_SOURCE<TV_input>
{
    typedef TV_input TV;typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<T,2> TV2;typedef VECTOR<int,3> TV_INT3;
    typedef WATER_SOURCE<TV> BASE;
public:
    using BASE::random;
    using BASE::length_scale;
    using BASE::thickness;  ////average source thickness
    using BASE::perturb_coef;
    using BASE::initial_speed;
    using BASE::eulerian_particles;
    using BASE::eulerian_particle_thicknesses;
    using BASE::initial_particle_thicknesses;

    TV center;
    T radius;
    int particle_num_per_ring;
    T theta_per_particle;

    enum SOURCE_TYPE{FISHBONE=1,POLYGON,CIRCLE} source_type;

    ////parameters for impinging jets
    T Q;
    T sheet_speed;
    T incoming_jets_angle;
    T miller_coef_a;
    T miller_coef_b;
    int polygon_n;

    WATER_RING_SOURCE(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,CODIMENSIONAL_MESH& mesh,const TV center_input,const T radius_input,const T length_scale_input,const int source_type_input,const T Q_input,const T sheet_speed_input,const T incoming_jets_angle_input,const int polygon_n_input=0);

    virtual TV Source_Velocity(const TV& pos=TV());
    virtual int Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles){PHYSBAM_NOT_IMPLEMENTED();}
    virtual int Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,CODIMENSIONAL_MESH& mesh);
    virtual int Reseed_Source_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles){PHYSBAM_NOT_IMPLEMENTED();}
    virtual bool Particle_Is_Inside_Source(const TV& location) const {PHYSBAM_NOT_IMPLEMENTED();}
    virtual bool Particle_Is_Fixed(const TV& location) const {return false;}
    virtual bool Reseed_Necessary(const ARRAY_VIEW<TV>& positions) const {PHYSBAM_NOT_IMPLEMENTED();}
    virtual RANGE<TV> Bounding_Box() const {PHYSBAM_NOT_IMPLEMENTED();}

    void Initialize_Parameters();
    T Flux_Per_Radian_Fishbone(const T theta);  ////Miller 1960
    T Flux_Per_Radian_Polygon(const T theta);
    T Theta(const TV& pos);
    T Thickness(const T theta);
};
}

#endif
