//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Math_Tools/pow.h>
#include "../Meshing/CODIMENSIONAL_HELPER.h"
#include "../Misc/WATER_SOURCE.h"

using namespace PhysBAM;
//using namespace CODIMENSIONAL;

template<class TV>
RANDOM_NUMBERS<typename TV::SCALAR> PhysBAM::WATER_SOURCE<TV>::random;

template<class TV>
PhysBAM::WATER_SOURCE<TV>::WATER_SOURCE(const TV& closed_end,const TV& open_end,const T& length_scale_,const T& thickness_,const T& perturb_coef_,const T& initial_speed_)
    :endpoint1(closed_end),endpoint2(open_end),/*direction(TV::Axis_Vector(1)),*/length_scale(length_scale_),thickness(thickness_),perturb_coef(perturb_coef_),initial_speed(initial_speed_),parity(false),
    on(true)
{
    random.Set_Seed(123910185);

    VECTOR<T,3> e1(endpoint1),e2(endpoint2);
    VECTOR<T,3> direction3;
    if(e1==e2)direction3=VECTOR<T,3>();
    else direction3=(e2-e1).Normalized();

    direction=Codimensional::TV_Compatible<T,TV::m>(direction3);
}

template<class TV>
TV PhysBAM::WATER_SOURCE<TV>::Source_Velocity(const TV& pos/*=TV()*/)
{
    return initial_speed*direction;
}

template<class TV>
TV PhysBAM::WATER_SOURCE<TV>::Source_Flux(const TV& pos/*=TV()*/)
{
    return initial_speed*(T)pi*pow<2>((T).5*thickness)*direction;   ////TODO: thickness is the pipe radius?
}

template<class TV>
bool PhysBAM::WATER_SOURCE<TV>::Is_Eulerian_Particle(int guid)
{
    return eulerian_particles.Contains(guid);
}

template<class TV>
void PhysBAM::WATER_SOURCE<TV>::Update_Eulerian_Particle_Mass(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            TV dim_mass=particles.dimensional_mass(i);
            eulerian_particle_mass.Set(guid,dim_mass);
        }
    }
}

template<class TV>
void PhysBAM::WATER_SOURCE<TV>::Update_Eulerian_Particle_Thickness(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            T thickness=particles.thickness(i);
            eulerian_particle_thicknesses.Set(guid,thickness);
        }
    }
}

template<class TV>
void PhysBAM::WATER_SOURCE<TV>::Update_Eulerian_Particle_Velocities()
{
    for(HASHTABLE_ITERATOR<int,TV> iter(eulerian_particles);iter.Valid();iter.Next()){
        int guid=iter.Key();
        TV pos=iter.Data();
        TV v=Source_Velocity(pos);
        eulerian_particle_velocities.Set(guid,v);
    }
}

template class WATER_SOURCE<VECTOR<float,2> >;
template class WATER_SOURCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class WATER_SOURCE<VECTOR<double,2> >;
template class WATER_SOURCE<VECTOR<double,3> >;
#endif
