//#####################################################################
// Copyright 2013, Bo Zhu, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __WATER_SOURCE__
#define __WATER_SOURCE__

#include <PhysBAM_Geometry/Basic_Geometry/CYLINDER.h>
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include <PhysBAM_Tools/Data_Structures/PAIR.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Vectors/VECTOR.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>

namespace PhysBAM{

template<class TV>
class CODIMENSIONAL_FLUID_PARTICLES;

template<class TV>
class WATER_SOURCE
{
    typedef typename TV::SCALAR T;
public:
    static RANDOM_NUMBERS<T> random;
    TV endpoint1;
    TV endpoint2;
    TV direction;
    T length_scale;
    T thickness;
    T perturb_coef;
    T initial_speed;
    bool parity;
    TV kinematic_velocity;
    bool on;

    HASHTABLE<int,TV> eulerian_particles;//map of guid -> fixed position
    HASHTABLE<int,TV> eulerian_particle_mass;    //map of guid -> fixed mass
    HASHTABLE<int,TV> eulerian_particle_velocities; //map of guid -> fixed velocity
    HASHTABLE<int,T> eulerian_particle_thicknesses; //map of guid -> fixed thickness
    HASHTABLE<int,T> initial_particle_thicknesses;   //only used for initializing mass

    WATER_SOURCE(const TV& closed_end,const TV& open_end,const T& length_scale_,const T& thickness_,const T& perturb_coef_,const T& initial_speed_);

    virtual TV Source_Velocity(const TV& pos=TV());
    virtual TV Source_Flux(const TV& pos=TV());
    virtual bool Is_Eulerian_Particle(int guid);
    virtual int Seed_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles) = 0;
    virtual int Reseed_Source_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles) = 0;
    virtual bool Particle_Is_Inside_Source(const TV& location) const = 0;
    virtual bool Particle_Is_Fixed(const TV& location) const = 0;
    virtual bool Reseed_Necessary(const ARRAY_VIEW<TV>& positions) const = 0;
    virtual RANGE<TV> Bounding_Box() const = 0;

    void Update_Eulerian_Particle_Mass(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);   ////update eulerian particles' mass from particles.mass
    void Update_Eulerian_Particle_Thickness(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles);  ////update eulerian particles' thickness from particles.mass
    void Update_Eulerian_Particle_Velocities(); ////update eulerian particle velocities from Source_Velocity
};
}

#endif //__WATER_SOURCE__
