//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Meshing/CODIMENSIONAL_HELPER.h"
#include "../Meshing/CODIMENSIONAL_MESH_CHECKER.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::MPI_CODIMENSIONAL<TV>::MPI_CODIMENSIONAL(MPI_UNIFORM_GRID<GRID<TV> >& mpi_grid_input,const int number_of_ghost_cells_input,
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,CODIMENSIONAL_MESH& mesh_input)
    :mpi_grid(mpi_grid_input),number_of_ghost_cells(number_of_ghost_cells_input),region_count_offset(1000000),particles(particles_input),mesh(mesh_input)
{
    domain=mpi_grid.local_grid.Domain();
    domain_ext=mpi_grid.local_grid.Ghost_Domain(number_of_ghost_cells);
    domain_shrink=mpi_grid.local_grid.Ghost_Domain(-number_of_ghost_cells);

    node_id=mpi_grid.rank+1;
    first_region_number=1+node_id*region_count_offset;
    dX=mpi_grid.local_grid.dX.Max();
    neighbor_ranks=mpi_grid.all_neighbor_ranks;
    neighbor_directions=mpi_grid.all_neighbor_directions;
    ARRAY<RANGE<TV_INT> > boundary_region_indices;
    RANGE<TV_INT> sentinels=RANGE<TV_INT>::Zero_Box();
    mpi_grid.Find_Boundary_Regions(boundary_region_indices,sentinels,false,RANGE<VECTOR<int,1> >(0,number_of_ghost_cells-1),true,false,mpi_grid.local_grid);
    boundary_regions.Resize(boundary_region_indices.m);
    for(int i=1;i<=boundary_region_indices.m;i++){
        boundary_regions(i)=RANGE<TV>(mpi_grid.local_grid.Node(boundary_region_indices(i).min_corner),mpi_grid.local_grid.Node(boundary_region_indices(i).max_corner+1));
    }
    Initialize_Boundary_Shrink_Regions(boundary_regions,boundary_shrink_regions);

    particles.Initialize_Mpi_Guid_Seed(node_id);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Initialize_Boundary_Shrink_Regions(const ARRAY<RANGE<TV> >& boundary_regions,ARRAY<RANGE<TV> >& boundary_shrink_regions)
{
    boundary_shrink_regions.Resize(boundary_regions.m);
    for(int i=1;i<=boundary_regions.m;i++)boundary_shrink_regions(i)=boundary_regions(i);
    ////corners: 1, 3, 7, 9, 18, 20, 24, 26, no change

    ////edges: 2, 4, 6, 8, 10, 12, 15, 17, 19, 21, 23, 25
    Shrink_Edge_Region_Helper(boundary_shrink_regions(2),boundary_regions(1),boundary_regions(3),1);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(4),boundary_regions(1),boundary_regions(7),2);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(6),boundary_regions(3),boundary_regions(9),2);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(8),boundary_regions(7),boundary_regions(9),1);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(10),boundary_regions(1),boundary_regions(18),3);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(12),boundary_regions(3),boundary_regions(20),3);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(15),boundary_regions(7),boundary_regions(24),3);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(17),boundary_regions(9),boundary_regions(26),3);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(19),boundary_regions(18),boundary_regions(20),1);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(21),boundary_regions(18),boundary_regions(24),2);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(23),boundary_regions(20),boundary_regions(26),2);
    Shrink_Edge_Region_Helper(boundary_shrink_regions(25),boundary_regions(24),boundary_regions(26),1);

    ////faces: 5, 11, 13, 14, 16, 22
    Shrink_Face_Region_Helper(boundary_shrink_regions(5),boundary_regions(1),boundary_regions(9),1,2);
    Shrink_Face_Region_Helper(boundary_shrink_regions(11),boundary_regions(1),boundary_regions(20),1,3);
    Shrink_Face_Region_Helper(boundary_shrink_regions(13),boundary_regions(1),boundary_regions(24),2,3);
    Shrink_Face_Region_Helper(boundary_shrink_regions(14),boundary_regions(3),boundary_regions(26),2,3);
    Shrink_Face_Region_Helper(boundary_shrink_regions(16),boundary_regions(7),boundary_regions(26),1,3);
    Shrink_Face_Region_Helper(boundary_shrink_regions(22),boundary_regions(18),boundary_regions(26),1,2);

    boundary_shrink_to_boundary_maps.Resize(boundary_regions.m);
    ////corners: 1, 3, 7, 9, 18, 20, 24, 26
    {int nbs[]={1,2,4,10,5,11,13};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(1).Append(nbs[i]);}
    {int nbs[]={3,2,6,12,5,11,14};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(3).Append(nbs[i]);}
    {int nbs[]={7,4,8,15,5,13,16};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(7).Append(nbs[i]);}
    {int nbs[]={9,6,8,17,5,14,16};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(9).Append(nbs[i]);}
    {int nbs[]={18,10,19,21,11,13,22};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(18).Append(nbs[i]);}
    {int nbs[]={20,12,19,23,11,14,22};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(20).Append(nbs[i]);}
    {int nbs[]={24,15,21,25,13,16,22};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(24).Append(nbs[i]);}
    {int nbs[]={26,17,23,25,14,16,22};for(int i=0;i<7;i++)boundary_shrink_to_boundary_maps(26).Append(nbs[i]);}

    ////edges: 2, 4, 6, 8, 10, 12, 15, 17, 19, 21, 23, 25
    {int nbs[]={2,1,3,5,11};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(2).Append(nbs[i]);}
    {int nbs[]={4,1,7,5,13};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(4).Append(nbs[i]);}
    {int nbs[]={6,3,9,5,14};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(6).Append(nbs[i]);}
    {int nbs[]={8,7,9,5,16};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(8).Append(nbs[i]);}
    {int nbs[]={10,1,18,11,13};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(10).Append(nbs[i]);}
    {int nbs[]={12,3,20,11,14};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(12).Append(nbs[i]);}
    {int nbs[]={15,7,24,13,16};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(15).Append(nbs[i]);}
    {int nbs[]={17,9,26,14,16};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(17).Append(nbs[i]);}
    {int nbs[]={19,18,20,11,22};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(19).Append(nbs[i]);}
    {int nbs[]={21,18,24,13,22};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(21).Append(nbs[i]);}
    {int nbs[]={23,20,26,14,22};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(23).Append(nbs[i]);}
    {int nbs[]={25,24,26,16,22};for(int i=0;i<5;i++)boundary_shrink_to_boundary_maps(25).Append(nbs[i]);}

    ////faces: 5, 11, 13, 14, 16, 22
    boundary_shrink_to_boundary_maps(5).Append(5);
    boundary_shrink_to_boundary_maps(11).Append(11);
    boundary_shrink_to_boundary_maps(13).Append(13);
    boundary_shrink_to_boundary_maps(14).Append(14);
    boundary_shrink_to_boundary_maps(16).Append(16);
    boundary_shrink_to_boundary_maps(22).Append(22);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Shrink_Edge_Region_Helper(RANGE<TV>& shrink_region,const RANGE<TV>& min_region,const RANGE<TV>& max_region,const int axis)
{
    shrink_region.min_corner(axis)=min_region.max_corner(axis);
    shrink_region.max_corner(axis)=max_region.min_corner(axis);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Shrink_Face_Region_Helper(RANGE<TV>& shrink_region,const RANGE<TV>& min_region,const RANGE<TV>& max_region,const int axis_1,const int axis_2)
{
    shrink_region.min_corner(axis_1)=min_region.max_corner(axis_1);
    shrink_region.min_corner(axis_2)=min_region.max_corner(axis_2);
    shrink_region.max_corner(axis_1)=max_region.min_corner(axis_1);
    shrink_region.max_corner(axis_2)=max_region.min_corner(axis_2);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Initialize_From_External_Mesh_And_Positions(const CODIMENSIONAL_MESH& ext_mesh,const ARRAY_VIEW<TV>& ext_X)
{
    ARRAY<int> ext_guid(ext_X.m);ext_guid.Fill(0);
    ////generate the same guid for all mpi nodes
    //int node_num=mpi_grid.number_of_processes;
    int global_guid_start=0;
    for(int i=1;i<=ext_guid.m;i++){ext_guid(i)=global_guid_start+i;}

    ////add particles inside domain_ext
    ARRAY<int> ghost_particle_indices;
    for(int i=1;i<=ext_X.m;i++){
        if(Inside_Domain(ext_X(i))){
            int index=particles.array_collection->Add_Element();
            particles.X(index)=ext_X(i);
            particles.guid(index)=ext_guid(i);
        }
        else if(Inside_Domain_Ext(ext_X(i))){
            ghost_particle_indices.Append(i);
        }
    }
    for(int i=1;i<=ghost_particle_indices.m;i++){
        int p=ghost_particle_indices(i);
        int index=particles.array_collection->Add_Element();
        particles.X(index)=ext_X(p);
        particles.guid(index)=ext_guid(p);
    }

    for(int i=1;i<=particles.X.m;i++){
        guid_vertex_hashtable.Set(particles.guid(i),i);
    }

    Initialize_Mesh_Elements_From_Global_External_Mesh_Helper(mesh.tetrahedron_mesh.elements,ext_mesh.tetrahedron_mesh.elements,ext_guid);
    Initialize_Mesh_Elements_From_Global_External_Mesh_Helper(mesh.triangle_mesh.elements,ext_mesh.triangle_mesh.elements,ext_guid);
    Initialize_Mesh_Elements_From_Global_External_Mesh_Helper(mesh.segment_mesh.elements,ext_mesh.segment_mesh.elements,ext_guid);

    Check_Invalid_Vertex(mesh.tetrahedron_mesh.elements,1,particles.X.m);
    Check_Invalid_Vertex(mesh.triangle_mesh.elements,1,particles.X.m);
    Check_Invalid_Vertex(mesh.segment_mesh.elements,1,particles.X.m);

    Update_Boundary_And_Ghost_Particle_Indices();
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Initialize_Mesh_Elements_From_Global_External_Mesh_Helper(ARRAY<TV_INTN>& elements,const ARRAY<TV_INTN>& external_elements,const ARRAY<int>& guid)
{
    elements.Remove_All();
    for(int i=1;i<=external_elements.m;i++){
        const TV_INTN& external_e=external_elements(i);
        TV_INTN e_guid=Element(guid,external_e);
        TV_INTN e=Guid_To_Vtx(e_guid);
        if(!e.Contains(0))elements.Append(e);
    }
}

template<class TV>
int PhysBAM::MPI_CODIMENSIONAL<TV>::Boundary_Shrink_Region_Index(const TV& pos) const
{
    TV_INT neighbor_offset=Neighbor_Offset(pos,domain_shrink);
    return Boundary_Shrink_Region_Index(neighbor_offset);
}

template<class TV>
int PhysBAM::MPI_CODIMENSIONAL<TV>::Boundary_Shrink_Region_Index(const TV_INT& neighbor_offset) const
{
    if(neighbor_offset!=TV_INT()){
        return GRID<TV>::One_Ring_Neighbor_Index(neighbor_offset);
    }
    else return 0;
}

template<class TV>
RANGE<TV> PhysBAM::MPI_CODIMENSIONAL<TV>::Boundary_Shrink_Region(const TV_INT& nb_offset) const
{
    if(nb_offset==TV_INT())return domain_shrink;
    else return boundary_shrink_regions(GRID<TV>::One_Ring_Neighbor_Index(nb_offset));
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_And_Append_Particle_Indices(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,
    ARRAY<int>& send_p_numbers,ARRAY<RANGE_INT1>& recv_p_ranges,ARRAY<int>& recv_p_numbers)
{
    ARRAY<MPI::Request> requests;
    ////find particles to send
    Clear_Arrays(send_p);Clear_Values(send_p_numbers);
    for(int i=1;i<=particles.X.m;i++){int p=i;
        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(particles.X(p));
        if(boundary_shrink_region_index!=0){
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                send_p(boundary_region_indices(j)).Append(p);
            }
        }
    }
    for(int i=1;i<=send_p.m;i++)send_p_numbers(i)=send_p(i).m;

    ////send particle numbers
    Mpi_Send_Values(send_p_numbers,requests);

    ////receive particle numbers
    Clear_Values(recv_p_numbers);
    Mpi_Recv_Values(recv_p_numbers,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);

    ////allocate memory for new particles
    recv_p_ranges.Resize(boundary_regions.m);recv_p_ranges.Fill(RANGE_INT1(TV_INT1(0),TV_INT1(0)));
    for(int i=1;i<=recv_p_numbers.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||recv_p_numbers(i)==0)continue;
        int start=particles.X.m+1;
        int end=particles.X.m+recv_p_numbers(i);
        particles.array_collection->Add_Elements(end-start+1);
        recv_p_ranges(i)=RANGE_INT1(TV_INT1(start),TV_INT1(end));
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particles_And_Elements(ARRAY<int>* p_added,ARRAY<int>* p_to_rmv,ARRAY<int>* p_to_exc,
    ARRAY<int>* tet_added,ARRAY<int>* tet_to_rmv, ARRAY<int>* tri_added,ARRAY<int>* tri_to_rmv,ARRAY<int>* seg_added,ARRAY<int>* seg_to_rmv)
{
    //if(p_added)LOG::cout<<"N"<<node_id<<" p_added: "<<*p_added<<std::endl;
    //if(p_to_rmv)LOG::cout<<"N"<<node_id<<" p_to_rmv: "<<*p_to_rmv<<std::endl;
    //if(tet_added)LOG::cout<<"N"<<node_id<<" tet_added: "<<*tet_added<<std::endl;
    //if(tet_to_rmv)LOG::cout<<"N"<<node_id<<" tet_to_rmv: "<<*tet_to_rmv<<std::endl;
    Update_Guid_Vertex_Hashtable();
    if(p_added){
        ARRAY2_INT send_particle_indices;ARRAY_INT send_particle_numbers;ARRAY2_INT recv_particle_indices;ARRAY_INT recv_particle_numbers;ARRAY_RANGE recv_particle_index_ranges;
        Add_Particles(*p_added,send_particle_indices,send_particle_numbers,recv_particle_indices,recv_particle_numbers,recv_particle_index_ranges);
    }
    
    if(p_to_exc){
        ARRAY2_INT send_particle_indices;ARRAY_INT send_particle_numbers;ARRAY2_INT recv_particle_indices;ARRAY_INT recv_particle_numbers;
        Exchange_Particle_Indices2(*p_to_exc,send_particle_indices,send_particle_numbers,recv_particle_indices,recv_particle_numbers);
        Exchange_Particle_Attributes_All(send_particle_indices,recv_particle_indices);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft1 tet "<<i<<std::endl;}
    //}

    if(tet_added){
        ARRAY2_INT4 send_tets;ARRAY_INT send_tet_numbers;ARRAY2_INT4 recv_tets;ARRAY_INT recv_tet_numbers;
        Add_Elements(mesh.tetrahedron_mesh.elements,*tet_added,send_tets,send_tet_numbers,recv_tets,recv_tet_numbers);
    }
    
    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft2 tet "<<i<<std::endl;}
    //}

    //for(int i=1;i<=tet_to_rmv->m;i++){
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements((*tet_to_rmv)(i));
    //    LOG::cout<<"Node "<<node_id<<" rmv tet "<<i<<", "<<(*tet_to_rmv)(i)<<": "<<v<<" guid "<<Element(particles.guid,v)<<std::endl;
    //}

    if(tet_to_rmv){
        ARRAY2_INT4 send_tets;ARRAY_INT send_tet_numbers;ARRAY2_INT4 recv_tets;ARRAY_INT recv_tet_numbers;
        Remove_Elements(mesh.tetrahedron_mesh.elements,*tet_to_rmv,send_tets,send_tet_numbers,recv_tets,recv_tet_numbers);
    }
    
    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft3 tet "<<i<<std::endl;}
    //}

    if(tri_added){
        ARRAY2_INT3 send_tris;ARRAY_INT send_tri_numbers;ARRAY2_INT3 recv_tris;ARRAY_INT recv_tri_numbers;
        Add_Elements(mesh.triangle_mesh.elements,*tri_added,send_tris,send_tri_numbers,recv_tris,recv_tri_numbers);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft4 tet "<<i<<std::endl;}
    //}

    if(tri_to_rmv){
        ARRAY2_INT3 send_tris;ARRAY_INT send_tri_numbers;ARRAY2_INT3 recv_tris;ARRAY_INT recv_tri_numbers;
        Remove_Elements(mesh.triangle_mesh.elements,*tri_to_rmv,send_tris,send_tri_numbers,recv_tris,recv_tri_numbers);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft5 tet "<<i<<std::endl;}
    //}

    if(seg_added){
        ARRAY2_INT2 send_segs;ARRAY_INT send_seg_numbers;ARRAY2_INT2 recv_segs;ARRAY_INT recv_seg_numbers;
        Add_Elements(mesh.segment_mesh.elements,*seg_added,send_segs,send_seg_numbers,recv_segs,recv_seg_numbers);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft6 tet "<<i<<std::endl;}
    //}

    if(seg_to_rmv){
        ARRAY2_INT2 send_segs;ARRAY_INT send_seg_numbers;ARRAY2_INT2 recv_segs;ARRAY_INT recv_seg_numbers;
        Remove_Elements(mesh.segment_mesh.elements,*seg_to_rmv,send_segs,send_seg_numbers,recv_segs,recv_seg_numbers);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft7 tet "<<i<<std::endl;}
    //}

    if(p_to_rmv){
        ARRAY2_INT send_particle_indices;ARRAY_INT send_particle_numbers;ARRAY2_INT recv_particle_indices;ARRAY_INT recv_particle_numbers;
        Remove_Particles(*p_to_rmv,send_particle_indices,send_particle_numbers,recv_particle_indices,recv_particle_numbers);
    }

    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){   ////dbg
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);
    //    if(v.Contains(0)){LOG::cout<<"tet contains 0-N"<<node_id<<" aft8 tet "<<i<<std::endl;}
    //}
}

template<class TV> template<class ARRAY_T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Values(ARRAY_T2& X,ARRAY2_INT& send_p,ARRAY<RANGE_INT1>& recv_p_ranges)
{
    typedef typename ARRAY_T2::ELEMENT T2;
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    ////send attributes
    ARRAY<INDIRECT_ARRAY<ARRAY_VIEW<T2> >* > send_X_indir_arrays(send_p.m);
    for(int i=1;i<=send_p.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||send_p(i).m==0)continue;
        send_X_indir_arrays(i)=new INDIRECT_ARRAY<ARRAY_VIEW<T2> >(X,send_p(i));
        MPI_PACKAGE package(*send_X_indir_arrays(i));packages.Append(package);
        requests.Append(package.Isend(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Send_Tag(neighbor_directions(i))));
    }
    ////receive attributes
    for(int i=1;i<=send_p.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||(recv_p_ranges(i).min_corner.x==0&&recv_p_ranges(i).max_corner.x==0))continue;
        MPI_PACKAGE package(X,recv_p_ranges(i));packages.Append(package);
        requests.Append(package.Irecv(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Recv_Tag(neighbor_directions(i))));
    }
    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);
    send_X_indir_arrays.Delete_Pointers_And_Clean_Memory();
}

template<class TV> template<class ARRAY_T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Values(ARRAY_T2& X,ARRAY2_INT& send_p,ARRAY2_INT& recv_p)
{
    typedef typename ARRAY_T2::ELEMENT T2;
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    ARRAY<INDIRECT_ARRAY<ARRAY_VIEW<T2> >* > send_X_indir_arrays(send_p.m);
    ARRAY<INDIRECT_ARRAY<ARRAY_VIEW<T2> >* > recv_X_indir_arrays(recv_p.m);
    ////send attributes
    for(int i=1;i<=send_p.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||send_p(i).m==0)continue;
        send_X_indir_arrays(i)=new INDIRECT_ARRAY<ARRAY_VIEW<T2> >(X,send_p(i));
        MPI_PACKAGE package(*send_X_indir_arrays(i));packages.Append(package);
        requests.Append(package.Isend(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Send_Tag(neighbor_directions(i))));
        //LOG::cout<<"N"<<node_id<<" sends "<<send_p(i).m<<" to "<<neighbor_ranks(i)+1<<" end "<<std::endl;
    }
    ////receive attributes
    for(int i=1;i<=send_p.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||recv_p(i).m==0)continue;
        recv_X_indir_arrays(i)=new INDIRECT_ARRAY<ARRAY_VIEW<T2> >(X,recv_p(i));
        MPI_PACKAGE package(*recv_X_indir_arrays(i));packages.Append(package);
        requests.Append(package.Irecv(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Recv_Tag(neighbor_directions(i))));
        //LOG::cout<<"N"<<node_id<<" recvs "<<recv_p(i).m<<" from "<<neighbor_ranks(i)+1<<" end "<<std::endl;
    }
    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);
    send_X_indir_arrays.Delete_Pointers_And_Clean_Memory();
    recv_X_indir_arrays.Delete_Pointers_And_Clean_Memory();
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particle_Attributes_All(ARRAY2_INT& send_p,ARRAY<RANGE_INT1>& recv_p_ranges)
{
    Exchange_Values(particles.guid,send_p,recv_p_ranges);
    Exchange_Values(particles.X,send_p,recv_p_ranges);
    Exchange_Values(particles.V,send_p,recv_p_ranges);
    Exchange_Values(particles.mass,send_p,recv_p_ranges);
    Exchange_Values(particles.dimensional_mass,send_p,recv_p_ranges);
    Exchange_Values(particles.dimensional_size,send_p,recv_p_ranges);
    Exchange_Values(particles.thickness,send_p,recv_p_ranges);
    Exchange_Values(particles.length_scale,send_p,recv_p_ranges);
    Exchange_Values(particles.contact,send_p,recv_p_ranges);
    
    if(particles.phis_number>0){
        for(int i=1;i<=particles.phis.m;i++)Exchange_Values(particles.phis(i),send_p,recv_p_ranges);
    }
    if(particles.parents_size>0){
        for(int i=1;i<=particles.parents_guid.m;i++){
            Exchange_Values(particles.parents_guid(i),send_p,recv_p_ranges);
            Exchange_Values(particles.parents_weight(i),send_p,recv_p_ranges);
        }
        Exchange_Values(particles.parents_number,send_p,recv_p_ranges);
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particle_Attributes_All(ARRAY2_INT& send_p,ARRAY2_INT& recv_p)
{
    Exchange_Values(particles.guid,send_p,recv_p);
    Exchange_Values(particles.X,send_p,recv_p);
    Exchange_Values(particles.V,send_p,recv_p);
    Exchange_Values(particles.mass,send_p,recv_p);
    Exchange_Values(particles.dimensional_mass,send_p,recv_p);
    Exchange_Values(particles.dimensional_size,send_p,recv_p);
    Exchange_Values(particles.thickness,send_p,recv_p);
    Exchange_Values(particles.length_scale,send_p,recv_p);
    Exchange_Values(particles.contact,send_p,recv_p);

    if(particles.phis_number>0){
        for(int i=1;i<=particles.phis.m;i++)Exchange_Values(particles.phis(i),send_p,recv_p);
    }
    if(particles.parents_size>0){
        for(int i=1;i<=particles.parents_guid.m;i++){
            Exchange_Values(particles.parents_guid(i),send_p,recv_p);
            Exchange_Values(particles.parents_weight(i),send_p,recv_p);
        }
        Exchange_Values(particles.parents_number,send_p,recv_p);
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particle_Indices(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,
    ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers,bool exc_guid/*=false*/)
{
    //LOG::cout<<"node "<<node_id<<" exc "<<p_to_exc<<" finish exc"<<std::endl;
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    ////find particles to send
    Clear_Arrays(send_p);Clear_Values(send_p_numbers);
    for(int i=1;i<=p_to_exc.m;i++){
        int p=p_to_exc(i);
        int p_guid=particles.guid(p);
        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(particles.X(p));
        if(boundary_shrink_region_index!=0){
            //LOG::cout<<"node "<<node_id<<" p_in_bd "<<p<<", "<<boundary_shrink_region_index<<std::endl;
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                send_p(boundary_region_indices(j)).Append(exc_guid?p_guid:p);
            }
        }
    }
    for(int i=1;i<=send_p.m;i++)send_p_numbers(i)=send_p(i).m;

    ////send particle numbers
    Mpi_Send_Values(send_p_numbers,requests);

    ////receive particle numbers
    Clear_Values(recv_p_numbers);
    Mpi_Recv_Values(recv_p_numbers,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);//Free_All(requests);

    ////allocate memory for received particles
    Clear_Arrays(recv_p);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_p_numbers(i)>0){
            recv_p(i).Resize(recv_p_numbers(i));
        }
    }

    ////send particles
    Mpi_Send_Arrays(send_p,send_p_numbers,requests,packages);
    ////receive tets
    Mpi_Recv_Arrays(recv_p,recv_p_numbers,requests,packages);

    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particle_Indices2(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers)
{
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    ////find particles to send
    Clear_Arrays(send_p);Clear_Values(send_p_numbers);
    ARRAY2_INT send_p_guids;Clear_Arrays(send_p_guids);
    for(int i=1;i<=p_to_exc.m;i++){
        int p=p_to_exc(i);
        int p_guid=particles.guid(p);
        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(particles.X(p));
        if(boundary_shrink_region_index!=0){
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                send_p(boundary_region_indices(j)).Append(p);
                send_p_guids(boundary_region_indices(j)).Append(p_guid);
            }
        }
    }
    for(int i=1;i<=send_p.m;i++)send_p_numbers(i)=send_p(i).m;

    ////send particle numbers
    Mpi_Send_Values(send_p_numbers,requests);

    ////receive particle numbers
    Clear_Values(recv_p_numbers);
    Mpi_Recv_Values(recv_p_numbers,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);//Free_All(requests);

    ////allocate memory for received particles
    Clear_Arrays(recv_p);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_p_numbers(i)>0){
            recv_p(i).Resize(recv_p_numbers(i));
        }
    }

    ////send particle guids
    Mpi_Send_Arrays(send_p_guids,send_p_numbers,requests,packages);
    ////receive particle guids
    Mpi_Recv_Arrays(recv_p,recv_p_numbers,requests,packages);

    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);

    ////translate guid to id
    for(int i=1;i<=boundary_regions.m;i++){
        for(int j=1;j<=recv_p(i).m;j++){
            int guid=recv_p(i)(j);
            recv_p(i)(j)=Guid_To_Vtx(recv_p(i)(j));
            if(recv_p(i)(j)==0){
                LOG::cout<<"&&&&&&&&&&& invalid recvp, N"<<node_id<<": "<<i<<", "<<j<<": "<<guid<<std::endl;
            }
            PHYSBAM_ASSERT(recv_p(i)(j)!=0);
        }
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Remove_Particles(const ARRAY<int>& p_to_rmv,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,
    ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers)
{
    //LOG::cout<<"Node "<<node_id<<", send p_to_rmv: ";
    //for(int i=1;i<=p_to_rmv.m;i++){LOG::cout<<particles.guid(p_to_rmv(i))<<" ";}LOG::cout<<" finish send"<<std::endl;

    Exchange_Particle_Indices(p_to_rmv,send_p,send_p_numbers,recv_p,recv_p_numbers,/*exchange guid*/true);
    
    //////dbg
    //LOG::cout<<"node "<<node_id<<" check send"<<std::endl;
    //LOG::cout<<"node "<<node_id<<" send_p_n "<<send_p_numbers<<std::endl;
    //for(int i=1;i<=send_p.m;i++){
    //    for(int j=1;j<=send_p(i).m;j++){
    //        const int vtx_guid=send_p(i)(j);
    //        LOG::cout<<"node "<<node_id<<" send "<<vtx_guid<<" to "<<neighbor_ranks(i)+1<<std::endl;
    //    }
    //}
    //LOG::cout<<"node "<<node_id<<" finish check send"<<std::endl;

    HASHTABLE<int> particles_to_remove_hashtable;
    for(int i=1;i<=p_to_rmv.m;i++){
        int p=p_to_rmv(i);particles_to_remove_hashtable.Set(p);
    }
    for(int i=1;i<=recv_p.m;i++){
        for(int j=1;j<=recv_p(i).m;j++){
            const int vtx_guid=recv_p(i)(j);
            //LOG::cout<<"xnode "<<node_id<<" recv_rmv "<<vtx_guid<<std::endl;
            int vtx=Guid_To_Vtx(vtx_guid);  ////translate from guid to id
            if(vtx!=0)particles_to_remove_hashtable.Set(vtx);
            //else LOG::cout<<"???? node "<<node_id<<" invalid recv_rmv "<<vtx_guid<<std::endl;
        }
    }

    ARRAY<int> particles_to_remove;
    for(HASHTABLE_ITERATOR<int> iter(particles_to_remove_hashtable);iter.Valid();iter.Next()){
        particles_to_remove.Append(iter.Key());
    }
    Sort(particles_to_remove);

    ////update guid_vtx_hashtable
    for(int i=1;i<=particles_to_remove.m;i++){
        int p=particles_to_remove(i);
        guid_vertex_hashtable.Delete_If_Present(particles.guid(p));
    }

    //////dbg: check whether there is tet containing particles to remove
    //for(int i=1;i<=mesh.tetrahedron_mesh.elements.m;i++){
    //    const TV_INT4& v=mesh.tetrahedron_mesh.elements(i);bool has_bad=false;int bad_id=0;
    //    for(int j=1;j<=particles_to_remove.m;j++){if(v.Contains(particles_to_remove(j))){has_bad=true;bad_id=particles_to_remove(j);break;}}
    //    if(has_bad){LOG::cout<<"Node "<<node_id<<" bad_ele: "<<i<<", "<<v<<": guid "<<Element(particles.guid,v)<<": bad_p: "<<bad_id<<", "<<particles.guid(bad_id)<<std::endl;}
    //}

    //////dbg: output particles_to_remove guid
    //LOG::cout<<"start rmv N"<<node_id<<": p ";
    //for(int i=1;i<=particles_to_remove.m;i++){LOG::cout<<particles.guid(particles_to_remove(i))<<" ";}
    //LOG::cout<<"finish rmv"<<std::endl;
    
    Remove_Sorted_Particles(particles,particles_to_remove,mesh,/*don't remove elements*/false);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Add_Particles(const ARRAY<int>& p_added,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers, 
    ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers,ARRAY<RANGE_INT1>& recv_p_ranges)
{
    Exchange_Particle_Indices(p_added,send_p,send_p_numbers,recv_p,recv_p_numbers);
    ////allocate memory for new particles
    //int add_first_idx=particles.X.m+1;
    recv_p_ranges.Resize(boundary_regions.m);recv_p_ranges.Fill(RANGE_INT1(TV_INT1(0),TV_INT1(0)));
    for(int i=1;i<=recv_p_numbers.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||recv_p_numbers(i)==0)continue;
        int start=particles.X.m+1;
        int end=particles.X.m+recv_p_numbers(i);
        particles.array_collection->Add_Elements(end-start+1);
        recv_p_ranges(i)=RANGE_INT1(TV_INT1(start),TV_INT1(end));
    }

    Exchange_Particle_Attributes_All(send_p,recv_p_ranges);
    Update_Guid_Vertex_Hashtable();
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_to_exc,
    ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers,ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers)
{
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;

    ////find elements to send
    Clear_Arrays(send_e);Clear_Values(send_e_numbers);
    for(int i=1;i<=e_to_exc.m;i++){
        int p=e_to_exc(i);
        const TV_INTN& vtx=elements(p);
        ARRAY<int> send_region_indices;
        for(int d=1;d<=TV_INTN::m;d++){
            const TV& pos=particles.X(vtx(d));
            TV_INT nb_offset=Neighbor_Offset(pos,domain_shrink);
            int boundary_shrink_region_index=Boundary_Shrink_Region_Index(nb_offset);
            if(boundary_shrink_region_index!=0){
                const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
                for(int j=1;j<=boundary_region_indices.m;j++){
                    send_region_indices.Append_Unique(boundary_region_indices(j));
                }
            }
        }

        for(int j=1;j<=send_region_indices.m;j++){
            TV_INTN e_guid=Element(particles.guid,vtx);
            send_e(send_region_indices(j)).Append(Unique_Ordered(e_guid));
        }
    }
    for(int i=1;i<=send_e.m;i++){
        send_e_numbers(i)=send_e(i).m;
    }

    ////send element numbers
    
    Mpi_Send_Values(send_e_numbers,requests);

    ////receive element numbers
    recv_e_numbers.Resize(boundary_regions.m,true,false);
    Mpi_Recv_Values(recv_e_numbers,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);//Free_All(requests);

    ////allocate memory for received tets
    Clear_Arrays(recv_e);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_e_numbers(i)>0){
            recv_e(i).Resize(recv_e_numbers(i));
        }
    }

    ////send tets
    Mpi_Send_Arrays(send_e,send_e_numbers,requests,packages);
    ////receive tets
    Mpi_Recv_Arrays(recv_e,recv_e_numbers,requests,packages);

    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Remove_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_to_rmv,
    ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers,ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers)
{
    Exchange_Elements(elements,e_to_rmv,send_e,send_e_numbers,recv_e,recv_e_numbers);

    ////update element-index hashtable
    HASHTABLE<TV_INTN,int> e_idx_hashtable;
    for(int i=1;i<=elements.m;i++){
        e_idx_hashtable.Set(Unique_Ordered(elements(i)),i);
    }

    ////merge all elements to remove 
    ////translate from guid to id
    HASHTABLE<int> elements_to_remove_hashtable;
    for(int i=1;i<=e_to_rmv.m;i++)elements_to_remove_hashtable.Set(e_to_rmv(i));

    for(int i=1;i<=recv_e.m;i++){
        for(int j=1;j<=recv_e(i).m;j++){
            const TV_INTN& vtx_guid=recv_e(i)(j);
            TV_INTN vtx=Guid_To_Vtx(vtx_guid);
            int idx=0;
            if(e_idx_hashtable.Get(Unique_Ordered(vtx),idx))elements_to_remove_hashtable.Set(idx);
        }
    }
    ARRAY<int> elements_to_remove;
    for(HASHTABLE_ITERATOR<int> iter(elements_to_remove_hashtable);iter.Valid();iter.Next()){
        elements_to_remove.Append(iter.Key());
    }
    Sort(elements_to_remove);

    //////dbg
    //for(int i=1;i<=elements_to_remove.m;i++){
    //    TV_INTN guid=Element(particles.guid,elements(elements_to_remove(i)));
    //    if(guid.Contains(16)){
    //        LOG::cout<<"++++++++++++ N"<<node_id<<" rmv ele "<<guid<<std::endl;
    //    }
    //}

    //LOG::cout<<"Node "<<node_id<<" recv e_to_rmv ";
    //for(int i=1;i<=elements_to_remove.m;i++){
    //    LOG::cout<<elements(elements_to_remove(i))<<" guid "<<Element(particles.guid,elements(elements_to_remove(i)))<<"\n";
    //}
    //LOG::cout<<" Node "<<node_id<<" recv finish"<<std::endl;

    elements.Remove_Sorted_Indices(elements_to_remove);
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Add_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_added,
    ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers, ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers)
{
    Exchange_Elements(elements,e_added,send_e,send_e_numbers,recv_e,recv_e_numbers);

    HASHTABLE<TV_INTN> e_to_add_hashtable;
    for(int i=1;i<=recv_e.m;i++){
        for(int j=1;j<=recv_e(i).m;j++){
            const TV_INTN& vtx_guid=recv_e(i)(j);
            TV_INTN vtx=Guid_To_Vtx(vtx_guid);  ////translate guid to id
            if(!vtx.Contains(0)){
                e_to_add_hashtable.Set(Unique_Ordered(vtx));
            }
        }
    }

    //////dbg
    //for(int i=1;i<=e_added.m;i++){
    //    const TV_INTN& v=elements(e_added(i));
    //    TV_INTN guid=Element(particles.guid,v);
    //    if(guid.Contains(16)){
    //        LOG::cout<<"$$$$$$$$$$$$$$$$ N"<<node_id<<" added "<<v<<", "<<guid<<std::endl;
    //    }
    //}

    //for(int i=1;i<=send_e.m;i++){
    //    for(int j=1;j<=send_e(i).m;j++){
    //        const TV_INTN& vtx_guid=send_e(i)(j);
    //        if(vtx_guid.Contains(16)){
    //            LOG::cout<<"!!!!!!!!!!!!!!!! N"<<node_id<<" send "<<vtx_guid<<std::endl;
    //        }
    //    }
    //}
    //for(int i=1;i<=recv_e.m;i++){
    //    for(int j=1;j<=recv_e(i).m;j++){
    //        const TV_INTN& vtx_guid=recv_e(i)(j);
    //        if(vtx_guid.Contains(16)){
    //            LOG::cout<<"!!!!!!!!!!!!!!!! N"<<node_id<<" recv "<<vtx_guid<<std::endl;
    //        }
    //    }
    //}

    ////add elements
    //if(node_id==2){
    //    LOG::cout<<"node 2 add e: ";
    //    for(HASHTABLE_ITERATOR<TV_INTN> iter(e_to_add_hashtable);iter.Valid();iter.Next()){
    //        LOG::cout<<iter.Key()<<", ";
    //    }
    //    LOG::cout<<"finish add e"<<std::endl;
    //}

    for(HASHTABLE_ITERATOR<TV_INTN> iter(e_to_add_hashtable);iter.Valid();iter.Next()){
        elements.Append(iter.Key());
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Particles_And_Elements_Simple()
{
    ARRAY2_INT send_particle_indices;ARRAY_INT send_particle_numbers;
    ARRAY2_INT recv_particle_indices;ARRAY_INT recv_particle_numbers;
    ARRAY_RANGE recv_particle_index_ranges;
    Update_Guid_Vertex_Hashtable();
    Update_Vertex_Element_Hashtable_In_Boundary_Regions();
    Remove_Out_Of_Domain_Particles_And_Elements();
    ARRAY<int> p_indices(particles.X.m);for(int i=1;i<=p_indices.m;i++)p_indices(i)=i;
    Exchange_And_Append_Particle_Indices(p_indices,send_particle_indices,send_particle_numbers,recv_particle_index_ranges,recv_particle_numbers);
    Exchange_Particle_Attributes_All(send_particle_indices,recv_particle_index_ranges);
    Update_Guid_Vertex_Hashtable();
    Exchange_Elements_Simple(send_particle_indices);

    ////boundary_particle_indices and ghost_particle_indices are not updated

    Update_Boundary_And_Ghost_Particle_Indices();
    mesh.Update_States(particles.X.m,this);
    //{ARRAY_VIEW<int> node_types_array_view(mesh.node_types);Exchange_Boundary_And_Ghost_Particle_Attributes(node_types_array_view);}
}

template<class TV>
typename TV::template REBIND<int>::TYPE PhysBAM::MPI_CODIMENSIONAL<TV>::Neighbor_Offset(const TV& pos,const RANGE<TV>& box) const
{
    TV_INT offset=TV_INT::Constant_Vector(0);
    for(int d=1;d<=TV::m;d++){
        if(pos(d)<box.min_corner(d))offset(d)=-1;
        else if(pos(d)>=box.max_corner(d))offset(d)=1;  ////half open
    }
    return offset;
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Remove_Out_Of_Domain_Particles_And_Elements()
{
    ////find particles to remove
    ARRAY<int> particles_to_remove;
    for(int i=1;i<=particles.X.m;i++){
        if(!Inside_Domain(particles.X(i))){particles_to_remove.Append(i);}
    }
    ////remove particles and incident elements
    Remove_Sorted_Particles(particles,particles_to_remove,mesh,true);
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Elements_Simple(const ARRAY2_INT& send_p)
{
    ARRAY2_INT4 send_tets;ARRAY_INT send_tet_numbers;ARRAY2_INT4 recv_tets;ARRAY_INT recv_tet_numbers;
    ARRAY2_INT3 send_tris;ARRAY_INT send_tri_numbers;ARRAY2_INT3 recv_tris;ARRAY_INT recv_tri_numbers;
    ARRAY2_INT2 send_segs;ARRAY_INT send_seg_numbers;ARRAY2_INT2 recv_segs;ARRAY_INT recv_seg_numbers;
    //ARRAY2_INT2 send_rim_segs;ARRAY_INT send_rim_seg_numbers;ARRAY2_INT2 recv_rim_segs;ARRAY_INT recv_rim_seg_numbers;

    Exchange_Mesh_Elements_Helper(mesh.tetrahedron_mesh.elements,vtx_tet_hashtable,send_p,send_tets,recv_tets,send_tet_numbers,recv_tet_numbers);
    Exchange_Mesh_Elements_Helper(mesh.triangle_mesh.elements,vtx_tri_hashtable,send_p,send_tris,recv_tris,send_tri_numbers,recv_tri_numbers);
    Exchange_Mesh_Elements_Helper(mesh.segment_mesh.elements,vtx_seg_hashtable,send_p,send_segs,recv_segs,send_seg_numbers,recv_seg_numbers);
    //Exchange_Mesh_Elements_Helper(mesh.rim_segment_mesh.elements,vtx_rim_seg_hashtable,send_p,send_rim_segs,recv_rim_segs,send_rim_seg_numbers,recv_rim_seg_numbers);
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Send_Values(ARRAY<T2>& send_values,ARRAY<MPI::Request>& requests)
{
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    for(int i=1;i<=send_values.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL)continue;
        requests.Append(mpi_grid.comm->Isend(&send_values(i),1,type,neighbor_ranks(i),mpi_grid.Get_Send_Tag(neighbor_directions(i))));
        //LOG::cout<<"N"<<node_id<<" mpi send value "<<send_values(i)<<" to "<<neighbor_ranks(i)+1<<", req "<<requests.m<<std::endl;
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Recv_Values(ARRAY<T2>& recv_values,ARRAY<MPI::Request>& requests)
{
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    for(int i=1;i<=recv_values.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL)continue;
        requests.Append(mpi_grid.comm->Irecv(&recv_values(i),1,type,neighbor_ranks(i),mpi_grid.Get_Recv_Tag(neighbor_directions(i))));
        //LOG::cout<<"N"<<node_id<<" mpi recv value "<<recv_values(i)<<" from "<<neighbor_ranks(i)+1<<", req "<<requests.m<<std::endl;
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Send_Arrays(ARRAY<ARRAY<T2> >& send_arrays,const ARRAY<int>& send_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages)
{
    for(int i=1;i<=send_arrays.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||send_numbers(i)==0)continue;
        MPI_PACKAGE package(send_arrays(i));packages.Append(package);
        requests.Append(package.Isend(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Send_Tag(neighbor_directions(i))));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Recv_Arrays(ARRAY<ARRAY<T2> >& recv_arrays,const ARRAY<int>& recv_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages)
{
    for(int i=1;i<=recv_arrays.m;i++){
        if(neighbor_ranks(i)==MPI::PROC_NULL||recv_numbers(i)==0)continue;
        MPI_PACKAGE package(recv_arrays(i));packages.Append(package);
        requests.Append(package.Irecv(*mpi_grid.comm,neighbor_ranks(i),mpi_grid.Get_Recv_Tag(neighbor_directions(i))));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Broadcast_Send_Values(T2& broadcast_send_value,ARRAY<MPI::Request>& requests)
{
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    int mpi_node_number=mpi_grid.number_of_processes;
    for(int i=1;i<=mpi_node_number;i++){
        if(i==node_id)continue;int send_node_rank=i-1; ////node_id=rank+1
        requests.Append(mpi_grid.comm->Isend(&broadcast_send_value,1,type,send_node_rank,0));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Broadcast_Recv_Values(ARRAY<T2>& broadcast_recv_values,ARRAY<MPI::Request>& requests)
{
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    for(int i=1;i<=broadcast_recv_values.m;i++){
        if(i==node_id)continue;int recv_node_rank=i-1; ////node_id=rank+1
        requests.Append(mpi_grid.comm->Irecv(&broadcast_recv_values(i),1,type,recv_node_rank,0));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Broadcast_Send_Arrays(ARRAY<T2>& broadcast_send_array,const int send_number,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages)
{
    if(send_number==0)return;
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    int mpi_node_number=mpi_grid.number_of_processes;
    for(int i=1;i<=mpi_node_number;i++){
        if(i==node_id)continue;int send_node_rank=i-1; ////node_id=rank+1
        MPI_PACKAGE package(broadcast_send_array);packages.Append(package);
        requests.Append(package.Isend(*mpi_grid.comm,send_node_rank,0));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Broadcast_Recv_Arrays(ARRAY<ARRAY<T2> >& broadcast_recv_arrays,const ARRAY<int>& recv_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages)
{
    MPI::Datatype type=MPI_UTILITIES::Datatype<T2>();
    for(int i=1;i<=broadcast_recv_arrays.m;i++){
        if(i==node_id||recv_numbers(i)==0)continue;int recv_node_rank=i-1; ////node_id=rank+1
        MPI_PACKAGE package(broadcast_recv_arrays(i));packages.Append(package);
        requests.Append(package.Irecv(*mpi_grid.comm,recv_node_rank,0));
    }
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Mpi_Broadcast_Arrays(ARRAY<T2>& broadcast_send_array,ARRAY<ARRAY<T2> >& broadcast_recv_arrays)
{
    //int broadcast_send_value=0;
    int broadcast_send_value=broadcast_send_array.m;
    //ARRAY<int> broadcast_send_array;
    //ARRAY<int> broadcast_recv_values(mpi_codim->mpi_grid.number_of_processes);
    broadcast_recv_arrays.Resize(mpi_grid.number_of_processes);
    broadcast_recv_arrays.Fill(ARRAY<T2>());
    ARRAY<int> broadcast_recv_values(mpi_grid.number_of_processes);
    //ARRAY<ARRAY<int> > broadcast_recv_arrays(mpi_codim->mpi_grid.number_of_processes);
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;

    ////broadcast array size
    Mpi_Broadcast_Send_Values(broadcast_send_value,requests);

    ////receive array size
    broadcast_recv_values.Fill(0);
    Mpi_Broadcast_Recv_Values(broadcast_recv_values,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);

    ////allocate memory for array
    for(int i=1;i<=broadcast_recv_arrays.m;i++){
        if(broadcast_recv_values(i)>0){
            broadcast_recv_arrays(i).Resize(broadcast_recv_values(i));
        }
    }

    ////broadcast arrays
    Mpi_Broadcast_Send_Arrays(broadcast_send_array,broadcast_send_value,requests,packages);
    ////receive arrays
    Mpi_Broadcast_Recv_Arrays(broadcast_recv_arrays,broadcast_recv_values,requests,packages);

    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Realloc_Arrays(ARRAY<ARRAY<T2> >& alloc_arrays,const ARRAY<int>& numbers)
{
    alloc_arrays.Resize(boundary_regions.m);for(int i=1;i<=alloc_arrays.m;i++)alloc_arrays(i).Remove_All();
    for(int i=1;i<=boundary_regions.m;i++){
        if(numbers(i)>0){
            alloc_arrays(i).Resize(numbers(i));
        }
    }
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Mesh_Elements_Helper(ARRAY<TV_INTN>& elements,HASHTABLE<int,ARRAY<TV_INTN> >& vtx_e_hashtable,const ARRAY2_INT& send_p,
    ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& send_e_numbers,ARRAY<int>& recv_e_numbers)
{
    ////find elements to send
    Clear_Arrays(send_e);Clear_Values(send_e_numbers);
    for(int i=1;i<=send_p.m;i++){
        HASHTABLE<TV_INTN> e_to_send_hashtable;
        for(int j=1;j<=send_p(i).m;j++){
            int p=send_p(i)(j);
            int guid=particles.guid(p);
            if(vtx_e_hashtable.Contains(guid)){
                const ARRAY<TV_INTN>& nb_tets=vtx_e_hashtable.Get(guid);
                for(int k=1;k<=nb_tets.m;k++){
                    e_to_send_hashtable.Set(Unique_Ordered(nb_tets(k)));
                }
            }
        }
        for(HASHTABLE_ITERATOR<TV_INTN> iter(e_to_send_hashtable);iter.Valid();iter.Next()){
            send_e(i).Append(iter.Key());
        }
        send_e_numbers(i)=send_e(i).m;
    }

    ////send element numbers
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    Mpi_Send_Values(send_e_numbers,requests);

    ////receive element numbers
    recv_e_numbers.Resize(boundary_regions.m,true,false);
    Mpi_Recv_Values(recv_e_numbers,requests);

    ////finish
    MPI_UTILITIES::Wait_All(requests);//Free_All(requests);

    ////allocate memory for received tets
    recv_e.Resize(boundary_regions.m);for(int i=1;i<=recv_e.m;i++)recv_e(i).Remove_All();
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_e_numbers(i)>0){
            recv_e(i).Resize(recv_e_numbers(i));
        }
    }

    ////send elements
    Mpi_Send_Arrays(send_e,send_e_numbers,requests,packages);
    ////receive elements
    Mpi_Recv_Arrays(recv_e,recv_e_numbers,requests,packages);

    ////finish
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);

    ////add elements
    ////TODO: need to check duplicity here?
    HASHTABLE<TV_INTN> e_to_add_hashtable;
    for(int i=1;i<=recv_e.m;i++){
        for(int j=1;j<=recv_e(i).m;j++){
            const TV_INTN& vtx_guid=recv_e(i)(j);
            TV_INTN vtx=Guid_To_Vtx(vtx_guid);  ////translate from guid to id
            if(!vtx.Contains(0))e_to_add_hashtable.Set(Unique_Ordered(vtx));
            //else{
            //    ////dbg
            //    if(vtx_guid.Contains(16)){
            //        LOG::cout<<",,,,,,,,,,,,,,, N"<<node_id<<" "<<vtx_guid<<": "<<vtx<<std::endl;
            //    }
            //}
        }
    }
    for(HASHTABLE_ITERATOR<TV_INTN> iter(e_to_add_hashtable);iter.Valid();iter.Next()){
        elements.Append(iter.Key());
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Update_Boundary_And_Ghost_Particle_Indices()
{
    ARRAY_INT particles_to_exchange;ARRAY_INT send_particle_numbers;ARRAY_INT recv_particle_numbers;
    
    domain_particle_indices.Remove_All();
    for(int i=1;i<=particles.X.m;i++){
        if(Inside_Domain_Boundary(particles.X(i)))particles_to_exchange.Append(i);
        if(Inside_Domain(particles.X(i)))domain_particle_indices.Append(i);
    }

    Clear_Arrays(ghost_particle_indices);Clear_Arrays(boundary_particle_indices);
    Clear_Values(send_particle_numbers);Clear_Values(recv_particle_numbers);
    Exchange_Particle_Indices2(particles_to_exchange,boundary_particle_indices,send_particle_numbers,ghost_particle_indices,recv_particle_numbers);
}

template<class TV> template<class T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<T2>& X)
{
    Exchange_Values(X,boundary_particle_indices,ghost_particle_indices);
}

template<class TV> template<class ARRAY_T2>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Exchange_Boundary_And_Ghost_Pressure_Values(ARRAY_T2& X,ARRAY2_INT& boundary_pressure_indices,ARRAY2_INT& ghost_pressure_indices)
{
    Exchange_Values(X,boundary_pressure_indices,ghost_pressure_indices);
}

template<class TV>
bool PhysBAM::MPI_CODIMENSIONAL<TV>::Valid_Particle_Pressure_Sample_To_Send(const int p,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh,bool all_pressures_on_nodes)
{
    for(int i=1;i<=(*solver_mesh.tetrahedron_mesh.incident_elements)(p).m;i++){
        int t=(*solver_mesh.tetrahedron_mesh.incident_elements)(p)(i);
        const TV_INT4& v=solver_mesh.tetrahedron_mesh.elements(t);
        bool valid_element=true;
        for(int j=1;j<=4;j++){
            int bs=Boundary_Shrink_Region_Index(particles.X(v(j)));if(bs==0){valid_element=false;break;}
            const ARRAY<int>& b_indices=boundary_shrink_to_boundary_maps(bs);
            if(!b_indices.Contains(boundary_region_idx)){valid_element=false;break;}
        }
        if(valid_element)return true;   ////return true if it has at least one element fully inside the domain it will send to 
    }
    if(all_pressures_on_nodes){
        for(int i=1;i<=(*solver_mesh.triangle_mesh.incident_elements)(p).m;i++){
            int t=(*solver_mesh.triangle_mesh.incident_elements)(p)(i);
            const TV_INT3& v=solver_mesh.triangle_mesh.elements(t);
            bool valid_element=true;
            for(int j=1;j<=3;j++){
                int bs=Boundary_Shrink_Region_Index(particles.X(v(j)));if(bs==0){valid_element=false;break;}
                const ARRAY<int>& b_indices=boundary_shrink_to_boundary_maps(bs);
                if(!b_indices.Contains(boundary_region_idx)){valid_element=false;break;}
            }
            if(valid_element)return true;
        }
        for(int i=1;i<=(*solver_mesh.segment_mesh.incident_elements)(p).m;i++){
            int t=(*solver_mesh.segment_mesh.incident_elements)(p)(i);
            const TV_INT2& v=solver_mesh.segment_mesh.elements(t);
            bool valid_element=true;
            for(int j=1;j<=2;j++){
                int bs=Boundary_Shrink_Region_Index(particles.X(v(j)));if(bs==0){valid_element=false;break;}
                const ARRAY<int>& b_indices=boundary_shrink_to_boundary_maps(bs);
                if(!b_indices.Contains(boundary_region_idx)){valid_element=false;break;}
            }
            if(valid_element)return true;
        }
    }
    return false;
}

template<class TV>
bool PhysBAM::MPI_CODIMENSIONAL<TV>::Valid_Triangle_Pressure_Sample_To_Send(const int t,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh)
{
    const TV_INT3& v=solver_mesh.triangle_mesh.elements(t);
    for(int i=1;i<=3;i++){
        int bs=Boundary_Shrink_Region_Index(particles.X(v(i)));if(bs==0){return false;}
        const ARRAY<int>& b_indices=boundary_shrink_to_boundary_maps(bs);
        if(!b_indices.Contains(boundary_region_idx))return false;
    }
    return true;
}

template<class TV>
bool PhysBAM::MPI_CODIMENSIONAL<TV>::Valid_Segment_Pressure_Sample_To_Send(const int s,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh)
{
    const TV_INT2& v=solver_mesh.segment_mesh.elements(s);
    for(int i=1;i<=2;i++){
        int bs=Boundary_Shrink_Region_Index(particles.X(v(i)));if(bs==0){return false;}
        const ARRAY<int>& b_indices=boundary_shrink_to_boundary_maps(bs);
        if(!b_indices.Contains(boundary_region_idx))return false;
    }
    return true;
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Update_Boundary_And_Ghost_Pressure_Indices(const ARRAY<CODIMENSIONAL_PRESSURE>& pressures,const CODIMENSIONAL_MESH& solver_mesh,bool all_pressures_on_nodes,
    ARRAY2_INT& boundary_pressure_indices,ARRAY2_INT& ghost_pressure_indices,ARRAY<int>& domain_pressure_indices)
{
    p_guid_to_pressure_hashtable.Remove_All();
    seg_guid_to_pressure_hashtable.Remove_All();
    tri_guid_to_pressure_hashtable.Remove_All();
    domain_pressure_indices.Remove_All();

    VECTOR<ARRAY2_INT,4> send_pressure_indices;
    VECTOR<ARRAY2_INT,4> recv_pressure_indices;
    VECTOR<ARRAY_INT,4> all_pressures_to_send;
    VECTOR<ARRAY_INT,4> all_elements_to_send;

    ARRAY2_INT send_particle_indices;ARRAY_INT send_particle_numbers;ARRAY2_INT recv_particle_indices;ARRAY_INT recv_particle_numbers;ARRAY_RANGE recv_particle_index_ranges;
    ARRAY2_INT4 send_tets;ARRAY_INT send_tet_numbers;ARRAY2_INT4 recv_tets;ARRAY_INT recv_tet_numbers;
    ARRAY2_INT3 send_tris;ARRAY_INT send_tri_numbers;ARRAY2_INT3 recv_tris;ARRAY_INT recv_tri_numbers;
    ARRAY2_INT2 send_segs;ARRAY_INT send_seg_numbers;ARRAY2_INT2 recv_segs;ARRAY_INT recv_seg_numbers;

    //LOG::cout<<"seg_n N"<<node_id<<" : "<<solver_mesh.segment_mesh.elements.m<<std::endl;

    for(int i=1;i<=pressures.m;i++){
        int type=pressures(i).type;int idx=pressures(i).index;
        switch(type){
        case SIMPLEX_PARTICLE:{
                TV pos=particles.X(idx);
                if(Inside_Domain_Boundary(pos)){
                    all_pressures_to_send(SIMPLEX_PARTICLE).Append(i);
                    all_elements_to_send(SIMPLEX_PARTICLE).Append(idx);
                }

                int guid=particles.guid(idx);
                p_guid_to_pressure_hashtable.Set(guid,i);
                //LOG::cout<<"N"<<node_id<<" set_p "<<guid<<", "<<i<<std::endl;

                if(Inside_Domain(pos)){domain_pressure_indices.Append(i);}
            }break;
        case SIMPLEX_SEGMENT:{
                const TV_INT2& v=solver_mesh.segment_mesh.elements(idx);
                //if(Inside_Domain_Boundary(pos)){
                if(Inside_Domain_Boundary(particles.X(v(1)))&&Inside_Domain_Boundary(particles.X(v(2)))){
                    all_pressures_to_send(SIMPLEX_SEGMENT).Append(i);
                    all_elements_to_send(SIMPLEX_SEGMENT).Append(idx);
                }

                TV_INT2 guid=Unique_Ordered(Element(particles.guid,solver_mesh.segment_mesh.elements(idx)));
                seg_guid_to_pressure_hashtable.Set(guid,i);
                //LOG::cout<<"N"<<node_id<<" set_s v "<<v<<" guid "<<guid<<", "<<i<<std::endl;

                if(Inside_Domain(particles.X(v(1)))&&Inside_Domain(particles.X(v(2)))){domain_pressure_indices.Append(i);}
            }break;
        case SIMPLEX_TRIANGLE:{
                const TV_INT3& v=solver_mesh.triangle_mesh.elements(idx);
                if(Inside_Domain_Boundary(particles.X(v(1)))&&Inside_Domain_Boundary(particles.X(v(2)))&&Inside_Domain_Boundary(particles.X(v(3)))){
                    all_pressures_to_send(SIMPLEX_TRIANGLE).Append(i);
                    all_elements_to_send(SIMPLEX_TRIANGLE).Append(idx);
                }

                TV_INT3 guid=Unique_Ordered(Element(particles.guid,solver_mesh.triangle_mesh.elements(idx)));
                tri_guid_to_pressure_hashtable.Set(guid,i);
                //LOG::cout<<"N"<<node_id<<" set_tri "<<guid<<", "<<i<<std::endl;

                if(Inside_Domain(particles.X(v(1)))&&Inside_Domain(particles.X(v(2)))&&Inside_Domain(particles.X(v(3)))){domain_pressure_indices.Append(i);}
            }break;
        case SIMPLEX_TETRAHEDRON:{/*no pressure sample on tetrahedron center*/}break;
        }
    }

    //LOG::cout<<"N"<<node_id<<" all_p_to_send ";
    //for(int i=1;i<=all_pressures_to_send(SIMPLEX_PARTICLE).m;i++){
    //    LOG::cout<<particles.guid(pressures(all_pressures_to_send(SIMPLEX_PARTICLE)(i)).index)<<" ";
    //}
    //LOG::cout<<std::endl;

    ////send particles
    ARRAY<MPI::Request> requests;ARRAY<MPI_PACKAGE> packages;
    Clear_Arrays(send_particle_indices);Clear_Values(send_particle_numbers);
    Clear_Arrays(send_pressure_indices(SIMPLEX_PARTICLE));
    for(int i=1;i<=all_pressures_to_send(SIMPLEX_PARTICLE).m;i++){
        int pressure_id=all_pressures_to_send(SIMPLEX_PARTICLE)(i);
        int particle_id=all_elements_to_send(SIMPLEX_PARTICLE)(i);
        int particle_guid=particles.guid(particle_id);
        
        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(particles.X(particle_id));
        if(boundary_shrink_region_index!=0){
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                if(Valid_Particle_Pressure_Sample_To_Send(particle_id,boundary_region_indices(j),solver_mesh,all_pressures_on_nodes)){
                    send_pressure_indices(SIMPLEX_PARTICLE)(boundary_region_indices(j)).Append(pressure_id);
                    send_particle_indices(boundary_region_indices(j)).Append(particle_guid);
                }
                //else{if(node_id==1)LOG::cout<<"N"<<node_id<<" invalid particle index to send: "<<pressure_id<<", "<<particle_id<<", "<<particle_guid<<std::endl;}
            }
        }
    }
    for(int i=1;i<=send_particle_indices.m;i++)send_particle_numbers(i)=send_particle_indices(i).m;

    ////send segments
    Clear_Arrays(send_segs);Clear_Values(send_seg_numbers);
    Clear_Arrays(send_pressure_indices(SIMPLEX_SEGMENT));
    for(int i=1;i<=all_pressures_to_send(SIMPLEX_SEGMENT).m;i++){
        int pressure_id=all_pressures_to_send(SIMPLEX_SEGMENT)(i);
        int segment_id=all_elements_to_send(SIMPLEX_SEGMENT)(i);
        TV_INT2 segment_guid=Unique_Ordered(Element(particles.guid,solver_mesh.segment_mesh.elements(segment_id)));
        TV segment_center=Center(particles.X,solver_mesh.segment_mesh.elements(segment_id));

        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(segment_center);
        if(boundary_shrink_region_index!=0){
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                if(Valid_Segment_Pressure_Sample_To_Send(segment_id,boundary_shrink_region_index,solver_mesh)){
                    send_pressure_indices(SIMPLEX_SEGMENT)(boundary_region_indices(j)).Append(pressure_id);
                    send_segs(boundary_region_indices(j)).Append(segment_guid);
                }
                //else{LOG::cout<<"N"<<node_id<<" invalid segment index to send: "<<pressure_id<<", "<<segment_id<<", "<<segment_guid<<std::endl;}
            }
        }
    }
    for(int i=1;i<=send_segs.m;i++)send_seg_numbers(i)=send_segs(i).m;

    ////send triangles
    Clear_Arrays(send_tris);Clear_Values(send_tri_numbers);
    Clear_Arrays(send_pressure_indices(SIMPLEX_TRIANGLE));
    for(int i=1;i<=all_pressures_to_send(SIMPLEX_TRIANGLE).m;i++){
        int pressure_id=all_pressures_to_send(SIMPLEX_TRIANGLE)(i);
        int triangle_id=all_elements_to_send(SIMPLEX_TRIANGLE)(i);
        TV_INT3 triangle_guid=Unique_Ordered(Element(particles.guid,solver_mesh.triangle_mesh.elements(triangle_id)));
        TV triangle_center=Center(particles.X,solver_mesh.triangle_mesh.elements(triangle_id));

        int boundary_shrink_region_index=Boundary_Shrink_Region_Index(triangle_center);
        if(boundary_shrink_region_index!=0){
            const ARRAY<int>& boundary_region_indices=boundary_shrink_to_boundary_maps(boundary_shrink_region_index);
            for(int j=1;j<=boundary_region_indices.m;j++){
                if(Valid_Triangle_Pressure_Sample_To_Send(triangle_id,boundary_shrink_region_index,solver_mesh)){
                    send_pressure_indices(SIMPLEX_TRIANGLE)(boundary_region_indices(j)).Append(pressure_id);
                    send_tris(boundary_region_indices(j)).Append(triangle_guid);
                }
                //else{LOG::cout<<"N"<<node_id<<" invalid triangle index to send: "<<pressure_id<<", "<<triangle_id<<", "<<triangle_guid<<std::endl;}
            }
        }
    }
    //LOG::cout<<"!!!!!!!!!!!chk tri_p_indices N"<<node_id;
    //for(int i=1;i<=send_pressure_indices(SIMPLEX_TRIANGLE).m;i++){
    //    for(int j=1;j<=send_pressure_indices(SIMPLEX_TRIANGLE)(i).m;j++){
    //        LOG::cout<<send_pressure_indices(SIMPLEX_TRIANGLE)(i)(j)<<" ";
    //    }
    //}
    //LOG::cout<<" finish chk~~~~~~~~~~"<<std::endl;

    for(int i=1;i<=send_tris.m;i++)send_tri_numbers(i)=send_tris(i).m;

    ////update boundary pressure indices
    Clear_Arrays(boundary_pressure_indices);
    for(int i=1;i<=boundary_regions.m;i++){
        for(int j=1;j<=send_pressure_indices(SIMPLEX_PARTICLE)(i).m;j++)boundary_pressure_indices(i).Append(send_pressure_indices(SIMPLEX_PARTICLE)(i)(j));
        for(int j=1;j<=send_pressure_indices(SIMPLEX_SEGMENT)(i).m;j++)boundary_pressure_indices(i).Append(send_pressure_indices(SIMPLEX_SEGMENT)(i)(j));
        for(int j=1;j<=send_pressure_indices(SIMPLEX_TRIANGLE)(i).m;j++)boundary_pressure_indices(i).Append(send_pressure_indices(SIMPLEX_TRIANGLE)(i)(j));
    }

    //////send and receive particle numbers
    Mpi_Send_Values(send_particle_numbers,requests);
    Clear_Values(recv_particle_numbers);
    Mpi_Recv_Values(recv_particle_numbers,requests);
    MPI_UTILITIES::Wait_All(requests);

    //////send and receive segment numbers
    Mpi_Send_Values(send_seg_numbers,requests);
    Clear_Values(recv_seg_numbers);
    Mpi_Recv_Values(recv_seg_numbers,requests);
    MPI_UTILITIES::Wait_All(requests);

    ////send and receive triangle numbers
    Mpi_Send_Values(send_tri_numbers,requests);
    Clear_Values(recv_tri_numbers);
    Mpi_Recv_Values(recv_tri_numbers,requests);
    MPI_UTILITIES::Wait_All(requests);

    //LOG::cout<<"N"<<node_id<<" send_par_numbers "<<send_particle_numbers<<std::endl;
    //LOG::cout<<"N"<<node_id<<" send_seg_numbers "<<send_seg_numbers<<std::endl;
    //LOG::cout<<"N"<<node_id<<" send_tri_numbers "<<send_tri_numbers<<std::endl;

    //LOG::cout<<"N"<<node_id<<" recv_par_numbers "<<recv_particle_numbers<<std::endl;
    //LOG::cout<<"N"<<node_id<<" recv_seg_numbers "<<recv_seg_numbers<<std::endl;
    //LOG::cout<<"N"<<node_id<<" recv_tri_numbers "<<recv_tri_numbers<<std::endl;

    //LOG::cout<<"N"<<node_id<<" recv_tri_numbers_aft "<<recv_tri_numbers<<std::endl;

    ////allocate memory for received particles, segments, and triangles
    Clear_Arrays(recv_particle_indices);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_particle_numbers(i)>0){
            recv_particle_indices(i).Resize(recv_particle_numbers(i));
        }
    }
    Clear_Arrays(recv_segs);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_seg_numbers(i)>0){
            recv_segs(i).Resize(recv_seg_numbers(i));
        }
    }
    Clear_Arrays(recv_tris);
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_tri_numbers(i)>0){
            recv_tris(i).Resize(recv_tri_numbers(i));
        }
    }

    ////send and receive particle guids
    Mpi_Send_Arrays(send_particle_indices,send_particle_numbers,requests,packages);
    Mpi_Recv_Arrays(recv_particle_indices,recv_particle_numbers,requests,packages);
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);

    ////send and receive segment guids
    Mpi_Send_Arrays(send_segs,send_seg_numbers,requests,packages);
    Mpi_Recv_Arrays(recv_segs,recv_seg_numbers,requests,packages);
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);

    ////send and receive triangle guids
    Mpi_Send_Arrays(send_tris,send_tri_numbers,requests,packages);
    Mpi_Recv_Arrays(recv_tris,recv_tri_numbers,requests,packages);
    MPI_UTILITIES::Wait_All(requests);MPI_PACKAGE::Free_All(packages);

    ////translate particle guid to id
    Clear_Arrays(recv_pressure_indices(SIMPLEX_PARTICLE));
    Clear_Arrays(recv_pressure_indices(SIMPLEX_SEGMENT));
    Clear_Arrays(recv_pressure_indices(SIMPLEX_TRIANGLE));
    for(int i=1;i<=boundary_regions.m;i++){
        if(recv_particle_numbers(i)>0){
            recv_pressure_indices(SIMPLEX_PARTICLE)(i).Resize(recv_particle_indices(i).m);
            for(int j=1;j<=recv_particle_indices(i).m;j++){
                int particle_guid=recv_particle_indices(i)(j);
                int pressure_id=0;p_guid_to_pressure_hashtable.Get(particle_guid,pressure_id);
                if(pressure_id==0){LOG::cout<<std::endl<<"ABOUT TO FAIL, node id: "<<node_id<<", guid: "<<particle_guid<<std::endl<<std::endl;}
                PHYSBAM_ASSERT(pressure_id!=0);
                recv_pressure_indices(SIMPLEX_PARTICLE)(i)(j)=pressure_id;
            }
        }

        if(recv_seg_numbers(i)>0){
            recv_pressure_indices(SIMPLEX_SEGMENT)(i).Resize(recv_segs(i).m);
            for(int j=1;j<=recv_segs(i).m;j++){
                const TV_INT2& segment_guid=recv_segs(i)(j);
                int pressure_id=0;seg_guid_to_pressure_hashtable.Get(segment_guid,pressure_id);
                if(pressure_id==0)LOG::cout<<"N"<<node_id<<" invalid seg "<<segment_guid<<std::endl;
                PHYSBAM_ASSERT(pressure_id!=0);
                recv_pressure_indices(SIMPLEX_SEGMENT)(i)(j)=pressure_id;
            }
        }

        if(recv_tri_numbers(i)>0){
            recv_pressure_indices(SIMPLEX_TRIANGLE)(i).Resize(recv_tris(i).m);
            for(int j=1;j<=recv_tris(i).m;j++){
                const TV_INT3& triangle_guid=recv_tris(i)(j);
                int pressure_id=0;tri_guid_to_pressure_hashtable.Get(triangle_guid,pressure_id);
                PHYSBAM_ASSERT(pressure_id!=0);
                recv_pressure_indices(SIMPLEX_TRIANGLE)(i)(j)=pressure_id;
            }
        }
    }

    ////update ghost pressure indices
    Clear_Arrays(ghost_pressure_indices);
    for(int i=1;i<=boundary_regions.m;i++){
        for(int j=1;j<=recv_pressure_indices(SIMPLEX_PARTICLE)(i).m;j++)ghost_pressure_indices(i).Append(recv_pressure_indices(SIMPLEX_PARTICLE)(i)(j));
        for(int j=1;j<=recv_pressure_indices(SIMPLEX_SEGMENT)(i).m;j++)ghost_pressure_indices(i).Append(recv_pressure_indices(SIMPLEX_SEGMENT)(i)(j));
        for(int j=1;j<=recv_pressure_indices(SIMPLEX_TRIANGLE)(i).m;j++)ghost_pressure_indices(i).Append(recv_pressure_indices(SIMPLEX_TRIANGLE)(i)(j));
    }

    //LOG::cout<<"@@@@@@@@@ chk bdr pi: N"<<node_id<<" ";
    //for(int i=1;i<=boundary_regions.m;i++){
    //    for(int j=1;j<=boundary_pressure_indices(i).m;j++){
    //        LOG::cout<<boundary_pressure_indices(i)(j)<<" ";
    //    }
    //}
    //LOG::cout<<" finish chk bdr N"<<node_id<<" @@@@@@@@@@@@"<<std::endl;

    //LOG::cout<<"@@@@@@@@@ chk gst pi: N"<<node_id<<" ";
    //for(int i=1;i<=boundary_regions.m;i++){
    //    for(int j=1;j<=ghost_pressure_indices(i).m;j++){
    //        LOG::cout<<ghost_pressure_indices(i)(j)<<" ";
    //    }
    //}
    //LOG::cout<<" finish chk gst N"<<node_id<<" @@@@@@@@@@@@"<<std::endl;
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Update_Vertex_Element_Hashtable_In_Boundary_Regions()
{
    Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(mesh.tetrahedron_mesh.elements,vtx_tet_hashtable);
    Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(mesh.triangle_mesh.elements,vtx_tri_hashtable);
    Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(mesh.segment_mesh.elements,vtx_seg_hashtable);
    //Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(mesh.rim_segment_mesh.elements,vtx_rim_seg_hashtable);
}

template<class TV> template<class TV_INTN>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(const ARRAY<TV_INTN>& elements,/*result*/HASHTABLE<int,ARRAY<TV_INTN> >& vtx_e_hashtable)
{
    vtx_e_hashtable.Remove_All();
    for(int i=1;i<=elements.m;i++){
        const TV_INTN& vtx=elements(i);
        for(int d=1;d<=TV_INTN::m;d++){
            int bdry_shrink_idx=Boundary_Shrink_Region_Index(particles.X(vtx(d)));
            if(bdry_shrink_idx!=0){
                TV_INTN vtx_guid=Element(particles.guid,vtx);
                ARRAY<TV_INTN>& nbs=vtx_e_hashtable.Get_Or_Insert(particles.guid(vtx(d)));
                nbs.Append(vtx_guid);
            }
        }
    }
}

template<class TV>
void PhysBAM::MPI_CODIMENSIONAL<TV>::Update_Guid_Vertex_Hashtable()
{
    guid_vertex_hashtable.Remove_All();
    for(int i=1;i<=particles.X.m;i++){
        guid_vertex_hashtable.Set(particles.guid(i),i);
    }
}

#define INSTANTIATION_HELPER(T,d)                                        \
    template class MPI_CODIMENSIONAL<VECTOR<T,d> >;                \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<int>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<bool>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<VECTOR<int,d> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<T>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<VECTOR<T,d> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Pressure_Values(ARRAY_VIEW<int>&,ARRAY<ARRAY<int> >&,ARRAY<ARRAY<int> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Pressure_Values(ARRAY_VIEW<VECTOR<T,d> >&,ARRAY<ARRAY<int> >&,ARRAY<ARRAY<int> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Pressure_Values(ARRAY_VIEW<T>&,ARRAY<ARRAY<int> >&,ARRAY<ARRAY<int> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Exchange_Boundary_And_Ghost_Pressure_Values(VECTOR_ND<T>&,ARRAY<ARRAY<int> >&,ARRAY<ARRAY<int> >&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Send_Arrays(ARRAY<T>&,const int,ARRAY<MPI::Request>&,ARRAY<MPI_PACKAGE>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Send_Arrays(ARRAY<int>&,const int,ARRAY<MPI::Request>&,ARRAY<MPI_PACKAGE>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Recv_Arrays(ARRAY<ARRAY<T> >&,const ARRAY<int>&,ARRAY<MPI::Request>&,ARRAY<MPI_PACKAGE>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Recv_Arrays(ARRAY<ARRAY<int> >&,const ARRAY<int>&,ARRAY<MPI::Request>&,ARRAY<MPI_PACKAGE>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Send_Values(T&,ARRAY<MPI::Request>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Send_Values(int&,ARRAY<MPI::Request>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Recv_Values(ARRAY<T>&,ARRAY<MPI::Request>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Recv_Values(ARRAY<int>&,ARRAY<MPI::Request>&); \
    template void MPI_CODIMENSIONAL<VECTOR<T,d> >::Mpi_Broadcast_Arrays(ARRAY<int>&,ARRAY<ARRAY<int> >&); \
    template float MPI_CODIMENSIONAL<VECTOR<T,d> >::Sum(float); \
    template double MPI_CODIMENSIONAL<VECTOR<T,d> >::Sum(double); \
    template int MPI_CODIMENSIONAL<VECTOR<T,d> >::Sum(int); \
    template float MPI_CODIMENSIONAL<VECTOR<T,d> >::Max(float); \
    template double MPI_CODIMENSIONAL<VECTOR<T,d> >::Max(double); \
    template int MPI_CODIMENSIONAL<VECTOR<T,d> >::Max(int); \
    template float MPI_CODIMENSIONAL<VECTOR<T,d> >::Min(float); \
    template double MPI_CODIMENSIONAL<VECTOR<T,d> >::Min(double); \
    template int MPI_CODIMENSIONAL<VECTOR<T,d> >::Min(int);

INSTANTIATION_HELPER(float,3)
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
INSTANTIATION_HELPER(double,3)
#endif
