//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __MPI_CODIMENSIONAL__
#define __MPI_CODIMENSIONAL__

#include <PhysBAM_Tools/Parallel_Computation/PARTITION_ID.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_UNIFORM_GRID.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_PACKAGE.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_UTILITIES.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"

namespace MPI{class Group;class Intracomm;class Request;class Status;class Op;}

namespace PhysBAM{
template<class TV>
class MPI_CODIMENSIONAL:public NONCOPYABLE
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef VECTOR<int,1> TV_INT1;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef RANGE<VECTOR<int,1> > RANGE_INT1;
    typedef ARRAY<ARRAY<int> > ARRAY2_INT;typedef ARRAY<ARRAY<TV_INT2> > ARRAY2_INT2;typedef ARRAY<ARRAY<TV_INT3> > ARRAY2_INT3;typedef ARRAY<ARRAY<TV_INT4> > ARRAY2_INT4;
    typedef ARRAY<int> ARRAY_INT;typedef ARRAY<RANGE_INT1> ARRAY_RANGE;

public:
    MPI_UNIFORM_GRID<GRID<TV> >& mpi_grid;
    int number_of_ghost_cells;    
    int node_id;
    int region_count_offset; ////each node will start counting its regions region_count_offset numbers away from the previous node
    int first_region_number;
    T dX;
    RANGE<TV> domain;
    RANGE<TV> domain_ext;
    RANGE<TV> domain_shrink;
    ARRAY<RANGE<TV> > boundary_regions;
    ARRAY<RANGE<TV> > boundary_shrink_regions;  ////boundary regions without overlap
    ARRAY2_INT boundary_shrink_to_boundary_maps;    ////mapping indices from boundary shrink regions to boundary regions
    ARRAY<int> neighbor_ranks;
    ARRAY<TV_INT> neighbor_directions;
    
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    CODIMENSIONAL_MESH& mesh;

    ARRAY2_INT boundary_particle_indices;
    ARRAY2_INT ghost_particle_indices;
    ARRAY<int> domain_particle_indices;

    HASHTABLE<int,int> guid_vertex_hashtable;   ////map guid to id
    HASHTABLE<int,int> p_guid_to_pressure_hashtable; ////map guid to pressure id
    HASHTABLE<TV_INT2,int> seg_guid_to_pressure_hashtable;  ////map seg guid to pressure id
    HASHTABLE<TV_INT3,int> tri_guid_to_pressure_hashtable;  ////map tri guid to pressure id

    HASHTABLE<int,ARRAY<TV_INT4> > vtx_tet_hashtable;   ////map vertex to incident tet list, stored in guid
    HASHTABLE<int,ARRAY<TV_INT3> > vtx_tri_hashtable;
    HASHTABLE<int,ARRAY<TV_INT2> > vtx_seg_hashtable;
    //HASHTABLE<int,ARRAY<TV_INT2> > vtx_rim_seg_hashtable;

    ////mpi communication temporal data structures
    //ARRAY2_INT send_particle_indices;ARRAY<int> send_particle_numbers;ARRAY2_INT recv_particle_indices;ARRAY<int> recv_particle_numbers;
    //ARRAY<RANGE_INT1> recv_particle_index_ranges;
    //ARRAY<ARRAY<TV_INT4> > send_tets;ARRAY<int> send_tet_numbers;ARRAY<ARRAY<TV_INT4> > recv_tets;ARRAY<int> recv_tet_numbers;
    //ARRAY<ARRAY<TV_INT3> > send_tris;ARRAY<int> send_tri_numbers;ARRAY<ARRAY<TV_INT3> > recv_tris;ARRAY<int> recv_tri_numbers;
    //ARRAY<ARRAY<TV_INT2> > send_segs;ARRAY<int> send_seg_numbers;ARRAY<ARRAY<TV_INT2> > recv_segs;ARRAY<int> recv_seg_numbers;
    
    MPI_CODIMENSIONAL(MPI_UNIFORM_GRID<GRID<TV> >& mpi_grid_input,const int number_of_ghost_cells_input,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles_input,CODIMENSIONAL_MESH& mesh_input);
    ~MPI_CODIMENSIONAL(){}
        
    void Initialize_From_External_Mesh_And_Positions(const CODIMENSIONAL_MESH& ext_mesh,const ARRAY_VIEW<TV>& ext_X);

    ////exchange indices
    void Update_Boundary_And_Ghost_Particle_Indices();
    void Exchange_Particle_Indices(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers,bool exc_guid=false);
    void Exchange_Particle_Indices2(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers);    ////comm by guid and translate to id
    void Exchange_And_Append_Particle_Indices(const ARRAY<int>& p_to_exc,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,ARRAY<RANGE_INT1>& recv_p_ranges,ARRAY<int>& recv_p_numbers);

    ////exchange particle attributes
    void Exchange_Particle_Attributes_All(ARRAY2_INT& send_p,ARRAY<RANGE_INT1>& recv_p_ranges);
    void Exchange_Particle_Attributes_All(ARRAY2_INT& send_p,ARRAY2_INT& recv_p);
    template<class ARRAY_T2> void Exchange_Values(ARRAY_T2& X,ARRAY2_INT& send_p,ARRAY<RANGE_INT1>& recv_p_ranges);
    template<class ARRAY_T2> void Exchange_Values(ARRAY_T2& X,ARRAY2_INT& send_p,ARRAY2_INT& recv_p);
    template<class T2> void Exchange_Boundary_And_Ghost_Particle_Attributes(ARRAY_VIEW<T2>& X);

    ////add and remove particles
    void Remove_Particles(const ARRAY<int>& p_to_rmv,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,
        ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers);
    void Add_Particles(const ARRAY<int>& p_added,ARRAY2_INT& send_p,ARRAY<int>& send_p_numbers,
        ARRAY2_INT& recv_p,ARRAY<int>& recv_p_numbers,ARRAY<RANGE_INT1>& recv_p_ranges);

    ////exchange elements
    void Exchange_Elements_Simple(const ARRAY2_INT& send_p); 
    template<class TV_INTN> void Exchange_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_to_exc,ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers,
        ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers);
    template<class TV_INTN> void Remove_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_to_rmv,ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers,
        ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers);
    template<class TV_INTN> void Add_Elements(ARRAY<TV_INTN>& elements,const ARRAY<int>& e_added,ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<int>& send_e_numbers,
        ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& recv_e_numbers);

    ////exchange particles and elements
    void Exchange_Particles_And_Elements_Simple();////first remove then add
    void Exchange_Particles_And_Elements(ARRAY<int>* p_added,ARRAY<int>* p_to_rmv,ARRAY<int>* p_to_exc,ARRAY<int>* tet_added,ARRAY<int>* tet_to_rmv,
        ARRAY<int>* tri_added,ARRAY<int>* tri_to_rmv,ARRAY<int>* seg_added,ARRAY<int>* seg_to_rmv);    
    void Remove_Out_Of_Domain_Particles_And_Elements();

    ////exchange pressure samples
    void Update_Boundary_And_Ghost_Pressure_Indices(const ARRAY<CODIMENSIONAL_PRESSURE>& pressures,const CODIMENSIONAL_MESH& solver_mesh,bool all_pressures_on_nodes,
        ARRAY2_INT& boundary_pressure_indices,ARRAY2_INT& ghost_pressure_indices,ARRAY<int>& domain_pressure_indices);
    template<class ARRAY_T2> void Exchange_Boundary_And_Ghost_Pressure_Values(ARRAY_T2& X,ARRAY2_INT& boundary_pressure_indices,ARRAY2_INT& ghost_pressure_indices);

    ////update auxiliary structures
    void Update_Guid_Vertex_Hashtable();
    void Update_Vertex_Element_Hashtable_In_Boundary_Regions();

    ////mpi communication functions
    template<class T2> void Mpi_Send_Arrays(ARRAY<ARRAY<T2> >& send_arrays,const ARRAY<int>& send_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages);
    template<class T2> void Mpi_Recv_Arrays(ARRAY<ARRAY<T2> >& recv_arrays,const ARRAY<int>& recv_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages);
    template<class T2> void Mpi_Send_Values(ARRAY<T2>& send_values,ARRAY<MPI::Request>& requests);
    template<class T2> void Mpi_Recv_Values(ARRAY<T2>& recv_values,ARRAY<MPI::Request>& requests);
    template<class T2> void Mpi_Broadcast_Send_Arrays(ARRAY<T2>& broadcast_send_array,const int send_number,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages);
    template<class T2> void Mpi_Broadcast_Recv_Arrays(ARRAY<ARRAY<T2> >& broadcast_recv_arrays,const ARRAY<int>& recv_numbers,ARRAY<MPI::Request>& requests,ARRAY<MPI_PACKAGE>& packages);
    template<class T2> void Mpi_Broadcast_Send_Values(T2& broadcast_send_value,ARRAY<MPI::Request>& requests);
    template<class T2> void Mpi_Broadcast_Recv_Values(ARRAY<T2>& broadcast_recv_values,ARRAY<MPI::Request>& requests);
    template<class T2> void Mpi_Broadcast_Arrays(ARRAY<T2>& broadcast_send_array,ARRAY<ARRAY<T2> >& broadcast_recv_arrays);

    template<class T2> void Realloc_Arrays(ARRAY<ARRAY<T2> >& alloc_arrays,const ARRAY<int>& numbers); 
    template<class T2> void Clear_Arrays(ARRAY<ARRAY<T2> >& arrays){arrays.Resize(boundary_regions.m);for(int i=1;i<=arrays.m;i++)arrays(i).Remove_All();}
    template<class T2> void Clear_Values(ARRAY<T2>& values){values.Resize(boundary_regions.m);values.Fill((T2)0);}
    void Free_All(ARRAY<MPI::Request>& requests){for(int i=1;i<=requests.m;i++)requests(i).Free();requests.Clean_Memory();}

    template<class T2> T2 Sum(const T2 local_sum){T2 global_sum=(T2)0;MPI_UTILITIES::Reduce(local_sum,global_sum,MPI::SUM,*mpi_grid.comm);return global_sum;}
    template<class T2> T2 Max(const T2 local_max){T2 global_max=(T2)0;MPI_UTILITIES::Reduce(local_max,global_max,MPI::MAX,*mpi_grid.comm);return global_max;}
    template<class T2> T2 Min(const T2 local_min){T2 global_min=(T2)0;MPI_UTILITIES::Reduce(local_min,global_min,MPI::MIN,*mpi_grid.comm);return global_min;}
    void Barrier(){MPI_Barrier(*mpi_grid.comm);}

    ////position
    bool Inside_Domain(const TV& pos) const {return domain.Lazy_Inside_Half_Open(pos);}
    bool Inside_Domain_Ext(const TV& pos) const {return domain_ext.Lazy_Inside_Half_Open(pos);}
    bool Inside_Domain_Boundary(const TV& pos) const {return domain.Lazy_Inside_Half_Open(pos)&&!domain_shrink.Lazy_Inside_Half_Open(pos);}
    bool Inside_Domain_Ghost(const TV& pos) const {return domain_ext.Lazy_Inside_Half_Open(pos)&&!domain.Lazy_Inside_Half_Open(pos);}
    bool Inside_Domain_Shrink(const TV& pos) const {return domain_shrink.Lazy_Inside_Half_Open(pos);}
    
    TV_INT Neighbor_Offset(const TV& pos,const RANGE<TV>& box) const; 
    int Boundary_Shrink_Region_Index(const TV& pos) const;
    int Boundary_Shrink_Region_Index(const TV_INT& neighbor_offset) const;
    RANGE<TV> Boundary_Shrink_Region(const TV_INT& nb_offset) const;

    ////guid
    template<class TV_INTN> bool Contains_All_Guid(const TV_INTN& vtx_guid) const
    {bool contains_all=true;for(int d=1;d<=TV_INTN::m;d++)if(!guid_vertex_hashtable.Contains(vtx_guid(d))){contains_all=false;break;}return contains_all;}
    template<class TV_INTN> TV_INTN Guid_To_Vtx(const TV_INTN& vtx_guid) const
    {TV_INTN vtx;for(int d=1;d<=TV_INTN::m;d++){guid_vertex_hashtable.Get(vtx_guid(d),vtx(d));}return vtx;}
    int Guid_To_Vtx(const int vtx_guid) const{int vtx=0;guid_vertex_hashtable.Get(vtx_guid,vtx);return vtx;}

    ////helper functions
    template<class TV_INTN> void Initialize_Mesh_Elements_From_Global_External_Mesh_Helper(ARRAY<TV_INTN>& elements,const ARRAY<TV_INTN>& external_elements,const ARRAY<int>& guid);
    template<class TV_INTN> void Update_Vertex_Element_Hashtable_In_Boundary_Regions_Helper(const ARRAY<TV_INTN>& elements,HASHTABLE<int,ARRAY<TV_INTN> >& vtx_tet_hashtable);
    template<class TV_INTN> void Exchange_Mesh_Elements_Helper(ARRAY<TV_INTN>& elements,HASHTABLE<int,ARRAY<TV_INTN> >& vtx_e_hashtable,const ARRAY2_INT& send_p,
        ARRAY<ARRAY<TV_INTN> >& send_e,ARRAY<ARRAY<TV_INTN> >& recv_e,ARRAY<int>& send_e_numbers,ARRAY<int>& recv_e_numbers);
    void Initialize_Boundary_Shrink_Regions(const ARRAY<RANGE<TV> >& boundary_regions,ARRAY<RANGE<TV> >& boundary_shrink_regions);
    void Shrink_Edge_Region_Helper(RANGE<TV>& shrink_region,const RANGE<TV>& min_region,const RANGE<TV>& max_region,const int axis);
    void Shrink_Face_Region_Helper(RANGE<TV>& shrink_region,const RANGE<TV>& min_region,const RANGE<TV>& max_region,const int axis_1,const int axis_2);

    bool Valid_Particle_Pressure_Sample_To_Send(const int p,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh,bool all_pressures_on_nodes);
    bool Valid_Triangle_Pressure_Sample_To_Send(const int t,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh);
    bool Valid_Segment_Pressure_Sample_To_Send(const int s,const int boundary_region_idx,const CODIMENSIONAL_MESH& solver_mesh);
};
}

#endif
