//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Utilities/FILE_UTILITIES.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "../Solver/CODIMENSIONAL_FLUID_DRIVER.h"
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::CODIMENSIONAL_FLUID_DRIVER()
    :test(1),output_directory(TV::m==2?"output":"output3"),model_directory("../model"),stream_type(T()),first_frame(0),last_frame(4000),current_frame(0),restart_frame(0),
    write_substep(false),current_substep_frame(0),last_substep_frame(-1),frame_rate(200),time(0),current_time(0),cfl((T)1.0),use_mpi(false),
    length_scale((T).01),adaptive_length_scale_min(length_scale),adaptive_length_scale_min2(length_scale),thickness_scale((T).1*length_scale),rim_thickness_scale(thickness_scale),max_thickness_factor((T)1.5),min_thickness_factor((T).1),
    max_rim_thickness_factor((T)1.5),min_rim_thickness_factor((T).1),surface_tension_factor((T)1),gravity_factor((T)1),length_scale_min_factor((T)1),
    scale((T)1),gravity(TV::Axis_Vector(2)*(T)-9.8),density((T)1000.),surface_tension((T).073),tri_surface_tension((T)-1),seg_surface_tension((T)-1),kinematic_viscosity((T)1e-6),kinematic_viscosity_min(kinematic_viscosity),rigid_viscosity((T)10),
    elastic_modulus((T)1),plastic_clamp_min((T).6),plastic_clamp_max((T)1.4),friction_magnitude((T).5),spatial_partition_voxel_size_coef((T)2),time_scale((T)1),environment_temperature((T)0),heat_conductivity((T)1),max_iter_number(2000),fixed_cfl((T)0),
    remesh(true),has_gravity(true),has_surface_tension(true),has_viscosity(true),
    use_implicit_viscosity(true),use_shear_dependent_variable_viscosity(false),has_elasticity(false),has_plasticity(false),use_implicit_elasticity(false),
    has_fiber_fluid_coupling_force(false),has_adhesion(false),has_friction(false),has_kinematic_particles(false),has_incompressibility(true),has_rim(true),use_implicit_all(false),use_explicit_all(false),
    use_centric_gravity(false),use_spatial_partition(false),check_negative_volume_tetrahedra(false),use_inactive_particle(false),
    use_correcting_particle_position_on_solids(true),use_non_slip_on_solids(false),use_adaptive_length_scale_on_codim(false),use_adaptive_length_scale_on_levelset(false),use_adaptive_length_scale_on_rim(false),
    use_adaptive_length_scale_on_fiber(false),use_adaptive_length_scale_based_on_test(false),
    has_rim_segments_for_viscosity(false),has_rim_triangles_for_viscosity(true),use_heat_transfer(false),use_paint_trace_boundary(false),use_friction_boundary(false),use_viscosity_dependent_friction(false),
    use_damped_friction(true),use_correcting_rigid_normal(false),exp_part_viscosity(true),
    rigid_body_collection(new RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES<TV>(),new COLLISION_GEOMETRY_COLLECTION<TV>()),rigid_standard_tests(model_directory,stream_type,rigid_body_collection),tank(0),rigid_body_speed_factor((T)1),use_inside_sim_box(false)
{
    Initialize_Particles();
    Initialize_Read_Write_General_Structures();
    Use_Rim(has_rim);
    random.Set_Seed(12010221);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Initialize()
{
    Initialize_Test_Parameters();
    Initialize_Mpi();
    Initialize_Mesh();
    Initialize_Solver();
    Initialize_Kinematic_Positions();

    if(restart_frame==0) Write_Output_Files(current_frame);
    else{Restart();Initialize_After_Restart();}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Execute_Main_Program()
{
    Simulate_To_Frame(last_frame);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Simulate_To_Frame(const int frame)
{
    while(current_frame<frame){
        current_frame++;
        Advance_To_Target_Time(Time_At_Frame(current_frame));
        if(!write_substep) Write_Output_Files(current_frame);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Advance_To_Target_Time(const T target_time)
{
    int timer_id=example.mpi_codim?example.mpi_codim->node_id:1;TIMER timer;timer.Reset_Time(timer_id);
    bool done=false;

    if(example.mesher.use_particle_parents)example.mesher.Reset_Particle_Parents();

    for(int substep=1;!done;substep++){
        if(write_substep){
            if(example.mesher.use_particle_parents)example.mesher.Reset_Particle_Parents();
            current_substep_frame++;
        }

        T dt=CFL();
        if(fixed_cfl>(T)0) dt=fixed_cfl;    ////for timestep test

        if(time+dt>=target_time){dt=target_time-time;done=true;}
        else if(time+(T)2*dt>=target_time){dt=(T).5*(target_time-time);}
        
        Advance_One_Timestep(dt,time);
        if(write_substep){Write_Output_Files(current_substep_frame);}
        time+=dt;
    }
    double elapsed_time=timer.Peek_Time(timer_id)/1.e3;
    if(!example.mpi_codim||example.mpi_codim->node_id==1)LOG::cout<<"adv to time: "<<time<<", sim time: "<<elapsed_time<<std::endl;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::CFL()
{
    T convection_cfl=(T)1e-10;
    for(int i=1;i<=example.particles.X.m;i++){
        if(example.particles.id(i)==0)continue;   ////ignore the discrete particles in cfl
        T dx=example.particles.length_scale(i)*(T).25;
        //if(example.mpi_codim->node_id==1)LOG::cout<<"N"<<example.mpi_codim->node_id<<": "<<i<<": V "<<example.particles.V(i).L1_Norm()<<", dx: "<<dx<<std::endl;
        convection_cfl=max(convection_cfl,example.particles.V(i).L1_Norm()/dx);
    }
    T net_force_cfl=(T)1e-10;
    for(int i=1;i<=example.particles.X.m;i++){
        T dx=example.particles.length_scale(i)*(T).25;
        //if(example.mpi_codim->node_id==1)LOG::cout<<"N"<<example.mpi_codim->node_id<<": "<<i<<": F "<<example.particles.F(i).L1_Norm()<<", M "<<example.particles.mass(i)<<", dx "<<dx<<std::endl;
        net_force_cfl=max(net_force_cfl,example.particles.F(i).L1_Norm()/(example.particles.mass(i)*dx));
        //LOG::cout<<i<<" F: "<<example.particles.F(i)<<", mass: "<<example.particles.mass(i)<<std::endl;
    }
    T total_cfl=(T).5*(convection_cfl+sqrt(pow<2>(convection_cfl)+4*net_force_cfl));
    T dt=(T)FLT_MAX;
    if(total_cfl>(T)0)dt=(T)1.0/total_cfl;
    //if(example.mpi_codim->node_id==1)LOG::cout<<"N"<<example.mpi_codim->node_id<<" dt "<<dt<<": "<<total_cfl<<", n "<<net_force_cfl<<", c "<<convection_cfl<<std::endl;
    if(example.mpi_codim)dt=example.mpi_codim->Min(dt);
    //LOG::cout<<"cfl "<<convection_cfl<<", "<<net_force_cfl<<", "<<total_cfl<<", dt "<<dt<<std::endl;
    T cfl_dt=cfl*dt;
    if(check_negative_volume_tetrahedra)cfl_dt=Valid_Timestep_For_Tetrahedron_Mesh(cfl_dt);
    return cfl_dt;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Valid_Timestep_For_Tetrahedron_Mesh(const T dt)
{
    ////TODO: didn't consider solid position correction
    ARRAY<TV_INT4>& elements=example.geometrical_codimensional_mesh.tetrahedron_mesh.elements;
    HASHTABLE<int> neg_tet_hashtable;
    for(int i=1;i<=elements.m;i++){
        const TV_INT4& v=elements(i);
        if(example.mpi_codim&&!Has_Vertex_Inside_Region(example.particles.X,v,example.mpi_codim->domain,(T)0)){continue;}
        VECTOR<TV,4> new_pos;for(int d=1;d<=4;d++)new_pos(d)=example.particles.X(v(d))+example.particles.V(v(d))*dt;
        bool is_negative_vol_tet=Is_Negative_Volume_Tetrahedron(new_pos(1),new_pos(2),new_pos(3),new_pos(4));
        if(is_negative_vol_tet)neg_tet_hashtable.Set(i);
    }
    T subdivided_dt=dt;int max_subdivide_num=10;int subdivide_num=0;
    bool has_neg_vol_tet=neg_tet_hashtable.Size()>0;
    while(has_neg_vol_tet&&subdivide_num<max_subdivide_num){
        subdivided_dt*=(T).5;subdivide_num++;bool has_negative_vol_tet=false;
        for(HASHTABLE_ITERATOR<int> iter(neg_tet_hashtable);iter.Valid();iter.Next()){
            int t=iter.Key();const TV_INT4& v=elements(t);
            VECTOR<TV,4> new_pos;for(int d=1;d<=4;d++)new_pos(d)=example.particles.X(v(d))+example.particles.V(v(d))*subdivided_dt;
            bool is_negative_vol_tet=Is_Negative_Volume_Tetrahedron(new_pos(1),new_pos(2),new_pos(3),new_pos(4));
            if(is_negative_vol_tet){has_negative_vol_tet=true;break;}
        }
        if(!has_negative_vol_tet)break;
    }
    if(example.mpi_codim)subdivided_dt=example.mpi_codim->Min(subdivided_dt);
    if(subdivided_dt<dt){
        LOG::cout<<"subdivided_dt: "<<subdivided_dt<<", dt: "<<dt<<std::endl;
        LOG::cout<<"neg tets: ";
        for(HASHTABLE_ITERATOR<int> iter(neg_tet_hashtable);iter.Valid();iter.Next()){
            int t=iter.Key();const TV_INT4& v=elements(t);const TV_INT4& guid=Element(example.particles.guid,v);
            LOG::cout<<"v: "<<v<<" guid "<<guid<<", ";
        }LOG::cout<<", N"<<(example.mpi_codim?example.mpi_codim->node_id:0)<<std::endl;
    }
    return subdivided_dt;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Advance_One_Timestep(const T dt,const T time)
{
    ////const int dbg_guid=52351;int dbg_step=1;
    ////Debug_Output_Particle_Guid(example.particles,dbg_guid,dbg_step++);

    static int time_step_n=0;
    time_step_n++;//if(time_step_n>=6)PHYSBAM_FATAL_ERROR("~~~~~~~~~~~Finish~~~~~~~~~~~~~");

    static double avg_conv=0.;static double avg_mesh=0.;static double avg_force=0.;static double avg_incomp=0.;static double avg_total=0.;
    double conv_t=0.;double mesh_t=0.;double force_t=0.;double incomp_t=0.;double total_t=0.;
    TIMER timer;int id=10;

    timer.Reset_Time(id);
    Convection(dt,time);
    //Debug_Output_Particle_Guid(example.particles,dbg_guid,dbg_step++);
    Update_Kinematic_Positions(dt,time);
    
    //Debug_Output_Particle_Guid(example.particles,dbg_guid,dbg_step++);

    if(use_mpi){
        example.mpi_codim->Exchange_Particles_And_Elements_Simple();
        example.mesher.Update_Rim_Segment_Mesh();
    }
    conv_t=timer.Peek_And_Reset_Time(id)/1000.;
    
    Update_Mesh();
    mesh_t=timer.Peek_And_Reset_Time(id)/1000.;
    //Debug_Output_Particle_Guid(example.particles,dbg_guid,dbg_step++);

    Update_Position_Based_State(time,true);
    if(has_elasticity){
        Update_Boundary_Conditions(example.elasticity_force,false);
        //example.elasticity_force.Update_Deformation_State(dt,time);
        example.elasticity_force.Update_Deformation_State_Part_One(dt,time);
    }

    ////Update_Boundary_Conditions();
    Update_Target_Divergence();
    Add_Velocity_Independent_Forces(dt,time);//Apply_Forces(example.particles.F,dt,time);
    Update_Boundary_Conditions();
    if(has_viscosity&&use_shear_dependent_variable_viscosity)example.viscosity_force.Update_Position_Based_State(time,true);
    Add_Velocity_Dependent_Forces(dt,time);
    if(use_heat_transfer){Update_Heat_Transfer(dt,time);}

    force_t=timer.Peek_And_Reset_Time(id)/1000.;

    if(has_incompressibility){
        ARRAY<TV> old_u(example.particles.X.m);for(int i=1;i<=example.particles.X.m;i++)old_u(i)=example.particles.V(i);
        example.incompressible.Make_Divergence_Free(dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.pressure_force(i)=((example.particles.V(i)-old_u(i))*example.particles.mass(i))/dt;
    }
    //Enforce_Solid_Boundary_Velocities();
    incomp_t=timer.Peek_And_Reset_Time(id)/1000.;

    if(has_elasticity){
        example.elasticity_force.Update_Deformation_State_Part_Two(dt,time);
    }

    if(time_step_n>1){
        total_t+=(conv_t+mesh_t+force_t+incomp_t);
        avg_conv+=conv_t;
        avg_mesh+=mesh_t;
        avg_force+=force_t;
        avg_incomp+=incomp_t;
        avg_total+=total_t;
        LOG::cout<<"time step "<<time_step_n<<std::endl;
        LOG::cout<<"conv_t\t "<<conv_t<<" \tmesh_t\t "<<mesh_t<<" \tforce_t\t "<<force_t<<" \tincomp_t\t "<<incomp_t<<" \ttotal_t\t "<<total_t<<std::endl;
        LOG::cout<<"conv_s\t "<<avg_conv<<" \tmesh_s\t "<<avg_mesh<<" \tforce_s\t "<<avg_force<<" \tincomp_s\t "<<avg_incomp<<" \ttotal_s\t "<<avg_total<<std::endl;
        LOG::cout<<"convcr\t "<<conv_t/total_t<<" \tmeshvr\t "<<mesh_t/total_t<<" \tforcevr\t "<<force_t/total_t<<" \tincompcr\t "<<incomp_t/total_t<<std::endl;
        LOG::cout<<"conv_r\t "<<avg_conv/avg_total<<" \tmesh_r\t "<<avg_mesh/avg_total<<" \tforce_r\t "<<avg_force/avg_total<<" \tincomp_r\t "<<avg_incomp/avg_total<<std::endl;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Write_Output_Files(const int frame)
{
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);

    FILE_UTILITIES::Create_Directory(output_directory);
    FILE_UTILITIES::Create_Directory(output_directory+"/common");
    FILE_UTILITIES::Create_Directory(output_directory+"/"+f);
    example.Write_Output_Files(frame);
    if(rigid_body_collection.rigid_body_particle.X.m>0)rigid_body_collection.Write(stream_type,output_directory,frame);
    if(kinematic_particle_velocity_hashtable.Size()>0)FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/kinematic_particles",kinematic_particle_velocity_hashtable);
    //if(tubes.m>0)FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/tube_eulerian_particles",tubes(1)->eulerian_particles);
    if(write_substep){FILE_UTILITIES::Write_To_Text_File(output_directory+"/"+f+"/substep_time",time,"\n");}
    if(frame==0) FILE_UTILITIES::Write_To_Text_File(output_directory+"/common/first_frame",frame,"\n");
    if(frame==0&&example.mpi_codim)FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/common/grid",example.mpi_codim->mpi_grid.local_grid);

    FILE_UTILITIES::Write_To_Text_File(output_directory+"/common/last_frame",frame,"\n");

    if(use_mpi)LOG::cout<<"---Node "<<example.mpi_codim->node_id<<" write frame "<<frame<<" ---"<<std::endl;
    else{if(write_substep) LOG::cout<<"--- Write substep "<<frame<<" ---"<<std::endl;else LOG::cout<<"--- Write frame "<<frame<<" ---"<<std::endl;}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Read_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    example.Read_Output_Files(frame);
    if(rigid_bodies.m>0)rigid_body_collection.Read(stream_type,output_directory,frame);
    std::string file_name=output_directory+"/"+f+"/kinematic_particles";
    if(FILE_UTILITIES::File_Exists(file_name)){
        FILE_UTILITIES::Read_From_File(stream_type,file_name,kinematic_particle_velocity_hashtable);
        LOG::cout<<"read kin p: "<<kinematic_particle_velocity_hashtable.Size()<<std::endl;
    }

    //std::string file_name=output_directory+"/"+f+"/tube_eulerian_particles";
    //if(FILE_UTILITIES::File_Exists(file_name)&&tubes.m>0){
    //    FILE_UTILITIES::Read_From_File(stream_type,file_name,tubes(1)->eulerian_particles);
    //    LOG::cout<<"read tube eulerian p: "<<tubes(1)->eulerian_particles.Size()<<std::endl;
    //}

    ////TODO: read mpi_grid?
    if(use_mpi)LOG::cout<<"---Node "<<example.mpi_codim->node_id<<" read frame "<<frame<<" ---"<<std::endl;
    else LOG::cout<<"--- Read frame "<<frame<<" ---"<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Set_Output_Directory(const std::string output_directory_input)
{
    output_directory=output_directory_input;
    example.output_directory=output_directory_input;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Initialize_Mpi()
{
    TV_INT global_grid_counts=TV_INT(global_grid_domain.Edge_Lengths()/((T)1.11*length_scale));
    LOG::cout<<"global grid counts: "<<global_grid_counts<<std::endl;
    global_grid.Initialize(global_grid_counts,global_grid_domain,false);
    number_of_ghost_cells=3;

    if(use_mpi){
        if(processes_per_dimension==TV_INT::Constant_Vector(0)){processes_per_dimension=TV_INT::Constant_Vector(1);processes_per_dimension.x*=2;}
        example.mpi_grid=new MPI_UNIFORM_GRID<GRID<TV> >(global_grid,number_of_ghost_cells,false,processes_per_dimension);
        example.mpi_codim=new MPI_CODIMENSIONAL<TV>(*example.mpi_grid,number_of_ghost_cells,example.particles,example.geometrical_codimensional_mesh);
        output_directory=output_directory+STRING_UTILITIES::string_sprintf("/%d",(example.mpi_grid->rank+1));
        example.mesher.Initialize_Mpi(example.mpi_codim);
        example.surface_tension_force.Initialize_Mpi(example.mpi_codim);
        example.viscosity_force.Initialize_Mpi(example.mpi_codim);
        example.elasticity_force.Initialize_Mpi(example.mpi_codim);
        example.incompressible.Initialize_Mpi(example.mpi_codim);
        example.heat_transfer.Initialize_Mpi(example.mpi_codim);
    }

    Set_Output_Directory(output_directory);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Initialize_Mesh()
{
    example.mesher.Initialize_Length_Scale(length_scale,adaptive_length_scale_min,thickness_scale,rim_thickness_scale,max_thickness_factor,min_thickness_factor,max_rim_thickness_factor,min_rim_thickness_factor);
    example.mesher.Initialize_From_External_Mesh_And_Positions(initial_mesh,ARRAY_VIEW<TV>(initial_X));
    example.mesher.Update_Rim_Segment_Mesh();
    if(tank)example.mesher.Initialize_Tank(tank->tank_box,tank->tank_thickness,tank->is_open_tank);
    if(use_spatial_partition)example.spatial_partition.Set_Voxel_Size(length_scale*spatial_partition_voxel_size_coef);
    if(initial_V.m>0)for(int i=1;i<=example.particles.X.m;i++)example.particles.V(i)=initial_V(i);
    if(initial_thickness.m>0)for(int i=1;i<=example.particles.X.m;i++)example.particles.thickness(i)=initial_thickness(i);
    if(initial_texture_coordinates.m>0)for(int i=1;i<=example.particles.X.m;i++)example.particles.texture_coordinate(i)=initial_texture_coordinates(i);
    for(int i=1;i<=example.particles.X.m;i++){example.particles.length_scale(i)=length_scale;}
    example.mesher.Update_Incompressibility_Mesh(example.incompressibility_codimensional_mesh);
    example.mesher.Update_Viscosity_Mesh(example.viscosity_codimensional_mesh,has_rim_segments_for_viscosity,has_rim_triangles_for_viscosity);
    example.mesher.Initialize_Particle_Mass_Based_On_Geometry(thickness_scale,rim_thickness_scale,has_rim,initial_thickness.m==0);
    example.mesher.rigid_bodies=rigid_bodies;
    example.mesher.Update_Particle_Thickness(has_rim);
    if(example.mesher.use_particle_parents){example.particles.Store_Parents(8);}

    example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_1D());
    example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_2D());
    if(TV::m==3)example.deformable_geometry_collection.Add_Structure(example.geometrical_codimensional_mesh_object.Get_Mesh_Object_3D());
    if(example.deformable_geometry_collection.structures.m>0)example.deformable_geometry_collection.structures.Last()->Update_Number_Nodes();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Initialize_Solver()
{
    ////simulation parameters
    example.surface_tension_force.surface_tension=surface_tension;
    if(tri_surface_tension>(T)0)example.surface_tension_force.triangle_surface_tension=tri_surface_tension;
    if(seg_surface_tension>(T)0)example.surface_tension_force.segment_surface_tension=seg_surface_tension;
    example.surface_tension_force.thickness_scale=thickness_scale;
    example.surface_tension_force.length_scale=length_scale;
    if(tank)example.surface_tension_force.Initialize_Tank(tank->tank_box,tank->tank_thickness,tank->is_open_tank);

    ////viscosity force
    example.viscosity_force.density=density;
    example.viscosity_force.length_scale=length_scale;example.viscosity_force.solver.length_scale=length_scale;
    example.viscosity_force.thickness_scale=thickness_scale;example.viscosity_force.solver.thickness_scale=thickness_scale;
    example.viscosity_force.kinematic_viscosity=kinematic_viscosity;
    example.viscosity_force.use_shear_dependent_variable_viscosity=use_shear_dependent_variable_viscosity;
    example.viscosity_force.kinematic_viscosity_min=kinematic_viscosity_min;
    example.viscosity_force.rigid_viscosity=rigid_viscosity;
    example.viscosity_force.solver.max_iter_number=max_iter_number;

    ////elasticity force
    example.elasticity_force.density=density;
    example.elasticity_force.length_scale=length_scale;example.elasticity_force.solver.length_scale=length_scale;
    example.elasticity_force.thickness_scale=thickness_scale;example.elasticity_force.solver.thickness_scale=thickness_scale;
    example.elasticity_force.has_plasticity=has_plasticity;
    example.elasticity_force.youngs_modulus=elastic_modulus;
    example.elasticity_force.plastic_clamp_min=plastic_clamp_min;
    example.elasticity_force.plastic_clamp_max=plastic_clamp_max;
    example.elasticity_force.Initialize_Deformation_State();
    example.elasticity_force.solver.max_iter_number=max_iter_number;

    ////incompressibility
    example.incompressible.density=density;
    example.incompressible.length_scale=length_scale;
    example.incompressible.thickness_scale=thickness_scale;
    example.incompressible.max_iter_number=max_iter_number;

    ////heat transfer
    example.heat_transfer.density=density;
    example.heat_transfer.length_scale=length_scale;example.heat_transfer.solver.length_scale=length_scale;
    example.heat_transfer.thickness_scale=thickness_scale;example.heat_transfer.solver.thickness_scale=thickness_scale;
    example.heat_transfer.kinematic_viscosity=heat_conductivity;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Convection(const T dt,const T time)
{
    for(int i=1;i<=example.particles.X.m;i++){
        int guid=example.particles.guid(i);
        if(kinematic_particle_velocity_hashtable.Contains(guid)||example.particles.inactive(i))continue;    ////skip kinematic and inactive particles

        example.particles.X(i)+=example.particles.V(i)*dt;
    }
    //if(use_mpi)example.mpi_codim->Exchange_Particles_And_Elements_Simple();   ////move to after update_kinematic_positions
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Mesh()
{
    if(remesh){
        Update_Particle_Thickness();
        Update_Particle_Contact();
        Update_Particle_Inactive();
        example.mesher.Update_Mesh();
        Correct_Particle_Normals();
        if(use_spatial_partition)Update_Spatial_Partition();
        Update_Particle_Incompressibility();
        Update_Particle_Length_Scale();
        example.mesher.Update_Incompressibility_Mesh(example.incompressibility_codimensional_mesh);
        example.mesher.Update_Viscosity_Mesh(example.viscosity_codimensional_mesh,has_rim_segments_for_viscosity,has_rim_triangles_for_viscosity);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Position_Based_State(const T time,const bool is_position_update)
{
    for(int i=1;i<=example.particles.X.m;i++){Clear_Forces(i);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Add_Velocity_Independent_Forces(const T dt,const T time)
{
    Update_Velocity_Independent_Forces(dt,time);
    Apply_Forces(example.particles.F,dt,time);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Velocity_Independent_Forces(const T dt,const T time)
{
    //Enforce_Solid_Boundary_Velocities();
    if(has_gravity){
        for(int i=1;i<=example.particles.X.m;i++){
            if(use_centric_gravity)example.particles.body_force(i)+=example.particles.mass(i)*(TV()-example.particles.X(i)).Normalized()*(T)9.8;
            else example.particles.body_force(i)+=example.particles.mass(i)*gravity;
            if(!example.particles.inactive(i))example.particles.F(i)+=example.particles.body_force(i);
        }
    }
    if((has_viscosity&&!use_implicit_viscosity)){
        example.viscosity_force.Add_Velocity_Independent_Forces(example.particles.viscosity_force,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.viscosity_force(i);
    }
    if(has_elasticity&&!use_implicit_elasticity){
        example.elasticity_force.Add_Velocity_Independent_Forces(example.particles.elasticity_force,dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.elasticity_force(i);
    }

    if(has_surface_tension){
        example.surface_tension_force.Update_Position_Based_State(time,true);
        example.surface_tension_force.Add_Velocity_Independent_Forces(example.particles.surface_tension_force,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.surface_tension_force(i);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Add_Velocity_Dependent_Forces(const T dt,const T time)
{
    if(has_elasticity&&use_implicit_elasticity){
        Enforce_Solid_Boundary_Velocities();
        example.elasticity_force.Add_Velocity_Dependent_Forces(example.particles.V,example.particles.elasticity_force,dt,time);     
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.elasticity_force(i);
    }
    if(has_viscosity&&use_implicit_viscosity){
        if(exp_part_viscosity)example.viscosity_force.Add_Velocity_Dependent_Forces_Explicit_Part(example.particles.V,example.particles.viscosity_force,dt,time);
        Enforce_Solid_Boundary_Velocities();
        example.viscosity_force.Add_Velocity_Dependent_Forces(example.particles.V,example.particles.viscosity_force,dt,time);
        for(int i=1;i<=example.particles.X.m;i++)example.particles.F(i)+=example.particles.viscosity_force(i);
    }
    //Enforce_Solid_Boundary_Velocities();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Apply_Forces(ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    for(int i=1;i<=example.particles.X.m;i++){example.particles.V(i)+=F(i)/example.particles.mass(i)*dt;}
    //Enforce_Solid_Boundary_Velocities();

    if(example.mpi_codim){
        if(has_viscosity)example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.viscosity_force);
        if(has_elasticity)example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.elasticity_force);
        if(has_surface_tension)example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.surface_tension_force);
        example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.F);
        example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.V);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Boundary_Conditions()
{
    Update_Boundary_Conditions(example.incompressible);
    if(has_viscosity)Update_Boundary_Conditions(example.viscosity_force,use_implicit_viscosity);
    if(has_elasticity)Update_Boundary_Conditions(example.elasticity_force,false);   ////TODO: correct bc for implicit solver
    if(use_heat_transfer)Update_Boundary_Conditions(example.heat_transfer);
}

//template<class TV>
//bool PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Solid_Inside_And_Value(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour/*=(T)0*/)
//{
//    for(int r=1;r<=rigid_bodies.m;r++){
//        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
//        if(rigid_body->Implicit_Geometry_Lazy_Inside_And_Value(pos,phi,contour)){
//            rigid_normal=rigid_body->Implicit_Geometry_Normal(pos);
//            rigid_v=RIGID_GEOMETRY<TV>::Pointwise_Object_Velocity(rigid_body->Twist(),rigid_body->X(),pos);
//            id=r;
//            return true;
//        }
//    }
//    return false;
//}
//
//template<class TV>
//bool PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Solid_Inside(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,const T contour/*=(T)0*/)
//{
//    for(int r=1;r<=rigid_bodies.m;r++){
//        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
//        if(rigid_body->Implicit_Geometry_Lazy_Inside(pos,contour)){id=r;return true;}
//    }
//    return false;
//}
//
//template<class TV>
//TV PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Solid_Extended_Normal(const TV& pos,const ARRAY<RIGID_BODY<TV>* > rigid_bodies,/*rst*/int& id,/*rst*/T& phi)
//{
//    T phi_min=(T)FLT_MAX;TV normal_min=TV();
//    for(int r=1;r<=rigid_bodies.m;r++){
//        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
//        T phi=(T)0;TV normal=rigid_body->Implicit_Geometry_Extended_Normal(pos,phi);
//        if(phi<phi_min){phi_min=phi;normal_min=normal;id=r;}
//    }
//    phi=phi_min;return normal_min;
//}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Boundary_Conditions(CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible)
{   ////modifed version in PAINT_DRIVER::Update_Boundary_Conditions
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=incompressible.particles;
    //CODIMENSIONAL_MESH& solver_mesh=incompressible.mesh;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N_on_mesh=incompressible.psi_N_on_mesh;
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D_on_mesh=incompressible.psi_D_on_mesh;
    psi_N_on_mesh.Remove_All();psi_D_on_mesh.Remove_All();
    for(int i=1;i<=particles.X.m;i++){
        ////rigid body
        int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;//bool is_contact_boundary_particle=false;

        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(particles.X(i),rigid_bodies,id,phi,rigid_normal,rigid_v)){
            ////is_contact_boundary_particle=Is_On_Tetrahedron_Mesh_Boundary(solver_mesh.node_types(i))||Is_On_Triangle_Mesh_Boundary(solver_mesh.node_types(i))||Is_On_Segment_Mesh_Boundary(solver_mesh.node_types(i));
            ////if(is_contact_boundary_particle){for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),rigid_v(d)));}
            ////else{T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));}

            //if(Is_On_Tetrahedron_Mesh_Boundary(solver_mesh.node_types(i))){for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),rigid_v(d)));}
            //else{T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));}
            
            //if(use_non_slip_on_solids){LOG::cout<<"boundary "<<i<<": "<<rigid_v<<std::endl;for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),rigid_v(d)));}
            //else{T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));}
            
            //is_contact_boundary_particle=Is_On_Tetrahedron_Mesh_Boundary(solver_mesh.node_types(i));
            //if(is_contact_boundary_particle){for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),rigid_v(d)));}
            //else{T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,rigid_normal,v_in_normal_direction));}
        }

        ////kinematic particle
        if(kinematic_particle_velocity_hashtable.Contains(particles.guid(i))){
            const TV& kinematic_v=kinematic_particle_velocity_hashtable.Get(particles.guid(i));
            for(int d=1;d<=TV::m;d++)psi_N_on_mesh.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),kinematic_v(d)));
        }
    }

    ////free surface: 0 psi_D

    //for(int i=1;i<=particles.X.m;i++){particles.psi_N(i)=false;particles.psi_D(i)=false;}
    //{Update_Particle_Psi_N(particles,psi_N_on_mesh);Update_Particle_Psi_D(particles,psi_D_on_mesh);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Boundary_Conditions(CODIMENSIONAL_VISCOSITY_FORCE<TV>& viscosity_force,bool implicit/*=true*/)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=viscosity_force.particles;
    CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object=example.viscosity_force.codimensional_mesh_object;   ////use the geometrical mesh instead of the viscosity mesh
    CODIMENSIONAL_MESH& mesh=mesh_object.mesh;
    VECTOR<ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >,TV::m>& psi_Ns_on_mesh=viscosity_force.psi_Ns_on_mesh;
    VECTOR<ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >,TV::m>& psi_Ds_on_mesh=viscosity_force.psi_Ds_on_mesh;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_Ns_on_surface=viscosity_force.psi_Ns_on_surface;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_Ns_on_codim_boundary=viscosity_force.psi_Ns_on_codim_boundary;

    for(int d=1;d<=TV::m;d++){psi_Ns_on_mesh(d).Remove_All();psi_Ds_on_mesh(d).Remove_All();}
    psi_Ns_on_surface.Remove_All();
    
    for(int i=1;i<=particles.X.m;i++){
        int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;bool is_contact_boundary_particle=false;
        bool non_slip_solid=use_non_slip_on_solids;
        if(implicit&&SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(particles.X(i),rigid_bodies,id,phi,rigid_normal,rigid_v)){ ////contact particles
            //is_contact_boundary_particle=Is_On_Tetrahedron_Mesh_Boundary(mesh.node_types(i))||Is_On_Triangle_Mesh_Boundary(mesh.node_types(i))||Is_On_Segment_Mesh_Boundary(mesh.node_types(i));
            //if(is_contact_boundary_particle){for(int d=1;d<=3;d++)psi_Ds_on_mesh(d).Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,SIMPLEX_PARTICLE,rigid_v(d)));}
            Process_Rigid_V_And_Normal(rigid_v,rigid_normal);

            if(use_friction_boundary)rigid_v=particles.V(i);
            if(use_correcting_rigid_normal){    ////hack for piz example
                if(rigid_normal.y>(T).7)rigid_normal=TV::Axis_Vector(2);else rigid_normal=TV();
            }

            non_slip_solid|=(solid_non_slip_hashtable.Contains(id)&&/*nonslip*/solid_non_slip_hashtable.Get(id).x);
            is_contact_boundary_particle=Is_On_Tetrahedron_Mesh_Boundary(mesh.node_types(i))||Is_On_Segment_Mesh(mesh.node_types(i))||Is_On_Triangle_Mesh_Boundary(mesh.node_types(i));
            if(is_contact_boundary_particle){
                if(non_slip_solid||use_paint_trace_boundary||use_friction_boundary) for(int d=1;d<=3;d++)psi_Ds_on_mesh(d).Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,SIMPLEX_PARTICLE,rigid_v(d)));
                else for(int d=1;d<=3;d++){if(rigid_normal(d)!=(T)0)psi_Ds_on_mesh(d).Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,SIMPLEX_PARTICLE,rigid_v(d)));}
            }
        }

        int p_type=mesh.node_types(i);
        bool on_free_surface=Is_On_Codimensional_Mesh_Free_Surface(p_type);
        bool on_codim_boundary=Is_On_Codimensional_Mesh_Free_Surface(p_type);
        TV area_weighted_normal=TV();
        ARRAY<TV> free_surface_normals;
        if(on_free_surface){
            //T epsilon=(T)1e-5;
            //TV tet_normal=mesh_object.Nodal_Area_Weighted_Tetrahedron_Normal(i);
            //if(Is_On_Tetrahedron_Mesh(p_type)&&tet_normal.Magnitude()<epsilon*pow<2>(length_scale))tet_normal=TV();
            //if(tet_normal!=TV())free_surface_normals.Append(tet_normal.Normalized());

            //TV tri_normal=mesh_object.Nodal_Length_Weighted_Triangle_Normal(i);
            //if(Is_On_Triangle_Mesh(p_type)&&tri_normal.Magnitude()<epsilon*length_scale)tri_normal=TV();
            //if(tri_normal!=TV())free_surface_normals.Append(tri_normal.Normalized());

            //TV seg_normal=mesh_object.Nodal_Segment_Normal(i);
            //if(Is_On_Segment_Mesh(p_type)&&seg_normal.Magnitude()<epsilon)seg_normal=TV();
            //if(seg_normal!=TV())free_surface_normals.Append(seg_normal.Normalized());

            //if(free_surface_normals.m>0)for(int j=1;j<=free_surface_normals.m;j++){
            //    TV normal=free_surface_normals(j);
            //    psi_Ns_on_surface.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
            //    if(viscosity_force.use_constant_extrapolation&&!on_codim_boundary){
            //        psi_Ns_on_codim_boundary.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
            //    }
            //}

            T epsilon=(T)1e-5;
            TV tet_normal=mesh_object.Nodal_Area_Weighted_Tetrahedron_Normal(i);
            if(Is_On_Tetrahedron_Mesh(p_type)&&tet_normal.Magnitude()<epsilon*pow<2>(length_scale))tet_normal=TV();
            //if(tet_normal!=TV())free_surface_normals.Append(tet_normal.Normalized());

            TV tri_normal=mesh_object.Nodal_Length_Weighted_Triangle_Normal(i);
            if(Is_On_Triangle_Mesh(p_type)&&tri_normal.Magnitude()<epsilon*length_scale)tri_normal=TV();
            //if(tri_normal!=TV())free_surface_normals.Append(tri_normal.Normalized());

            TV seg_normal=mesh_object.Nodal_Segment_Normal(i);
            if(Is_On_Segment_Mesh(p_type)&&seg_normal.Magnitude()<epsilon)seg_normal=TV();
            //if(seg_normal!=TV())free_surface_normals.Append(seg_normal.Normalized());

            T particle_thickness=particles.thickness(i);
            if(!viscosity_force.use_particle_thickness) particle_thickness=viscosity_force.thickness_scale;
            area_weighted_normal=tet_normal+particle_thickness*tri_normal+(T).25*(T)pi*pow<2>(particle_thickness)*seg_normal;

            if(area_weighted_normal!=TV()){
                TV normal=area_weighted_normal.Normalized();
                psi_Ns_on_surface.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
                if(viscosity_force.use_constant_extrapolation&&!on_codim_boundary){
                    psi_Ns_on_codim_boundary.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
                }
            }

            //if(free_surface_normals.m>0)for(int j=1;j<=free_surface_normals.m;j++){
            //    TV normal=free_surface_normals(j);
            //    psi_Ns_on_surface.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
            //    if(viscosity_force.use_constant_extrapolation&&!on_codim_boundary){
            //        psi_Ns_on_codim_boundary.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
            //    }
            //}

            //if(particles.guid(i)==4403){
            //    LOG::cout<<"free surface p "<<i<<": "<<free_surface_normals<<": "<<tet_normal<<", "<<tri_normal<<", "<<seg_normal<<std::endl;
            //}
        }

        bool add_free_surface_psi_N=false;
        if(viscosity_force.use_constant_extrapolation){
            add_free_surface_psi_N=(Is_On_Triangle_Mesh_Interior(mesh.node_types(i))||Is_On_Segment_Mesh_Interior(mesh.node_types(i)))&&!is_contact_boundary_particle;
        }
        else{add_free_surface_psi_N=on_free_surface/*&&!is_contact_boundary_particle*/;}

        if(add_free_surface_psi_N){ ////free surface particles
            //if(Is_Joint<TV::m>(mesh.node_types(i))){continue;}
            //if(Is_On_Segment_Mesh_Interior(mesh.node_types(i))){continue;}

            if(area_weighted_normal!=TV()){    ////check if the area weighted normal is zero
                TV normal=area_weighted_normal.Normalized();
                for(int d=1;d<=3;d++){
                    bool on_nonslip_solids=is_contact_boundary_particle&&(non_slip_solid||use_paint_trace_boundary||rigid_normal(d)!=(T)0);
                    if(!on_nonslip_solids)psi_Ns_on_mesh(d).Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
                    //LOG::cout<<"vis aw normal: "<<i<<": "<<normal<<", "<<area_weighted_normal<<", "<<area_scale<<", "<<area_weighted_normal.Magnitude()<<" > "<<area_scale*epsilon<<std::endl;
                }
            }

            //if(free_surface_normals.m>0)for(int j=1;j<=free_surface_normals.m;j++){
            //    TV normal=free_surface_normals(j);
            //    for(int d=1;d<=3;d++){
            //        bool on_nonslip_solids=is_contact_boundary_particle&&(non_slip_solid||use_paint_trace_boundary||rigid_normal(d)!=(T)0);
            //        if(!on_nonslip_solids)psi_Ns_on_mesh(d).Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
            //        //LOG::cout<<"vis aw normal: "<<i<<": "<<normal<<", "<<area_weighted_normal<<", "<<area_scale<<", "<<area_weighted_normal.Magnitude()<<" > "<<area_scale*epsilon<<std::endl;
            //    }
            //}
        }

        ////kinematic particle
        if(implicit&&kinematic_particle_velocity_hashtable.Contains(particles.guid(i))){
            const TV& kinematic_v=kinematic_particle_velocity_hashtable.Get(particles.guid(i));
            for(int d=1;d<=3;d++)psi_Ds_on_mesh(d).Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,SIMPLEX_PARTICLE,kinematic_v(d)));
        }
    }

    //for(int i=1;i<=particles.X.m;i++){particles.psi_N(i)=false;particles.psi_D(i)=false;}
    //int d=1;{Update_Particle_Psi_N(particles,psi_Ns_on_mesh(d));Update_Particle_Psi_D(particles,psi_Ds_on_mesh(d));}

    //LOG::cout<<"get bc: psi_D "<<psi_Ds_on_mesh(1).m<<", psi_N "<<psi_Ns_on_mesh(1).m<<std::endl;
}


template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Boundary_Conditions(CODIMENSIONAL_HEAT_TRANSFER<TV>& heat_transfer)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=heat_transfer.particles;
    CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object=heat_transfer.codimensional_mesh_object;
    CODIMENSIONAL_MESH& mesh=mesh_object.mesh;
    VECTOR<ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >,TV::m>& psi_Ns_on_mesh=heat_transfer.psi_Ns_on_mesh;
    VECTOR<ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >,TV::m>& psi_Ds_on_mesh=heat_transfer.psi_Ds_on_mesh;
    for(int d=1;d<=TV::m;d++){psi_Ns_on_mesh(d).Remove_All();psi_Ds_on_mesh(d).Remove_All();}
    for(int i=1;i<=particles.X.m;i++){
        if(Is_On_Codimensional_Mesh_Free_Surface(mesh.node_types(i))){ ////free surface particles
            psi_Ds_on_mesh(1).Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(i,SIMPLEX_PARTICLE,environment_temperature));
            //LOG::cout<<"psi_N "<<i<<": "<<environment_temperature<<std::endl;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Add_Friction_Forces(ARRAY_VIEW<TV>& friction_forces,const T dt,const T time)
{
    if(use_damped_friction){
        for(int i=1;i<=example.particles.X.m;i++){
            int id=0;T phi=(T)0;TV normal;TV v_body;//bool is_contact_boundary_particle=false;
            if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(example.particles.X(i),rigid_bodies,id,phi,normal,v_body)){ ////contact particles
                Process_Rigid_V_And_Normal(v_body,normal);
                const TV& v=example.particles.V(i);
                T vn=TV::Dot_Product(v,normal);
                T vn_body=TV::Dot_Product(v_body,normal);
                T vn_final=max(vn,vn_body);
                TV vt=v-vn*normal;
                TV vt_body=v_body-vn_body*normal;
                TV vt_relative=vt-vt_body;
                TV vt_final=vt_body+vt_relative*((T)1-friction_magnitude);
                TV v_final=vn_final*normal+vt_final;
                TV friction_f=example.particles.mass(i)*(v_final-v)/dt;
                friction_forces(i)+=friction_f;
            }
        }
    }
    else{
        for(int i=1;i<=example.particles.X.m;i++){
            int id=0;T phi=(T)0;TV normal;TV v_body;//bool is_contact_boundary_particle=false;
            if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(example.particles.X(i),rigid_bodies,id,phi,normal,v_body)){ ////contact particles
                Process_Rigid_V_And_Normal(v_body,normal);
                //T fric_mag=friction_magnitude;
                //if(solid_non_slip_hashtable.Contains(id)&&/*slip*/!solid_non_slip_hashtable.Get(id).x){
                //    T fric_mag_factor=solid_non_slip_hashtable.Get(id).y;fric_mag*=fric_mag_factor;fric_mag=clamp(fric_mag,(T)0,(T)1);
                //}

                T local_coefficient_of_friction=friction_magnitude;
                if(use_viscosity_dependent_friction){
                    T epsilon=(T)1e-10;
                    T max_vis=example.viscosity_force.kinematic_viscosity;
                    T min_vis=example.viscosity_force.kinematic_viscosity_min;
                    T fric_max=local_coefficient_of_friction;
                    T fric_min=local_coefficient_of_friction*(T).01;
                    T frac=(T)1;
                    if(max_vis>min_vis){
                        frac=(example.particles.viscosity(i)-min_vis)/(max_vis-min_vis+epsilon);
                        frac=clamp(frac,(T)0,(T)1);
                    }
                    local_coefficient_of_friction=((T)1-frac)*fric_min+frac*fric_max;
                }

                const TV& v=example.particles.V(i);
                T vn=TV::Dot_Product(v,normal);
                T vn_body=TV::Dot_Product(v_body,normal);
                T vn_final=max(vn,vn_body);
                TV vt=v-vn*normal;
                TV vt_body=v_body-vn_body*normal;
                TV vt_relative=vt-vt_body;
                T vt_relative_magnitude=vt_relative.Magnitude();
                T normal_force=vn_final-vn;
                T frac=(T)1;if(local_coefficient_of_friction*normal_force<vt_relative_magnitude)frac=local_coefficient_of_friction*normal_force/vt_relative_magnitude;
                TV vt_final=vt_body+vt_relative*((T)1-frac);
                TV v_final=vn_final*normal+vt_final;
                TV friction_f=example.particles.mass(i)*(v_final-v)/dt;
                friction_forces(i)+=friction_f;
                //if(friction_f!=TV())LOG::cout<<"friction "<<i<<": "<<friction_f<<", v_final: "<<v_final<<", "<<v<<std::endl;
            }
        }
    }

    if(use_mpi)example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(friction_forces);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Correct_Particle_Positions()
{
    T epsilon=(T)1e-3*length_scale; ////positive value to keep particles contact
    for(int i=1;i<=example.particles.X.m;i++){
        ////correct particle positions if contacting with solid
        TV rigid_normal;
        if(Is_Contact_With_Solid(example.particles.contact(i))){
            int id=0;T phi=(T)0;TV rigid_normal=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Extended_Normal(example.particles.X(i),rigid_bodies,id,phi);
            if(phi>(T)0)example.particles.X(i)-=((T)1+epsilon)*rigid_normal*phi;    ////adhesion, stick the contact particles from last timestp to solid surface
            else if(phi<(T)0){example.particles.X(i)-=((T)1-epsilon)*rigid_normal*phi;} ////push inside particles onto solid surface
        }

        ////correct particle positions if contacting with tank
        if(tank!=0&&Is_Contact_With_Tank(example.particles.contact(i))){
            TV wall_normal=TV();TV wall_pos=TV();wall_normal=SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tank_Normal_And_Position(example.particles.X(i),*tank,wall_pos);
            example.particles.X(i)=wall_pos;
        }
    }
    ////do not need mpi sync
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Particle_Thickness()
{
    ////after advection, update particle thickness to discrete thickness for detecting contact and remeshing
    for(int i=1;i<=example.particles.X.m;i++){example.particles.thickness(i)=example.mesher.Discrete_Particle_Thickness(i,example.mesher.has_rim);} 

    ////mpi sync
    if(example.mpi_codim){example.mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(example.particles.thickness);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Particle_Contact()
{
    for(int i=1;i<=example.particles.X.m;i++){
        Clear_Contact(example.particles.contact(i),CONTACT_FIBER|CONTACT_SOLID2|CONTACT_SOLID3|CONTACT_BOUNDARY|CONTACT_TANK);    ////No CONTACT_FLUID
        if(!has_adhesion)Clear_Contact(example.particles.contact(i),CONTACT_SOLID);

        ////update particle contacting with solid
        int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;
        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(example.particles.X(i),rigid_bodies,id,phi,rigid_normal,rigid_v)){
            Set_Contact_With_Solid(example.particles.contact(i));
        }

        ////update particle contacting with tank
        if(tank!=0&&SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Contact_Tank(example.particles.X(i),*tank)) Set_Contact_With_Tank(example.particles.contact(i));
    }
    ////do not need mpi sync
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Correct_Particle_Normals()
{
    for(int i=1;i<=example.particles.X.m;i++){
        int id=0;T phi=(T)0;TV rigid_normal;TV rigid_v;
        if(SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(example.particles.X(i),rigid_bodies,id,phi,rigid_normal,rigid_v)){
            Process_Rigid_V_And_Normal(rigid_v,rigid_normal);
            //example.particles.area_weighted_normal(i)=example.particles.area_weighted_normal(i).Magnitude()*rigid_normal;
            example.particles.normal(i)=rigid_normal;
        }
    }
    ////do not need mpi sync
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Enforce_Solid_Boundary_Velocities()
{
    example.incompressible.Apply_Boundary_Conditions();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Restart()
{
    LOG::cout<<"Restart from frame "<<restart_frame<<std::endl;

    if(!write_substep){
        current_frame=restart_frame;
        time=Time_At_Frame(current_frame);
        LOG::cout<<"restart time "<<time<<std::endl;
    }
    else{
        current_substep_frame=restart_frame;
        std::string f=STRING_UTILITIES::string_sprintf("%d",current_substep_frame);
        std::string file_name=output_directory+"/"+f+"/substep_time";
        if(FILE_UTILITIES::File_Exists(file_name)){FILE_UTILITIES::Read_From_Text_File(file_name,time);current_frame=Frame_At_Time(time);}
        else{current_frame=restart_frame;time=Time_At_Frame(current_frame);}
        LOG::cout<<"restart time substep: "<<time<<std::endl;
    }

    Read_Output_Files(restart_frame);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Clear_Forces(const int i)
{
    example.particles.F(i)=TV();
    example.particles.body_force(i)=TV();
    example.particles.pressure_force(i)=TV();
    example.particles.surface_tension_force(i)=TV();
    example.particles.viscosity_force(i)=TV();
    example.particles.elasticity_force(i)=TV();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Merge_Mesh_And_Attributes_In_Different_Regions(/*rst*/CODIMENSIONAL_MESH& merged_mesh,const ARRAY<CODIMENSIONAL_MESH>& meshes,/*rst*/ARRAY<TV>& merged_X,const ARRAY<ARRAY<TV> >& Xs,
    /*rst*/ARRAY<TV>* merged_V/*=0*/,const ARRAY<ARRAY<TV> >* Vs/*=0*/,/*rst*/ARRAY<ARRAY<T> >* merged_phis/*=0*/,const ARRAY<ARRAY<ARRAY<T> > >* phis/*=0*/,
    /*rst*/ARRAY<T>* merged_thickness/*=0*/,const ARRAY<ARRAY<T> >* thicknesses/*=0*/)   ////merge mesh from different regions, for initialization
{
    if(phis!=0){merged_phis->Resize((*phis)(1).m);}

    int size=merged_X.m;
    for(int i=1;i<=meshes.m;i++){
        for(int j=1;j<=meshes(i).tetrahedron_mesh.elements.m;j++){
            TV_INT4 v=meshes(i).tetrahedron_mesh.elements(j)+TV_INT4::Constant_Vector(size);
            merged_mesh.tetrahedron_mesh.elements.Append(v);
        }
        for(int j=1;j<=meshes(i).triangle_mesh.elements.m;j++){
            TV_INT3 v=meshes(i).triangle_mesh.elements(j)+TV_INT3::Constant_Vector(size);
            merged_mesh.triangle_mesh.elements.Append(v);
        }
        for(int j=1;j<=meshes(i).segment_mesh.elements.m;j++){
            TV_INT2 v=meshes(i).segment_mesh.elements(j)+TV_INT2::Constant_Vector(size);
            merged_mesh.segment_mesh.elements.Append(v);
        }
        for(int j=1;j<=Xs(i).m;j++){
            TV pos=Xs(i)(j);merged_X.Append(pos);
            if(merged_V!=0&&Vs!=0){TV v=(*Vs)(i)(j);(*merged_V).Append(v);}
            if(merged_phis!=0&&phis!=0)for(int c=1;c<=(*phis).m;c++){T phi=(*phis)(i)(c)(j);(*merged_phis)(c).Append(phi);}
            if(merged_thickness!=0&&thicknesses!=0){T thickness=(*thicknesses)(i)(j);(*merged_thickness).Append(thickness);}
        }
        size+=Xs(i).m;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Particle_Psi_N(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N_on_mesh)
{for(int i=1;i<=psi_N_on_mesh.m;i++){int p=psi_N_on_mesh(i).index;particles.psi_N(p)=true;}}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Particle_Psi_D(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D_on_mesh)
{for(int i=1;i<=psi_D_on_mesh.m;i++){int p=psi_D_on_mesh(i).index;particles.psi_D(p)=true;}}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Spatial_Partition()
{
    SEGMENT_MESH& segment_mesh=example.geometrical_codimensional_mesh.segment_mesh;
    TRIANGLE_MESH& triangle_mesh=example.geometrical_codimensional_mesh.triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh=example.geometrical_codimensional_mesh.tetrahedron_mesh;
    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>& spatial_partition=example.spatial_partition;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;

    spatial_partition.Clear();
    int object_id=1;    ////TODO: store voxel_range
    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& e=segment_mesh.elements(i);RANGE<TV_INT> voxel_range;
        spatial_partition.Insert_Into_Hashtable(particles.X,e,i,object_id,voxel_range);
    }
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        const TV_INT3& e=triangle_mesh.elements(i);RANGE<TV_INT> voxel_range;
        spatial_partition.Insert_Into_Hashtable(particles.X,e,i,object_id,voxel_range);
    }
    for(int i=1;i<=tetrahedron_mesh.elements.m;i++){
        const TV_INT4& e=tetrahedron_mesh.elements(i);RANGE<TV_INT> voxel_range;
        spatial_partition.Insert_Into_Hashtable(particles.X,e,i,object_id,voxel_range);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Heat_Transfer(const T dt,const T time)
{
    example.heat_transfer.Update_Heat_Transfer(example.particles.temperature,dt,time);
    //LOG::cout<<"T: "<<example.particles.temperature<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Process_Rigid_V_And_Normal(TV& rigid_v,TV& rigid_normal)
{
    rigid_v*=rigid_body_speed_factor;   ////for splash example
    if(use_correcting_rigid_normal){    ////hack for piz example
        if(rigid_normal.y>(T).7)rigid_normal=TV::Axis_Vector(2);else rigid_normal=TV();
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Remove_Particles_In_Invalid_Boxes()
{
    ARRAY<int> p_to_rmv;
    for(int i=1;i<=example.particles.X.m;i++){
        bool inside=false;
        for(int j=1;j<=sim_particle_boxes.m;j++){
            const BOX<TV>& box=sim_particle_boxes(j);
            if(box.Lazy_Inside(example.particles.X(i))){inside=true;break;}
        }
        if(inside!=use_inside_sim_box)p_to_rmv.Append(i);
    }
    LOG::cout<<"rmv particles in restart: "<<p_to_rmv<<std::endl;

    Remove_Sorted_Particles(example.particles,p_to_rmv,example.geometrical_codimensional_mesh,true);

    example.geometrical_codimensional_mesh.Update_States(example.particles.X.m,example.mpi_codim);
    example.mesher.Update_Rim_Segment_Mesh();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_DRIVER<TV>::Update_Particles_In_Boundary_Boxes()
{
    boundary_particle_hashtable.Remove_All();
    for(int i=1;i<=example.particles.X.m;i++){
        bool inside=false;
        for(int j=1;j<=boundary_particle_boxes.m;j++){
            const BOX<TV>& box=boundary_particle_boxes(j);
            if(box.Lazy_Inside(example.particles.X(i))){inside=true;break;}
        }
        if(inside){boundary_particle_hashtable.Set(example.particles.guid(i));}
    }
}

//template class CODIMENSIONAL_FLUID_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_FLUID_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_FLUID_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_FLUID_DRIVER<VECTOR<double,3> >;
#endif
