//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_FLUID_DRIVER__
#define __CODIMENSIONAL_FLUID_DRIVER__
#include <PhysBAM_Tools/Grids_Uniform_Arrays/ARRAYS_ND.h>
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Utilities/TIMER.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Standard_Tests/RIGIDS_STANDARD_TESTS.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY_COLLECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include <PhysBAM_Geometry/Basic_Geometry/ELLIPSE.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include "../Solver/CODIMENSIONAL_SURFACE_TENSION_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOELASTIC_FORCE.h"
#include "../Misc/WATER_CYLINDRICAL_SOURCE.h"
#include "../Misc/WATER_ANNULAR_GAP_SOURCE.h"
#include "../Misc/WATER_SOURCE.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Solver/CODIMENSIONAL_FLUID_EXAMPLE.h"
#include "../Viscoelastic/SOLID_FLUID_CONTACT_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_FLUID_DRIVER
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    int test;
    std::string output_directory;
    std::string model_directory;
    STREAM_TYPE stream_type;
    int first_frame,last_frame,current_frame,restart_frame;
    bool write_substep;
    int current_substep_frame,last_substep_frame;
    T frame_rate;
    T time,current_time;
    T cfl;
    bool use_mpi;
    CODIMENSIONAL_FLUID_EXAMPLE<TV> example;
    
    T length_scale;
    T adaptive_length_scale_min;
    T adaptive_length_scale_min2;
    T thickness_scale;
    T rim_thickness_scale;
    T max_thickness_factor;
    T min_thickness_factor;
    T max_rim_thickness_factor;
    T min_rim_thickness_factor;
    T surface_tension_factor;
    T gravity_factor;
    T length_scale_min_factor;
    T scale;
    TV gravity;
    T density;
    T surface_tension;
    T tri_surface_tension;
    T seg_surface_tension;
    T kinematic_viscosity;
    T kinematic_viscosity_min;  ////for variable viscosity
    T rigid_viscosity;          ////for Bingham plastic   
    T elastic_modulus;
    T plastic_clamp_min;
    T plastic_clamp_max;
    T friction_magnitude;
    T spatial_partition_voxel_size_coef; 
    T time_scale;
    T environment_temperature;
    T heat_conductivity;
    int max_iter_number;
    T fixed_cfl;

    ////external mesh
    CODIMENSIONAL_MESH initial_mesh;
    ARRAY<TV> initial_X;
    ARRAY<TV> initial_V;
    ARRAY<T> initial_thickness;
    ARRAY<TV> initial_texture_coordinates;
    ARRAY<ARRAY<T> > initial_phis;

    ////options
    bool remesh;
    bool has_gravity;
    bool has_surface_tension;
    bool has_viscosity;
    bool use_implicit_viscosity;
    bool use_shear_dependent_variable_viscosity;
    bool has_elasticity;
    bool has_plasticity;
    bool use_implicit_elasticity;
    bool has_fiber_fluid_coupling_force;
    bool has_adhesion;
    bool has_friction;
    bool has_kinematic_particles;
    bool has_incompressibility;
    bool has_rim;

    bool use_implicit_all;
    bool use_explicit_all;
    bool use_centric_gravity;
    bool use_spatial_partition;
    bool check_negative_volume_tetrahedra;
    bool use_inactive_particle;
    bool use_correcting_particle_position_on_solids;
    bool use_non_slip_on_solids;
    bool use_adaptive_length_scale_on_codim;
    bool use_adaptive_length_scale_on_levelset;
    bool use_adaptive_length_scale_on_rim;
    bool use_adaptive_length_scale_on_fiber;
    bool use_adaptive_length_scale_based_on_test;
    bool has_rim_segments_for_viscosity;
    bool has_rim_triangles_for_viscosity;
    bool use_heat_transfer;
    bool use_paint_trace_boundary;  ////for paint example
    bool use_friction_boundary; ////for bunny example
    bool use_viscosity_dependent_friction;
    bool use_damped_friction;
    bool use_correcting_rigid_normal;
    bool exp_part_viscosity;

    ////rigid objects
    RIGID_BODY_COLLECTION<TV> rigid_body_collection;
    RIGIDS_STANDARD_TESTS<TV> rigid_standard_tests;
    ARRAY<RIGID_BODY<TV>*> rigid_bodies;
    WATER_TANK<TV>* tank;

    HASHTABLE<int,PAIR<bool,T> > solid_non_slip_hashtable;   ////<non_slip,friction_magnitude_factor>
    T rigid_body_speed_factor;
    bool use_inside_sim_box;
    ARRAY<BOX<TV> > sim_particle_boxes;
    ARRAY<BOX<TV> > boundary_particle_boxes;
    HASHTABLE<int> boundary_particle_hashtable;     ////boundary static particles
    HASHTABLE<int,TV> boundary_particle_hashtable2; ////boundary kinematic particles

    ////kinematic particles
    HASHTABLE<int,TV> kinematic_particle_velocity_hashtable;

    ////random
    RANDOM_NUMBERS<T> random;

    ////mpi
    TV_INT processes_per_dimension;
    RANGE<TV> global_grid_domain;
    int number_of_ghost_cells;
    GRID<TV> global_grid;

    CODIMENSIONAL_FLUID_DRIVER();
    virtual ~CODIMENSIONAL_FLUID_DRIVER(){}
    virtual void Initialize();
    virtual void Restart();
    void Execute_Main_Program();
    void Simulate_To_Frame(const int frame);
    void Advance_To_Target_Time(const T target_time);
    T Time_At_Frame(int frame){return (T)frame/frame_rate;}
    int Frame_At_Time(T time){return (int)(time*frame_rate);}
    void Set_Output_Directory(const std::string output_directory_input);

    virtual void Advance_One_Timestep(const T dt,const T time);
    virtual T CFL();
    virtual T Valid_Timestep_For_Tetrahedron_Mesh(const T dt);
    virtual void Write_Output_Files(const int frame);
    virtual void Read_Output_Files(const int frame);
    virtual void Initialize_Test_Parameters(){}
    virtual void Initialize_Mpi();
    virtual void Initialize_Mesh();
    virtual void Initialize_Solver();
    virtual void Initialize_Kinematic_Positions(){}
    virtual void Initialize_After_Restart(){}

    virtual void Convection(const T dt,const T time);
    virtual void Update_Mesh();
    virtual void Update_Position_Based_State(const T time,const bool is_position_update);
    virtual void Add_Velocity_Independent_Forces(const T dt,const T time);
    virtual void Update_Velocity_Independent_Forces(const T dt,const T time);
    virtual void Add_Velocity_Dependent_Forces(const T dt,const T time);
    virtual void Add_Friction_Forces(ARRAY_VIEW<TV>& friction_forces,const T dt,const T time);
    virtual void Apply_Forces(ARRAY_VIEW<TV>& F,const T dt,const T time);
    virtual void Update_Boundary_Conditions();
    virtual void Update_Particle_Contact();
    virtual void Update_Particle_Inactive(){}
    virtual void Update_Particle_Incompressibility(){}
    virtual void Update_Target_Divergence(){}
    virtual void Update_Particle_Length_Scale(){}
    virtual void Update_Particle_Thickness();
    virtual void Correct_Particle_Positions();
    virtual void Correct_Particle_Normals();
    virtual void Enforce_Solid_Boundary_Velocities();
    virtual void Update_Kinematic_Positions(const T dt,const T time){}
    virtual void Update_Heat_Transfer(const T dt,const T time);
    virtual void Remove_Particles_In_Invalid_Boxes();
    virtual void Update_Particles_In_Boundary_Boxes();

    void Use_Rim(bool has_rim_input){has_rim=has_rim_input;example.mesher.has_rim=has_rim;example.mesher.use_rim_mass_redistribution=has_rim;}
    
    virtual void Update_Boundary_Conditions(CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible);
    virtual void Update_Boundary_Conditions(CODIMENSIONAL_VISCOSITY_FORCE<TV>& viscosity_force,bool implicit=true);
    virtual void Update_Boundary_Conditions(CODIMENSIONAL_HEAT_TRANSFER<TV>& heat_transfer);
    virtual void Update_Spatial_Partition();

protected:
    ////helper functions
    void Clear_Forces(const int i);
    void Merge_Mesh_And_Attributes_In_Different_Regions(/*rst*/CODIMENSIONAL_MESH& merged_mesh,const ARRAY<CODIMENSIONAL_MESH>& meshes,/*rst*/ARRAY<TV>& merged_X,const ARRAY<ARRAY<TV> >& Xs,
        /*rst*/ARRAY<TV>* merged_V=0,const ARRAY<ARRAY<TV> >* Vs=0,/*rst*/ARRAY<ARRAY<T> >* merged_phis=0,const ARRAY<ARRAY<ARRAY<T> > >* phis=0,
        /*rst*/ARRAY<T>* merged_thickness=0,const ARRAY<ARRAY<T> >* thicknesses=0);   ////merge mesh from different regions, for initialization
    void Update_Particle_Psi_N(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N);
    void Update_Particle_Psi_D(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D);
    void Process_Rigid_V_And_Normal(TV& rigid_v,TV& rigid_normal);  ////for different examples
};
}
#endif
