//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Utilities/FILE_UTILITIES.h>
#include <PhysBAM_Tools/Read_Write/Arrays/READ_WRITE_ARRAY.h>
#include <PhysBAM_Tools/Read_Write/Data_Structures/READ_WRITE_HASHTABLE.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include "../Solver/CODIMENSIONAL_FLUID_EXAMPLE.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_FLUID_EXAMPLE<TV>::CODIMENSIONAL_FLUID_EXAMPLE()
    :output_directory(TV::m==2?"output":"output3"),geometrical_codimensional_mesh(&segment_mesh,&triangle_mesh,&tetrahedron_mesh),geometrical_codimensional_mesh_object(geometrical_codimensional_mesh,particles),
    incompressibility_codimensional_mesh(&incompressibility_segment_mesh,&incompressibility_triangle_mesh,&tetrahedron_mesh),incompressibility_codimensional_mesh_object(incompressibility_codimensional_mesh,particles),
    viscosity_codimensional_mesh(&viscosity_segment_mesh,&viscosity_triangle_mesh,&tetrahedron_mesh),viscosity_codimensional_mesh_object(viscosity_codimensional_mesh,particles),
    deformable_geometry_collection(particles,false),mesher(geometrical_codimensional_mesh_object),
    surface_tension_force(geometrical_codimensional_mesh_object,(T)0,(T)0,0),viscosity_force(viscosity_codimensional_mesh_object,0),elasticity_force(viscosity_codimensional_mesh_object,0),
    incompressible(incompressibility_codimensional_mesh_object),heat_transfer(geometrical_codimensional_mesh_object),mpi_grid(0),mpi_codim(0)
{
    //particles.Update_Store_Attributes(CODIMENSIONAL_FLUID_PARTICLES<TV>::PAINT_ATTRIBUTES);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_EXAMPLE<TV>::Write_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    if(particles.X.m>0){
        FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/codimensional_fluid_particles",particles);
        FILE_UTILITIES::Write_To_Text_File(output_directory+"/"+f+"/next_guid",particles.next_guid,"\n");
        deformable_geometry_collection.Write(stream_type,output_directory,frame,frame,true);

        if(mesher.use_mesh_edge_connect){
            FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/connected_segment_guids",geometrical_codimensional_mesh.connected_segment_hashtable);
        }

        if(mesher.use_particle_parents){
            FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/particle_parents",particles.parents_hashtable);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_FLUID_EXAMPLE<TV>::Read_Output_Files(const int frame)
{
    STREAM_TYPE stream_type((T()));
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);
    std::string file_name=output_directory+"/"+f+"/codimensional_fluid_particles";
    if(FILE_UTILITIES::File_Exists(file_name)){
        FILE_UTILITIES::Read_From_File(stream_type,file_name,particles);
        FILE_UTILITIES::Read_From_Text_File(output_directory+"/"+f+"/next_guid",particles.next_guid);
        deformable_geometry_collection.Read(stream_type,output_directory,frame,frame,true);

        geometrical_codimensional_mesh.Update_States(particles.X.m,mpi_codim);
        mesher.Update_Rim_Segment_Mesh();

        if(mesher.use_mesh_edge_connect){
            FILE_UTILITIES::Read_From_File(stream_type,output_directory+"/"+f+"/connected_segment_guids",geometrical_codimensional_mesh.connected_segment_hashtable);
        }

        if(mesher.use_particle_parents){
            FILE_UTILITIES::Read_From_File(stream_type,output_directory+"/"+f+"/particle_parents",particles.parents_hashtable);
        }
    }
}

//template class CODIMENSIONAL_FLUID_EXAMPLE<VECTOR<float,2> >;
template class CODIMENSIONAL_FLUID_EXAMPLE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_FLUID_EXAMPLE<VECTOR<double,2> >;
template class CODIMENSIONAL_FLUID_EXAMPLE<VECTOR<double,3> >;
#endif
