//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_FLUID_EXAMPLE__
#define __CODIMENSIONAL_FLUID_EXAMPLE__
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE.h>
#include <PhysBAM_Geometry/Solids_Geometry/DEFORMABLE_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_UNIFORM_GRID.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Solver/CODIMENSIONAL_SURFACE_TENSION_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOELASTIC_FORCE.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE.h"
#include "../Solver/CODIMENSIONAL_HEAT_TRANSFER.h"
#include "../Intersection/CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_FLUID_EXAMPLE
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;

public:
    std::string output_directory;

    CODIMENSIONAL_FLUID_PARTICLES<TV> particles;  

    SEGMENT_MESH segment_mesh;
    TRIANGLE_MESH triangle_mesh;
    TETRAHEDRON_MESH tetrahedron_mesh;
    CODIMENSIONAL_MESH geometrical_codimensional_mesh;
    CODIMENSIONAL_MESH_OBJECT<TV> geometrical_codimensional_mesh_object;
    SEGMENT_MESH incompressibility_segment_mesh;
    TRIANGLE_MESH incompressibility_triangle_mesh;
    CODIMENSIONAL_MESH incompressibility_codimensional_mesh;
    CODIMENSIONAL_MESH_OBJECT<TV> incompressibility_codimensional_mesh_object;
    SEGMENT_MESH viscosity_segment_mesh;
    TRIANGLE_MESH viscosity_triangle_mesh;
    CODIMENSIONAL_MESH viscosity_codimensional_mesh;
    CODIMENSIONAL_MESH_OBJECT<TV> viscosity_codimensional_mesh_object;

    DEFORMABLE_GEOMETRY_COLLECTION<TV> deformable_geometry_collection;

    CODIMENSIONAL_MESHING_3D<T> mesher;
    CODIMENSIONAL_SURFACE_TENSION_FORCE<TV> surface_tension_force;
    CODIMENSIONAL_VISCOSITY_FORCE<TV> viscosity_force;
    CODIMENSIONAL_VISCOELASTIC_FORCE<TV> elasticity_force;
    CODIMENSIONAL_INCOMPRESSIBLE<TV> incompressible;
    CODIMENSIONAL_HEAT_TRANSFER<TV> heat_transfer;
    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV> spatial_partition;

    ////mpi
    MPI_UNIFORM_GRID<GRID<TV> >* mpi_grid;
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_FLUID_EXAMPLE();

    virtual void Write_Output_Files(const int frame);
    virtual void Read_Output_Files(const int frame);
};
}
#endif
