//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Solver/CODIMENSIONAL_HEAT_TRANSFER.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_HEAT_TRANSFER<TV>::CODIMENSIONAL_HEAT_TRANSFER(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :BASE(codimensional_mesh_object_input,mpi_codim_input)
{
}

template<class TV>
void PhysBAM::CODIMENSIONAL_HEAT_TRANSFER<TV>::Update_Heat_Transfer(ARRAY_VIEW<T>& heat,const T dt,const T time)
{
    Reset_Solver();
    Update_G(psi_Ns_on_mesh(1),psi_Ds_on_mesh(1));
    Update_M_inv();
    Update_A();

    VECTOR_ND<T> vol(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        vol(i)=codimensional_mesh_object.Element_Volume(p,type);
    }

    VECTOR_ND<T> u_old(solver.pressure_samples.m);

    //Update_Dirichlet_Boundary_Samples(psi_Ds_on_mesh(d),solver);
    Interpolate_Particle_Attribute_To_Pressure_Sample(particles.temperature,u_old);

    SPARSE_MATRIX_FLAT_NXN<T>& B=solver.A;
    VECTOR_ND<T>& b=solver.b;
    b.Resize(solver.pressure_samples.m);

    for(int i=1;i<=solver.pressure_samples.m;i++){
        T coef=vol(i)/dt;
        B(i,i)+=coef;b(i)=coef*u_old(i);
    }

    solver.Modify_b_With_Dirichlet_Boundary_Conditions(b,solver.psi_D);
    solver.Solve();
    VECTOR_ND<T>& u_new=solver.p;ARRAY<T> unused(particles.X.m);ARRAY_VIEW<T> unused_arrayview(unused);
    Update_Velocity_And_Force_Component_FLIP(heat,unused_arrayview,u_old,u_new,dt);

    if(mpi_codim){
        ARRAY_VIEW<T> heat_array_view(heat);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(heat);
    }
}

//template class CODIMENSIONAL_HEAT_TRANSFER<VECTOR<float,2> >;
template class CODIMENSIONAL_HEAT_TRANSFER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_HEAT_TRANSFER<VECTOR<double,2> >;
template class CODIMENSIONAL_HEAT_TRANSFER<VECTOR<double,3> >;
#endif
