//#####################################################################
// Copyright 2013, Matthew Cong, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_INCOMPRESSIBLE__
#define __CODIMENSIONAL_INCOMPRESSIBLE__
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Matrices/MATRIX_2X2.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_NXN.h>
#include <PhysBAM_Tools/Krylov_Solvers/PCG_SPARSE.h>
#include <PhysBAM_Tools/Math_Tools/FACTORIAL.h>
#include <PhysBAM_Tools/Arrays/INDIRECT_ARRAY.h>
#include <PhysBAM_Tools/Parallel_Computation/THREADED_UNIFORM_GRID.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Constitutive_Models/STRAIN_MEASURE.h>
#include <PhysBAM_Dynamics/Particles/CODIMENSIONAL_FLUID_PARTICLES.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_INCOMPRESSIBLE
{
    typedef typename TV::SCALAR T;
    typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef MATRIX<T,TV::m> T_MATRIX;
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object;   ////mesh object for the incompressible solver
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;   
    CODIMENSIONAL_MESH& mesh;
    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;

    T density;
    T length_scale;
    T thickness_scale;
    T tolerance,relative_tolerance,absolute_tolerance;
    int number_of_regions;

    bool use_target_div;
    bool use_particle_thickness;
    bool use_all_pressure_sample_on_nodes;
    int max_iter_number;

    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_N_on_mesh;
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > psi_D_on_mesh;
    ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> > tar_div_on_mesh;
    HASHTABLE<int> psi_D_on_mesh_hashtable;
    ARRAY<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> > subsystems;

    VECTOR<ARRAY<int>,4> element_index_in_pressure_samples;
    ARRAY<int> particle_index_in_velocity_samples;
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_INCOMPRESSIBLE(CODIMENSIONAL_MESH_OBJECT<TV>& solver_codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input);

    void Make_Divergence_Free(const T dt,const T time);
    void Find_Subsystems();
    void Reset_Subsystems();
    void Update_Subsystem_Matrices();
    void Solve_Subsystems();
    void Apply_Pressures();
    void Apply_Boundary_Conditions();
    void Find_Null_Space_Basis();
    void Find_Null_Space_Basis_2D();
    //void Find_Null_Space_Basis_3D();

    void Check_Divergence();
};
}
#endif
