//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_NODE.h>
#include <PhysBAM_Tools/Grids_Uniform/UNIFORM_GRID_ITERATOR_CELL.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Tools/Math_Tools/clamp.h>
#include <PhysBAM_Tools/Random_Numbers/NOISE.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Tessellation/TORUS_TESSELLATION.h>
#include <PhysBAM_Geometry/Tessellation/SPHERE_TESSELLATION.h>
#include <PhysBAM_Geometry/Basic_Geometry/TORUS.h>
#include <PhysBAM_Geometry/Basic_Geometry/TETRAHEDRON.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "CODIMENSIONAL_INCOMPRESSIBLE_DRIVER.h"

using namespace PhysBAM;

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Initialize_Data()
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=example.particles;

    ARRAY<PAIR<TV,TV> > X_and_V;ARRAY<T> R;ARRAY<T> mass;ARRAY<bool> boundary;
    switch(test){
        //////////////////////////////////////////////////////////////////////////
        ////basic 2d sims, triangles and segments
    case 1: ////one triangle
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2()),TV(TV2((T)(-0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1))));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale,0)),TV(TV2((T)(0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1))));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.5,length_scale*(T)(0.5*sqrt(3.0)))),TV(TV2(0,(T)0.1))));
        reseed=false;remesh=false;
        }break;
    case 2: ////two triangles sharing one edge
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.8,0)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4,length_scale*(T)0.7)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4,length_scale*(T)-0.7)),random.template Get_Direction<TV>()*(T)0.1));
        reseed=false;remesh=false;
        }break;
    case 3:////two triangles sharing one node
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.5,(T)0)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.25,length_scale*(T)0.9)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2((T)0,length_scale*(T)1.5)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.5,length_scale*(T)1.5)),random.template Get_Direction<TV>()*(T)0.1));
        reseed=false;remesh=false;
        }break;
    case 4:////three triangles sharing two edges
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.8,0)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.2,length_scale*(T)0.7)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.6,0)),random.template Get_Direction<TV>()*(T)0.1));
        reseed=false;remesh=false;
        }break;
    case 5:////three triangles sharing one node
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.1,length_scale*(T)-0.4)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.3,length_scale*(T)-0.4)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.25,length_scale*(T)0.9)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)-1.0,length_scale*(T)1.3)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)-0.9,length_scale*(T)1.4)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.2,length_scale*(T)1.4)),random.template Get_Direction<TV>()*(T)0.1));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.3,length_scale*(T)1.3)),random.template Get_Direction<TV>()*(T)0.1));
        reseed=false;remesh=false;
        }break;
    case 6: ////one triangle with one segment
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),TV()));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.8,0)),TV(TV2(0,(T)0.1))));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4,length_scale*(T)0.7)),TV(TV2((T)0.023,(T)0.072))));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(0,length_scale*(T)1.0)),TV(TV2((T)-0.04,(T)0.1))));
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.7,length_scale*(T)0.7)),TV(TV2((T)0.088,(T)-0.04))));
        reseed=false;remesh=false;
        }break;
    case 7: ////one triangle with three segments
        {
            X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T)0.1));
            X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.8,0)),random.template Get_Direction<TV>()*(T)0.1));
            X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4,length_scale*(T)0.7)),random.template Get_Direction<TV>()*(T)0.1));
            X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)-0.7,length_scale*(T)-0.7)),random.template Get_Direction<TV>()*(T)0.1));
            X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.5,length_scale*(T)-0.7)),random.template Get_Direction<TV>()*(T)0.1));
            X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.4,length_scale*(T)1.7)),random.template Get_Direction<TV>()*(T)0.1));
            reseed=false;remesh=false;
        }break;
    case 8:////one segment
        {PHYSBAM_ASSERT(TV::m==2);
        int particle_number=2;
        for(int np=1;np<=particle_number;np++){TV x=TV(TV2(length_scale*(T)np,0));X_and_V.Append(PAIR<TV,TV>(x,TV()));}
        X_and_V(1).y=TV::Axis_Vector(1)*(T)-0.1;X_and_V(2).y=TV::Axis_Vector(1)*(T)0.1;
        }break;
    case 9:////one triangle, testing target divergence
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),TV(TV2((T)(-0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1))));R.Append(length_scale);mass.Append((T)1.0);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale,0)),TV(TV2((T)(0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1))));R.Append(length_scale*(T).5);mass.Append((T)1.0);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.5,length_scale*(T)sqrt(3.0)*(T)0.5)),TV(TV2(0,(T)0.1))));R.Append(length_scale*(T).2);mass.Append((T)1.0);
        initialize_mass_from_geometry=false;reseed=false;remesh=false;
        }break;
    case 10:////two triangles sharing one edge, testing target divergence
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),/*TV((T)(-0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1)*/TV()));R.Append((T).6*length_scale);mass.Append((T)1.0);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale,0)),/*TV((T)(0.5*sqrt(3.0))*(T)0.1,(T)(-0.5)*(T)0.1)*/TV()));R.Append((T).6*length_scale);mass.Append((T)1.0);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)0.5,length_scale*(T)sqrt(3.0)*(T)0.5)),/*TV(0,(T)0.1)*/TV()));R.Append((T).6*length_scale);mass.Append((T)1.0);
        X_and_V.Append(PAIR<TV,TV>(TV(TV2(length_scale*(T)1.5,length_scale*(T)sqrt(3.0)*(T)0.5)),/*TV(0,(T)0.1)*/TV()));R.Append((T).6*length_scale);mass.Append((T)1.0);
        initialize_mass_from_geometry=false;reseed=false;remesh=false;
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////examples of particles
    case 21: ////one-ring of particles
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),TV()));
        int ring_particle_number=6;
        for(int np=1;np<=ring_particle_number;np++){
            T theta=(T)two_pi*(T)np/(T)ring_particle_number;TV location=length_scale*TV(TV2(cos(theta),sin(theta)));
            TV x=location;TV v=random.Get_Uniform_Vector(RANGE<TV>(TV::Constant_Vector((T)-0.1),TV::Constant_Vector((T)0.1)));X_and_V.Append(PAIR<TV,TV>(x,v));
        }
        }break;
    case 22: ////two-ring of particles
        {PHYSBAM_ASSERT(TV::m==2);
        X_and_V.Append(PAIR<TV,TV>(TV(),TV()));
        int ring_particle_number=10;
        for(int np=1;np<=ring_particle_number;np++){
            T theta=(T)two_pi*((T)np+random.Get_Uniform_Number((T)-0.2,(T)0.2))/(T)ring_particle_number;TV location=length_scale*TV(TV2(cos(theta),sin(theta)));
            TV x=location;TV v=random.Get_Uniform_Vector(RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1)));X_and_V.Append(PAIR<TV,TV>(x,v));
        }
        ring_particle_number=16;
        for(int np=1;np<=ring_particle_number;np++){
            T theta=(T)two_pi*((T)np+(T)0.5+random.Get_Uniform_Number((T)-0.2,(T)0.2))/(T)ring_particle_number;TV location=(T)2*length_scale*TV(TV2(cos(theta),sin(theta)));
            TV x=location;TV v=random.Get_Uniform_Vector(RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1)));X_and_V.Append(PAIR<TV,TV>(x,v));
        }
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////low dimensional examples
    case 41: ////thread
        {PHYSBAM_ASSERT(TV::m==2);
        int particle_number=10;
        T initial_distance=length_scale;
        for(int np=1;np<=particle_number;np++){
            TV x=TV(TV2(initial_distance*(T)np,0));
            TV v=random.Get_Uniform_Vector(RANGE<TV>(TV::Constant_Vector((T)-.1),TV::Constant_Vector((T).1)));
            X_and_V.Append(PAIR<TV,TV>(x,v));
        }
        }break;
    case 42: ////band
        {PHYSBAM_ASSERT(TV::m==2);
        int particle_number=100;int layer_number=3;
        T initial_distance=length_scale*(T)0.8;
        for(int nl=1;nl<=layer_number;nl++){
            for(int np=1;np<=particle_number;np++){
                TV x=TV(TV2(initial_distance*(T)np,initial_distance*(T)nl));TV v=random.Get_Uniform_Vector(RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1)));
                X_and_V.Append(PAIR<TV,TV>(x,TV()));
            }
        }
        }break;
    case 43:////circle
        {PHYSBAM_ASSERT(TV::m==2);
        int particle_number=20;
        //T initial_distance=length_scale*(T)0.8;
        T r=length_scale/((T)two_pi/(T)particle_number);
        for(int np=1;np<=particle_number;np++){
            T theta=T(np-1)*(T)two_pi/(T)particle_number;
            TV x=TV(TV2(r*cos(theta),r*sin(theta)));TV v=(T)0.1*TV(TV2((T)cos(theta),(T)sin(theta)));
            X_and_V.Append(PAIR<TV,TV>(x,v));
        }
        }break;
    case 44:////shrinking line segments
        {PHYSBAM_ASSERT(TV::m==2);
        int particle_number=8;
        for(int np=1;np<=particle_number;np++){
            TV x=TV(TV2(length_scale*(T)(np-particle_number/2),0));
            TV v=(x.x>0?TV::Axis_Vector(1)*((T)-0.1):TV::Axis_Vector(1)*((T)0.1));
            X_and_V.Append(PAIR<TV,TV>(x,v));				
        }
        }break;
    case 45:////another shrinking 
        {PHYSBAM_ASSERT(TV::m==2);

        int particle_number=32;
        for(int np=1;np<=particle_number;np++){
            TV x=TV(TV2(length_scale*(T)(np-particle_number/2),0));
            TV v=(TV::Axis_Vector(1)*(T)(-0.1)*sin(x.x*(T)2))+(TV::Axis_Vector(2)*(T)(0.001)*random.Get_Uniform_Number((T)-1,(T)1));
            X_and_V.Append(PAIR<TV,TV>(x,v));								
        }
        }break;
    case 46:////shrinking and stretching at the same time
        {PHYSBAM_ASSERT(TV::m==2);

        int particle_number=128;
        for(int np=1;np<=particle_number;np++){
            TV x=TV(TV2(length_scale*(T)(np-particle_number/2),0));
            TV v=(TV::Axis_Vector(1)*(T)0.1*sin(x.x*(T)2));
            X_and_V.Append(PAIR<TV,TV>(x,v));				
        }
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////surface tension
    case 51:    ////rod with one end fixed, with gravity
        {
            int particle_number=10;
            for(int np=1;np<=particle_number;np++){
                TV x=TV(TV2(0,-length_scale*(T)np));
                X_and_V.Append(PAIR<TV,TV>(x,TV()));
            }
            for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(1,TV::Axis_Vector(d),(T)0));
            has_gravity=true;
        }break;
    case 52:    ////rod with two ends fixed, no gravity
        {
            int particle_number=10;
            for(int np=1;np<=particle_number;np++){
                TV x=TV(TV2(length_scale*sin((T)pi*(T)(np-1)/(T)(particle_number-1)),-length_scale*(T)np));
                X_and_V.Append(PAIR<TV,TV>(x,TV()));
            }
            for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(1,TV::Axis_Vector(d),(T)0));
            for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(particle_number,TV::Axis_Vector(d),(T)0));
            has_surface_tension=true;
        }break;
    case 53:    ////rod with two free ends, no gravity
        {
            int particle_number=20;
            for(int np=1;np<=particle_number;np++){
                TV x=TV(TV2(length_scale*(T)np,0));
                X_and_V.Append(PAIR<TV,TV>(x,TV()));
            }
            has_surface_tension=true;
        }break;
    case 54:    ////ellipsoid
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)8);radii.x*=(T)2;
            DIAGONAL_MATRIX<T,TV::m> radii_matrix=DIAGONAL_MATRIX<T,TV::m>(radii);
            T_ELLIPSOID ellipsoid(TV(),radii_matrix);
            TV_INT counts=TV_INT(ellipsoid.Bounding_Box().Edge_Lengths()/length_scale);
            example.Add_Ellipsoid_Particles(X,boundary,ellipsoid,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 55:    ////cube
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4);
            BOX<TV> box=BOX<TV>(-radii,radii);
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 56:    ////plane
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2*data_scale);
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.y=1;
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 57:    ////plane (volumetric)
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2*data_scale);radii.y=length_scale;
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 58:    ////membrane with two ends fixed (rectangular)
        {
            ARRAY<TV> X;
            TV size=TV::Constant_Vector(length_scale*(T)2*data_scale);size.x=length_scale;size.y*=(T)2;
            BOX<TV> box=BOX<TV>(TV(),size);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.x=1;
            example.Add_Box_Particles(X,boundary,box,counts,(T)1e-3);
            fixed_boundary_box=box;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 59:    ////membrane with two ends fixed (rectangular, volumetric)
        {
            ARRAY<TV> X;
            TV size=TV::Constant_Vector(length_scale*(T)8*data_scale);size.x=length_scale;size.y*=(T)2;
            BOX<TV> box=BOX<TV>(TV(),size);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);
            example.Add_Box_Particles(X,boundary,box,counts,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                X(i).x+=(T)4*length_scale*sin((T)pi*(X(i).y-box.min_corner.y)/box.Edge_Lengths().y);
            }
            fixed_boundary_box=box;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;
        }break;
    case 60:    ////oscillating membrane (circular)
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4*data_scale);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                T r=sqrt(pow<2>(X(i).x)+pow<2>(X(i).y));
                T z=(T)6*length_scale*sin((T)half_pi*(circle.radius-r)/circle.radius);
                X(i)=TV_Compatible<T,TV::m>(X(i).x,X(i).y,z);
            }
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).005;
        }break;
    case 61:    ////membrane in gravity (circular)
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4*data_scale);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,(T)1e-3);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).005;
            has_gravity=true;gravity=TV_Compatible<T,TV::m>((T)0,(T)0,(T)-2);
            use_fixed_mass=true;
            //TORUS<T> torus(TV3(),TV3::Axis_Vector(3),length_scale*(T).25,radii.x);
            //TRIANGULATED_SURFACE<T>* torus_triangulated_surface=TESSELLATION::Generate_Triangles(torus,80,60);
            //FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/common/torus.tri",*torus_triangulated_surface);
        }break;
    case 63:    ////shrinking plane
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)2*data_scale);radii.x*=(T)2;
            BOX<TV> box=BOX<TV>(-radii,radii);  
            TV_INT counts=TV_INT(box.Edge_Lengths()/length_scale);counts.y=1;
            example.Add_Box_Particles(X,boundary,box,counts);
            for(int i=1;i<=X.m;i++){
                TV v=TV::Axis_Vector(1)*(T)(-0.1)*sin(X(i).x*(T)2)+(TV::Axis_Vector(2)*(T)(0.001)*random.Get_Uniform_Number((T)-1,(T)1));
                X_and_V.Append(PAIR<TV,TV>(X(i),v));
            }
            //has_surface_tension=true;//surface_tension_coefficient*=(T).04;
        }break;
    case 64:    ////shrinking circle
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4*data_scale);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,(T)1e-3);
            for(int i=1;i<=X.m;i++){
                T r=sqrt(pow<2>(X(i).x)+pow<2>(X(i).y));
                TV v=r>(T).9*radii.x?(-X(i))*(T).5:TV();
                T z=(T).01*length_scale*random.Get_Uniform_Number((T)-1,(T)1);
                X(i)=TV_Compatible<T,TV::m>(X(i).x,X(i).y,z);
                X_and_V.Append(PAIR<TV,TV>(X(i),v));
            }
            has_surface_tension=true;
        }break;
    case 65:    ////ellipsoid single bubble
        {
            //TRIANGULATED_SURFACE<T>& sphere_triangulated_surface=*(TESSELLATION::Generate_Triangles(sphere));
            TRIANGULATED_SURFACE<T> ellipsolid_triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            FILE_UTILITIES::Read_From_File(stream_type,"data/ellipsoid.tri",ellipsolid_triangulated_surface);
            external_mesh.triangle_mesh.Initialize_Mesh(ellipsolid_triangulated_surface.mesh);
            external_mesh.Update_States(ellipsolid_triangulated_surface.particles.X.m);
            for(int i=1;i<=ellipsolid_triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(ellipsolid_triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_air_pressure_force=true;inside_bubble_pressures=(T)40;bubble_pV=inside_bubble_pressures*Bubble_Volume(ellipsolid_triangulated_surface);outside_air_pressure=(T)40;
            has_surface_tension=true;surface_tension_coefficient*=(T).001;
        }break;
    case 66:    ////sphere bubble with pressure force only 
        {
            SPHERE<TV3> sphere(TV3(),length_scale*(T)8);
            TRIANGULATED_SURFACE<T>& sphere_triangulated_surface=*(TESSELLATION::Generate_Triangles(sphere));
            external_mesh.triangle_mesh.Initialize_Mesh(sphere_triangulated_surface.mesh);
            external_mesh.Update_States(sphere_triangulated_surface.particles.X.m);
            for(int i=1;i<=sphere_triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(sphere_triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_air_pressure_force=true;inside_bubble_pressures=(T)10;bubble_pV=inside_bubble_pressures*Bubble_Volume(sphere_triangulated_surface);outside_air_pressure=(T)15;
        }break;
    case 67:    ////real water bubbles
        {
            length_scale=(T).002;thickness_scale=(T)1e-5;
            TRIANGULATED_SURFACE<T> triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            //FILE_UTILITIES::Read_From_File(stream_type,"../data/sphere.tri",triangulated_surface);
            FILE_UTILITIES::Read_From_File(stream_type,"../data/ellipsoid4.tri",triangulated_surface);
            //SPHERE<TV3> sphere(TV3(),(T).02);
            //TRIANGULATED_SURFACE<T>& triangulated_surface=*(TESSELLATION::Generate_Triangles(sphere));
            external_mesh.triangle_mesh.Initialize_Mesh(triangulated_surface.mesh);
            external_mesh.Update_States(triangulated_surface.particles.X.m);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_air_pressure_force=true;inside_bubble_pressures=air_pressure;bubble_pV=inside_bubble_pressures*Bubble_Volume(triangulated_surface);outside_air_pressure=air_pressure;
            has_surface_tension=true;
        }break;

    case 62:    ////blowing bubbles
        {
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)4*data_scale);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,(T)1e-3);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T).005;
            use_fixed_mass=true;

            has_wind_force=true;
        }break;
    case 68:    ////blowing bubble in wind field solved by grid solver
        {
            length_scale=(T).001;thickness_scale=(T)1e-5;
            ARRAY<TV> X;
            TV radii=TV::Constant_Vector(length_scale*(T)8*data_scale);
            radii=TV_Compatible<T,TV::m>(radii.x,radii.y,length_scale);
            BOX<TV> box=BOX<TV>(-radii*(T)1.2,radii*(T)1.2);
            SPHERE<TV2> circle(TV2(),radii.x);
            TV_INT counts=TV_INT(ceil(box.Edge_Lengths()/length_scale));
            counts=TV_Compatible<int,TV::m>(counts.x,counts.y,1);
            example.Add_Circle_Particles(X,boundary,circle,counts,(T)1e-3);
            fixed_boundary_circle=circle;has_fixed_boundary_condition=true;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_surface_tension=true;surface_tension_coefficient*=(T)1e-2;
            use_fixed_mass=true;

            has_wind_force=true;
            has_dynamic_wind_field=true;
        }break;
    case 69:    ////dumbbell single bubble
        {
            length_scale=(T).002;thickness_scale=(T)1e-5;
            TRIANGULATED_SURFACE<T> triangulated_surface(*(new TRIANGLE_MESH()),*(new GEOMETRY_PARTICLES<TV3>()));
            FILE_UTILITIES::Read_From_File(stream_type,"../data/dumbbell2.tri",triangulated_surface);
            //FILE_UTILITIES::Write_To_File(stream_type,"../data/dumbbell3.tri",triangulated_surface);
            external_mesh.triangle_mesh.Initialize_Mesh(triangulated_surface.mesh);
            external_mesh.Update_States(triangulated_surface.particles.X.m);
            for(int i=1;i<=triangulated_surface.particles.X.m;i++){
                X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(triangulated_surface.particles.X(i)),TV()));
            }

            use_external_mesh=true;
            has_air_pressure_force=true;inside_bubble_pressures=(T)40;bubble_pV=inside_bubble_pressures*Bubble_Volume(triangulated_surface);outside_air_pressure=(T)40;
            has_surface_tension=true;surface_tension_coefficient*=(T).001;
        }break;
        //////////////////////////////////////////////////////////////////////////
        ////water
    case 71:////hydrostatic test
        {PHYSBAM_ASSERT(TV::m==2);
        T tank_width=length_scale*(T)2;
        T tank_height=tank_width*(T)2;
        TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
        TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
        tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);
        TV tank_center=tank_box.Center();
        tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
        tank->X()=tank_center;
        has_tank=true;

        ////water bulk
        TV water_size=TV::Constant_Vector(tank_width);water_size.y=water_size.x*(T).5;
        BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_size);
        TV_INT particle_counts=2*TV_INT(water_size/length_scale);
        ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
        for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

        has_gravity=true;
        }break;
    case 72:    ////droplet falling onto water
        {
            T tank_width=length_scale*(T)80;
            T tank_height=tank_width;
            TV tank_size=TV::Constant_Vector(tank_width);tank_size.y=tank_height;
            TV tank_min_corner=TV::Constant_Vector(-(T).5*tank_width);tank_min_corner.y=(T)0;
            tank_box=BOX<TV>(tank_min_corner,tank_min_corner+tank_size);
            TV tank_center=tank_box.Center();
            tank=&rigid_standard_tests.Add_Analytic_Box(tank_size);
            tank->X()=tank_center;
            has_tank=true;

            ////water bulk
            TV water_box_size=TV::Constant_Vector(tank_width);water_box_size.y=water_box_size.x*(T).2;
            BOX<TV> water_box=BOX<TV>(tank_min_corner,tank_min_corner+water_box_size);
            TV_INT particle_counts=2*TV_INT(water_box_size/length_scale);
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,water_box,particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}

            ////water sphere
            TV water_sphere_center=tank_center;water_sphere_center.y=water_box_size.y*(T)3;
            SPHERE<TV> water_sphere=SPHERE<TV>(water_sphere_center,tank_width*(T).025);
            TV_INT sphere_particle_counts=2*TV_INT(water_sphere.Bounding_Box().Edge_Lengths()/length_scale);
            X.Remove_All();
            example.Add_Sphere_Particles(X,boundary,water_sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(2)*(T)-2));}

            has_gravity=true;has_surface_tension=true;
        }break;

        //////////////////////////////////////////////////////////////////////////
        ////codimensional examples
    case 81:    ////impinging jets
        {
            water_pipe.Add_Source(TV(TV2((T)0,(T)1)),TV(TV2((T)1,(T)0)));
            water_pipe.Add_Source(TV(TV2((T)5,(T)1)),TV(TV2((T)-1,(T)0)));
            water_pipe.time_interval=(T).075;
            water_pipe.particle_mass=density*(T)pi*pow<2>((T).25*length_scale);
            for(int i=1;i<=water_pipe.locations.m;i++){X_and_V.Append(PAIR<TV,TV>(water_pipe.locations(i),water_pipe.velocities(i)));}
            has_water_pipe=true;has_gravity=true;has_surface_tension=true;
        }break;
    case 82:    ////adhesion force
        {
            T s=(T)1;
            for(int i=1;i<=8*(int)s;i++){water_pipe.Add_Source(TV(TV2((T)length_scale*(T)(i-1),(T)0)),TV());}
            water_pipe.time_interval=(T).2;
            water_pipe.particle_mass=density*(T)pi*pow<2>((T).25*length_scale);
            for(int i=1;i<=water_pipe.locations.m;i++){X_and_V.Append(PAIR<TV,TV>(water_pipe.locations(i),water_pipe.velocities(i)));}
            has_water_pipe=true;has_gravity=true;has_surface_tension=true;

            rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(length_scale*(T)10*s, (T)1.0);
            rigid_body->X()=TV(TV2(length_scale*(T)4*s,(T)-1.3*s));
            has_solid=true;has_adhesion_force=true;
        }break;
    case 83:    ////fluid polygon
        {
            T s=(T)4;
            int pipe_width=4*(int)s;
            for(int i=1;i<=pipe_width;i++){water_pipe.Add_Source(TV(TV2((T)length_scale*(T)(i-1),(T)1)),TV::Axis_Vector(2)*(T)-1);}
            water_pipe.time_interval=(T).025;
            water_pipe.particle_mass=density*(T)pi*pow<2>((T).25*length_scale);
            for(int i=1;i<=water_pipe.locations.m;i++){X_and_V.Append(PAIR<TV,TV>(water_pipe.locations(i),water_pipe.velocities(i)));}
            has_water_pipe=true;has_gravity=true;has_surface_tension=true;

            rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(length_scale*(T)20*s));
            rigid_body->X()=TV(TV2(length_scale*(T).5*(T)pipe_width,(T)-4*s));
            has_solid=true;
        }break;


        //////////////////////////////////////////////////////////////////////////
        ////basic 3d sim
    case 101: ////one tetrahedron
        {PHYSBAM_ASSERT(TV::m==3);
        X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale,(T)0,(T)0),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T)(.5*sqrt(3.)),(T)0),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T)(.25*sqrt(3.)),length_scale*(T).5),random.template Get_Direction<TV>()*(T).1));
        for(int i=1;i<=X_and_V.m;i++)mass.Append((T)1.0);
        reseed=false;remesh=false;
        }break;
    case 102: ////one tetrahedron with 3 segments
        {PHYSBAM_ASSERT(TV::m==3);
        X_and_V.Append(PAIR<TV,TV>(TV(),/*random.template Get_Direction<TV>()*(T).1*/TV()));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale,(T)0,(T)0),/*random.template Get_Direction<TV>()*(T).1*/TV()));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T)(.5*sqrt(3.)),(T)0),/*random.template Get_Direction<TV>()*(T).1*/TV()));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T)(.25*sqrt(3.)),length_scale*(T).5),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV()-TV::Constant_Vector(length_scale*(T).5),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale,(T)0,(T)0)+TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T).5,-length_scale*(T).5),random.template Get_Direction<TV>()*(T).1));
        for(int i=1;i<=X_and_V.m;i++)mass.Append((T)1.0);			
        reseed=false;remesh=false;
        }break;
    case 103:   ////one triangle
        {PHYSBAM_ASSERT(TV::m==3);
        X_and_V.Append(PAIR<TV,TV>(TV(),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale,(T)0,(T)0),random.template Get_Direction<TV>()*(T).1));
        X_and_V.Append(PAIR<TV,TV>(TV_Compatible<T,TV::m>(length_scale*(T).5,length_scale*(T)(.5*sqrt(3.)),(T)0),random.template Get_Direction<TV>()*(T).1));
        for(int i=1;i<=X_and_V.m;i++)mass.Append((T)1.0);
        reseed=false;remesh=false;
        }break;
    case 104:  ////impinging jet
        {
            water_pipe.Add_Source(TV(TV2((T)0,(T)1)),TV(TV2((T)2,(T)0)));
            water_pipe.Add_Source(TV(TV2((T)3,(T)1)),TV(TV2((T)-2,(T)0)));
            water_pipe.time_interval=(T).02;
            water_pipe.particle_mass=density*(T)(4./3.*pi)*pow<3>((T).25*length_scale);
            for(int i=1;i<=water_pipe.locations.m;i++){X_and_V.Append(PAIR<TV,TV>(water_pipe.locations(i),water_pipe.velocities(i)));}
            has_water_pipe=true;has_gravity=true;has_surface_tension=false;
        }break;
    case 105:
        {
            TV plane_center=TV();
            TV box_r=TV::Constant_Vector((T)8*length_scale);
            BOX<TV> box=BOX<TV>(plane_center-box_r,plane_center+box_r);
            TV_INT counts=(TV_INT)(box.Edge_Lengths()/length_scale)+1;counts.y=1;
            ARRAY<TV> X;example.Add_Box_Particles(X,boundary,box,counts);
            T v_perturb=(T)1e-3;
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),X(i)+random.Get_Uniform_Vector(-TV::Constant_Vector(v_perturb),TV::Constant_Vector(v_perturb))));}
        }break;
    case 106:	////two colliding water balls
        {
            TV sphere_center=TV();
            SPHERE<TV> sphere=SPHERE<TV>(sphere_center,length_scale*(T)2);
            TV_INT sphere_particle_counts=(TV_INT)(sphere.Bounding_Box().Edge_Lengths()/length_scale);
            ARRAY<TV> X;
            example.Add_Sphere_Particles(X,boundary,sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(1)*(T)1));}

            X.Remove_All();
            sphere.center.x+=(T)5;
            example.Add_Sphere_Particles(X,boundary,sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV::Axis_Vector(1)*(T)-1));}
        }break;
    case 107:   ////water ball falling onto a cube
        {
            TV sphere_center=TV();
            SPHERE<TV> sphere=SPHERE<TV>(sphere_center,length_scale*(T)4);
            TV_INT sphere_particle_counts=(TV_INT)(sphere.Bounding_Box().Edge_Lengths()/length_scale);
            ARRAY<TV> X;
            example.Add_Sphere_Particles(X,boundary,sphere,sphere_particle_counts);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_gravity=true;
            has_surface_tension=true;surface_tension_coefficient=(T)100;

            rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(length_scale*(T)20));
            rigid_body->X()=TV();rigid_body->X().y=-length_scale*(T)30;
            has_solid=true;
        }break;
    case 108:   ////one tet falling onto a cube
        {
            TV tet_center=TV();
            ARRAY<TV> X;
            example.Add_Regular_Simplex(X,tet_center,length_scale);
            for(int i=1;i<=X.m;i++){X_and_V.Append(PAIR<TV,TV>(X(i),TV()));}
            has_gravity=true;

            rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(length_scale*(T)20));
            rigid_body->X()=TV();rigid_body->X().y=-length_scale*(T)30;
            has_solid=true;
        }break;
        //////////////////////////////////////////////////////////////////////////
    default:PHYSBAM_FATAL_ERROR("invalid test id");break;
    }

    for(int i=1;i<=X_and_V.m;i++){
        int index=particles.Add_Particle();
        particles.X(index)=X_and_V(i).x;particles.V(index)=X_and_V(i).y;
    }
    if(R.m>0)for(int i=1;i<=particles.X.m;i++)particles.radius(i)=R(i);
    else for(int i=1;i<=particles.X.m;i++)particles.radius(i)=(T).5*length_scale;
    if(mass.m>0)for(int i=1;i<=particles.X.m;i++){particles.mass(i)=mass(i);}
    else for(int i=1;i<=particles.X.m;i++){particles.mass(i)=(T)1;}
    if(boundary.m>0)for(int i=1;i<=particles.X.m;i++)particles.on_boundary(i)=boundary(i);
    else for(int i=1;i<=particles.X.m;i++)particles.on_boundary(i)=false;

    if(has_wind_force)Initialize_Wind_Force();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Initialize_Wind_Force()
{
    switch(test){
    case 62:    ////blowing soap bubble
        {
            T radius=fixed_boundary_circle.radius;
            //TV wind=TV::Axis_Vector(3)*(T)-1;wind_force.Use_Constant_Wind((T)0,wind);

            TV wind_box_min_corner=TV::Constant_Vector(-radius*(T)2);wind_box_min_corner.x=-radius*(T)2;wind_box_min_corner.y=-radius*(T)2;
            TV wind_box_max_corner=TV::Constant_Vector(radius*(T)8);wind_box_max_corner.x=radius*(T)2;wind_box_max_corner.y=radius*(T)2;
            BOX<TV> wind_box(wind_box_min_corner,wind_box_max_corner);
            TV_INT wind_counts=TV_INT(ceil(wind_box.Edge_Lengths()/length_scale));
            wind_grid.Initialize(wind_counts,wind_box,true);
            wind_velocities.Resize(wind_grid.Domain_Indices(1));

            BOX<TV> turbulence_box(TV::Constant_Vector(-radius*(T)8),TV::Constant_Vector(radius*(T)8));
            TV_INT turbulence_counts=TV_INT::Constant_Vector(16);
            GRID<TV> turbulence_grid;turbulence_grid.Initialize(turbulence_counts,turbulence_box,false);

            turbulence.Initialize_Grid(turbulence_grid);
            turbulence.Generate_Initial_Turbulence();

            TV global_wind=TV::Axis_Vector(3);

            for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(wind_grid);iter.Valid();iter.Next()){
                TV pos=iter.Location();
                T z=TV3(pos).z;

                TV turbulent_wind=(T).5*turbulence.Turbulent_Velocity(pos,0);
                TV wind=global_wind+turbulent_wind;
                wind_velocities(iter.Cell_Index())=wind;
            }

            wind_density=(T)400.;
        }break;
    case 68:    ////blowing bubble 2
        {
            wind_speed=(T)1;
            wind_density=(T)4;

            T radius=fixed_boundary_circle.radius;
            TV wind_box_min_corner=TV::Constant_Vector(-radius*(T)2);wind_box_min_corner.x=-radius*(T)2;wind_box_min_corner.y=-radius*(T)2;
            TV wind_box_max_corner=TV::Constant_Vector(radius*(T)8);wind_box_max_corner.x=radius*(T)2;wind_box_max_corner.y=radius*(T)2;
            BOX<TV> wind_box(wind_box_min_corner,wind_box_max_corner);
            T cell_size=radius*(T).2;
            TV_INT wind_counts=TV_INT(ceil(wind_box.Edge_Lengths()/cell_size));
            wind_grid.Initialize(wind_counts,wind_box,true);
            wind_velocities.Resize(wind_grid.Domain_Indices(1));
            wind_field.Initialize(example.geometrical_codimensional_mesh_object,wind_grid);
            
            TV source_dx=wind_grid.dX*(T).4;
            TV3 source_box_min_corner=TV3(wind_box.min_corner);TV3 source_box_max_corner=TV3(wind_box.max_corner);
            source_box_min_corner.x*=(T).5;source_box_min_corner.y*=(T).5;
            source_box_max_corner.x*=(T).5;source_box_max_corner.y*=(T).5;
            source_box_max_corner.z=source_box_min_corner.z+cell_size;
            BOX<TV> source_box=BOX<TV>(TV_Compatible<T,TV::m>(source_box_min_corner)-source_dx,TV_Compatible<T,TV::m>(source_box_max_corner)+source_dx);
            TV3 source_velocity=TV3::Axis_Vector(3)*wind_speed;
            wind_field.Initialize_Source(source_box,TV_Compatible<T,TV::m>(source_velocity));
            
            wind_field.vorticity_confinement_coeff=(T).5;

            //////initialize wind speed
            //for(UNIFORM_GRID_ITERATOR_FACE<TV> iter(wind_grid);iter.Valid();iter.Next()){
            //    wind_field.velocities(iter.Full_Index())=source_velocity(iter.Axis())*(T).1;
            //}
        }break;
    }

    if(has_dynamic_wind_field){
        for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(wind_grid);iter.Valid();iter.Next()){
            for(int d=1;d<=TV::m;d++){
                wind_velocities(iter.Cell_Index())(d)=(T).5*(wind_field.velocities(d,iter.First_Face_Index(d))+wind_field.velocities(d,iter.Second_Face_Index(d)));
            }
        }
    }

    wind_force.Set_Wind_Density(wind_density);
    wind_force.Use_Spatially_Varying_Wind(/*viscosity*/(T)0,wind_grid,wind_velocities);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Update_Wind_Force(const T time)
{
    if(has_dynamic_wind_field){
        for(UNIFORM_GRID_ITERATOR_CELL<TV> iter(wind_grid);iter.Valid();iter.Next()){
            for(int d=1;d<=TV::m;d++){
                wind_velocities(iter.Cell_Index())(d)=(T).5*(wind_field.velocities(d,iter.First_Face_Index(d))+wind_field.velocities(d,iter.Second_Face_Index(d)));
            }
        }
    }

    switch(test){
    case 62:    ////blowing soap bubble
        {
            TV global_wind=TV::Axis_Vector(3);
            for(UNIFORM_GRID_ITERATOR_NODE<TV> iter(wind_grid);iter.Valid();iter.Next()){
                TV pos=iter.Location();T z=TV3(pos).z;
                TV turbulent_wind=(T).5*turbulence.Turbulent_Velocity(pos,0);
                TV wind=global_wind+turbulent_wind;wind_velocities(iter.Node_Index())=wind;
            }
        }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<TV>::Update_Fixed_Boundary_Conditions()
{
    switch(test){
    case 58:    ////membrane with two ends fixed (rectangular)
        {
            for(int i=1;i<=example.particles.X.m;i++){
                if(abs(example.particles.X(i).y-fixed_boundary_box.min_corner.y)<length_scale*(T).1||abs(example.particles.X(i).y-fixed_boundary_box.max_corner.y)<length_scale*(T).1){
                    for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    if(use_fixed_mass){
                        example.particles.mass(i)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                    }
                }
            }
        }break;
    case 59:    ////membrane with two ends fixed (rectangular, volumetric)
        {
            for(int i=1;i<=example.particles.X.m;i++){
                if(abs(example.particles.X(i).y-fixed_boundary_box.min_corner.y)<length_scale*(T).1||abs(example.particles.X(i).y-fixed_boundary_box.max_corner.y)<length_scale*(T).1){
                    for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    if(use_fixed_mass){
                        example.particles.mass(i)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                    }
                }
            }
        }break;
    case 60:    ////membrane (circular)
    case 61:    ////membrane with gravity
    case 62:    ////blowing soap bubble
    case 68:    ////another blowing bubble
        {
            for(int i=1;i<=example.particles.X.m;i++){
                TV3 pos=TV3(example.particles.X(i));
                if(abs(TV2(pos.x,pos.y).Magnitude()-fixed_boundary_circle.radius)<length_scale*(T).1&&abs(pos.z)<length_scale*(T).1){
                    for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_INCOMPRESSIBLE_NEUMANN_BOUNDARY_SAMPLE<TV>(i,TV::Axis_Vector(d),(T)0));
                    example.particles.F(i)=TV();example.particles.surface_tension_force(i)=TV();example.particles.pressure_force(i)=TV();
                    if(use_fixed_mass){
                        example.particles.mass(i)=example.geometrical_codimensional_mesh_object.Nodal_Triangle_Area_Sum(i)*thickness_scale*density;
                    }
                }
            }
        }break;
    }
    //LOG::cout<<"fixed psi_N: "<<example.incompressible.psi_N.m<<std::endl;
    //for(int i=1;i<=example.incompressible.psi_N.m;i++){
    //    LOG::cout<<example.incompressible.psi_N(i).index<<", "<<example.incompressible.psi_N(i).v<<std::endl;
    //}
}

template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,2> >::Initialize_Data();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,3> >::Initialize_Data();
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,2> >::Initialize_Data();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,3> >::Initialize_Data();
#endif

template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,2> >::Initialize_Wind_Force();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,3> >::Initialize_Wind_Force();
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,2> >::Initialize_Wind_Force();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,3> >::Initialize_Wind_Force();
#endif

template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,2> >::Update_Fixed_Boundary_Conditions();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<float,3> >::Update_Fixed_Boundary_Conditions();
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,2> >::Update_Fixed_Boundary_Conditions();
template void CODIMENSIONAL_INCOMPRESSIBLE_DRIVER<VECTOR<double,3> >::Update_Fixed_Boundary_Conditions();
#endif
