//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Tools/Math_Tools/clamp.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Solver/CODIMENSIONAL_SOLVER_DRIVER.h"

#include <string>
#include <sstream>
#include <fstream>
using std::string;
using std::stringstream;

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::CODIMENSIONAL_SOLVER_DRIVER()
    :test(1),output_directory(TV::m==2?"output":"output3"),model_directory("../model"),stream_type(T()),first_frame(0),last_frame(1),current_frame(0),restart_frame(0),
    write_substep(false),current_substep_frame(0),last_substep_frame(-1),frame_rate(4000),time(0),current_time(0),cfl((T)1.),
    analytic_function(1),
    equation(POISSON),length_scale((T)1.),thickness_scale((T)1.),codimensional_mesh(&segment_mesh,&triangle_mesh,&tetrahedron_mesh),codimensional_mesh_object(codimensional_mesh,particles),deformable_geometry_collection(particles,false),
    mesher(codimensional_mesh_object),viscosity_solver(0),use_staggered_dof(false),use_mpi(false),mpi_grid(0),mpi_codim(0)
{
    Initialize_Particles();
    Initialize_Read_Write_General_Structures();
}

template<class TV>
PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::~CODIMENSIONAL_SOLVER_DRIVER()
{
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize()
{
    Initialize_Test_Parameters();
    Initialize_Mpi();
    Initialize_Mesh();
    Initialize_Boundary_Conditions();
    Initialize_Solver();
    Write_Output_Files(0);

    //Test_Area_Weighted_Normals();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Test_Parameters()
{
    switch(test){
        ////solving Poisson equation on a squared triangle mesh
    case 1:
    case 2:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1.2),TV::Constant_Vector((T)1.2));

            TV domain_size=TV((T)2,(T)2,(T)0);length_scale=(T).01;thickness_scale=(T)1;
            m=(int)(domain_size.y/length_scale)+1;n=(int)(domain_size.x/length_scale)+1;
            TV min_corner=(T)-.5*domain_size;min_corner.z=(T)0;

            CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,min_corner,length_scale);

            equation=POISSON;
            use_staggered_dof=false;
        }break;
    case 3: ////solving viscosity on a squared triangle mesh
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));

            domain_size=(T)1;int resolution=20;m=resolution+1;n=resolution+1;
            TV min_corner=TV::Constant_Vector((T)-.5*domain_size);min_corner.z=(T)0;
            length_scale=domain_size/(T)resolution;thickness_scale=(T)1;

            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,min_corner,length_scale);

            equation=VISCOSITY;
        }break;
    case 4: ////solving heat equation on a rectangular mesh
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-10),TV::Constant_Vector((T)10));

            domain_size=(T)1;int resolution=20;
            m=resolution+1;n=resolution+1;
            TV min_corner=TV::Constant_Vector((T)-.5*domain_size);min_corner.z=(T)0;
            length_scale=domain_size/(T)resolution;thickness_scale=length_scale;

            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,min_corner,length_scale);

            equation=HEAT;
            use_staggered_dof=false;
            analytic_function=2;
        }break;
    case 5:
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-10),TV::Constant_Vector((T)10));

            domain_size=(T)1;int resolution=10;
            m=resolution+1;n=resolution+1;
            TV min_corner=TV::Constant_Vector((T)-.5*domain_size);//min_corner.z*=(T).5;
            TV max_corner=TV::Constant_Vector((T).5*domain_size);//max_corner.z*=(T).5;
            length_scale=domain_size/(T)resolution;thickness_scale=length_scale;

            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,min_corner,max_corner,length_scale);

            equation=HEAT;
            use_staggered_dof=false;
            analytic_function=2;
        }break;
    case 6: ////solve poisson equation on filament
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            domain_size=(T)1;int resolution=50;
            TV e1=TV::Axis_Vector(1)*((T)-.5*domain_size);TV e2=e1+TV::Axis_Vector(1)*domain_size;
            length_scale=domain_size/(T)resolution;thickness_scale=(T)1;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);
            
            equation=POISSON;
            use_staggered_dof=true;
            //analytic_function=1;
        }break;
    case 7: ////solve poisson equation on triangles and filaments
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            domain_size=(T)1;int resolution=5;
            TV e1=TV::Axis_Vector(1)*((T)-.5*domain_size);TV e2=e1+TV::Axis_Vector(1)*domain_size;
            length_scale=domain_size/(T)resolution;thickness_scale=(T)1;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Square_Sticks(initial_mesh,initial_X,resolution,resolution,TV(),length_scale);

            equation=POISSON;
            use_staggered_dof=true;
        }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Mpi()
{
    TV_INT global_grid_counts=TV_INT(global_grid_domain.Edge_Lengths()/((T)1.51*length_scale));
    global_grid.Initialize(global_grid_counts,global_grid_domain,false);
    number_of_ghost_cells=3;

    if(use_mpi){
        if(processes_per_dimension==TV_INT::Constant_Vector(0)){processes_per_dimension=TV_INT::Constant_Vector(1);processes_per_dimension.x*=2;}
        mpi_grid=new MPI_UNIFORM_GRID<GRID<TV> >(global_grid,number_of_ghost_cells,false,processes_per_dimension);
        mpi_codim=new MPI_CODIMENSIONAL<TV>(*mpi_grid,number_of_ghost_cells,particles,codimensional_mesh);
        output_directory=output_directory+STRING_UTILITIES::string_sprintf("/%d",(mpi_grid->rank+1));
        LOG::cout<<"node: "<<mpi_codim->node_id<<": "<<mpi_codim->mpi_grid.local_grid.domain<<", "<<mpi_codim->mpi_grid.local_grid.counts<<std::endl;
        mesher.Initialize_Mpi(mpi_codim);
    }

    Set_Output_Directory(output_directory);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Mesh()
{
    mesher.Initialize_Length_Scale(length_scale,length_scale,thickness_scale);
    ARRAY_VIEW<TV> init_X_array_view(initial_X);mesher.Initialize_From_External_Mesh_And_Positions(initial_mesh,ARRAY_VIEW<TV>(init_X_array_view));
    mesher.Initialize_Particle_Mass_Based_On_Geometry(thickness_scale,thickness_scale,false);
    mesher.Update_Particle_Thickness(false);

    deformable_geometry_collection.Add_Structure(codimensional_mesh_object.Get_Mesh_Object_1D());
    deformable_geometry_collection.Add_Structure(codimensional_mesh_object.Get_Mesh_Object_2D());
    deformable_geometry_collection.Add_Structure(codimensional_mesh_object.Get_Mesh_Object_3D());
    if(deformable_geometry_collection.structures.m>0)deformable_geometry_collection.structures.Last()->Update_Number_Nodes();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Get_Boundary_Particles_From_Herring_Bone_Triangle_Mesh_Particles(int ring_num,HASHTABLE<int>& boundary_particle_hashtable)
{
    for(int r=0;r<ring_num;r++){
        for(int i=m*r+1;i<=(r+1)*m;i++){boundary_particle_hashtable.Set(i);}
        for(int i=(n-1-r)*m+1;i<=(n-r)*m;i++){boundary_particle_hashtable.Set(i);}
        for(int i=1+r;i<=(n-1)*m+1+r;i+=m){boundary_particle_hashtable.Set(i);}
        for(int i=m-r;i<=n*m-r;i+=m){boundary_particle_hashtable.Set(i);}
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Boundary_Conditions()
{
    switch(test){
    case 1: 
        {
            HASHTABLE<int> boundary_particle_hashtable;Get_Boundary_Particles_From_Herring_Bone_Triangle_Mesh_Particles(1,boundary_particle_hashtable);
            for(HASHTABLE_ITERATOR<int> iter(boundary_particle_hashtable);iter.Valid();iter.Next()){
                int p=iter.Key();
                if(mpi_codim)p=mpi_codim->Guid_To_Vtx(p);
                if(p!=0){
                    if(true){
                        T boundary_value=Analytic_Solution(particles.X(p));   
                        psi_D.Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(p,SIMPLEX_PARTICLE,boundary_value));
                    }
                    else{
                        TV boundary_gradient=Analytic_Gradient(particles.X(p));
                        psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(p,boundary_gradient.Normalized(),boundary_gradient.Magnitude()));
                        //LOG::cout<<"psi_N: "<<p<<", "<<boundary_gradient.Normalized()<<", "<<boundary_gradient.Magnitude()<<std::endl;
                    }
                }
            }
        }break;
    case 2:
        {
            HASHTABLE<int> boundary_particle_hashtable;Get_Boundary_Particles_From_Herring_Bone_Triangle_Mesh_Particles(1,boundary_particle_hashtable);

            TV center=TV();
            for(int i=1;i<=particles.X.m;i++){
                T distance=(particles.X(i)-center).Magnitude();
                if(distance<(T)2*length_scale){
                    boundary_particle_hashtable.Set(i);
                }
            }

            for(HASHTABLE_ITERATOR<int> iter(boundary_particle_hashtable);iter.Valid();iter.Next()){
                int p=iter.Key();
                T boundary_value=Analytic_Solution(particles.X(p));
                psi_D.Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(p,SIMPLEX_PARTICLE,boundary_value));
            }
        }break;
    case 3:
        {
            int p=1;
            if(mpi_codim)p=mpi_codim->Guid_To_Vtx(p);
            if(p!=0)particles.V(p).z=(T)1;
        }break;
    case 4:
    case 5:
        {
            for(int i=1;i<=particles.X.m;i++){
                if(particles.X(i).Magnitude()<(T).2){
                    particles.V(i).z=Analytic_Solution(particles.X(i),(T)0);
                }
            }
        }break;
    case 6:
        {
            HASHTABLE<int> boundary_particle_hashtable;
            for(int i=1;i<=1;i++){boundary_particle_hashtable.Set(i);boundary_particle_hashtable.Set(particles.X.m+1-i);}
            for(HASHTABLE_ITERATOR<int> iter(boundary_particle_hashtable);iter.Valid();iter.Next()){
                int p=iter.Key();
                if(mpi_codim)p=mpi_codim->Guid_To_Vtx(p);
                if(p!=0){
                    T boundary_value=Analytic_Solution(particles.X(p));   
                    psi_D.Append(CODIMENSIONAL_DIRICHLET_BOUNDARY<T>(p,SIMPLEX_PARTICLE,boundary_value));
                } 
            }
        }break;
    case 7:
        {

        }break;
    }

    //LOG::cout<<"psi_D: "<<psi_D.m<<std::endl;
    //for(int i=1;i<=psi_D.m;i++){
    //    LOG::cout<<"["<<psi_D(i).index<<","<<psi_D(i).v<<"] ";
    //}LOG::cout<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Update_Boundary_Conditions(const T time)
{
    switch(test){
    case 4:
    case 5:
        {
            for(int i=1;i<=particles.X.m;i++){
                if(particles.X(i).Magnitude()<(T).2){
                    particles.V(i).z=Analytic_Solution(particles.X(i),(T)0);
                }
            }
        }break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Execute_Main_Program()
{
    Solve();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Write_Output_Files(const int frame)
{
    std::string f=STRING_UTILITIES::string_sprintf("%d",frame);

    FILE_UTILITIES::Create_Directory(output_directory);
    FILE_UTILITIES::Create_Directory(output_directory+"/common");
    FILE_UTILITIES::Create_Directory(output_directory+"/"+f);
    FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/"+f+"/codimensional_fluid_particles",particles);
    deformable_geometry_collection.Write(stream_type,output_directory,frame,frame,true);

    if(frame==0) FILE_UTILITIES::Write_To_Text_File(output_directory+"/common/first_frame",frame,"\n");
    if(frame==0&&mpi_codim)FILE_UTILITIES::Write_To_File(stream_type,output_directory+"/common/grid",mpi_codim->mpi_grid.local_grid);
    FILE_UTILITIES::Write_To_Text_File(output_directory+"/common/last_frame",frame,"\n");

    LOG::cout<<"---------------------------------------- Finish Frame "<<frame<<" ----------------------------------------"<<std::endl;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Laplacian(const TV& location)
{
    switch(analytic_function){
    case 1:return (T)(2*2);
    case 2:return exp(location.x)*((T)4*location.x*sin(location.y)+location.y*location.y+(T)2*sin(location.y))+(T)2;
    case 3:return exp(-location.Magnitude_Squared())*((T)4*location.Magnitude_Squared()-(T)2*(T)2);
    case 4:return -(T)2*(T)pi*(T)pi*sin((T)pi*location.x)*sin((T)pi*location.y);
    case 5:{T c=(T)5;return (4*c*c*pow<2>(location.x)+4*c*c*pow<2>(location.y)-4*c)*exp(-(pow<2>(location.x)+pow<2>(location.y))*c);}
    case 6:return (T)2;
    case 7:return (T)0;
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");
    }
}

template<class TV>
TV PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Gradient(const TV& location)
{
    switch(analytic_function){
    case 1:return TV((T)2*location.x,(T)2*location.y,(T)0);
    case 6:return TV((T)2*location.x,(T)0,(T)0);
    //case 6:return TV((T)0,(T)2*location.y,(T)0);
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");
    }
    return TV();
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Solution(const TV& location)
{
    switch(analytic_function){
    case 1:return location.Magnitude_Squared();
    case 2:return exp(location.x)*(location.x*location.x*sin(location.y)+location.y*location.y);
    case 3:return exp(-location.Magnitude_Squared());
    case 4:return sin((T)pi*location.x)*sin((T)pi*location.y);
    case 5:{T c=(T)5;return exp(-(pow<2>(location.x)+pow<2>(location.y))*c);}
    case 6:return location.x*location.x;
    //case 6:return location.y*location.y;
    case 7:return (T)0;
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");
    }
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Solution(const TV& location,const T t)
{
    switch(analytic_function){
    case 1:return exp((T)-2*t)*sin(location.x)*sin(location.y);break;
    case 2:return (T)1;break;
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");break;
    }
}

template<class TV>
TV PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Gradient(const TV& location,const T t)
{
    switch(analytic_function){
    case 1:{TV g;g.x=exp((T)-2*t)*cos(location.x)*sin(location.y);g.y=exp((T)-2*t)*sin(location.x)*cos(location.y);return g;}break;
    case 2:return TV();break;
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");break;
    }
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Analytic_Laplacian(const TV& location,const T t)
{
    switch(analytic_function){
    case 1:return (T)-2*exp((T)-2*t)*sin(location.x)*sin(location.y);break;
    case 2:return (T)0;break;
    default:PHYSBAM_FATAL_ERROR("invalid analytic function");break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Solver()
{
    switch(equation){
    case POISSON:Initialize_Poisson_Solver();break;
    case VISCOSITY:Initialize_Viscosity_Solver();break;
    case HEAT:Initialize_Heat_Solver();break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Solve()
{
    switch(equation){
    case POISSON:Solve_Poisson_Equation();break;
    case VISCOSITY:Solve_Viscosity_Equation();break;
    case HEAT:Solve_Heat_Equation();break;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Poisson_Solver()
{
    solver.mpi_codim=mpi_codim;
    solver.Clear();
    solver.use_all_pressure_sample_on_nodes=!use_staggered_dof;
    solver.use_particle_thickness=false;
    solver.length_scale=length_scale;
    solver.thickness_scale=thickness_scale;    
    solver.codimensional_mesh_object=&codimensional_mesh_object;
    solver.element_index_in_pressure_samples=&element_index_in_pressure_samples;
    solver.particle_index_in_velocity_samples=&particle_index_in_velocity_samples;
    solver.psi_D_on_mesh_hashtable=new HASHTABLE<int>();

    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(solver,&psi_N,&psi_D);

    if(mpi_codim){
        mpi_codim->Update_Boundary_And_Ghost_Pressure_Indices(solver.pressure_samples,codimensional_mesh,solver.use_all_pressure_sample_on_nodes,solver.boundary_pressure_indices,solver.ghost_pressure_indices,solver.domain_pressure_indices);
        LOG::cout<<"node "<<mpi_codim->node_id<<": bdry: "<<solver.boundary_pressure_indices.m<<", ghost: "<<solver.ghost_pressure_indices.m<<": domain: "<<solver.domain_pressure_indices.m<<std::endl;
    }

    solver.Update_G(solver.G);
    solver.Modify_G_With_Boundary_Conditions(solver.G,solver.psi_N);

    ////update M_inv as 1/V
    //solver.M_inv.Resize(solver.velocity_samples.m*TV::m);
    //for(int i=1;i<=solver.velocity_samples.m;i++){
    //    int p=solver.velocity_samples(i);
    //    T volume=codimensional_mesh_object.Nodal_Volume(p);
    //    T one_over_volume=(T)1/volume;
    //    for(int d=1;d<=TV::m;d++)solver.M_inv((TV::m*(i-1))+d)=one_over_volume;
    //}

    solver.Update_V_inv_With_Dirichlet_Boundary_Conditions(solver.M_inv);

    solver.Update_A(solver.G,solver.M_inv,solver.A);

    solver.b.Resize(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;int type=solver.pressure_samples(i).type;
        T volume=codimensional_mesh_object.Element_Partial_Volume(p,type,(*solver.psi_D_on_mesh_hashtable));
        TV pos=codimensional_mesh_object.Element_Center(p,type);
        solver.b(i)=-volume*Analytic_Laplacian(pos);
    }

    //LOG::cout<<"b0: "<<solver.b<<std::endl;

    solver.Modify_b_With_Dirichlet_Boundary_Conditions(solver.b,solver.psi_D);
    solver.Modify_b_With_Neumann_Boundary_Conditions(solver.b,solver.psi_N);

    //LOG::cout<<"G: "<<solver.G<<std::endl;
    //LOG::cout<<"A: "<<solver.A<<std::endl;
    //LOG::cout<<"M_inv: "<<solver.M_inv<<std::endl;    
    //LOG::cout<<"b: "<<solver.b<<std::endl;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Solve_Poisson_Equation()
{
    solver.Solve();
    //LOG::cout<<"p: "<<solver.p<<std::endl;

    ////visualize results
    ARRAY<T> u;CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Interpolate_Pressure_To_Particles(solver,u);

    for(int i=1;i<=psi_D.m;i++){
        int p=psi_D(i).index;//int type=psi_D(i).type;
        T v=psi_D(i).v;
        if(CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Full_Ring_Dirichlet_Boundary(codimensional_mesh,(*solver.psi_D_on_mesh_hashtable),p,SIMPLEX_PARTICLE)){u(p)=v;}
    }

    //LOG::cout<<"u: "<<u<<std::endl;
    for(int i=1;i<=particles.X.m;i++){particles.X(i).z=u(i);}

    Write_Output_Files(1);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Viscosity_Solver()
{
    viscosity_solver=new CODIMENSIONAL_VISCOSITY_FORCE<TV>(codimensional_mesh_object);
    viscosity_solver->kinematic_viscosity=(T)1;
    viscosity_solver->density=(T)1;
    viscosity_solver->thickness_scale=(T)1;
    viscosity_solver->Use_Staggered_DoF(use_staggered_dof);
    viscosity_solver->Initialize_Mpi(mpi_codim);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Solve_Viscosity_Equation()
{
    viscosity_solver->Update_Position_Based_State(0,false);
    viscosity_solver->Add_Velocity_Independent_Forces(particles.F,(T)0);

    Write_Output_Files(1);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Initialize_Heat_Solver()
{
    viscosity_solver=new CODIMENSIONAL_VISCOSITY_FORCE<TV>(codimensional_mesh_object);
    viscosity_solver->kinematic_viscosity=(T)1;
    viscosity_solver->density=(T)1;
    viscosity_solver->thickness_scale=(T)1;
    viscosity_solver->Use_Staggered_DoF(use_staggered_dof);
    viscosity_solver->Initialize_Mpi(mpi_codim);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Solve_Heat_Equation()
{
    T time=(T)0;
    for(int frame=first_frame+1;frame<=last_frame;frame++){
        for(int k=1;k<=10;k++){
            T dt=(T)2;
            Update_Boundary_Conditions(time);
            if(mpi_codim)mpi_codim->Barrier();
            viscosity_solver->Update_Position_Based_State(0,false);
            viscosity_solver->Add_Velocity_Dependent_Forces(particles.V,particles.F,dt,time);

            //for(int i=1;i<=particles.X.m;i++){particles.F(i)=TV();}
            //viscosity_solver->Add_Velocity_Independent_Forces(particles.F,(T)0);
            //for(int i=1;i<=particles.X.m;i++){particles.V(i)+=particles.F(i)*dt;}

            Update_Boundary_Conditions(time);
            time+=dt;
        }

        Write_Output_Files(frame);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_SOLVER_DRIVER<TV>::Test_Area_Weighted_Normals()
{
    ARRAY<TV> X;X.Append(TV());X.Append(TV::Axis_Vector(1));X.Append(TV::Axis_Vector(2));X.Append(TV::Axis_Vector(3));ARRAY_VIEW<TV> X_array_view(X);
    ARRAY<int> contact;contact.Append(1);contact.Append(1);contact.Append(0);contact.Append(0);ARRAY_VIEW<int> contact_array_view(contact);
    TV_INT4 tet_e(1,2,3,4);
    
    TV tet_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,tet_e,1);
    TV tet_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,tet_e,2);
    TV tet_n3=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,tet_e,3);
    TV tet_n4=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,tet_e,4);

    TV_INT3 tri_e(1,2,3);
    TV tri_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,tri_e,1);
    TV tri_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,tri_e,2);
    TV tri_n3=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,tri_e,3);

    TV_INT2 seg_e(1,2);
    TV seg_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(X_array_view,seg_e,1);
    TV seg_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(X_array_view,seg_e,2);

    LOG::cout<<"tet n: "<<tet_n1<<", "<<tet_n2<<", "<<tet_n3<<", "<<tet_n4<<std::endl;
    LOG::cout<<"tri n: "<<tri_n1<<", "<<tri_n2<<", "<<tri_n3<<std::endl;
    LOG::cout<<"seg n: "<<seg_n1<<", "<<seg_n2<<std::endl;

    tet_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,contact_array_view,tet_e,1);
    tet_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,contact_array_view,tet_e,2);
    tet_n3=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,contact_array_view,tet_e,3);
    tet_n4=CODIMENSIONAL_MESH_OBJECT<TV>::Face_Area_Weighted_Volumetric_Vertex_Normal(X_array_view,contact_array_view,tet_e,4);

    tri_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,contact_array_view,tri_e,1);
    tri_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,contact_array_view,tri_e,2);
    tri_n3=CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(X_array_view,contact_array_view,tri_e,3);

    seg_n1=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(X_array_view,contact_array_view,seg_e,1);
    seg_n2=CODIMENSIONAL_MESH_OBJECT<TV>::Segment_Vertex_Normal(X_array_view,contact_array_view,seg_e,2);

    LOG::cout<<"contact tet n: "<<tet_n1<<", "<<tet_n2<<", "<<tet_n3<<", "<<tet_n4<<std::endl;
    LOG::cout<<"contact tri n: "<<tri_n1<<", "<<tri_n2<<", "<<tri_n3<<std::endl;
    LOG::cout<<"contact seg n: "<<seg_n1<<", "<<seg_n2<<std::endl;
}

//template class CODIMENSIONAL_SOLVER_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_SOLVER_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_SOLVER_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_SOLVER_DRIVER<VECTOR<double,3> >;
#endif
