//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_SOLVER_DRIVER__
#define __CODIMENSIONAL_SOLVER_DRIVER__
#include <PhysBAM_Tools/Grids_Uniform_Arrays/ARRAYS_ND.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Solids_Geometry/DEFORMABLE_GEOMETRY_COLLECTION.h>
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_SOLVER_DRIVER
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,SEGMENT_MESH,TRIANGLE_MESH>::TYPE T_SURFACE_MESH;
    typedef typename IF<TV::m==2,SEGMENTED_CURVE_2D<T>,TRIANGULATED_SURFACE<T> >::TYPE T_SURFACE_MESH_OBJECT;
    typedef typename IF<TV::m==2,TRIANGULATED_AREA<T>,TETRAHEDRALIZED_VOLUME<T> >::TYPE T_VOLUME_MESH_OBJECT;
public:
    int test;
    std::string output_directory;
    std::string model_directory;
    STREAM_TYPE stream_type;
    int first_frame,last_frame,current_frame,restart_frame;
    bool write_substep;
    int current_substep_frame,last_substep_frame;
    T frame_rate;
    T time,current_time;
    T cfl;
    int analytic_function;

    enum EQUATION{POISSON,VISCOSITY,HEAT} equation;
    T domain_size;
    int m;
    int n;
    T length_scale;
    T thickness_scale;
    CODIMENSIONAL_FLUID_PARTICLES<TV> particles;  
    SEGMENT_MESH segment_mesh;
    TRIANGLE_MESH triangle_mesh;
    TETRAHEDRON_MESH tetrahedron_mesh;
    CODIMENSIONAL_MESH codimensional_mesh;
    CODIMENSIONAL_MESH_OBJECT<TV> codimensional_mesh_object;
    DEFORMABLE_GEOMETRY_COLLECTION<TV> deformable_geometry_collection;
    CODIMENSIONAL_MESHING_3D<T> mesher; 
    
    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> solver;
    VECTOR<ARRAY<int>,4> element_index_in_pressure_samples;
    ARRAY<int> particle_index_in_velocity_samples;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_N;
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > psi_D;

    CODIMENSIONAL_VISCOSITY_FORCE<TV>* viscosity_solver;
    bool use_staggered_dof;

    ////external mesh
    CODIMENSIONAL_MESH initial_mesh;
    ARRAY<TV> initial_X;

    ////mpi
    bool use_mpi;
    TV_INT processes_per_dimension;
    RANGE<TV> global_grid_domain;
    int number_of_ghost_cells;
    GRID<TV> global_grid;
    MPI_UNIFORM_GRID<GRID<TV> >* mpi_grid;
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_SOLVER_DRIVER();
    ~CODIMENSIONAL_SOLVER_DRIVER();
    void Set_Output_Directory(const std::string& output_dir){output_directory=output_dir;}
    void Initialize();
    void Initialize_Mpi();
    void Initialize_Test_Parameters();
    void Initialize_Mesh();
    void Initialize_Solver();
    void Initialize_Boundary_Conditions();
    void Update_Boundary_Conditions(const T time);
    void Solve();
    void Initialize_Poisson_Solver();
    void Solve_Poisson_Equation();
    void Initialize_Viscosity_Solver();
    void Solve_Viscosity_Equation();
    void Initialize_Heat_Solver();
    void Solve_Heat_Equation();

    void Execute_Main_Program();
    void Write_Output_Files(const int frame);
    T Time_At_Frame(int frame){return (T)frame/frame_rate;}
    int Frame_At_Time(T time){return (int)(time*frame_rate);}

    T Analytic_Laplacian(const TV& location);
    TV Analytic_Gradient(const TV& location);
    T Analytic_Solution(const TV& location);

    T Analytic_Laplacian(const TV& location,const T t);
    TV Analytic_Gradient(const TV& location,const T t);
    T Analytic_Solution(const TV& location,const T t);

    void Get_Boundary_Particles_From_Herring_Bone_Triangle_Mesh_Particles(int ring_num,HASHTABLE<int>& boundary_particle_hashtable);

    void Test_Area_Weighted_Normals();
};

}
#endif
