//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Solver/CODIMENSIONAL_SURFACE_TENSION_FORCE.h"
using namespace PhysBAM;

template<class T_input>
void PhysBAM::CODIMENSIONAL_SURFACE_TENSION_FORCE<VECTOR<T_input,3> >::Add_Velocity_Independent_Forces(ARRAY_VIEW<TV> F,const T time)
{
    if(viscous){Add_Velocity_Independent_Forces_Viscous(F,time);return;}

    T tet_surface_tension=surface_tension;
    T tri_surface_tension=triangle_surface_tension>(T)0?triangle_surface_tension:surface_tension;
    T seg_surface_tension=segment_surface_tension>(T)0?segment_surface_tension:surface_tension;

    ////boundary of tetrahedron mesh
    for(int i=1;i<=(*tetrahedron_mesh.boundary_mesh).elements.m;i++){
        const TV_INT3& v=(*tetrahedron_mesh.boundary_mesh).elements(i);
        if(water_tank&&water_tank->Is_Tank_Contact_Element(v))continue;   ////triangle contacting the solid boundary
        if(Element_Or(particles.inactive,v))continue;

        for(int d=1;d<=3;d++){
            int ix=(d-1)%3+1;int iy=(d)%3+1;int iz=(d+1)%3+1;
            T surface_tension=tet_surface_tension;
            if(contact) F(v(d))+=surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(particles.X,particles.contact,v(ix),v(iy),v(iz)));
            else F(v(d))+=surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(particles.X,v(ix),v(iy),v(iz)));
        }
    }

    ////triangle mesh
    //ARRAY<TV> thickness_F;thickness_F.Resize(F.m);thickness_F.Fill(TV());
    for(int i=1;i<=triangle_mesh.elements.m;i++){
        if(Element_Or(particles.inactive,triangle_mesh.elements(i)))continue;

        VECTOR<TV,3> p;VECTOR<TV,3> thickened_p;
        for(int d=1;d<=3;d++){
            int v=triangle_mesh.elements(i)(d);p(d)=particles.X(v);
            //thickened_p(d)=p(d)+particles.area_weighted_normal(v)*(T).5*particles.thickness(v);
        }
        for(int d=1;d<=3;d++){
            int v=triangle_mesh.elements(i)(d);
            int i1=(d-1)%3+1;int i2=(d)%3+1;int i3=(d+1)%3+1;
            bool c1=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i1)));
            bool c2=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i2)));
            bool c3=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i3)));
            //F(v)+=(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(p(ix),p(iy),p(iz)));
            //thickness_F(v)+=(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(thickened_p(ix),thickened_p(iy),thickened_p(iz)));

            //bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(v));
            bool apply_surface_tension=true;
            if(apply_surface_tension){
                if(contact) F(v)+=(T)2*tri_surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(c1,c2,c3,p(i1),p(i2),p(i3)));
                else F(v)+=(T)2*tri_surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(p(i1),p(i2),p(i3)));
            }
        }
    }

    ////rim surface tension, boundary of triangle mesh
    for(int i=1;i<=mesh.rim_segment_mesh.elements.m;i++){
        const TV_INT2& segment=mesh.rim_segment_mesh.elements(i);
        if(Element_Or(particles.inactive,segment))continue;
        int a=segment(1);int c=segment(2);
        T R=(T).25*(particles.thickness(a)+particles.thickness(c));
        T tangent_boundary_coefficient=(T)2*((T)pi-(T)1)*R;
        TV f_tangent=tangent_boundary_coefficient*tri_surface_tension*(particles.X(c)-particles.X(a)).Normalized();
        for(int j=1;j<=2;j++){T sign=(j==1?(T)1:(T)-1);
            //bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(segment(j)));        
            F(segment(j))+=sign*f_tangent;}
    }


    ////segment mesh
    for(int i=1;i<=segment_mesh.elements.m;i++){
        const TV_INT2& segment=segment_mesh.elements(i);int a=segment(1);int c=segment(2);
        if(Element_Or(particles.inactive,segment))continue;

        T R=(T).25*(particles.thickness(a)+particles.thickness(c));
        T segment_coefficient=(T)two_pi*R;
        TV f_tangent=segment_coefficient*seg_surface_tension*(particles.X(c)-particles.X(a)).Normalized();

        for(int j=1;j<=2;j++){T sign=(j==1?(T)1:(T)-1);
            bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(segment(j)));
            if(apply_surface_tension)F(segment(j))+=sign*f_tangent;
        }
    }

    if(mpi_codim){ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);}
}

template<class T_input>
void PhysBAM::CODIMENSIONAL_SURFACE_TENSION_FORCE<VECTOR<T_input,3> >::Add_Velocity_Independent_Forces_Viscous(ARRAY_VIEW<TV> F,const T time)
{
    T tet_surface_tension=surface_tension;
    T tri_surface_tension=triangle_surface_tension>(T)0?triangle_surface_tension:surface_tension;
    T seg_surface_tension=segment_surface_tension>(T)0?segment_surface_tension:surface_tension;

    ////boundary of tetrahedron mesh
    for(int i=1;i<=(*tetrahedron_mesh.boundary_mesh).elements.m;i++){
        const TV_INT3& v=(*tetrahedron_mesh.boundary_mesh).elements(i);
        if(water_tank&&water_tank->Is_Tank_Contact_Element(v))continue;   ////triangle contacting the solid boundary
        if(Element_Or(particles.inactive,v))continue;

        for(int d=1;d<=3;d++){
            int ix=(d-1)%3+1;int iy=(d)%3+1;int iz=(d+1)%3+1;
            T surface_tension=tet_surface_tension;
            T coef=(T)1;if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(v(d)))coef=rim_coef_1;
            if(contact) F(v(d))+=surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(particles.X,particles.contact,v(ix),v(iy),v(iz)))*coef;
            else F(v(d))+=surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(particles.X,v(ix),v(iy),v(iz)))*coef;
        }
    }

    if(st_on_tris){
        ////triangle mesh
        //ARRAY<TV> thickness_F;thickness_F.Resize(F.m);thickness_F.Fill(TV());
        for(int i=1;i<=triangle_mesh.elements.m;i++){
            if(Element_Or(particles.inactive,triangle_mesh.elements(i)))continue;

            VECTOR<TV,3> p;VECTOR<TV,3> thickened_p;
            for(int d=1;d<=3;d++){
                int v=triangle_mesh.elements(i)(d);p(d)=particles.X(v);
                //thickened_p(d)=p(d)+particles.area_weighted_normal(v)*(T).5*particles.thickness(v);
            }
            for(int d=1;d<=3;d++){
                int v=triangle_mesh.elements(i)(d);
                int i1=(d-1)%3+1;int i2=(d)%3+1;int i3=(d+1)%3+1;
                bool c1=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i1)));
                bool c2=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i2)));
                bool c3=Is_Contact_With_Solid(particles.contact(triangle_mesh.elements(i)(i3)));
                //F(v)+=(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(p(ix),p(iy),p(iz)));
                //thickness_F(v)+=(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(thickened_p(ix),thickened_p(iy),thickened_p(iz)));

                ////for viscous, weaken/strengthen the retraction force on triangle mesh boundary
                T surface_tension=tri_surface_tension;
                T coef=(T)1;if(mesh.Is_Boundary_Particle_On_Triangle_Mesh(v)&&!Is_On_Segment_Mesh(mesh.node_types(v)))coef=rim_coef_1;

                if(Is_On_Tetrahedron_Mesh(mesh.node_types(v)))surface_tension=tet_surface_tension;

                bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(v));
                if(apply_surface_tension){
                    if(contact) F(v)+=coef*(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(c1,c2,c3,p(i1),p(i2),p(i3)));
                    else F(v)+=coef*(T)2*surface_tension*(-CODIMENSIONAL_MESH_OBJECT<TV>::Length_Weighted_Triangle_Vertex_Normal(p(i1),p(i2),p(i3)));
                }
            }
        }

        ////rim surface tension, boundary of triangle mesh
        for(int i=1;i<=mesh.rim_segment_mesh.elements.m;i++){
            const TV_INT2& segment=mesh.rim_segment_mesh.elements(i);
            if(Element_Or(particles.inactive,segment))continue;
            int a=segment(1);int c=segment(2);

            for(int j=1;j<=2;j++){T sign=(j==1?(T)1:(T)-1);
            T R=particles.thickness(segment(j))*(T).5;
            T coef=rim_coef_2;
            T tangent_boundary_coefficient=coef*(T)2*((T)pi-(T)1)*R;
            TV f_tangent=tangent_boundary_coefficient*tri_surface_tension*(particles.X(c)-particles.X(a)).Normalized();

            //bool apply_surface_tension=!viscous||(!Is_On_Segment_Mesh(mesh.node_types(segment(j)))&&!Is_On_Tetrahedron_Mesh(mesh.node_types(segment(j))));
            bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(segment(j)));
            if(apply_surface_tension){F(segment(j))+=sign*f_tangent;}
            }
        }

        ////TODO: fix the direction of surface tension from triangles or tetrahedra
        ////rescale the thin tet surface tension for viscous fluid
        if(viscous){HASHTABLE<int> tet_seg_vertices;
            for(int i=1;i<=segment_mesh.elements.m;i++){
                const TV_INT2& segment=segment_mesh.elements(i);
                for(int j=1;j<=2;j++){int v=segment(j);
                    if(Is_On_Tetrahedron_Mesh(mesh.node_types(v))&&!tet_seg_vertices.Contains(v)){
                        T R=thickness_scale*(T).5;T segment_coefficient=(T)two_pi*R;
                        TV f_direction=-particles.normal(v);
                        F(v)=segment_coefficient*seg_surface_tension*f_direction;tet_seg_vertices.Set(v);
                    }
                }
            }
        }
    }

    if(st_on_segs){
        ////segment mesh
        for(int i=1;i<=segment_mesh.elements.m;i++){
            const TV_INT2& segment=segment_mesh.elements(i);int a=segment(1);int c=segment(2);
            if(Element_Or(particles.inactive,segment))continue;

            for(int j=1;j<=2;j++){
                ////for viscous, turn off surface tension on endpoint of segment mesh
                bool apply_surface_tension=!Is_Contact_With_Solid(particles.contact(segment(j)));

                if(apply_surface_tension){
                    T R=(T).5*particles.thickness(segment(j));  ////for viscous, weaken flow along filaments
                    T segment_coefficient=(T)two_pi*R;
                    TV f_tangent=segment_coefficient*seg_surface_tension*(particles.X(c)-particles.X(a)).Normalized();
                    T sign=(j==1?(T)1:(T)-1);
                    F(segment(j))+=sign*f_tangent;
                }
            }
        }
    }

    if(mpi_codim){ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);}
}

//template class CODIMENSIONAL_SURFACE_TENSION_FORCE_2D<float>;
template class CODIMENSIONAL_SURFACE_TENSION_FORCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_SURFACE_TENSION_FORCE_2D<double>;
template class CODIMENSIONAL_SURFACE_TENSION_FORCE<VECTOR<double,3> >;
#endif

