//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_SURFACE_TENSION_FORCE__
#define __CODIMENSIONAL_SURFACE_TENSION_FORCE__
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Misc/WATER_TANK.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_SURFACE_TENSION_FORCE_BASE
{
    typedef typename TV::SCALAR T;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    T surface_tension;
    T triangle_surface_tension;
    T segment_surface_tension;
    T length_scale;
    T thickness_scale;
    CODIMENSIONAL_MESH& mesh;
    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;
    WATER_TANK<TV>* water_tank;
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_SURFACE_TENSION_FORCE_BASE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,T surface_tension_input,T thickness_scale_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0)
        :codimensional_mesh_object(codimensional_mesh_object_input),particles(codimensional_mesh_object.particles),surface_tension(surface_tension_input),triangle_surface_tension((T)-1),segment_surface_tension((T)-1),
        thickness_scale(thickness_scale_input),mesh(codimensional_mesh_object.mesh),segment_mesh(mesh.segment_mesh),triangle_mesh(mesh.triangle_mesh),tetrahedron_mesh(mesh.tetrahedron_mesh),
        water_tank(0),mpi_codim(mpi_codim_input){}

    virtual void Update_Position_Based_State(const T time,const bool is_position_update){PHYSBAM_NOT_IMPLEMENTED();}
    virtual void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV> F,const T time){PHYSBAM_NOT_IMPLEMENTED();}

    void Initialize_Tank(const BOX<TV>& tank_box_input,const T tank_thickness_input,bool is_open_tank_input=false)
    {if(water_tank)delete water_tank;water_tank=new WATER_TANK<TV>(tank_box_input,tank_thickness_input,is_open_tank_input,&codimensional_mesh_object);}

    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input){mpi_codim=mpi_codim_input;}
protected:
};

template<class TV> class CODIMENSIONAL_SURFACE_TENSION_FORCE;
template<class T_input>
class CODIMENSIONAL_SURFACE_TENSION_FORCE<VECTOR<T_input,3> >: public CODIMENSIONAL_SURFACE_TENSION_FORCE_BASE<VECTOR<T_input,3> >
{
    typedef T_input T;typedef VECTOR<T,3> TV;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef CODIMENSIONAL_SURFACE_TENSION_FORCE_BASE<TV> BASE;
public:
    using BASE::mesh;
    using BASE::codimensional_mesh_object;using BASE::particles;
    using BASE::surface_tension;using BASE::triangle_surface_tension;using BASE::segment_surface_tension;
    using BASE::thickness_scale;
    using BASE::segment_mesh;using BASE::triangle_mesh;using BASE::tetrahedron_mesh;
    using BASE::water_tank;using BASE::mpi_codim;
    
    bool viscous;
    bool contact;
    bool st_on_tris;
    bool st_on_segs;
    T rim_coef_1;
    T rim_coef_2;

    CODIMENSIONAL_SURFACE_TENSION_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,T surface_tension_input,T thickness_scale_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0)
        :BASE(codimensional_mesh_object_input,surface_tension_input,thickness_scale_input,mpi_codim_input),viscous(true),contact(true),st_on_tris(true),st_on_segs(true),rim_coef_1((T)1),rim_coef_2((T)1){}

    virtual void Update_Position_Based_State(const T time,const bool is_position_update){}
    virtual void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV> F,const T time);
    virtual void Add_Velocity_Independent_Forces_Viscous(ARRAY_VIEW<TV> F,const T time);
};
}
#endif
