//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Solver/CODIMENSIONAL_VISCOELASTIC_FORCE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::CODIMENSIONAL_VISCOELASTIC_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
:BASE(codimensional_mesh_object_input,mpi_codim_input),youngs_modulus((T)1),plastic_clamp_min((T).6),plastic_clamp_max((T)1.4),
has_plasticity(false),relaxation((T)1),use_green_strain(true),use_semi_implicit(true)
{Use_Constant_Extrapolation(true);}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Position_Based_State(const T time,const bool is_position_update)
{BASE::Update_Position_Based_State(time,is_position_update);}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    LOG::cout<<"explicit elastic force"<<std::endl;

    //Reset_Solver();
    //Update_G(psi_Ns_on_mesh(1),psi_Ds_on_mesh(1));
    //solver.Update_V_inv(solver.M_inv);

    //LOG::cout<<"elas psi_N: "<<psi_Ns_on_mesh(1).m<<std::endl;
    //for(int i=1;i<=psi_Ns_on_mesh(1).m;i++){
    //    int p=psi_Ns_on_mesh(1)(i).index;
    //    int guid=particles.guid(p);
    //    LOG::cout<<i<<": "<<guid<<", "<<psi_Ns_on_mesh(1)(i).direction<<", "<<psi_Ns_on_mesh(1)(i).v<<std::endl;
    //}

    //Update_Deformation_State(dt,time);

    ////linear stress

    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> strain_component(solver.velocity_samples.m*TV::m);
        for(int i=1;i<=solver.velocity_samples.m;i++){
            int p=solver.velocity_samples(i);
            //const T_MATRIX& strain=particles.strain(p);
            //for(int k=1;k<=TV::m;k++)strain_component((i-1)*TV::m+k)=strain(d,k);

            T trace=particles.strain(p).Trace();
            T_MATRIX strain_deviatoric=particles.strain(p)-(T)one_third*trace*T_MATRIX::Identity_Matrix();
            for(int k=1;k<=TV::m;k++)strain_component((i-1)*TV::m+k)=strain_deviatoric(d,k);
        }

        VECTOR_ND<T> vol_weighted_div_strain(solver.pressure_samples.m);
        solver.GT.Times(strain_component,vol_weighted_div_strain);vol_weighted_div_strain*=(T)-1;
        ARRAY<T> vol_weighted_div_strain_on_particles(particles.X.m);vol_weighted_div_strain_on_particles.Fill((T)0);
        ARRAY_VIEW<T> vw_div_strain_on_particles_arrayview(vol_weighted_div_strain_on_particles);
        Interpolate_Pressure_Sample_To_Particle_Attribute(vol_weighted_div_strain,vw_div_strain_on_particles_arrayview);

        for(int i=1;i<=solver.pressure_samples.m;i++){
            int p=solver.pressure_samples(i).index;
            F(p)(d)+=youngs_modulus*vol_weighted_div_strain_on_particles(p);
        }
    }
    
    if(mpi_codim){ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    if(use_semi_implicit) Semi_Implicit_Update(V,F,dt,time);
    else Fully_Implicit_Update(V,F,dt,time);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Semi_Implicit_Update(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    LOG::cout<<"implicit elasticity: "<<youngs_modulus<<" psi_N: "<<psi_Ns_on_mesh(1).m<<", psi_D: "<<psi_Ds_on_mesh(1).m<<", rlx: "<<relaxation<<", plas: "<<has_plasticity<<
        ", thk: "<<use_particle_thickness<<", den: "<<density<<", staggered dof: "<<use_staggered_dof<<std::endl;

    //////explicit
    //TV total_f=TV();
    //for(int d=1;d<=TV::m;d++){
    //    VECTOR_ND<T> strain_component(solver.velocity_samples.m*TV::m);
    //    for(int i=1;i<=solver.velocity_samples.m;i++){
    //        int p=solver.velocity_samples(i);
    //        //const T_MATRIX& strain=particles.strain(p);
    //        //for(int k=1;k<=TV::m;k++)strain_component((i-1)*TV::m+k)=strain(d,k);

    //        T trace=particles.strain(p).Trace();
    //        T_MATRIX strain_deviatoric=particles.strain(p)-(T)one_third*trace*T_MATRIX::Identity_Matrix();
    //        for(int k=1;k<=TV::m;k++)strain_component((i-1)*TV::m+k)=strain_deviatoric(d,k);
    //    }

    //    VECTOR_ND<T> vol_weighted_div_strain(solver.pressure_samples.m);
    //    solver.GT.Times(strain_component,vol_weighted_div_strain);vol_weighted_div_strain*=(T)-1;
    //    //if(d==1){LOG::cout<<"vol_w_div_strain: "<<std::endl;for(int i=1;i<=solver.pressure_samples.m;i++){LOG::cout<<vol_weighted_div_strain(i)<<", ";}LOG::cout<<std::endl;}

    //    ARRAY<T> vol_weighted_div_strain_on_particles(particles.X.m);vol_weighted_div_strain_on_particles.Fill((T)0);
    //    ARRAY_VIEW<T> vw_div_strain_on_particles_arrayview(vol_weighted_div_strain_on_particles);
    //    Interpolate_Pressure_Sample_To_Particle_Attribute(vol_weighted_div_strain,vw_div_strain_on_particles_arrayview);
    //    //if(d==1){LOG::cout<<"type: "<<std::endl;for(int i=1;i<=solver.pressure_samples.m;i++){LOG::cout<<solver.pressure_samples(i).type<<", ";}LOG::cout<<std::endl;}
    //    //if(d==1){LOG::cout<<"arrayview: "<<std::endl;for(int i=1;i<=solver.pressure_samples.m;i++){LOG::cout<<vol_weighted_div_strain_on_particles(i)<<", ";}LOG::cout<<std::endl;}
    //    for(int i=1;i<=solver.pressure_samples.m;i++){
    //        int p=solver.pressure_samples(i).index;
    //        F(p)(d)+=youngs_modulus*vol_weighted_div_strain_on_particles(p);
    //        total_f(d)+=F(p)(d);
    //    }
    //    for(int i=1;i<=particles.X.m;i++){particles.V(i)(d)+=F(i)(d)/particles.mass(i)*dt;}
    //}
    //LOG::cout<<"total_f: "<<total_f<<std::endl;

    //Update_Deformation_State(dt,time);

    //Reset_Solver();
    //Update_G(psi_Ns_on_mesh(1),psi_Ds_on_mesh(1));
    //ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    //Update_G(psi_Ns_on_surface,empty_psi_Ds);
    //solver.Update_V_inv(solver.M_inv);
    solver.Update_FM_inv(particles.elastic_deformation_gradient,solver.M_inv,solver.FM_inv);
    //solver.Update_FM_inv(particles.elastic_deformation_gradient,solver.M_inv,solver.asym_FM_inv,/*use asymmetric part of F*/false);
    //solver.Update_GT(solver.G,solver.GT);
    solver.Update_A(solver.G,solver.FM_inv,solver.GT,solver.A);

    ////initialize volume
    VECTOR_ND<T> vol(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        //vol(i)=codimensional_mesh_object.Element_Volume(p,type);
        vol(i)=(use_particle_thickness?codimensional_mesh_object.Element_Volume(p,type):codimensional_mesh_object.Element_Volume_With_Thickness(p,type,thickness_scale));
        //vol(i)=particles.mass(p)/density;   ////dbg
    }

    Ensure_Matrix_Symmetry(solver.A);   ////fix the numerical errors in GT*FM_inv*G
    SPARSE_MATRIX_FLAT_NXN<T> A_copy=solver.A;

    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> strain_vec(solver.velocity_samples.m*TV::m);
        VECTOR_ND<T> u0(solver.pressure_samples.m);
        VECTOR_ND<T> GTe(solver.pressure_samples.m);
        VECTOR_ND<T> u1(solver.pressure_samples.m);
        VECTOR_ND<T> D(solver.pressure_samples.m);D.Fill((T)0);

        ////initialize strain component
        for(int i=1;i<=solver.velocity_samples.m;i++){
            int p=solver.velocity_samples(i);
            //const T_MATRIX& strain=particles.strain(p);
            //for(int k=1;k<=TV::m;k++)strain_vec((i-1)*TV::m+k)=strain(d,k);
            T trace=particles.strain(p).Trace();
            T_MATRIX strain_deviatoric=particles.strain(p)-(T)one_third*trace*T_MATRIX::Identity_Matrix();
            for(int k=1;k<=TV::m;k++)strain_vec((i-1)*TV::m+k)=strain_deviatoric(d,k);
        }

        ////initialize u0
        Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u0);

        ////initialize neg_div_strain_u
        solver.GT.Times(strain_vec,GTe);

        SPARSE_MATRIX_FLAT_NXN<T>& B=solver.A;B=A_copy;
        VECTOR_ND<T>& b=solver.b;b.Resize(solver.pressure_samples.m);
        solver.p.Resize(b.n);solver.p.Fill((T)0);
        B*=(relaxation*dt);
        for(int i=1;i<=solver.pressure_samples.m;i++){D(i)=(T)2*density*vol(i)/(dt*youngs_modulus);} 
        for(int i=1;i<=solver.pressure_samples.m;i++){B(i,i)+=D(i);}

        ////initialize b=Bu-GTe
        if(use_green_strain){B.Times(u0,b);}else b=u0*D;
        //for(int i=1;i<=solver.pressure_samples.m;i++){b(i)=u0(i)*D(i);} ////dbg
        for(int i=1;i<=solver.pressure_samples.m;i++){b(i)-=(T)2*GTe(i);}

        //if(d==1){LOG::cout<<"GTe: "<<std::endl;for(int i=1;i<=solver.pressure_samples.m;i++){LOG::cout<<GTe(i)<<", ";}LOG::cout<<std::endl;}

        //solver.Modify_b_With_Dirichlet_Boundary_Conditions(b,solver.psi_D);

        solver.Solve();
        Update_Velocity_And_Force_Component_FLIP(V,F,u0,solver.p,d,dt);

        //////dbg
        //for(int i=1;i<=solver.pressure_samples.m;i++){solver.p(i)=b(i)/D(i);}
        //for(int i=1;i<=solver.pressure_samples.m;i++){V(i)(d)=solver.p(i);}
    }

    if(mpi_codim){
        ARRAY_VIEW<TV> V_array_view(V);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(V_array_view);
        ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Fully_Implicit_Update(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    LOG::cout<<"start fully implicit solve"<<std::endl;
    TIMER timer;int id=5;timer.Reset_Time(id);
    solver.Update_Implicit_Elasticity_Matrices(particles.elastic_deformation_gradient);
    double t1=timer.Peek_And_Reset_Time(id);

    SPARSE_MATRIX_FLAT_NXN<T> B=solver.Ah;
    VECTOR_ND<T> V_vec(solver.pressure_samples.m*TV::m);
    VECTOR<VECTOR_ND<T>, TV::m> u0;
    for(int d=1;d<=TV::m;d++){
        u0(d).Resize(solver.pressure_samples.m);
        Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u0(d));
        for(int i=1;i<=u0(d).n;i++){int idx=(d-1)*u0(d).n+i;V_vec(idx)=u0(d)(i);}
    }
    
    VECTOR_ND<T> vol(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;int type=solver.pressure_samples(i).type;//vol(i)=particles.mass(p)/density;
        vol(i)=codimensional_mesh_object.Element_Volume(p,type);
    }
    VECTOR_ND<T> alpha;alpha.Resize(vol.n*TV::m);
    for(int i=1;i<=vol.n;i++){
        T a=(T)2*density*vol(i)/(dt*dt*youngs_modulus);
        for(int d=1;d<=TV::m;d++){int idx=(d-1)*vol.n+i;alpha(idx)=a;}
    }
    for(int i=1;i<=B.n;i++)B(i,i)+=alpha(i);

    VECTOR_ND<T> b;b.Resize(V_vec.n);b=alpha*V_vec;
    VECTOR_ND<T> strain_vec;Vectorize(particles.strain,strain_vec);
    VECTOR_ND<T> GTe(solver.pressure_samples.m*TV::m);solver.GhT.Times(strain_vec,GTe);
    GTe*=((T)2/dt);b-=GTe;
    VECTOR_ND<T> p_vec(solver.pressure_samples.m*TV::m);
    //LOG::cout<<"v:\n"<<V_vec<<std::endl;
    //LOG::cout<<"GTe:\n"<<GTe<<std::endl;
    //LOG::cout<<"b:\n"<<b<<std::endl;
    //solver.Solve_Linear_System(B,p_vec,b);
    //LOG::cout<<"p:\n"<<p_vec<<std::endl;

    //////dbg, check matrices
    ////for(int i=1;i<=solver.Mh_inv.m;i++)solver.Mh_inv(i,i)=(T)1e1;
    //SPARSE_MATRIX_FLAT_MXN<T> M1;
    //VECTOR_ND<T> u1(solver.pressure_samples.m*TV::m);u1.Fill((T)1);
    //M1=/*solver.Mh_inv**/solver.I_plus_P*solver.Gh;
    //VECTOR_ND<T> v1(M1.m);
    //M1.Times(u1,v1);
    ////LOG::cout<<"v1:\n"<<v1<<std::endl;

    //ARRAY<T_MATRIX> grad_uT(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());
    //for(int d=1;d<=TV::m;d++){
    //    VECTOR_ND<T> u(solver.pressure_samples.m);u.Fill((T)1);//Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);
    //    
    //    VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
    //    solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
    //    //grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u

    //    for(int i=1;i<=solver.velocity_samples.m;i++){
    //        TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
    //        grad_uT(i).Column(d)=grad_u_i;
    //    }
    //}
    //for(int i=1;i<=solver.velocity_samples.m;i++){
    //    grad_uT(i)=grad_uT(i)+grad_uT(i).Transposed();
    //}
    //VECTOR_ND<T> v2;ARRAY_VIEW<T_MATRIX> grad_uT_arrayview=ARRAY_VIEW<T_MATRIX>(grad_uT);
    //Vectorize(grad_uT_arrayview,v2);
    ////LOG::cout<<"v2:\n"<<v2<<std::endl;
    ////LOG::cout<<"dif:\n"<<v2-v1<<std::endl;

    //////explicit, dbg
    //B*=(T)0;
    //for(int i=1;i<=B.n;i++)B(i,i)+=alpha(i);
    //VECTOR_ND<T> Au(V_vec.n);solver.Ah.Times(V_vec,Au);b-=Au;
    //solver.Solve_Linear_System(B,p_vec,b);
    ////LOG::cout<<"B:\n"<<B<<std::endl;
    ////LOG::cout<<"b:\n"<<b<<std::endl;
    ////LOG::cout<<"p:\n"<<p_vec<<std::endl;
    //for(int i=1;i<=B.A.m;i++){
        //if(!Is_Valid_Number(B.A(i).a)){
            //LOG::cout<<"invalid "<<B.A(i).j<<", "<<B.A(i).a<<std::endl;
        //}
    //}
    //for(int i=1;i<=B.n;i++){
        //if(B(i,i)==(T)0)LOG::cout<<"zero diag "<<i<<": "<<B(i,i)<<std::endl;
        //LOG::cout<<B(i,i)<<", ";
    //}LOG::cout<<std::endl;

    double t2=timer.Peek_And_Reset_Time(id);

    T absolute_tolerance=(T)1e-30;T relative_tolerance=(T)1e-10;
    T b_max_abs=b.Max_Abs();////TODO: mpi
    T tolerance=max(relative_tolerance*b_max_abs,absolute_tolerance);
    int max_iterations=2000;
    VECTOR_ND<T> &x=p_vec;VECTOR_ND<T> last_x(x);
    for(int k=1;k<=max_iterations;k++){
        B.Gauss_Seidel_Single_Iteration(x,b);
        T residual=0;for(int j=1;j<=x.n;j++){residual+=sqr(last_x(j)-x(j));last_x(j)=x(j);}
        LOG::cout<<"res "<<k<<": "<<residual<<", ";
        if(residual<tolerance){LOG::cout<<"Gauss-Seidel iterations = "<<k<<std::endl;break;}
        if(k==max_iterations)LOG::cout<<"G-S iterations = "<<residual<<": "<<tolerance<<std::endl;
    }
   
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u1(solver.pressure_samples.m);
        for(int i=1;i<=solver.pressure_samples.m;i++){
            int idx=(d-1)*solver.pressure_samples.m+i;u1(i)=p_vec(idx);
        }
        Update_Velocity_And_Force_Component_FLIP(V,F,u0(d),u1,d,dt);
    }
    double t3=timer.Peek_And_Reset_Time(id);
    LOG::cout<<"imp time: "<<t1<<", "<<t2<<", "<<t3<<std::endl;

    if(mpi_codim){
        ARRAY_VIEW<TV> V_array_view(V);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(V_array_view);
        ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Deformation_State(const T dt,const T time)
{
    if(has_plasticity){
        Update_Deformation_Gradients(particles.V,particles.elastic_deformation_gradient,particles.plastic_deformation_gradient,dt,time);
        Update_Strains(particles.elastic_deformation_gradient,particles.strain);
        //Update_Stress(particles.strain,particles.stress);
    }
    else{
        Update_Deformation_Gradients(particles.V,particles.elastic_deformation_gradient,dt,time);
        Update_Strains(particles.elastic_deformation_gradient,particles.strain);
        //Update_Stress(particles.strain,particles.stress);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Deformation_Gradients(ARRAY_VIEW<TV>& V,ARRAY_VIEW<T_MATRIX>& FT,const T dt,const T time)
{
    //////constant extrapolation
    //bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    //solver.use_constant_extrapolation=true;
    //Reset_Solver();
    //ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > empty_psi_Ns;
    //ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    //Update_G(empty_psi_Ns,empty_psi_Ds);
    //solver.Update_V_inv(solver.M_inv);

    ////no constant extrapolation
    bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    solver.use_constant_extrapolation=false;
    Reset_Solver();
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    Update_G(psi_Ns_on_surface,empty_psi_Ds);
    solver.Update_V_inv(solver.M_inv);

    ////update grad_uT
    grad_uT.Resize(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u;Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);
        //u.Fill((T)1);       ////dbg
        //LOG::cout<<"V "<<particles.V<<std::endl;
        //LOG::cout<<"u "<<u<<std::endl;

        VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
        solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
        grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u

        for(int i=1;i<=solver.velocity_samples.m;i++){
            TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
            grad_uT(i).Column(d)=grad_u_i;
        }
    }

    INDIRECT_ARRAY<ARRAY_VIEW<T_MATRIX> > FT_indir_array(FT,solver.velocity_samples);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        FT_indir_array(i)+=FT_indir_array(i)*grad_uT(i)*dt;
        T_MATRIX Q;T_SYMMETRIC_MATRIX S;
        FT_indir_array(i).Fast_Indefinite_Polar_Decomposition(Q,S);
        FT_indir_array(i)=S;
    }

    for(int i=1;i<=solver.velocity_samples.m;i++){
        //particles.strain(i)+=grad_uT(i).Symmetric_Part()*dt;
        int idx=solver.velocity_samples(i);
        particles.elastic_deformation_gradient(idx)=FT_indir_array(i);
        particles.gradient_V(idx)=grad_uT(i);
    }

    solver.use_constant_extrapolation=use_constant_extrapolation_copy;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Deformation_State_Part_One(const T dt,const T time)
{
    bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    solver.use_constant_extrapolation=false;
    Reset_Solver();
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    Update_G(psi_Ns_on_surface,empty_psi_Ds);
    solver.Update_V_inv(solver.M_inv);
    solver.use_constant_extrapolation=use_constant_extrapolation_copy;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Deformation_State_Part_Two(const T dt,const T time)
{
    ////update gradu^T_{n+1}
    grad_uT.Resize(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u;Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);

        VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
        solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
        grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u

        for(int i=1;i<=solver.velocity_samples.m;i++){
            TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
            grad_uT(i).Column(d)=grad_u_i;
        }
    }

    ////epsilon_{n+1}=dt*FT_n*.5*(gradu_{n+1}^T+gradu_{n+1})*F_n+epsilon_n
    ////F_{n+1}=(I+dt*grad_u)*F_n
    ARRAY_VIEW<T_MATRIX>& FT=particles.elastic_deformation_gradient;
    ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain=particles.strain;
    INDIRECT_ARRAY<ARRAY_VIEW<T_MATRIX> > FT_indir_array(FT,solver.velocity_samples);
    INDIRECT_ARRAY<ARRAY_VIEW<T_SYMMETRIC_MATRIX> > strain_indir_array(strain,solver.velocity_samples);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        T_MATRIX deformation_rate=(T).5*(grad_uT(i)+grad_uT(i).Transposed());
        T_SYMMETRIC_MATRIX delta_strain=T_SYMMETRIC_MATRIX::Multiply_With_Symmetric_Result(FT(i)*deformation_rate,FT(i).Transposed());
        strain_indir_array(i)+=delta_strain*dt;
        FT_indir_array(i)+=FT_indir_array(i)*grad_uT(i)*dt;
    }

    ////particle attributes
    for(int i=1;i<=solver.velocity_samples.m;i++){
        int idx=solver.velocity_samples(i);
        particles.elastic_deformation_gradient(idx)=FT_indir_array(i);
        particles.gradient_V(idx)=grad_uT(i);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Deformation_Gradients(ARRAY_VIEW<TV>& V,ARRAY_VIEW<T_MATRIX>& FT_e,ARRAY_VIEW<T_MATRIX>& FT_p,const T dt,const T time)
{
    INDIRECT_ARRAY<ARRAY_VIEW<T_MATRIX> > FT_e_indir_array(FT_e,solver.velocity_samples);
    INDIRECT_ARRAY<ARRAY_VIEW<T_MATRIX> > FT_p_indir_array(FT_p,solver.velocity_samples);
    Update_Deformation_Gradients(V,FT_e,dt,time);
    //plastic_clamp_min=(T).8;plastic_clamp_max=(T)1.2;
    for(int i=1;i<=solver.velocity_samples.m;i++){
        T_MATRIX U,V;T_DIAGONAL_MATRIX D;
        T_MATRIX& FT_ei=FT_e_indir_array(i);
        T_MATRIX& FT_pi=FT_p_indir_array(i);
        T_MATRIX FT_i=FT_ei*FT_pi;  ////F_n+1
        FT_ei.Fast_Singular_Value_Decomposition(U,D,V);
        for(int j=1;j<=TV::m;j++){/*clamp D*/
            if(D(j)<plastic_clamp_min)D(j)=plastic_clamp_min;
            else if(D(j)>plastic_clamp_max){/*LOG::cout<<"plastic clamp "<<D(j)<<std::endl;*/D(j)=plastic_clamp_max;}
        }
        FT_ei=U*D*V.Transposed(); ////F_e+1
        FT_pi=V*D.Inverse()*U.Transposed()*FT_i; ////F_p+1
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain)
{
    if(use_green_strain)Update_Green_Strains(FT,strain);
    else Update_Cauchy_Strains(FT,strain);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Cauchy_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain)
{
    for(int i=1;i<=strain.m;i++){
        const T_MATRIX& FT_i=FT(i);
        strain(i)=(T).5*(FT_i.Symmetric_Part()-T_SYMMETRIC_MATRIX::Identity_Matrix());
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Update_Green_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain)
{
    for(int i=1;i<=strain.m;i++){
        const T_MATRIX& FT_i=FT(i);
        strain(i)=(T).5*(T_SYMMETRIC_MATRIX::Times_Transpose_With_Symmetric_Result(FT_i,FT_i)-T_SYMMETRIC_MATRIX::Identity_Matrix());
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Initialize_Deformation_State()
{
    for(int i=1;i<=particles.X.m;i++){
        particles.deformation_gradient(i)=T_MATRIX::Identity_Matrix();
        particles.elastic_deformation_gradient(i)=particles.deformation_gradient(i);
        particles.plastic_deformation_gradient(i)=T_MATRIX::Identity_Matrix();
        particles.strain(i).Set_Zero_Matrix();
        particles.stress(i).Set_Zero_Matrix();
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Vectorize(const ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain,/*rst*/VECTOR_ND<T>& strain_vec)
{
    strain_vec.Resize(strain.m*TV::m*TV::m);
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=strain.m;i++){
            int r0=(d-1)*strain.m*TV::m+(i-1)*TV::m;
            for(int j=1;j<=TV::m;j++){T value=strain(i)(d,j);int idx=r0+j;strain_vec(idx)=value;}
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Vectorize(const ARRAY_VIEW<T_MATRIX>& F,/*rst*/VECTOR_ND<T>& F_vec)
{
    F_vec.Resize(F.m*TV::m*TV::m);
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=F.m;i++){
            int r0=(d-1)*F.m*TV::m+(i-1)*TV::m;
            for(int j=1;j<=TV::m;j++){T value=F(i)(d,j);int idx=r0+j;F_vec(idx)=value;}
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Vectorize(const ARRAY_VIEW<TV>& V,/*rst*/VECTOR_ND<T>& V_vec)
{
    V_vec.Resize(V.m*TV::m);
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=V.m;i++){int idx=(d-1)*V.m+i;V_vec(idx)=V(i)(d);}
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Unvectorize(const VECTOR_ND<T>& strain_vec,/*rst*/ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain)
{
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=strain.m;i++){
            int r0=(d-1)*strain.m*TV::m+(i-1)*TV::m;
            for(int j=1;j<=TV::m;j++){int idx=r0+j;strain(i)(d,j)=strain_vec(idx);}
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_FORCE<TV>::Unvectorize(const VECTOR_ND<T>& V_vec,ARRAY_VIEW<TV>& V)
{
    for(int d=1;d<=TV::m;d++){
        for(int i=1;i<=V.m;i++){int idx=(d-1)*V.m+i;V(i)(d)=V_vec(idx);}
    }
}

//template class CODIMENSIONAL_VISCOELASTIC_FORCE<VECTOR<float,2> >;
template class CODIMENSIONAL_VISCOELASTIC_FORCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_VISCOELASTIC_FORCE<VECTOR<double,2> >;
template class CODIMENSIONAL_VISCOELASTIC_FORCE<VECTOR<double,3> >; 
#endif
