//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_VISCOELASTIC_FORCE__
#define __CODIMENSIONAL_VISCOELASTIC_FORCE__
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include <PhysBAM_Tools/Matrices/MATRIX.h>
#include <PhysBAM_Tools/Matrices/SYMMETRIC_MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_NXN.h>
#include <PhysBAM_Tools/Matrices/MATRIX_MXN.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_VISCOELASTIC_FORCE : public CODIMENSIONAL_VISCOSITY_FORCE<TV>
{
    typedef CODIMENSIONAL_VISCOSITY_FORCE<TV> BASE;
    typedef typename TV::SCALAR T;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef MATRIX<T,TV::m> T_MATRIX;typedef DIAGONAL_MATRIX<T,TV::m> T_DIAGONAL_MATRIX;typedef SYMMETRIC_MATRIX<T,TV::m> T_SYMMETRIC_MATRIX;
public:
    using BASE::codimensional_mesh_object;using BASE::particles;using BASE::segment_mesh;using BASE::triangle_mesh;using BASE::tetrahedron_mesh;
    using BASE::kinematic_viscosity;using BASE::density;using BASE::thickness_scale;using BASE::psi_Ds_on_mesh;using BASE::psi_Ns_on_mesh;using BASE::element_index_in_pressure_samples;using BASE::particle_index_in_velocity_samples;
    using BASE::solver;using BASE::use_constant_extrapolation;using BASE::use_staggered_dof;using BASE::mpi_codim;
    using BASE::use_particle_thickness;using BASE::psi_Ns_on_surface;using BASE::grad_uT;
    using BASE::Add_Velocity_Independent_Forces;using BASE::Use_Constant_Extrapolation;using BASE::Reset_Solver;using BASE::Update_G;
    T youngs_modulus;
    T plastic_clamp_min;
    T plastic_clamp_max;
    bool has_plasticity;
    T relaxation;
    bool use_green_strain;
    bool use_semi_implicit;

    CODIMENSIONAL_VISCOELASTIC_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    virtual void Update_Position_Based_State(const T time,const bool is_position_update);
    virtual void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T dt,const T time);
    virtual void Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);

    ////all deformation gradients here are deformation gradient transpose
    void Initialize_Deformation_State();
    void Update_Deformation_State(const T dt,const T time);
    void Update_Deformation_Gradients(ARRAY_VIEW<TV>& V,ARRAY_VIEW<T_MATRIX>& FT,const T dt,const T time);
    void Update_Deformation_Gradients(ARRAY_VIEW<TV>& V,ARRAY_VIEW<T_MATRIX>& FT_e,ARRAY_VIEW<T_MATRIX>& FT_p,const T dt,const T time);
    void Update_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain);
    void Update_Cauchy_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain);
    void Update_Green_Strains(ARRAY_VIEW<T_MATRIX>& FT,ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain);

    void Semi_Implicit_Update(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);
    void Fully_Implicit_Update(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);

    void Update_Deformation_State_Part_One(const T dt,const T time);    ////update matrices
    void Update_Deformation_State_Part_Two(const T dt,const T time);    ////update deformation gradient and strain

    ////helper functions
    void Vectorize(const ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain,/*rst*/VECTOR_ND<T>& strain_vec);
    void Vectorize(const ARRAY_VIEW<T_MATRIX>& F,/*rst*/VECTOR_ND<T>& F_vec);
    void Vectorize(const ARRAY_VIEW<TV>& V,/*rst*/VECTOR_ND<T>& V_vec);
    void Unvectorize(const VECTOR_ND<T>& strain_vec,/*rst*/ARRAY_VIEW<T_SYMMETRIC_MATRIX>& strain);
    void Unvectorize(const VECTOR_ND<T>& V_vec,ARRAY_VIEW<TV>& V);
};
}
#endif
