//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::CODIMENSIONAL_VISCOSITY_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :codimensional_mesh_object(codimensional_mesh_object_input),particles(codimensional_mesh_object.particles),
    segment_mesh(codimensional_mesh_object.mesh.segment_mesh),triangle_mesh(codimensional_mesh_object.mesh.triangle_mesh),tetrahedron_mesh(codimensional_mesh_object.mesh.tetrahedron_mesh),
    kinematic_viscosity((T)1e-6),density((T)1000.),thickness_scale((T)1),variable_viscosity_mode(SHEAR_THINNING),kinematic_viscosity_min(kinematic_viscosity),rigid_viscosity((T)10),shear_rate_max((T)20),
    use_constant_extrapolation(false),use_staggered_dof(true),use_variable_viscosity(false),mpi_codim(mpi_codim_input)
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);
        solver.use_target_div=false;
        solver.codimensional_mesh_object=&codimensional_mesh_object;
        solver.element_index_in_pressure_samples=&element_index_in_pressure_samples;
        solver.particle_index_in_velocity_samples=&particle_index_in_velocity_samples;
        solver.psi_D_on_mesh_hashtable=new HASHTABLE<int>();
        solver.thickness_scale=thickness_scale;
        solver.mpi_codim=mpi_codim;
    }
    Use_Constant_Extrapolation(use_constant_extrapolation);
    Use_Staggered_DoF(use_staggered_dof);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Position_Based_State(const T time,const bool is_position_updated)
{
    Update_G();
    Update_M_inv();
    Update_A();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_G()
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);
        solver.Clear();
        solver.use_particle_thickness=true;
        solver.use_all_pressure_sample_on_nodes=!use_staggered_dof;
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(solver,&psi_N,&psi_Ds(d));

        if(mpi_codim)mpi_codim->Update_Boundary_And_Ghost_Pressure_Indices(solver.pressure_samples,solver.codimensional_mesh_object->mesh,solver.use_all_pressure_sample_on_nodes,
            solver.boundary_pressure_indices,solver.ghost_pressure_indices,solver.domain_pressure_indices);

        solver.Update_G(solver.G);
        //Check_Matrix_Invalid_Index(solver.G);

        solver.Modify_G_With_Boundary_Conditions(solver.G,solver.psi_N);
        solver.Update_GT(solver.G,solver.GT);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Dirichlet_Boundary_Samples(const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver)
{
    solver.psi_D.Remove_All();
    for(int i=1;i<=psi_D.m;i++){
        int p=psi_D(i).index;int type=psi_D(i).type;T v=psi_D(i).v;
        int region=codimensional_mesh_object.mesh.Particle_Region_Index(p);if(region==0)continue;
        if(type==SIMPLEX_PARTICLE&&solver.Find_Velocity_Sample(p)!=0){  ////currently all psi_D are on particles
            solver.Add_Dirichlet_Boundary_Sample(p,type,v);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Neumann_Boundary_Samples(const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N,const HASHTABLE<int>& invalid_indices,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver)
{
    solver.psi_N.Remove_All();
    for(int i=1;i<=psi_N.m;i++){
        int p=psi_N(i).index;
        if(invalid_indices.Contains(p)){/*LOG::cout<<"skip "<<p<<std::endl;*/continue;}
        int region=codimensional_mesh_object.mesh.Particle_Region_Index(p);if(region==0)continue;
        TV direction=psi_N(i).direction;T v=psi_N(i).v;
        if(solver.Find_Velocity_Sample(p)!=0){////p is the velocity sample in the subsystem
            solver.Add_Neumann_Boundary_Sample(p,direction,v);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Neumann_Boundary_On_Free_Surface(ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N)
{
    for(int i=1;i<=particles.X.m;i++){
        int region=codimensional_mesh_object.mesh.Particle_Region_Index(i);if(region==0)continue;
        
        if(Is_On_Codimensional_Mesh_Boundary(codimensional_mesh_object.mesh.node_types(i))){
            if(Is_Joint<TV::m>(codimensional_mesh_object.mesh.node_types(i))){continue;}
            TV normal=codimensional_mesh_object.Nodal_Area_Weighted_Normal(i).Normalized();
            psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
        }

        //if(Is_On_Tetrahedron_Mesh_Boundary(codimensional_mesh_object.mesh.node_types(i))){
        //    TV normal=codimensional_mesh_object.Nodal_Area_Weighted_Tetrahedron_Normal(i).Normalized();
        //    psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
        //}
        //if(Is_On_Triangle_Mesh(codimensional_mesh_object.mesh.node_types(i))){
        //    TV normal=codimensional_mesh_object.Nodal_Length_Weighted_Triangle_Normal(i).Normalized();
        //    psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
        //}
        //if(Is_On_Segment_Mesh()){
        //    TV normal=codimensional_mesh_object.Nodal_Segment_Normal(i).Normalized();
        //    psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,normal,(T)0));
        //}
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_M_inv()
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);

        solver.Update_V_inv(solver.M_inv);
        if(use_variable_viscosity){
            ARRAY<T> variable_kinematic_viscosity;
            Update_Variable_Viscosity(variable_kinematic_viscosity);
            for(int i=1;i<=solver.velocity_samples.m;i++){
                for(int d=1;d<=TV::m;d++)solver.M_inv((TV::m*(i-1))+d)*=variable_kinematic_viscosity(i);
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Interpolate_Velocity_Component_To_Pressure_Samples(VECTOR_ND<T>& u,ARRAY_VIEW<TV>& V,const int d)
{
    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);
    u.Resize(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:{u(i)=V(p)(d);}break;
        case SIMPLEX_SEGMENT:
            {const TV_INT2& vtx=segment_mesh.elements(p);
            u(i)=(T).5*(V(vtx(1))(d)+V(vtx(2))(d));}break;
        case SIMPLEX_TRIANGLE:
            {const TV_INT3& vtx=triangle_mesh.elements(p);
            u(i)=(T)one_third*(V(vtx(1))(d)+V(vtx(2))(d)+V(vtx(3))(d));}break;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Variable_Viscosity(ARRAY<T>& variable_kinematic_viscosity,const int d)
{
    Update_Shear_Rate(particles.shear_rate);    ////assume the current M_inv is V_inv
    variable_kinematic_viscosity.Resize(solver.velocity_samples.m);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        int p=solver.velocity_samples(i);
        T shear_rate=particles.shear_rate(p);
        T vis=kinematic_viscosity;
        switch(variable_viscosity_mode){
        case SHEAR_THINNING:{
            T mu_0=kinematic_viscosity;T mu_inf=kinematic_viscosity_min;//T lamda=(T).2;T alpha=(T).8;T n=(T).2; 
            //T vis=Carreau_Yasuda_Viscosity(shear_rate,mu_0,mu_inf,lamda,alpha,n);
            vis=Linear_Viscosity(shear_rate,mu_0,mu_inf,shear_rate_max);
        }break;
        case BINGHAM_PLASTIC:{
            vis=Bingham_Viscosity_Dual(shear_rate,shear_rate_max,(T)0,/*rigid_viscosity*/kinematic_viscosity,kinematic_viscosity_min);
        }break;
        }
        variable_kinematic_viscosity(i)=vis;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Shear_Rate(ARRAY_VIEW<T>& shear_rate)
{
    ARRAY<T_MATRIX> grad_uT;grad_uT.Resize(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());

    ////update grad_uT
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u;Interpolate_Velocity_Component_To_Pressure_Samples(u,particles.V,d);

        VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
        solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
        grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u

        for(int i=1;i<=solver.velocity_samples.m;i++){
            TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
            grad_uT(i).Column(d)=grad_u_i;
        }
    }

    INDIRECT_ARRAY<ARRAY_VIEW<T> > shear_rate_indir_array(shear_rate,solver.velocity_samples);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        T_MATRIX strain_rate=grad_uT(i)+grad_uT(i).Transposed();
        shear_rate_indir_array(i)=sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
        //shear_rate_indir_array(i)=(T).8*shear_rate_indir_array(i)+(T).2*sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
    }
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Carreau_Yasuda_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T lamda,const T alpha,const T n)
{
    return mu_inf+(mu_0-mu_inf)*pow(1+pow(lamda*shear_rate,alpha),(n-(T)1)/alpha);
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Linear_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T max_shear_rate)
{
    T s=shear_rate/max_shear_rate;  ////normalize
    T f=(T)1-(T)1/((T)1+exp((T)5-(T)20*s));
    return mu_inf+(mu_0-mu_inf)*f;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Bingham_Viscosity_Regularized(const T shear_rate,const T plastic_viscosity,const T yield_stress,const T transition)
{
    T epsilon=(T)1e-7;
    return plastic_viscosity+yield_stress*((T)1-exp(-transition*shear_rate))/(shear_rate+epsilon);
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Bingham_Viscosity_Dual(const T shear_rate,T yield_shear_rate,T yield_shear_stress,const T rigid_viscosity,const T plastic_viscosity)
{
    if(shear_rate<yield_shear_rate)return rigid_viscosity;
    else {//LOG::cout<<"plas "<<shear_rate<<std::endl;
    return plastic_viscosity+yield_shear_stress/shear_rate;}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_A()
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);
        solver.Update_A(solver.G,solver.M_inv,solver.A);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T time)
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);
        VECTOR_ND<T> u(solver.pressure_samples.m);
        VECTOR_ND<T> neg_laplace_u(solver.pressure_samples.m);
        Interpolate_Velocity_Component_To_Pressure_Samples(u,particles.V,d);
        solver.A.Times(u,neg_laplace_u);
        for(int i=1;i<=solver.pressure_samples.m;i++){
            int p=solver.pressure_samples(i).index;
            int type=solver.pressure_samples(i).type;
            T f=(use_variable_viscosity?(T)1:kinematic_viscosity)*density*-neg_laplace_u(i);
            switch(type){
            case SIMPLEX_PARTICLE:{F(p)(d)+=f;}break;
            case SIMPLEX_SEGMENT:{const TV_INT2& vtx=segment_mesh.elements(p);for(int j=1;j<=2;j++)F(vtx(j))(d)+=(T).5*f;}break;
            case SIMPLEX_TRIANGLE:{const TV_INT3& vtx=triangle_mesh.elements(p);for(int j=1;j<=3;j++)F(vtx(j))(d)+=(T)one_third*f;}break;
            }
        }
    }

    if(mpi_codim){ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<TV>& V,/*rst*/ARRAY_VIEW<TV>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const int d,const T dt)
{
    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);

    VECTOR_ND<T> delta_u(solver.pressure_samples.m);
    VECTOR_ND<T> particle_u(solver.velocity_samples.m);particle_u.Fill((T)0);
    VECTOR_ND<T> particle_delta_u(solver.velocity_samples.m);particle_delta_u.Fill((T)0);
    VECTOR_ND<T> particle_weight(solver.velocity_samples.m);particle_weight.Fill((T)0);

    for(int i=1;i<=solver.pressure_samples.m;i++){delta_u(i)=u1(i)-u0(i);}

    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:
            {
                T particle_volume=codimensional_mesh_object.Nodal_Volume(p);
                int p_in_vs=particle_index_in_velocity_samples(p);
                particle_u(p_in_vs)+=u1(i)*particle_volume;
                particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                particle_weight(p_in_vs)+=particle_volume;
            }break;
        case SIMPLEX_SEGMENT:
            {
                const TV_INT2& vtx=segment_mesh.elements(p);
                for(int d=1;d<=2;d++){
                    T particle_volume=codimensional_mesh_object.Nodal_Segment_Volume(vtx(d),p);
                    int p_in_vs=particle_index_in_velocity_samples(vtx(d));
                    particle_u(p_in_vs)+=u1(i)*particle_volume;
                    particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                    particle_weight(p_in_vs)+=particle_volume;
                }
            }break;
        case SIMPLEX_TRIANGLE:
            {
                const TV_INT3& vtx=triangle_mesh.elements(p);
                for(int d=1;d<=3;d++){
                    T particle_volume=codimensional_mesh_object.Nodal_Triangle_Volume(vtx(d),p);
                    int p_in_vs=particle_index_in_velocity_samples(vtx(d));
                    particle_u(p_in_vs)+=u1(i)*particle_volume;
                    particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                    particle_weight(p_in_vs)+=particle_volume;
                }
            }break;
        }
    }

    if(use_staggered_dof){////deltu_u for FLIP
        T interp_weight=(T).0;
        for(int i=1;i<=solver.velocity_samples.m;i++){
            if(particle_weight(i)!=(T)0){particle_u(i)/=particle_weight(i);particle_delta_u(i)/=particle_weight(i);}
            int p=solver.velocity_samples(i);
            T old_particle_u=V(p)(d);
            V(p)(d)=((T)1-interp_weight)*particle_u(i)+interp_weight*(V(p)(d)+particle_delta_u(i));
            F(p)(d)=(V(p)(d)-old_particle_u)*particles.mass(p)/dt;
        }
    }
    else{
        for(int i=1;i<=solver.velocity_samples.m;i++){
            if(particle_weight(i)!=(T)0)particle_u(i)/=particle_weight(i);
            int p=solver.velocity_samples(i);
            T old_particle_u=V(p)(d);
            //if(d==2)LOG::cout<<p<<": "<<V(p)(d)<<", "<<particle_u(i)<<std::endl;
            V(p)(d)=particle_u(i);
            F(p)(d)=(V(p)(d)-old_particle_u)*particles.mass(p)/dt;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    for(int d=1;d<=TV::m;d++){
        CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver=solvers(d);

        ////initialize volume
        VECTOR_ND<T> vol(solver.pressure_samples.m);
        for(int i=1;i<=solver.pressure_samples.m;i++){
            int p=solver.pressure_samples(i).index;
            int type=solver.pressure_samples(i).type;
            vol(i)=codimensional_mesh_object.Element_Volume(p,type);
        }

        VECTOR_ND<T> u_old(solver.pressure_samples.m);

        Update_Dirichlet_Boundary_Samples(psi_Ds(d),solver);
        Interpolate_Velocity_Component_To_Pressure_Samples(u_old,particles.V,d);
        
        VECTOR_ND<T>& b=solver.b;
        b.Resize(solver.pressure_samples.m);

        for(int i=1;i<=solver.pressure_samples.m;i++){
            T coef=vol(i)/(dt*(use_variable_viscosity?(T)1:kinematic_viscosity));
            solver.A(i,i)+=coef;b(i)=coef*u_old(i);
        }

        solver.Modify_b_With_Dirichlet_Boundary_Conditions(b,solver.psi_D);

        solver.Solve();
        VECTOR_ND<T>& u_new=solver.p;
        Update_Velocity_And_Force_Component_FLIP(V,F,u_old,u_new,d,dt);
    }

    if(mpi_codim){
        ARRAY_VIEW<TV> V_array_view(V);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(V_array_view);
        ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);
    }
}

//template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<float,2> >;
template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<double,2> >;
template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<double,3> >; 
#endif

