    //#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_VISCOSITY_FORCE__
#define __CODIMENSIONAL_VISCOSITY_FORCE__
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_VISCOSITY_FORCE
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef MATRIX<T,TV::m> T_MATRIX;typedef DIAGONAL_MATRIX<T,TV::m> T_DIAGONAL_MATRIX;typedef SYMMETRIC_MATRIX<T,TV::m> T_SYMMETRIC_MATRIX;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;
    T kinematic_viscosity;
    T density;
    T thickness_scale;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_N;
    VECTOR<ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >,TV::m> psi_Ds;
    VECTOR<ARRAY<int>,4> element_index_in_pressure_samples;
    ARRAY<int> particle_index_in_velocity_samples;
    VECTOR<CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>,3> solvers;

    ////variable viscosity
    enum VARIABLE_VISCOSITY_MODE{SHEAR_THINNING=1,BINGHAM_PLASTIC} variable_viscosity_mode;
    T kinematic_viscosity_min;
    T rigid_viscosity;
    T shear_rate_max;

    bool use_constant_extrapolation;
    bool use_staggered_dof;
    bool use_variable_viscosity;

    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_VISCOSITY_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    virtual void Update_Position_Based_State(const T time,const bool is_position_update);
    virtual void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T time);
    virtual void Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);
    virtual void Update_Neumann_Boundary_On_Free_Surface(ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N);

    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input){mpi_codim=mpi_codim_input;for(int d=1;d<=TV::m;d++)solvers(d).mpi_codim=mpi_codim;}
    void Use_Constant_Extrapolation(bool use_constant_extrapolation_input=true)
    {use_constant_extrapolation=use_constant_extrapolation_input;solver.use_constant_extrapolation=use_constant_extrapolation;}
    void Use_Staggered_DoF(bool use_staggered_dof_input=true)
    {use_staggered_dof=use_staggered_dof_input;solver.use_all_pressure_sample_on_nodes=!use_staggered_dof;}
    void Use_Variable_Viscosity(bool use_variable_viscosity=true){use_variable_viscosity=true;}

protected:
    virtual void Update_G();
    virtual void Update_M_inv();
    virtual void Update_A();

    ////variable viscosity
    virtual void Update_Variable_Viscosity(ARRAY<T>& variable_kinematic_viscosity);
    void Update_Shear_Rate(ARRAY_VIEW<T>& shear_rate);
    T Carreau_Yasuda_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T lambda,const T alpha,const T n);
    T Linear_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T max_shear_rate);
    T Bingham_Viscosity_Regularized(const T shear_rate,const T plastic_viscosity,const T yield_stress,const T transition);
    T Bingham_Viscosity_Dual(const T shear_rate,T yield_shear_rate,T yield_shear_stress,const T rigid_viscosity,const T plastic_viscosity);
    void Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<TV>& V,/*rst*/ARRAY_VIEW<TV>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const int d,const T dt);

    ////helper functions
    void Interpolate_Velocity_Component_To_Pressure_Samples(VECTOR_ND<T>& u,ARRAY_VIEW<TV>& V,const int d);
    void Update_Dirichlet_Boundary_Samples(const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver);
    void Update_Neumann_Boundary_Samples(const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N,const HASHTABLE<int>& invalid_indices,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver);
};
}
#endif