//#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Meshing/CODIMENSIONAL_MESH_HASHTABLE.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::CODIMENSIONAL_VISCOSITY_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input/*=0*/)
    :codimensional_mesh_object(codimensional_mesh_object_input),particles(codimensional_mesh_object.particles),
    segment_mesh(codimensional_mesh_object.mesh.segment_mesh),triangle_mesh(codimensional_mesh_object.mesh.triangle_mesh),tetrahedron_mesh(codimensional_mesh_object.mesh.tetrahedron_mesh),
    kinematic_viscosity((T)1e-6),density((T)1000.),length_scale((T)1e-3),length_scale_min((T)1e-3),thickness_scale((T)1e-4),variable_viscosity_mode(SHEAR_THINNING_LINEAR),kinematic_viscosity_min(kinematic_viscosity),rigid_viscosity((T)10),
    shear_rate_max((T)20),bp_alpha((T)1),use_particle_thickness(true),use_constant_extrapolation(false),use_staggered_dof(true),use_shear_dependent_variable_viscosity(false),use_thickness_dependent_variable_viscosity(false),use_temperature_dependent_variable_viscosity(false),mpi_codim(mpi_codim_input)
{
    solver.use_target_div=false;
    solver.codimensional_mesh_object=&codimensional_mesh_object;
    solver.element_index_in_pressure_samples=&element_index_in_pressure_samples;
    solver.particle_index_in_velocity_samples=&particle_index_in_velocity_samples;
    solver.psi_D_on_mesh_hashtable=new HASHTABLE<int>();
    solver.mpi_codim=mpi_codim;
    Use_Constant_Extrapolation(use_constant_extrapolation);
    Use_Staggered_DoF(use_staggered_dof);
    solver.use_precoditioner=true;

    lambda_thinning=(T)1;
    alpha_thinning=(T)1;
    n_thinning=(T).4;

    lambda_thickening=(T).05;
    alpha_thickening=(T).8;
    n_thickening=(T)6;

    tetrahedron_viscosity_factor=(T)1;
    triangle_viscosity_factor=(T)1;
    segment_viscosity_factor=(T)1;
    
    viscosity_min_clamp=(T)1e-10;
    viscosity_max_clamp=(T)1e5;
    pic_flip_weight=(T).8;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Position_Based_State(const T time,const bool is_position_updated)
{
    Update_Shear_Rate2();
    if(use_shear_dependent_variable_viscosity)Update_Variable_Viscosity();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Reset_Solver()
{
    solver.Clear();
    solver.use_particle_thickness=use_particle_thickness;
    solver.use_all_pressure_sample_on_nodes=!use_staggered_dof;

    solver.use_target_div=false;
    solver.codimensional_mesh_object=&codimensional_mesh_object;
    solver.length_scale=length_scale;
    solver.length_scale_min=length_scale_min;
    solver.thickness_scale=thickness_scale;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_G(ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N_on_mesh,ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D_on_mesh)
{
    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>::Initialize(solver,&psi_N_on_mesh,&psi_D_on_mesh);

    if(mpi_codim)mpi_codim->Update_Boundary_And_Ghost_Pressure_Indices(solver.pressure_samples,solver.codimensional_mesh_object->mesh,solver.use_all_pressure_sample_on_nodes,
        solver.boundary_pressure_indices,solver.ghost_pressure_indices,solver.domain_pressure_indices);
    //LOG::cout<<"viscosity solver.psi_N: "<<solver.psi_N.m<<", psi_D: "<<solver.psi_D.m<<std::endl;

    solver.Update_G(solver.G);
    solver.Modify_G_With_Boundary_Conditions(solver.G,solver.psi_N);
    solver.Update_GT(solver.G,solver.GT);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Dirichlet_Boundary_Samples(const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D_on_mesh,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver)
{
    solver.psi_D.Remove_All();
    for(int i=1;i<=psi_D_on_mesh.m;i++){
        int p=psi_D_on_mesh(i).index;int type=psi_D_on_mesh(i).type;T v=psi_D_on_mesh(i).v;
        int region=codimensional_mesh_object.mesh.Particle_Region_Index(p);if(region==0)continue;
        if(type==SIMPLEX_PARTICLE&&solver.Find_Velocity_Sample(p)!=0){  ////currently all psi_D are on particles
            solver.Add_Dirichlet_Boundary_Sample(p,type,v);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Neumann_Boundary_Samples(const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N_on_mesh,const HASHTABLE<int>& invalid_indices,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver)
{
    solver.psi_N.Remove_All();
    for(int i=1;i<=psi_N_on_mesh.m;i++){
        int p=psi_N_on_mesh(i).index;
        if(invalid_indices.Contains(p)){/*LOG::cout<<"skip "<<p<<std::endl;*/continue;}
        int region=codimensional_mesh_object.mesh.Particle_Region_Index(p);if(region==0)continue;
        TV direction=psi_N_on_mesh(i).direction;T v=psi_N_on_mesh(i).v;
        if(solver.Find_Velocity_Sample(p)!=0){////p is the velocity sample in the subsystem
            solver.Add_Neumann_Boundary_Sample(p,direction,v);
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_M_inv()
{
    solver.Update_V_inv(solver.M_inv);
    
    if(use_shear_dependent_variable_viscosity){
        for(int i=1;i<=solver.velocity_samples.m;i++){
            for(int d=1;d<=TV::m;d++)solver.M_inv((TV::m*(i-1))+d)*=variable_kinematic_viscosity(i);
        }
    }
    else{
        for(int i=1;i<=solver.velocity_samples.m;i++){
            for(int d=1;d<=TV::m;d++)solver.M_inv((TV::m*(i-1))+d)*=kinematic_viscosity;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Interpolate_Particle_Attribute_To_Pressure_Sample(ARRAY_VIEW<TV>& V,const int d,VECTOR_ND<T>& u)
{
    u.Resize(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:{u(i)=V(p)(d);}break;
        case SIMPLEX_SEGMENT:
            {const TV_INT2& vtx=segment_mesh.elements(p);
            //u(i)=(T).5*(V(vtx(1))(d)+V(vtx(2))(d));
            T w1=(T).5;
            if(use_particle_thickness){w1=pow<2>(particles.thickness(vtx(1)));T w2=pow<2>(particles.thickness(vtx(2)));w1/=(w1+w2);}
            u(i)=w1*V(vtx(1))(d)+((T)1-w1)*V(vtx(2))(d);
            }break;
        case SIMPLEX_TRIANGLE:
            {const TV_INT3& vtx=triangle_mesh.elements(p);
            //u(i)=(T)one_third*(V(vtx(1))(d)+V(vtx(2))(d)+V(vtx(3))(d));
            T w1=(T)one_third;T w2=(T)one_third;T w3=(T)one_third;
            if(use_particle_thickness){
                w1=particles.thickness(vtx(1));w2=particles.thickness(vtx(2));w3=particles.thickness(vtx(3));
                T sum=w1+w2+w3;w1/=sum;w2/=sum;w3/=sum;}
            u(i)=w1*V(vtx(1))(d)+w2*V(vtx(2))(d)+w3*V(vtx(3))(d);
            }break;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Interpolate_Particle_Attribute_To_Pressure_Sample(ARRAY_VIEW<T>& V,/*rst*/VECTOR_ND<T>& u)
{
    u.Resize(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:{u(i)=V(p);}break;
        case SIMPLEX_SEGMENT:
            {const TV_INT2& vtx=segment_mesh.elements(p);
            //u(i)=(T).5*(V(vtx(1))+V(vtx(2)));
            T w1=(T).5;if(use_particle_thickness){w1=pow<2>(particles.thickness(vtx(1)));T w2=pow<2>(particles.thickness(vtx(2)));w1/=(w1+w2);}
            u(i)=w1*V(vtx(1))+((T)1-w1)*V(vtx(2));}break;
        case SIMPLEX_TRIANGLE:
            {const TV_INT3& vtx=triangle_mesh.elements(p);
            //u(i)=(T)one_third*(V(vtx(1))+V(vtx(2))+V(vtx(3)));}break;
            T w1=(T)one_third;T w2=(T)one_third;T w3=(T)one_third;
            if(use_particle_thickness){
                w1=particles.thickness(vtx(1));w2=particles.thickness(vtx(2));w3=particles.thickness(vtx(3));
                T sum=w1+w2+w3;w1/=sum;w2/=sum;w3/=sum;}
            u(i)=w1*V(vtx(1))+w2*V(vtx(2))+w3*V(vtx(3));}break;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Interpolate_Pressure_Sample_To_Particle_Attribute(VECTOR_ND<T>& p,/*rst*/ARRAY_VIEW<T>& v)
{
    for(int i=1;i<=v.m;i++)v(i)=(T)0;
    ARRAY<T> v_weights(v.m);v_weights.Fill((T)0);
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int idx=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:{
            v(idx)+=p(i);v_weights(idx)+=(T)1;}break;
        case SIMPLEX_SEGMENT:{
            const TV_INT2& vtx=segment_mesh.elements(idx);
            //for(int d=1;d<=2;d++){v(vtx(d))+=(T).5*p(i);v_weights(vtx(d))+=(T).5;}
            T w1=(T).5;if(use_particle_thickness){w1=pow<2>(particles.thickness(vtx(1)));T w2=pow<2>(particles.thickness(vtx(2)));w1/=(w1+w2);}
            v(vtx(1))+=w1*p(i);v_weights(vtx(1))+=w1;
            v(vtx(2))+=((T)1-w1)*p(i);v_weights(vtx(2))+=((T)1-w1);
        }break;
        case SIMPLEX_TRIANGLE:{
            const TV_INT3& vtx=triangle_mesh.elements(idx);
            //for(int d=1;d<=3;d++){v(vtx(d))+=(T)one_third*p(i);v_weights(vtx(d))+=(T)one_third;}
            TV3 w=TV::Constant_Vector((T)one_third);
            if(use_particle_thickness){
                T sum=(T)0;for(int d=1;d<=3;d++){w(d)=particles.thickness(vtx(d));sum+=w(d);}
                if(sum>(T)0)for(int d=1;d<=3;d++)w(d)/=sum;
            }
            for(int d=1;d<=3;d++){v(vtx(d))+=w(d)*p(i);v_weights(vtx(d))+=w(d);}
        }break;
        }
    }
    //LOG::cout<<"v_weights: "<<v_weights<<std::endl;
    for(int i=1;i<=v.m;i++){if(v_weights(i)!=(T)0)v(i)/=v_weights(i);}
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Variable_Viscosity()
{
    ARRAY<T> codim_coef;
    if(use_variable_viscosity_for_codim){
        codim_coef.Resize(particles.X.m);codim_coef.Fill((T)1);
        for(int i=1;i<=particles.X.m;i++){
            int p_type=codimensional_mesh_object.mesh.node_types(i);
            if(Is_On_Tetrahedron_Mesh(p_type))codim_coef(i)=tetrahedron_viscosity_factor;
            else if(Is_On_Triangle_Mesh(p_type))codim_coef(i)=triangle_viscosity_factor;
            else if(Is_On_Segment_Mesh(p_type)){codim_coef(i)=segment_viscosity_factor;}

            //HASHTABLE<TV_INT2> edge_hashtable;ARRAY<RANGE<TV> > domain;domain.Append(RANGE<TV>(TV::Constant_Vector((T)-FLT_MAX),TV::Constant_Vector((T)FLT_MAX)));
            //Update_Edge_Hashtable(segment_mesh.elements,particles.X,domain,(T)0,edge_hashtable);
            //int iter_num=3;
            //for(int iter=1;iter<=iter_num;iter++)for(HASHTABLE_ITERATOR<TV_INT2> iter(edge_hashtable);iter.Valid();iter.Next()){
            //    const TV_INT2& v=iter.Key();
            //    T coef1=codim_coef(v(1));T coef2=codim_coef(v(2));
            //    T intp_coef=(T).5*max(coef1,coef2)+(T).5*min(coef1,coef2);
            //    codim_coef(v(1))=codim_coef(v(2))=intp_coef;
            //}
        }
    }

    //Update_Shear_Rate(particles.shear_rate);    ////assume the current M_inv is V_inv
    variable_kinematic_viscosity.Resize(solver.velocity_samples.m);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        int p=solver.velocity_samples(i);
        T shear_rate=particles.shear_rate(p);
        T vis=kinematic_viscosity;

        // carreau-yasuda 'good' values:
        // mu_0: 1000
        // mu_inf: 1
        // lambda: 0.4
        // alpha: 0.8
        // n: 0.2

        // shear-thinning 'good' values:
        // mu_0: 1000
        // mu_inf: 1
        // max_shear_rate: 100

        switch(variable_viscosity_mode){
        case SHEAR_THINNING_LINEAR:{
            T mu_0=kinematic_viscosity;T mu_inf=kinematic_viscosity_min;//T lamda=(T).2;T alpha=(T).8;T n=(T).2; 
            //T vis=Carreau_Yasuda_Viscosity(shear_rate,mu_0,mu_inf,lamda,alpha,n);
            vis=Linear_Viscosity(shear_rate,mu_0,mu_inf,shear_rate_max);
        }break;
        case BINGHAM_PLASTIC_DUAL_2:{
            T tau_0=(T)60;
            vis=Bingham_Viscosity_Dual2(shear_rate,tau_0,/*rigid_viscosity*/kinematic_viscosity,kinematic_viscosity_min);
        }break;
        case CARREAU_YASUDA_SHEAR_THINNING:{
            //T lambda=(T).05;T alpha=(T)0.8;T n=(T)0.2;    ////for sphere example and paint example
            //T lambda=(T)1;T alpha=(T)1;T n=(T).4;           ////for bunny example
            //LOG::cout<<"param: "<<lambda_thinning<<", "<<alpha_thinning<<", "<<n_thinning<<std::endl;
            vis=Carreau_Yasuda_Viscosity(shear_rate,kinematic_viscosity,kinematic_viscosity_min,lambda_thinning,alpha_thinning,n_thinning);
        }break;
        case CARREAU_YASUDA_SHEAR_THICKENING:{
            //T lambda=(T).05;T alpha=(T).8;T n=(T)6;       ////fo sphere example
            //T lambda=(T).05;T alpha=(T).8;T n=(T)10;
            //T lambda=(T).02;T alpha=(T)1;T n=(T)10;
            //T lambda=(T)1;T alpha=(T)1;T n=(T)3.2;
            //T lambda=(T).8;T alpha=(T)1;T n=(T)6;         
            //T lambda=(T).35;T alpha=(T)1;T n=(T)8;        ////for bunny example
            vis=Carreau_Yasuda_Viscosity(shear_rate,kinematic_viscosity,kinematic_viscosity_min,lambda_thickening,alpha_thickening,n_thickening);
            //LOG::cout<<"thickening: "<<shear_rate<<": "<<kinematic_viscosity<<", "<<kinematic_viscosity_min<<": "<<vis<<std::endl;
        }break;
        case BINGHAM_PLASTIC_DUAL:{
            vis=Bingham_Viscosity_Dual(shear_rate,shear_rate_max,kinematic_viscosity,kinematic_viscosity_min,bp_alpha);
        }break;
        }

        if(use_temperature_dependent_variable_viscosity){
            T max_temperature=(T)1000;
            T frac=Temperature_Dependent_Viscosity_Coefficient(particles.temperature(i),max_temperature);
            vis*=frac;
        }

        if(use_thickness_dependent_variable_viscosity){
            T frac=(T)1;
            if(Is_On_Segment_Mesh(codimensional_mesh_object.mesh.node_types(p)))frac=std::pow(particles.thickness(p)/thickness_scale,(T)2);
            else if(Is_On_Triangle_Mesh(codimensional_mesh_object.mesh.node_types(p)))frac=particles.thickness(p)/thickness_scale;
            frac=clamp(frac,(T)0,(T)1);
            vis*=frac;
        }

        if(use_variable_viscosity_for_codim){vis*=codim_coef(p);}

        vis=clamp(vis,viscosity_min_clamp,viscosity_max_clamp);
        variable_kinematic_viscosity(i)=vis;
    }

    if(particles.store_viscosity)for(int i=1;i<=solver.velocity_samples.m;i++){
        int p=solver.velocity_samples(i);particles.viscosity(p)=variable_kinematic_viscosity(i);
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Test_Viscosity_Models()
{
    T shear_rate_0=(T)0;T shear_rate_1=(T)100;
    for(int i=1;i<=100;i++){
        T frac=(T)i/(T)100;
        T shear_rate=((T)1-frac)*shear_rate_0+frac*shear_rate_1;
        T lambda=(T).05;T alpha=(T).8;T n=(T)5;
        T mu_0=(T)2e2;T mu_1=1;
        T vis=Carreau_Yasuda_Viscosity(shear_rate,mu_0,mu_1,lambda,alpha,n);
        LOG::cout<<"cy "<<shear_rate<<": "<<vis<<std::endl;
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Shear_Rate()
{
    ARRAY_VIEW<T>& shear_rate=particles.shear_rate;
    grad_uT.Resize(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());

    ////update grad_uT
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u;Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);

        VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
        solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
        grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u

        for(int i=1;i<=solver.velocity_samples.m;i++){
            TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
            grad_uT(i).Column(d)=grad_u_i;
        }
    }

    INDIRECT_ARRAY<ARRAY_VIEW<T> > shear_rate_indir_array(shear_rate,solver.velocity_samples);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        T_MATRIX strain_rate=grad_uT(i)+grad_uT(i).Transposed();
        shear_rate_indir_array(i)=sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
        //shear_rate_indir_array(i)=(T).8*shear_rate_indir_array(i)+(T).2*sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Shear_Rate2()
{
    //for(int i=1;i<=particles.X.m;i++)particles.thickness(i)=(T)1;   ////dbg

    //////constant extrapolation
    //bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    //solver.use_constant_extrapolation=true;
    //Reset_Solver();
    //ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > empty_psi_Ns;
    //ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    //Update_G(empty_psi_Ns,empty_psi_Ds);
    //solver.Update_V_inv(solver.M_inv);

    ////no constant extrapolation
    bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    solver.use_constant_extrapolation=false;
    Reset_Solver();
    ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    Update_G(psi_Ns_on_surface,empty_psi_Ds);
    solver.Update_V_inv(solver.M_inv);

    //LOG::cout<<"psi_N: "<<psi_Ns_on_surface.m<<std::endl;
    //for(int i=1;i<=psi_Ns_on_surface.m;i++){
    //    int index=psi_Ns_on_surface(i).index;
    //    if(particles.guid(index)==4403)
    //        LOG::cout<<i<<": "<<psi_Ns_on_surface(i).direction<<", "<<psi_Ns_on_surface(i).v<<std::endl;
    //}

    //LOG::cout<<"psi_N: "<<psi_Ns_on_mesh(1).m<<std::endl;
    //for(int i=1;i<=psi_Ns_on_mesh(1).m;i++){
    //    LOG::cout<<i<<": "<<psi_Ns_on_mesh(1)(i).direction<<", "<<psi_Ns_on_mesh(1)(i).v<<std::endl;
    //}
    //LOG::cout<<"psi_D: "<<psi_Ds_on_mesh(1).m<<std::endl;
    //for(int i=1;i<=psi_Ds_on_mesh(1).m;i++){
    //    LOG::cout<<i<<": "<<psi_Ds_on_mesh(1)(i).type<<", "<<psi_Ds_on_mesh(1)(i).v<<std::endl;
    //}

    //LOG::cout<<"psi_N: "<<solver.psi_N.m<<std::endl;
    //for(int i=1;i<=solver.psi_N.m;i++){
    //    LOG::cout<<solver.psi_N(i).index<<": "<<solver.psi_N(i).direction<<", "<<solver.psi_N(i).v<<std::endl;
    //}
    //LOG::cout<<"psi_D: "<<solver.psi_D.m<<std::endl;
    //for(int i=1;i<=solver.psi_D.m;i++){
    //    LOG::cout<<i<<": "<<solver.psi_D(i).type<<", "<<solver.psi_D(i).v<<std::endl;
    //}

    ////update grad_uT
    ARRAY_VIEW<T>& shear_rate=particles.shear_rate;
    grad_uT.Resize(solver.velocity_samples.m);grad_uT.Fill(T_MATRIX());
    for(int d=1;d<=TV::m;d++){
        VECTOR_ND<T> u;Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);
        //u.Fill((T)1);       ////dbg
        //LOG::cout<<"V "<<particles.V<<std::endl;
        //LOG::cout<<"u "<<u<<std::endl;

        VECTOR_ND<T> grad_u_vec(solver.velocity_samples.m*TV::m);
        solver.G.Times(u,grad_u_vec);   ////volume weighted gradient u
        grad_u_vec=solver.M_inv*grad_u_vec; ////gradient u
        
        for(int i=1;i<=solver.velocity_samples.m;i++){
            TV grad_u_i(grad_u_vec((i-1)*TV::m+1),grad_u_vec((i-1)*TV::m+2),grad_u_vec((i-1)*TV::m+3));
            grad_uT(i).Column(d)=grad_u_i;
        }
    }
    INDIRECT_ARRAY<ARRAY_VIEW<T> > shear_rate_indir_array(shear_rate,solver.velocity_samples);
    for(int i=1;i<=solver.velocity_samples.m;i++){
        T_MATRIX strain_rate=grad_uT(i)+grad_uT(i).Transposed();
        shear_rate_indir_array(i)=sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
        //shear_rate_indir_array(i)=(T).8*shear_rate_indir_array(i)+(T).2*sqrt((T).5*(strain_rate*strain_rate.Transposed()).Trace());
    }

    //////update grad_vis
    //ARRAY_VIEW<T>& viscosity=particles.viscosity;
    //grad_vis.Resize(solver.velocity_samples.m);grad_vis.Fill(TV());
    ////LOG::cout<<"vel sample: "<<solver.velocity_samples.m<<std::endl;
    //{
    //    VECTOR_ND<T> vis;Interpolate_Particle_Attribute_To_Pressure_Sample(viscosity,vis);
    //    VECTOR_ND<T> grad_vis_vec(solver.velocity_samples.m*TV::m);
    //    //vis.Fill((T)1); ////dbg
    //    //LOG::cout<<"vis: "<<vis<<std::endl;
    //    //LOG::cout<<"G: "<<solver.G<<std::endl;
    //    
    //    solver.G.Times(vis,grad_vis_vec);   ////volume weighted gradient vis
    //    //LOG::cout<<"vol_w_grad_vis_vec: "<<grad_vis_vec<<std::endl;
    //    grad_vis_vec=solver.M_inv*grad_vis_vec; ////gradient vis

    //    //LOG::cout<<"grad_vis_vec: "<<grad_vis_vec<<std::endl;
    //    //LOG::cout<<"M_inv: "<<solver.M_inv<<std::endl;

    //    for(int i=1;i<=solver.velocity_samples.m;i++){
    //        TV grad_vis_i(grad_vis_vec((i-1)*TV::m+1),grad_vis_vec((i-1)*TV::m+2),grad_vis_vec((i-1)*TV::m+3));
    //        grad_vis(i)=grad_vis_i;
    //    }
    //}

    solver.use_constant_extrapolation=use_constant_extrapolation_copy;

    ////no constant extrapolation
    //bool use_constant_extrapolation_copy=solver.use_constant_extrapolation;
    //solver.use_constant_extrapolation=false;
    //Reset_Solver();
    ////ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > empty_psi_Ns;
    ////ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> > empty_psi_Ds;
    //Update_G(psi_Ns_on_mesh(1),psi_Ds_on_mesh(1));
    //solver.Update_V_inv(solver.M_inv);
    //Update_Shear_Rate(particles.shear_rate);
    //solver.use_constant_extrapolation=use_constant_extrapolation_copy;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Carreau_Yasuda_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T lamda,const T alpha,const T n)
{
    return mu_inf+(mu_0-mu_inf)*pow(1+pow(lamda*shear_rate,alpha),(n-(T)1)/alpha);
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Linear_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T max_shear_rate)
{
    T s=shear_rate/max_shear_rate;  ////normalize
    T f=(T)1-(T)1/((T)1+exp((T)5-(T)20*s));
    return mu_inf+(mu_0-mu_inf)*f;
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Bingham_Viscosity_Dual(const T shear_rate,T yield_shear_rate,const T rigid_viscosity,const T plastic_viscosity,const T alpha)
{
    T yield_shear_stress=(rigid_viscosity-plastic_viscosity)*yield_shear_rate;
    if(shear_rate<yield_shear_rate)return rigid_viscosity;
    else return plastic_viscosity+yield_shear_stress/(std::pow(shear_rate,alpha));
}

//// same as Bingham_Viscosity_Dual, only accepts tau_0 as an argument instead of critical shear rate
template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Bingham_Viscosity_Dual2(const T shear_rate,T yield_stress,const T rigid_viscosity,const T plastic_viscosity)
{
    T& tau_0=yield_stress;
    T critical_shear_rate=tau_0/(rigid_viscosity-plastic_viscosity);
    T apparent_viscosity;
    if(shear_rate<critical_shear_rate) apparent_viscosity=rigid_viscosity;
    else apparent_viscosity=plastic_viscosity+tau_0/shear_rate;
    return apparent_viscosity;
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_A()
{
    solver.Update_A(solver.G,solver.M_inv,solver.A);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T time)
{
    VECTOR_ND<T> u(solver.pressure_samples.m);
    VECTOR_ND<T> neg_laplace_u(solver.pressure_samples.m);
    for(int d=1;d<=TV::m;d++){
        Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u);
        solver.A.Times(u,neg_laplace_u);
        for(int i=1;i<=solver.pressure_samples.m;i++){
            int p=solver.pressure_samples(i).index;
            int type=solver.pressure_samples(i).type;
            T f=(use_shear_dependent_variable_viscosity?(T)1:kinematic_viscosity)*density*-neg_laplace_u(i);
            switch(type){
            case SIMPLEX_PARTICLE:{F(p)(d)+=f;}break;
            case SIMPLEX_SEGMENT:{const TV_INT2& vtx=segment_mesh.elements(p);for(int j=1;j<=2;j++)F(vtx(j))(d)+=(T).5*f;}break;
            case SIMPLEX_TRIANGLE:{const TV_INT3& vtx=triangle_mesh.elements(p);for(int j=1;j<=3;j++)F(vtx(j))(d)+=(T)one_third*f;}break;
            }
        }
    }

    if(mpi_codim){ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);}
}

template<class TV>
typename TV::SCALAR PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Temperature_Dependent_Viscosity_Coefficient(const T temperature,const T max_temperature)
{
    //T frac=temperature/max_temperature;T c0=(T)1;T c1=(T).1;return c0*((T)1-frac)+c1*frac;   ////linear
    T b=(T)1e-2;return exp(-b*temperature);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Velocity_And_Force_Component_FLIP_Helper(const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,/*rst*/VECTOR_ND<T>& particle_u,/*rst*/VECTOR_ND<T>& particle_delta_u,/*rst*/VECTOR_ND<T>& particle_weight)
{
    VECTOR_ND<T> delta_u(solver.pressure_samples.m);
    for(int i=1;i<=solver.pressure_samples.m;i++){delta_u(i)=u1(i)-u0(i);}
    for(int i=1;i<=solver.pressure_samples.m;i++){
        int p=solver.pressure_samples(i).index;
        int type=solver.pressure_samples(i).type;
        switch(type){
        case SIMPLEX_PARTICLE:
            {
                T particle_volume=codimensional_mesh_object.Nodal_Volume(p);
                int p_in_vs=particle_index_in_velocity_samples(p);
                particle_u(p_in_vs)+=u1(i)*particle_volume;
                particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                particle_weight(p_in_vs)+=particle_volume;
            }break;
        case SIMPLEX_SEGMENT:
            {
                const TV_INT2& vtx=segment_mesh.elements(p);
                for(int j=1;j<=2;j++){
                    T particle_volume=codimensional_mesh_object.Nodal_Segment_Volume(vtx(j),p);
                    int p_in_vs=particle_index_in_velocity_samples(vtx(j));
                    particle_u(p_in_vs)+=u1(i)*particle_volume;
                    particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                    particle_weight(p_in_vs)+=particle_volume;
                }
            }break;
        case SIMPLEX_TRIANGLE:
            {
                const TV_INT3& vtx=triangle_mesh.elements(p);
                for(int j=1;j<=3;j++){
                    int p_in_vs=particle_index_in_velocity_samples(vtx(j));
                    T particle_volume=codimensional_mesh_object.Nodal_Triangle_Volume(vtx(j),p);

                    //if(particles.guid(vtx(j))==60)LOG::cout<<"~~~~~~~~~ aaaa guid 60~~         i "<<i<<" p "<<p<<": "<<", pu: "<<particle_u(p_in_vs)<<", pdu: "<<particle_delta_u(i)
                    //    <<", pv: "<<particle_volume<<", pw: "<<particle_weight(p_in_vs)<<", u1: "<<u1(i)<<", u0: "<<u0(i)<<std::endl;     ////dbg
                    
                    particle_u(p_in_vs)+=u1(i)*particle_volume;
                    particle_delta_u(p_in_vs)+=delta_u(i)*particle_volume;
                    particle_weight(p_in_vs)+=particle_volume;
                }
            }break;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<TV>& V,/*rst*/ARRAY_VIEW<TV>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const int d,const T dt)
{
    VECTOR_ND<T> particle_u(solver.velocity_samples.m);particle_u.Fill((T)0);
    VECTOR_ND<T> particle_delta_u(solver.velocity_samples.m);particle_delta_u.Fill((T)0);
    VECTOR_ND<T> particle_weight(solver.velocity_samples.m);particle_weight.Fill((T)0);

    Update_Velocity_And_Force_Component_FLIP_Helper(u0,u1,particle_u,particle_delta_u,particle_weight);

    if(use_staggered_dof){////deltu_u for FLIP
        for(int i=1;i<=solver.velocity_samples.m;i++){
            int p=solver.velocity_samples(i);
            if(particle_weight(i)!=(T)0){particle_u(i)/=particle_weight(i);particle_delta_u(i)/=particle_weight(i);}
            T old_particle_u=V(p)(d);

            //////TODO: fix the hack
            //if(use_variable_viscosity_for_codim){
            //    int p_type=codimensional_mesh_object.mesh.node_types(p);
            //    if(Is_On_Segment_Mesh(p_type))particle_delta_u(i)*=segment_viscosity_factor;
            //    else if(Is_On_Triangle_Mesh(p_type))particle_delta_u(i)*=triangle_viscosity_factor;
            //}
            
            ////TODO: fix this 
            if(!Is_Valid_Number(particle_u(i)))particle_u(i)=old_particle_u;
            if(!Is_Valid_Number(particle_delta_u(i)))particle_delta_u(i)=(T)0;

            V(p)(d)=((T)1-pic_flip_weight)*particle_u(i)+pic_flip_weight*(V(p)(d)+particle_delta_u(i));
            F(p)(d)=(V(p)(d)-old_particle_u)*particles.mass(p)/dt;

            //////dbg
            //if(particles.guid(p)!=(39800)&&particles.guid(p)!=(45795)&&particles.guid(p)!=(47182)&&particles.guid(p)!=(37884)&&particles.guid(p)!=(18676)){continue;}
            //else{
            //    LOG::cout<<"intp velocity "<<d<<" guid "<<particles.guid(p)<<" w: "<<pic_flip_weight<<" V: "<<V(p)(d)
            //        <<", old_p_u: "<<old_particle_u<<", pu: "<<particle_u(i)<<", pdeltau: "<<particle_delta_u(i)<<", F: "<<F(p)(d)<<std::endl;
            //}
        }
    }
    else{
        for(int i=1;i<=solver.velocity_samples.m;i++){
            if(particle_weight(i)!=(T)0)particle_u(i)/=particle_weight(i);
            int p=solver.velocity_samples(i);
            T old_particle_u=V(p)(d);
            V(p)(d)=particle_u(i);
            F(p)(d)=(V(p)(d)-old_particle_u)*particles.mass(p)/dt;
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<T>& V,/*rst*/ARRAY_VIEW<T>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const T dt)
{
    VECTOR_ND<T> particle_u(solver.velocity_samples.m);particle_u.Fill((T)0);
    VECTOR_ND<T> particle_delta_u(solver.velocity_samples.m);particle_delta_u.Fill((T)0);
    VECTOR_ND<T> particle_weight(solver.velocity_samples.m);particle_weight.Fill((T)0);

    Update_Velocity_And_Force_Component_FLIP_Helper(u0,u1,particle_u,particle_delta_u,particle_weight);

    if(use_staggered_dof){////deltu_u for FLIP
        for(int i=1;i<=solver.velocity_samples.m;i++){
            if(particle_weight(i)!=(T)0){particle_u(i)/=particle_weight(i);particle_delta_u(i)/=particle_weight(i);}
            int p=solver.velocity_samples(i);
            T old_particle_u=V(p);

            ////TODO: fix this
            if(!Is_Valid_Number(particle_u(i)))particle_u(i)=old_particle_u;
            if(!Is_Valid_Number(particle_delta_u(i)))particle_delta_u(i)=(T)0;

            V(p)=((T)1-pic_flip_weight)*particle_u(i)+pic_flip_weight*(V(p)+particle_delta_u(i));
            F(p)=(V(p)-old_particle_u)*particles.mass(p)/dt;
        }
    }
    else{
        for(int i=1;i<=solver.velocity_samples.m;i++){
            if(particle_weight(i)!=(T)0)particle_u(i)/=particle_weight(i);
            int p=solver.velocity_samples(i);
            T old_particle_u=V(p);
            V(p)=particle_u(i);
            F(p)=(V(p)-old_particle_u)*particles.mass(p)/dt;
        }
    }
}


template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Add_Velocity_Dependent_Forces_Explicit_Part(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    if(use_shear_dependent_variable_viscosity){
        INDIRECT_ARRAY<ARRAY_VIEW<TV> > V_indir_array(V,solver.velocity_samples);
        //INDIRECT_ARRAY<ARRAY_VIEW<TV> > exp_F_indir_array(particles.elasticity_force,solver.velocity_samples);  ////dbg
        INDIRECT_ARRAY<ARRAY_VIEW<TV> > F_indir_array(particles.F,solver.velocity_samples);  ////for cfl
        //for(int i=1;i<=solver.velocity_samples.m;i++){
        //    //TV f_over_m=(T).0*(grad_uT(i).Transposed()+grad_uT(i))*grad_vis(i);
        //    TV f_over_m=grad_uT(i).Transposed()*grad_vis(i);
        //    V_indir_array(i)+=f_over_m*dt;
        //    exp_F_indir_array(i)=f_over_m*particles.mass(solver.velocity_samples(i));   ////dbg
        //    //exp_F_indir_array(i)=grad_vis(i);   ////dbg
        //    F_indir_array(i)+=f_over_m*particles.mass(solver.velocity_samples(i));
        //}
        ////LOG::cout<<"grad_vis: "<<grad_vis<<std::endl;

        for(int d=1;d<=3;d++){
            VECTOR_ND<T> neg_div_mu_grad_uT_component(solver.pressure_samples.m);
            ARRAY<T> exp_F(particles.X.m);exp_F.Fill((T)0);ARRAY_VIEW<T> exp_F_arrayview(exp_F);
            VECTOR_ND<T> neg_mu_grad_uT_component(solver.velocity_samples.m*TV::m);
            //neg_mu_grad_uT_component.Fill((T)1);    ////dbg
            for(int i=1;i<=solver.velocity_samples.m;i++){
                neg_mu_grad_uT_component((i-1)*TV::m+1)=grad_uT(i)(d,1)*variable_kinematic_viscosity(i);
                neg_mu_grad_uT_component((i-1)*TV::m+2)=grad_uT(i)(d,2)*variable_kinematic_viscosity(i);
                neg_mu_grad_uT_component((i-1)*TV::m+3)=grad_uT(i)(d,3)*variable_kinematic_viscosity(i);
            }
            solver.GT.Times(neg_mu_grad_uT_component,neg_div_mu_grad_uT_component);
            Interpolate_Pressure_Sample_To_Particle_Attribute(neg_div_mu_grad_uT_component,exp_F_arrayview);

            for(int i=1;i<=solver.velocity_samples.m;i++){
                T f_over_m=-exp_F(solver.velocity_samples(i))*solver.M_inv((i-1)*TV::m+1);
                V_indir_array(i)(d)+=f_over_m*dt;
                //exp_F_indir_array(i)(d)=f_over_m*particles.mass(solver.velocity_samples(i));   ////dbg
                F_indir_array(i)(d)+=f_over_m*particles.mass(solver.velocity_samples(i));
            }
        }
    }
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOSITY_FORCE<TV>::Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time)
{   
    for(int d=1;d<=TV::m;d++){
        Reset_Solver();
        Update_G(psi_Ns_on_mesh(d),psi_Ds_on_mesh(d));
        //if(use_shear_dependent_variable_viscosity){Update_Variable_Viscosity(variable_kinematic_viscosity);}
        Update_M_inv();
        Update_A();
        
        VECTOR_ND<T> vol(solver.pressure_samples.m);
        for(int i=1;i<=solver.pressure_samples.m;i++){
            int p=solver.pressure_samples(i).index;
            int type=solver.pressure_samples(i).type;
            //vol(i)=codimensional_mesh_object.Element_Volume(p,type);
            vol(i)=(use_particle_thickness?codimensional_mesh_object.Element_Volume(p,type):codimensional_mesh_object.Element_Volume_With_Thickness(p,type,thickness_scale));

            ////dbg
            //if(!(vol(i)>(T)0)){
            //    LOG::cout<<"neg vol!!!!!!!!! "<<i<<", "<<p<<", "<<type<<", "<<vol(i)<<std::endl;
            //    if(type==SIMPLEX_TRIANGLE){
            //        TV_INT3 v=triangle_mesh.elements(p);
            //        TV_INT3 guid=Element(particles.guid,v);
            //        LOG::cout<<"neg vol v: "<<guid<<std::endl;
            //    }
            //}
        }

        VECTOR_ND<T> u_old(solver.pressure_samples.m);

        //Update_Dirichlet_Boundary_Samples(psi_Ds_on_mesh(d),solver);
        Interpolate_Particle_Attribute_To_Pressure_Sample(particles.V,d,u_old);
        
        //if(solver.pressure_samples.m>14644)LOG::cout<<"u_old: "<<u_old(14644)<<std::endl;

        //////dbg
        //u_old.Fill((T)1);

        SPARSE_MATRIX_FLAT_NXN<T>& B=solver.A;
        VECTOR_ND<T>& b=solver.b;
        b.Resize(solver.pressure_samples.m);

        for(int i=1;i<=solver.pressure_samples.m;i++){
            T coef=vol(i)/dt;
            B(i,i)+=coef;b(i)=coef*u_old(i);
        }

        solver.Modify_b_With_Dirichlet_Boundary_Conditions(b,solver.psi_D);
        //LOG::cout<<"vis boundary d "<<d<<" psi_D: "<<solver.psi_D.m<<", psi_N: "<<solver.psi_N.m<<std::endl;
        
        //////dbg
        //if(solver.pressure_samples.m>14644){
        //    LOG::cout<<"B: "<<std::endl;
        //    for(int i=1;i<=B.A.m;i++){
        //        if(B(14644,i)!=(T)0)LOG::cout<<i
        //    }
        //    LOG::cout<<"b: "<<b(14644)<<std::endl;
        //}

        solver.Solve();
        VECTOR_ND<T>& u_new=solver.p;

        //////dbg
        //for(int i=1;i<=V.m;i++)V(i)=TV::Constant_Vector((T)1);

        Update_Velocity_And_Force_Component_FLIP(V,F,u_old,u_new,d,dt);

        //////dbg
        //for(int i=1;i<=solver.pressure_samples.m;i++){
        //    int p=solver.pressure_samples(i).index;
        //    int type=solver.pressure_samples(i).type;

        //    if(type==SIMPLEX_SEGMENT){
        //        TV_INT2 v=segment_mesh.elements(p);
        //        TV_INT2 guid=Element(particles.guid,v);

        //        ////dbg
        //        if(!guid.Contains(39800)&&!guid.Contains(45795)&&!guid.Contains(47182)&&!guid.Contains(37884)&&!guid.Contains(18676)){
        //            continue;
        //        }
        //        else{
        //            LOG::cout<<"pressure "<<d<<", "<<guid<<", u_old: "<<u_old(i)<<", u_new: "<<u_new(i)<<std::endl;
        //        }
        //    }
        //}

        //////dbg
        //for(int i=1;i<=particles.X.m;i++){
        //    ////dbg
        //    if(particles.guid(i)!=(39800)&&particles.guid(i)!=(45795)&&particles.guid(i)!=(47182)&&particles.guid(i)!=(37884)&&particles.guid(i)!=(18676)){
        //        continue;
        //    }
        //    else{
        //        LOG::cout<<"velocity "<<d<<", "<<particles.guid(i)<<", V: "<<V(i)(d)<<", F: "<<F(i)(d)<<std::endl;
        //    }
        //}
    }

    if(mpi_codim){
        ARRAY_VIEW<TV> V_array_view(V);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(V_array_view);
        ARRAY_VIEW<TV> F_array_view(F);mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F_array_view);
    }
}

//template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<float,2> >;
template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<double,2> >;
template class CODIMENSIONAL_VISCOSITY_FORCE<VECTOR<double,3> >; 
#endif

