    //#####################################################################
// Copyright 2013, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_VISCOSITY_FORCE__
#define __CODIMENSIONAL_VISCOSITY_FORCE__
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/TETRAHEDRON_MESH.h>
#include "../Meshing/CODIMENSIONAL_MESH_OBJECT_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM.h"
#include "../Mpi/MPI_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_VISCOSITY_FORCE
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef MATRIX<T,TV::m> T_MATRIX;typedef DIAGONAL_MATRIX<T,TV::m> T_DIAGONAL_MATRIX;typedef SYMMETRIC_MATRIX<T,TV::m> T_SYMMETRIC_MATRIX;
public:
    CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles;
    SEGMENT_MESH& segment_mesh;
    TRIANGLE_MESH& triangle_mesh;
    TETRAHEDRON_MESH& tetrahedron_mesh;
    T kinematic_viscosity;
    T density;
    T length_scale;
    T length_scale_min;
    T thickness_scale;
    VECTOR<ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >,TV::m> psi_Ns_on_mesh;
    VECTOR<ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >,TV::m> psi_Ds_on_mesh;
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_Ns_on_surface;    
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> > psi_Ns_on_codim_boundary;    ////for const extrapolation

    VECTOR<ARRAY<int>,4> element_index_in_pressure_samples;
    ARRAY<int> particle_index_in_velocity_samples;
    CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV> solver;

    ////variable viscosity
    enum VARIABLE_VISCOSITY_MODE{SHEAR_THINNING_LINEAR=1,BINGHAM_PLASTIC_DUAL_2,CARREAU_YASUDA_SHEAR_THINNING,CARREAU_YASUDA_SHEAR_THICKENING,BINGHAM_PLASTIC_DUAL} variable_viscosity_mode;
    T kinematic_viscosity_min;
    T rigid_viscosity;
    T shear_rate_max;
    T bp_alpha;
    T lambda_thinning;
    T alpha_thinning;
    T n_thinning;
    T lambda_thickening;
    T alpha_thickening;
    T n_thickening;
    T tetrahedron_viscosity_factor;
    T triangle_viscosity_factor;
    T segment_viscosity_factor;
    T viscosity_min_clamp;
    T viscosity_max_clamp;
    T pic_flip_weight;

    bool use_particle_thickness;
    bool use_constant_extrapolation;
    bool use_staggered_dof;
    bool use_shear_dependent_variable_viscosity;
    bool use_thickness_dependent_variable_viscosity;
    bool use_variable_viscosity_for_codim;
    bool use_temperature_dependent_variable_viscosity;

    ARRAY<T> variable_kinematic_viscosity;
    ARRAY<T_MATRIX> grad_uT;
    ARRAY<TV> grad_vis;

    MPI_CODIMENSIONAL<TV>* mpi_codim;

    CODIMENSIONAL_VISCOSITY_FORCE(CODIMENSIONAL_MESH_OBJECT<TV>& codimensional_mesh_object_input,MPI_CODIMENSIONAL<TV>* mpi_codim_input=0);

    virtual void Update_Position_Based_State(const T time,const bool is_position_update);
    virtual void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV>& F,const T time);
    virtual void Add_Velocity_Dependent_Forces_Explicit_Part(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);
    virtual void Add_Velocity_Dependent_Forces(ARRAY_VIEW<TV>& V,ARRAY_VIEW<TV>& F,const T dt,const T time);

    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input){mpi_codim=mpi_codim_input;solver.mpi_codim=mpi_codim;}
    void Use_Constant_Extrapolation(bool use_constant_extrapolation_input=true)
    {use_constant_extrapolation=use_constant_extrapolation_input;solver.use_constant_extrapolation=use_constant_extrapolation;}
    void Use_Staggered_DoF(bool use_staggered_dof_input=true)
    {use_staggered_dof=use_staggered_dof_input;solver.use_all_pressure_sample_on_nodes=!use_staggered_dof;}

    ////variable viscosity
    virtual void Update_Variable_Viscosity();

    void Update_Shear_Rate();
    void Update_Shear_Rate2();
    
    T Carreau_Yasuda_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T lambda,const T alpha,const T n);
    T Linear_Viscosity(const T shear_rate,const T mu_0,const T mu_inf,const T max_shear_rate);
    T Bingham_Viscosity_Dual(const T shear_rate,T yield_shear_rate,const T rigid_viscosity,const T plastic_viscosity,const T lambda);
    T Bingham_Viscosity_Dual2(const T shear_rate,T yield_shear_stress,const T rigid_viscosity,const T plastic_viscosity);
    T Temperature_Dependent_Viscosity_Coefficient(const T temperature,const T max_temperature);

protected:
    virtual void Reset_Solver();
    virtual void Update_G(ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N,ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_Ds);
    virtual void Update_M_inv();
    virtual void Update_A();

    ////helper functions
    void Interpolate_Particle_Attribute_To_Pressure_Sample(ARRAY_VIEW<TV>& V,const int d,/*rst*/VECTOR_ND<T>& u);
    void Interpolate_Particle_Attribute_To_Pressure_Sample(ARRAY_VIEW<T>& V,/*rst*/VECTOR_ND<T>& u);
    void Interpolate_Pressure_Sample_To_Particle_Attribute(VECTOR_ND<T>& p,/*rst*/ARRAY_VIEW<T>& v);
    void Update_Dirichlet_Boundary_Samples(const ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >& psi_D,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver);
    void Update_Neumann_Boundary_Samples(const ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& psi_N,const HASHTABLE<int>& invalid_indices,CODIMENSIONAL_INCOMPRESSIBLE_SUBSYSTEM<TV>& solver);
    void Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<TV>& V,/*rst*/ARRAY_VIEW<TV>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const int d,const T dt);
    void Update_Velocity_And_Force_Component_FLIP(/*rst*/ARRAY_VIEW<T>& V,/*rst*/ARRAY_VIEW<T>& F,const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,const T dt);
    void Update_Velocity_And_Force_Component_FLIP_Helper(const VECTOR_ND<T>& u0,const VECTOR_ND<T>& u1,/*rst*/VECTOR_ND<T>& particle_u,/*rst*/VECTOR_ND<T>& particle_delta_u,/*rst*/VECTOR_ND<T>& particle_weight);
    void Test_Viscosity_Models();
};
}
#endif
