//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Read_Write/Grids_Uniform/READ_WRITE_GRID.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Grids_Uniform_Arrays/READ_WRITE_FACE_ARRAYS.h>
#include <PhysBAM_Tools/Read_Write/Octave/OCTAVE_OUTPUT.h>
#include <PhysBAM_Dynamics/Geometry/GENERAL_GEOMETRY_FORWARD.h>
#include <PhysBAM_Dynamics/Particles/PARTICLES_FORWARD.h>
#include <PhysBAM_Dynamics/Read_Write/Particles/READ_WRITE_PARTICLES.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_AREA.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TETRAHEDRALIZED_VOLUME.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_STRUCTURE.h>
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Forces_And_Torques/RIGIDS_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include "../Meshing/CODIMENSIONAL_MESH_GENERATOR.h"
#include "../Viscoelastic/CODIMENSIONAL_VISCOELASTIC_DRIVER.h"
#include "../Viscoelastic/CODIMENSIONAL_VISCOELASTIC_DRIVER_TESTS.h"

using namespace PhysBAM;

template<class TV>
PhysBAM::CODIMENSIONAL_VISCOELASTIC_DRIVER<TV>::CODIMENSIONAL_VISCOELASTIC_DRIVER()
    :BASE(),youngs_modulus_factor((T)1),viscosity_factor((T)1),viscosity_min_factor((T)1)
{
    Use_Rim(false);
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_DRIVER<TV>::Initialize_Solver()
{
    BASE::Initialize_Solver();
}

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_DRIVER<TV>::Update_Kinematic_Positions(const T dt,const T time)
{
    kinematic_particle_velocity_hashtable.Remove_All();
    if(has_kinematic_particles){
        if(kinematic_boxes.m>0)for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=kinematic_boxes.m;j++){
                if(kinematic_boxes(j).Inside(example.particles.X(i),(T)0)){
                    const TV& v=kinematic_box_velocities.m>0?kinematic_box_velocities(j):TV();
                    
                    ////update kinematic particle hashtable
                    kinematic_particle_velocity_hashtable.Set(i,v);
                    ////update psi_N
                    //for(int d=1;d<=TV::m;d++)example.incompressible.psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,TV::Axis_Vector(d),v(d)));
                    Clear_Forces(i);
                }
            }
        }

        ////update kinematic particles in mesher
        example.mesher.Clear_Eulerian_Particles();
        for(HASHTABLE_ITERATOR<int,TV> iter(kinematic_particle_velocity_hashtable);iter.Valid();iter.Next()){
            int p=iter.Key();int guid=example.particles.guid(p);
            example.mesher.Add_Eulerian_Particle(guid);
        }

        ////update kinematic boxes
        if(kinematic_box_velocities.m>0)for(int i=1;i<=kinematic_boxes.m;i++){
            const TV& v=kinematic_box_velocities(i);kinematic_boxes(i).min_corner+=v*dt;kinematic_boxes(i).max_corner+=v*dt;
        }
    }

    kinematic_particle_velocity_hashtable.Remove_All();
    if(test==17){
        for(int i=1;i<=example.particles.X.m;i++){
            T s=(T)(i*i)*(T).01;
            TV v=TV::Axis_Vector(2)*s;//+TV::Axis_Vector(3)*(T).5*s;
            kinematic_particle_velocity_hashtable.Set(i,v);
        }
    }
}

//template class CODIMENSIONAL_VISCOELASTIC_DRIVER<VECTOR<float,2> >;
template class CODIMENSIONAL_VISCOELASTIC_DRIVER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class CODIMENSIONAL_VISCOELASTIC_DRIVER<VECTOR<double,2> >;
template class CODIMENSIONAL_VISCOELASTIC_DRIVER<VECTOR<double,3> >;
#endif
