//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __CODIMENSIONAL_VISCOELASTIC_DRIVER__
#define __CODIMENSIONAL_VISCOELASTIC_DRIVER__
#include <PhysBAM_Tools/Grids_Uniform_Arrays/ARRAYS_ND.h>
#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Utilities/TIMER.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Standard_Tests/RIGIDS_STANDARD_TESTS.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY_COLLECTION.h>
#include <PhysBAM_Geometry/Basic_Geometry/SPHERE.h>
#include <PhysBAM_Geometry/Basic_Geometry/ELLIPSE.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include "../Solver/CODIMENSIONAL_SURFACE_TENSION_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "../Solver/CODIMENSIONAL_VISCOELASTIC_FORCE.h"
#include "../Misc/WATER_CYLINDRICAL_SOURCE.h"
#include "../Misc/WATER_ANNULAR_GAP_SOURCE.h"
#include "../Misc/WATER_SOURCE.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Solver/CODIMENSIONAL_FLUID_DRIVER.h"

namespace PhysBAM
{
template<class TV>
class CODIMENSIONAL_VISCOELASTIC_DRIVER : public CODIMENSIONAL_FLUID_DRIVER<TV>
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef CODIMENSIONAL_FLUID_DRIVER<TV> BASE;
public:
    using BASE::test;using BASE::output_directory;using BASE::model_directory;using BASE::stream_type;
    using BASE::first_frame;using BASE::last_frame;using BASE::current_frame;using BASE::restart_frame;
    using BASE::write_substep;using BASE::current_substep_frame;using BASE::last_substep_frame;using BASE::frame_rate;
    using BASE::time;using BASE::current_time;using BASE::cfl;using BASE::use_mpi;using BASE::example;
    using BASE::length_scale;using BASE::thickness_scale;using BASE::rim_thickness_scale;using BASE::gravity;
    using BASE::density;using BASE::surface_tension;using BASE::kinematic_viscosity;using BASE::elastic_modulus;
    using BASE::plastic_clamp_min;using BASE::plastic_clamp_max;using BASE::initial_mesh;using BASE::initial_X;
    using BASE::remesh;using BASE::has_gravity;using BASE::has_surface_tension;using BASE::has_viscosity;
    using BASE::use_implicit_viscosity;using BASE::has_elasticity;using BASE::has_plasticity;using BASE::has_fiber_fluid_coupling_force;using BASE::has_kinematic_particles;
    using BASE::rigid_body_collection;using BASE::rigid_standard_tests;using BASE::rigid_bodies;
    using BASE::processes_per_dimension;using BASE::global_grid_domain;using BASE::number_of_ghost_cells;using BASE::global_grid;
    using BASE::kinematic_viscosity_min;using BASE::use_implicit_elasticity;
    using BASE::has_incompressibility;using BASE::use_implicit_all;using BASE::use_explicit_all;using BASE::use_centric_gravity;
    using BASE::kinematic_particle_velocity_hashtable;
    using BASE::Use_Rim;using BASE::Clear_Forces;

    ARRAY<BOX<TV> > kinematic_boxes;
    ARRAY<TV> kinematic_box_velocities;

    ////viscoelastic solver
    T youngs_modulus_factor;
    T viscosity_factor;
    T viscosity_min_factor;

    CODIMENSIONAL_VISCOELASTIC_DRIVER();
    ~CODIMENSIONAL_VISCOELASTIC_DRIVER(){}

    virtual void Initialize_Test_Parameters();
    virtual void Initialize_Solver();
    virtual void Update_Kinematic_Positions(const T dt,const T time);
};
}
#endif