//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/CODIMENSIONAL_VISCOELASTIC_DRIVER.h"

template<class TV>
void PhysBAM::CODIMENSIONAL_VISCOELASTIC_DRIVER<TV>::Initialize_Test_Parameters()
{
    switch(test){
    case 5: ////box falling onto ground
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)3);
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            T height=(T).6*box_edge_length.y;box_min.y+=height;box_max.y+=height;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)1000;
            use_implicit_viscosity=true;
            has_surface_tension=false;
            has_elasticity=true;
            use_implicit_elasticity=false;
            elastic_modulus=(T)100000;
            has_plasticity=false;
            plastic_clamp_min=(T).8;
            plastic_clamp_max=(T)1.2;
            remesh=false;
            has_incompressibility=true;

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 6: ////sphere falling onto ground
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-3;
            T radius=length_scale*(T)3;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,TV::Axis_Vector(2)*radius*(T)1.2,radius,length_scale);

            //T length=radius*(T)5;
            //TV e1=TV::Axis_Vector(2)*radius*(T)1.2-TV::Axis_Vector(1)*(T).5*length;TV e2=e1+TV::Axis_Vector(1)*length;
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)33000;
            use_implicit_viscosity=true;
            has_surface_tension=false;
            has_elasticity=false;
            has_plasticity=false;
            remesh=true;
            elastic_modulus=(T).1;

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).5));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.25;rigid_bodies.Append(rigid_body);
        }break;
    case 7: ////cylinder with two ends fixed
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)2e-3;
            T radius=length_scale*(T)1;T length=length_scale*(T)10;
            TV e1=TV::Axis_Vector(1)*(T)-.5*length;e1.y+=(T)5*radius;
            TV e2=TV::Axis_Vector(1)*(T).5*length;e2.y+=(T)5*radius;
            LOG::cout<<"e1e2: "<<e1<<", "<<e2<<std::endl;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)33000;
            use_implicit_viscosity=true;
            has_surface_tension=true;
            has_elasticity=true;
            has_plasticity=false;
            remesh=true;
            elastic_modulus=(T)10;

            has_kinematic_particles=true;
            TV box1_min=e1-TV::Constant_Vector(radius*(T)1.1);box1_min.x=e1.x-radius*(T).2;TV box1_max=e1+TV::Constant_Vector(radius*(T)1.1);box1_max.x=e1.x+radius*(T).2;
            TV box2_min=e2-TV::Constant_Vector(radius*(T)1.1);box2_min.x=e2.x-radius*(T).2;TV box2_max=e2+TV::Constant_Vector(radius*(T)1.1);box2_max.x=e2.x+radius*(T).2;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));kinematic_boxes.Append(BOX<TV>(box2_min,box2_max));

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 8: ////cylinder with one end fixed
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)2e-3;
            T radius=length_scale*(T)2;T length=length_scale*(T)10;
            TV e1=TV::Axis_Vector(1)*(T)-.5*length;e1.y+=(T)5*radius;
            TV e2=TV::Axis_Vector(1)*(T).5*length;e2.y+=(T)5*radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)33;
            use_implicit_viscosity=true;
            has_surface_tension=false;
            has_elasticity=true;
            has_plasticity=true;
            remesh=false;
            elastic_modulus=(T)100;elastic_modulus*=youngs_modulus_factor;

            has_kinematic_particles=true;
            TV box1_min=e1-TV::Constant_Vector(radius*(T)1.1);box1_min.x=e1.x-radius*(T).2;TV box1_max=e1+TV::Constant_Vector(radius*(T)1.1);box1_max.x=e1.x+radius*(T).2;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 9: ////cylinder hanging with one end fixed
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)2e-3;
            T radius=length_scale*(T)2;T length=length_scale*(T)10;
            TV e1=TV::Axis_Vector(2)*length;e1.y+=(T)5*radius;
            TV e2=TV();e2.y+=(T)5*radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)1e3;
            use_implicit_viscosity=true;
            has_surface_tension=true;
            has_elasticity=false;
            has_plasticity=false;
            use_implicit_elasticity=true;
            plastic_clamp_min=(T).8;
            plastic_clamp_max=(T)1.2;
            remesh=true;
            elastic_modulus=(T).1;

            has_kinematic_particles=true;
            TV box1_min=e1-TV::Constant_Vector(radius*(T)1.1);box1_min.y=e1.y-radius*(T).2;TV box1_max=e1+TV::Constant_Vector(radius*(T)1.1);box1_max.y=e1.y+radius*(T).2;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 10:    ////elastic box under centric gravity
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)3);
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)1000;
            use_implicit_viscosity=true;
            has_surface_tension=false;
            has_elasticity=true;
            use_implicit_elasticity=true;
            elastic_modulus=(T)1e5;
            has_plasticity=false;
            plastic_clamp_min=(T).8;
            plastic_clamp_max=(T)1.2;
            remesh=false;
            use_centric_gravity=true;
            has_incompressibility=false;
        }break;
    case 11:    ////elastic square under centric gravity
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            int n=5;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)(n-1));
            TV box_min=-(T).5*box_edge_length;box_min.z=(T)0;
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Centrosymmetric_Herring_Bone_Rectangle(initial_mesh,initial_X,n,n,box_min,TV(),length_scale);
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Centrosymmetric_Herring_Bone_Rectangle(initial_mesh,initial_X,n,n,box_min,TV(),length_scale,/*reversed*/true);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,n,n,box_min,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)1000;
            use_implicit_viscosity=true;
            has_surface_tension=false;
            has_elasticity=true;
            use_implicit_elasticity=true;
            elastic_modulus=(T)1e5;
            has_plasticity=false;
            plastic_clamp_min=(T).8;
            plastic_clamp_max=(T)1.2;
            remesh=false;
            use_centric_gravity=true;
            has_incompressibility=false;
        }break;
    case 12:    ////shear thinning
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)2e-3;
            T radius=length_scale*(T)2;T length=length_scale*(T)5;
            TV e1=TV::Axis_Vector(2)*length;e1.y+=(T).15*radius;
            TV e2=TV();e2.y+=(T).15*radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)1e3;
            //kinematic_viscosity_min*=(T)1e2;
            use_implicit_viscosity=true;
            has_gravity=true;
            has_surface_tension=true;
            has_elasticity=false;
            has_plasticity=false;
            remesh=true;

            //has_kinematic_particles=true;T box_half_length=radius;
            //TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            //TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            //kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T).2);

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 13:    ////viscous sheet
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)1);box_edge_length.x*=(T)20;box_edge_length.y*=(T)5;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            T height=(T)10*box_edge_length.y;box_min.y+=height;box_max.y+=height;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)1e3;
            //kinematic_viscosity_min*=(T)1e2;
            use_implicit_viscosity=true;
            has_gravity=true;
            has_surface_tension=false;
            has_elasticity=false;
            has_plasticity=false;
            remesh=true;

            has_kinematic_particles=true;T box_half_length=box_edge_length.Max()*(T).5;
            TV e1=(T).5*(box_min+box_max);e1.y=box_max.y;
            TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 14:    ////test elasticity
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)1);box_edge_length.x*=(T)5;box_edge_length.y*=(T)10;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            T height=(T)10*box_edge_length.y;box_min.y+=height;box_max.y+=height;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)1e3;
            use_implicit_viscosity=true;
            has_gravity=false;
            has_surface_tension=false;
            has_incompressibility=false;
            has_elasticity=true;
            has_plasticity=false;
            remesh=false;

            has_kinematic_particles=true;T box_half_length=box_edge_length.Max()*(T).5;
            TV e1=(T).5*(box_min+box_max);e1.y=box_max.y;
            TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T).2);
            TV e2=(T).5*(box_min+box_max);e2.y=box_min.y;
            TV box2_min=e2-TV::Constant_Vector(box_half_length*(T)1.1);box2_min.y=e2.y-box_half_length*(T).1;
            TV box2_max=e2+TV::Constant_Vector(box_half_length*(T)1.1);box2_max.y=e2.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box2_min,box2_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T)-.2);

            RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector((T).1));
            rigid_body->X()=TV();rigid_body->X().y=(T)-.05;rigid_bodies.Append(rigid_body);
        }break;
    case 15:    ////testing strain visualization
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)1);box_edge_length.x*=(T)5;box_edge_length.y*=(T)10;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            T height=(T)10*box_edge_length.y;box_min.y+=height;box_max.y+=height;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

            has_viscosity=false;
            kinematic_viscosity*=(T)1e3;
            use_implicit_viscosity=true;
            has_gravity=false;
            has_surface_tension=false;
            has_incompressibility=true;
            has_elasticity=true;
            has_plasticity=false;
            remesh=true;

            has_kinematic_particles=true;T box_half_length=box_edge_length.Max()*(T).5;
            TV e1=(T).5*(box_min+box_max);e1.y=box_max.y;
            TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T).2);
            TV e2=(T).5*(box_min+box_max);e2.y=box_min.y;
            TV box2_min=e2-TV::Constant_Vector(box_half_length*(T)1.1);box2_min.y=e2.y-box_half_length*(T).1;
            TV box2_max=e2+TV::Constant_Vector(box_half_length*(T)1.1);box2_max.y=e2.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box2_min,box2_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T)-.2);
        }break;
    case 16:    ////stretching thin sheet
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            int m=11;int n=11;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)(n-1));box_edge_length.y=length_scale*(T)(m-1);
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Centrosymmetric_Herring_Bone_Rectangle(initial_mesh,initial_X,m,n,box_min,TV(),length_scale,true);

            has_viscosity=true;
            kinematic_viscosity*=(T)1e3;
            use_implicit_viscosity=true;
            has_gravity=true;
            has_surface_tension=false;
            has_incompressibility=false;
            has_elasticity=true;
            has_plasticity=false;
            remesh=false;

            has_kinematic_particles=true;T box_half_length=box_edge_length.Max()*(T).5;
            TV e1=(T).5*(box_min+box_max);e1.y=box_max.y;
            TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));//kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T).05);
            //TV e2=(T).5*(box_min+box_max);e2.y=box_min.y;
            //TV box2_min=e2-TV::Constant_Vector(box_half_length*(T)1.1);box2_min.y=e2.y-box_half_length*(T).1;
            //TV box2_max=e2+TV::Constant_Vector(box_half_length*(T)1.1);box2_max.y=e2.y+box_half_length*(T).1;
            //kinematic_boxes.Append(BOX<TV>(box2_min,box2_max));kinematic_box_velocities.Append(TV::Axis_Vector(2)*(T)-.05);
        }break;
    case 17:    ////testing strain visualization: 
        {
            global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
            length_scale=(T)1e-2;
            int n=10;
            TV box_edge_length=TV::Constant_Vector(length_scale*(T)n);box_edge_length.y=length_scale*(T)n;
            TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,box_min,box_min+TV::Axis_Vector(1)*box_edge_length,length_scale);

            has_viscosity=true;
            kinematic_viscosity*=(T)1e3;
            use_implicit_viscosity=true;
            has_gravity=false;
            has_surface_tension=false;
            has_incompressibility=false;
            has_elasticity=true;
            has_plasticity=false;
            remesh=false;

            //has_kinematic_particles=true;T box_half_length=box_edge_length.Max()*(T).5;
            //TV e1=(T).5*(box_min+box_max);e1.y=box_max.y;
            //TV box1_min=e1-TV::Constant_Vector(box_half_length*(T)1.1);box1_min.y=e1.y-box_half_length*(T).1;
            //TV box1_max=e1+TV::Constant_Vector(box_half_length*(T)1.1);box1_max.y=e1.y+box_half_length*(T).1;
            //kinematic_boxes.Append(BOX<TV>(box1_min,box1_max));kinematic_box_velocities.Append(TV::Axis_Vector(1)*(T).05);
        }break;
    default:
        {
            PHYSBAM_FATAL_ERROR("invalid test");
        }break;
    }

    ////rescale parameters based on arguments
    kinematic_viscosity*=viscosity_factor;kinematic_viscosity_min*=viscosity_min_factor;if(viscosity_factor==(T)0)has_viscosity=false;
    elastic_modulus*=youngs_modulus_factor;if(youngs_modulus_factor==(T)0)has_elasticity=false;

    if(use_implicit_all){use_implicit_viscosity=true;use_implicit_elasticity=true;}
    else if(use_explicit_all){use_implicit_viscosity=false;use_implicit_elasticity=false;}
}

