//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __FIBER__
#define __FIBER__

#include <PhysBAM_Tools/Grids_Uniform/GRID.h>
#include <PhysBAM_Tools/Vectors/TWIST.h>
#include <PhysBAM_Tools/Matrices/FRAME.h>
#include <PhysBAM_Tools/Matrices/QUATERNION.h>
#include <PhysBAM_Tools/Random_Numbers/RANDOM_NUMBERS.h>
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Geometry/Basic_Geometry/BOX.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE.h>
#include <PhysBAM_Geometry/Geometry_Particles/RIGID_GEOMETRY_PARTICLES.h>
#include <PhysBAM_Geometry/Solids_Geometry/DEFORMABLE_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Solids_Geometry/RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Solids_Geometry_Evolution/RIGID_GEOMETRY_EXAMPLE_VELOCITIES.h>
#include <PhysBAM_Geometry/Solids_Geometry/RIGID_GEOMETRY.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Implicit_Objects/IMPLICIT_OBJECT.h>
#include <PhysBAM_Geometry/Implicit_Objects/IMPLICIT_OBJECT_TRANSFORMED.h>
#include <PhysBAM_Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <PhysBAM_Geometry/Basic_Geometry/TRIANGLE_3D.h>
#include <PhysBAM_Geometry/Standard_Tests/RIGIDS_GEOMETRY_STANDARD_TESTS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Particles/PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Deformable_Objects/DEFORMABLES_PARAMETERS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/DEFORMABLES_EXAMPLE.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/DEFORMABLES_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/SEGMENT_BENDING_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/TRIANGLE_BENDING_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Deformables_Evolution/DEFORMABLES_EVOLUTION.h>
#include "../Meshing/CODIMENSIONAL_MESH.h"
#include "../Meshing/CODIMENSIONAL_MESHING_3D.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE.h"
#include "../Viscoelastic/FIBER_EVOLUTION.h"
#include "../Intersection/CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION.h"

namespace PhysBAM
{
template<class TV> class IMPLICIT_FIBER_GEOMETRY
{public:
    IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >* implicit_object_transformed;
    FRAME<TV>* frame;
    IMPLICIT_FIBER_GEOMETRY(IMPLICIT_OBJECT_TRANSFORMED<TV,FRAME<TV> >* implicit_object_transformed_input=0,FRAME<TV>* frame_input=0):implicit_object_transformed(implicit_object_transformed_input),frame(frame_input){}
};

template<class TV> class PRESCRIBED_MOTION
{
    typedef typename TV::SCALAR T;
public:
    int motion_bspline_order;
    ARRAY<T> motion_t;
    ARRAY<TV> motion_X;
    ARRAY<QUATERNION<T> > motion_Q;
    BSPLINE<T,TV>* motion_X_bspline;
    BSPLINE<T,QUATERNION<T> >* motion_Q_bspline;

    PRESCRIBED_MOTION<TV>():motion_bspline_order(4){}
    void Initialize_Prescribed_Motion();
    void Add_Prescribed_Motion_Key_Frame(const T t,const TV& x,const ROTATION<TV>& r);
    void Initialize_Prescribed_Motion_From_File_With_Constant_Rotation(const std::string file_name,const ROTATION<TV>& r,const T s0);
    FRAME<TV> Prescribed_Frame(const T time);
    TWIST<TV> Prescribed_Twist(const T dt,const T time);
};

template<class TV>
class FIBER
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; 
    typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef SEGMENT_MESH T_MESH;typedef typename IF<TV::m==2,SEGMENTED_CURVE_2D<T>,SEGMENTED_CURVE<TV> >::TYPE T_GEOMETRY;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TRIANGLE_3D<T> >::TYPE T_TRIANGLE;

public:
    PARTICLES<TV> fiber_particles;
    T_MESH fiber_mesh;
    T_GEOMETRY fiber_geometry;
    T length_scale;

    ////setting kinematic particle positions and velocities
    class FIBER_DEFORMABLES_EXAMPLE : public DEFORMABLES_EXAMPLE<TV>{typedef typename TV::SCALAR T;typedef DEFORMABLES_EXAMPLE<TV> BASE;public:
        using BASE::deformables_parameters;using BASE::deformable_body_collection;using BASE::collision_body_list;using BASE::rigid_geometry_collection;
        using RIGID_GEOMETRY_EXAMPLE_VELOCITIES<TV>::Set_External_Velocities;using RIGID_GEOMETRY_EXAMPLE_VELOCITIES<TV>::Set_External_Positions;
        FIBER<TV>* fiber;RIGIDS_GEOMETRY_STANDARD_TESTS<TV> rigid_standard_tests;
        FIBER_DEFORMABLES_EXAMPLE(FIBER<TV>* fiber_input):BASE(STREAM_TYPE(T()),fiber_input->fiber_particles),fiber(fiber_input),rigid_standard_tests(*this,rigid_geometry_collection){}
        virtual void Set_External_Velocities(ARRAY_VIEW<TV> V,const T velocity_time,const T current_position_time){fiber->Update_Kinematic_Velocities(V);}
        virtual void Set_External_Positions(ARRAY_VIEW<TV> X,const T time){fiber->Update_Kinematic_Positions(X);}};
    FIBER_DEFORMABLES_EXAMPLE fiber_deformables_example;

    DEFORMABLE_BODY_COLLECTION<TV>& deformable_body_collection;
    DEFORMABLE_GEOMETRY_COLLECTION<TV>& deformable_geometry_collection;
    RIGID_GEOMETRY_COLLECTION<TV>& rigid_geometry_collection;
    COLLISION_GEOMETRY_COLLECTION<TV>& collision_body_list;
    DEFORMABLES_PARAMETERS<TV>& deformables_parameters;

    ////object attributes
    FRAME<TV> object_frame;     ////translation and rotation
    TWIST<TV> object_twist;     ////linear and angular velocity
    ARRAY<PAIR<int,TV> > kinematic_particle_object_space_positions;
    ARRAY<int> fiber_start_indices; ////start index for each fiber
    ARRAY<int> fiber_end_indices;   ////end index for each fiber
    TV handle_center;
    ARRAY<RIGID_BODY<TV>* > handle_rigid_bodies;     ////brush handle
    enum PAINT_TUBE_RIGID_BODY_TYPE{NONE=0,PLACEHOLDER,RENDERABLE} rigid_body_type;

    HASHTABLE<int> near_end_segment_hashtable;
    BOX<TV> bounding_box;

    ////fiber attributes
    T fiber_radius; ////constant radius
    T fiber_length;
    T fiber_distance;
    T density;
    T fiber_bending_start_frac;
    TV fiber_bending_perturb_frac;
    ARRAY<TV> fiber_particle_X_old;
    ARRAY_VIEW<T> fiber_particle_thickness;
    ARRAY_VIEW<TV> fiber_particle_forces;

    ////collision detection
    HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> > implicit_fiber_geometry_hashtable;
    int implicit_fiber_segment_number;
    ARRAY<int> implicit_fiber_elements;
    ARRAY<RANGE<TV_INT> > fiber_element_voxel_ranges;
    HASHTABLE<int> fiber_trace_segment_hashtable;

    ////simulation attributes
    enum FIBER_DEFORMABLE_MODEL{RIGID,MASS_SPRING,READ_FROM_FILE} fiber_deformable_model;    ////simulation model
    FIBER_EVOLUTION<TV> fiber_evolution;
    bool use_gravity;
    T linear_stiffness_coefficient;
    T linear_damping_coefficient;
    T bending_stiffness_coefficient;
    T bending_damping_coefficient;

    ////guided fibers
    bool use_guided_fibers;
    enum FIBER_SHAPE{GUIDE_CYLINDER,GUIDE_BOX};
    PARTICLES<TV> guided_fiber_particles;
    T_MESH guided_fiber_mesh;
    DEFORMABLE_GEOMETRY_COLLECTION<TV> guided_deformable_geometry_collection;
    T_GEOMETRY guided_fiber_geometry;
    T guided_fiber_radius;
    T guided_fiber_distance;
    ARRAY<int> guided_fiber_start_indices;
    ARRAY<int> guided_fiber_end_indices;
    ARRAY<ARRAY<int> > guided_p_indices;
    ARRAY<ARRAY<T> > guided_p_weights;
    HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> > implicit_guided_fiber_geometry_hashtable;
    ARRAY<int> implicit_guided_fiber_elements;

    ////spatial partition
    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>* spatial_partition;

    ////prescribed motion
    bool use_prescribed_motion;
    PRESCRIBED_MOTION<TV> prescribed_motion;

    ////misc
    RANDOM_NUMBERS<T> random;
    bool use_fiber_length_variation_on_x_axis;

    ////motion from file
    ARRAY<ARRAY<TV> > file_position_data;
    ARRAY<ARRAY<TV> > file_velocity_data;
    T file_start_time;
    T file_end_time;
    T motion_start_time;
    int file_frame_count;

    bool use_virtual_springs;
    T_MESH virtual_mesh;
    T virtual_linear_stiffness_coefficient;
    T virtual_linear_damping_coefficient;
    ARRAY<T> virtual_spring_restlength;
    int virtual_spring_force_id;
    LINEAR_SPRINGS<TV>* virtual_linear_springs;
    
    ////mpi
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    FIBER();
    void Initialize_Structure();
    void Initialize_Forces();
    void Initialize_Mass_Spring_Forces();
    void Initialize_Geometry(const ARRAY<TV>& positions,const T length_scale_input,const T fiber_radius_input,const T fiber_length_input,const T guided_fiber_radius_input=(T)0,const T guided_fiber_distance_input=(T)0);
    void Initialize_Mass_Based_On_Geometry();
    void Initialize_Guided_Fiber_Weights();
    void Initialize_Implicit_Geometry();
    
    void Initialize_Velocity(const TV& v0);
    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input){mpi_codim=mpi_codim_input;}
    void Advance(const T dt,const T time);

    ////fiber-fluid coupling
    template<class T2> bool Interpolate_Value(const TV& position,const T range,ARRAY_VIEW<T2>& values,ARRAY<int>& indices,ARRAY<T>& weights,/*result*/T2& interpolated_value);
    template<class T2> bool Interpolate_Value(const TV& position,const T range,ARRAY_VIEW<T2>& values,/*result*/T2& interpolated_value);
    template<class T2> bool Interpolate_Value(const int p,const T range,const CODIMENSIONAL_MESH& mesh,ARRAY_VIEW<TV>& X,ARRAY_VIEW<T2>& values,/*result*/T2& interpolated_value);
    T Signed_Distance(const TV& pos) const;
    T Signed_Distance(const TV& pos,/*rst*/int& nearest_fiber_segment_index) const;
    T Signed_Distance_To_Segment(const TV& pos,const int segment_index) const;
    int Nearest_Fiber_Segment_Index(const TV& pos) const;

    void Update_Implicit_Geometry();
    void Update_Bounding_Box();

    ////guided fibers
    void Update_Guided_Fibers();
    void Store_Old_States();

    ////kinematics
    void Update_Kinematic_Velocities(ARRAY_VIEW<TV>& V);
    void Update_Kinematic_Positions(ARRAY_VIEW<TV>& X);
    void Update_Rigid_Body_Positions();

    bool Fiber_Segment_Intersect(const TV& p1,const TV& p2,const int segment_index,/*rst*/T& t) const;
    bool Fiber_Segment_Inside(const TV& pos,const int segment_index) const;
    bool Fiber_Inside(const TV& pos,/*rst*/int& segment_index) const;

    ////motion from file
    void Initialize_Attributes_From_File(const std::string& directory,const T t0,const T t1,const T motion_t0=(T)0);
    void Advance_Using_File_Data(const T& time);

    ////virtual springs
    void Span_Fibers(const T span_coef=(T)1.);
    void Use_Virtual_Springs(bool use=true);

    bool Is_End_Segment(const int seg_id);

protected:
    void Add_Fiber(PARTICLES<TV>& particles,const TV& start_pos,const TV& dir,const T length_scale,const T radius,const T fiber_length,const T& perturb_start_frac,const TV& perturb_frac,
        ARRAY<int>& start_indices,ARRAY<int>& end_indices,T_MESH& mesh,HASHTABLE<int>* fiber_near_end_segment_hashtable=0,ARRAY<PAIR<int,TV> >* kinematic_positions=0);
    void Randomize_Fiber(PARTICLES<TV>& particles,const int first,const int last,const int control_point_n,const T& perturb_start_frac,const TV& perturb);

    void Initialize_Implicit_Geometry_Helper(HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry_hashtable,ARRAY<int>& implicit_fiber_elements,
        const ARRAY<int>& end_indices,const PARTICLES<TV>& particles,const SEGMENT_MESH& segment_mesh,const FIBER_SHAPE fiber_shape,int segment_number_to_initialize=1);
    void Initialize_Trace_Segment_Helper(HASHTABLE<int>& fiber_trace_segment_hashtable,const int end_index,int segment_number_to_initialize=1);

    void Update_Implicit_Geometry_Helper(HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry_hashtable,
        const ARRAY<int>& end_indices,const PARTICLES<TV>& particles,const SEGMENT_MESH& segment_mesh,const FIBER_SHAPE fiber_shape);
    void Initialize_Mass_Based_On_Geometry_Helper(const T_MESH& mesh,PARTICLES<TV>& particles,const T radius);
    T Signed_Distance_Helper(const TV& pos,/*rst*/int& nearest_fiber_segment_index,const HASHTABLE<int,IMPLICIT_FIBER_GEOMETRY<TV> >& implicit_geometry,CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>* spatial_partition=0) const;

private:
    ////unused
    RIGID_GEOMETRY_EXAMPLE_VELOCITIES<TV> rigid_geometry_example_velocities;
    DEFORMABLES_EVOLUTION_CALLBACKS<TV> deformables_evolution_callbacks;
};
}
#endif
