//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/DEFORMABLES_EXAMPLE_FORCES_AND_VELOCITIES.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/DEFORMABLE_GRAVITY.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/LINEAR_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/SEGMENT_BENDING_SPRINGS.h>
#include "../Viscoelastic/FIBER_EVOLUTION.h"

using namespace PhysBAM;


template<class TV>
PhysBAM::FIBER_EVOLUTION<TV>::FIBER_EVOLUTION(DEFORMABLES_PARAMETERS<TV>& deformables_parameters_input,DEFORMABLE_BODY_COLLECTION<TV>& deformable_body_collection_input, 
    RIGID_GEOMETRY_COLLECTION<TV>& rigid_geometry_collection_input,T_GEOMETRY& fiber_geometry_input)
    :BASE(deformables_parameters_input,deformable_body_collection_input,rigid_geometry_collection_input),
    particles(deformable_body_collection_input.particles),fiber_mesh(fiber_geometry_input.mesh),fiber_geometry(fiber_geometry_input),use_ether_damping(false),ether_damping((T)50)
{
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Advance_One_Time_Step_Position(const T dt,const T time)
{
    BASE::Advance_One_Time_Step_Position(dt,time);
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Advance_One_Time_Step_Velocity(const T dt,const T time)
{
    BASE::Advance_One_Time_Step_Velocity(dt,time);

    if(use_ether_damping)for(int i=1;i<=particles.X.m;i++){
        TV dir=particles.V(i).Normalized();
        particles.V(i)*=((T)1-dt*ether_damping);
    }
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Advance(const T dt,const T time)
{
    bool done=false;
    T fiber_time=time;
    T target_time=time+dt;
    T cfl=(T)100;
    for(int substep=1;!done;substep++){
        T fiber_dt=FLT_MAX;
        fiber_dt=cfl*min(fiber_dt,deformable_body_collection.CFL(/*verbose_dt*/false));
        if(fiber_time+fiber_dt>=target_time){fiber_dt=target_time-fiber_time;done=true;}
        else if(fiber_time+(T)2*fiber_dt>=target_time){fiber_dt=(T).5*(target_time-fiber_time);}
        //if(fiber_dt<dt)LOG::cout<<"dt "<<fiber_dt<<": "<<dt<<std::endl;
        Advance_One_Time_Step_Position(fiber_dt,fiber_time);
        Advance_One_Time_Step_Velocity(fiber_dt,fiber_time);
        fiber_time+=fiber_dt;
    }
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Add_External_Forces(ARRAY<TV>& F,const T time)
{
    for(int k=1;k<=deformable_body_collection.deformables_forces.m;k++){
        DEFORMABLE_GRAVITY<TV>* gravity=dynamic_cast<DEFORMABLE_GRAVITY<TV>* >(deformable_body_collection.deformables_forces(k));
        if(gravity!=0){gravity->Add_Velocity_Independent_Forces(F,time);}   
    }
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Add_Explicit_Spring_Forces(ARRAY<TV>& F,const T time)
{
    for(int k=1;k<=deformable_body_collection.deformables_forces.m;k++){
        SEGMENT_BENDING_SPRINGS<TV>* bending_spring=dynamic_cast<SEGMENT_BENDING_SPRINGS<TV>* >(deformable_body_collection.deformables_forces(k));
        if(bending_spring!=0){deformable_body_collection.deformables_forces(k)->Add_Velocity_Independent_Forces(F,time);continue;}

        LINEAR_SPRINGS<TV>* linear_spring=dynamic_cast<LINEAR_SPRINGS<TV>* >(deformable_body_collection.deformables_forces(k));
        if(linear_spring!=0){deformable_body_collection.deformables_forces(k)->Add_Velocity_Independent_Forces(F,time);continue;}
    }
}

template<class TV>
void PhysBAM::FIBER_EVOLUTION<TV>::Forward_Euler_Step_Velocity(const ARRAY<TV>& F,const T dt,const T time)
{
    for(int i=1;i<=particles.X.m;i++){particles.V(i)+=dt*F(i)*particles.one_over_mass(i);}
    Set_External_Velocities(particles.V,time+dt,time);
}

//template class FIBER_EVOLUTION<VECTOR<float,2> >;
template class FIBER_EVOLUTION<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class FIBER_EVOLUTION<VECTOR<double,2> >;
template class FIBER_EVOLUTION<VECTOR<double,3> >;
#endif