//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __FIBER_EVOLUTION__
#define __FIBER_EVOLUTION__
#include <PhysBAM_Tools/Arrays/ARRAY.h>
#include <PhysBAM_Tools/Matrices/MATRIX_2X2.h>
#include <PhysBAM_Tools/Matrices/MATRIX_3X3.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_MXN.h>
#include <PhysBAM_Tools/Matrices/SPARSE_MATRIX_FLAT_NXN.h>
#include <PhysBAM_Tools/Krylov_Solvers/PCG_SPARSE.h>
#include <PhysBAM_Tools/Math_Tools/FACTORIAL.h>
#include <PhysBAM_Tools/Arrays/INDIRECT_ARRAY.h>
#include <PhysBAM_Geometry/Topology/TRIANGLE_MESH.h>
#include <PhysBAM_Geometry/Topology/SEGMENT_MESH.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/TRIANGULATED_SURFACE.h>
#include <PhysBAM_Geometry/Topology_Based_Geometry/SEGMENTED_CURVE_2D.h>
#include <PhysBAM_Geometry/Constitutive_Models/STRAIN_MEASURE.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Particles/PARTICLES.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/LINEAR_SPRINGS.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Deformables_Evolution/DEFORMABLES_EVOLUTION.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>

namespace PhysBAM
{
template<class TV>
class FIBER_EVOLUTION : public DEFORMABLES_EVOLUTION<TV>
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT;
    typedef typename IF<TV::dimension==2,SEGMENTED_CURVE_2D<T>,SEGMENTED_CURVE<TV> >::TYPE T_GEOMETRY;
    typedef DEFORMABLES_EVOLUTION<TV> BASE;
public:
    using BASE::deformable_body_collection;
    using BASE::Save_Velocity;using BASE::Save_Position;using BASE::Backward_Euler_Step_Velocity_Helper;using BASE::Backward_Euler_Step_Velocity;
    using BASE::Euler_Step_Position;

    PARTICLES<TV>& particles;   
    SEGMENT_MESH& fiber_mesh;
    T_GEOMETRY& fiber_geometry;

    bool use_ether_damping;
    T ether_damping;

    FIBER_EVOLUTION(DEFORMABLES_PARAMETERS<TV>& deformables_parameters_input,DEFORMABLE_BODY_COLLECTION<TV>& deformable_body_collection_input,
        RIGID_GEOMETRY_COLLECTION<TV>& rigid_geometry_collection_input,T_GEOMETRY& fiber_geometry_input);

    virtual void Advance_One_Time_Step_Position(const T dt,const T time);
    virtual void Advance_One_Time_Step_Velocity(const T dt,const T time);

    void Advance(const T dt,const T time);

    void Add_External_Forces(ARRAY<TV>& F,const T time);
    void Add_Explicit_Spring_Forces(ARRAY<TV>& F,const T time);
    void Forward_Euler_Step_Velocity(const ARRAY<TV>& F,const T dt,const T time);
};
}
#endif