//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <fstream>

#include "../Viscoelastic/KNIFE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::KNIFE<TV>::KNIFE()
    :motion_type(NONE)
{
    //no-op
}

template<class TV>
PhysBAM::KNIFE<TV>::~KNIFE()
{
    //no-op
}

template<class TV>
void PhysBAM::KNIFE<TV>::Set_Prescribed_Motion_For_Test()
{
    switch(motion_type){
    case SPREAD:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r;

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3));
        //ROTATION<TV> current_rotation=r;

        T current_time=(T)0;
        //T dt=(T)0.025;
        //T dt=(T)0.2;
        //T dt=(T)0.4;
        //T dt=(T)3;
        T dt=(T)1;

        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        t=TV((T).175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)-half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        t=TV((T)-.175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        t=TV((T).175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame((T)0.225,x,r);
    }break;
    case STRAIGHT_SPREAD:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r;

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3));
        //ROTATION<TV> current_rotation=r;

        T current_time=(T)0;
        //T dt=(T)0.025;
        //T dt=(T)0.2;
        //T dt=(T)0.4;
        T dt=(T)1;

        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t=TV((T).2,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

    }break;
    case TURN_SPREAD:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r;

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3));
        //ROTATION<TV> current_rotation=r;

        T current_time=(T)0;
        //T dt=(T)0.025;
        //T dt=(T)0.2;
        //T dt=(T)0.4;
        T dt=(T)1;

        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        current_rotation=ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(2))*current_rotation;
        t=TV((T)-.08,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        current_rotation=ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(2))*current_rotation;
        t=TV((T)-.08,(T)0,(T).02);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

    }break;
    case CIRCLE:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(1));
        T current_time=(T)0;
        T dt=(T)0.1;

        for(int i=1;i<=17;i++){
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            //r=r*ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(2));
            r=ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(2))*r;
            current_time+=dt;
        }
    }break;
    case SLICE:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        T dt_down=(T)0.05;
        T dt_up=(T)3*dt_down;
        TV dx=TV((T)0.01,(T)0,(T)0);
        TV up=TV::Axis_Vector(2);
        TV right=TV::Axis_Vector(1);
        T down_angle=(T)-pi/(T)8;
        T up_angle=(T)pi/(T)8;
        //T side_angle=(T)one_fourth_pi;

        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt_down;

        for(int i=1;i<=20;i++){
            //down
            r=ROTATION<TV>((i==1?down_angle:down_angle-up_angle),right)*r;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt_down;

            //up
            x+=dx;
            //r=ROTATION<TV>(side_angle,up)*r;
            //right=ROTATION<TV>(side_angle,up).Rotate(right);
            r=ROTATION<TV>(up_angle-down_angle,right)*r;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt_up;
        }
    }break;
    case STAMP:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)pi/(T)2.2,TV::Axis_Vector(3));
        T current_time=(T)0;
        T dt=(T)0.1;
        T dx=(T)0.005;

        x.y+=dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;

        x.y-=dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;

        x.y+=dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
    }break;
    case STATIONARY:{
        Add_Prescribed_Motion_Key_Frame((T)0,object_frame.t,object_frame.r);
    }break;
    case FLIP_AND_SPREAD:{
        FRAME<TV> original_frame=object_frame;
        TV x=object_frame.t;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)-half_pi,TV::Axis_Vector(3));
        T current_time=(T)0;
        T dt=(T)0.05;
        T dx=(T)-0.02;

        for(int i=1;i<=3;i++){
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt;
        }

        r=r*ROTATION<TV>((T)pi*(T)3/(T)4,TV::Axis_Vector(3));
        x.x+=dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;

        x.x+=dx;
        x.y-=(T)0.002;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;


        for(int i=1;i<=5;i++){
            x.x+=dx;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt;
        }
    }break;
    case ROTATE:{
        TV x=object_frame.t;
        ROTATION<TV> r=object_frame.r;
        T current_time=(T)0;
        T dt=(T)0.1;
        T angle=(T)pi/(T)4;

        for(int i=1;i<=20;i++){
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt;
            r=r*ROTATION<TV>(angle,TV::Axis_Vector(2));
        }

    }break;
    case TRANSLATE:{
        TV x=object_frame.t;
        ROTATION<TV> r=object_frame.r;
        T current_time=(T)0;
        T dt=(T)0.1;
        T dx=(T)0.1;

        for(int i=1;i<=20;i++){
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            current_time+=dt;
            x.x+=dx;
        }
    }break;
    case MULTIPLE_SPREAD:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r;

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3));
        //ROTATION<TV> current_rotation=r;

        T current_time=(T)0;
        //T dt=(T)0.025;
        T dt=(T)0.05;
        //T dt=(T)0.4;
        //T dt=(T)3;
        //T dt=(T)1;

        //T dx1=(T).2;
        //T dx2=(T).075;
        //T dy=(T).02;
        T dx1=(T).1;
        T dx2=(T).03;
        T dy=(T).01;


        t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        x(1)+=dx1;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        //t=TV((T).175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)+=dx2;x(2)+=dy;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)-=dx2;x(2)-=dy;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)-=dx1;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)-half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        //t=TV((T)-.175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)-=dx2;x(2)+=dy;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //t=TV((T)-.1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)+=dx2;x(2)-=dy;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)+=dx1;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_rotation=current_rotation*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));

        current_time+=dt;
        //t=TV((T).175,(T).02,(T)0);x=original_frame.t+t;r=current_rotation;
        r=current_rotation;
        x(1)+=dx2;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame((T)0.225,x,r);
    }break;
    case SWIFT_LIGHTNING_HURRICANE_TIGER_STRIKE:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        //T dt_factor=(T).1;
        T dt_factor=(T).2;
        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(3));

        T current_time=(T)0;
        //T dt=(T)0.05;
        //T dt=(T)1;

        TV original_x=original_frame.t;
        TV x=original_x;

        //start position
        TV dx=TV((T).07610,(T).01297,(T).03269);
        x=original_x+dx;
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //scoop on slice 1
        T dt=(T)1*dt_factor;
        current_time+=dt;
        dx=TV((T).00017,(T).01122,(T).03269);
        x=original_x+dx;
        r=ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3))*r;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //lift up on slice 1
        dt=(T)1*dt_factor;
        current_time+=dt;
        dx=TV((T)-.06603,(T).03002,(T)-.03269);
        x=original_x+dx;
        r=ROTATION<TV>((T)-half_pi,TV::Axis_Vector(3))*r;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //reposition
        dt=(T)1*dt_factor;
        current_time+=dt;
        //dx=TV((T)-.02334,(T).03002,(T).00461);
        dx=TV((T)-.02334,(T).02002,(T).00461);
        x=original_x+dx;
        r=ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(2))*r;
        r=r*ROTATION<TV>((T)half_pi+(T)one_fourth_pi,TV::Axis_Vector(3));
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //scoop 2
        dt=(T)1*dt_factor;
        current_time+=dt;
        //dx=TV((T)-.08956,(T)0.01161,(T)-.00230);
        dx=TV((T)-.08956,(T)0.00875,(T)-.00230);
        x=original_x+dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //scoop 2
        dt=(T)1*dt_factor;
        current_time+=dt;
        dx=TV((T)-.141185,(T).00875,(T).034845);
        x=original_x+dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //lift 2
        dt=(T)1*dt_factor;
        current_time+=dt;
        //dx=TV((T)-.19281,(T).02665,(T).06739);
        dx=TV((T)-.19281,(T).012,(T).06739);
        x=original_x+dx;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //current_time+=dt;
        //t=TV((T).1,(T)0,(T)0);x=original_frame.t+t;r=current_rotation;
        //x(1)+=dx1;r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame(current_time,x,r);


    }break;
    case DO_IT_IN_BLENDER:{

        T current_time=(T)0;

        //T dt=(T)0.05;
        T dt=(T)0.1;
        
        ARRAY<TV> x;
        {
            TV pos;
            std::string motion_data="./Motion_Data/mayo_knife.txt";
            std::ifstream fin(motion_data.c_str());
            if(fin.fail())PHYSBAM_FATAL_ERROR("could not open knife motion data file.");

            while(fin>>pos.x>>pos.y>>pos.z){
                pos.x-=(T)5;
                pos.z+=(T)5;
                pos/=(T)100;
                x.Append(pos);
            }
            fin.close();
        }

        ARRAY<TV> radians;
        {
            T x,y,z;
            std::string rotation_data="./Motion_Data/mayo_knife_rotation.txt";
            std::ifstream fin(rotation_data.c_str());
            if(fin.fail())PHYSBAM_FATAL_ERROR("could not open knife rotation data file.");

            while(fin>>x>>y>>z){
                TV degrees(x,y,z);
                radians.Append((T)pi/(T)180*degrees);
            }
            fin.close();
        }

        //ROTATION<TV> r=object_frame.r;
        //ROTATION<TV> current_rotation=r*ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(3));


        for(int i=1;i<=x.m;i++){
            //ROTATION<TV> r=object_frame.r*ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(3));
            ROTATION<TV> r=object_frame.r;
            int rotation_idx=(i<=radians.m?i:radians.m);
            //for(int d=3;d>0;d--)r=ROTATION<TV>(radians(rotation_idx)(d),TV::Axis_Vector(d))*r;
            for(int d=3;d>0;d--)r=r*ROTATION<TV>(radians(rotation_idx)(d),TV::Axis_Vector(d));
            Add_Prescribed_Motion_Key_Frame(current_time,x(i),r);
            current_time+=dt;
        }

    }break;
    case DO_IT_IN_BLENDER2_THE_RECKONING:{

        T current_time=(T)0;

        T dt=(T)0.05;
        
        ARRAY<TV> x;
        {
            TV pos;
            std::string motion_data="./Motion_Data/mayo_knife.txt";
            std::ifstream fin(motion_data.c_str());
            if(fin.fail())PHYSBAM_FATAL_ERROR("could not open knife motion data file.");

            while(fin>>pos.x>>pos.y>>pos.z){
                pos.x-=(T)5;
                pos.z+=(T)5;
                pos/=(T)100;
                x.Append(pos);
            }
            fin.close();
        }

        ARRAY<TV> v(x.m);
        {
            for(int i=1;i<=x.m;i++){
                if(i==1){
                    v(i)=(x(i+1)-x(i)).Normalized();
                }
                else if(i==x.m){
                    v(i)=(x(i)-x(i-1)).Normalized();
                }
                else{
                    v(i)=((T)0.5*(x(i+1)-x(i-1))).Normalized();
                }
            }
        }

        ROTATION<TV> r=object_frame.r;
        //ROTATION<TV> current_rotation=r*ROTATION<TV>((T)one_fourth_pi,TV::Axis_Vector(3));


        ROTATION<TV> r_about_z((T)-pi/(T)6,TV::Axis_Vector(3));
        TV initial_dir=r_about_z.Rotate((T)-1*TV::Axis_Vector(1)).Normalized();
        for(int i=1;i<=x.m;i++){
            r=ROTATION<TV>::From_Rotated_Vector(initial_dir,v(i));
            Add_Prescribed_Motion_Key_Frame(current_time,x(i),r);
            current_time+=dt;
        }

    }break;
    default:{
        LOG::cout<<"WARNING: no KNIFE motion type specified"<<std::endl;
    }break;
    }

    Initialize_Prescribed_Motion();
}

template<class TV>
void PhysBAM::KNIFE<TV>::Update_Kinematic_Velocities_And_Positions(const T& dt,const T& time)
{
    if(motion_type!=STATIONARY){
        BASE::Update_Kinematic_Velocities_And_Positions(dt,time);
    }
}

template class KNIFE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class KNIFE<VECTOR<double,3> >;
#endif
