//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/LADLE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::LADLE<TV>::LADLE()
    :motion_type(CIRCLE),random(1337)
{
    //no-op
}

template<class TV>
PhysBAM::LADLE<TV>::~LADLE()
{
    //no-op
}

template<class TV>
void PhysBAM::LADLE<TV>::Set_Prescribed_Motion_For_Test()
{
    switch(motion_type){
    case CIRCLE:{
        Add_Prescribed_Motion_Key_Frame((T)0,object_frame.t,object_frame.r);
        //TODO...
    }break;
    case ZIGZAG:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        T dt=(T)0.01;
        T dz=(T)0.01;
        T dx=(T)0.1;
        TV x=original_frame.t;

        for(int i=1;i<=10;i++){
            T sign=(i%2==1?(T)1:(T)-1);
            x(1)+=sign*(i>2&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }
    }break;
    case COMBO:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        //T dt=(T)0.01;
        //T dz=(T)0.01;
        //T dx=(T)0.1;
        TV x=original_frame.t;
        //const T k=(T)50;
        const T k=(T)20;
        T dt_factor=(T)1;

        {//descend
            T dt=(T)0.01*dt_factor;
            //T dy=(T)0.002;
            T dy=(T)0.005;
            x(2)+=dy;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            
            x(2)-=dy;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=10;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            T radius=k*current_time;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            for(int i=1;i<=10;i++){
                current_time+=dt;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.5*(x(1)-(T)0.05);
            x(2)+=dy;
            x(3)=(T)0.5*(x(3)-(T)0.05);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//zigzag
            T dt=(T)0.01*dt_factor;
            T dz=(T)0.01;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.06;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            //for(int i=1;i<=10;i++){
            for(int i=1;i<=6;i++){
                T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

    }break;
    case COMBO2:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        //T dt=(T)0.01;
        //T dz=(T)0.01;
        //T dx=(T)0.1;
        TV x=original_frame.t;
        //const T k=(T)50;
        const T k=(T)20;
        T dt_factor=(T)1;

        {//move to origin
            T dt=(T)0.01*dt_factor;
            x(1)=(T)0.03;
            x(3)=(T)0;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.015;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            
            x(1)=(T)0;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=10;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            T radius=k*current_time;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            for(int i=1;i<=10;i++){
                current_time+=dt;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.5*(x(1)-(T)0.05);
            x(2)+=dy;
            x(3)=(T)0.5*(x(3)-(T)0.05);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//zigzag
            T dt=(T)0.01*dt_factor;
            T dz=(T)0.01;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.06;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            //for(int i=1;i<=10;i++){
            for(int i=1;i<=6;i++){
                T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

    }break;
    case COMBO2_SLOWMO:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        //T dt=(T)0.01;
        //T dz=(T)0.01;
        //T dx=(T)0.1;
        TV x=original_frame.t;
        //const T k=(T)50;
        const T k=(T)10;
        T dt_factor=(T)1;

        {//move to origin
            T dt=(T)0.01*dt_factor;
            x(1)=(T)0.03;
            x(3)=(T)0;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.02;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.01;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);


            
            x(1)=(T)0;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            T radius=k*current_time;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.75*x(1)-(T)0.25*(T)0.05;
            x(2)+=dy;
            x(3)=(T)0.75*x(3)-(T)0.25*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.25*x(1)-(T)0.75*(T)0.05;
            x(2)+=dy;
            x(3)=(T)0.25*x(3)-(T)0.75*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        }

        {//zigzag
            T dt=(T)0.005*dt_factor;
            T dz=(T)0.005;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.03;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            //for(int i=1;i<=10;i++){
            T sign=(T)-1;
            for(int i=1;i<=12;i++){
                if(i%2==1)sign*=(T)-1;
                //T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

    }break;
    case COMBO2_SLOWMO_THE_RECKONING:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        T current_time=(T)0;
        TV x=original_frame.t;
        //const T k=(T)50;
        const T k=(T)10;
        T dt_factor=(T)2;

        {//move to origin
            T dt=(T)0.01*dt_factor;
            x(1)=(T)0.03;
            x(3)=(T)0;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.02;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.01;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);


            
            x(1)=(T)0;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            T radius=k*current_time;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T scale=(T)100;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.75*x(1)-(T)0.25*(T)0.05;
            x(2)+=dy;
            x(3)=(T)0.75*x(3)-(T)0.25*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.25*x(1)-(T)0.75*(T)0.05;
            x(2)+=dy;
            x(3)=(T)0.25*x(3)-(T)0.75*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        }

        {//zigzag
            T dt=(T)0.005*dt_factor;
            T dz=(T)0.005;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.03;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            //for(int i=1;i<=10;i++){
            T sign=(T)-1;
            for(int i=1;i<=12;i++){
                if(i%2==1)sign*=(T)-1;
                //T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

    }break;
    case COMBO3_THE_JOURNEY_HOME:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        r=ROTATION<TV>((T)pi/(T)4,TV::Axis_Vector(2))*r;
        T current_time=(T)0;
        TV x=original_frame.t;
        //const T k=(T)50;
        //const T k=(T)5;
        const T k=(T)2;
        T dt_factor=(T)2;

        {//move to origin
            T dt=(T)0.01*dt_factor;
            x(1)=(T)0.03;
            x(3)=(T)0;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.02;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.01;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);


            
            x(1)=(T)0;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100/dt_factor;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            T radius=k*current_time;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T scale=(T)100/dt_factor;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.75*x(1)-(T)0.25*(T)0.05;
            x(2)+=dy;
            x(3)=(T)0.75*x(3)-(T)0.25*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.25*x(1)-(T)0.75*(T)0.05;
            //x(2)+=dy;
            x(3)=(T)0.25*x(3)-(T)0.75*(T)0.05;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        }

        {//zigzag
            T dt=(T)0.005*dt_factor;
            T dz=(T)0.005;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.03;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            T sign=(T)-1;
            for(int i=1;i<=12;i++){
                if(i%2==1)sign*=(T)-1;
                //T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

    }break;
    case COMBO4_A_NEW_HOPE:{
        FRAME<TV> original_frame=object_frame;
        ROTATION<TV> r=original_frame.r;
        r=ROTATION<TV>((T)pi/(T)4,TV::Axis_Vector(2))*r;
        T current_time=(T)0;
        TV x=original_frame.t;
        //const T k=(T)50;
        //const T k=(T)5;
        const T k=(T)2;
        T dt_factor=(T)2;
        //TV direction;
        //T speed=(T)0.01;

        {//move to origin
            T dt=(T)0.01*dt_factor;
            x(1)=(T)0.03;
            x(3)=(T)0;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.02;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.01;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0;
            x(3)=(T)0;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }

        {//expanding circle
            T dt=(T)0.01*dt_factor;
            for(int i=1;i<=20;i++){
                current_time+=dt;
                T radius=k*current_time;
                T scale=(T)100/dt_factor;
                T dx=radius*(T)-1*sin(scale*current_time)*dt;
                T dz=radius*cos(scale*current_time)*dt;
                //direction(1)=dx;direction(2)=(T)0;direction(3)=dz;
                //direction.Normalize();
                //x+=speed*direction;
                x(1)+=dx;x(3)+=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }

        {//reverse direction circle
            //T radius=k*current_time;
            //const T k_decay=(T)1;
            //T dt_sum=(T)0;
            T dt=(T)0.01*dt_factor;
            T reverse_time=current_time;
            reverse_time-=(T)0.25*dt;//?
            //for(int i=1;i<=20;i++){
            for(int i=1;i<=18;i++){
                current_time+=dt;
                T radius=k*reverse_time;
                //T radius=k*reverse_time-k_decay*dt_sum;
                T scale=(T)100/dt_factor;
                //scale*=(T).5;
                T dx=radius*(T)-1*sin(scale*reverse_time)*dt;
                T dz=radius*(T)1*cos(scale*reverse_time)*dt;
                //direction(1)=dx;direction(2)=(T)0;direction(3)=dz;
                //direction.Normalize();
                //x-=speed*direction;
                x(1)-=dx;x(3)-=dz;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                reverse_time-=dt;
                //dt_sum+=dt;
            }
        }

        {//lift up and move
            T dt=(T)0.01*dt_factor;
            T dy=(T)0.01;

            x(1)=(T)0.75*x(1)-(T)0.25*(T)0.03;
            x(2)+=dy;
            x(3)=(T)0.75*x(3)-(T)0.25*(T)0.03;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            x(1)=(T)0.25*x(1)-(T)0.75*(T)0.03;
            //x(2)+=dy;
            x(3)=(T)0.25*x(3)-(T)0.75*(T)0.03;
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        }

        {//zigzag
            //T dt=(T)0.005*dt_factor;
            T dt=(T)0.02*dt_factor;
            T dz=(T)0.005;
            T dy=(T)0.01;
            //T dx=(T)0.1;
            T dx=(T)0.03;

            //x=TV((T)-0.05,x(2)-dy,(T)-0.05);
            x=TV((T)-0.03,x(2)-dy,(T)-0.03);
            current_time+=dt;
            Add_Prescribed_Motion_Key_Frame(current_time,x,r);

            T sign=(T)-1;
            for(int i=1;i<=12;i++){
                if(i%2==1)sign*=(T)-1;
                //T sign=(i%2==1?(T)1:(T)-1);
                //x(1)+=sign*(i>1&&i<10?(T)2:(T)1)*dx;x(3)+=dz;current_time+=dt;
                x(1)+=sign*dx;x(3)+=dz;current_time+=dt;
                Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }


    }break;
    default:{
        LOG::cout<<"WARNING: no LADLE motion type specified"<<std::endl;
    }break;
    }

    Initialize_Prescribed_Motion();
}

template<class TV>
FRAME<TV> PhysBAM::LADLE<TV>::Prescribed_Frame(const T time)
{
    FRAME<TV> frame;
    if(motion_type==CIRCLE){
        TV pos;
        ARRAY<T> times;
        times.Append((T)0.1);
        times.Append((T)0.2);
        times.Append((T)0.3);

        const T k=(T).5;

        LOG::cout<<"time: "<<time<<std::endl;

        if(time<times(1)){
            LOG::cout<<"case 1."<<std::endl;
            T r=k*time;
            T time_factor=(T)100;
            T x=r*cos(time_factor*time);
            T z=r*sin(time_factor*time);
            T rand_x=(T)0;
            T rand_z=(T)0;
            pos(1)=x+rand_x;
            pos(3)=z+rand_z;
        }
        else if(time<times(2)){
            T r=k*times(1);
            LOG::cout<<"case 2, r=="<<r<<std::endl;
            T time_factor=(T)-100;
            T x=r*cos(time_factor*time);
            T z=r*sin(time_factor*time);
            T rand_x=(T)0;
            T rand_z=(T)0;
            pos(1)=x+rand_x;
            pos(3)=z+rand_z;

        }
        else if(time<times(3)){

            LOG::cout<<"case 3."<<std::endl;
            T r=k*times(1);

            T time_factor=(T)100;
            T x=r*cos(time_factor*time);
            T z=r*sin(time_factor*time);
            T rand_x=(T)0;
            T rand_z=(T)0;
            pos(1)=x+rand_x;
            pos(3)=z+rand_z;
        }

        /*
        //const T k=(T).5;
        T r=k*time;
        T time_factor=(T)100;
        T x=r*cos(time_factor*time);
        T z=r*sin(time_factor*time);
        //TV pos=rigid_body->X();
        T max_perturb=(T)2e-3;
        T rand_x=(random.Get_Number()-(T)0.5)*max_perturb;
        T rand_z=(random.Get_Number()-(T)0.5)*max_perturb;
        //TV pos;
        pos(1)=x+rand_x;
        pos(3)=z+rand_z;
        */

        ROTATION<TV> rot=ROTATION<TV>::From_Quaternion(motion_Q_bspline->Clamped_Evaluate(time));
        frame.t=pos;
        frame.r=rot;
    }
    else{
        frame=BASE::Prescribed_Frame(time);
    }
    return frame;
}

template class LADLE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class LADLE<VECTOR<double,3> >;
#endif
