//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/OBJECT.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::OBJECT<TV>::OBJECT()
    :motion_bspline_order(2),motion_X_bspline(0),motion_Q_bspline(0),rigid_body(0),use_slow_motion(false),slow_motion_threshold((T)0),slow_motion_factor((T)1)
{
    //no-op
}

template<class TV>
PhysBAM::OBJECT<TV>::~OBJECT()
{
    //if(rigid_body!=0)delete rigid_body;
    if(motion_Q_bspline!=0)delete motion_Q_bspline;
    if(motion_X_bspline!=0)delete motion_X_bspline;
}

template<class TV>
void PhysBAM::OBJECT<TV>::Initialize_Prescribed_Motion()
{
    motion_X_bspline=new BSPLINE<T,TV>(motion_t,motion_X,motion_bspline_order);
    motion_Q_bspline=new BSPLINE<T,QUATERNION<T> >(motion_t,motion_Q,motion_bspline_order);
}

template<class TV>
void PhysBAM::OBJECT<TV>::Add_Prescribed_Motion_Key_Frame(const T t,const TV& x,const ROTATION<TV>& r)
{
    if(motion_t.m==0){
        for(int j=1;j<=motion_bspline_order;j++){
            motion_t.Append(t);
            motion_X.Append(x);
            motion_Q.Append(r.Quaternion());
        }
    }
    else{
        for(int j=1;j<=motion_bspline_order;j++){
            motion_t.Pop();
            motion_X.Pop();
            motion_Q.Pop();
        }
    }
    motion_t.Append(t);
    motion_X.Append(x);
    motion_Q.Append(r.Quaternion());
    for(int j=1;j<=motion_bspline_order;j++){
        motion_t.Append(t);
        motion_X.Append(x);
        motion_Q.Append(r.Quaternion());
    }
}

template<class TV>
FRAME<TV> PhysBAM::OBJECT<TV>::Prescribed_Frame(const T time)
{
    T effective_time=time;
    if(use_slow_motion&&time>slow_motion_threshold){
        effective_time=slow_motion_threshold+slow_motion_factor*(time-slow_motion_threshold);
    }
    TV pos=motion_X_bspline->Clamped_Evaluate(effective_time);
    ROTATION<TV> rot=ROTATION<TV>::From_Quaternion(motion_Q_bspline->Clamped_Evaluate(effective_time));
    return FRAME<TV>(pos,rot);
}


template<class TV>
TWIST<TV> PhysBAM::OBJECT<TV>::Prescribed_Twist(const T& dt,const T& time)
{
    T effective_time=time;
    T effective_dt=dt;
    if(use_slow_motion&&time>slow_motion_threshold){
        effective_time=slow_motion_threshold+slow_motion_factor*(time-slow_motion_threshold);
        effective_dt=dt*slow_motion_factor;
    }
    FRAME<TV> frame1=Prescribed_Frame(effective_time);
    FRAME<TV> frame2=Prescribed_Frame(effective_time+effective_dt);
    return (frame2.Delta() - frame1.Delta())/effective_dt;
    //return (frame1.Inverse_Times(frame2)).Delta()/dt;
    //TWIST<TV> twist;
    //twist.linear=(frame2.t-frame1.t)/dt;
    //twist.angular=((frame1.Inverse_Times(frame2)).Delta()/dt).angular;
    //return twist;
}

template<class TV>
void PhysBAM::OBJECT<TV>::Update_Kinematic_Velocities_And_Positions(const T& dt,const T& time)
{
    object_frame=Prescribed_Frame(time);
    object_twist=Prescribed_Twist(dt,time);

    rigid_body->X()=object_frame.t;
    rigid_body->Rotation()=object_frame.r;
    rigid_body->Set_Twist(object_twist);

    /*
    //// velocity test
    //TV direction=TV::Axis_Vector(3);
    TV direction=rigid_body->Rotation().Rotate(TV::Axis_Vector(3)).Normalized();

    ARRAY<TV> x;
    x.Append((T)2*direction);
    x.Append((T)1*direction);
    x.Append(TV());
    x.Append((T)-1*direction);
    x.Append((T)-2*direction);

    ARRAY<TV> v;
    for(int i=1;i<=x.m;i++){
        v.Append(RIGID_GEOMETRY<TV>::Pointwise_Object_Velocity(rigid_body->Twist(),rigid_body->X(),x(i)));
    }

    LOG::cout<<std::endl;
    LOG::cout<<"direction: "<<direction<<std::endl;
    LOG::cout<<"Extremes: "<<v(1)<<std::endl;
    LOG::cout<<"          "<<v(5)<<std::endl;
    LOG::cout<<"Interior: "<<v(2)<<std::endl;
    LOG::cout<<"          "<<v(4)<<std::endl;
    LOG::cout<<"Center:   "<<v(3)<<std::endl;
    LOG::cout<<std::endl;
    */
}

template class OBJECT<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class OBJECT<VECTOR<double,3> >;
#endif
