//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __OBJECT_H__
#define __OBJECT_H__
#include <PhysBAM_Tools/Interpolation/BSPLINE.h>
#include <PhysBAM_Tools/Matrices/FRAME.h>
#include <PhysBAM_Tools/Matrices/QUATERNION.h>
#include <PhysBAM_Tools/Vectors/TWIST.h>
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>

namespace PhysBAM
{
template<class TV>
class OBJECT
{
    typedef typename TV::SCALAR T;
public:

    ////prescribed motion
    int motion_bspline_order;
    ARRAY<T> motion_t;
    ARRAY<TV> motion_X;
    ARRAY<QUATERNION<T> > motion_Q;
    BSPLINE<T,TV>* motion_X_bspline;
    BSPLINE<T,QUATERNION<T> >* motion_Q_bspline;

    ////object attributes
    FRAME<TV> object_frame; //translation & rotation
    TWIST<TV> object_twist; //linear & angular velocity
    RIGID_BODY<TV>* rigid_body;

    bool use_slow_motion;
    T slow_motion_threshold;
    T slow_motion_factor;

    OBJECT();
    virtual ~OBJECT();

    virtual void Initialize_Prescribed_Motion();
    virtual void Set_Prescribed_Motion_For_Test() = 0;
    virtual void Add_Prescribed_Motion_Key_Frame(const T t,const TV& x,const ROTATION<TV>& r);
    virtual FRAME<TV> Prescribed_Frame(const T time);
    virtual TWIST<TV> Prescribed_Twist(const T& dt,const T& time);
    virtual void Update_Kinematic_Velocities_And_Positions(const T& dt,const T& time);
};
}
#endif //__OBJECT_H__
