//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __PAINT_DRIVER__
#define __PAINT_DRIVER__
#include "../Solver/CODIMENSIONAL_FLUID_DRIVER.h"
#include "../Viscoelastic/FIBER.h"
#include "../Viscoelastic/KNIFE.h"
#include "../Viscoelastic/LADLE.h"
#include "../Viscoelastic/ROCK.h"
#include "../Viscoelastic/SCRAPER.h"
#include "../Viscoelastic/SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL.h"
#include "../Viscoelastic/PAINT_TUBE.h"
#include "../Levelset/LEVELSET_MULTIPLE_CODIMENSIONAL.h"

namespace PhysBAM
{
template<class TV>
class PAINT_DRIVER : public CODIMENSIONAL_FLUID_DRIVER<TV>
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; typedef typename TV::SPIN T_SPIN;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef CODIMENSIONAL_FLUID_DRIVER<TV> BASE;
public:
    using BASE::Time_At_Frame;using BASE::Use_Rim;using BASE::Update_Boundary_Conditions;using BASE::Correct_Particle_Normals;using BASE::Update_Particle_Contact;using BASE::Update_Particle_Thickness;using BASE::Enforce_Solid_Boundary_Velocities;
    using BASE::test;using BASE::output_directory;using BASE::model_directory;using BASE::stream_type;
    using BASE::first_frame;using BASE::last_frame;using BASE::current_frame;using BASE::restart_frame;
    using BASE::write_substep;using BASE::current_substep_frame;using BASE::last_substep_frame;using BASE::frame_rate;
    using BASE::time;using BASE::current_time;using BASE::cfl;using BASE::use_mpi;using BASE::example;
    using BASE::length_scale;using BASE::adaptive_length_scale_min;using BASE::adaptive_length_scale_min2;using BASE::thickness_scale;using BASE::rim_thickness_scale;using BASE::max_thickness_factor;using BASE::min_thickness_factor;
    using BASE::max_rim_thickness_factor;using BASE::min_rim_thickness_factor;using BASE::surface_tension_factor;using BASE::gravity_factor;using BASE::length_scale_min_factor;
    using BASE::scale;using BASE::gravity;using BASE::density;using BASE::surface_tension;using BASE::tri_surface_tension;using BASE::seg_surface_tension;using BASE::kinematic_viscosity;using BASE::elastic_modulus;using BASE::plastic_clamp_min;using BASE::plastic_clamp_max;using BASE::friction_magnitude;
    using BASE::initial_mesh;using BASE::initial_X;using BASE::initial_thickness;using BASE::initial_texture_coordinates;using BASE::initial_phis;using BASE::remesh;using BASE::has_gravity;using BASE::has_surface_tension;using BASE::has_viscosity;
    using BASE::use_implicit_viscosity;using BASE::has_elasticity;using BASE::has_plasticity;using BASE::use_implicit_elasticity;using BASE::has_fiber_fluid_coupling_force;using BASE::has_adhesion;using BASE::has_friction;using BASE::has_rim;
    using BASE::rigid_body_collection;using BASE::rigid_standard_tests;using BASE::rigid_bodies;using BASE::tank;using BASE::rigid_body_speed_factor;using BASE::solid_non_slip_hashtable;
    using BASE::random;using BASE::processes_per_dimension;using BASE::global_grid_domain;using BASE::number_of_ghost_cells;using BASE::global_grid;
    using BASE::kinematic_viscosity_min;using BASE::rigid_viscosity;using BASE::has_incompressibility;
    using BASE::initial_V;using BASE::use_spatial_partition;using BASE::boundary_particle_hashtable;using BASE::boundary_particle_hashtable2;using BASE::kinematic_particle_velocity_hashtable;using BASE::use_inactive_particle;
    using BASE::use_correcting_particle_position_on_solids;using BASE::use_non_slip_on_solids;using BASE::use_adaptive_length_scale_on_codim;using BASE::use_adaptive_length_scale_on_levelset;using BASE::use_adaptive_length_scale_on_rim;
    using BASE::use_adaptive_length_scale_on_fiber;using BASE::use_adaptive_length_scale_based_on_test;
    using BASE::has_rim_segments_for_viscosity;using BASE::has_rim_triangles_for_viscosity;using BASE::use_heat_transfer;using BASE::use_paint_trace_boundary;
    using BASE::use_friction_boundary;using BASE::use_viscosity_dependent_friction;using BASE::use_damped_friction;using BASE::use_correcting_rigid_normal;using BASE::time_scale;using BASE::environment_temperature;using BASE::heat_conductivity;

    FIBER<TV> fiber;
    FRAME<TV> fiber_initial_frame;
    ARRAY<TV> fiber_positions;
    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV> coupled_evolution;
    ARRAY<PAINT_TUBE<TV>* > tubes;
    KNIFE<TV>* knife;
    LADLE<TV>* ladle;
    ROCK<TV>* rock;
    SCRAPER<TV>* scraper;

    bool use_levelset;
    LEVELSET_MULTIPLE_CODIMENSIONAL<TV> levelset;
    int levelset_region_number;
    bool update_slow_interface_phis;

    T fiber_length_scale;
    T fiber_radius;
    T fiber_length;
    T fiber_distance;
    T guided_fiber_radius;
    T guided_fiber_distance;

    T viscosity_factor;
    T viscosity_min_factor;
    T elastic_modulus_factor;
    T band_width_min;
    T band_width_max;
    T band_width_min_factor;

    T min_tri_thickness_factor;
    T min_seg_thickness_factor;

    T adaptive_length_scale_min_factor;
    T adaptive_length_scale_min_factor2;

    T lambda_cy;
    T alpha_cy;
    T n_cy;

    ARRAY<RIGID_BODY<TV>* > kinematic_rigid_bodies;
    TV canvas_normal;
    
    enum PARTICLE_INACTIVE_MODE{SMALL_VELOCITY=1,FAR_FROM_FOCUS} particle_inactive_mode;
    bool use_particle_incompressibility_for_solids;
    bool use_particle_incompressibility_for_fiber;

    ////parameters for different tests
    int paint_tube_test;    ////test 210
    int bunny_test;         ////test 219,220
    T paint_height_factor;  ////test 214
    bool use_1d_tube;       ////test 210
    bool use_tube_rigid;    ////test 210
    bool use_overlapping_tube;  ////test 210
    bool use_full_letters;  ////test 211
    bool use_letter_r;             ////test 211
    bool use_low_fric_near_fiber;   ////test 211
    T piz_thickness_min_factor; ////test 205, 206    
    T piz_thickness_max_factor; ////test 205, 206
    int ring_source_peak_number;    ////test 213
    bool use_ring_source_randomness;    ////test 213
    T cylinder_angular_velocity;     ////test 222
    int paint_trace_test;           ////test 211
    int piz_init_frame;         ////test 205        
    T piz_init_angle;           ////test 205
    T piz_mesh_noise_scale;     ////test 205
    TV piz_mesh_noise_offset;    ////test 205
    std::string piz_mesh_filename;  ////test 205
    int piz_test;               ////test 205
    int spray_init_frame;           ////test 222
    std::string spray_filename;     ////test 222
    int trace_init_frame;       ////test 211
    std::string trace_filename; ////test 211
    bool use_sim_box; ////test 211

    PAINT_DRIVER();
    virtual ~PAINT_DRIVER();
    virtual void Convection(const T dt,const T time);
    virtual void Write_Output_Files(const int frame);
    virtual void Read_Output_Files(const int frame);
    virtual void Initialize_Test_Parameters();
    virtual void Initialize_Test_Parameters2();
    virtual void Initialize_Solver();
    virtual void Initialize_Kinematic_Positions();
    virtual void Initialize_Mpi();
    virtual void Initialize_Mesh();
    virtual void Initialize_After_Restart();
    virtual void Update_Mesh();
    virtual void Update_Boundary_Conditions();
    virtual void Update_Boundary_Conditions(CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible);
    virtual void Update_Particle_Incompressibility();
    virtual void Update_Target_Divergence();
    virtual void Update_Particle_Length_Scale();
    virtual void Update_Kinematic_Positions(const T dt,const T time);
    virtual void Update_Position_Based_State(const T time,const bool is_position_update);
    virtual void Update_Spatial_Partition();
    virtual void Update_Particle_Contact();
    virtual void Update_Particle_Inactive();
    virtual void Update_Particle_Meshing_Direction();
    virtual void Correct_Particle_Positions();
    virtual void Add_Velocity_Independent_Forces(const T dt,const T time);
    virtual void Add_Velocity_Dependent_Forces(const T dt,const T time);

    void Initialize_Prescribed_Object_Motion();
    void Initialize_Paint_Color();
    void Initialize_Paint_Phis();
    void Initialize_Paint_Textures();
    void Update_Kinematic_Object_Velocities_And_Positions(const T dt,const T time);
    void Update_Kinematic_Paint_Velocities_And_Positions(const T dt,const T time);
    void Squash_Fluid();
    void Add_Mesh_Objects();

    ////functions for specific examples
    void Add_Particles_And_Mesh(const ARRAY<TV>& added_X,const CODIMENSIONAL_MESH& added_mesh,const ARRAY<TV>* added_V=0,const ARRAY<T>* added_thickness=0,const ARRAY<ARRAY<T> >* added_phis=0);
    void Initialize_Pizza(const ARRAY<TV>& X,/*rst*/ARRAY<T>& thickness,/*rst*/ARRAY<TV>& texture_coordinates,const TV& center,const T radius,const T max_thickness,const T min_thickness);
    void Add_Cheese_Filaments();
    void Add_Cheese_Filaments_Helper(ARRAY<TV>& added_X,CODIMENSIONAL_MESH& added_mesh,ARRAY<TV>* added_V=0,ARRAY<T>* added_thickness=0,ARRAY<ARRAY<T> >* added_phis=0);
    TV Read_Tube_Initial_Position_From_File(const std::string& file_name);
    void Write_Tube_Eulerian_Particles_To_File(PAINT_TUBE<TV>& tube,const std::string& file_name);
    void Read_Tube_Eulerian_Particles_From_File(PAINT_TUBE<TV>& tube,const std::string& file_name);
    void Initialize_Part_Pizza_Mesh();
    void Initialize_Pizza_Texture_Coordinates(const ARRAY<TV>& X,const TV& center,const T radius,/*rst*/ARRAY<TV>& texture_coordinates);

protected:
    ////helper functions
    void Add_Canvas(const T size=(T).2,const TV& dir=TV::Axis_Vector(2),const TV center=TV());
    void Add_Bread_Slices(const T size=(T).1);
    void Add_Tube_Rigid_Body(PAINT_TUBE<TV>& tube);
    void Add_Fiber_Rigid_Body(const ARRAY<TV>& fiber_positions,FIBER<TV>& fiber);
    void Add_Toothbrush_Rigid_Body(const ARRAY<TV>& fiber_positions,FIBER<TV>& fiber);
    
    //////fluid solid intersection
    //bool Tall_Particle_Intersect_Solid(const int p,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id);
    //bool Tall_Particle_Intersect_Solid_And_Value(const int p,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour=(T)0);
    //bool Prism_Intersect_Solid(const TV_INT3& v,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id);
    //T Prism_Lost_Volume(const TV_INT3& v,const RIGID_BODY<TV>& rigid_body);
    //T Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const RIGID_BODY<TV>& rigid_body);

    T Distance_Fiber_To_Canvas();
    void Update_Particle_Length_Scale_From_Levelset(const T length_scale_min,const T length_scale_max,const T band_width_min,const T band_width_max);
    void Update_Particle_Length_Scale_Based_On_Tests();
    void Update_Particle_Length_Scale_From_Codimensions();
    void Update_Particle_Length_Scale_From_Rim();
    void Update_Inactive_Particle_Based_On_Tests();
};
}
#endif
