//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Viscoelastic/PAINT_TUBE.h"
#include <fstream>
using namespace PhysBAM;


template<class TV>
TV PhysBAM::PAINT_TUBE<TV>::Cylindrical_Normal(const TV& pos)
{
    TV d1=pos-cylinder.plane1.x1;TV n1=(d1-TV::Dot_Product(d1,dir)*dir).Normalized();return n1;
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Cylindrical_Coordinates(const TV& pos,/*rst*/T& axial_distance,/*rst*/T& radical_distance,/*rst*/T& angle)
{
    TV v=pos-cylinder.plane1.x1;
    axial_distance=TV::Dot_Product(v,dir);
    radical_distance=(v-axial_distance*dir).Magnitude();

    TV d1=axis1;TV d2=Cylindrical_Normal(pos);
    T cos_theta=TV::Dot_Product(d1,d2);
    T sign=VECTOR<T,3>::Dot_Product(VECTOR<T,3>::Cross_Product(d1,d2),dir)>(T)0?(T)1:(T)-1;
    angle=sign*acos(cos_theta)+(T)pi;
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Enable(bool enable,CODIMENSIONAL_FLUID_PARTICLES<TV>* particles/*=0*/)
{
    if(enabled==enable)return;
    enabled=enable;
    if(enabled&&particles)Initialzie_Eulerian_Particles(*particles,eulerian_frac);
}

template<class TV>
TV PhysBAM::PAINT_TUBE<TV>::Source_Velocity(const T r,const T a,const T time)
{
    if(!use_ring_source){
        return dir*source_speed+object_twist.linear;  ////TODO: consider angular velocity
    }
    else{
        ROTATION<TV> rot(a,dir);TV n1=rot.Rotate(axis1);
        //T r=random.Get_Uniform_Number((T)0,(T)1);T scale=(T)1+(T).25*r;
        //T spatial_variation=((T)1+sin((T)5*a))*scale;
        //return -n1*source_speed*spatial_variation*cos(ring_source_angle)+dir*source_speed*sin(ring_source_angle)+object_twist.linear;
        ////return -n1*source_speed*cos(ring_source_angle)+dir*source_speed*sin(ring_source_angle);

        if(ring_source_speed_bspline){
            T spatial_variation=ring_source_speed_bspline->Evaluate(clamp(a/(T)two_pi,(T)0,(T)1));
            T rand_x=random.Get_Uniform_Number((T)0,(T)1);T rescale_x=(T).75+(T).25*rand_x;
            T rand_y=random.Get_Uniform_Number((T)0,(T)1);T rescale_y=(T).75+(T).25*rand_y;
            //LOG::cout<<"dir "<<dir<<": "<<source_speed<<std::endl;
            return -n1*source_speed*spatial_variation*cos(ring_source_angle)*rescale_x+dir*source_speed*sin(ring_source_angle)*rescale_y+object_twist.linear;
        }
        else{
            //T r=random.Get_Uniform_Number((T)0,(T)1);T scale=(T)1+(T).25*r;
            //T spatial_variation=((T)1+(T).5*sin((T)8*a))*scale;
            //return -n1*source_speed*spatial_variation*cos(ring_source_angle)+dir*source_speed*sin(ring_source_angle)+object_twist.linear;

            T phase=(T)ring_source_peak_num*a;
            //T normal_variation=(T)1.2+(T).7*sin(phase);

            T frac_min=(T).8;T frac_max=(T)1.8;
            T normal_variation=frac_min+(frac_max-frac_min)*(T).5*(sin(phase)+(T)1);

            //T y_variation=(T)1+(T).5*cos(phase);
            //T y_variation=(T)1+(T).55*cos(phase);

            T y_frac_min=(T).3;T y_frac_max=(T)1.5;
            T y_variation=y_frac_min+(y_frac_max-y_frac_min)*(T).5*(sin(phase-(T)half_pi)+(T)1);

            //T y_variation=(T).7+(T).8*cos(phase);

            //double rdm_scale[]={.8,1.2,1,.7,1.3,1,1.2,1.4,1.1,.7,1.2,.8,.9,1.4};
            double rdm_scale[]={1.0,.7,1.1,.8,1.,.85,1.15,.79,1.1,.9,1.2,.9,1.2};

            int peak_ids=(int)floor(phase/(T)two_pi);
            normal_variation*=(T)rdm_scale[peak_ids];

            if(use_ring_source_randomness){
                T rnd_1=random.Get_Uniform_Number((T)0,(T)1);
                T scale_1=(T).85+(T).3*rnd_1;
                T rnd_2=random.Get_Uniform_Number((T)0,(T)1);
                T scale_2=(T).95+(T).1*rnd_2; 
                normal_variation*=scale_1;
                y_variation*=scale_2;
            }

            return -n1*source_speed*normal_variation+dir*source_speed*y_variation+object_twist.linear;
        }
    }
}

template<class TV>
typename TV::SCALAR PhysBAM::PAINT_TUBE<TV>::Angular_Noise(const T angle)
{
    return (T)0;
}

template<class TV>
typename TV::SCALAR PhysBAM::PAINT_TUBE<TV>::Source_Thickness_Factor(const T r/*=(T)0*/,const T a/*=(T)0*/,const T time/*=0*/)
{
    //T coef=(T)1;
    if(source_thickness_factor_bspline){
        if(a<(T)half_pi*(T).5)return source_thickness_factor;
        else return source_thickness_factor_bspline->Evaluate((T)0);
    }
    //else return source_thickness_factor;
    else{
        T frac_min=(T).5;T frac_max=(T)1;
        T phase=(T)ring_source_peak_num*a;
        T normal_variation=frac_min+(frac_max-frac_min)*(T).5*(sin(phase)+(T)1);
        //if(time<(T).08)normal_variation=(T)1;
        //LOG::cout<<"factor "<<a<<": "<<normal_variation<<", "<<time<<std::endl;
        return source_thickness_factor*normal_variation;
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialzie_Eulerian_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,T frac)
{
    eulerian_particles.Remove_All();
    T cylinder_frac=frac;if(!enabled)cylinder_frac=(T)3;
    for(int i=1;i<=particles.X.m;i++){
        T axial_distance=(T)0;T radical_distance=(T)0;T angle=(T)0;Cylindrical_Coordinates(particles.X(i),axial_distance,radical_distance,angle); 
        bool is_eulerian_particle=false;
        if(use_ring_source) is_eulerian_particle=radical_distance<=cylinder.radius+(ring_source_radius_max-cylinder.radius)*frac;
        else is_eulerian_particle=radical_distance<cylinder.radius*(T)1.1&&axial_distance<cylinder.height*cylinder_frac;

        if(is_eulerian_particle){
            TV object_space_pos=object_frame.Inverse_Times(particles.X(i));
            EULERIAN_PARTICLE ep(particles.guid(i),radical_distance,angle,object_space_pos,particles.phis.m);
            for(int j=1;j<=particles.phis.m;j++){ep.phis(j)=particles.phis(j)(i);}
            ep.Set_Constant_Thickness(particles.thickness(i));
            eulerian_particles.Set(particles.guid(i),ep);
            //LOG::cout<<"ep: "<<i<<": "<<particles.thickness(i)<<": "<<object_space_pos<<std::endl;
        }
    }

    if(use_texture_coordinates)Initialize_Texture_Coordinates(particles);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Texture_Coordinates(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    if(initialize_texture_coordinates_box){
        TV minimum=particles.X(1),maximum=particles.X(1);
        for(int i=2;i<=particles.X.m;i++)for(int d=1;d<=3;d++){
            if(particles.X(i)(d)<minimum(d))minimum(d)=particles.X(i)(d);
            if(particles.X(i)(d)>maximum(d))maximum(d)=particles.X(i)(d);
        }

        TV range=maximum-minimum;

        for(int i=1;i<=particles.X.m;i++){
            TV tex_coord=TV();
            for(int d=1;d<=3;d++)tex_coord(d)=(particles.X(i)(d)-minimum(d))/range(d);
            particles.texture_coordinate(i)=tex_coord;

            if(eulerian_particles.Contains(particles.guid(i))){
                EULERIAN_PARTICLE& ep=eulerian_particles.Get(particles.guid(i));
                ep.texture_coordinates=tex_coord;
            }
        }
    }
    else{
        for(int i=1;i<=particles.X.m;i++){
            T axial_distance=(T)0;T radical_distance=(T)0;T angle=(T)0;
            Cylindrical_Coordinates(particles.X(i),axial_distance,radical_distance,angle); 
            bool is_tube_particle=radical_distance<=cylinder.radius*(T)1.2&&
                axial_distance<=cylinder.height*(T)1.5;
            if(!is_tube_particle)continue;
            
            TV tex_coord=TV();tex_coord.x=angle/(T)two_pi;tex_coord.y=height-axial_distance;
            particles.texture_coordinate(i)=tex_coord;

            if(eulerian_particles.Contains(particles.guid(i))){
                EULERIAN_PARTICLE& ep=eulerian_particles.Get(particles.guid(i));
                ep.texture_coordinates=tex_coord;
            }
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Ring_Phis(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const int number_of_regions)
{
    for(int i=1;i<=particles.X.m;i++){
        T axial_distance=(T)0;T radical_distance=(T)0;T angle=(T)0;
        Cylindrical_Coordinates(particles.X(i),axial_distance,radical_distance,angle); 
        bool is_tube_particle=radical_distance<=cylinder.radius*(T)2&&
            axial_distance<=cylinder.height*(T)2;
        if(!is_tube_particle)continue;

        T frac=angle/(T)two_pi;
        int p=(frac>(T).2&&frac<(T).3)?1:2;
        for(int c=1;c<=number_of_regions;c++){
            T sign=1;if(c==p)sign=-1;
            T phi=sign*(T)1e10;
            particles.phis(c)(i)=phi;
        }

        //if(eulerian_particles.Contains(particles.guid(i))){
        //    EULERIAN_PARTICLE& ep=eulerian_particles.Get(particles.guid(i));
        //    for(int c=1;c<=number_of_regions;c++)ep.phis(c)=particles.phis(c)(i);
        //}
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Advance_Texture_Coordinates(const T dt,const T time)
{
    for(HASHTABLE_ITERATOR<int,EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
        TV tex_coord=iter.Data().texture_coordinates;
        tex_coord.y+=source_speed*dt*texture_coordinate_space_scale.y;
        iter.Data().texture_coordinates=tex_coord;
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Update_Eulerian_Particle_Positions_And_Mass(CODIMENSIONAL_MESH_OBJECT<TV>& mesh_object,T min_tri_thickness,T min_seg_thickness,T time/*=(T)0*/)
{
    CODIMENSIONAL_FLUID_PARTICLES<TV>& particles=mesh_object.particles;
    T density=(T)1000.;
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            EULERIAN_PARTICLE& ep=eulerian_particles.Get(guid);
            TV pos=object_frame*ep.object_space_pos;
            particles.X(i)=pos;
            if(source_speed>(T)0&&ep.has_constant_thickness){
                T thickness=ep.thickness*Source_Thickness_Factor(ep.r,ep.angle,time);
                //LOG::cout<<"ep "<<ep.angle<<", "<<Source_Thickness_Factor(ep.r,ep.angle)<<", "<<thickness<<std::endl;
                int p_type=mesh_object.mesh.node_types(i);
                if(Is_On_Triangle_Mesh(p_type)){thickness=clamp_min(min_tri_thickness,thickness);}
                if(Is_On_Segment_Mesh(p_type)){thickness=clamp_min(min_seg_thickness,thickness);}

                particles.thickness(i)=thickness;
                //LOG::cout<<"thk "<<thickness<<std::endl;

                particles.dimensional_mass(i)=mesh_object.Nodal_Dimensional_Volume_With_Thickness(i,thickness)*density;
                particles.mass(i)=particles.dimensional_mass(i).Sum();
            }
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Reinitialize_Eulerian_Particles(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    if(!use_reinitialize_eulerian_particles)return;
    //LOG::cout<<"reinit ep "<<std::endl;
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            EULERIAN_PARTICLE& ep=eulerian_particles.Get(guid);
            //if(abs(ep.object_space_pos.x)>object_space_x_max||abs(ep.object_space_pos.z)>object_space_r){
            //    eulerian_particles.Delete(guid);
            //    LOG::cout<<"~~~~~~~~~~~~~~ delete ep "<<guid<<std::endl;
            //}
            //else{
                ep.object_space_pos.z=ep.object_space_pos0.z*object_space_z_rescale;
                ep.object_space_pos.x=ep.object_space_pos0.x*object_space_x_rescale;
                //ep.object_space_pos.z*=object_space_z_rescale;                
            //}
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Update_Eulerian_Particle_Velocities(HASHTABLE<int,TV>& particle_velocities)
{
    for(HASHTABLE_ITERATOR<int,EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
        int guid=iter.Key();T r=iter.Data().r;T a=iter.Data().angle;
        TV v=Source_Velocity(r,a);particle_velocities.Set(guid,v);
        //LOG::cout<<"ep???? : "<<guid<<": "<<v<<std::endl;
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Update_Eulerian_Particle_Phis(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            EULERIAN_PARTICLE& ep=eulerian_particles.Get(guid);
            for(int j=1;j<=particles.phis.m;j++){
                particles.phis(j)(i)=ep.phis(j);
            }
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Update_Eulerian_Particle_Texture_Coordinates(CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    for(int i=1;i<=particles.X.m;i++){
        int guid=particles.guid(i);
        if(eulerian_particles.Contains(guid)){
            EULERIAN_PARTICLE& ep=eulerian_particles.Get(guid);
            particles.texture_coordinate(i)=ep.texture_coordinates;
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_V_Helper(const ARRAY<TV>& X,/*rst*/ARRAY<TV>& V)
{
    V.Resize(X.m);for(int i=1;i<=X.m;i++){
        T axial_distance=(T)0;T radical_distance=(T)0;T angle=(T)0;
        Cylindrical_Coordinates(X(i),axial_distance,radical_distance,angle); 
        V(i)=Source_Velocity(radical_distance,angle);
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Volumetric_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T hemisphere_frac/*=(T).8*/)
{
    LOG::cout<<"vol src: "<<e1<<", height: "<<height<<", dir: "<<dir<<", e2: "<<e1+height*dir<<std::endl;
    if(hemisphere_frac==(T).0)CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(mesh,X,e1,e1+height*dir,radius,length_scale);
    else CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder_With_Rounded_End(mesh,X,e1,e1+height*dir,radius,length_scale,hemisphere_frac);
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Volumetric_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T thickness)
{
    TV e2=e1+height*dir;
    TV box_min=e2-TV(thickness*(T).5,(T)0,radius);TV box_max=e1+TV(thickness*(T).5,(T)0,radius);
    CODIMENSIONAL_MESH_GENERATOR<TV>::Box(mesh,X,box_min,box_max,length_scale);
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V,const T thickness)
{
    TV e2=e1+height*dir;
    TV box_min=e2-TV(thickness*(T).5,(T)0,radius);TV box_max=e1+TV(thickness*(T).5,(T)0,radius);
    CODIMENSIONAL_MESH_GENERATOR<TV>::Smoothed_Box(mesh,X,box_min,box_max,length_scale);
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Thin_Sheet_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V)
{
    TV e2=e1+height*dir;
    TV min_corner=e2-TV::Axis_Vector(3)*radius;VECTOR<T,2> size=VECTOR<T,2>(radius*(T)2,height);
    CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(mesh,X,min_corner,size,length_scale,1);
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Filament_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V)
{
    CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(mesh,X,e1,e1+height*dir,length_scale);
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Ring_Mesh_And_X_And_V(CODIMENSIONAL_MESH& mesh,ARRAY<TV>& X,ARRAY<TV>& V)
{
    int number_of_rings=3;
    CODIMENSIONAL_MESH_GENERATOR<TV>::Ring(mesh,X,e1,dir,radius,number_of_rings,length_scale);
    ring_source_radius_max=radius+number_of_rings*length_scale;
    use_ring_source=true;
    Initialize_V_Helper(X,V);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Advance(const T dt,const T time)
{
    if(use_texture_coordinates)Advance_Texture_Coordinates(dt,time);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Update_Position_Based_States()
{
    e0=object_frame*object_space_e0;
    //e1=object_frame*object_space_e1;
    dir=object_frame.Rotation().Rotate(object_space_dir).Normalized();
    axis1=object_frame.Rotation().Rotate(object_space_axis1).Normalized();
    e1=e0+dir*tube_height;
    TV e2=e1+dir*height;
    cylinder.Set_Endpoints(e1,e2);

    if(rigid_bodies.m>0)for(int i=1;i<=rigid_bodies.m;i++){
        if(rigid_bodies(i)->name=="tube_cylinder")rigid_bodies(i)->Rotation()=object_frame.Rotation()*ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        else rigid_bodies(i)->Rotation()=object_frame.Rotation();
        rigid_bodies(i)->X()=e1-dir*(T).5*rigid_body_height;
        rigid_bodies(i)->V()=object_twist.linear;
        rigid_bodies(i)->Angular_Velocity()=object_twist.angular;
    }
}

//template<class TV>
//void PhysBAM::PAINT_TUBE<TV>::Initialize_Bspline_Helper(ARRAY<VECTOR<T,2> >& bspline_time_value,BSPLINE<T,T>*& bspline,int order/*=1*/)
//{
//    ARRAY<T> bspline_time;ARRAY<T> bspline_value;
//    for(int i=1;i<=bspline_time_value.m;i++){
//        bspline_time.Append(bspline_time_value(i).x);
//        bspline_value.Append(bspline_time_value(i).y);
//    }
//    bspline=new BSPLINE<T,T>(bspline_time,bspline_value,order);
//}
//
//template<class TV>
//void PhysBAM::PAINT_TUBE<TV>::Add_Bspline_Time_Value_Helper(const T time,const T value,ARRAY<VECTOR<T,2> >& bspline_time_value)
//{
//    int n=2;
//    if(bspline_time_value.m==0)for(int j=1;j<=n;j++){bspline_time_value.Append(TV2(time,value));}
//    else for(int j=1;j<=n;j++){bspline_time_value.Pop();}////first node
//    bspline_time_value.Append(TV2(time,value));
//    for(int j=1;j<=n;j++){bspline_time_value.Append(TV2(time,value));}////last node
//}


template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Source_Speed_Bspline()
{
    Initialize_Bspline_Helper(source_speed_bspline_time_value,source_speed_bspline);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Source_Speed_Bspline_From_File(const std::string& file_name)
{
    std::string str;STREAM_TYPE stream_type((T()));std::fstream in(file_name.c_str(),std::ios::in);
    if(!in){LOG::cout<<"cannot open source speed bspline file "<<file_name<<std::endl;}

    T t=(T)0;T speed=(T)0;
    while(in>>t>>speed){Add_Source_Speed_Bspline_Time_Value(t,speed);}
    Initialize_Source_Speed_Bspline();
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Add_Source_Speed_Bspline_Time_Value(const T time,const T value)
{
    Add_Bspline_Time_Value_Helper(time,value,source_speed_bspline_time_value);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Ring_Source_Speed_Bspline()
{
    Initialize_Bspline_Helper(ring_source_speed_bspline_time_value,ring_source_speed_bspline);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Add_Ring_Source_Speed_Bspline_Time_Value(const T time,const T value)
{
    Add_Bspline_Time_Value_Helper(time,value,ring_source_speed_bspline_time_value);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Initialize_Source_Thickness_Factor_Bspline()
{
    Initialize_Bspline_Helper(source_thickness_factor_bspline_time_value,source_thickness_factor_bspline);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Add_Source_Thickness_Factor_Bspline_Time_Value(const T time,const T value)
{
    Add_Bspline_Time_Value_Helper(time,value,source_thickness_factor_bspline_time_value);
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Add_Eulerian_Particles_From_Mesher(const HASHTABLE<int>& mesher_eulerian_particles_hashtable,HASHTABLE<int,PAIR<VECTOR<int,2>,T> > added_eulerian_particle_parents_hashtable,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles)
{
    //LOG::cout<<"add ep from mesher"<<std::endl;
    for(int i=1;i<=particles.X.m;i++){
        if(mesher_eulerian_particles_hashtable.Contains(particles.guid(i))&&!eulerian_particles.Contains(particles.guid(i))){
            //PAIR<VECTOR<int,2>,T> parent_weight=added_eulerian_particle_parents_hashtable.Get(particles.guid(i));
            //EULERIAN_PARTICLE parent_1=eulerian_particles.Get(parent_weight.x(1));
            //EULERIAN_PARTICLE parent_2=eulerian_particles.Get(parent_weight.x(2));
            //T weight=parent_weight.y;
            //TV object_space_pos0=((T)1-weight)*parent_1.object_space_pos0+weight*parent_2.object_space_pos0;

            T axial_distance=(T)0;T radical_distance=(T)0;T angle=(T)0;
            Cylindrical_Coordinates(particles.X(i),axial_distance,radical_distance,angle); 
            TV object_space_pos=object_frame.Inverse_Times(particles.X(i));
            EULERIAN_PARTICLE ep(particles.guid(i),radical_distance,angle,object_space_pos,particles.phis.m);
            for(int j=1;j<=particles.phis.m;j++){ep.phis(j)=particles.phis(j)(i);}
            ep.Set_Constant_Thickness(particles.thickness(i));
            ep.object_space_pos0=object_space_pos;
            if(object_space_x_rescale!=(T)0)ep.object_space_pos0.x=object_space_pos.x/object_space_x_rescale;
            if(object_space_z_rescale!=(T)0)ep.object_space_pos0.z=object_space_pos.z/object_space_z_rescale;
            if(use_texture_coordinates)ep.texture_coordinates=particles.texture_coordinate(i);
            eulerian_particles.Set(particles.guid(i),ep);
            //LOG::cout<<"add ep: "<<particles.guid(i)<<": "<<particles.thickness(i)<<": "<<object_space_pos<<std::endl;
        }
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Set_Constant_Thickness(const T thickness)
{
    for(HASHTABLE_ITERATOR<int,EULERIAN_PARTICLE> iter(eulerian_particles);iter.Valid();iter.Next()){
        EULERIAN_PARTICLE& ep=iter.Data();
        ep.thickness=thickness;
    }
}

template<class TV>
void PhysBAM::PAINT_TUBE<TV>::Remove_All_Eulerian_Particles()
{
    eulerian_particles.Remove_All();
}

//template class PAINT_TUBE<VECTOR<float,2> >;
template class PAINT_TUBE<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
//template class PAINT_TUBE<VECTOR<double,2> >;
template class PAINT_TUBE<VECTOR<double,3> >;
#endif
