//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/ROCK.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::ROCK<TV>::ROCK()
    :motion_type(FALL1),initial_height((T)0),final_height((T)0),initial_velocity((T)0),gravity((T)-9.81)
{
    //no-op
}

template<class TV>
PhysBAM::ROCK<TV>::~ROCK()
{
    //no-op
}

template<class TV>
void PhysBAM::ROCK<TV>::Set_Prescribed_Motion_For_Test()
{
    switch(motion_type){
    case FALL1:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)pi/(T)2,TV::Axis_Vector(1));

        //ROTATION<TV> current_rotation=r*ROTATION<TV>((T)-one_fourth_pi,TV::Axis_Vector(3));
        ROTATION<TV> current_rotation=r;

        T current_time=(T)0;
        T dt=(T)0.05;
        //T dt=(T)1;
        T dx=(T)0.001;

        t=TV((T)0,(T)4*dx,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)5,TV::Axis_Vector(3));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);


        current_time+=dt;
        current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;
        current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;
        current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)6,TV::Axis_Vector(1));
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //current_rotation=current_rotation*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));
    }break;
    case FALL2:{
        FRAME<TV> original_frame=object_frame;
        TV x=original_frame.t;
        TV t;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)pi/(T)2,TV::Axis_Vector(1));

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)pi/(T)2,TV::Axis_Vector(1));

        T current_time=(T)0;
        T dt=(T)0.05;
        T dx=(T)0.05;

        t=TV((T)0,(T)4*dx,(T)0);x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt;
        current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)12,TV::Axis_Vector(3));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt/2;
        current_rotation=current_rotation*ROTATION<TV>((T)-pi/(T)12,TV::Axis_Vector(3));
        t(2)-=dx;x=original_frame.t+t;r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        current_time+=dt/2;
        current_rotation=current_rotation*ROTATION<TV>((T)-pi/(T)12,TV::Axis_Vector(3));
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        current_time+=dt;
    }break;
    case GRAVITY_WITH_ROTATION:{
        FRAME<TV> original_frame=object_frame;
        TV x;
        ROTATION<TV> r=original_frame.r*ROTATION<TV>((T)pi/(T)2,TV::Axis_Vector(1));

        ROTATION<TV> current_rotation=r*ROTATION<TV>((T)pi/(T)2,TV::Axis_Vector(1));

        T current_time=(T)0;
        //T dt=(T)0.05;

        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        //current_time+=dt;
        //current_time+=(T)0.02;
        current_time+=(T)0.0015;
        r=current_rotation;
        Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        ////current_time=(T)0.172;
        ////current_time=(T)0.032;
        ////current_time=(T)0.052;
        //current_time=(T)0.004;
        //current_rotation=current_rotation*ROTATION<TV>((T)pi/(T)12,TV::Axis_Vector(3));
        //r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        ////current_time+=dt/2;
        //current_time+=dt/2;
        //current_rotation=current_rotation*ROTATION<TV>((T)-pi/(T)12,TV::Axis_Vector(3));
        //r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame(current_time,x,r);

        ////current_time+=dt/2;
        //current_time+=dt/3;
        ////current_rotation=current_rotation*ROTATION<TV>((T)-pi/(T)12,TV::Axis_Vector(3));
        //current_rotation=current_rotation*ROTATION<TV>((T)-pi/(T)24,TV::Axis_Vector(3));
        //r=current_rotation;
        //Add_Prescribed_Motion_Key_Frame(current_time,x,r);
    }break;
    default:{
        LOG::cout<<"WARNING: no ROCK motion type specified"<<std::endl;
    }break;
    }

    Initialize_Prescribed_Motion();
}

template<class TV>
FRAME<TV> PhysBAM::ROCK<TV>::Prescribed_Frame(const T time)
{
    FRAME<TV> frame;
    if(motion_type==GRAVITY_WITH_ROTATION){
        //TV pos=rigid_body->X();
        TV pos;
        pos.y=initial_height+initial_velocity*time+(T)0.5*gravity*time*time;
        pos.y=clamp_min(pos.y,final_height);

        //LOG::cout<<std::endl<<"analytical velocity: "<<initial_velocity*gravity*time<<std::endl<<std::endl;

        ROTATION<TV> rot=ROTATION<TV>::From_Quaternion(motion_Q_bspline->Clamped_Evaluate(time));
        frame.t=pos;
        frame.r=rot;
    }
    else{
        frame=BASE::Prescribed_Frame(time);
    }
    return frame;
}

template class ROCK<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class ROCK<VECTOR<double,3> >;
#endif
