//#####################################################################
// Copyright 2014, Ed Quigley
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/SCRAPER.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::SCRAPER<TV>::SCRAPER()
    :motion_type(HORIZONTAL_SWEEP),use_spline(true),initial_X(),initial_V(),acceleration()
{
    //no-op
}

template<class TV>
PhysBAM::SCRAPER<TV>::~SCRAPER()
{
    //no-op
}

template<class TV>
void PhysBAM::SCRAPER<TV>::Set_Prescribed_Motion_For_Test()
{
    switch(motion_type){
    case HORIZONTAL_SWEEP:{
        use_spline=false;
        //initial_V=TV((T).1,(T)-.01,(T)0);
        //initial_V=TV((T).1,(T)0,(T)0);
        object_frame.r=ROTATION<TV>((T)half_pi,TV::Axis_Vector(2));
        object_frame.r=ROTATION<TV>((T)pi/(T)-6,TV::Axis_Vector(3))*object_frame.r;
    }break;
    default:{
        LOG::cout<<"WARNING: no SCRAPER motion type specified"<<std::endl;
    }break;
    }
    Initialize_Prescribed_Motion();
}

template<class TV>
FRAME<TV> PhysBAM::SCRAPER<TV>::Prescribed_Frame(const T time)
{
    FRAME<TV> frame;
    if(use_spline){
        frame=BASE::Prescribed_Frame(time);
    }
    else{
        TV X=initial_X+initial_V*time+(T)0.5*acceleration*time*time;
        LOG::cout<<std::endl<<"analytical velocity: "<<initial_V+acceleration*time<<std::endl<<std::endl;
        frame.t=X;
        frame.r=object_frame.r;
    }
    return frame;
}

template class SCRAPER<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class SCRAPER<VECTOR<double,3> >;
#endif
