//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Geometry/Level_Sets/IMPLICIT_OBJECT_ON_A_RAY.h>
#include <PhysBAM_Tools/Nonlinear_Equations/ITERATIVE_SOLVER.h>
#include "../Viscoelastic/SOLID_FLUID_CONTACT_CODIMENSIONAL.h"
using namespace PhysBAM;

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside_And_Value(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour/*=(T)0*/)
{
    for(int r=1;r<=rigid_bodies.m;r++){
        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
        if(rigid_body->Implicit_Geometry_Lazy_Inside_And_Value(pos,phi,contour)){
            rigid_normal=rigid_body->Implicit_Geometry_Normal(pos);
            //rigid_v=RIGID_GEOMETRY<TV>::Pointwise_Object_Velocity(rigid_body->Twist(),rigid_body->X(),rigid_body->Object_Space_Point(pos));
            rigid_v=rigid_body->Pointwise_Object_Velocity(pos);
            id=r;
            return true;
        }
    }
    return false;
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,const T contour/*=(T)0*/)
{
    for(int r=1;r<=rigid_bodies.m;r++){
        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
        if(rigid_body->Implicit_Geometry_Lazy_Inside(pos,contour)){id=r;return true;}
    }
    return false;
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside(const TV& pos,const RIGID_BODY<TV>& rigid_body,/*rst*/int& unused,const T contour/*=(T)0*/)
{
    return rigid_body.Implicit_Geometry_Lazy_Inside(pos,contour);
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Inside(const TV& pos,const FIBER<TV>& fiber,/*rst*/int& id,const T contour/*=(T)0*/)
{   ///TODO: contour is unused
    return fiber.Fiber_Inside(pos,id);    
}

template<class TV>
TV PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Solid_Extended_Normal(const TV& pos,const ARRAY<RIGID_BODY<TV>* > rigid_bodies,/*rst*/int& id,/*rst*/T& phi)
{
    T phi_min=(T)FLT_MAX;TV normal_min=TV();
    for(int r=1;r<=rigid_bodies.m;r++){
        RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
        T phi=(T)0;TV normal=rigid_body->Implicit_Geometry_Extended_Normal(pos,phi);
        if(phi<phi_min){phi_min=phi;normal_min=normal;id=r;}
    }
    phi=phi_min;return normal_min;
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id)
{
    /*
    const TV canvas_normal=TV::Axis_Vector(2);
    TV endpoint_pos=particles.X(p)+canvas_normal*particles.thickness(p);
    return Solid_Inside(endpoint_pos,rigid_bodies,id);
    */

    //sample points along particle's height to find intersection
    const TV canvas_normal=TV::Axis_Vector(2);
    bool inside=false;
    const int num_steps=5;
    for(int i=0;i<=num_steps&&!inside;i++){
        TV pos=particles.X(p)+((T)i)/((T)num_steps)*canvas_normal*particles.thickness(p);
        inside=Solid_Inside(pos,rigid_bodies,id);
    }
    return inside;
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const FIBER<TV>& fiber,/*rst*/int& id)
{
    const TV canvas_normal=TV::Axis_Vector(2);
    TV endpoint_pos=particles.X(p)+canvas_normal*particles.thickness(p);
    return Solid_Inside(endpoint_pos,fiber,id);
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Intersect_Solid_And_Value(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour/*=(T)0*/)
{
    /*
    const TV canvas_normal=TV::Axis_Vector(2);
    TV endpoint_pos=particles.X(p)+canvas_normal*particles.thickness(p);
    return Solid_Inside_And_Value(endpoint_pos,rigid_bodies,id,phi,rigid_normal,rigid_v,contour);
    */

    //sample points along particle's height to find intersection
    bool inside=false;
    const int num_steps=5;
    const TV canvas_normal=TV::Axis_Vector(2);
    for(int i=0;i<=num_steps&&!inside;i++){
        TV pos=particles.X(p)+((T)i)/((T)num_steps)*canvas_normal*particles.thickness(p);
        inside=Solid_Inside_And_Value(pos,rigid_bodies,id,phi,rigid_normal,rigid_v,contour);
    }
    return inside;
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Intersect_Solid(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id)
{
    /*
    const TV canvas_normal=TV::Axis_Vector(2);
    for(int i=1;i<=3;i++){
        TV endpoint_pos=particles.X(v(i))+canvas_normal*particles.thickness(v(i));
        if(Solid_Inside(endpoint_pos,rigid_bodies,id))return true;
    }
    return false;
    */

    //sample points along particle's height to find intersection
    const int num_steps=5;
    const TV canvas_normal=TV::Axis_Vector(2);
    for(int i=1;i<=3;i++){
        for(int j=0;j<=num_steps;j++){
            TV pos=particles.X(v(i))+((T)j)/((T)num_steps)*canvas_normal*particles.thickness(v(i));
            if(Solid_Inside(pos,rigid_bodies,id))return true;
        }
    }
    return false;

}

template<class TV>
typename TV::SCALAR PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Lost_Volume(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const RIGID_BODY<TV>& rigid_body)
{
    const TV canvas_normal=TV::Axis_Vector(2);
    TV center=(T)one_third*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3)));
    TV triangle_normal=canvas_normal;   ////TODO: handle non-contact case
    T averaged_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
    T tolerance=averaged_thickness*(T)1e-4;
    T triangle_area=TRIANGLE_3D<T>::Area(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)));
    T t_intersect=Segment_Intersection_Fraction_With_Solid(center,center+triangle_normal*averaged_thickness,rigid_body,tolerance);
    return ((T)1-t_intersect)*averaged_thickness*triangle_area;
}

//template<class TV>
//typename TV::SCALAR PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Prism_Lost_Volume(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const FIBER<TV>& fiber)
//{
//    const TV canvas_normal=TV::Axis_Vector(2);
//    TV center=(T)one_third*(particles.X(v(1))+particles.X(v(2))+particles.X(v(3)));
//    TV triangle_normal=canvas_normal;   ////TODO: handle non-contact case
//    T averaged_thickness=(T)one_third*(particles.thickness(v(1))+particles.thickness(v(2))+particles.thickness(v(3)));
//    T tolerance=averaged_thickness*(T)1e-4;
//    T triangle_area=TRIANGLE_3D<T>::Area(particles.X(v(1)),particles.X(v(2)),particles.X(v(3)));
//    int seg_id=0;T t_intersect=(T)0;
//    if(fiber.Fiber_Segment_Intersect(center,center+triangle_normal*averaged_thickness,seg_id,t_intersect))return ((T)1-t_intersect)*averaged_thickness*triangle_area;
//    else return (T)0;
//}

template<class TV>
typename TV::SCALAR PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const RIGID_BODY<TV>& rigid_body,const T tolerance/*=(T)0*/)
{
    RAY<TV> ray(SEGMENT_3D<T>(p1,p2));
    IMPLICIT_OBJECT_ON_A_RAY<IMPLICIT_OBJECT_TRANSFORMED<TV,RIGID_GEOMETRY<TV> > > implicit_surface_on_a_ray(*rigid_body.implicit_object,ray); 
    ITERATIVE_SOLVER<T> iterative_solver;iterative_solver.tolerance=tolerance;
    T t_intersect=iterative_solver.Bisection_Secant_Root(implicit_surface_on_a_ray,(T)0,ray.t_max)/(p1-p2).Magnitude();;return t_intersect;
}

//template<class TV>
//T PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const FIBER<TV>& fiber,const T tolerance/*=(T)0*/)
//{
//    RAY<TV> ray(SEGMENT_3D<T>(p1,p2));
//    IMPLICIT_OBJECT_ON_A_RAY<IMPLICIT_OBJECT_TRANSFORMED<TV,RIGID_GEOMETRY<TV> > > implicit_surface_on_a_ray(*rigid_body.implicit_object,ray); 
//    ITERATIVE_SOLVER<T> iterative_solver;iterative_solver.tolerance=tolerance;
//    T t_intersect=iterative_solver.Bisection_Secant_Root(implicit_surface_on_a_ray,(T)0,ray.t_max)/(p1-p2).Magnitude();;return t_intersect;
//}

//template<class TV>
//typename TV::SCALAR PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Edge_Intersection_Fraction_With_Solid(const TV_INT2& edge,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const RIGID_BODY<TV>& rigid_body)
//{
//    TV canvas_normal=TV::Axis_Vector(2);T tolerance=(T)1e-2;int a=edge(1);int c=edge(2);
//    T t_min=(T)0;T t_max=(T)1;
//    TV pos_min=particles.X(a);TV pos_max=particles.X(c);
//    T thickness_min=particles.thickness(a);T thickness_max=particles.thickness(c);
//    bool intersect_min=Solid_Inside(pos_min+canvas_normal*thickness_min,rigid_body);
//    bool intersect_max=Solid_Inside(pos_max+canvas_normal*thickness_max,rigid_body);
//    if(intersect_min==intersect_max)return (T)-1;
//
//    while(t_max-t_min>=tolerance){
//        T t_mid=(T).5*(t_min+t_max);
//        TV pos_mid=(T).5*(pos_min+pos_max);
//        T thickness_mid=(T).5*(thickness_min+thickness_max);
//        bool intersect_mid=Solid_Inside(pos_mid+canvas_normal*thickness_mid,rigid_body);
//        if(intersect_mid==intersect_min){t_min=t_mid;pos_min=pos_mid;thickness_min=thickness_mid;}
//        else{t_max=t_mid;pos_max=pos_mid;thickness_max=thickness_mid;}
//        LOG::cout<<"tall intersect binary search: "<<t_mid<<", "<<t_min<<", "<<t_max<<": "<<intersect_mid<<", "<<intersect_min<<", "<<intersect_max<<std::endl;
//    }
//    return t_min;
//}


template<class TV> template<class T_SOLID>
typename TV::SCALAR PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tall_Particle_Edge_Intersection_Fraction_With_Solid(const TV_INT2& edge,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,T_SOLID& solid)
{
    TV canvas_normal=TV::Axis_Vector(2);T tolerance=(T)1e-2;int a=edge(1);int c=edge(2);
    T t_min=(T)0;T t_max=(T)1;
    TV pos_min=particles.X(a);TV pos_max=particles.X(c);
    T thickness_min=particles.thickness(a);T thickness_max=particles.thickness(c);int unused=0;
    bool intersect_min=Solid_Inside(pos_min+canvas_normal*thickness_min,solid,unused);
    bool intersect_max=Solid_Inside(pos_max+canvas_normal*thickness_max,solid,unused);
    if(intersect_min==intersect_max){LOG::cout<<"what? "<<particles.contact(a)<<", "<<particles.contact(c)<<": "<<intersect_min<<", "<<intersect_max<<std::endl;return (T)-1;}

    while(t_max-t_min>=tolerance){
        T t_mid=(T).5*(t_min+t_max);
        TV pos_mid=(T).5*(pos_min+pos_max);
        T thickness_mid=(T).5*(thickness_min+thickness_max);
        bool intersect_mid=Solid_Inside(pos_mid+canvas_normal*thickness_mid,solid,unused);
        if(intersect_mid==intersect_min){t_min=t_mid;pos_min=pos_mid;thickness_min=thickness_mid;}
        else{t_max=t_mid;pos_max=pos_mid;thickness_max=thickness_mid;}
        //LOG::cout<<"tall intersect binary search: "<<t_mid<<", "<<t_min<<", "<<t_max<<": "<<intersect_mid<<", "<<intersect_min<<", "<<intersect_max<<std::endl;
    }
    return t_min;
}

template<class TV>
TV PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Tank_Normal_And_Position(const TV& pos,const WATER_TANK<TV>& tank,/*rst*/TV& wall_pos)
{
    TV corrected_pos=pos;TV wall_normal;
    const BOX<TV>& tank_box=tank.tank_box;
    for(int d=1;d<=TV::m;d++){
        if(!tank.is_open_tank){
            if(pos(d)<=tank_box.min_corner(d)){
                //TV wall_normal=TV::Axis_Vector(d);
                wall_normal(d)=(T)1;
                corrected_pos(d)=tank_box.min_corner(d);
            }
            if(pos(d)>=tank_box.max_corner(d)){
                wall_normal(d)=(T)-1;
                corrected_pos(d)=tank_box.max_corner(d);
            }
        }
        else{   ////open tank
            if(d==2){
                if(pos(d)<=tank_box.min_corner(d)){
                    wall_normal(d)=(T)1;
                    corrected_pos(d)=tank_box.min_corner(d);
                }
            }
            else{
                T tank_thickness=tank.tank_thickness;
                T y=pos.y;
                if(y>=tank_box.min_corner.y&&y<=tank_box.max_corner.y){
                    if(pos(d)<=tank_box.min_corner(d)&&
                        pos(d)>=tank_box.min_corner(d)-tank_thickness){
                            wall_normal(d)=(T)1;        
                            corrected_pos(d)=tank_box.min_corner(d);
                    }
                    if(pos(d)>=tank_box.max_corner(d)&&
                        pos(d)<=tank_box.max_corner(d)+tank_thickness){
                            wall_normal(d)=(T)-1;
                            corrected_pos(d)=tank_box.max_corner(d);
                    }
                }
            }
        }
    }
    wall_pos=corrected_pos;
    return Normalize_With_Zero(wall_normal);
}

template<class TV>
bool PhysBAM::SOLID_FLUID_CONTACT_CODIMENSIONAL<TV>::Contact_Tank(const TV& pos,const WATER_TANK<TV>& tank)
{
    if(tank.is_open_tank){
        if(pos(2)<=tank.tank_box.min_corner(2))return true;
        for(int d=1;d<=TV::m;d++){if(d==2)continue;if(pos(d)<=tank.tank_box.min_corner(d)||pos(d)>=tank.tank_box.max_corner(d))return true;}return false;
    }
    else for(int d=1;d<=TV::m;d++){if(pos(d)<=tank.tank_box.min_corner(d)||pos(d)>=tank.tank_box.max_corner(d))return true;}return false;
}

#define INSTANTIATION_HELPER(T,d)                                        \
    template class SOLID_FLUID_CONTACT_CODIMENSIONAL<VECTOR<T,d> >;                \
    template T SOLID_FLUID_CONTACT_CODIMENSIONAL<VECTOR<T,d> >::Tall_Particle_Edge_Intersection_Fraction_With_Solid(const VECTOR<int,2>& edge,const CODIMENSIONAL_FLUID_PARTICLES<VECTOR<T,d> >& particles,RIGID_BODY<VECTOR<T,d> >& solid); \
    template T SOLID_FLUID_CONTACT_CODIMENSIONAL<VECTOR<T,d> >::Tall_Particle_Edge_Intersection_Fraction_With_Solid(const VECTOR<int,2>& edge,const CODIMENSIONAL_FLUID_PARTICLES<VECTOR<T,d> >& particles,FIBER<VECTOR<T,d> >& solid); 

INSTANTIATION_HELPER(float,3)
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
    INSTANTIATION_HELPER(double,3)
#endif
