//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __SOLID_FLUID_CONTACT_CODIMENSIONAL__
#define __SOLID_FLUID_CONTACT_CODIMENSIONAL__
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Viscoelastic/FIBER.h"
#include "../Intersection/CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION.h"
#include "../Misc/WATER_TANK.h"

namespace PhysBAM
{
template<class TV>
class SOLID_FLUID_CONTACT_CODIMENSIONAL
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; typedef typename TV::SPIN T_SPIN;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TRIANGLE_3D<T> >::TYPE T_TRIANGLE;

public:
    CODIMENSIONAL_MESH& fluid_mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& fluid_particles;
    CODIMENSIONAL_MESHING_3D<T>& mesher;
    ////solids
    ARRAY<RIGID_BODY<TV>* >* rigid_bodies;
    FIBER<TV>* fiber;

    SOLID_FLUID_CONTACT_CODIMENSIONAL(CODIMENSIONAL_MESHING_3D<T>& mesher_input,ARRAY<RIGID_BODY<TV>* >* rigid_bodies_input=0,FIBER<TV>* fiber_input=0)
        :fluid_mesh(mesher_input.mesh),fluid_particles(mesher_input.particles),mesher(mesher_input),rigid_bodies(rigid_bodies_input),fiber(fiber_input){}

    static bool Solid_Inside_And_Value(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour=(T)0);
    //static bool Solid_Inside_And_Value(const TV& pos,const FIBER<TV>& fiber,/*rst*/int& id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour=(T)0);
    static bool Solid_Inside(const TV& pos,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id,const T contour=(T)0);
    static bool Solid_Inside(const TV& pos,const RIGID_BODY<TV>& rigid_body,/*rst*/int& unused,const T contour=(T)0);
    static TV Solid_Extended_Normal(const TV& pos,const ARRAY<RIGID_BODY<TV>* > rigid_bodies,/*rst*/int& id,/*rst*/T& phi);

    ////tall particle and prism
    static bool Tall_Particle_Intersect_Solid(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id);
    static bool Tall_Particle_Intersect_Solid_And_Value(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int id,/*rst*/T& phi,/*rst*/TV& rigid_normal,/*rst*/TV& rigid_v,const T contour=(T)0);
    static bool Prism_Intersect_Solid(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const ARRAY<RIGID_BODY<TV>* >& rigid_bodies,/*rst*/int& id);
    static T Prism_Lost_Volume(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const RIGID_BODY<TV>& rigid_body);
    //static T Tall_Particle_Edge_Intersection_Fraction_With_Solid(const TV_INT2& edge,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const RIGID_BODY<TV>& rigid_body);
    static T Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const RIGID_BODY<TV>& rigid_body,const T tolerance=(T)0);
    
    static bool Solid_Inside(const TV& pos,const FIBER<TV>& fiber,/*rst*/int& id,const T contour=(T)0);
    
    static bool Tall_Particle_Intersect_Solid(const int p,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const FIBER<TV>& fiber,/*rst*/int& id);

    //static T Segment_Intersection_Fraction_With_Solid(const TV& p1,const TV& p2,const FIBER<TV>& fiber,const T tolerance=(T)0);
    //static T Prism_Lost_Volume(const TV_INT3& v,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const FIBER<TV>& fiber);
    
    template<class T_SOLID> static T Tall_Particle_Edge_Intersection_Fraction_With_Solid(const TV_INT2& edge,const CODIMENSIONAL_FLUID_PARTICLES<TV>& particles, T_SOLID& solid);

    static TV Tank_Normal_And_Position(const TV& pos,const WATER_TANK<TV>& tank,/*rst*/TV& wall_pos);
    static bool Contact_Tank(const TV& pos,const WATER_TANK<TV>& tank);
};
}
#endif