//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "../Viscoelastic/SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL.h"
#include "../Intersection/CODIMENSIONAL_INTERSECTION.h"
#include "../Meshing/CODIMENSIONAL_MESH_HASHTABLE.h"
using namespace PhysBAM;

template<class TV>
PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL(FIBER<TV>& fiber_input,CODIMENSIONAL_MESHING_3D<T>& mesher_input,CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible_input,
    CODIMENSIONAL_VISCOSITY_FORCE<TV>& viscosity_force_input,CODIMENSIONAL_VISCOELASTIC_FORCE<TV>& elasticity_force_input,ARRAY<RIGID_BODY<TV>*>& rigid_bodies_input,CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>& spatial_partition_input)
    :fiber(fiber_input),fluid_mesh(mesher_input.mesh),fluid_particles(mesher_input.particles),mesher(mesher_input),incompressible(incompressible_input),
    viscosity_force(viscosity_force_input),elasticity_force(elasticity_force_input),rigid_bodies(rigid_bodies_input),spatial_partition(spatial_partition_input),v_interp_weight((T).9),use_fiber_fluid_coupling_force(true),
    use_squash_force(false),use_trace_fiber_segments(false),apply_force_on_fiber(true),fiber_adhesion_forces(0),mpi_codim(0)
{
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Set_Triangle_Length_Scale(T length_scale_min_input,T length_scale_max_input,T band_width_min_input,T band_width_max_input)
{triangle_length_scale_min=length_scale_min_input;triangle_length_scale_max=length_scale_max_input;triangle_band_width_min=band_width_min_input;triangle_band_width_max=band_width_max_input;}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Set_Tetrahedron_Length_Scale(T length_scale_min_input,T length_scale_max_input,T band_width_min_input,T band_width_max_input)
{tetrahedron_length_scale_min=length_scale_min_input;tetrahedron_length_scale_max=length_scale_max_input;tetrahedron_band_width_min=band_width_min_input;tetrahedron_band_width_max=band_width_max_input;}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Advance_One_Timestep(const T dt,const T time)
{
    Advance_One_Timestep_Solid(dt,time);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Advance_One_Timestep_Solid(const T dt,const T time)
{
    fiber.Store_Old_States();
    fiber.Advance(dt,time);
    if(fiber.use_guided_fibers)fiber.Update_Guided_Fibers();
    fiber.Update_Implicit_Geometry();
    if(use_squash_force)Squash_Fluid();
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(const CODIMENSIONAL_MESH& mesh,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const int iter_num/*=2*/,bool length_scale_coarsen/*=true*/,ARRAY<bool>* flag_changeable/*=0*/)
{
    HASHTABLE<TV_INT2> edge_hashtable;ARRAY<RANGE<TV> > domain;domain.Append(RANGE<TV>(TV::Constant_Vector((T)-FLT_MAX),TV::Constant_Vector((T)FLT_MAX)));
    Update_Edge_Hashtable(mesh.triangle_mesh.elements,particles.X,domain,(T)0,edge_hashtable);
    for(int iter=1;iter<=iter_num;iter++)for(HASHTABLE_ITERATOR<TV_INT2> iter(edge_hashtable);iter.Valid();iter.Next()){
        const TV_INT2& v=iter.Key();
        for(int i=1;i<=2;i++){
            if(!flag_changeable||(*flag_changeable)(v(i))){int other_v=(i==1?v(2):v(1));T frac=(T).5;
                if(length_scale_coarsen&&particles.length_scale(v(i))<(T).75*particles.length_scale(other_v)){ ////length scale up
                    particles.length_scale(v(i))=((T)1-frac)*particles.length_scale(v(i))+frac*particles.length_scale(other_v);
                }
                else if(!length_scale_coarsen&&particles.length_scale(v(i))>(T)1.5*particles.length_scale(other_v)){   ////length scale down
                    particles.length_scale(v(i))=((T)1-frac)*particles.length_scale(v(i))+frac*particles.length_scale(other_v);
                }
            }
        }
    }
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Length_Scale()
{
    //ARRAY<bool> flag_changeable_near_fiber(fluid_particles.X.m);flag_changeable_near_fiber.Fill(true);
    //ARRAY<bool> flag_changeable_high_dim(fluid_particles.X.m);flag_changeable_high_dim.Fill(true);
    //for(int i=1;i<=fluid_particles.X.m;i++){
    //    T band_width_min=triangle_band_width_min;T band_width_max=triangle_band_width_max;T length_scale_min=triangle_length_scale_min;T length_scale_max=triangle_length_scale_max;
    //    if(Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(i))){band_width_min=tetrahedron_band_width_min;band_width_max=tetrahedron_band_width_max;length_scale_min=tetrahedron_length_scale_min;length_scale_max=tetrahedron_length_scale_max;}
    //    int p_type=fluid_mesh.node_types(i);if(Is_On_Tetrahedron_Mesh(p_type)){fluid_particles.length_scale(i)=length_scale;flag_changeable_high_dim(i)=false;}

    //    int fiber_segment_index=0;T phi=fiber.Signed_Distance(fluid_particles.X(i),fiber_segment_index);phi=clamp(phi,band_width_min,triangle_band_width_max);
    //    if(!fiber.near_end_segment_hashtable.Contains(fiber_segment_index))continue;
    //    if(use_trace_fiber_segments&&!fiber.fiber_trace_segment_hashtable.Contains(fiber_segment_index))continue;

    //    ////all particles near trace fibers
    //    T frac=(phi-band_width_min)/(band_width_max-band_width_min);T particle_length_scale=((T)1-frac)*length_scale_min+frac*length_scale_max;
    //    if(fluid_particles.length_scale(i)>particle_length_scale){fluid_particles.length_scale(i)=particle_length_scale;flag_changeable_near_fiber(i)=false;flag_changeable_high_dim(i)=true;}
    //}
    //Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*refine*/false,&flag_changeable_near_fiber);
    //Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*coarsen*/true,&flag_changeable_high_dim);
    //if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.length_scale);

    ARRAY<bool> flag_changeable_near_fiber(fluid_particles.X.m);flag_changeable_near_fiber.Fill(true);
    ARRAY<bool> flag_changeable_high_dim(fluid_particles.X.m);flag_changeable_high_dim.Fill(true);
    for(int i=1;i<=fluid_particles.X.m;i++){
        T band_width_min=triangle_band_width_min;T band_width_max=triangle_band_width_max;T length_scale_min=triangle_length_scale_min;T length_scale_max=triangle_length_scale_max;
        //if(Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(i))){band_width_min=tetrahedron_band_width_min;band_width_max=tetrahedron_band_width_max;length_scale_min=tetrahedron_length_scale_min;length_scale_max=tetrahedron_length_scale_max;}
        int p_type=fluid_mesh.node_types(i);if(Is_On_Tetrahedron_Mesh(p_type)){fluid_particles.length_scale(i)=length_scale;flag_changeable_high_dim(i)=false;flag_changeable_near_fiber(i)=false;continue;}

        int fiber_segment_index=0;T phi=fiber.Signed_Distance(fluid_particles.X(i),fiber_segment_index);phi=clamp(phi,band_width_min,triangle_band_width_max);
        if(!fiber.near_end_segment_hashtable.Contains(fiber_segment_index))continue;
        if(use_trace_fiber_segments&&!fiber.fiber_trace_segment_hashtable.Contains(fiber_segment_index))continue;

        T frac=(phi-band_width_min)/(band_width_max-band_width_min);T particle_length_scale=((T)1-frac)*length_scale_min+frac*length_scale_max;
        if(fluid_particles.length_scale(i)>particle_length_scale){fluid_particles.length_scale(i)=particle_length_scale;flag_changeable_near_fiber(i)=false;}
    }
    Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*refine*/false,&flag_changeable_near_fiber);
    Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*coarsen*/true,&flag_changeable_high_dim);
    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.length_scale);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Length_Scale_Using_Rigid_Body(RIGID_BODY<TV>* rb)
{
    ARRAY<bool> flag_changeable_near_fiber(fluid_particles.X.m);flag_changeable_near_fiber.Fill(true);
    ARRAY<bool> flag_changeable_high_dim(fluid_particles.X.m);flag_changeable_high_dim.Fill(true);
    for(int i=1;i<=fluid_particles.X.m;i++){
        T band_width_min=triangle_band_width_min;T band_width_max=triangle_band_width_max;T length_scale_min=triangle_length_scale_min;T length_scale_max=triangle_length_scale_max;
        if(Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(i))){band_width_min=tetrahedron_band_width_min;band_width_max=tetrahedron_band_width_max;length_scale_min=tetrahedron_length_scale_min;length_scale_max=tetrahedron_length_scale_max;}

        T phi=(T)0;
        //rb->Implicit_Geometry_Lazy_Inside_And_Value(fluid_particles.X(i),phi);
        phi=rb->Implicit_Geometry_Extended_Value(fluid_particles.X(i));
        phi=clamp(phi,band_width_min,triangle_band_width_max);

        T frac=(phi-band_width_min)/(band_width_max-band_width_min);T particle_length_scale=((T)1-frac)*length_scale_min+frac*length_scale_max;
        if(fluid_particles.length_scale(i)>particle_length_scale){fluid_particles.length_scale(i)=particle_length_scale;flag_changeable_near_fiber(i)=false;}

        int p_type=fluid_mesh.node_types(i);if(Is_On_Tetrahedron_Mesh(p_type)){fluid_particles.length_scale(i)=length_scale;flag_changeable_high_dim(i)=false;}
    }
    Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*refine*/false,&flag_changeable_near_fiber);
    Smooth_Particle_Length_Scale(fluid_mesh,fluid_particles,2,/*coarsen*/true,&flag_changeable_high_dim);
    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.length_scale);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Contact()
{
    TV canvas_normal=TV::Axis_Vector(2);
    //T min_triangle_thickness=mesher.thickness_scale*(T)1e-3;
    for(int i=1;i<=fluid_particles.X.m;i++){
        int nearest_fiber_segment_index=fiber.Nearest_Fiber_Segment_Index(fluid_particles.X(i));
        if(!fiber.near_end_segment_hashtable.Contains(nearest_fiber_segment_index))continue;
        //TV p1=fluid_particles.X(i);TV p2=fluid_particles.X(i)+fluid_particles.thickness(i)*canvas_normal;T t=(T)0;
        //if(fiber.Fiber_Segment_Intersect(p1,p2,nearest_fiber_segment_index,t)){Set_Contact_With_Fiber(fluid_particles.contact(i));}
        TV endpoint_pos=fluid_particles.X(i)+fluid_particles.thickness(i)*canvas_normal;
        if(fiber.Fiber_Segment_Inside(endpoint_pos,nearest_fiber_segment_index)){Set_Contact_With_Fiber(fluid_particles.contact(i));}
    }

    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.contact);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Inactive()
{
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(fiber.bounding_box.Lazy_Inside(fluid_particles.X(i)))fluid_particles.inactive(i)=false;
        else fluid_particles.inactive(i)=true;
    }
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Squash_Fluid()
{
    ARRAY<T> target_thickness;target_thickness.Resize(fluid_particles.X.m);
    for(int i=1;i<=fluid_particles.X.m;i++)target_thickness(i)=fluid_particles.thickness(i);
    //LOG::cout<<"squash "<<collision_pairs.m<<std::endl;

    TV canvas_normal=TV::Axis_Vector(2);
    T min_triangle_thickness=mesher.thickness_scale*(T)1e-3;
    for(int i=1;i<=fluid_particles.X.m;i++){
        int nearest_fiber_segment_index=fiber.Nearest_Fiber_Segment_Index(fluid_particles.X(i));
        if(!fiber.near_end_segment_hashtable.Contains(nearest_fiber_segment_index))continue;
        if(use_trace_fiber_segments&&!fiber.fiber_trace_segment_hashtable.Contains(nearest_fiber_segment_index))continue;

        TV p1=fluid_particles.X(i);TV p2=fluid_particles.X(i)+fluid_particles.thickness(i)*canvas_normal;T t=(T)0;T tolerance=(T)1e-2;
        if(fiber.Fiber_Segment_Intersect(p1,p2,nearest_fiber_segment_index,t)){
            target_thickness(i)=clamp_min(fluid_particles.thickness(i)*(t+tolerance),min_triangle_thickness);
        }
    }

    ////TODO: bug fix for thickness changes during iterations
    ARRAY<T> delta_mass;delta_mass.Resize(fluid_particles.X.m);delta_mass.Fill((T)0);int number_of_rings=3;
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(target_thickness(i)>=fluid_particles.thickness(i))continue;
        T particle_delta_frac=clamp_min((fluid_particles.thickness(i)-target_thickness(i))/fluid_particles.thickness(i),(T)0);
        T particle_delta_mass=fluid_particles.dimensional_mass(i)(TV::m-1)*particle_delta_frac;
        delta_mass(i)-=particle_delta_mass;

        T max_distance_squared=pow<2>(fluid_particles.length_scale(i)*(T)number_of_rings);
        ARRAY<int> nbs;Find_Neighbor_Particles_On_Mesh(fluid_mesh.triangle_mesh,i,nbs,number_of_rings);
        ARRAY<T> weights(nbs.m);T total_weight=(T)0;
        for(int j=1;j<=nbs.m;j++){int nb_p=nbs(j);if(nb_p==i)continue;
            if(target_thickness(nb_p)<fluid_particles.thickness(nb_p))continue;
            T distance_squared=(fluid_particles.X(i)-fluid_particles.X(nb_p)).Magnitude_Squared();
            T normalized_distance_squared=clamp(distance_squared/max_distance_squared,(T)0,(T)1);
            T frac=Gaussian_Kernel(normalized_distance_squared);
            weights(j)=frac;total_weight+=frac;
        }
        if(total_weight>(T)0){
            for(int j=1;j<=nbs.m;j++){weights(j)/=total_weight;}
            for(int j=1;j<=nbs.m;j++){int nb_p=nbs(j);delta_mass(nb_p)+=particle_delta_mass*weights(j);}
        }
    }

    for(int i=1;i<=fluid_particles.X.m;i++){
        fluid_particles.dimensional_mass(i)(TV::m-1)+=(T).99*delta_mass(i);
        fluid_particles.mass(i)=fluid_particles.dimensional_mass(i).Sum();
    }   
    mesher.Update_Particle_Thickness(mesher.has_rim);

    if(mpi_codim){
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.dimensional_mass);
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.mass);
        mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.thickness);
    }
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Boundary_Conditions()
{
    ARRAY<CODIMENSIONAL_NEUMANN_BOUNDARY<TV> >& incompressible_psi_N=incompressible.psi_N_on_mesh;
    VECTOR<ARRAY<CODIMENSIONAL_DIRICHLET_BOUNDARY<T> >,TV::m>& viscosity_psi_Ds=viscosity_force.psi_Ds_on_mesh;

    //////update boundary conditions for incompressible and viscosity solvers
    ARRAY<TV> fiber_v(fluid_particles.X.m);fiber_v.Fill(TV());ARRAY<T> weights(fluid_particles.X.m);weights.Fill((T)0);
    for(int i=1;i<=collision_pairs.m;i++){
        const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
        int fiber_segment=pair.fiber_segment_index;const TV_INT2& fiber_segment_vtx=fiber.fiber_mesh.elements(fiber_segment);
        const TV segment_V=(T).5*(fiber.fiber_particles.V(fiber_segment_vtx(1))+fiber.fiber_particles.V(fiber_segment_vtx(2)));
        const TV segment_X=(T).5*(fiber.fiber_particles.X(fiber_segment_vtx(1))+fiber.fiber_particles.X(fiber_segment_vtx(2)));

        switch(pair.fluid_element_type){
        case SIMPLEX_TETRAHEDRON:{
            const TV_INT4& v=fluid_mesh.tetrahedron_mesh.elements(pair.fluid_element_index);
            for(int d=1;d<=4;d++){
                if(Is_On_Tetrahedron_Mesh_Boundary(fluid_mesh.node_types(v(d)))){
                    T distance_squared=(segment_X-fluid_particles.X(v(d))).Magnitude_Squared();fiber_v(v(d))+=segment_V*distance_squared;weights(v(d))+=distance_squared;
                }
            }
        }break;
        case SIMPLEX_TRIANGLE:{
            const TV_INT3& v=fluid_mesh.triangle_mesh.elements(pair.fluid_element_index);
            TV canvas_normal=TV::Axis_Vector(2);
            for(int d=1;d<=3;d++){
                //TV p1=fluid_particles.X(v(d));TV p2=fluid_particles.X(v(d))+fluid_particles.thickness(v(d))*canvas_normal;T t=(T)0;
                //if(fiber.Fiber_Segment_Intersect(p1,p2,fiber_segment,t)){
                    T distance_squared=(segment_X-fluid_particles.X(v(d))).Magnitude_Squared();fiber_v(v(d))+=segment_V*distance_squared;weights(v(d))+=distance_squared;
                //}
            }
        }break;
        case SIMPLEX_SEGMENT:{}break;
        case SIMPLEX_PARTICLE:{}break;
        }
    }

    ////update psi_N for incompressibility and psi_D for viscosity
    for(int i=1;i<=fluid_particles.X.m;i++)if(weights(i)!=(T)0){
        fiber_v(i)/=weights(i);

        ARRAY<int> psi_N_indices;   ////psi_N already defined on this particle
        for(int j=1;j<=incompressible_psi_N.m;j++){if(incompressible_psi_N(j).index==i)psi_N_indices.Append(j);}

        //for(int d=1;d<=3;d++){
        //    TV dir=TV::Axis_Vector(d);T v=fiber_v(i)(d);

        //    bool has_psi_N_on_axis=false;
        //    for(int j=1;j<=psi_N_indices.m;j++){if(incompressible_psi_N(psi_N_indices(j)).direction==dir){has_psi_N_on_axis=true;break;}}
        //    if(has_psi_N_on_axis)continue;

        //    incompressible_psi_N.Append(CODIMENSIONAL_NEUMANN_BOUNDARY<TV>(i,dir,v));
        //    //LOG::cout<<"fiber psi_N "<<i<<", dir: "<<dir<<", v: "<<v<<std::endl;
        //}
        
        for(int d=1;d<=3;d++){
            CODIMENSIONAL_DIRICHLET_BOUNDARY<T> viscosity_psi_D_sample(i,SIMPLEX_PARTICLE,fiber_v(i)(d));
            if(!viscosity_psi_Ds(d).Contains(viscosity_psi_D_sample)){
                viscosity_psi_Ds(d).Append(viscosity_psi_D_sample); 
            }
            else{
                int p=viscosity_psi_Ds(d).Find(viscosity_psi_D_sample);
                viscosity_psi_Ds(d)(p)=viscosity_psi_D_sample;
            } ////overwrite psi_D
        }
        //LOG::cout<<"set psi_D: "<<i<<": "<<fiber_v(i)<<std::endl;
    }

    ////TODO: mpi sync
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Target_Divergence()
{
    ARRAY<CODIMENSIONAL_TARGET_DIVERGENCE<T> >& tar_div_on_mesh=incompressible.tar_div_on_mesh;

    ARRAY<TV> fiber_v(fluid_particles.X.m);fiber_v.Fill(TV());ARRAY<T> weights(fluid_particles.X.m);weights.Fill((T)0);
    HASHTABLE<TV_INT3,T> tri_tar_div_hashtable;
    for(int i=1;i<=collision_pairs.m;i++){
        const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
        int fiber_segment=pair.fiber_segment_index;const TV_INT2& fiber_segment_vtx=fiber.fiber_mesh.elements(fiber_segment);
        const TV segment_V=(T).5*(fiber.fiber_particles.V(fiber_segment_vtx(1))+fiber.fiber_particles.V(fiber_segment_vtx(2)));
        const TV segment_X=(T).5*(fiber.fiber_particles.X(fiber_segment_vtx(1))+fiber.fiber_particles.X(fiber_segment_vtx(2)));

        switch(pair.fluid_element_type){
        case SIMPLEX_TETRAHEDRON:{}break;
        case SIMPLEX_TRIANGLE:{
            int t=pair.fluid_element_index;
            const TV_INT3& v=fluid_mesh.triangle_mesh.elements(t);
            //TV canvas_normal=TV::Axis_Vector(2);
            //TV triangle_normal=canvas_normal;
            //T normal_fiber_vel=TV::Dot_Product(segment_V,triangle_normal);
            //T normal_rel_vel=(T).1;
            //T A=T_TRIANGLE::Area(fluid_particles.X(v(1)),fluid_particles.X(v(2)),fluid_particles.X(v(3)));
            //T t_div=normal_rel_vel*A;
            T coef=(T)1e1;T t_div=Prism_Lost_Volume(v,fiber_segment)*coef;
            if(t_div>(T)0){
                TV_INT guid_key=Unique_Ordered(Element(fluid_particles.guid,v));
                tri_tar_div_hashtable.Set(guid_key,t_div);
            }
        }break;
        case SIMPLEX_SEGMENT:{}break;
        case SIMPLEX_PARTICLE:{}break;
        }
    }

    ////update target div
    CODIMENSIONAL_MESH& solver_mesh=incompressible.mesh;
    for(int i=1;i<=solver_mesh.triangle_mesh.elements.m;i++){
        const TV_INT3& v=solver_mesh.triangle_mesh.elements(i);
        TV_INT guid_key=Unique_Ordered(Element(fluid_particles.guid,v));
        T* tri_tar_div=tri_tar_div_hashtable.Get_Pointer(guid_key);
        if(tri_tar_div){
            tar_div_on_mesh.Append(CODIMENSIONAL_TARGET_DIVERGENCE<T>(i,SIMPLEX_TRIANGLE,*tri_tar_div));
            //LOG::cout<<"tar_div: "<<i<<": "<<*tri_tar_div<<std::endl;
        }
    }

    ////TODO: mpi sync
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Correct_Fiber_Particle_Positions()
{
    ////fiber-solid interaction
    if(rigid_bodies.m>0){
        for(int i=1;i<=fiber.fiber_particles.X.m;i++){
            for(int r=1;r<=rigid_bodies.m;r++){
                RIGID_BODY<TV>* rigid_body=rigid_bodies(r);
                T phi=(T)0;//T contour_value=(T)length_scale*(T).1;
                if(rigid_body->Implicit_Geometry_Lazy_Inside_And_Value(fiber.fiber_particles.X(i),phi)){
                    TV rigid_normal=rigid_body->Implicit_Geometry_Normal(fiber.fiber_particles.X(i));
                    TV rigid_tangential_1=rigid_normal.Unit_Orthogonal_Vector();
                    TV rigid_tangential_2=TV::Cross_Product(rigid_normal,rigid_tangential_1);

                    fiber.fiber_particles.X(i)-=rigid_normal*phi;
                    TV rigid_v=RIGID_GEOMETRY<TV>::Pointwise_Object_Velocity(rigid_body->Twist(),rigid_body->X(),fiber.fiber_particles.X(i));
                    T v_in_normal_direction=TV::Dot_Product(rigid_normal,rigid_v);
                    //T v_in_tangential_direction=(T)0;
                    //LOG::cout<<"contact fiber v "<<i<<": "<<fiber.fiber_particles.V(i)<<" + "<<(v_in_normal_direction-TV::Dot_Product(rigid_normal,fiber.fiber_particles.V(i)))*rigid_normal<<std::endl;
                    fiber.fiber_particles.V(i)+=(v_in_normal_direction-TV::Dot_Product(rigid_normal,fiber.fiber_particles.V(i)))*rigid_normal;

                    //fiber.fiber_particles.V(i)+=(v_in_tangential_direction-TV::Dot_Product(rigid_tangential_1,fiber.fiber_particles.V(i)))*rigid_tangential_1;
                    //fiber.fiber_particles.V(i)+=(v_in_tangential_direction-TV::Dot_Product(rigid_tangential_2,fiber.fiber_particles.V(i)))*rigid_tangential_2;
                    //fiber.fiber_particles.V(i)=TV();
                }
            }
        }
    }
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Incompressibility()
{
    //ARRAY<bool>& triangle_incompressibilities=mesher.triangle_incompressibilities;
    //for(int i=1;i<=fluid_mesh.triangle_mesh.elements.m;i++){
    //    const TV_INT3& v=fluid_mesh.triangle_mesh.elements(i);
    //    TV pos1=(T)one_third*(fluid_particles.X(v(1))+fluid_particles.X(v(2))+fluid_particles.X(v(3)));
    //    TV pos2=pos1+TV::Axis_Vector(2)*(T)one_third*(fluid_particles.thickness(v(1))+fluid_particles.thickness(v(2))+fluid_particles.thickness(v(3)));////TODO: fix this with particle normal
    //    //int intersected_fiber_p1=0;T phi1=Signed_Distance(pos1,intersected_fiber_p1);
    //    int intersected_fiber_p2=0;T phi2=fiber.Signed_Distance(pos2,intersected_fiber_p2);

    //    if(phi2<=(T)0&&intersected_fiber_p2!=0){
    //        triangle_incompressibilities(i)=true;
    //        for(int d=1;d<=3;d++)fluid_particles.incompressibility(v(d))=true;
    //    }
    //}

    ARRAY<bool>& triangle_incompressibilities=mesher.triangle_incompressibilities;
    //for(int i=1;i<=collision_pairs.m;i++){
    //    const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
    //    switch(pair.fluid_element_type){
    //    case SIMPLEX_TRIANGLE:{const TV_INT3& v=fluid_mesh.triangle_mesh.elements(pair.fluid_element_index);
    //        if(!Is_Contact_With_Solid(fluid_particles.contact(v(1)))||!Is_Contact_With_Solid(fluid_particles.contact(v(2)))||!Is_Contact_With_Solid(fluid_particles.contact(v(3)))||
    //            Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(v(1)))||Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(v(2)))||Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(v(3))))continue;
    //        triangle_incompressibilities(pair.fluid_element_index)=true;for(int d=1;d<=3;d++)fluid_particles.incompressibility(v(d))=true;                      
    //    }break;
    //    case SIMPLEX_SEGMENT:{/*TODO:implement this*/}break;
    //    }
    //}

    TV canvas_normal=TV::Axis_Vector(2);
    ARRAY<bool> particle_incomp;particle_incomp.Resize(fluid_particles.X.m);particle_incomp.Fill(false);
    ARRAY<bool> particle_checked;particle_checked.Resize(fluid_particles.X.m);particle_checked.Fill(false);
    for(int i=1;i<=collision_pairs.m;i++){
        const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
        switch(pair.fluid_element_type){
        case SIMPLEX_TRIANGLE:{
            int tri=pair.fluid_element_index;
            const TV_INT3& v=fluid_mesh.triangle_mesh.elements(tri);
            for(int j=1;j<=3;j++){
                int p=v(j);if(particle_checked(p))continue;
                int nearest_fiber_segment_index=fiber.Nearest_Fiber_Segment_Index(fluid_particles.X(p));
                //if(!fiber.near_end_segment_hashtable.Contains(nearest_fiber_segment_index))continue;
                //if(use_trace_fiber_segments&&!fiber.fiber_trace_segment_hashtable.Contains(nearest_fiber_segment_index))continue;

                TV p1=fluid_particles.X(p);TV p2=fluid_particles.X(p)+fluid_particles.thickness(p)*canvas_normal;T t=(T)0;T tolerance=(T)1e-2;
                if(fiber.Fiber_Segment_Intersect(p1,p2,nearest_fiber_segment_index,t)){particle_incomp(p)=true;}
                particle_checked(p)=true;
            }
        }break;
        case SIMPLEX_SEGMENT:{/*TODO:implement this*/}break;
        }
    }

    //for(int i=1;i<=fluid_particles.X.m;i++){
    //    int nearest_fiber_segment_index=fiber.Nearest_Fiber_Segment_Index(fluid_particles.X(i));
    //    //if(!fiber.near_end_segment_hashtable.Contains(nearest_fiber_segment_index))continue;
    //    //if(use_trace_fiber_segments&&!fiber.fiber_trace_segment_hashtable.Contains(nearest_fiber_segment_index))continue;

    //    TV p1=fluid_particles.X(i);TV p2=fluid_particles.X(i)+fluid_particles.thickness(i)*canvas_normal;T t=(T)0;T tolerance=(T)1e-2;
    //    if(fiber.Fiber_Segment_Intersect(p1,p2,nearest_fiber_segment_index,t)){particle_incomp(i)=true;}
    //}

    for(int i=1;i<=fluid_mesh.triangle_mesh.elements.m;i++){
        const TV_INT3& v=fluid_mesh.triangle_mesh.elements(i);
        if(particle_incomp(v(1))&&particle_incomp(v(2))&&particle_incomp(v(3))){
            triangle_incompressibilities(i)=true;for(int d=1;d<=3;d++)fluid_particles.incompressibility(v(d))=true;                      
        }
    }

    ////TODO: mpi sync
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Detect_Fiber_Fluid_Collisions(ARRAY<FIBER_FLUID_COLLISION_PAIR>& collision_pairs)
{
    for(int i=1;i<=fiber.implicit_fiber_elements.m;i++){
        int s=fiber.implicit_fiber_elements(i);
        const TV_INT2& fiber_segment=fiber.fiber_mesh.elements(s);

        const RANGE<TV_INT>& voxel_range=fiber.fiber_element_voxel_ranges(s);
        ARRAY<CODIMENSIONAL_ELEMENT_INDEX> potential_collision_elements;
        spatial_partition.Get_From_Cell(voxel_range,1,potential_collision_elements);
        for(int j=1;j<=potential_collision_elements.m;j++){
            const CODIMENSIONAL_ELEMENT_INDEX& element=potential_collision_elements(j);
            switch(element.type){
            case SIMPLEX_TETRAHEDRON:{
                const TV_INT4& fluid_tet=fluid_mesh.tetrahedron_mesh.elements(element.index);
                //bool intersect=CODIMENSIONAL_INTERSECTION<T>::Geometries_Intersect(fiber.fiber_particles.X,fluid_particles.X,fiber.fiber_particle_thickness,fluid_particles.thickness,fiber_segment,fluid_tet);
                //if(intersect){
                //    collision_pairs.Append(FIBER_FLUID_COLLISION_PAIR(s,element.index,element.type));
                //}
                
                T fiber_thickness=(fiber.fiber_particle_thickness(fiber_segment(1))+fiber.fiber_particle_thickness(fiber_segment(2)))*(T).5;
                T coef=(T)3;
                for(int i=1;i<=4;i++){
                    int p=fluid_tet(i);T unused=(T)0;
                    T distance_vtx_to_fiber=Distance_From_Point_To_Segment(fluid_particles.X(p),SEGMENT_3D<T>(fiber.fiber_particles.X(fiber_segment(1)),fiber.fiber_particles.X(fiber_segment(2))),unused);
                    if(distance_vtx_to_fiber<fiber_thickness*coef){
                        collision_pairs.Append(FIBER_FLUID_COLLISION_PAIR(s,element.index,element.type));
                    }
                }
            }break;
            case SIMPLEX_TRIANGLE:{
                const TV_INT3& fluid_tri=fluid_mesh.triangle_mesh.elements(element.index);
                TV canvas_normal=TV::Axis_Vector(2);
                bool intersect=CODIMENSIONAL_INTERSECTION<T>::Capsule_Triangle_Thickness_Intersection(fiber.fiber_particles.X,fluid_particles.X,fiber.fiber_particle_thickness,fluid_particles.thickness,fiber_segment,fluid_tri,canvas_normal);
                if(intersect){collision_pairs.Append(FIBER_FLUID_COLLISION_PAIR(s,element.index,element.type));}
            }break;
            case SIMPLEX_SEGMENT:{
                ////TODO: implement this
            }break;
            case SIMPLEX_PARTICLE:{
                ////TODO: implement this
            }break;
            }
        }
    }
    //LOG::cout<<"collision pairs detected: "<<collision_pairs.m<<std::endl;
}

template<class TV>
typename TV::SCALAR PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Prism_Lost_Volume(const TV_INT3& fluid_tri_v,const int& fiber_seg_index)
{
    TV canvas_normal=TV::Axis_Vector(2);
    TV p1=(T)one_third*(fluid_particles.X(fluid_tri_v(1))+fluid_particles.X(fluid_tri_v(2))+fluid_particles.X(fluid_tri_v(3)));
    T averaged_thickness=(T)one_third*(fluid_particles.thickness(fluid_tri_v(1))+fluid_particles.thickness(fluid_tri_v(2))+fluid_particles.thickness(fluid_tri_v(3)));
    TV p2=p1+canvas_normal*averaged_thickness;
    T t=(T)0;T lost_vol=(T)0;
    if(fiber.Fiber_Segment_Intersect(p1,p2,fiber_seg_index,t)){
        T triangle_area=TRIANGLE_3D<T>::Area(fluid_particles.X(fluid_tri_v(1)),fluid_particles.X(fluid_tri_v(2)),fluid_particles.X(fluid_tri_v(3)));
        lost_vol=((T)1-t)*averaged_thickness*triangle_area;
    }
    return lost_vol;
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Add_Fiber_Fluid_Coupling_Forces(ARRAY_VIEW<TV>& F,const T dt,const T time)
{
    ARRAY<ARRAY<int> > fiber_to_fluid_interpolation_indices(fluid_particles.X.m);
    ARRAY<ARRAY<T> > fiber_to_fluid_interpolation_weights(fluid_particles.X.m);
    for(int i=1;i<=collision_pairs.m;i++){
        const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
        int fiber_segment_index=pair.fiber_segment_index;const TV_INT2& fiber_segment_vtx=fiber.fiber_mesh.elements(fiber_segment_index);

        switch(pair.fluid_element_type){
        case SIMPLEX_TETRAHEDRON:{
            const TV_INT4& v=fluid_mesh.tetrahedron_mesh.elements(pair.fluid_element_index);
            for(int d=1;d<=4;d++){
                fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(1));
                fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(2));
            }
        }break;
        case SIMPLEX_TRIANGLE:{
            const TV_INT3& v=fluid_mesh.triangle_mesh.elements(pair.fluid_element_index);
            for(int d=1;d<=3;d++){
                //if(Is_Contact_With_Solid2(fluid_particles.contact(v(d)))){
                    fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(1));
                    fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(2));
                //}
            }
        }break;
        case SIMPLEX_SEGMENT:{}break;   
        case SIMPLEX_PARTICLE:{}break;
        }
    }

    T max_distance_squared=pow<2>(fiber.fiber_radius*(T)2);
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(fiber_to_fluid_interpolation_indices(i).m>0){
            fiber_to_fluid_interpolation_weights(i).Resize(fiber_to_fluid_interpolation_indices(i).m);
            T weight_sum=(T)0;
            for(int j=1;j<=fiber_to_fluid_interpolation_weights(i).m;j++){
                int fiber_p=fiber_to_fluid_interpolation_indices(i)(j);
                T distance_squared=(fiber.fiber_particles.X(fiber_p)-fluid_particles.X(i)).Magnitude_Squared();
                T weight=Gaussian_Kernel(clamp(distance_squared/max_distance_squared,(T)0,(T)1));
                fiber_to_fluid_interpolation_weights(i)(j)=weight;weight_sum+=weight;
            }
            for(int j=1;j<=fiber_to_fluid_interpolation_weights(i).m;j++){
                fiber_to_fluid_interpolation_weights(i)(j)/=weight_sum;
            }
        }
    }

    ////fiber to fluid impulse forces
    ARRAY<TV> fiber_to_fluid_impulse_force(fluid_particles.X.m);fiber_to_fluid_impulse_force.Fill(TV());
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(fiber_to_fluid_interpolation_indices(i).m>0){
            TV fiber_v=TV();
            for(int j=1;j<=fiber_to_fluid_interpolation_indices(i).m;j++){
                int fiber_p=fiber_to_fluid_interpolation_indices(i)(j);
                T fiber_w=fiber_to_fluid_interpolation_weights(i)(j);
                fiber_v+=fiber.fiber_particles.V(fiber_p)*fiber_w;
            }
            //TV target_v=(T).5*(fiber_v+fluid_particles.V(i));
            TV target_v=fluid_particles.V(i)+v_interp_weight*(fiber_v-fluid_particles.V(i));
            //if(Is_On_Triangle_Mesh(fluid_mesh.node_types(i))&&!Is_On_Tetrahedron_Mesh(fluid_mesh.node_types(i)))target_v=fluid_particles.V(i)+(T).9*(fiber_v-fluid_particles.V(i));    ////hack
            fiber_to_fluid_impulse_force(i)=fluid_particles.mass(i)*(target_v-fluid_particles.V(i))/dt;
        }
    }

    if(apply_force_on_fiber){
        ////fluid to fiber impulse forces
        ARRAY<TV> fluid_to_fiber_impulse_force(fiber.fiber_particles.X.m);fluid_to_fiber_impulse_force.Fill(TV());
        for(int i=1;i<=fluid_particles.X.m;i++){
            if(fiber_to_fluid_interpolation_indices(i).m>0){
                const TV& impulse_f=fiber_to_fluid_impulse_force(i);
                for(int j=1;j<=fiber_to_fluid_interpolation_indices(i).m;j++){
                    int fiber_p=fiber_to_fluid_interpolation_indices(i)(j);
                    T fiber_w=fiber_to_fluid_interpolation_weights(i)(j);
                    fluid_to_fiber_impulse_force(fiber_p)+=fiber_w*(-impulse_f);
                }
            }
        }

        if(fiber_adhesion_forces){
            fiber_adhesion_forces->forces.Resize(fiber.fiber_particles.X.m);
            for(int i=1;i<=fluid_to_fiber_impulse_force.m;i++){
                fiber_adhesion_forces->forces(i)=fluid_to_fiber_impulse_force(i);
                fiber.fiber_particle_forces(i)=fiber_adhesion_forces->forces(i);
            }
        }
    }

    for(int i=1;i<=fiber_to_fluid_impulse_force.m;i++)F(i)+=fiber_to_fluid_impulse_force(i);

    ////TODO: synchronize the mpi forces for fiber
    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(F);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Update_Particle_Meshing_Direction()
{
    ARRAY<ARRAY<int> > fiber_to_fluid_interpolation_indices(fluid_particles.X.m);
    ARRAY<ARRAY<T> > fiber_to_fluid_interpolation_weights(fluid_particles.X.m);
    ARRAY<TV> meshing_direction(fluid_particles.X.m);meshing_direction.Fill(TV());
    ARRAY<int> flag(fluid_particles.X.m);flag.Fill(0);

    for(int i=1;i<=collision_pairs.m;i++){
        const FIBER_FLUID_COLLISION_PAIR& pair=collision_pairs(i);
        int fiber_segment_index=pair.fiber_segment_index;const TV_INT2& fiber_segment_vtx=fiber.fiber_mesh.elements(fiber_segment_index);

        switch(pair.fluid_element_type){
        case SIMPLEX_TETRAHEDRON:{}break;
        case SIMPLEX_TRIANGLE:{
            const TV_INT3& v=fluid_mesh.triangle_mesh.elements(pair.fluid_element_index);
            for(int d=1;d<=3;d++){
                fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(1));
                fiber_to_fluid_interpolation_indices(v(d)).Append_Unique(fiber_segment_vtx(2));
            }
        }break;
        case SIMPLEX_SEGMENT:{}break;   
        case SIMPLEX_PARTICLE:{}break;
        }
    }

    T max_distance_squared=pow<2>(fiber.fiber_radius*(T)2);
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(fiber_to_fluid_interpolation_indices(i).m>0){
            fiber_to_fluid_interpolation_weights(i).Resize(fiber_to_fluid_interpolation_indices(i).m);
            T weight_sum=(T)0;
            for(int j=1;j<=fiber_to_fluid_interpolation_weights(i).m;j++){
                int fiber_p=fiber_to_fluid_interpolation_indices(i)(j);
                T distance_squared=(fiber.fiber_particles.X(fiber_p)-fluid_particles.X(i)).Magnitude_Squared();
                T weight=Gaussian_Kernel(clamp(distance_squared/max_distance_squared,(T)0,(T)1));
                fiber_to_fluid_interpolation_weights(i)(j)=weight;weight_sum+=weight;
            }
            for(int j=1;j<=fiber_to_fluid_interpolation_weights(i).m;j++){
                fiber_to_fluid_interpolation_weights(i)(j)/=weight_sum;
            }
        }
    }

    ARRAY<TV> fiber_to_fluid_impulse_force(fluid_particles.X.m);fiber_to_fluid_impulse_force.Fill(TV());int c=0;
    for(int i=1;i<=fluid_particles.X.m;i++){
        if(fiber_to_fluid_interpolation_indices(i).m>0){
            TV fiber_v=TV();
            for(int j=1;j<=fiber_to_fluid_interpolation_indices(i).m;j++){
                int fiber_p=fiber_to_fluid_interpolation_indices(i)(j);
                T fiber_w=fiber_to_fluid_interpolation_weights(i)(j);
                fiber_v+=fiber.fiber_particles.V(fiber_p)*fiber_w;
            }
            meshing_direction(i)=fiber_v.Normalized();flag(i)=c+1;
        }
    }

    ////diffuse direction
    HASHTABLE<TV_INT2> edge_hashtable;ARRAY<RANGE<TV> > domain;domain.Append(RANGE<TV>(TV::Constant_Vector((T)-FLT_MAX),TV::Constant_Vector((T)FLT_MAX)));
    //Update_Edge_Hashtable(fluid_mesh.tetrahedron_mesh.elements,fluid_particles.X,domain,(T)0,edge_hashtable);
    Update_Edge_Hashtable(fluid_mesh.triangle_mesh.elements,fluid_particles.X,domain,(T)0,edge_hashtable);
    for(int iter=1;iter<=2;iter++){c++;for(HASHTABLE_ITERATOR<TV_INT2> iter(edge_hashtable);iter.Valid();iter.Next()){
        const TV_INT2& v=iter.Key();if(flag(v(1))!=c&&flag(v(2))!=c)continue;
        TV new_dir=(meshing_direction(v(1))+meshing_direction(v(2)));if(new_dir!=TV())new_dir=new_dir.Normalized();
        for(int i=1;i<=2;i++){meshing_direction(v(i))=new_dir;if(flag(v(i))==0)flag(v(i))=c+1;}
    }}
    for(int i=1;i<=fluid_particles.X.m;i++)if(meshing_direction(i)!=TV())fluid_particles.meshing_direction(i)=meshing_direction(i);

    if(mpi_codim)mpi_codim->Exchange_Boundary_And_Ghost_Particle_Attributes(fluid_particles.meshing_direction);
}

template<class TV>
void PhysBAM::SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Initialize_Forces()
{
    if(use_fiber_fluid_coupling_force){
        fiber_adhesion_forces=new DEFORMABLE_ADHESION_FORCE<TV>(fiber.fiber_particles,fiber.fiber_mesh);
        fiber.deformable_body_collection.Add_Force(fiber_adhesion_forces);
    }
}

template class SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<VECTOR<float,3> >;
#ifndef COMPILE_WITHOUT_DOUBLE_SUPPORT
template class SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<VECTOR<double,3> >;
#endif


