//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL__
#define __SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL__
#include <PhysBAM_Solids/PhysBAM_Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <PhysBAM_Solids/PhysBAM_Deformables/Forces/POINTWISE_DEFORMABLE_FORCE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE.h>
#include <PhysBAM_Tools/Data_Structures/HASHTABLE_ITERATOR.h>
#include "../Viscoelastic/FIBER.h"
#include "../Solver/CODIMENSIONAL_INCOMPRESSIBLE.h"
#include "../Solver/CODIMENSIONAL_VISCOELASTIC_FORCE.h"
#include "../Intersection/CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION.h"

namespace PhysBAM
{
class FIBER_FLUID_COLLISION_PAIR
{public:
    int fiber_segment_index;int fluid_element_index;int fluid_element_type;
    FIBER_FLUID_COLLISION_PAIR(int fiber_segment_index=0,int fluid_element_index=0,int fluid_element_type=0):fiber_segment_index(fiber_segment_index),fluid_element_index(fluid_element_index),fluid_element_type(fluid_element_type){}
};

template<class TV>
class DEFORMABLE_ADHESION_FORCE: public POINTWISE_DEFORMABLE_FORCE<TV>
{typedef typename TV::SCALAR T;typedef POINTWISE_DEFORMABLE_FORCE<TV> BASE;
public:
    ARRAY<TV> forces;
    template<class T_MESH> DEFORMABLE_ADHESION_FORCE(PARTICLES<TV>& particles_input,const T_MESH& mesh):BASE(particles_input,mesh){}
    void Add_Velocity_Dependent_Forces(ARRAY_VIEW<const TV> V,ARRAY_VIEW<TV> F,const T time) const PHYSBAM_OVERRIDE{}
    void Add_Velocity_Independent_Forces(ARRAY_VIEW<TV> F,const T time) const PHYSBAM_OVERRIDE {for(int i=1;i<=forces.m;i++)F(i)+=forces(i);}
};

template<class TV>
class SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL
{
    typedef typename TV::SCALAR T;typedef typename TV::template REBIND<int>::TYPE TV_INT; typedef typename TV::SPIN T_SPIN;
    typedef VECTOR<T,2> TV2;typedef VECTOR<T,3> TV3;typedef VECTOR<int,2> TV_INT2;typedef VECTOR<int,3> TV_INT3;typedef VECTOR<int,4> TV_INT4;
    typedef typename IF<TV::m==2,TRIANGLE_2D<T>,TRIANGLE_3D<T> >::TYPE T_TRIANGLE;

public:
    FIBER<TV>& fiber;
    CODIMENSIONAL_MESH& fluid_mesh;
    CODIMENSIONAL_FLUID_PARTICLES<TV>& fluid_particles;
    CODIMENSIONAL_MESHING_3D<T>& mesher;
    CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible;
    CODIMENSIONAL_VISCOSITY_FORCE<TV>& viscosity_force;
    CODIMENSIONAL_VISCOELASTIC_FORCE<TV>& elasticity_force;
    ARRAY<RIGID_BODY<TV>*>& rigid_bodies;
    CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>& spatial_partition;

    static const int fluid_object_id=1;
    static const int fiber_object_id=2;

    T length_scale;
    T triangle_length_scale_min;
    T triangle_length_scale_max;
    T triangle_band_width_min;
    T triangle_band_width_max;
    T tetrahedron_length_scale_min;
    T tetrahedron_length_scale_max;
    T tetrahedron_band_width_min;
    T tetrahedron_band_width_max;
    T v_interp_weight;

    bool use_fiber_fluid_coupling_force;
    bool use_squash_force;
    bool use_trace_fiber_segments;
    bool apply_force_on_fiber;
    ARRAY<FIBER_FLUID_COLLISION_PAIR> collision_pairs;
    DEFORMABLE_ADHESION_FORCE<TV>* fiber_adhesion_forces;

    ////mpi
    MPI_CODIMENSIONAL<TV>* mpi_codim;

    SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL(FIBER<TV>& fiber_input,CODIMENSIONAL_MESHING_3D<T>& mesher_input,CODIMENSIONAL_INCOMPRESSIBLE<TV>& incompressible_input,
        CODIMENSIONAL_VISCOSITY_FORCE<TV>& viscosity_force_input,CODIMENSIONAL_VISCOELASTIC_FORCE<TV>& elasticity_force_input,ARRAY<RIGID_BODY<TV>*>& rigid_bodies_input,
        CODIMENSIONAL_GEOMETRY_SPATIAL_PARTITION<TV>& spatial_partition_input);
    void Initialize_Mpi(MPI_CODIMENSIONAL<TV>* mpi_codim_input){mpi_codim=mpi_codim_input;}
    virtual void Advance_One_Timestep(const T dt,const T time);
    void Advance_One_Timestep_Solid(const T dt,const T time);
    void Advance_One_Timestep_Fluid(const T dt,const T time){}

    void Update_Particle_Length_Scale();
    void Update_Particle_Length_Scale_Using_Rigid_Body(RIGID_BODY<TV>* rb);
    void Update_Particle_Contact();
    void Update_Particle_Inactive();
    void Squash_Fluid(); ////squash fluid using a brute-force way, replaced by triangle incompressibility now
    void Update_Boundary_Conditions();
    void Update_Particle_Incompressibility();
    void Update_Particle_Meshing_Direction();
    void Update_Target_Divergence();

    void Set_Triangle_Length_Scale(T length_scale_min_input,T length_scale_max_input,T band_width_min_input,T band_width_max_input);
    void Set_Tetrahedron_Length_Scale(T length_scale_min_input,T length_scale_max_input,T band_width_min_input,T band_width_max_input);

    void Correct_Fiber_Particle_Positions();
    void Detect_Fiber_Fluid_Collisions(ARRAY<FIBER_FLUID_COLLISION_PAIR>& collision_pairs);
    
    void Initialize_Forces();
    void Add_Fiber_Fluid_Coupling_Forces(ARRAY_VIEW<TV>& adhesion_forces,const T dt,const T time);

    T Prism_Lost_Volume(const TV_INT3& fluid_tri_v,const int& fiber_seg_index);

    static void Smooth_Particle_Length_Scale(const CODIMENSIONAL_MESH& mesh,CODIMENSIONAL_FLUID_PARTICLES<TV>& particles,const int iter_num=2,bool length_scale_coarsen=true,ARRAY<bool>* flag_changeable=0);
};
}
#endif
