//#####################################################################
// Copyright 2012 Matthew Cong.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <string>
#include <PhysBAM_Tools/Parallel_Computation/MPI_UNIFORM_GRID.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_WORLD.h>
#include <PhysBAM_Tools/Parallel_Computation/THREAD_QUEUE.h>
#include <PhysBAM_Tools/Parsing/PARSE_ARGS.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_WORLD.h>
#include <PhysBAM_Tools/Parallel_Computation/MPI_UNIFORM_GRID.h>
#include <PhysBAM_Geometry/Geometry_Particles/REGISTER_GEOMETRY_READ_WRITE.h>
#include <PhysBAM_Geometry/Basic_Geometry/BASIC_GEOMETRY_FORWARD.h>
#include "Solver/CODIMENSIONAL_SOLVER_DRIVER.h"
#include "Jets/CODIMENSIONAL_JETS_DRIVER.h"
#include "Viscoelastic/PAINT_DRIVER.h"
#include "Solver/CODIMENSIONAL_FLUID_DRIVER.h"
#include "Meshing/CODIMENSIONAL_MESHING_DRIVER.h"
#include "Viscoelastic/CODIMENSIONAL_VISCOELASTIC_DRIVER.h"
#include "Solver/CODIMENSIONAL_VISCOSITY_FORCE.h"
#include "Intersection/CODIMENSIONAL_INTERSECTION_DRIVER.h"
#include "Levelset/CODIMENSIONAL_LEVELSET_DRIVER.h"

using namespace PhysBAM;

template<class TV> void Execute_Main_Program(STREAM_TYPE& stream_type,PARSE_ARGS& parse_args,MPI_WORLD& mpi_world){
    typedef typename TV::SCALAR T;

    int driver=parse_args.Get_Integer_Value("-driver");

    switch(driver){
    case 1: ////jets driver
        {
            CODIMENSIONAL_JETS_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.restart_frame=parse_args.Get_Integer_Value("-restart");
            driver.cfl=(T)parse_args.Get_Double_Value("-cfl");
            driver.frame_rate=(T)parse_args.Get_Double_Value("-fps");

            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            if(driver.use_mpi)driver.Set_Output_Directory(driver.output_directory+"_m");

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 2: ////solver driver
        {
            CODIMENSIONAL_SOLVER_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.analytic_function=parse_args.Get_Integer_Value("-af");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.use_mpi=mpi_world.initialized;
            VECTOR<double,3> mpi_subdivision=parse_args.Get_Vector_3D_Value("-mpi_sub");
            if(mpi_subdivision!=VECTOR<double,3>())driver.processes_per_dimension=VECTOR<int,3>(mpi_subdivision);
            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            if(driver.use_mpi)driver.Set_Output_Directory(driver.output_directory+"_m");

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 3: ////paint driver
        {
            PAINT_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.restart_frame=parse_args.Get_Integer_Value("-restart");
            driver.cfl=(T)parse_args.Get_Double_Value("-cfl");
            driver.frame_rate=(T)parse_args.Get_Double_Value("-fps");
            driver.write_substep=parse_args.Get_Option_Value("-write_substep");
            driver.viscosity_min_factor=(T)parse_args.Get_Double_Value("-minvis");
            driver.viscosity_factor=(T)parse_args.Get_Double_Value("-vis");
            driver.example.viscosity_force.bp_alpha=(T)parse_args.Get_Double_Value("-bp_alpha");
            driver.gravity_factor=(T)parse_args.Get_Double_Value("-g");
            driver.example.viscosity_force.shear_rate_max=parse_args.Get_Double_Value("-maxshear");
            int varvis=parse_args.Get_Integer_Value("-varvis");driver.use_shear_dependent_variable_viscosity=(varvis!=0);
            driver.elastic_modulus_factor=(T)parse_args.Get_Double_Value("-elas");
            driver.use_implicit_elasticity=!parse_args.Get_Option_Value("-exp");
            if(varvis)driver.example.viscosity_force.variable_viscosity_mode=typename CODIMENSIONAL_VISCOSITY_FORCE<TV>::VARIABLE_VISCOSITY_MODE(varvis);
            driver.example.elasticity_force.relaxation=parse_args.Get_Double_Value("-rlx");
            driver.scale=(T)parse_args.Get_Double_Value("-s");
            driver.plastic_clamp_min=(T)parse_args.Get_Double_Value("-minplas");
            driver.plastic_clamp_max=(T)parse_args.Get_Double_Value("-maxplas");
            driver.fiber.linear_stiffness_coefficient=parse_args.Get_Double_Value("-fb_s");
            driver.fiber.linear_damping_coefficient=parse_args.Get_Double_Value("-fb_d");
            driver.fiber.bending_stiffness_coefficient=parse_args.Get_Double_Value("-fb_bs");
            driver.fiber.bending_damping_coefficient=parse_args.Get_Double_Value("-fb_bd");

            driver.adaptive_length_scale_min=parse_args.Get_Double_Value("-ls_min");
            driver.surface_tension_factor=parse_args.Get_Double_Value("-st");
            driver.rigid_body_speed_factor=parse_args.Get_Double_Value("-speed");
            driver.example.surface_tension_force.rim_coef_1=parse_args.Get_Double_Value("-rc1");
            driver.example.surface_tension_force.rim_coef_2=parse_args.Get_Double_Value("-rc2");
            driver.friction_magnitude=parse_args.Get_Double_Value("-fric");

            driver.paint_tube_test=parse_args.Get_Integer_Value("-tube_test");
            driver.bunny_test=parse_args.Get_Integer_Value("-bunny_test");
            driver.use_1d_tube=parse_args.Get_Option_Value("-1d_tube");
            driver.use_tube_rigid=parse_args.Get_Option_Value("-tube_rigid");
            driver.paint_height_factor=parse_args.Get_Double_Value("-paint_height");
            driver.use_overlapping_tube=parse_args.Get_Option_Value("-overlap_tube");
            driver.time_scale=parse_args.Get_Double_Value("-time_scale");
            driver.example.viscosity_force.triangle_viscosity_factor=parse_args.Get_Double_Value("-tri_vis");
            driver.example.viscosity_force.segment_viscosity_factor=parse_args.Get_Double_Value("-seg_vis");
            driver.min_tri_thickness_factor=parse_args.Get_Double_Value("-min_tri_thk");
            driver.min_seg_thickness_factor=parse_args.Get_Double_Value("-min_seg_thk");
            driver.adaptive_length_scale_min_factor=parse_args.Get_Double_Value("-adp");
            driver.adaptive_length_scale_min_factor2=parse_args.Get_Double_Value("-adp2");
            driver.use_full_letters=parse_args.Get_Option_Value("-full_letters");
            driver.example.mesher.use_triangle_mesh=!parse_args.Get_Option_Value("-no_tri");
            driver.example.mesher.use_segment_mesh=!parse_args.Get_Option_Value("-no_seg");
            driver.example.surface_tension_force.st_on_tris=!parse_args.Get_Option_Value("-no_st_on_tri");
            driver.example.surface_tension_force.st_on_segs=!parse_args.Get_Option_Value("-no_st_on_seg");
            driver.use_low_fric_near_fiber=parse_args.Get_Option_Value("-low_fric_near_fiber");
            driver.piz_thickness_min_factor=parse_args.Get_Double_Value("-piz_min");
            driver.piz_thickness_max_factor=parse_args.Get_Double_Value("-piz_max");
            driver.exp_part_viscosity=!parse_args.Get_Option_Value("-no_exp_vis");
            driver.example.mesher.use_fix_poping=parse_args.Get_Option_Value("-fix_poping");
            driver.example.viscosity_force.pic_flip_weight=parse_args.Get_Double_Value("-pfw");
            driver.example.elasticity_force.pic_flip_weight=parse_args.Get_Double_Value("-pfw_e");
            driver.ring_source_peak_number=parse_args.Get_Integer_Value("-ring_peak");
            driver.max_iter_number=parse_args.Get_Integer_Value("-max_itr");
            driver.use_particle_incompressibility_for_fiber=parse_args.Get_Option_Value("-fb_pt_td");
            driver.cylinder_angular_velocity=parse_args.Get_Double_Value("-cy_a");
            driver.use_letter_r=parse_args.Get_Option_Value("-use_letter_r");
            driver.lambda_cy=parse_args.Get_Double_Value("-lambda_cy");
            driver.alpha_cy=parse_args.Get_Double_Value("-alpha_cy");
            driver.n_cy=parse_args.Get_Double_Value("-n_cy");
            driver.paint_trace_test=parse_args.Get_Integer_Value("-trace_test");
            driver.piz_init_frame=parse_args.Get_Integer_Value("-piz_init_frame");
            driver.piz_init_angle=parse_args.Get_Double_Value("-piz_init_angle");
            driver.piz_mesh_noise_scale=parse_args.Get_Double_Value("-piz_scale");
            driver.piz_mesh_noise_offset=parse_args.Get_Vector_3D_Value("-piz_offset");
            driver.piz_mesh_filename=parse_args.Get_String_Value("-piz_mesh");
            driver.piz_test=parse_args.Get_Integer_Value("-piz_test");
            driver.spray_init_frame=parse_args.Get_Integer_Value("-spray_init_frame");
            driver.spray_filename=parse_args.Get_String_Value("-spray_filename");
            driver.trace_init_frame=parse_args.Get_Integer_Value("-trace_init_frame");
            driver.trace_filename=parse_args.Get_String_Value("-trace_filename");
            driver.use_sim_box=parse_args.Get_Option_Value("-sim_box");
            driver.use_inside_sim_box=parse_args.Get_Option_Value("-inside_sim_box");
            driver.example.mesher.use_allowed_break_seg_num=parse_args.Get_Option_Value("-use_max_break_seg");
            driver.example.mesher.allowed_break_seg_num_max=parse_args.Get_Integer_Value("-max_break_seg");
            driver.example.mesher.break_seg_num=parse_args.Get_Integer_Value("-break_seg");
            driver.fixed_cfl=parse_args.Get_Double_Value("-fixed_cfl");
            driver.example.elasticity_force.use_semi_implicit=!parse_args.Get_Option_Value("-fully_imp");
            T y_in_the_air_min=parse_args.Get_Double_Value("-y_air_min");
            if(y_in_the_air_min>(T)0){
                driver.example.mesher.use_remove_codim_elements_in_the_air=true;
                driver.example.mesher.y_in_the_air_min=y_in_the_air_min;
            }

            VECTOR<double,3> mpi_subdivision=parse_args.Get_Vector_3D_Value("-mpi_sub");
            if(mpi_subdivision!=VECTOR<double,3>())driver.processes_per_dimension=VECTOR<int,3>(mpi_subdivision);
            driver.use_mpi=mpi_world.initialized;

            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            if(driver.use_mpi)driver.Set_Output_Directory(driver.output_directory+"_m");

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 4: ////meshing driver
        {
            CODIMENSIONAL_MESHING_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.restart_frame=parse_args.Get_Integer_Value("-restart");
            driver.cfl=(T)parse_args.Get_Double_Value("-cfl");
            driver.frame_rate=(T)parse_args.Get_Double_Value("-fps");
            VECTOR<double,3> mpi_subdivision=parse_args.Get_Vector_3D_Value("-mpi_sub");
            if(mpi_subdivision!=VECTOR<double,3>())driver.processes_per_dimension=VECTOR<int,3>(mpi_subdivision);
            driver.use_mpi=mpi_world.initialized;
            
            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            if(driver.use_mpi)driver.Set_Output_Directory(driver.output_directory+"_m");

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 5: ////viscoelastic fluid driver
        {
            CODIMENSIONAL_VISCOELASTIC_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.restart_frame=parse_args.Get_Integer_Value("-restart");
            driver.cfl=(T)parse_args.Get_Double_Value("-cfl");
            driver.frame_rate=(T)parse_args.Get_Double_Value("-fps");
            driver.write_substep=parse_args.Get_Option_Value("-write_substep");

            driver.youngs_modulus_factor=(T)parse_args.Get_Double_Value("-elas");
            driver.viscosity_factor=(T)parse_args.Get_Double_Value("-vis");
            driver.viscosity_min_factor=(T)parse_args.Get_Double_Value("-minvis");
            driver.use_implicit_all=parse_args.Get_Option_Value("-imp");
            driver.use_explicit_all=parse_args.Get_Option_Value("-exp");
            int varvis=parse_args.Get_Integer_Value("-varvis");driver.use_shear_dependent_variable_viscosity=(varvis!=0);
            if(varvis)driver.example.viscosity_force.variable_viscosity_mode=typename CODIMENSIONAL_VISCOSITY_FORCE<TV>::VARIABLE_VISCOSITY_MODE(varvis);
            driver.example.viscosity_force.shear_rate_max=parse_args.Get_Double_Value("-maxshear");
            driver.example.elasticity_force.relaxation=parse_args.Get_Double_Value("-rlx");
            driver.example.elasticity_force.use_green_strain=!parse_args.Get_Option_Value("-cauchy");
            
            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            if(driver.use_mpi)driver.Set_Output_Directory(driver.output_directory+"_m");

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 6: ////intersection driver
        {
            CODIMENSIONAL_INTERSECTION_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            std::string input=parse_args.Get_String_Value("-input");
            if(!input.empty())driver.input_filename=input;

            driver.Initialize();
            driver.Execute_Main_Program();
        }break;
    case 7: ////levelset driver
        {
            CODIMENSIONAL_LEVELSET_DRIVER<TV> driver;
            driver.test=parse_args.Get_Integer_Value("-test");
            driver.last_frame=parse_args.Get_Integer_Value("-e");
            driver.restart_frame=parse_args.Get_Integer_Value("-restart");
            driver.first_frame=driver.restart_frame;
            driver.current_frame=driver.restart_frame;
            driver.read_phis_from_first_frame=parse_args.Get_Option_Value("-read_phis");
            driver.update_slow_interface_phis=!parse_args.Get_Option_Value("-fix_slow");
            driver.levelset_region_number=parse_args.Get_Integer_Value("-phi_regions");
            driver.ignore_tetrahedron_interior=parse_args.Get_Option_Value("-ignore_interior");
            driver.only_update_interface=parse_args.Get_Option_Value("-interface");
            driver.passive_update=parse_args.Get_Option_Value("-passive");
            driver.narrow_band_width=parse_args.Get_Integer_Value("-band_width");
            if(driver.narrow_band_width>-1)driver.use_narrow_band=true;
            std::string output_directory=parse_args.Get_String_Value("-o");
            if(!output_directory.empty())driver.Set_Output_Directory(output_directory);
            driver.Initialize_Test_Parameters();
            driver.Execute_Main_Program();
        }break;
    default:
        PHYSBAM_FATAL_ERROR("Invalid driver number.");break;
    }
}

int main(int argc, char **argv){
    typedef double T;
    typedef double RW;
    STREAM_TYPE stream_type((RW()));

    PARSE_ARGS parse_args;

    ////common options
    parse_args.Add_Integer_Argument("-dim",3,"dimension");
    parse_args.Add_Integer_Argument("-restart",0,"specify the restart frame");
    parse_args.Add_Integer_Argument("-e",200,"specify the final frame of the simulation");
    parse_args.Add_Integer_Argument("-substep_e",-1,"specify the final frame of the substep");
    parse_args.Add_Integer_Argument("-r",0,"specify the restart frame");
    parse_args.Add_Integer_Argument("-driver",1,"specify the driver");
    parse_args.Add_Integer_Argument("-test",1,"select the test to run");
    parse_args.Add_Double_Argument("-cfl",1,"specify the cfl number (and consequently the maximum timestep)");
    parse_args.Add_Double_Argument("-fps",1000,"frame rate");
    parse_args.Add_Option_Argument("-write_substep","write sub step");
    parse_args.Add_String_Argument("-o","","output directory");
    parse_args.Add_Vector_3D_Argument("-mpi_sub",VECTOR<double,3>(),"mpi domain subdivision");
    parse_args.Add_Double_Argument("-s",1,"scale");
    parse_args.Add_Double_Argument("-g",1,"gravity_factor");
    parse_args.Add_Integer_Argument("-max_itr",2000,"max interation number for cg solver");

    ////options for solver driver
    parse_args.Add_Integer_Argument("-af",1,"analytic function");

    ////options for viscoelastic driver
    parse_args.Add_Double_Argument("-elas",1.,"elastic modulus factor");
    parse_args.Add_Double_Argument("-vis",1.,"viscosity factor");
    parse_args.Add_Double_Argument("-minvis",1.,"min viscosity factor, for shear thinning");
    parse_args.Add_Double_Argument("-maxshear",20.,"max shear rate, for shear thinning");
    parse_args.Add_Option_Argument("-exp","use explicit solver");
    parse_args.Add_Option_Argument("-imp","use implicit solver");
    parse_args.Add_Double_Argument("-rlx",1.,"relaxation");
    parse_args.Add_Option_Argument("-cauchy","use cauchy strain");
    parse_args.Add_Integer_Argument("-varvis",0,"variable viscosity, 1-shear thinning, 2-bingham plastic");
    parse_args.Add_Double_Argument("-st",1.,"surface tension factor");
    parse_args.Add_Double_Argument("-minplas",.6,"min plastic clamp");
    parse_args.Add_Double_Argument("-maxplas",1.4,"max plastic clamp");

    ////options for intersection driver
    parse_args.Add_String_Argument("-input","","input file");
    
    ////options for paint driver
    parse_args.Add_Double_Argument("-fb_s",10.,"fiber linear spring stiffness");
    parse_args.Add_Double_Argument("-fb_d",1.,"fiber linear spring damping");
    parse_args.Add_Double_Argument("-fb_bs",10.,"fiber bending spring stiffness");
    parse_args.Add_Double_Argument("-fb_bd",1.,"fiber bending spring damping");
    parse_args.Add_Double_Argument("-ls_min",.2,"length scale min factor");
    parse_args.Add_Double_Argument("-speed",1,"rigid body speed factor");
    parse_args.Add_Double_Argument("-rc1",1.,"rim coef 1");
    parse_args.Add_Double_Argument("-rc2",1.,"rim coef 2");
    parse_args.Add_Double_Argument("-bp_alpha",1.,"bp alpha");
    parse_args.Add_Double_Argument("-paint_height",1.,"paint height");
    parse_args.Add_Double_Argument("-fric",.5,"friction magnitude");
    parse_args.Add_Double_Argument("-time_scale",1,"time scale");
    parse_args.Add_Double_Argument("-tri_vis",1,"triangle viscosity factor");
    parse_args.Add_Double_Argument("-seg_vis",1,"segment viscosity factor");
    parse_args.Add_Double_Argument("-min_tri_thk",1,"min triangle thickness factor");
    parse_args.Add_Double_Argument("-min_seg_thk",1,"min segment thickness factor");
    parse_args.Add_Double_Argument("-adp",1,"adaptive length scale min factor");
    parse_args.Add_Double_Argument("-adp2",1,"adaptive length scale min factor2");
    parse_args.Add_Double_Argument("-lambda_cy",0.,"lambda cy");
    parse_args.Add_Double_Argument("-alpha_cy",0.,"alpha cy");
    parse_args.Add_Double_Argument("-n_cy",0.,"n cy");
    parse_args.Add_Double_Argument("-y_air_min",0.,"y in the air min");
    
    ////options for examples of paint driver
    parse_args.Add_Integer_Argument("-tube_test",1,"paint tube test");
    parse_args.Add_Integer_Argument("-bunny_test",1,"bunny test, test 219");
    parse_args.Add_Option_Argument("-1d_tube","use 1d tube, test 210");
    parse_args.Add_Option_Argument("-tube_rigid","add tube rigid body, test 210");
    parse_args.Add_Option_Argument("-overlap_tube","use overlapped tube, test 210");
    parse_args.Add_Option_Argument("-full_letters","use full letters, test 211");
    parse_args.Add_Option_Argument("-no_tri","no triangle");
    parse_args.Add_Option_Argument("-no_seg","no segment");
    parse_args.Add_Option_Argument("-no_st_on_tri","no st on triangle");
    parse_args.Add_Option_Argument("-no_st_on_seg","no st on segment");
    parse_args.Add_Option_Argument("-low_fric_near_fiber","low friction near fiber");
    parse_args.Add_Double_Argument("-piz_min",1.,"piz thickness min");
    parse_args.Add_Double_Argument("-piz_max",1.,"piz thickness max");
    parse_args.Add_Option_Argument("-no_exp_vis","no explicit part viscosity");
    parse_args.Add_Option_Argument("-fix_poping","use fix poping");
    parse_args.Add_Double_Argument("-pfw",.8,"pic-flip weight");
    parse_args.Add_Double_Argument("-pfw_e",.8,"pic-flip weight elasticity");
    parse_args.Add_Integer_Argument("-ring_peak",5,"ring source peak number");
    parse_args.Add_Option_Argument("-ring_rnd","use ring randomness");
    parse_args.Add_Option_Argument("-fb_pt_td","fiber-paint target divergnce");
    parse_args.Add_Double_Argument("-cy_a",200.,"cylinder angular velocity");
    parse_args.Add_Option_Argument("-use_letter_r","use letter r");
    parse_args.Add_Integer_Argument("-trace_test",1,"paint trace test");
    parse_args.Add_Integer_Argument("-piz_init_frame",3,"piz initial frame");
    parse_args.Add_Double_Argument("-piz_init_angle",0.,"piz initial angle");
    parse_args.Add_Double_Argument("-piz_scale",400.,"piz noise scale");
    parse_args.Add_Vector_3D_Argument("-piz_offset",VECTOR<double,3>::Constant_Vector(10.),"piz noise offset");
    parse_args.Add_String_Argument("-piz_mesh","./piz_mesh","piz mesh filename");
    parse_args.Add_Integer_Argument("-piz_test",1,"piz test");
    parse_args.Add_Integer_Argument("-spray_init_frame",1080,"spray initial frame");
    parse_args.Add_String_Argument("-spray_filename","./j5_cy_s1");
    parse_args.Add_Integer_Argument("-trace_init_frame",600,"trace initial frame");
    parse_args.Add_String_Argument("-trace_filename","./paint_simple_2_restart");
    parse_args.Add_Option_Argument("-sim_box","use sim box");
    parse_args.Add_Option_Argument("-inside_sim_box","use inside sim box");
    parse_args.Add_Option_Argument("-use_max_break_seg","use max break seg");
    parse_args.Add_Integer_Argument("-max_break_seg",1000,"max break seg num");
    parse_args.Add_Integer_Argument("-break_seg",0,"break seg num");
    parse_args.Add_Double_Argument("-fixed_cfl",0.,"fixed cfl");
    parse_args.Add_Option_Argument("-fully_imp","fully implicit elasticity");

    ////options for levelset driver
    parse_args.Add_Integer_Argument("-phi_regions",4,"number of levelset regions");
    parse_args.Add_Option_Argument("-read_phis","read phi values from first frame");
    parse_args.Add_Option_Argument("-fix_slow","fix phi values for slow particles");
    parse_args.Add_Option_Argument("-ignore_interior","ignore interior of tetrahedron mesh during fast marching");
    parse_args.Add_Option_Argument("-interface","only update interface phis; no fast marching");
    parse_args.Add_Option_Argument("-passive","do not perform levelset reinitialization");
    parse_args.Add_Integer_Argument("-band_width",-1,"graph distance threshold for narrow band update");

    parse_args.Parse(argc,argv);
    parse_args.Print_Arguments(argc,argv);

    Initialize_Geometry_Particle();
    Initialize_Read_Write_Structures();

    std::string output_directory=parse_args.Get_Integer_Value("-dim")==2?"output":"output3";
    std::string output_directory_from_args=parse_args.Get_String_Value("-o");
    if(!output_directory_from_args.empty()){output_directory=output_directory_from_args;}
    //if(mpi_world.initialized)output_directory+="_m/1";

    if(!FILE_UTILITIES::File_Exists(output_directory)) FILE_UTILITIES::Create_Directory(output_directory);
    if(!FILE_UTILITIES::File_Exists(output_directory+"/common")) FILE_UTILITIES::Create_Directory(output_directory+"/common");
    LOG::Initialize_Logging(false,false,1<<30,true,1);
    LOG::Instance()->Copy_Log_To_File(output_directory+"/common/log",parse_args.Is_Value_Set("-restart"));

    std::string params=parse_args.Print_Arguments(argc,argv);
    LOG::cout<<params<<std::endl;

    MPI_WORLD mpi_world(argc,argv);

	//if(parse_args.Get_Integer_Value("-dim")==2){
	//	Execute_Main_Program<VECTOR<T,2> >(stream_type,parse_args,mpi_world);
	//}
	//else{
		Execute_Main_Program<VECTOR<T,3> >(stream_type,parse_args,mpi_world);
	//}
    
    return 0;
}
