//#####################################################################
// Copyright 2014, Bo Zhu
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <PhysBAM_Tools/Random_Numbers/NOISE.h>
#include <PhysBAM_Geometry/Basic_Geometry/PLANE.h>
#include <PhysBAM_Geometry/Solids_Geometry/RIGID_GEOMETRY_COLLECTION.h>
#include <PhysBAM_Geometry/Collisions/COLLISION_GEOMETRY_COLLECTION.h>
#include "../Viscoelastic/PAINT_DRIVER.h"
#include "../Meshing/CODIMENSIONAL_MESH_HASHTABLE.h"
#include <PhysBAM_Geometry/Read_Write/Geometry/READ_WRITE_TRIANGULATED_SURFACE.h>

#include <fstream>

static const int TEST_MAYO=201;
static const int TEST_TOOTHPASTE=202;
static const int TEST_CHEESE=205;
static const int TEST_CHEESE_ON_BREAD=206;
static const int TEST_CHEESE_HEATING=207;
static const int TEST_CHEESE_POURING=208;
static const int TEST_PREPROCESS_CHEESE_MESH=209;
static const int TEST_PAINT_TUBE=210;
static const int TEST_PAINT_TUBE_MULTIPLE_COLOR=216;
static const int TEST_PAINT_TRACE=211;
static const int TEST_PAINT_TRACE2=223;
static const int TEST_PAINT_DIP=212;
static const int TEST_PAINT_SPRAY=213;
static const int TEST_PAINT_SPLASH=214;
static const int TEST_PAINT_SPRAY2=218;
static const int TEST_MATERIALS=215;
static const int TEST_MATERIALS_2=217;
static const int TEST_EXPOSITION_1=219;
static const int TEST_EXPOSITION_2=220;
static const int TEST_PAINT_SPRAY3=221;
static const int TEST_PAINT_SPRAY4=222;

#define ACCEL_WINDOWS_COMPILING 1

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Test_Parameters()
{
    global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-1),TV::Constant_Vector((T)1));
    switch(test){
    //////////////////////////////////////////////////////////////////////////
    ////tests for paint example, paint tube, shear thinning, fiber-paint interaction, levelset, etc.
    ////tests for paint tube
    case 20:{   ////tube
        length_scale=(T)1e-3;
        T radius=length_scale*(T)2;

        T tube_height=length_scale*(T)10;
        T paint_height=length_scale*(T)1;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)5;
        TV dir=TV((T).5,(T)-1,(T)0);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        levelset_region_number=2;
        Add_Canvas();
   }break;
#if !defined(_WIN32) || !defined(ACCEL_WINDOWS_COMPILING)
   case 21:{   ////tube
        length_scale=(T)1e-3;
        T radius=length_scale*(T)2;

        T tube_height=length_scale*(T)10;
        T paint_height=length_scale*(T)1;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)1.5;
        TV dir=TV((T).2,(T)-1,(T)0);
        T source_speed=(T).1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,source_speed,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_viscosity=true;
        kinematic_viscosity*=(T)10000;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        Add_Canvas();
    }break;
    case 22:{   ////1D tube
        length_scale=(T)1e-3;

        ////tube
        T radius=length_scale*(T)2;
        T tube_height=length_scale*(T)10;
        T paint_height=length_scale;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)5;
        TV dir=TV((T).5,(T)-1,(T)0);
        //TV dir=TV::Axis_Vector(2)*(T)-1;
        T source_speed=(T).2;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,source_speed,(T).1);
        tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_gravity=true;
        has_viscosity=true;
        kinematic_viscosity*=(T)10000;
        use_implicit_viscosity=true;
        has_surface_tension=false;
        surface_tension*=(T).2;
        has_incompressibility=false;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        levelset_region_number=2;
        //Add_Canvas((T).02);
        Add_Canvas((T).2);
    }break;
    case 23:{   ////tube prototype
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T tube_height=length_scale*(T)5;
        T paint_height=length_scale*(T)15;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)15;
        TV dir=TV::Axis_Vector(2)*(T)-1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 24:{   ////test multiple tubes
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;

        ////paint tube1
        T tube_height=length_scale*(T)5;
        T paint_height=length_scale*(T)2;
        T radius=length_scale*(T)2;
        {TV e1=TV(length_scale*(T)-5,length_scale*(T)10,(T)0);
        TV dir=TV((T)2,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).08);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();}

        ////paint tube2
        {TV e1=TV(length_scale*(T)5,length_scale*(T)5,(T)0);
        TV dir=TV((T)-1.5,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tube->Enable(false);}

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=true;
        levelset_region_number=2;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    ////tests for shear thinning
    case 124:{  ////test codimensional shear thinning, prototype for the paint tube example
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;

        ////paint
        T tube_height=(T)5e-3;
        T paint_height=(T)15e-3;
        TV e1=TV::Axis_Vector(2)*(T)15e-3;
        TV dir=TV::Axis_Vector(2)*(T)-1;
        T radius=(T)2e-3;

        //volumetric
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,(T)0);

        //////volumetric thin sheet
        //radius=(T)4e-3;PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        //tube->Initialize_Volumetric_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,length_scale);   
        
        ////thin sheet
        //PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        //tube->Initialize_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);

        ////filament
        //PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);

        tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        //surface_tension*=(T).005;
        //tri_surface_tension=(T).005;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 125:{   ////test 1d shear thinning, 
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;
        max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).1;

        ////paint
        T tube_height=length_scale*(T)5;
        T paint_height=length_scale*(T)15;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)15;
        TV dir=TV::Axis_Vector(2)*(T)-1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).0,(T).1);
        tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 126:{  ////test 1d shear-thinning, static
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;
        max_thickness_factor=(T)2;

        ////paint
        T height=length_scale*(T)15;
        //T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)15;TV dir=TV((T)0,(T)-1,(T)0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e1+dir*height,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 127:{  ////test 2d shear-thinning, 
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;
        max_thickness_factor=(T)2;

        ////paint
        T tube_height=length_scale*(T)5;
        T paint_height=length_scale*(T)15;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)15;
        TV dir=TV::Axis_Vector(2)*(T)-1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;

    ////tests for solid-paint interaction
    case 26:{   ////cube squashes volumetric paint
        length_scale=(T)1e-3;

        T height=length_scale*(T)2;T radius=length_scale*(T)3;
        TV e1=TV::Axis_Vector(2)*height;TV e2=TV();e2.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

        ////kinematic solid
        //TV rigid_body_size=TV::Constant_Vector(radius*(T)3);
        //kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        //kinematic_rigid_body->X()=TV();kinematic_rigid_body->X().y=(T).5*rigid_body_size.y+(T)length_scale*(T).3;kinematic_rigid_body->X().x=(T)-.5*rigid_body_size.x+(T)-.5*radius-length_scale*(T)2;
        //rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";

        T cylinder_height=radius*(T)3;T cylinder_radius=radius*(T)2;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius);
        kinematic_rigid_body->X()=TV();
        kinematic_rigid_body->X().x=(T)-cylinder_radius-radius-length_scale*(T)2;
        kinematic_rigid_body->X().y=(T).5*cylinder_height+(T)length_scale*(T).1;
        kinematic_rigid_body->Rotation()=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
    }break;
    case 27:{   ////
        length_scale=(T)1e-3;
        Add_Canvas((T).1);

        ////kinematic solid
        TV rigid_body_size=TV::Constant_Vector(length_scale*(T)10);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        kinematic_rigid_body->X()=TV();
        kinematic_rigid_body->X().y=(T).5*rigid_body_size.y;
        //kinematic_rigid_body->X().x=(T)-.5*rigid_body_size.x+(T)-.5*radius-length_scale*(T)2;
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T height=length_scale*(T)2;T radius=length_scale*(T)3;
        TV e1=TV::Axis_Vector(2)*(height+rigid_body_size.y);TV e2=TV::Axis_Vector(2)*rigid_body_size.y;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=false;
        surface_tension*=(T)1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        //T cylinder_height=radius*(T)3;T cylinder_radius=radius*(T)2;
        //kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_height,cylinder_radius);
        //kinematic_rigid_body->X()=TV();
        //kinematic_rigid_body->X().x=(T)-cylinder_radius-radius-length_scale*(T)2;
        //kinematic_rigid_body->X().y=(T).5*cylinder_height+(T)length_scale*(T).1;
        //kinematic_rigid_body->Rotation()=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        //rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";
    }break;
    case 37:{   ////two boxes with a gap interacting with volumetric paint
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T paint_size=(T)2e-3;TV paint_box_size=TV(paint_size,paint_size*(T).5,paint_size*(T)1);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////kinematic solid
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV::Constant_Vector(paint_size);rigid_body_size.x*=(T).25;rigid_body_size.z*=(T).25;
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();
            kinematic_rigid_body->X().x=box_min.x-rigid_body_size.x*(T).52;
            kinematic_rigid_body->X().y=(T).55*rigid_body_size.y;
            if(i==1)kinematic_rigid_body->X().z-=((T).5*rigid_body_size.z+paint_box_size.z*(T).25);
            else kinematic_rigid_body->X().z+=((T).5*rigid_body_size.z+paint_box_size.z*(T).25);
            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);
        }

        TV domain_size=TV(paint_box_size.x*(T)10,paint_box_size.y*(T)4,paint_box_size.z*(T)4);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 38:{   ////two cylinders interacting with volumetric paint
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T paint_size=(T)2e-3;TV paint_box_size=TV(paint_size,paint_size*(T).5,paint_size*(T)1);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////kinematic solid
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV::Constant_Vector(paint_size);rigid_body_size.x*=(T).4;rigid_body_size.z*=(T).4;
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(rigid_body_size.y,rigid_body_size.x*(T).5);
            kinematic_rigid_body->X()=TV();
            kinematic_rigid_body->X().x=box_min.x-rigid_body_size.x*(T).55;
            kinematic_rigid_body->X().y=(T).5*rigid_body_size.y;
            if(i==1)kinematic_rigid_body->X().z-=((T).5*rigid_body_size.z+paint_box_size.z*(T).25);
            else kinematic_rigid_body->X().z+=((T).5*rigid_body_size.z+paint_box_size.z*(T).25);
            T rot_angle=(T)-.6*(T)half_pi;if(i==2)rot_angle*=(T)-1;
            kinematic_rigid_body->Rotation()=ROTATION<TV>(rot_angle,TV::Axis_Vector(1));
            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);
        }

        TV domain_size=TV(paint_box_size.x*(T)10,paint_box_size.y*(T)4,paint_box_size.z*(T)4);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 39:{   ////two spheres interacting with volumetric paint
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)6;

        ////paint
        //T paint_size=(T)2e-3;TV paint_box_size=TV(paint_size,paint_size*(T).5,paint_size*(T)1);
        //TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        //TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        T paint_size=(T)1e-2;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////kinematic solid
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV::Constant_Vector(paint_size*(T).8);
            //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(rigid_body_size.y,rigid_body_size.x*(T).5);
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(rigid_body_size.Min()*(T).5,(T)1);
            //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();
            if(i==1)kinematic_rigid_body->X().z=-rigid_body_size.Min()*(T).5;
            else kinematic_rigid_body->X().z=rigid_body_size.Min()*(T).5;
            kinematic_rigid_body->X().y=rigid_body_size.Min()*(T).5+thickness_scale+(T).5*length_scale;
            //T rot_angle=(T)half_pi;kinematic_rigid_body->Rotation()=ROTATION<TV>(rot_angle,TV::Axis_Vector(1));
            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);
        }

        //////brush
        //fiber_length_scale=(T)1e-3;
        //fiber_radius=fiber_length_scale*(T).1;fiber_distance=fiber_radius*(T)7;fiber_length=fiber_length_scale*(T)20;
        //fiber.object_frame=FRAME<TV>(TV(0,fiber_length*(T).85,(T)0),ROTATION<TV>((T)-.0*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        //fiber.implicit_fiber_segment_number=10;
        //fiber.fiber_bending_start_frac=(T).6;   
        //fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,2,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_length_scale*(T)0);
        //Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(paint_box_size.x*(T)10,paint_box_size.y*(T)4,paint_box_size.z*(T)4);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 50:{   ////another cube squash thin sheet
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=(T)1e-3;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)10e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        //TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)2*paint_size;paint_box_min.y=(T)0;
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.mesher.use_contact_aware_triangle_remeshing=true;

        ////kinematic solid
        //TV rigid_body_size=TV::Constant_Vector((T)20e-3);
        //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        //kinematic_rigid_body->X()=TV();
        //kinematic_rigid_body->X().x+=(T).5*length_scale;
        //kinematic_rigid_body->X().y=(T).72*rigid_body_size.y+thickness_scale*(T)1.5;
        ////kinematic_rigid_body->Rotation()=ROTATION<TV>((T)half_pi*(T)-.5,TV((T).2,(T).2,(T)1).Normalized());
        //kinematic_rigid_body->Rotation()=ROTATION<TV>((T)half_pi*(T).5,TV::Axis_Vector(3));
        
        TV rigid_body_size=TV((T)5e-3,(T)10e-3,(T)1e-3);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        kinematic_rigid_body->X()=TV();
        kinematic_rigid_body->X().x+=(T).5*length_scale;kinematic_rigid_body->X().z+=(T).5*length_scale;
        kinematic_rigid_body->X().y=(T).5*rigid_body_size.y+thickness_scale*(T)1.5;
        kinematic_rigid_body->Rotation()=ROTATION<TV>((T)half_pi*(T)-.2,TV::Axis_Vector(2));

        kinematic_rigid_body->name="kin_obj";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;

    ////tests for fiber-paint interaction
    case 6:case 13:{    ////brush on triangle mesh
        ////fluid
        length_scale=(T)1e-2;thickness_scale=length_scale*(T).2;
        int m=20;int n=20;
        TV box_edge_length=TV();box_edge_length.x=length_scale*(T)n;box_edge_length.z=length_scale*(T)m;
        TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,box_min,length_scale,2);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_triangle_remeshing=true;

        ////fiber
        fiber_length_scale=length_scale;
        fiber_radius=length_scale*(T).5;fiber_length=length_scale*(T)15;fiber_distance=fiber_radius*(T)4;
        TV brush_center=TV();brush_center.z=(T).02;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles(fiber_positions,1,1,brush_center,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3));
        fiber.object_frame=FRAME<TV>(TV::Axis_Vector(2)*(fiber_length*(T).89),ROTATION<TV>((T)0,TV::Axis_Vector(2)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        //fiber.use_guided_fibers=false;
        //guided_fiber_radius=(T).1*fiber_radius;guided_fiber_distance=guided_fiber_radius*(T)4;

        //if(test==6)levelset_region_number=2;else levelset_region_number=4;

        Add_Canvas((T).5);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 8:{    ////brush squashing fluid
        //Add_Canvas();
        T scale=(T).1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)1;
        ////fluid
        int m=40;int n=40;
        TV box_edge_length=TV();box_edge_length.x=length_scale*(T)n;box_edge_length.z=length_scale*(T)m;
        TV box_min=-(T).5*box_edge_length;TV box_max=(T).5*box_edge_length;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,box_min,length_scale,2);
        has_viscosity=false;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=false;
        surface_tension*=(T)1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;
        example.mesher.use_adaptive_triangle_remeshing=true;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).2;fiber_distance=fiber_radius*(T)7;fiber_length=fiber_length_scale*(T)20;
        //fiber.object_frame=FRAME<TV>(TV((T)0,fiber_length*(T).9,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));
        fiber.object_frame=FRAME<TV>(TV((T)0,fiber_length*(T).79,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));
        //fiber.object_frame=FRAME<TV>(TV((T)0,fiber_length*(T).75,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=10;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,1,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*position perturb*/fiber_length_scale*(T)0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        Add_Canvas();
    }break;
    case 28:{   ////brush motion again
        T scale=(T).5;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////brush
        T fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).1;fiber_length=fiber_length_scale*(T)20;fiber_distance=fiber_radius*(T)2;
        //fiber.object_frame=FRAME<TV>(TV(length_scale*(T)-5,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)2,fiber_length*(T).8,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=10;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,2,4,TV(),fiber_length_scale,TV::Axis_Vector(1),TV::Axis_Vector(3),16,fiber_length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,1,TV(),fiber_length_scale,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_length_scale);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////paint
        T height=one_over_scale*length_scale*(T)2;T radius=one_over_scale*length_scale*(T)3;
        TV e1=TV::Axis_Vector(2)*height;TV e2=TV();e2.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        Add_Canvas();
    }break;
    case 29:{   ////brush interacting with volumetric paint
        T scale=(T).1;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).1;fiber_distance=fiber_radius*(T)7;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4.5,fiber_length*(T).85,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=10;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,4,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_length_scale*(T)0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////paint
        //T radius=one_over_scale*length_scale*(T)1;
        //TV e1=TV::Axis_Vector(2)*height;TV e2=TV();e2.y=(T)0;TV box_min=TV(-radius,(T)0,-radius);
        T height=one_over_scale*length_scale*(T)1;TV box_size=TV(height,height,height*(T)3);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        Add_Canvas();
    }break;
    case 30:{   ////brush interacting with paint triangle surface
        T scale=(T).1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)1;
        ////fluid
        int m=20;int n=20;
        TV paint_box_size=TV();paint_box_size.x=length_scale*(T)n;paint_box_size.z=length_scale*(T)m;
        TV paint_box_min=-(T).5*paint_box_size;TV paint_box_max=(T).5*paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);
        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=false;
        surface_tension*=(T)1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;
        example.mesher.use_adaptive_triangle_remeshing=true;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).1;fiber_distance=fiber_radius*(T)7;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(0,fiber_length*(T)1,(T)0),ROTATION<TV>((T)-.0*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
 
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,2,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_length_scale*(T)0);

        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(paint_box_size.x*(T)10,paint_box_size.x*(T)2,paint_box_size.z*(T)4);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Canvas();
    }break;
    case 31:{   ////brush interacting with paint surface, testing brush shape
        T scale=(T).5;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        //T radius=one_over_scale*length_scale*(T)1;
        //TV e1=TV::Axis_Vector(2)*height;TV e2=TV();e2.y=(T)0;TV box_min=TV(-radius,(T)0,-radius);
        T height=one_over_scale*length_scale*(T)1;TV box_size=TV(height,height,height*(T)3);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,radius,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////adaptive meshing
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        //length_scale_min=(T).5;
        //band_width_max=(T)1;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).05;fiber_distance=fiber_radius*(T)2;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4,fiber_length*(T).8,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;
        fiber.fiber_bending_perturb_frac=TV((T).005,(T)0,(T)0.002);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,24,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        Add_Canvas();
    }break;

    ////tests for traces
    case 40:{   /////rigid brush forming traces on thin sheet paint
        T scale=(T).1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)4;

        //////paint
        //T paint_height=(T)2e-3;TV paint_box_size=TV(paint_height,paint_height,paint_height*(T)1);
        //TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        //TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        T paint_size=(T)5e-3;TV paint_box_size=TV(paint_size*(T).5,paint_size,paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=-paint_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).4;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)2;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).76,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        int fiber_number_row=1;int fiber_number_column=75;
        T fiber_width=fiber_number_column*fiber_distance;
        //fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T).0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(fiber_width*(T)10,fiber_width*(T)2,fiber_width*(T)1.2);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 41:{   ////mass-spring brush forming traces on thin sheet paint
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;

        //////paint
        //T paint_height=(T)2e-3;TV paint_box_size=TV(paint_height,paint_height,paint_height*(T)1);
        //TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        //TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        T paint_size=(T)5e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)-2*paint_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_triangle_remeshing=true;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)8;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;
        int fiber_number_row=1;int fiber_number_column=3;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).02);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 42:case 43:{   ////a few fibers forming traces on thin sheet paint
        //T scale=(T).25;//T one_over_scale=(T)1/scale;
        //length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        //max_thickness_factor=(T)2;

        //////paint
        //T paint_size=(T)5e-3;TV paint_box_size=TV((T)paint_size,paint_size,(T)2*paint_size);
        //int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        //TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)2*paint_size;paint_box_min.y=(T)0;
        ////TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)0;paint_box_min.y=(T)0;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        //has_gravity=true;fiber.use_gravity=has_gravity;
        //has_incompressibility=true;
        //has_surface_tension=false;
        //has_viscosity=true;
        //use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        //use_levelset=false;
        //example.mesher.use_adaptive_triangle_remeshing=true;
        //example.mesher.use_segment_mesh=false;
        //example.mesher.use_continuous_meshing_operations=true;

        //////brush
        //fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        //fiber.implicit_fiber_segment_number=5;
        //fiber.fiber_bending_start_frac=(T).6;   
        //fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        //fiber.use_gravity=true;

        ////int fiber_number_row=4;int fiber_number_column=12;
        //int fiber_number_row=1;int fiber_number_column=3;
        ////T fiber_width=fiber_number_column*fiber_distance;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        //Add_Fiber_Rigid_Body(fiber_positions,fiber);

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();

        //////////////////////////////////////////////////////////////////////////

        length_scale=scale*(T)1e-3;thickness_scale=(T).1e-3;
        adaptive_length_scale_min=(T).25e-3;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)10e-3;TV paint_box_size=TV((T).5*paint_size,(T).3*paint_size,(T)1.5*paint_size);
        TV box_min=TV((T)8e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        has_fiber_fluid_coupling_force=true;
        has_friction=true;
        coupled_evolution.use_squash_force=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.mesher.use_contact_aware_triangle_remeshing=false;
        example.mesher.use_anisotropic_triangle_remeshing=true;
        example.incompressible.use_target_div=false;
        use_correcting_particle_position_on_solids=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber_radius=(T).2e-3;fiber_distance=(T).6e-3;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;band_width_max=(T)2e-3;

        //////Predefined motion
        //fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
        //std::string dir="../model/Motion_Data/brush_coarse_curve";//T t0=(T).0;T t1=(T).6;T motion_t0=(T).2;
        //fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T)0);

        //int fiber_number_row=1;int fiber_number_column=2;
        int fiber_number_row=3;int fiber_number_column=20;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 90:{       ////test anisotropic meshing
        T scale=(T).2;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=(T).05e-3;
        adaptive_length_scale_min=(T).05e-3;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)5e-3;TV paint_box_size=TV((T)paint_size,(T).3*paint_size,(T)paint_size);
        TV box_min=TV((T)8e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=false;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        has_fiber_fluid_coupling_force=true;
        has_friction=false;
        coupled_evolution.use_squash_force=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.mesher.use_contact_aware_triangle_remeshing=false;
        example.mesher.use_anisotropic_triangle_remeshing=true;
        example.incompressible.use_target_div=false;
        use_correcting_particle_position_on_solids=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber_radius=(T).3e-3;fiber_distance=(T).6e-3;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        band_width_min=(T)2*fiber_radius;band_width_max=(T)2e-3;

        //////Predefined motion
        //fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
        //std::string dir="../model/Motion_Data/brush_coarse_curve";//T t0=(T).0;T t1=(T).6;T motion_t0=(T).2;
        //fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T)0);

        //int fiber_number_row=1;int fiber_number_column=2;
        int fiber_number_row=1;int fiber_number_column=1;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 81:case 82:{   ///brush tearing paint volume 
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=(T).25e-3;
        //length_scale_min=(T).05e-3;
        adaptive_length_scale_min=(T).25e-3;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)10e-3;TV paint_box_size=TV((T).5*paint_size,(T).5*paint_size,paint_size);
        TV box_min=TV((T)8e-3,(T)0,(T)-10e-3);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).05;
        tri_surface_tension=surface_tension*(T).2;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        has_fiber_fluid_coupling_force=true;
        has_friction=true;
        coupled_evolution.use_squash_force=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.mesher.use_contact_aware_triangle_remeshing=false;
        example.incompressible.use_target_div=false;
        use_correcting_particle_position_on_solids=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber_radius=(T).2e-3;fiber_distance=(T).6e-3;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;band_width_max=(T)2e-3;

        int fiber_number_row=6;int fiber_number_column=10;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 44:case 45:{   ////brush interacting volumetric paint
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).4*paint_size,paint_size);
        TV box_min=TV(paint_size,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.incompressible.use_all_pressure_sample_on_nodes=true;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;

        int fiber_number_row=4;int fiber_number_column=12;
        //int fiber_number_row=1;int fiber_number_column=3;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 70:{

    }break;
    case 48:case 49:{   ////brush forming traces on codimensional paint, prototype for trace example
        T scale=(T).5;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).3*paint_size,paint_size);
        TV box_min=TV((T)12e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)40e-3,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;
        band_width_min=(T)2*fiber_radius;

        int fiber_number_row=2;int fiber_number_column=12;
        //int fiber_number_row=4;int fiber_number_column=12;
        //int fiber_number_row=1;int fiber_number_column=1;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;

    ////tests for dents
    case 46:{   ////dents on paint, brush moving upwards
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,paint_size,(T)1*paint_size);
        TV box_min=TV((T)0,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.incompressible.use_all_pressure_sample_on_nodes=true;
        
        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T)1.02,(T)0),ROTATION<TV>((T).0*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=5;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;

        int fiber_number_row=4;int fiber_number_column=12;
        //int fiber_number_row=1;int fiber_number_column=1;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 47:{   ////dents on paint, brush moves down and then go up 
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).4*paint_size,paint_size);
        TV box_min=TV((T)0,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.incompressible.use_all_pressure_sample_on_nodes=true;
        use_inactive_particle=true;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).95,(T)0),ROTATION<TV>((T).0*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=5;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;

        //int fiber_number_row=4;int fiber_number_column=12;
        int fiber_number_row=1;int fiber_number_column=3;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;

    ////tests for levelset
    case 9:{    ////volumes of different colors
        length_scale=(T)5e-3;thickness_scale=length_scale*(T).2;

        ////fiber
        fiber_radius=length_scale*(T).5;fiber_length=length_scale*(T)20;fiber_distance=fiber_radius*(T)2;
        fiber.object_frame=FRAME<TV>(TV::Axis_Vector(2)*(fiber_length*(T).95),ROTATION<TV>((T)0,TV::Axis_Vector(2)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles(fiber_positions,1,1,TV(),length_scale*(T)2,TV::Axis_Vector(1),TV::Axis_Vector(3));

        ////fluid
        int paint_bulk_num=2;
        ARRAY<CODIMENSIONAL_MESH> paint_meshes;paint_meshes.Resize(paint_bulk_num);
        ARRAY<ARRAY<TV> > paint_Xs;paint_Xs.Resize(paint_bulk_num);
        ARRAY<ARRAY<ARRAY<T> > > paint_phis;paint_phis.Resize(paint_bulk_num);
        {int i=1;CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(paint_meshes(i),paint_Xs(i),0,0,0,"../data/fluid/paint0");
        TV t=TV();for(int j=1;j<=paint_Xs(i).m;j++)paint_Xs(i)(j)+=t;}
        {int i=2;CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(paint_meshes(i),paint_Xs(i),0,0,0,"../data/fluid/paint1");
        TV t=TV::Axis_Vector(1)*length_scale*(T)20;for(int j=1;j<=paint_Xs(i).m;j++)paint_Xs(i)(j)+=t;}
        Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,paint_meshes,initial_X,paint_Xs);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        surface_tension*=(T)10;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e5;

        Add_Canvas();
    }break;
    case 10:{   ////test levelset splash
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=s0*(T).5;max_thickness_factor=(T)2;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).1;

        ////paint
        T paint_radius=s0*(T)2;TV paint_center=TV((T)0,s0*(T)40,(T)0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_center-(T).5*TV::Constant_Vector(paint_radius),paint_center+(T).5*TV::Constant_Vector(paint_radius),length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,paint_center-(T).5*TV::Axis_Vector(2)*paint_radius,paint_center+(T).5*TV::Axis_Vector(2)*paint_radius,paint_radius,length_scale);
        
        //TV paint_box_size=TV((T)2*paint_radius,paint_radius*(T).4,(T)2*paint_radius);
        //TV paint_box_min=paint_center-(T).5*paint_box_size;TV paint_box_max=paint_center+(T).5*paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);
        initial_V.Resize(initial_X.m);for(int i=1;i<=initial_X.m;i++)initial_V(i)=TV::Axis_Vector(2)*(T)-2.5;

        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)-2;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        example.surface_tension_force.viscous=false;
        has_rim=true;Use_Rim(has_rim);
        has_friction=false;
        friction_magnitude=(T).002;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=false;
        has_plasticity=true;
        use_implicit_elasticity=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        
        use_non_slip_on_solids=false;
        use_correcting_particle_position_on_solids=false;
        use_levelset=true;
        levelset_region_number=2;

        adaptive_length_scale_min=length_scale*(T).25;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_levelset=false;

        //////sphere
        //{T sphere_radius=s0*(T)8;
        //RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(sphere_radius,(T)1);
        //rigid_body->X()=TV();
        //rigid_body->X().y=s0*(T)30;
        //rigid_body->Rotation()=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        //rigid_bodies.Append(rigid_body);}

        ////box
        {T size=s0*(T)3;
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Box(TV::Constant_Vector(size));
        rigid_body->X().y=s0*(T)35;
        rigid_bodies.Append(rigid_body);}

        //Add_Canvas();

        TV domain_size=TV::Constant_Vector((T)50*s0);domain_size.y=(T)200*s0;TV domain_center=paint_center;
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for misc, mpi, meshing, etc.
    case 32:{   ////test mpi
        T scale=(T).5;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        T domain_size=length_scale*(T)200;global_grid_domain=RANGE<TV>(TV::Constant_Vector((T)-.5*domain_size),TV::Constant_Vector((T).5*domain_size));

        ////paint
        T height=one_over_scale*length_scale*(T)1;TV box_size=TV(height,height,height*(T)10);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).05;fiber_distance=fiber_radius*(T)5;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4,fiber_length*(T).8,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;
        fiber.fiber_bending_perturb_frac=TV((T).005,(T)0,(T)0.002);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,24,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        Add_Canvas();
    }break;
    case 33:{   ////mpi brush
        T scale=(T).2;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T height=one_over_scale*length_scale*(T)1;TV box_size=TV(height,height,height*(T)10);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).05;fiber_distance=fiber_radius*(T)4;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4,fiber_length*(T).8,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,50,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(fiber_length_scale*(T)40,fiber_length_scale*(T)10,fiber_length_scale*(T)12);TV domain_center=TV(length_scale*(T)18,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 34:{   ////mpi, higher res, 8-core test
        T scale=(T).15;T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T height=one_over_scale*length_scale*(T)1;TV box_size=TV(height*(T)2,height,height*(T)10);
        TV box_min=TV(-box_size.x*(T).25,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).75,box_size.y,box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////brush
        fiber_length_scale=(T)1e-3;
        fiber_radius=fiber_length_scale*(T).05;fiber_distance=fiber_radius*(T)4;fiber_length=fiber_length_scale*(T)20;
        fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4,fiber_length*(T).8,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,1,50,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(fiber_length_scale*(T)40,fiber_length_scale*(T)10,fiber_length_scale*(T)12);TV domain_center=TV(length_scale*(T)18,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 35:{   /////mpi, better fiber layout, single trace
        T scale=(T).5;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T paint_height=(T)2e-3;TV paint_box_size=TV(paint_height,paint_height,paint_height*(T)1);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).4;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)2;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)7.7e-3,fiber_length*(T).75,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        int fiber_number_row=1;int fiber_number_column=75;
        T fiber_width=fiber_number_column*fiber_distance;
        //fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T).0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(fiber_width*(T)10,fiber_width*(T)2,fiber_width*(T)1.2);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 36:{   /////mpi, better fiber layout, multiple traces
        T length_scale_factor=(T).25*scale;//T one_over_length_scale_factor=(T)1/length_scale_factor;
        length_scale=length_scale_factor*(T)1e-3;thickness_scale=length_scale*(T).2;

        ////paint
        T paint_height=(T)2e-3;TV paint_box_size=TV(paint_height,paint_height,paint_height*(T)8);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).4;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        use_levelset=false;

        //////brush
        //fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)2;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)7.5e-3,fiber_length*(T).75,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        //fiber.implicit_fiber_segment_number=3;
        //fiber.fiber_bending_start_frac=(T).6;   
        //fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.005);
        //int fiber_number_row=1;int fiber_number_column=75;
        //T fiber_width=fiber_number_column*fiber_distance;
        ////fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T).0);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        //Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)2;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)7.7e-3,fiber_length*(T).77,(T)0),ROTATION<TV>((T)-.1*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        int fiber_number_row=1;int fiber_number_column=75;
        T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).0);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV(fiber_width*(T)10,fiber_width*(T)2,fiber_width*(T)1.2);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 101:{   ////mpi signed distance test
        T length_scale_factor=(T).25*scale;
        length_scale=length_scale_factor*(T)1e-3;
        thickness_scale=length_scale*(T).2;

        ////paint
        T paint_height=(T)2e-3;
        T paint_depth=paint_height;
        T paint_width=(T)10*paint_height;
        TV paint_box_size=TV(paint_width,paint_height,paint_depth);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;

        use_levelset=true;
        levelset_region_number=2;

        TV domain_size=TV((T)4*length_scale+paint_box_size.x,(T)4*length_scale+paint_box_size.y,(T)4*length_scale+paint_box_size.z); //?
        TV domain_center=TV((T)0,(T)0.5*paint_height,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 102:{  ////snap meshing operation
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;
        max_thickness_factor=(T)2;

        ////paint
        {T paint_size=(T)5e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.y=(T)0;TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);}
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);}
        int first_part_n=initial_X.m;

        {T paint_size=(T)5e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        //TV paint_box_min=(T)1*paint_box_size;paint_box_min.y=(T)0;
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)1.5*paint_box_size.x;paint_box_min.y=(T)0;TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);}
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);}

        initial_V.Resize(initial_X.m);for(int i=1;i<=initial_X.m;i++)if(i<=first_part_n)initial_V(i)=TV((T).1,(T)0,(T)0);else initial_V(i)=TV();

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;

        //////kinematic solid
        //TV rigid_body_size=TV::Constant_Vector((T)2e-3);
        //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        //kinematic_rigid_body->X()=TV();
        //kinematic_rigid_body->X().y=(T).5*rigid_body_size.y+thickness_scale*(T)1.5;
        ////rigid_bodies.Append(kinematic_rigid_body);
        //kinematic_rigid_body->name="kin_obj";
        //kinematic_rigid_bodies.Append(kinematic_rigid_body);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for paint spray example
    case 51:{   ////centripetal force, prototype for rotating cylinder example
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder
        T cylinder_length=(T)100e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        //T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).5*paint_size,paint_size);
        //TV box_min=TV((T)-paint_box_size.x*(T).5,cylinder_radius-length_scale*(T).2,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //TV e1=TV::Axis_Vector(2)*cylinder_radius;TV e2=e1+TV::Axis_Vector(2)*paint_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        //T tube_height=(T)12e-3;
        //T tube_radius=(T)4e-3;
        //T tube_height2=(T)80e-3;
        //TV tube_e1=TV(cylinder_radius*(T).5,tube_height2+cylinder_radius,(T)0);
        //TV tube_dir=TV::Axis_Vector(2)*(T)-1;
        //tube=new PAINT_TUBE<TV>(tube_e1,tube_dir,tube_height,tube_radius,length_scale,(T).0,(T).1);
        //tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        //Add_Tube_Rigid_Body(*tube);

        //{int i=1;T paint_radius=(T)5e-3;TV paint_center=TV((T)-.5*cylinder_radius,cylinder_radius+(T)40e-3,(T)0);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        //{int i=2;T paint_radius=(T)5e-3;TV paint_center=TV((T).2*cylinder_radius,cylinder_radius+(T)80e-3,(T)30e-3);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        //{int i=3;T paint_radius=(T)5e-3;TV paint_center=TV((T)-.9*cylinder_radius,cylinder_radius+(T)60e-3,(T)-20e-3);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        {/*int i=1;*/T paint_radius=(T)5e-3;TV paint_center=TV((T)-.5*cylinder_radius,cylinder_radius+(T)40e-3,(T)0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        {/*int i=2;*/T paint_radius=(T)5e-3;TV paint_center=TV((T).2*cylinder_radius,cylinder_radius+(T)80e-3,(T)30e-3);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        {/*int i=3;*/T paint_radius=(T)5e-3;TV paint_center=TV((T)-.6*cylinder_radius,cylinder_radius+(T)60e-3,(T)-20e-3);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        use_levelset=true;
        levelset_region_number=4;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 52:{
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder
        T cylinder_length=(T)100e-3;
        T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);

        //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("cylinder",1,0,true,true);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T paint_size=(T)15e-3;TV paint_box_size=TV((T)2*paint_size,(T).2*paint_size,(T)4*paint_size);
        TV box_min=TV((T)-paint_box_size.x*(T).5,cylinder_radius-(T)2e-4,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //TV e1=TV::Axis_Vector(2)*cylinder_radius;TV e2=e1+TV::Axis_Vector(2)*paint_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(true);
        use_levelset=true;
        levelset_region_number=3;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 53:{   ////test twisting viscous thin film
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder
        T cylinder_length=(T)100e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        //T paint_size=(T)20e-3;TV2 paint_box_size_zx=TV2((T)4*paint_size,paint_size);TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,paint_box_size_zx,length_scale,2);

        //T paint_size=(T)20e-3;TV paint_box_size=TV(paint_size,paint_size*(T).1,paint_size*(T)4);
        //TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);

        T paint_size=(T)10e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV paint_box_min=TV(paint_size*(T)-.5,paint_size*(T)-.5,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);
        TV e1=TV((T)cylinder_radius*(T).5,cylinder_radius*(T).5,cylinder_length*(T).5);TV e2=e1+TV::Axis_Vector(3)*paint_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,paint_size*(T).5,length_scale);

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(true);
        use_levelset=true;
        levelset_region_number=3;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 54:{   ////test twisting viscous thin film 2
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder 1
        T cylinder_length=(T)10e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T paint_size=(T)20e-3;TV2 paint_box_size_zx=TV2((T)4*paint_size,paint_size);TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,paint_box_size_zx,length_scale,2);

        //T paint_size=(T)20e-3;TV paint_box_size=TV(paint_size,paint_size*(T).1,paint_size*(T)4);
        //TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);

        //T paint_size=(T)10e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        //TV paint_box_min=TV(paint_size*(T)-.5,paint_size*(T)-.5,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);
        //TV e1=TV((T)cylinder_radius*(T).5,cylinder_radius*(T).5,cylinder_length*(T).5);TV e2=e1+TV::Axis_Vector(3)*paint_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,paint_size*(T).5,length_scale);

        ////cylinder 2
        {T cylinder_length=(T)100e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV((T)0,(T)0,cylinder_length+(T)4*paint_size);kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj2";}

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=false;
        has_surface_tension=false;
        surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(false);
        use_levelset=true;
        levelset_region_number=3;
        remesh=false;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 55:{   ////test twisting filament
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder 1
        T cylinder_length=(T)10e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T paint_size=(T)20e-3;TV2 paint_box_size_zx=TV2((T)4*paint_size,paint_size);
        TV e1=TV(paint_size*(T)-.5,(T)cylinder_radius*(T).5,cylinder_length*(T).5);TV e2=e1+TV::Axis_Vector(3)*(T)4*paint_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        ////cylinder 2
        {T cylinder_length=(T)10e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV((T)0,(T)0,cylinder_length+(T)4*paint_size);kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj2";}

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 56:{   ////filament twisting a cylinder
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder 1
        T cylinder_length=(T)100e-3;T cylinder_radius=(T)8e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).0,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T paint_size=(T)60e-3;TV2 paint_box_size_zx=TV2((T)4*paint_size,paint_size);
        TV e1=TV(-cylinder_radius-length_scale,(T)-.5*paint_size,(T)-.5*paint_size);TV e2=e1+TV((T)0,(T)1,(T).1)*paint_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);
        initial_V.Resize(initial_X.m);initial_V.Fill(TV((T).1,(T)0,(T)0));

        has_gravity=true;fiber.use_gravity=has_gravity;gravity=TV::Axis_Vector(1)*(T)9.8;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for cheese example, elasticity
    ////tests for elasticity
    case 131:{  ////test elasticity 2, standing
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        //////volume
        //T height=length_scale*(T)5;TV box_size=TV(height,height,height);
        //TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        //TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        ////film
        T height=length_scale*(T)5;TV2 box_size=TV2(height,height);TV box_min=TV();
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,box_size,length_scale);

        //////filament
        //T height=length_scale*(T)5;TV e1=TV();TV e2=e1+TV::Axis_Vector(2)*height;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        surface_tension*=(T).01;
        has_viscosity=false;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=true;
        use_levelset=false;
        remesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        example.elasticity_force.Use_Constant_Extrapolation(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 132:{  ////test elasticity 3, dragging
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;max_thickness_factor=(T)2;

        ////paint
        //////volume
        //T height=length_scale*(T)5;TV box_size=TV(height,height,height);
        //TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        //TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        ////film
        //T height=length_scale*(T)5;TV2 box_size=TV2(height,height);TV box_min=TV();
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,box_size,length_scale);
        //////filament
        //T height=length_scale*(T)5;TV e1=TV();TV e2=e1+TV::Axis_Vector(2)*height;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        ////paint
        T paint_size=(T)4e-3;TV paint_box_size=TV((T)8*paint_size,(T).2*paint_size,paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,paint_box_size_zx,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        surface_tension*=(T).01;
        has_viscosity=false;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=true;
        use_levelset=false;
        remesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        example.elasticity_force.Use_Constant_Extrapolation(true);

        ////kinematic solid
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV::Constant_Vector(paint_size*(T)1.5);
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();
            T offset=(T)14e-3;
            if(i==1)kinematic_rigid_body->X().x=rigid_body_size.x*(T).5+offset;
            else kinematic_rigid_body->X().x=-rigid_body_size.x*(T).5-offset;
            kinematic_rigid_body->X().y=-rigid_body_size.y*(T).5;

            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);
        }

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 133:{  ////test elasticity 4, twisting
        T scale=(T)1;length_scale=scale*(T)2e-3;thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).001;

        ////cylinder 1
        T cylinder_length=(T)10e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint
        T paint_size=(T)20e-3;TV2 paint_box_size_zx=TV2((T)4*paint_size,paint_size);TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,paint_box_min,paint_box_size_zx,length_scale,2);

        //T paint_size=(T)20e-3;TV paint_box_size=TV(paint_size,paint_size*(T).1,paint_size*(T)4);
        //TV paint_box_min=TV(paint_size*(T)-.5,(T)0,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);

        //T paint_size=(T)10e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        //TV paint_box_min=TV(paint_size*(T)-.5,paint_size*(T)-.5,cylinder_length*(T).5);TV paint_box_max=paint_box_min+paint_box_size;
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,paint_box_min,paint_box_max,length_scale);
        //TV e1=TV((T)cylinder_radius*(T).5,cylinder_radius*(T).5,cylinder_length*(T).5);TV e2=e1+TV::Axis_Vector(3)*paint_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Cylinder(initial_mesh,initial_X,e1,e2,paint_size*(T).5,length_scale);

        ////cylinder 2
        {T cylinder_length=(T)10e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV((T)0,(T)0,cylinder_length+(T)4*paint_size);kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj2";}

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=false;
        surface_tension*=(T).01;
        has_viscosity=false;
        use_implicit_viscosity=false;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=true;
        use_levelset=false;
        remesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        example.elasticity_force.Use_Constant_Extrapolation(true);
        remesh=false;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 134:{   ////test segment contact
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;
        max_thickness_factor=(T)2;min_rim_thickness_factor=(T).05;

        ////paint
        //T tube_height=length_scale*(T)5;
        //T paint_height=length_scale*(T)30;
        //T radius=length_scale*(T)2;
        T filament_length=length_scale*(T)20;
        TV dir1=TV::Axis_Vector(1);TV c1=TV();TV e1=c1-(T).5*filament_length*dir1;TV e2=c1+(T).5*filament_length*dir1;
        TV dir2=TV::Axis_Vector(3);TV c2=TV((T).3*length_scale,(T)5*length_scale,(T).7*length_scale);TV e3=c2-(T).5*filament_length*dir2;TV e4=c2+(T).5*filament_length*dir2;
        
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);int n=initial_X.m;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e3,e4,length_scale);
        initial_V.Resize(initial_X.m);for(int i=1;i<=initial_V.m;i++){if(i>n)initial_V(i)=TV::Axis_Vector(2)*(T)-.1;else initial_V(i)=TV();}

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=false;
        has_plasticity=true;
        use_implicit_elasticity=true;
        use_levelset=false;
        use_non_slip_on_solids=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_mesh_edge_connect=true;
        example.viscosity_force.Use_Staggered_DoF(false);
        example.viscosity_force.Use_Constant_Extrapolation(true);
        example.elasticity_force.Use_Staggered_DoF(false);  ////TODO: fix the staggered dof case for elasticity
        example.elasticity_force.Use_Constant_Extrapolation(true);
        remesh=true;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 57:{   ////pizza cheese prototype
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        //length_scale=scale*(T).5e-3;thickness_scale=length_scale*(T).5;
        //max_thickness_factor=(T)1.5;min_thickness_factor=(T).15;
        length_scale=scale*(T).5e-3;thickness_scale=length_scale*(T)1;
        max_thickness_factor=(T)1.5;min_thickness_factor=(T).01;

        ////paint
        T paint_size=(T)20e-3;TV paint_box_size=TV((T).4*paint_size,paint_size,(T).4*paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,paint_box_size_zx,length_scale,2);
        TV e1=TV(-paint_box_size.z*(T).5,(T)0,(T)0);TV e2=TV(paint_box_size.z*(T).5,(T)0,(T)0);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        //TV center=(T).5*(box_min+box_max);T r=paint_box_size_zx.x;Initialize_Pizza_Thickness(initial_X,initial_thickness,center,r,thickness_scale*max_thickness_factor,thickness_scale*min_thickness_factor);

        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).001;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        has_elasticity=false;
        has_plasticity=true;
        use_implicit_elasticity=true;
        elastic_modulus=(T)1e3;
        use_levelset=false;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        use_inactive_particle=false;
        example.viscosity_force.Use_Staggered_DoF(true);
        example.viscosity_force.Use_Constant_Extrapolation(false);

        ////kinematic solid
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV(paint_size*(T)1,paint_size*(T).1,paint_size*(T)1.5);
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();T offset=paint_box_size.x*(T).0;
            if(i==1)kinematic_rigid_body->X().x=rigid_body_size.x*(T).5+offset*(T).5;
            else kinematic_rigid_body->X().x=-rigid_body_size.x*(T).5-offset*(T).5;
            kinematic_rigid_body->X().y=-rigid_body_size.y*(T).5;

            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);

            ////ground
            if(i==2){
                TV ground_size=TV(paint_size*(T)10,paint_size*(T).2,paint_size*(T)10);
                RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
                ground_rigid_body->X().y=-ground_size.y*(T).5-rigid_body_size.y;
                rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";
            }
        }

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for paint splash
    case 100:{   /////tank of paint with different layers of color
        //based on test 36...
        T length_scale_factor=(T).25*scale;
        //length_scale=length_scale_factor*(T)1e-3;
        length_scale=length_scale_factor*(T)1e-3;
        thickness_scale=length_scale*(T).2;

        ////paint
        T paint_height=(T)2e-3;
        T paint_width=(T)4*paint_height;
        TV paint_box_size=TV(paint_width,paint_height,paint_width);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_viscosity=true;
        kinematic_viscosity*=(T)1e5;
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=(T).2;
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;

        use_levelset=true;
        levelset_region_number=2;

        TV domain_size=TV((T)2*paint_box_size.x,(T)2*paint_box_size.y,(T)2*paint_box_size.z); //?
        //TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        TV domain_center=TV((T)0,(T)0.5*paint_height,(T)0); //?
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for mayo
    case 256:{ ////knife striking fluid volume
        length_scale=(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->object_frame.t.z=(T).02;
        knife->object_frame.t.y=(T).0075;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;

        knife->motion_type=KNIFE<TV>::SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        //fluid
        has_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.mesher.use_break_thin_triangles=false;        

        T height=7*length_scale;TV box_size=TV(height,height,height*(T)2);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);

        box_min.x-=(T).075;
        box_max.x-=(T).075;

        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        TV domain_size=TV::Constant_Vector((T).21);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();

    }break;
    case 257:{ ////knife and fluid thin sheet
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->object_frame.t.z=(T).02;
        knife->object_frame.t.y=(T).0055;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;

        knife->motion_type=KNIFE<TV>::SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        T paint_size=(T)15e-3;TV paint_box_size=TV((T)4*paint_size,paint_size,(T)2*paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)-6*paint_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;

        TV domain_size=TV::Constant_Vector((T).21);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 258:{ ////knife circular movement
        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::CIRCLE;
        knife->Set_Prescribed_Motion_For_Test();
    }break;
    case 259:{ ////knife slicing movement
        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::SLICE;
        knife->Set_Prescribed_Motion_For_Test();
    }break;
    case 260:{ ////knife stamping movement
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->object_frame.t.z=(T)-.0005;//-.002;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::STAMP;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        T paint_size=(T)30e-3;TV paint_box_size=TV((T)paint_size,paint_size,(T)2*paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;
        paint_box_min.y=(T)0;
        paint_box_min.z=(T)-2*paint_size;
        //paint_box_min.x=(T)0.5*paint_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;

        //has_surface_tension=false;
        //example.mesher.use_continuous_meshing_operations=true;
        //example.incompressible.use_all_pressure_sample_on_nodes=true;

        TV domain_size=TV::Constant_Vector((T).21);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 261:{   ////mass-spring fiber with triangle paint
        T scale=(T).25;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)8;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;

        fiber.implicit_fiber_segment_number=3;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;
        int fiber_number_row=1;int fiber_number_column=3;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).02);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;

        std::string dir="nov14a";
        T t0=(T)0;
        T t1=(T)2;
        fiber.Initialize_Attributes_From_File(dir,t0,t1);


        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 262:{  ////viscosity test
        length_scale=scale*(T)1e-3;

        ////fluid
        //T paint_radius=(T)5e-3;TV paint_center=TV((T)-.5*cylinder_radius,cylinder_radius+(T)40e-3,(T)0);
        T paint_radius=(T)5e-3;
        TV paint_center=TV();
        paint_center.y=(T)3*paint_radius;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        //example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 263:{ ////paint falling onto knife
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        //knife->object_frame.t.z=(T)-.0005;//-.002;
        //knife->object_frame.t.x=(T)1;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        //knife->rigid_body->Rotation()=knife->object_frame.r;
        knife->object_frame.r=knife->object_frame.r*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));
        //knife->object_frame.r=ROTATION<TV>((T)pi/(T)4,TV::Axis_Vector(1));
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::STATIONARY;
        knife->Set_Prescribed_Motion_For_Test();


        ////fluid
        T paint_radius=(T)3e-3;
        TV paint_center=TV();
        paint_center.y=(T)3*paint_radius;
        paint_center.z=(T)-4*paint_radius;
        paint_center.x=(T)1*paint_radius;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
        paint_center.z=(T)-7*paint_radius;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        //example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 264:case 265:{ ////knife flipping over and spreading (and higher resolution variant)
        if(test==264)length_scale=scale*(T)1e-3;
        else length_scale=scale*(T)5e-4;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        if(test==264)kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        else kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_highres",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.005;
        T knife_height=(T)0.008;

        knife=new KNIFE<TV>();
        //knife->object_frame.t.z=(T)-.0005;//-.002;
        //knife->object_frame.t.x=(T)1;
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        //knife->rigid_body->Rotation()=knife->object_frame.r;
        //knife->object_frame.r=knife->object_frame.r*ROTATION<TV>((T)half_pi,TV::Axis_Vector(3));
        //knife->object_frame.r=ROTATION<TV>((T)pi/(T)4,TV::Axis_Vector(1));
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::FLIP_AND_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        if(test==264){
            T paint_radius=(T)3e-3;
            TV paint_center=TV();
            paint_center.y=(T)3*paint_radius+knife_height;
            paint_center.z=(T)-4*paint_radius;
            paint_center.x=(T)-1.2*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-7*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-10*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
        }
        else{
            T paint_radius=(T)3e-3;
            TV paint_center=TV();
            paint_center.y=(T)3*paint_radius+knife_height;
            paint_center.z=(T)-4*paint_radius;
            paint_center.x=(T)-1.2*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-6*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-8*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-10*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
        }


        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;//true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        //example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 266:{ ////knife rotation velocity test
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::ROTATE;
        knife->Set_Prescribed_Motion_For_Test();

    }break;
    case 267:{ ////knife translation velocity test
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        knife=new KNIFE<TV>();
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::TRANSLATE;
        knife->Set_Prescribed_Motion_For_Test();
    }break;
#if 0
    case 268:{ ////spread test w/ fluid starting on surface
        length_scale=scale*(T)1e-3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        T knife_height=(T)0.008;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        if(test==264){
            T paint_radius=(T)3e-3;
            TV paint_center=TV();
            paint_center.y=(T)3*paint_radius+knife_height;
            paint_center.z=(T)-4*paint_radius;
            paint_center.x=(T)-1.2*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-7*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
            paint_center.z=(T)-10*paint_radius;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);
        }

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        //example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
#endif
    case 269:case 270:case 271:case 272:case 273:{ ////knife trace test
        length_scale=scale*(T)1e-3;
        //thickness_scale=length_scale*(T)0.5;
        //thickness_scale=length_scale*(T)0.1;
        thickness_scale=length_scale*(T)0.5;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        if(test==269)     kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_edge_original",(T)1.0,(T)0.0,true,true,&float_stream_type);
        else if(test==270)kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_edge_mod1",(T)1.0,(T)0.0,true,true,&float_stream_type);
        else if(test==271)kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_edge_mod2",(T)1.0,(T)0.0,true,true,&float_stream_type);
        else              kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_edge_mod1pt5",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        T knife_height=(T)0.0007;
        if(test==271)knife_height=(T)0.0005;

        //knife_height=(T)0.0004;
        //knife_height=(T)0.00035;
        //knife_height=(T)0.0007;
        knife_height=(T).0008;

        if(test==273)knife_height=(T)0;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::STRAIGHT_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        //T paint_size=(T)15e-3;TV paint_box_size=TV((T)2*paint_size,paint_size,(T)2*paint_size);
        T paint_size=(T)15e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)-6.5*paint_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=false;
        //has_incompressibility=false;
        has_surface_tension=true;
        //has_surface_tension=false;
        surface_tension*=(T).01;
        has_viscosity=true;
        has_friction=false;
        //has_viscosity=false;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_break_thin_triangles=true;//false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;//false;
        //example.mesher.use_boundary_vertex_snap=true;

        ////canvas
        has_friction=false;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 274:{ ////knife turning
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife_edge_mod1pt5",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        T knife_height=(T)0.00035;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::TURN_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();

        ////fluid
        T paint_size=(T)15e-3;TV paint_box_size=TV((T)2*paint_size,paint_size,(T)2*paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;paint_box_min.x=(T)-6.5*paint_size;paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=false;//true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_break_thin_triangles=true;//false;        
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;//false;
        //example.mesher.use_boundary_vertex_snap=true;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 275:{ ////rock falling
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("rock2",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="rock";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        T rock_height=(T)0.001;

        rock=new ROCK<TV>();
        rock->motion_type=ROCK<TV>::FALL2;
        rock->object_frame.t.y=rock_height;
        rock->rigid_body=kinematic_rigid_body;
        rock->rigid_body->X()=rock->object_frame.t;
        rock->rigid_body->Rotation()=rock->object_frame.r;

        rock->use_slow_motion=true;
        rock->slow_motion_factor=(T)0.25;
        rock->slow_motion_threshold=(T)0.1;
        
        //rock->motion_type=TURN_SPREAD;
        rock->Set_Prescribed_Motion_For_Test();

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 276:{ ////mayo spread, initialize fluid from file
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife2",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        T knife_height=(T)0.007;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::STRAIGHT_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();


        ////fluid
        //example.mesher.use_triangle_mesh=false;
        //std::string path="blob_pt25";
        std::string path="blob";
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,0,1);
        for(int i=1;i<=initial_X.m;i++){
            initial_X(i)(1)-=(T)0.09;
            initial_X(i)(3)-=(T)0.03;
        }

        LOG::cout<<"# tets: "<<initial_mesh.tetrahedron_mesh.elements.m<<std::endl<<std::endl;

        /*
        T height=7*length_scale;TV box_size=TV(height,height,height*(T)2);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);

        box_min.x-=(T).075;
        box_max.x-=(T).075;

        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        */

        has_friction=true;
        use_correcting_particle_position_on_solids=true;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        //example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_adaptive_tetrahedron_remeshing=true; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        //example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;

        Use_Rim(true);

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        domain_size(1)=(T)0.25;
        domain_size(3)=(T)0.06;
        //domain_size(3)=(T)0.04;
        domain_center(3)=(T)-0.03;

        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Canvas();

    }break;
    case 277:{ ////levelset test
        T scale=(T).25;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).5;

        //////paint
        T paint_size=(T)30e-3;TV paint_box_size=TV(paint_size,paint_size,(T)0.25*paint_size);
        int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        TV paint_box_min=-(T).5*paint_box_size;//paint_box_min.x=(T)-2*paint_size;
        paint_box_min.y=(T)0;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,paint_box_min,length_scale,2);

        has_gravity=false;fiber.use_gravity=has_gravity;
        has_incompressibility=false;
        has_surface_tension=false;
        has_viscosity=false;

        use_levelset=true;
        levelset_region_number=2;

        has_gravity=false;
        has_incompressibility=false;
        has_surface_tension=false;
        has_viscosity=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_triangles=true;//false;        
        //example.mesher.use_adaptive_triangle_remeshing=true;
        //example.mesher.use_continuous_meshing_operations=true;//false;
        //example.mesher.preserve_thickness_in_remeshing=true;
        //example.mesher.use_boundary_vertex_snap=true;

        //update_slow_interface_phis=false;
        //update_slow_interface_phis=true;

        //example.mesher.use_adaptive_triangle_remeshing=true;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 278:{ ////bunny test
        length_scale=scale*(T)1e-3;

        ////fluid
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Tetrahedralized_Volume_From_File(initial_mesh,initial_X,"../../Public_Data/Archives/Tetrahedralized_Volumes/bunny.tet.gz",(T)25*length_scale);
        T initial_height=(T)0.01;
        for(int i=1;i<=initial_X.m;i++){
            initial_X(i)(2)+=initial_height;
        }

        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=false;
        //has_surface_tension=true;
        //surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        //kinematic_viscosity*=(T)1e3;
        has_friction=false;
        //use_levelset=true;
        //levelset_region_number=3;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        use_correcting_particle_position_on_solids=false;//?
        
        adaptive_length_scale_min=length_scale*(T).25;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_levelset=false;

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 279:case 280:case 281:case 282:case 283:case 284:case 289:{ ////sauce test
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;


        STREAM_TYPE float_stream_type(1.f);
        T height_offset=(T)5e-4;

        //rigid body
        if(test!=289){
            //RIGID_BODY<TV>* kinematic_rigid_body;
            //kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("ladle",(T)1.0,(T)0.0,true,true,&float_stream_type);
            //T sphere_radius=length_scale*(T)5;
            //T sphere_radius=(T)5e-3;
            //sphere_radius*=(T)0.85;
            //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(sphere_radius,(T)1);
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("ladle",(T)1.0,(T)0.0,true,true,&float_stream_type);
            kinematic_rigid_body->name="ladle";
            kinematic_rigid_body->X()=TV();
            rigid_bodies.Append(kinematic_rigid_body);
            kinematic_rigid_bodies.Append(kinematic_rigid_body);
            T ladle_height=(T)0.005+height_offset;

            ladle=new LADLE<TV>();
            ladle->object_frame.t.y=ladle_height;
            ladle->rigid_body=kinematic_rigid_body;
            ladle->rigid_body->X()=ladle->object_frame.t;
            ladle->rigid_body->Rotation()=ladle->object_frame.r;
            
            if(test==279)ladle->motion_type=LADLE<TV>::COMBO;
            else if(test==280)ladle->motion_type=LADLE<TV>::COMBO2;
            else if(test==281)ladle->motion_type=LADLE<TV>::COMBO2_SLOWMO;
            else if(test==282)ladle->motion_type=LADLE<TV>::COMBO2_SLOWMO_THE_RECKONING;
            else if(test==283)ladle->motion_type=LADLE<TV>::COMBO3_THE_JOURNEY_HOME;
            else if(test==284)ladle->motion_type=LADLE<TV>::COMBO4_A_NEW_HOPE;
            ladle->Set_Prescribed_Motion_For_Test();
        }

        ////pizza surface
        RIGID_BODY<TV>* pizza_base=&rigid_standard_tests.Add_Rigid_Body("pizza_flt",(T)1.,(T)0.,true,true,&float_stream_type);
        rigid_bodies.Append(pizza_base);
        pizza_base->name="pizza_base";


        ////fluid
        //example.mesher.use_triangle_mesh=false;
        //std::string path="blob_pt25";
        //std::string path="blob";
        //std::string path="sauce2";
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,path,0,1);
        //for(int i=1;i<=initial_X.m;i++){
        //    initial_X(i)(2)+=height_offset;
        //}
        std::string path="dec31_sauce_settle";
        if(scale<(T)1)path+="_hires";
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,140,1);

        LOG::cout<<"# tets: "<<initial_mesh.tetrahedron_mesh.elements.m<<std::endl<<std::endl;

        has_friction=true;
        use_correcting_particle_position_on_solids=false;//true;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=false;//true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;//true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        use_non_slip_on_solids=true;
        has_adhesion=false;
        use_correcting_particle_position_on_solids=false;

        Use_Rim(true);

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //domain_size(1)=(T)0.25;
        //domain_size(3)=(T)0.06;
        //domain_center(3)=(T)-0.03;

        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        //Add_Canvas();
    }break;
    case 285:{ ////mayo spread, initialize fluid from file
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife2",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        T knife_height=(T)0.007;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::MULTIPLE_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();


        ////fluid
        //example.mesher.use_triangle_mesh=false;
        //std::string path="blob_pt25";
        std::string path="blob";
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,0,1);
        for(int i=1;i<=initial_X.m;i++){
            initial_X(i)(1)-=(T)0.09;
            initial_X(i)(3)-=(T)0.03;
        }

        LOG::cout<<"# tets: "<<initial_mesh.tetrahedron_mesh.elements.m<<std::endl<<std::endl;

        /*
        T height=7*length_scale;TV box_size=TV(height,height,height*(T)2);
        TV box_min=TV(-box_size.x*(T).5,(T)0,-box_size.z*(T).5);
        TV box_max=TV(box_size.x*(T).5,box_size.y,box_size.z*(T).5);

        box_min.x-=(T).075;
        box_max.x-=(T).075;

        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        */

        has_friction=true;
        use_correcting_particle_position_on_solids=true;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=false;//true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        //example.mesher.use_boundary_vertex_snap=true;

        Use_Rim(true);

        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        domain_size(1)=(T)0.25;
        domain_size(3)=(T)0.06;
        domain_center(3)=(T)-0.03;

        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Canvas();

    }break;
    case 286:{ ////mayo source
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;

        //rigid body
        /*
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife3",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        T knife_height=(T)0.007;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::STRAIGHT_SPREAD;
        knife->Set_Prescribed_Motion_For_Test();
        */

        ////fluid
        has_friction=true;
        use_correcting_particle_position_on_solids=true;
        //use_correcting_particle_position_on_solids=false;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=use_1d_tube;
        //example.mesher.use_segment_mesh=true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;//false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;

        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_particle_parents=true;


        Use_Rim(true);

        ////tube
        //TV e1=TV::Axis_Vector(2)*length_scale*(T)10;
        //TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)10;
        TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)20;
        e1(1)=(T)-0.04;e1(3)=(T)-0.04;
        TV dir=(T)-1*TV::Axis_Vector(2);
        //T radius=length_scale*(T)16;
        //T radius=(T)16e-3;
        T radius=(T)0.5*(T)16e-3;
        //T height=length_scale*(T)2;
        T height=(T)2e-3;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->use_texture_coordinates=true;
        tube->initialize_texture_coordinates_box=true;
        T thickness=(T)8e-3;

        if(use_1d_tube) tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,thickness);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);

        tubes.Append(tube);
        LOG::cout<<"total particles: "<<initial_X.m<<std::endl;
        //tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;
        tube->rigid_body_type=PAINT_TUBE<TV>::NONE;
        Add_Tube_Rigid_Body(*tube);

        ////tube motion
        T current_time=(T)0;

        tube->use_prescribed_motion=true;
        FRAME<TV> original_frame=tube->object_frame;
        TV x=original_frame.t;
        ROTATION<TV> r=original_frame.r;
        ROTATION<TV> original_rotation=original_frame.r;
        ROTATION<TV> angled_rotation1=ROTATION<TV>((T)pi/(T)-10,TV::Axis_Vector(3));
        ROTATION<TV> angled_rotation2=ROTATION<TV>((T)pi/(T)20,TV::Axis_Vector(3));
        x=TV((T)-0.04,x.y,(T)-0.04);
        T sign=(T)-1;
        //T dt_factor=(T)4;
        T dt_factor=(T)10;
        T dt=(T)0.01*dt_factor;
        //T speed=(T).1;
        //T base_speed=(T).1;
        //T speed_factor=(T).2;
        //T sine_factor=(T)5;
        tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        //speed=base_speed+speed_factor*(sin(sine_factor*current_time)>(T).5?(T)1:(T)0);
        //tube->Add_Source_Speed_Bspline_Time_Value(current_time,speed);
        //for(int i=1;i<=2;i++){
        for(int i=1;i<=1;i++){
            current_time+=dt;
            tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            //speed=base_speed+speed_factor*(sin(sine_factor*current_time)>(T).5?(T)1:(T)0);
            //tube->Add_Source_Speed_Bspline_Time_Value(current_time,speed);
        }

        {////zigzag
            int steps=8;
            int steps_per_direction=2;
            T dx=(T)0.12/(T)steps_per_direction;
            T dz=(T)0.08/(T)steps;

            T dx_initial=(T)0.1/(T)steps_per_direction;
            //T dx_final=(T)0.08/(T)steps_per_direction;

            for(int i=1;i<=steps;i++){
                if(i%steps_per_direction==1)sign*=(T)-1;
                //if(sign>(T)0){r=angled_rotation1;}
                //else{r=angled_rotation2;}
                current_time+=dt;
                if(i<=steps_per_direction)x(1)+=sign*dx_initial;
                else x(1)+=sign*dx;
                //x(1)+=sign*dx;
                x(3)+=dz;
                tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
                //speed=base_speed+speed_factor*(sin(sine_factor*current_time)>(T).5?(T)1:(T)0);
                //tube->Add_Source_Speed_Bspline_Time_Value(current_time,speed);
            }
        }

        //current_time+=dt;

        sign*=(T)-1;
        {////zigzag back
            int steps=6;
            int steps_per_direction=2;
            T dx=(T)0.12/(T)steps_per_direction;
            T dz=(T)0.08/(T)steps;

            T dx_final=(T)0.1/(T)steps_per_direction;

            for(int i=1;i<=steps;i++){
                if(i%steps_per_direction==1)sign*=(T)-1;
                //if(sign<(T)0){r=angled_rotation1;}
                //else{r=angled_rotation2;}
                current_time+=dt;
                if(i<=steps-steps_per_direction)x(1)-=sign*dx;
                else x(1)-=sign*dx_final;
                //x(1)-=sign*dx;
                x(3)-=dz;
                if(i==steps)r=original_rotation;
                tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);

                //speed=base_speed+speed_factor*(sin(sine_factor*current_time)>(T).5?(T)1:(T)0);
                //if(i<=steps-2*steps_per_direction)tube->Add_Source_Speed_Bspline_Time_Value(current_time,speed);
                //else tube->Add_Source_Speed_Bspline_Time_Value(current_time,(T)0);
            }
        }

        ////tube speed
        {
            dt=dt*(T)0.25;
            T epsilon=(T)1e-7;
            //T speed=(T).1;
            //T higher_speed=(T).3;
            T speed=(T).2;
            T higher_speed=(T).27;
            int steps=(int)(current_time/dt);
            for(int i=0;i<=steps;i++){
                T time=(T)i*dt;
                if(i>=steps-4){
                    tube->Add_Source_Speed_Bspline_Time_Value(time,(T)0);
                    continue;
                }
                //if(i%6<3){
                //if(i%6>=3){
                if(i%5>=3){
                    if(i%5==0)tube->Add_Source_Speed_Bspline_Time_Value(time-epsilon,speed);
                    tube->Add_Source_Speed_Bspline_Time_Value(time,higher_speed);
                    if(i%5==1)tube->Add_Source_Speed_Bspline_Time_Value(time+epsilon,speed);
                }
                else{
                    tube->Add_Source_Speed_Bspline_Time_Value(time,speed);
                }
            }
            tube->Initialize_Source_Speed_Bspline();
        }

        {////break off and move offscreen
            current_time+=dt;
            tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            int steps=6;
            T dx=(T)0.12;
            T dy=(T)0.1;
            T dz=(T)-0.08;
            for(int i=1;i<=steps;i++){
                current_time+=dt;
                x(1)+=dx;
                x(2)+=dy;
                x(3)+=dz;
                tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
            }
        }



        //TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        TV domain_size=TV::Constant_Vector((T).2);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Bread_Slices((T)0.1);

    }break;
    case 287:{ ////knife part of previous example
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        //max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        max_thickness_factor=(T)10;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife3",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        //T knife_height=(T)0.007;
        T knife_height=(T)0;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::SWIFT_LIGHTNING_HURRICANE_TIGER_STRIKE;
        knife->Set_Prescribed_Motion_For_Test();


        ////fluid
        //std::string path="/phoenix/equigley/paint/sims/jan3/jan3_mayo_a";
        std::string path="/phoenix/equigley/paint/sims/jan3/jan3_mayo_i";
        //int frame=2400;
        int frame=2400;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,frame,1);
        //example.mesher.use_triangle_mesh=false;
        //std::string path="blob_pt25";
        //std::string path="dec22_e_volume";
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,1125,1);


        ////fluid
        has_friction=true;
        use_correcting_particle_position_on_solids=true;
        //use_correcting_particle_position_on_solids=false;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=use_1d_tube;
        //example.mesher.use_segment_mesh=true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;//false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;

        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_particle_parents=true;

        has_adhesion=true;


        Use_Rim(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        domain_size(1)=(T)0.25;
        domain_size(3)=(T)0.06;
        domain_center(3)=(T)-0.03;

        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        //Add_Canvas();
        Add_Bread_Slices((T)0.1);

    }break;
    case 288:{ ////generate box for level set tests
        length_scale=scale*(T)1e-3;
        T paint_size=(T).05;
        TV paint_box_size=TV(paint_size,(T).4*paint_size,paint_size);
        TV box_min=TV((T)-.025,(T)0,-paint_box_size.z*(T).5);
        TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        example.mesher.use_particle_parents=true;
    }break;
    case 290:{ ////spline test
        length_scale=scale*(T)1e-3;

        ////tube
        //TV e1=TV::Axis_Vector(2)*length_scale*(T)10;
        //TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)10;
        TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)20;
        e1(1)=(T)-0.04;e1(3)=(T)-0.04;
        TV dir=(T)-1*TV::Axis_Vector(2);
        //T radius=length_scale*(T)16;
        //T radius=(T)16e-3;
        T radius=(T)0.5*(T)16e-3;
        //T height=length_scale*(T)2;
        T height=(T)2e-3;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->use_texture_coordinates=true;
        tube->initialize_texture_coordinates_box=true;
        //T thickness=(T)8e-3;

        //std::string path="dec28_mayo_a_lores";
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,2400,1);

        //if(use_1d_tube) tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        //else tube->Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,thickness);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);

        tubes.Append(tube);
        //LOG::cout<<"total particles: "<<initial_X.m<<std::endl;
        tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;
        //tube->rigid_body_type=PAINT_TUBE<TV>::NONE;
        Add_Tube_Rigid_Body(*tube);

        ////tube motion
        T current_time=(T)0;

        tube->use_prescribed_motion=true;
        T dt_factor=(T)5;
        T dt=(T)0.01*dt_factor;

        FRAME<TV> original_frame=tube->object_frame;
        ROTATION<TV> r=original_frame.r;

        std::string motion_data="./Motion_Data/mayo_tube.txt";
        std::ifstream fin(motion_data.c_str());
        if(fin.fail())PHYSBAM_FATAL_ERROR("could not open tube motion data file.");

        TV x;
        while(fin>>x.x>>x.y>>x.z){
            x(1)-=(T)5;x/=(T)100;
            current_time+=dt;
            tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x,r);
        }
        fin.close();


        TV domain_size=TV::Constant_Vector((T).2);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Bread_Slices((T)0.1);

    }break;
    case 291:{ ////mayo with spline
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        adaptive_length_scale_min=length_scale*(T).3;

        ////fluid
        has_friction=true;
        use_correcting_particle_position_on_solids=true;
        //use_correcting_particle_position_on_solids=false;

        use_adaptive_length_scale_on_codim=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=use_1d_tube;
        //example.mesher.use_segment_mesh=true;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;//false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;

        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_particle_parents=true;

        Use_Rim(true);

        ////tube
        ARRAY<TV> x;

        int count=0;
        TV pos;
        std::string motion_data="./Motion_Data/mayo_tube.txt";
        std::ifstream fin(motion_data.c_str());
        if(fin.fail())PHYSBAM_FATAL_ERROR("could not open tube motion data file.");

        //PRESCRIBED_MOTION<TV> prescribed_motion;

        //T current_time_b=(T)0;
        //T dt_b=(T)0.05;

        while(fin>>pos.x>>pos.y>>pos.z){
            count++;
            pos.x-=(T)5;pos/=(T)100;
            T original_y=(T)20*(T)1e-3;
            if(count<31)pos.y=original_y;
            x.Append(pos);
            ROTATION<TV> r;
            //LOG::cout<<"pos: "<<pos<<std::endl;
            //prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time_b,pos,r);
            //current_time_b+=dt_b;
        }
        fin.close();
        //prescribed_motion.Initialize_Prescribed_Motion();

        //LOG::cout<<"x(1): "<<x(1)<<", pm: "<<prescribed_motion.Prescribed_Frame(Time_At_Frame(0)).t<<std::endl;

        //tube->prescribed_motion.Prescribed_Frame(scaled_time);

        //TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)20;
        //e1(1)=(T)-0.04;e1(3)=(T)-0.04;
        T height=(T)2e-3;
        TV e1=x(1);
        e1.y-=height;
        TV dir=(T)-1*TV::Axis_Vector(2);
        T radius=(T)0.5*(T)16e-3;
        //T height=(T)3e-3;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->use_texture_coordinates=true;
        tube->initialize_texture_coordinates_box=true;
        T thickness=(T)8e-3;

        if(use_1d_tube) tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,thickness);

        tubes.Append(tube);
        //tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;
        tube->rigid_body_type=PAINT_TUBE<TV>::NONE;
        Add_Tube_Rigid_Body(*tube);

        ////tube motion
        T current_time=(T)0;

        tube->use_prescribed_motion=true;
        T dt_factor=(T)5;
        T dt=(T)0.01*dt_factor;

        LOG::cout<<"original y: "<<tube->object_frame.t.y<<std::endl;

        FRAME<TV> original_frame=tube->object_frame;
        ROTATION<TV> r=original_frame.r;

        for(int i=1;i<=x.m;i++){
            tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x(i),r);
            current_time+=dt;
        }

        ////tube speed
        {
            dt=dt*(T)0.25;
            T epsilon=(T)1e-7;
            //T speed=(T).1;
            //T higher_speed=(T).3;
            T speed=(T).2;
            T higher_speed=(T).27;
            int steps=(int)(current_time/dt);
            for(int i=0;i<=steps;i++){
                T time=(T)i*dt;
                if(i>=steps-4){
                    tube->Add_Source_Speed_Bspline_Time_Value(time,(T)0);
                    continue;
                }
                //if(i%6<3){
                //if(i%6>=3){
                if(i%5>=3){
                    if(i%5==0)tube->Add_Source_Speed_Bspline_Time_Value(time-epsilon,speed);
                    tube->Add_Source_Speed_Bspline_Time_Value(time,higher_speed);
                    if(i%5==1)tube->Add_Source_Speed_Bspline_Time_Value(time+epsilon,speed);
                }
                else{
                    tube->Add_Source_Speed_Bspline_Time_Value(time,speed);
                }
            }
            tube->Initialize_Source_Speed_Bspline();
        }

        TV domain_size=TV::Constant_Vector((T).2);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        Add_Bread_Slices((T)0.1);

    }break;
    case 292:{ ////knife part of mayo example
        length_scale=scale*(T)1e-3;
        thickness_scale=length_scale*(T)0.5;
        rim_thickness_scale=thickness_scale;
        //max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        max_thickness_factor=(T)10;min_thickness_factor=(T).05;
        //max_thickness_factor=(T)10;min_thickness_factor=(T).5;//bigger min thickness to break triangles
        //max_thickness_factor=(T)20;min_thickness_factor=(T).1;
        //adaptive_length_scale_min=length_scale*(T).3;
        //adaptive_length_scale_min=length_scale*(T).1;
        //adaptive_length_scale_min=length_scale*(T).5;
        adaptive_length_scale_min=length_scale*(T).35;

        //min_rim_thickness_factor=(T).01;//get segments?
        //max_rim_thickness_factor=(T)10;

        //rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("knife3",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="knife";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        //T knife_height=(T)0.008;
        //T knife_height=(T)0.007;
        T knife_height=(T)0;

        knife=new KNIFE<TV>();
        knife->object_frame.t.y=knife_height;
        knife->rigid_body=kinematic_rigid_body;
        knife->rigid_body->X()=knife->object_frame.t;
        knife->rigid_body->Rotation()=knife->object_frame.r;
        
        knife->motion_type=KNIFE<TV>::DO_IT_IN_BLENDER;
        //knife->motion_type=KNIFE<TV>::DO_IT_IN_BLENDER2_THE_RECKONING;
        knife->Set_Prescribed_Motion_For_Test();


        ////fluid
        //std::string path="/phoenix/equigley/paint/sims/jan3/jan3_mayo_i";
        //int frame=2400;
        std::string path;
        if(scale<=(T)1.5)path="/phoenix/equigley/paint/sims/jan6/j6_mayo_a_s1.5";
        else path="/phoenix/equigley/paint/sims/jan5/j5_mayo_a_s2";
        int frame;
        if(scale<=(T)1.5)frame=2000;
        else frame=1700;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,path,frame,1);


        ////fluid
        has_friction=true;
        friction_magnitude=(T).9;
        //use_correcting_particle_position_on_solids=true;
        use_correcting_particle_position_on_solids=false; //don't die, sim....

        use_adaptive_length_scale_on_codim=true;

        example.mesher.use_segment_mesh=true;

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=false;
        example.mesher.use_segment_mesh=use_1d_tube;
        //example.mesher.use_break_thin_segments=true;
        //example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_triangle_mesh=true;//false;
        //example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        if(scale<=(T)1.5) example.mesher.use_adaptive_tetrahedron_remeshing=true;
        else example.mesher.use_adaptive_tetrahedron_remeshing=false; // true to get trace?
        example.mesher.use_continuous_meshing_operations=true;//false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;

        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_particle_parents=true;

        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;

        example.viscosity_force.use_variable_viscosity_for_codim=true;

        has_adhesion=true;


        Use_Rim(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        //domain_size(1)=(T)0.25;
        //domain_size(3)=(T)0.06;
        //domain_center(3)=(T)-0.03;

        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);

        //Add_Canvas();
        Add_Bread_Slices((T)0.1);

    }break;
    case 293:{////rigid body motion for mayo source
        ////tube
        ARRAY<TV> x;

        int count=0;
        TV pos;
        std::string motion_data="./Motion_Data/mayo_tube.txt";
        std::ifstream fin(motion_data.c_str());
        if(fin.fail())PHYSBAM_FATAL_ERROR("could not open tube motion data file.");

        //PRESCRIBED_MOTION<TV> prescribed_motion;

        //T current_time_b=(T)0;
        //T dt_b=(T)0.05;

        while(fin>>pos.x>>pos.y>>pos.z){
            count++;
            pos.x-=(T)5;pos/=(T)100;
            T original_y=(T)20*(T)1e-3;
            if(count<31)pos.y=original_y;
            x.Append(pos);
            ROTATION<TV> r;
            //LOG::cout<<"pos: "<<pos<<std::endl;
            //prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time_b,pos,r);
            //current_time_b+=dt_b;
        }
        fin.close();
        //prescribed_motion.Initialize_Prescribed_Motion();

        //LOG::cout<<"x(1): "<<x(1)<<", pm: "<<prescribed_motion.Prescribed_Frame(Time_At_Frame(0)).t<<std::endl;

        //tube->prescribed_motion.Prescribed_Frame(scaled_time);

        //TV e1=TV::Axis_Vector(2)*(T)1e-3*(T)20;
        //e1(1)=(T)-0.04;e1(3)=(T)-0.04;
        T height=(T)2e-3;
        TV e1=x(1);
        e1.y-=height;
        TV dir=(T)-1*TV::Axis_Vector(2);
        T radius=(T)0.5*(T)16e-3;
        //T height=(T)3e-3;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->use_texture_coordinates=true;
        tube->initialize_texture_coordinates_box=true;
        T thickness=(T)8e-3;

        if(use_1d_tube) tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Volumetric_Rounded_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,thickness);

        tubes.Append(tube);
        //tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;
        tube->rigid_body_type=PAINT_TUBE<TV>::RENDERABLE;
        Add_Tube_Rigid_Body(*tube);

        ////tube motion
        T current_time=(T)0;

        tube->use_prescribed_motion=true;
        T dt_factor=(T)5;
        T dt=(T)0.01*dt_factor;

        LOG::cout<<"original y: "<<tube->object_frame.t.y<<std::endl;

        FRAME<TV> original_frame=tube->object_frame;
        ROTATION<TV> r=original_frame.r;

        for(int i=1;i<=x.m;i++){
            tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame(current_time,x(i),r);
            current_time+=dt;
        }

        ////tube speed
        {
            dt=dt*(T)0.25;
            T epsilon=(T)1e-7;
            //T speed=(T).1;
            //T higher_speed=(T).3;
            T speed=(T).2;
            T higher_speed=(T).27;
            int steps=(int)(current_time/dt);
            for(int i=0;i<=steps;i++){
                T time=(T)i*dt;
                if(i>=steps-4){
                    tube->Add_Source_Speed_Bspline_Time_Value(time,(T)0);
                    continue;
                }
                //if(i%6<3){
                //if(i%6>=3){
                if(i%5>=3){
                    if(i%5==0)tube->Add_Source_Speed_Bspline_Time_Value(time-epsilon,speed);
                    tube->Add_Source_Speed_Bspline_Time_Value(time,higher_speed);
                    if(i%5==1)tube->Add_Source_Speed_Bspline_Time_Value(time+epsilon,speed);
                }
                else{
                    tube->Add_Source_Speed_Bspline_Time_Value(time,speed);
                }
            }
            tube->Initialize_Source_Speed_Bspline();
        }

        //Add_Bread_Slices((T)0.1);
    }break;

    //////////////////////////////////////////////////////////////////////////
    ////tests for toothpaste
    case 666:{
        length_scale=T(1e-3);
        T radius=length_scale*T(8); // No / here in the past checkin
        T height=length_scale*T(2);
        TV e1=TV::Axis_Vector(2)*length_scale*T(10);
        //TV dir=TV(.5,-1,0);
        TV dir=TV(1,0,0);
//        tube=new PAINT_TUBE<TV>(e1,dir,height,radius,length_scale/T(2),.1,.1);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        has_viscosity=true;
        kinematic_viscosity*=T(1e5);
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=T(.2);
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        //remesh=false;
        levelset_region_number=3;
        Add_Canvas();
    }break;
    case 667:{
        length_scale=T(1e-3);
        T radius=length_scale*T(3);
        T height=length_scale*T(2);
        TV e1=TV::Axis_Vector(2)*length_scale*T(10);
        TV dir=TV(1,0,0);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        has_viscosity=true;
        kinematic_viscosity*=T(1e5);
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=T(.2);
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        levelset_region_number=2;
        Add_Canvas();
    }break;
    case 668:{
        length_scale=T(1e-3);
        T radius=length_scale*T(3);
        T height=length_scale*T(2);
        TV e1=TV::Axis_Vector(2)*length_scale*T(10);
        TV dir=TV(1,0,0);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,height,height,radius,length_scale,(T).1,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        has_viscosity=true;
        kinematic_viscosity*=T(1e5);
        use_implicit_viscosity=true;
        has_surface_tension=true;
        surface_tension*=T(.2);
        has_incompressibility=true;
        has_elasticity=false;
        has_plasticity=false;
        remesh=true;
        levelset_region_number=4;
        Add_Canvas();
    }break;
    case 669:{   ////dents on paint, brush moves down and then go up 
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).4*paint_size,paint_size);
        TV box_min=TV((T)0,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.incompressible.use_all_pressure_sample_on_nodes=true;
        use_inactive_particle=true;

        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).95,(T)0),ROTATION<TV>((T).0*(T)half_pi,TV::Axis_Vector(3)));
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=5;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_gravity=true;

        //int fiber_number_row=4;int fiber_number_column=12;
        //int fiber_number_row=1;int fiber_number_column=3;
        //T fiber_width=fiber_number_column*fiber_distance;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,TV(),fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        //n=n;
        //m=m;
        //fiber.object_frame=FRAME<TV>(TV(fiber_length_scale*(T)4.5,fiber_length*(T).85,(T)0),ROTATION<TV>(half_pi/2,TV::Axis_Vector(3))*ROTATION<TV>((T)-1*(T)half_pi,TV::Axis_Vector(2)));
        // TODO remove this!
        //fiber.model=FIBER<TV>::RIGID;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry.tri",surface);
        surface->Rescale(4);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,4);
        std::cout<<fiber_positions<<std::endl;
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;
        T random_scale=(T).5;
        std::cout<<fiber_positions.m<<std::endl;
        for(int i=1;i<=fiber_positions.m;++i){
            T x=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
            T y=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
            fiber_positions(i)+=TV((T)x,(T)y,(T)0);
            fiber_positions(i)=rotate.Rotate(rotate2.Rotate(fiber_positions(i)*fiber_scale))+TV(0,0,0);
            //fiber_positions(i)=rotate.Rotate(fiber_positions(i)*fiber_scale)+TV(0,0,0);
            //fiber_positions(i)=fiber_positions(i)*fiber_scale;
        }
        //fiber.Randomize_Fiber(fiber_positions,1,fiber_positions.m,);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);
        // TODO free surface
        //delete surface;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 670:{ // Toothpaste zigzag
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;T paint_height=length_scale*(T)2;T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)5;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
//        tube->Initialize_Volumetric_Mesh_And_X_And_V_Bumpy_End(initial_mesh,initial_X,initial_V);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
//        tube->rigid_body_height=-0.005;
        Add_Tube_Rigid_Body(*tube);
//        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).1);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).05,(T).0);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
//        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)1);//.3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T)1);//.3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 671:{ // Toothpaste zigzag - HIGHRES!
        T scale=(T)1/T(4);length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*T(4);T paint_height=length_scale*(T)2*T(4);T radius=length_scale*(T)2*T(4);
        TV e1=TV::Axis_Vector(2)*length_scale*(T)20;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);//.2
        tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).5);//.2
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 672:{ // Toothpaste top to bottom vertical
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;
        T paint_height=length_scale*(T)2;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)50;
        TV dir=TV((T)0,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(9),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(10),(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(109),(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(110),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(119),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(120),(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        gravity=TV::Axis_Vector(2)*(T)-9.8;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 673:{ // Toothbrush fibers
    /*
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;T paint_height=length_scale*(T)2;T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
*/
        T hack=2;
        T scale=(T).5;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;T paint_height=length_scale*(T)2*hack;T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10*hack;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
//        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);


        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
//        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).14,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).15,(T).0);

/*
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).049,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).05,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);

*/

        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_fibers_bbox",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*(T)1,TV::Axis_Vector(2));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);

/*
        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)25e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        band_width_min_factor=(T)2*fiber_radius/length_scale;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(100);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,2);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
random_scale=random_scale;
        for(int i=1;i<=fiber_positions.m;++i){
//            T x=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            T y=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            fiber_positions(i)+=TV((T)x,(T)y,(T)0);
            fiber_positions(i)=rotate.Rotate(rotate2.Rotate(fiber_positions(i)*fiber_scale))+TV(-0.15,0.03,0);}
//        Add_Fiber_Rigid_Body(fiber_positions,fiber);
        // TODO free surface
        //delete surface;
        */
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 674:{ // Toothbrush fibers
    /*
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;T paint_height=length_scale*(T)2;T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
*/
        T hack=2;
        T scale=(T).5;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;T paint_height=length_scale*(T)2*hack;T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10*hack;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
//        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
//        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).074,(T).1);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).075,(T).0);

//        tube->Add_Source_Speed_Bspline_Time_Value((T).049,(T).1);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).050,(T).0);

        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(49),(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(50),(T).0);

        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_fibers_bbox",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,0,0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);

/*
        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)25e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        band_width_min_factor=(T)2*fiber_radius/length_scale;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(100);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,2);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
random_scale=random_scale;
        for(int i=1;i<=fiber_positions.m;++i){
//            T x=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            T y=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            fiber_positions(i)+=TV((T)x,(T)y,(T)0);
            fiber_positions(i)=rotate.Rotate(rotate2.Rotate(fiber_positions(i)*fiber_scale))+TV(-0.15,0.03,0);}
//        Add_Fiber_Rigid_Body(fiber_positions,fiber);
        // TODO free surface
        //delete surface;
        */
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 675:{ // Toothbrush fibers - HIGHRES
    /*
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;T paint_height=length_scale*(T)2;T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
*/
        T hack=8;
        T scale=(T).5/T(4);length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;T paint_height=length_scale*(T)2*hack;T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10*hack;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
//        tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        tube->Add_Source_Speed_Bspline_Time_Value((T).074,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).075,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_fibers_bbox",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);

/*
        ////brush
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)25e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        band_width_min_factor=(T)2*fiber_radius/length_scale;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(100);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,2);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
random_scale=random_scale;
        for(int i=1;i<=fiber_positions.m;++i){
//            T x=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            T y=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
//            fiber_positions(i)+=TV((T)x,(T)y,(T)0);
            fiber_positions(i)=rotate.Rotate(rotate2.Rotate(fiber_positions(i)*fiber_scale))+TV(-0.15,0.03,0);}
//        Add_Fiber_Rigid_Body(fiber_positions,fiber);
        // TODO free surface
        //delete surface;
        */
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 676:{ // paint trace with better brush + modified model
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T)/*.25*/.1;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).3*paint_size,paint_size);
        TV box_min=TV((T)12e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,box_min,length_scale,2);
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.incompressible.use_all_pressure_sample_on_nodes=false;

        ////brush
        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        ////Predefined motion
        //fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
        //std::string dir="../model/Motion_Data/brush_coarse_curve";//T t0=(T).0;T t1=(T).6;T motion_t0=(T).2;
        //fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T)0);
        fiber_length_scale=(T)4e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)25e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        band_width_min_factor=(T)2*fiber_radius/length_scale;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry.tri",surface);
        surface->Rescale(1.25);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,3);//4); // TODO change this to this if fibers look too sparse
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
//random_scale=random_scale;
//STREAM_TYPE st(1.f);
//FILE_UTILITIES::Write_To_File(st,"/data/mjlgg/PhysBAM/trunk/tri2obj/surface.tri",surface);
        for(int i=1;i<=fiber_positions.m;++i){
            T x=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
            T y=static_cast<T>(rand())/static_cast<T>(RAND_MAX)*random_scale-random_scale*T(.5);
            fiber_positions(i)+=TV((T)x,(T)y,(T)0);
            fiber_positions(i)=rotate.Rotate(rotate2.Rotate(fiber_positions(i)*fiber_scale))+TV((T)0.000,(T)0.004,(T)0.0);}
        Add_Fiber_Rigid_Body(fiber_positions,fiber);
        // TODO free surface
        //delete surface;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 677:{ // Toothpaste top to bottom vertical - HIGHRES
        T hack=4;
        T scale=(T)1/T(4);length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;
        T paint_height=length_scale*(T)2*hack;
        T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)50*hack;
        TV dir=TV((T)0,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(9),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(10),(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        gravity=TV::Axis_Vector(2)*T(-98);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 678:{ // Toothbrush fibers - with heightmap phi
        T hack=2;
        T scale=(T).5;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;T paint_height=length_scale*(T)2*hack;T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10*hack;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).074,(T).1);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).075,(T).0);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).049,(T).1);
//        tube->Add_Source_Speed_Bspline_Time_Value((T).050,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(49),(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(50),(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=true;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);
/*
        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("tbrush_heightmap",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1*rot3*rot2;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);
        */
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 679:{ // Toothbrush fibers - HIGHRES - with heightmap
        T hack=8;
        T scale=(T).5/T(4);length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25*hack;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10*hack;T paint_height=length_scale*(T)2*hack;T radius=length_scale*(T)2*hack;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)10*hack;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        tube->Add_Source_Speed_Bspline_Time_Value((T).074,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).075,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("tbrush_heightmap",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1*rot3*rot2;

        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 680:{ // GHOST FIBER!
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.Use_Staggered_DoF(true);

        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_fibers_bbox",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rott=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rott;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);

        ////brush
        fiber_length_scale=(T)14.9e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)15e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        band_width_min_factor=(T)2*fiber_radius/length_scale;
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(30);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,3);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
        random_scale=random_scale;
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        for(int i=1;i<=fiber_positions.m;++i){
            fiber_positions(i)=rot1.Rotate(rot3.Rotate(rot2.Rotate(fiber_positions(i)*fiber_scale)));} // TODO adjust x axis in raytracer... by 0.04?
        // TODO free surface
        //delete surface;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
        // TODO hacky code..
        fiber_radius=(T)4e-5;
    }break;
    case 681:{ // Toothpaste top to bottom vertical - VERTICAL MOTION
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;
        ////toothpaste
        T tube_height=length_scale*(T)10;
        T paint_height=length_scale*(T)2;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)50;
        TV dir=TV((T)0,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(9),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(10),(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(199),(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(200),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(209),(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(210),(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        gravity=TV::Axis_Vector(2)*(T)-9.8;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.preserve_thickness_in_remeshing=true;
        example.viscosity_force.Use_Staggered_DoF(true);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // These tests are valid after bo's recent checkin which broke most of above tests since 666
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    case 682:{ // Toothbrush fibers - with heightmap phi (revised)
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).01;
        T paint_height=s0*T(1);T radius=s0*(T)2;
        T tube_height=s0*T(.1);
        TV e1=TV::Axis_Vector(2)*s0*T(10);TV dir=TV((T)1,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,T(0),T(.1));
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(49),(T).2);
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(50),(T).0);
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(74),(T).2);//.3
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(75),(T).0);
        
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(99),(T).2);//.3
//        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(100),(T).0);

        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(149),(T).2);//.3
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(150),(T).0);

        tube->Initialize_Source_Speed_Bspline();
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_break_thin_triangles=false;
        has_adhesion=true;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).001;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        use_non_slip_on_solids=true;
        levelset_region_number=3;
        //if(paint_tube_test!=2&&paint_tube_test!=13)use_correcting_particle_position_on_solids=false;
        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("toothbrush_fibers_bbox_soft",(T).03,(T)0.,true,true,&double_stream_type);
//        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("tbrush_heightmap",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1*rot3*rot2;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 683:{ // Toothbrush fibers - with heightmap phi (using old heightmap)
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).01;
        T paint_height=s0*T(1);T radius=s0*(T)2;
        T tube_height=s0*T(.1);
        TV e1=TV::Axis_Vector(2)*s0*T(10);TV dir=TV((T)1,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,T(0),T(.1));
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(49),(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T)Time_At_Frame(50),(T).0);
        tube->Initialize_Source_Speed_Bspline();
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_break_thin_triangles=false;
        has_adhesion=true;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        use_non_slip_on_solids=true;
        //if(paint_tube_test!=2&&paint_tube_test!=13)use_correcting_particle_position_on_solids=false;
        STREAM_TYPE double_stream_type(1.0);
        ////handle
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Rigid_Body("tbrush_heightmap",(T).03,(T)0.,true,true,&double_stream_type);
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        rigid_body->X()=fiber.object_frame.t+TV((T)-0.04,(T)0,(T)0);
        rigid_body->Rotation()=fiber.object_frame.r*rot1*rot3*rot2;
        rigid_body->name="toothbrush_fibers";
        rigid_bodies.Append(rigid_body);
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 684:{ // GHOST FIBER!
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_break_thin_triangles=false;
        has_adhesion=true;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        use_non_slip_on_solids=true;
        //if(paint_tube_test!=2&&paint_tube_test!=13)use_correcting_particle_position_on_solids=false;
        ////brush
        fiber_length_scale=(T)14.9e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)15e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        band_width_min_factor=12e-5;//(T)2*fiber_radius/length_scale;
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(30);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,3);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
        random_scale=random_scale;
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi,TV::Axis_Vector(3));
        for(int i=1;i<=fiber_positions.m;++i){
            fiber_positions(i)=rot1.Rotate(rot3.Rotate(rot2.Rotate(fiber_positions(i)*fiber_scale)));} // TODO adjust x axis in raytracer... by 0.04?
        // TODO free surface
        //delete surface;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
        // TODO hacky code..
        fiber_radius=(T)12e-5;
    }break;
    case 685:{ // Toothpaste zigzag
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).01;
        T paint_height=s0*T(1);T radius=s0*(T)2;
        T tube_height=s0*T(.1);
        TV e1=TV::Axis_Vector(2)*s0*T(5);TV dir=TV((T)1,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,T(0),T(.1));
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_break_thin_triangles=false;
        has_adhesion=true;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        use_non_slip_on_solids=true;
        //if(paint_tube_test!=2&&paint_tube_test!=13)use_correcting_particle_position_on_solids=false;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 686:{ // Brushing
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_break_thin_triangles=false;
        has_adhesion=true;
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        use_non_slip_on_solids=true;
        ////brush
        fiber_length_scale=(T)14.9e-3;
        fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)15e-3;//45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)4e-3,(T)30e-3,(T)0),ROTATION<TV>((T)1*(T)pi,TV::Axis_Vector(1)));
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;
        band_width_min_factor=12e-5;//(T)2*fiber_radius/length_scale;
        fiber.fiber_bending_perturb_frac=TV((T).0,(T)0,(T)0);
        TRIANGULATED_SURFACE<T> *surface;
        FILE_UTILITIES::Create_From_File<T>("guide_geometry_sparse.tri",surface);
        surface->Rescale(30);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Brush_Particles(fiber_positions,*surface,3);
        ROTATION<TV> rotate((T)-half_pi,TV((T)0,(T)1,(T)0));
        ROTATION<TV> rotate2((T)-half_pi,TV((T)1,(T)0,(T)0));
        T fiber_scale=(T).001;T random_scale=(T).25;
        random_scale=random_scale;
        ROTATION<TV> rot1=ROTATION<TV>((T)half_pi*1,TV::Axis_Vector(2));
        ROTATION<TV> rot2=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        ROTATION<TV> rot3=ROTATION<TV>((T)pi+two_pi,TV::Axis_Vector(3));
        for(int i=1;i<=fiber_positions.m;++i){
            fiber_positions(i)=rot1.Rotate(rot3.Rotate(rot2.Rotate(fiber_positions(i)*fiber_scale)));} // TODO adjust x axis in raytracer... by 0.04?
        // TODO free surface
        //delete surface;
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
        // TODO hacky code..
        fiber_radius=(T)12e-5;
    }break;
#endif
    default:{   ////skip final tests
        if(test!=TEST_MAYO&&test!=TEST_TOOTHPASTE&&test!=TEST_PAINT_TUBE&&test!=TEST_PAINT_TUBE_MULTIPLE_COLOR&&
            test!=TEST_PAINT_TRACE&&test!=TEST_PAINT_TRACE2&&test!=TEST_PAINT_DIP&&test!=TEST_PAINT_SPRAY&&test!=TEST_PAINT_SPRAY2&&test!=TEST_PAINT_SPRAY3&&test!=TEST_PAINT_SPRAY4&&test!=TEST_PAINT_SPLASH&&
            test!=TEST_MATERIALS&&test!=TEST_MATERIALS_2&&test!=TEST_EXPOSITION_1&&test!=TEST_EXPOSITION_2&&
            test!=TEST_CHEESE&&test!=TEST_PREPROCESS_CHEESE_MESH&&test!=TEST_CHEESE_ON_BREAD&&test!=TEST_CHEESE_POURING&&test!=TEST_CHEESE_HEATING&&
            test!=82&&test!=83&&test!=130&&test!=228&&test!=229&&test!=230){PHYSBAM_FATAL_ERROR("Invalid Test Number.");}
    }break;
    }
    Initialize_Test_Parameters2();

    fiber_initial_frame=fiber.object_frame;  
    kinematic_viscosity*=viscosity_factor;kinematic_viscosity_min*=viscosity_min_factor;if(viscosity_factor==(T)0)has_viscosity=false;
    elastic_modulus*=elastic_modulus_factor;if(elastic_modulus_factor==(T)0)has_elasticity=false;
    surface_tension*=surface_tension_factor;seg_surface_tension*=surface_tension_factor;tri_surface_tension*=surface_tension_factor;if(surface_tension_factor==(T)0)has_surface_tension=false;
    gravity*=gravity_factor;if(gravity_factor==(T)0)has_gravity=false;
    //adaptive_length_scale_min*=length_scale_min_factor;if(length_scale_min_factor==(T)0)adaptive_length_scale_min=length_scale;
    adaptive_length_scale_min*=adaptive_length_scale_min_factor;
    adaptive_length_scale_min2*=adaptive_length_scale_min_factor2;

    ////initialize fiber structures
    fiber.Initialize_Geometry(fiber_positions,fiber_length_scale,fiber_radius,fiber_length,guided_fiber_radius,guided_fiber_distance);

    ////initialize prescribed motion
    Initialize_Prescribed_Object_Motion();

    ////set length scale
    coupled_evolution.length_scale=length_scale;
    coupled_evolution.Set_Triangle_Length_Scale(/*adaptive_length_scale_min*/adaptive_length_scale_min2,length_scale,band_width_min,band_width_max);
    
    if(test==292)coupled_evolution.Set_Tetrahedron_Length_Scale(adaptive_length_scale_min,length_scale,band_width_min,length_scale*(T)2);
    else coupled_evolution.Set_Tetrahedron_Length_Scale(length_scale*(T).5,length_scale,band_width_min,length_scale*(T)2);

    example.viscosity_force.length_scale_min=adaptive_length_scale_min;
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Test_Parameters2()
{
    switch(test){
    case 201:{

    }break;
    case 202:{
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;

        ////paint
        T tube_height=length_scale*(T)10;
        T paint_height=length_scale*(T)2;
        T radius=length_scale*(T)2;
        //TV e1=TV::Axis_Vector(2)*length_scale*(T)15;TV dir=TV((T).6,(T)-1,(T)0).Normalized();   ////high
        TV e1=TV::Axis_Vector(2)*length_scale*(T)5;
        TV dir=TV((T)1,(T)0,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);
        
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);

        tube->Initialize_Source_Speed_Bspline();
        tubes.Append(tube);
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        //example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.Use_Staggered_DoF(true);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 205:{  ////TEST_CHEESE
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=(T).5*s0;max_thickness_factor=(T)2;min_thickness_factor=(T).2*min_tri_thickness_factor;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).05*min_seg_thickness_factor;
        adaptive_length_scale_min=length_scale*(T).3;
        //example.mesher.clamp_segment_thickness_min=thickness_scale*(T).5;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)4;

        ////paint
        T paint_size=(T)50e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,paint_box_size_zx,length_scale,2);

        //CODIMENSIONAL_MESH_GENERATOR<TV>::Circle(initial_mesh,initial_X,TV(),TV::Axis_Vector(2),paint_size*(T).5,length_scale);TV center=TV();T r=paint_box_size_zx.x;
        //Initialize_Pizza(initial_X,initial_thickness,initial_texture_coordinates,center,r,thickness_scale*max_thickness_factor,thickness_scale*min_thickness_factor);

        //T piz_thickness_min=length_scale;T piz_thickness_max=length_scale*(T)3;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume(initial_mesh,initial_X,TV(),paint_size*(T).5,length_scale,piz_thickness_min,piz_thickness_max);

        //////pizza codimensional mesh
        //T piz_thickness_min=length_scale*(T).2;piz_thickness_min*=piz_thickness_min_factor;
        //T piz_thickness_max=length_scale*(T)1;piz_thickness_max*=piz_thickness_max_factor;
        //T size_x=s0*(T)40;T size_z=s0*(T)10;TV box2_min=TV((T)0,(T)0,(T)-.5*size_z);TV box2_size=TV(size_x,size_x*(T)2,size_z);TV box2_max=box2_min+box2_size;
        ////T size_x=s0*(T)50;T size_z=s0*(T)60;TV box2_min=TV((T)0,(T)0,(T)-.5*size_z);TV box2_size=TV(size_x,size_x*(T)2,size_z);TV box2_max=box2_min+box2_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume(initial_mesh,initial_X,box2_min,box2_max,length_scale,piz_thickness_min,piz_thickness_max);

        ////TV box3_min=TV(s0*(T)20,(T)0,(T)0);TV box3_size=TV::Constant_Vector(s0*(T)5);TV box3_max=box3_min+box3_size;
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box3_min,box3_max,length_scale);

        T piz_radius=s0*(T)40;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,piz_mesh_filename,piz_init_frame,1,0);
        Initialize_Part_Pizza_Mesh();
        Initialize_Pizza_Texture_Coordinates(initial_X,TV(),piz_radius,initial_texture_coordinates);

        //TV trans((T)0,(T)0,(T)0);
        //for(int i=1;i<=initial_X.m;i++){initial_X(i)+=trans;}
        //ROTATION<TV> rot((T)-pi/(T)8,TV::Axis_Vector(2));
        //for(int i=1;i<=initial_X.m;i++){initial_X(i)=rot.Rotate(initial_X(i));}

        ////pizza rigid body
        for(int i=1;i<=2;i++){
            TV rigid_body_size=TV((T).12,(T).004,(T).12);
            RIGID_BODY<TV>* kinematic_rigid_body=0;STREAM_TYPE float_stream_type(1.0f);
            if(i==1)kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("pizza_slice_flt",(T)2.,(T)0.,true,true,&float_stream_type);
            else kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("pizza_slice_missing_flt",(T)2.,(T)0.,true,true,&float_stream_type);

            T y_offset=length_scale*(T).1;
            kinematic_rigid_body->X()=TV()+y_offset;kinematic_rigid_body->X().y-=rigid_body_size.y*(T).5;
            kinematic_rigid_body->Rotation()=ROTATION<TV>((T).875*(T)pi,TV::Axis_Vector(2));
            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            //solid_non_slip_hashtable.Set(rigid_bodies.m,PAIR<bool,T>(false,(T)1));
            kinematic_rigid_bodies.Append(kinematic_rigid_body);

            ////ground
            if(i==2){
                TV ground_size=TV(paint_size*(T)10,paint_size*(T).2,paint_size*(T)10);
                RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
                ground_rigid_body->X().y=-ground_size.y*(T).5-rigid_body_size.y*(T).5;
                rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";
                solid_non_slip_hashtable.Set(rigid_bodies.m,PAIR<bool,T>(true,(T)0));
            }
        }

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_friction=true;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_levelset=false;
        rigid_body_speed_factor=(T).95;
        use_correcting_rigid_normal=true;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_variable_viscosity_for_codim=true;
        example.viscosity_force.use_particle_thickness=false;
        //example.viscosity_force.segment_viscosity_factor=(T).1;

        ////elasticity
        if(piz_test==2){
            ////elasticity
            has_elasticity=true;
            has_plasticity=true;
            //use_implicit_elasticity=true;
            example.elasticity_force.Use_Staggered_DoF(true);
            example.elasticity_force.Use_Constant_Extrapolation(false);
            example.elasticity_force.use_particle_thickness=false;
        }
        
        ////rim
        Use_Rim(true);

        ////meshing
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_enforce_thin_tri_break_constrained=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_half_cylindrical_rim=true;
        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_based_on_test=false;
        use_adaptive_length_scale_on_rim=false;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;
        use_inactive_particle=false;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).2);TV domain_center=TV();domain_center.x=(T).015;
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 209:{  ////preprocess pizza mesh
        T s0=(T)1e-3;length_scale=scale*s0;thickness_scale=(T).5*s0;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).2*min_tri_thickness_factor;
        max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).05*min_seg_thickness_factor;
        adaptive_length_scale_min=length_scale*(T).3;

        ////paint
        T paint_size=(T)50e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,paint_box_size_zx,length_scale,2);

        //CODIMENSIONAL_MESH_GENERATOR<TV>::Circle(initial_mesh,initial_X,TV(),TV::Axis_Vector(2),paint_size*(T).5,length_scale);TV center=TV();T r=paint_box_size_zx.x;
        //Initialize_Pizza(initial_X,initial_thickness,initial_texture_coordinates,center,r,thickness_scale*max_thickness_factor,thickness_scale*min_thickness_factor);

        //T piz_thickness_min=length_scale;T piz_thickness_max=length_scale*(T)3;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume(initial_mesh,initial_X,TV(),paint_size*(T).5,length_scale,piz_thickness_min,piz_thickness_max);

        T piz_thickness_min=s0*piz_thickness_min_factor;T piz_thickness_max=s0*piz_thickness_max_factor;
        T size_x=s0*(T)40;T size_z=s0*(T)10;TV box2_min=TV((T)0,(T)0,(T)-.5*size_z);TV box2_size=TV(size_x,size_x*(T)2,size_z);TV box2_max=box2_min+box2_size;
        //T size_x=s0*(T)50;T size_z=s0*(T)60;TV box2_min=TV((T)0,(T)0,(T)-.5*size_z);TV box2_size=TV(size_x,size_x*(T)2,size_z);TV box2_max=box2_min+box2_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume(initial_mesh,initial_X,box2_min,box2_max,length_scale,piz_thickness_min,piz_thickness_max);
        T radius=size_x;CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Cylinder_Volume(initial_mesh,initial_X,TV(),radius,length_scale,piz_thickness_min,piz_thickness_max,piz_mesh_noise_scale,piz_mesh_noise_offset);
        //TV box3_min=TV(s0*(T)20,(T)0,(T)0);TV box3_size=TV::Constant_Vector(s0*(T)5);TV box3_max=box3_min+box3_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box3_min,box3_max,length_scale);

        //TV trans((T)0,(T)0,(T)0);
        //for(int i=1;i<=initial_X.m;i++){initial_X(i)+=trans;}
        //ROTATION<TV> rot((T)-pi/(T)8,TV::Axis_Vector(2));
        //for(int i=1;i<=initial_X.m;i++){initial_X(i)=rot.Rotate(initial_X(i));}

        TV ground_size=TV(paint_size*(T)10,paint_size*(T).2,paint_size*(T)10);
        RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
        ground_rigid_body->X().y=-ground_size.y*(T).5;
        rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=false;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_friction=true;
        use_damped_friction=false;
        use_correcting_particle_position_on_solids=true;
        use_non_slip_on_solids=false;
        use_levelset=false;
        rigid_body_speed_factor=(T).95;
        use_correcting_rigid_normal=true;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.use_variable_viscosity_for_codim=true;
        //example.viscosity_force.segment_viscosity_factor=(T).1;

        //////elasticity
        //has_elasticity=true;
        //has_plasticity=true;
        ////use_implicit_elasticity=true;
        //example.elasticity_force.Use_Staggered_DoF(true);
        //example.elasticity_force.Use_Constant_Extrapolation(false);
        //example.elasticity_force.use_particle_thickness=false;

        ////rim
        Use_Rim(true);

        ////meshing
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.mesher.use_half_cylindrical_rim=true;
        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_based_on_test=false;
        use_adaptive_length_scale_on_rim=false;
        adaptive_length_scale_min=length_scale*(T).3;
        use_inactive_particle=false;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 206:{  ////TEST_CHEESE_ON_BREAD
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=s0*(T).4;max_thickness_factor=(T)4;min_thickness_factor=(T).1;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).05;
        adaptive_length_scale_min=length_scale*(T).3;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)3;

        ////paint
        //T paint_size=(T)20e-3;TV paint_box_size=TV((T).5*paint_size,(T).2*paint_size,(T)2*paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        //TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;

        T paint_size=(T)20e-3;TV paint_box_size=TV((T).5*paint_size,(T).2*paint_size,(T)2*paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;

        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //T piz_thickness_min=length_scale*(T).2;piz_thickness_min*=piz_thickness_min_factor;
        //T piz_thickness_max=length_scale*(T)1;piz_thickness_max*=piz_thickness_max_factor;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Noise_Box_Volume(initial_mesh,initial_X,box_min,box_max,length_scale,piz_thickness_min,piz_thickness_max);

        ////cheese
        {
            T cheese_size=(T)20*s0;TV cheese_box_size=TV(cheese_size,(T).05*cheese_size,cheese_size);

            //int cheese_n=1;
            //ARRAY<CODIMENSIONAL_MESH> cheese_meshes;cheese_meshes.Resize(cheese_n);
            //ARRAY<ARRAY<TV> > cheese_Xs;cheese_Xs.Resize(cheese_n);
            
            //{int idx=1;CODIMENSIONAL_MESH& cheese_mesh=cheese_meshes(idx);ARRAY<TV>& cheese_X=cheese_Xs(idx);
            //TV cheese_box_min=TV((T)cheese_box_size.x*(T)-.5,(T)0,cheese_box_size.z*(T)-.5);
            //TV cheese_box_max=cheese_box_min+cheese_box_size;
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(cheese_mesh,cheese_X,cheese_box_min,cheese_box_max,length_scale);}
            
            //{int idx=2;CODIMENSIONAL_MESH& cheese_mesh=cheese_meshes(idx);ARRAY<TV>& cheese_X=cheese_Xs(idx);
            //TV cheese_box_min=TV((T)cheese_box_size.x*(T)-.5,cheese_box_size.y*(T)1.15,cheese_box_size.z*(T)-.5);
            //TV cheese_box_max=cheese_box_min+cheese_box_size;
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(cheese_mesh,cheese_X,cheese_box_min,cheese_box_max,length_scale);
            //ROTATION<TV> rot((T)-pi/(T)8,TV::Axis_Vector(2));for(int i=1;i<=cheese_X.m;i++){cheese_X(i)=rot.Rotate(cheese_X(i));}}

            //Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,cheese_meshes,initial_X,cheese_Xs,0,0,0,0,0,0);

            ////multiple boxes
            ARRAY<BOX<TV> > boxes;ARRAY<ROTATION<TV> > rotations;
            {            
                TV cheese_box_min=TV((T)cheese_box_size.x*(T)-.5,(T)0,cheese_box_size.z*(T)-.5);
                TV cheese_box_max=cheese_box_min+cheese_box_size;
                boxes.Append(BOX<TV>(cheese_box_min,cheese_box_max));
                rotations.Append(ROTATION<TV>((T)0,TV::Axis_Vector(2)));
            }
            {
                TV cheese_box_min=TV((T)cheese_box_size.x*(T)-.5,cheese_box_size.y*(T)1.15,cheese_box_size.z*(T)-.5);
                TV cheese_box_max=cheese_box_min+cheese_box_size;
                boxes.Append(BOX<TV>(cheese_box_min,cheese_box_max));
                rotations.Append(ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(2)));
            }
            CODIMENSIONAL_MESH_GENERATOR<TV>::Multiple_Boxes(initial_mesh,initial_X,boxes,rotations,length_scale);
        }

        ////kinematic solid
        for(int i=1;i<=1;i++){
            //TV rigid_body_size=TV(paint_size*(T)1,paint_size*(T).1,paint_size*(T)2.2);
            //TV rigid_body_size=TV(paint_size*(T)2,paint_size*(T).1,paint_size*(T)2.2);
            TV rigid_body_size=TV(paint_size*(T)2,paint_size*(T).1,paint_size*(T)2.2);

            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();T offset=paint_box_size.x*(T).0;
            //if(i==1)kinematic_rigid_body->X().x=rigid_body_size.x*(T).5+offset*(T).5;
            //else kinematic_rigid_body->X().x=-rigid_body_size.x*(T).5-offset*(T).5;
            kinematic_rigid_body->X().y=-rigid_body_size.y*(T).5;

            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);

            ////ground
            if(i==2){
                TV ground_size=TV(paint_size*(T)20,paint_size*(T).2,paint_size*(T)20);
                RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
                ground_rigid_body->X().y=-ground_size.y*(T).5-rigid_body_size.y;
                rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";
            }
        }

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).005;
        has_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;
        example.viscosity_force.use_temperature_dependent_variable_viscosity=true;
        //example.viscosity_force.segment_viscosity_factor=(T).1;

        ////elasticity
        has_elasticity=true;
        has_plasticity=true;
        //use_implicit_elasticity=true;
        example.elasticity_force.Use_Staggered_DoF(true);
        example.elasticity_force.Use_Constant_Extrapolation(false);
        example.elasticity_force.use_particle_thickness=false;

        ////rim
        Use_Rim(true);

        ////meshing
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_based_on_test=true;
        use_adaptive_length_scale_on_rim=false;
        adaptive_length_scale_min=length_scale*(T).3;
        use_inactive_particle=true;

        ////heat transfer
        use_heat_transfer=true;
        environment_temperature=(T)0;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 130:{   ////test elasticity 1, falling
        T scale=(T)1;length_scale=scale*(T)1e-3;thickness_scale=length_scale*(T).75;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        max_rim_thickness_factor=(T)1.5;min_rim_thickness_factor=(T).05;

        ////paint
        T tube_height=length_scale*(T)5;
        T paint_height=length_scale*(T)10;
        T radius=length_scale*(T)2;
        TV e1=TV::Axis_Vector(2)*length_scale*(T)50;
        TV dir=TV::Axis_Vector(2)*(T)-1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).0,(T).01);
        //tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V,/*no hemisphere*/(T)0);
        if(use_1d_tube) tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        Add_Tube_Rigid_Body(*tube);tubes.Append(tube);

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=false;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        has_elasticity=true;
        has_plasticity=true;
        //use_implicit_elasticity=true;
        use_levelset=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.elasticity_force.Use_Staggered_DoF(true);  ////TODO: fix the staggered dof case for elasticity
        example.elasticity_force.Use_Constant_Extrapolation(false);
        example.elasticity_force.use_particle_thickness=false;

        remesh=true;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 208:{  ////TEST_CHEESE_POURING
        T s0=(T)1e-3;length_scale=scale*s0;thickness_scale=(T).5*s0;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).05;
        max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).05;
        adaptive_length_scale_min=length_scale*(T).3;

        ////pizza
        T pizza_radius=30*s0;T pizza_thickness=s0;TV rigid_body_size=TV(pizza_radius*(T)2,pizza_thickness*(T)2,pizza_radius*(T)2);
        RIGID_BODY<TV>* kinematic_rigid_body=0;STREAM_TYPE float_stream_type(1.0f);
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("pizza_flt",(T)1.,(T)0.,true,true,&float_stream_type);
        T y_offset=length_scale*(T).1;
        kinematic_rigid_body->X()=TV()+y_offset;kinematic_rigid_body->X().y-=rigid_body_size.y*(T).5;
        kinematic_rigid_body->Rotation()=ROTATION<TV>((T).875*(T)pi,TV::Axis_Vector(2));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////cheese source
        T tube_height=s0*(T)2;T source_height=s0*(T)5;T source_radius=s0*(T)8;
        TV e1=TV(s0*(T)-25,s0*(T)20,(T)0);TV source_dir=TV::Axis_Vector(2)*(T)-1;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,source_dir,source_height,tube_height,source_radius,length_scale,(T)0,(T)0);tubes.Append(tube);
        //tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);

        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T)1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T)1);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T)1);
        tube->Initialize_Source_Speed_Bspline();

        ////cheese
        T paint_size=(T)50e-3;TV paint_box_size=TV(paint_size,paint_size,paint_size);TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        //TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        Add_Cheese_Filaments_Helper(initial_X,initial_mesh,0,&initial_thickness,0);

        ////ground
        TV ground_size=TV(paint_size*(T)10,paint_size*(T).2,paint_size*(T)10);
        RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
        ground_rigid_body->X().y=-ground_size.y*(T).5-rigid_body_size.y;
        rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_friction=true;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_levelset=false;
        rigid_body_speed_factor=(T).95;
        use_correcting_rigid_normal=true;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_variable_viscosity_for_codim=true;
        example.viscosity_force.use_particle_thickness=false;
        ////elasticity
        has_elasticity=true;
        has_plasticity=true;
        use_implicit_elasticity=true;

        ////rim
        Use_Rim(true);

        ////meshing
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_mesh_edge_connect=true;
        ////adaptivity
        use_adaptive_length_scale_on_codim=false;
        use_adaptive_length_scale_on_rim=false;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 207:{  ////TEST_CHEESE_HEATING
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        max_rim_thickness_factor=(T)2;min_rim_thickness_factor=(T).05;
        adaptive_length_scale_min=length_scale*(T).3;

        ////paint
        T paint_size=(T)40e-3;TV paint_box_size=TV((T).2*paint_size,(T).5*paint_size,(T).2*paint_size);//TV2 paint_box_size_zx=TV2(paint_box_size.z,paint_box_size.x);
        TV box_min=TV((T)-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,paint_box_size_zx,length_scale,2);

        ////kinematic solid
        for(int i=1;i<=1;i++){
            TV rigid_body_size=TV((T)100e-3,(T)10e-3,(T)100e-3);
            RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
            kinematic_rigid_body->X()=TV();//T offset=paint_box_size.x*(T).0;
            //if(i==1)kinematic_rigid_body->X().x=rigid_body_size.x*(T).5+offset*(T).5;
            //else kinematic_rigid_body->X().x=-rigid_body_size.x*(T).5-offset*(T).5;
            kinematic_rigid_body->X().y=-rigid_body_size.y*(T).5;

            rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
            kinematic_rigid_bodies.Append(kinematic_rigid_body);

            ////ground
            TV ground_size=TV(paint_size*(T)20,paint_size*(T).2,paint_size*(T)20);
            RIGID_BODY<TV>* ground_rigid_body=&rigid_standard_tests.Add_Analytic_Box(ground_size);
            ground_rigid_body->X().y=-ground_size.y*(T).5-rigid_body_size.y;
            rigid_bodies.Append(ground_rigid_body);ground_rigid_body->name="ground";
        }

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)1;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).005;
        has_friction=false;
        use_levelset=false;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.use_temperature_dependent_variable_viscosity=true;

        ////elasticity
        has_elasticity=true;
        has_plasticity=true;
        use_implicit_elasticity=true;

        ////rim
        Use_Rim(true);

        ////heat transfer
        use_heat_transfer=true;
        environment_temperature=(T)0;

        ////meshing
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_half_cylindrical_rim=true;
        use_adaptive_length_scale_on_codim=false;
        use_adaptive_length_scale_on_rim=true;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 210:{  ////TEST_PAINT_TUBE
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).1;
        max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).05;

        ////paint
        T paint_height=s0;T radius=s0*(T)2;
        //paint_tube_test=12;
        
        switch(paint_tube_test){
        case 1001:{     ////S
            T tube_height=s0*(T)1;
            TV dir=TV((T)0,(T)-1,(T)-.5).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/s_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/s_source.txt");
            
            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1002:{     ////I
            T tube_height=s0*(T)1;
            TV e1=TV((T)0,s0*(T)6,s0*(T)-8);TV dir=TV((T)0,(T)-1,(T)-1).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);

            if(use_overlapping_tube){   ////read "S"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                BOX<TV> box(TV::Constant_Vector((T)-.036),TV::Constant_Vector((T).023));
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_s",300,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_s",360,1,&box);
                TV trans_X=TV((T)-.009,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).101,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();

            if(use_sim_box){
                //sim_particle_boxes.Append(BOX<TV>(TV((T)-1,(T)-1,(T)-1),TV((T)1,(T)1,(T).04)));
                sim_particle_boxes.Append(BOX<TV>(TV((T)-5,(T)-5,(T).04),TV((T)5,(T)5,(T)5)));
            }

            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            use_inactive_particle=true;
        }break;
        case 1003:{     ////G
            T tube_height=s0*(T)1;
            TV dir=TV((T)0,(T)-1,(T)-.2).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/g_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/g_source.txt");

            if(use_overlapping_tube){   ////read "SI"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                BOX<TV> box(TV::Constant_Vector((T)-.05),TV::Constant_Vector((T).05));//box.min_corner.x=(T)-.045;box.max_corner.x=(T).05;
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_si",250,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_si",370,1,&box);
                TV trans_X=TV((T)-.011,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            if(use_sim_box){sim_particle_boxes.Append(BOX<TV>(TV((T)-5,(T)-5,(T)-5),TV((T).05,(T)5,(T)5)));}

            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            //use_inactive_particle=true;
        }break;
        case 1004:{     ////G2
            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T).2).Normalized(); 
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/g2_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/g2_source.txt");

            //TV box3_min=TV(s0*(T)5,(T)0,s0*(T)-8);TV box3_size=TV::Constant_Vector(s0*(T)5);TV box3_max=box3_min+box3_size;
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box3_min,box3_max,length_scale);
            //initial_V.Resize(initial_X.m);initial_V.Fill(TV::Axis_Vector(2)*(T)-5);

            if(use_overlapping_tube){       ////read "SIG" or "G"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                //BOX<TV> box(TV::Constant_Vector((T)-.05),TV::Constant_Vector((T).023));
                BOX<TV> box(TV::Constant_Vector((T)-.02),TV::Constant_Vector((T).03));  ////for G
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_sig",400,1,&box);
                //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_sig",400,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high5_g",480/*180*/,1,&box);     ////for G
                
                TV trans_X=TV((T)-.014,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            if(use_sim_box){sim_particle_boxes.Append(BOX<TV>(TV((T)-5,(T)-5,(T)-5),TV((T).02,(T)5,(T)5)));}

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1005:{     ////R
            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T)-.4).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/r_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/r_source.txt");

            if(use_sim_box){sim_particle_boxes.Append(BOX<TV>(TV((T)-5,(T)-5,(T)-5),TV((T).04,(T)5,(T)5)));}

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1006:{     ///A
            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T)-.4).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/a_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/a_source.txt");

            if(use_overlapping_tube){   ////read "R"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                BOX<TV> box(TV((T)-.005,(T)-.02,(T)-.03),TV((T).025,(T).02,(T).02));
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_r",360,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_r",360,1,&box);
                TV trans_X=TV((T)-.025,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1007:{     ////P
            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T)-.4).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/p_motion.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/p_source.txt");

            if(use_overlapping_tube){   ////read "RA" or "A"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                //BOX<TV> box(TV((T)-.03,(T)-.02,(T)-.03),TV((T).025,(T).02,(T).02));
                //if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_ra",430,1,&box);
                //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_ra",430,1,&box);

                ////read "A"
                BOX<TV> box(TV((T)-.1,(T)-.1,(T)-.1),TV((T).03,(T).1,(T).1));
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_a",600,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_a",600,1,&box);

                TV trans_X=TV((T)-.007,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            if(use_sim_box){sim_particle_boxes.Append(BOX<TV>(TV((T)-5,(T)-5,(T)-5),TV((T)-.0255,(T)5,(T)5)));}

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1008:{     ////H
            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T)-.4).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/h_motion.txt")*s0+dir*tube_height;            
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/h_source.txt");

            if(use_overlapping_tube){   ////read "RAP" or "AP"
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                //BOX<TV> box(TV((T)-.035,(T)-.025,(T)-.03),TV((T).025,(T).02,(T).02));
                //if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high_rap",400,1,&box);
                //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./high3_rap",400,1,&box);

                ////read "AP"
                BOX<TV> box(TV((T)-.035,(T)-.03,(T)-.035),TV((T).035,(T).03,(T).035));
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_ap",540,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_ap",540,1,&box);
                TV trans_X=TV((T)-.013,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;

                letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1009:{ ////simple example tube 1
            T tube_height=s0*(T)1;
            TV dir=TV((T)0,(T)-1,(T)0).Normalized();
            //TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/sample_motion1.txt")*s0+dir*tube_height;
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/exp_motion1.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid){tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);}
            //tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/sample_source1.txt");
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/exp_source1.txt");

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            //min_rim_thickness_factor=(T).1;
        }break;
        case 1010:{ ////simple example tube 2
            //T tube_height=s0*(T)1;
            //TV dir=TV((T)0,(T)-1,(T)-.2).Normalized();
            //TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/sample_motion2.txt")*s0+dir*tube_height;
            //PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            //if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            //else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            //tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            //if(use_tube_rigid){tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);}
            //tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/sample_source2.txt");

            //if(use_overlapping_tube){
            //    ARRAY<CODIMENSIONAL_MESH> meshes(1);
            //    CODIMENSIONAL_MESH& prev_tube_mesh=meshes(1);
            //    ARRAY<TV> prev_tube_X;
            //    BOX<TV> box(TV::Constant_Vector((T)-.03),TV::Constant_Vector((T).03));
            //    CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(prev_tube_mesh,prev_tube_X,0,0,0,"./e2",600,1);
            //    ARRAY<TV> prev_tube_V(prev_tube_X.m);prev_tube_V.Fill(TV());
            //    TV trans_X=TV((T)0,(T)0,(T).001);for(int i=1;i<=prev_tube_X.m;i++)prev_tube_X(i)+=trans_X;

            //    ARRAY<ARRAY<TV> > Xs;Xs.Append(prev_tube_X);
            //    ARRAY<ARRAY<TV> > Vs;Vs.Append(prev_tube_V);
            //    Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,meshes,initial_X,Xs,&initial_V,&Vs);
            //}

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            min_rim_thickness_factor=(T).05;

            T tube_height=s0*(T)1;//T y=s0*(T)8;
            TV dir=TV((T)0,(T)-1,(T)-.2).Normalized(); 
            //TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/sample_motion2.txt")*s0+dir*tube_height;
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/exp_motion2.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            initial_thickness.Resize(initial_X.m);initial_thickness.Fill((T)0);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
            //tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/sample_source2.txt");
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/exp_source2.txt");

            if(use_overlapping_tube){       ////read first part
                int letter_n=1;ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
                ARRAY<ARRAY<TV> > letter_Vs;letter_Vs.Resize(letter_n);
                ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);
                int idx=1;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);
                ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<TV>& letter_V=letter_Vs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
                BOX<TV> box(TV::Constant_Vector((T)-.03),TV::Constant_Vector((T).04));
                //if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./e_a.7",560,1,&box);
                //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./e_a.7",560/*180*/,1,&box);     ////for G
                if(scale<=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./exp_1_h",590,1,&box);
                else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./exp_1_h",590,1,&box);     ////for G

                TV trans_X=TV((T)0,(T)0,(T).002);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;letter_V.Resize(letter_X.m);letter_V.Fill(TV());
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&initial_V,&letter_Vs,0,0,&initial_thickness,&letter_thicknesses);
            }

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 1011:{ ////simple example tube 3
            T tube_height=s0*(T)1;
            TV dir=TV((T)0,(T)-1,(T)-.2).Normalized();
            TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/sample_motion3.txt")*s0+dir*tube_height;
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
            tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            if(use_tube_rigid){tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);}
            tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/sample_source3.txt");

            if(use_overlapping_tube){
                ARRAY<CODIMENSIONAL_MESH> meshes(1);
                CODIMENSIONAL_MESH& prev_tube_mesh=meshes(1);
                ARRAY<TV> prev_tube_X;
                CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(prev_tube_mesh,prev_tube_X,0,0,0,"./jan1_tube2",400,1);
                ARRAY<TV> prev_tube_V(prev_tube_X.m);prev_tube_V.Fill(TV());

                ARRAY<ARRAY<TV> > Xs;Xs.Append(prev_tube_X);
                ARRAY<ARRAY<TV> > Vs;Vs.Append(prev_tube_V);
                Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,meshes,initial_X,Xs,&initial_V,&Vs);
            }

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            min_rim_thickness_factor=(T).05;
        }break;

        case 1:{    ////ssss
            T tube_height=s0*(T)10;
            //TV e1=TV::Axis_Vector(2)*length_scale*(T)15;TV dir=TV((T).6,(T)-1,(T)0).Normalized();   ////high
            TV e1=TV::Axis_Vector(2)*s0*(T)5;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            
            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2);
            tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).2);
            tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
        }break;
        case 2:{    ////c
            T tube_height=s0*(T)1;
            TV e1=TV::Axis_Vector(2)*s0*(T)3;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).08,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).09,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();

            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
        }break;
        case 14:{   ////minor c
            T tube_height=s0*(T)1;
            TV e1=TV::Axis_Vector(2)*s0*(T)3;TV dir=TV((T)1,(T)-.4,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).08,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).09,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();

            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
        }break;
        case 20:{   ////line
            T tube_height=s0*(T)1;
            TV e1=TV((T)0,s0*(T)5,s0*(T)-10);TV dir=TV((T)0,(T)-1,(T)-1).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3*time_scale,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).3*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).101,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();

            TV domain_min=TV((T)-.02,(T)-.001,(T)0);TV domain_max=TV((T).01,(T).01,(T).05);
            global_grid_domain=RANGE<TV>(domain_min,domain_max);
            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
        }break;
        case 3:{    ////alpha
            T tube_height=s0*(T)1;
            TV e1=TV::Axis_Vector(2)*s0*(T)10;TV dir=TV((T).5,(T)-1,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3);
            tube->Add_Source_Speed_Bspline_Time_Value((T).14,(T).3);
            tube->Add_Source_Speed_Bspline_Time_Value((T).15,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
            example.mesher.use_boundary_vertex_snap=true;

            TV domain_size=TV((T).02,(T).02,(T).08);TV domain_min=TV((T)-.01,(T)-.001,(T)-.05);
            global_grid_domain=RANGE<TV>(domain_min,domain_min+domain_size);
        }break;
        case 10:{   ////s
            T tube_height=s0*(T)10;
            TV e1=TV::Axis_Vector(2)*s0*(T)4;TV dir=TV((T)1,(T)-.3,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).3,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).3);
            tube->Add_Source_Speed_Bspline_Time_Value((T).17,(T).3);
            tube->Add_Source_Speed_Bspline_Time_Value((T).18,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
            if(restart_frame==0)tube->Enable(false);
        }break;
        case 11:{   ////line
            T tube_height=s0*(T)10;
            paint_height=s0*(T)1;
            TV e1=TV::Axis_Vector(2)*s0*(T)3;TV dir=TV((T)1,(T)-.5,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).1);
            tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).1);
            tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).2);
            tube->Add_Source_Speed_Bspline_Time_Value((T).26,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
        }break;
        case 12:{   ////sqz tail
            T tube_height=s0*(T)1;
            paint_height=s0*(T).5;
            TV e1=TV::Axis_Vector(2)*s0*(T)2;TV dir=TV((T)1,(T)-1,(T)0).Normalized();

            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).1);
            tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).1);
            tube->Add_Source_Speed_Bspline_Time_Value((T).11,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
            example.mesher.use_segment_mesh=false;
            example.mesher.use_break_thin_segments=false;
            example.mesher.use_break_thin_triangles=false;
            has_adhesion=true;
        }break;
        case 13:{   ////pinch off
            T tube_height=s0*(T)1;
            paint_height=s0*(T).5;
            TV e1=TV::Axis_Vector(2)*s0*(T)3;TV dir=TV((T).75,(T)-1,(T)0).Normalized();    ////low
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).13,(T).1);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
            Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).13);
            tube->Add_Source_Speed_Bspline_Time_Value((T).11,(T).13);
            tube->Add_Source_Speed_Bspline_Time_Value((T).12,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
            tube->Initialize_Source_Speed_Bspline();
            example.mesher.use_segment_mesh=false;
            example.mesher.use_break_thin_segments=false;
            example.mesher.use_break_thin_triangles=false;
            has_adhesion=true;
        }break;
        case 4:{    ////1d coiling
            T tube_height=s0*(T)10;max_rim_thickness_factor=(T)2;radius*=(T).5;
            TV e1=TV::Axis_Vector(2)*s0*(T)40;TV dir=TV((T).1,(T)-1,(T).2).Normalized();   ////high
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).4,(T).1);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            //example.mesher.use_mesh_edge_connect=false;

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
        }break;
        case 5:{    ////folding
            T tube_height=length_scale*(T)10;
            TV e1=TV::Axis_Vector(2)*length_scale*(T)15;TV dir=TV((T)0,(T)-1,(T)0).Normalized();   ////high
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).1,(T).1);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            tube->Initialize_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            //example.mesher.use_mesh_edge_connect=true;
        }break;
        case 6:{    ////random 1d coiling
            T tube_height=s0*(T)15;
            TV e1=TV::Axis_Vector(2)*s0*(T)22;TV dir=TV((T)0,(T)-1,(T)-.3).Normalized();   ////high
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
            //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
            example.mesher.use_mesh_edge_connect=true;

            T speed_scale=(T)1.25;
            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).05,(T).3*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).4*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).15,(T).2*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).3*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).39*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).26,(T).0*time_scale*speed_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0*time_scale*speed_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).25*time_scale*speed_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).35,(T).27*time_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).4,(T).23*time_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).45,(T).37*time_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).5,(T).49*time_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T).55,(T).35*time_scale);
            //tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).25*time_scale);
            tube->Initialize_Source_Speed_Bspline();

            has_adhesion=true;
            use_correcting_particle_position_on_solids=true;
            example.mesher.use_boundary_vertex_snap=true;
            min_rim_thickness_factor=(T).05;
        }break;
        case 7:{    ////jackson pollock
            thickness_scale=length_scale*(T).5;

            ////tube 1
            {T tube_height=length_scale*(T)10;
            TV e1=TV::Axis_Vector(2)*length_scale*(T)8;TV dir=TV((T)1,(T)-1,(T)0).Normalized();   ////high
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).2,(T).1);
            tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).2*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).4,(T).2*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).41,(T).0*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0*time_scale);            
            tube->Initialize_Source_Speed_Bspline();

            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)0,(T)1);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).05,(T)3);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).1,(T).2);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).12,(T)1);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)1,(T)1);
            //tube->Initialize_Source_Thickness_Factor_Bspline();
            }

            ////tube 2
            {T tube_height=length_scale*(T)10;
            TV e1=TV((T)length_scale*(T)-20,length_scale*(T)15,(T)0);TV dir=TV((T)-.5,(T)-1,(T)0).Normalized();   ////high
            PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).0,(T).1);
            tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

            tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).0*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).201,(T).2*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).6,(T).2*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T).601,(T).0*time_scale);
            tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0*time_scale);            
            tube->Initialize_Source_Speed_Bspline();

            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)0,(T)1);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).05,(T)3);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).1,(T).2);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).12,(T)1);
            //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)1,(T)1);
            //tube->Initialize_Source_Thickness_Factor_Bspline();
            }

            example.mesher.use_mesh_edge_connect=true;
            //length_scale*=(T).5;
        }break;
        }

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_friction=false;

        ////surface tension
        has_surface_tension=true;
        surface_tension*=(T).01;
        //example.surface_tension_force.rim_coef_1=(T)1;
        //example.surface_tension_force.rim_coef_2=(T)1;

        ////boundary
        use_non_slip_on_solids=true;
        if(paint_tube_test!=2&&paint_tube_test!=13)use_correcting_particle_position_on_solids=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;
        
        ////shear thinning
        example.viscosity_force.lambda_thinning=(T)1;
        example.viscosity_force.alpha_thinning=(T)1;
        example.viscosity_force.n_thinning=(T).4;

        ////rim
        Use_Rim(true);

        ////levelset
        if(paint_tube_test!=7)use_levelset=false;

        ////meshing
        //example.mesher.use_triangle_mesh=true;
        //example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        //example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_particle_parents=true;

        //example.mesher.use_thin_sheet_aware_tetrahedron_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 216:{   ////test multiple tubes
        T s0=(T)1e-3;length_scale=scale*(T)s0;thickness_scale=length_scale*(T).25;max_thickness_factor=(T)2;min_rim_thickness_factor=(T).01;

        ////paint tube1, left
        T tube_height=s0*(T)5;
        T paint_height=length_scale*(T).5;
        T radius=s0*(T)2;
        {TV e1=TV(s0*(T)-6,s0*(T)7,(T)0);
        TV dir=TV((T)2,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).12);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();}

        ////paint tube2, right
        {TV e1=TV(s0*(T)5,s0*(T)5,(T)0);
        TV dir=TV((T)-1.5,(T)-1,(T)0).Normalized();
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)0,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).11);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tube->Enable(false);}

        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_viscosity=true;
        use_implicit_viscosity=true;
        use_levelset=true;
        levelset_region_number=2;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_boundary_vertex_snap=true;
        use_non_slip_on_solids=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        //example.mesher.use_remove_invalid_elements=true;

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 211:{     ////TEST_PAINT_TRACE
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T)/*.25*/.1;max_thickness_factor=(T)8;min_thickness_factor=(T).05*min_tri_thickness_factor;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).05*min_seg_thickness_factor;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)4;

        ////paint
        //T paint_size=(T)40e-3;TV paint_box_size=TV((T).5*paint_size,(T).3*paint_size,(T)1*paint_size);
        //TV box_min=TV((T)-40e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        //{BOX<TV> box(TV::Constant_Vector((T)-.02),TV::Constant_Vector((T).02));
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,"./dec20_paint_c2",250,1,&box);
        //TV rescale_X=TV((T).7,(T).7,(T).7);for(int i=1;i<=initial_X.m;i++)initial_X(i)*=rescale_X;        
        //TV trans_X=TV((T).0075,(T)0,(T)-.002);for(int i=1;i<=initial_X.m;i++)initial_X(i)+=trans_X;}

        //int letter_n=1;
        //if(use_full_letters)letter_n=2;
        //ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
        //ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
        //ARRAY<ARRAY<ARRAY<T> > > letter_phis;letter_phis.Resize(letter_n);
        //ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);

        //////sigg
        //TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);int idx=1;
        //{CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
        //BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));
        //box.min_corner.x=(T)-.05;box.max_corner.x=(T).013;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./low_sigg",450,1,&box);
        //for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        //TV trans_X=TV();for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}

        ////////raph
        //if(use_full_letters){
        //    idx=2;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
        //    BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));
        //    box.min_corner.x=(T)-.048;box.max_corner.x=(T).029;
        //    CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,"./low_raph",400,1,&box);
        //    for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        //    TV trans_X=TV((T).06,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;
        //}

        //////test s
        //TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);int idx=1;
        //{CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);
        //BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));
        //box.min_corner.x=(T)-.05;box.max_corner.x=(T).013;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,0,"./low_s",300,1,&box);
        //for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        //TV trans_X=TV((T)-.03,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}

        ////////siggr
        if(paint_trace_test==1){
            int letter_n=3;
            ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
            ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
            ARRAY<ARRAY<ARRAY<T> > > letter_phis;letter_phis.Resize(letter_n);
            ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);

            ////new si
            TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);int idx=1;
            {CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
            BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).03));
            box.min_corner.x=(T)-.05;box.max_corner.x=(T).013;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_si",320,1,&box);
            for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
            TV trans_X=TV((T)-.024,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
            ////new gg
            {idx=2;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
            BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_gg",500,1,&box);
            for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
            TV trans_X=TV((T).005,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
            ////new r
            if(use_full_letters){idx=3;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
            BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;box.max_corner.x=(T).04;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_r",450,1,&box);
            for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
            TV trans_X=TV((T).0205,(T)0,(T).005);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}

            ////Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&letter_thicknesses);
            Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,0,0,0,0,&initial_thickness,&letter_thicknesses);
        }
        ////exposition
        else if(paint_trace_test==2){
            BOX<TV> box(TV::Constant_Vector((T)-.04),TV::Constant_Vector((T).04));
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,&initial_thickness,0,0,"./exp_2_h23",500,1,&box);
            CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,&initial_thickness,0,0,trace_filename,trace_init_frame,1,&box);
            TV trans_X=TV((T)-.025,(T)0,(T).007);for(int i=1;i<=initial_X.m;i++)initial_X(i)+=trans_X;
        }

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        //example.surface_tension_force.st_on_tris=false;
        //example.surface_tension_force.st_on_segs=false;
        has_friction=true;
        friction_magnitude=(T).8;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_paint_trace_boundary=true;
        coupled_evolution.use_squash_force=true;
        use_low_fric_near_fiber=true;

        ////levelset
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_fiber=true;
        adaptive_length_scale_min=length_scale*(T).5; 
        adaptive_length_scale_min2=length_scale*(T).2;
        example.mesher.use_anisotropic_triangle_remeshing=false;

        ////rim
        Use_Rim(false);

        ////brush
        T s1=s0*(T)1.4;
        fiber_length_scale=(T)4*s1;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber_radius=(T).25*s1;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45*s1;
        fiber.object_frame=FRAME<TV>(TV((T)-24*s1,(T)28.5*s1,(T)0*s1),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.object_frame=FRAME<TV>(TV((T)0,(T)0,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));

        fiber.implicit_fiber_segment_number=15;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;
        fiber.rigid_body_type=FIBER<TV>::NONE;

        coupled_evolution.use_trace_fiber_segments=true;
        coupled_evolution.use_squash_force=true;
        has_fiber_fluid_coupling_force=true;

        if(paint_trace_test==1){
            ////Predefined motion
            fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
            //std::string dir="../model/Motion_Data/dec25_fiber";
            std::string dir="../model/Motion_Data/j4_trace1";
            fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T).05);
        }
        else if(paint_trace_test==2){
            fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;

            if(use_sim_box){
                ////for trace_test3, restart 735
                //sim_particle_boxes.Append(BOX<TV>(TV((T)-1,(T)-1,(T)-1),TV((T).008,(T)1,(T)1)));
                //boundary_particle_boxes.Append(BOX<TV>(TV((T).008,(T)-1,(T)-1),TV((T).01,(T)1,(T)1)));

                ////for trace_test4, restart from frame 429
                sim_particle_boxes.Append(BOX<TV>(TV((T)-.0135,(T)-1,(T)-1),TV((T)1,(T)1,(T).018)));
                boundary_particle_boxes.Append(BOX<TV>(TV((T)-.0135,(T)-1,(T)-1),TV((T)-.013,(T)1,(T).018)));
            }
        }

        int fiber_number_row=6;int fiber_number_column=20;
        //int fiber_number_row=3;int fiber_number_column=20;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////mpi
        //if(use_full_letters){
        //    TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).2;TV domain_center=TV();domain_center.x=(T).0132;
        //    global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //}
        //else{
        //    TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).1;TV domain_center=TV();domain_center.x=(T)-.016;
        //    global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //}

        ////mpi
        if(use_full_letters){
            TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).16;TV domain_center=TV();
            global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        }
        else{
            TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).16;TV domain_center=TV();domain_center.x=(T)-.01;
            global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        }

        Add_Canvas();
    }break;
    case 223:{
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T)/*.25*/.1;max_thickness_factor=(T)8;min_thickness_factor=(T).05;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).05;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)4;

        ////paint
        T paint_size=(T)40e-3;TV paint_box_size=TV((T).5*paint_size,(T).3*paint_size,(T)1*paint_size);
        TV box_min=TV((T)-40e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        //////test s
        //TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);int idx=1;
        //{CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);
        //BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));
        //box.min_corner.x=(T)-.05;box.max_corner.x=(T).013;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,0,0,"./low_s",300,1,&box);
        //for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        //TV trans_X=TV((T)-.03,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}

        int letter_n=3;
        ARRAY<CODIMENSIONAL_MESH> letter_meshes;letter_meshes.Resize(letter_n);
        ARRAY<ARRAY<TV> > letter_Xs;letter_Xs.Resize(letter_n);
        ARRAY<ARRAY<ARRAY<T> > > letter_phis;letter_phis.Resize(letter_n);
        ARRAY<ARRAY<T> > letter_thicknesses;letter_thicknesses.Resize(letter_n);

        ////new si
        TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);int idx=1;
        {CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
        BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).03));
        box.min_corner.x=(T)-.05;box.max_corner.x=(T).013;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_si",320,1,&box);
        for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        TV trans_X=TV((T)-.024,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
        ////new gg
        {idx=2;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
        BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_gg",500,1,&box);
        for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        TV trans_X=TV((T).005,(T)0,(T)0);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
        ////new r
        if(use_full_letters){idx=3;CODIMENSIONAL_MESH& letter_mesh=letter_meshes(idx);ARRAY<TV>& letter_X=letter_Xs(idx);ARRAY<T>& letter_thickness=letter_thicknesses(idx);
        BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;box.max_corner.x=(T).04;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_r",450,1,&box);
        for(int i=1;i<=letter_X.m;i++)letter_X(i)*=rescale_X;        
        TV trans_X=TV((T).0205,(T)0,(T).005);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}

        //Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,&letter_thicknesses);
        Merge_Mesh_And_Attributes_In_Different_Regions(initial_mesh,letter_meshes,initial_X,letter_Xs,0,0,0,0,&initial_thickness,&letter_thicknesses);

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        example.surface_tension_force.st_on_tris=false;
        example.surface_tension_force.st_on_segs=false;
        has_friction=true;
        friction_magnitude=(T).8;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_paint_trace_boundary=true;
        coupled_evolution.use_squash_force=true;
        use_low_fric_near_fiber=true;

        ////levelset
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_fiber=true;
        adaptive_length_scale_min=length_scale*(T).5; 
        adaptive_length_scale_min2=length_scale*(T).2;
        example.mesher.use_anisotropic_triangle_remeshing=false;

        ////rim
        Use_Rim(false);

        ////brush
        T s1=s0*(T)1.4;
        fiber_length_scale=(T)4*s1;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber_radius=(T).25*s1;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45*s1;
        fiber.object_frame=FRAME<TV>(TV((T)-24*s1,(T)28.5*s1,(T)0*s1),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.object_frame=FRAME<TV>(TV((T)0,(T)0,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));

        //fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=15;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;
        fiber.rigid_body_type=FIBER<TV>::NONE;

        coupled_evolution.use_trace_fiber_segments=true;
        coupled_evolution.use_squash_force=true;
        has_fiber_fluid_coupling_force=true;

        ////Predefined motion
        fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
        //std::string dir="../model/Motion_Data/dec25_fiber";
        std::string dir="../model/Motion_Data/j4_trace1";
        fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T).05);

        int fiber_number_row=6;int fiber_number_column=20;
        //int fiber_number_row=3;int fiber_number_column=20;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////mpi
        //if(use_full_letters){
        //    TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).2;TV domain_center=TV();domain_center.x=(T).0132;
        //    global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //}
        //else{
        //    TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).1;TV domain_center=TV();domain_center.x=(T)-.016;
        //    global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //}

        ////mpi
        if(use_full_letters){
            TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).16;TV domain_center=TV();
            global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        }
        else{
            TV domain_size=TV::Constant_Vector((T).2);domain_size.x=(T).16;TV domain_center=TV();domain_center.x=(T)-.01;
            global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        }

        Add_Canvas();
    }break;
    case 83:{   ////lifting brush, tearing paint volume
        length_scale=scale*(T)1e-3;thickness_scale=(T).25e-3;
        //length_scale_min=(T).05e-3;
        adaptive_length_scale_min=(T).25e-3;
        max_thickness_factor=(T)2;

        //////paint
        //T paint_size=(T)10e-3;TV paint_box_size=TV((T).5*paint_size,(T).5*paint_size,paint_size);
        //TV box_min=TV((T)-2.5e-3,(T)0,(T)-5e-3);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        ////coiling
        TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);//int idx=1;
        {BOX<TV> box(TV::Constant_Vector((T)-.1),TV::Constant_Vector((T).1));box.max_corner.x=(T).01;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"./coiling5",500,1,&box);
        TV trans_X=TV((T)-.008,(T)0,(T).01);for(int i=1;i<=initial_X.m;i++)initial_X(i)+=trans_X;}

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        has_friction=true;
        friction_magnitude=(T).8;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_paint_trace_boundary=true;
        coupled_evolution.use_squash_force=true;
        has_fiber_fluid_coupling_force=true;

        ////levelset
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_variable_viscosity_for_codim=true;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        //example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_fiber=true;
        adaptive_length_scale_min=length_scale*(T).5; 
        adaptive_length_scale_min2=length_scale*(T).2;
        example.mesher.use_anisotropic_triangle_remeshing=false;

        ////rim
        Use_Rim(true);

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber_radius=(T).2e-3;fiber_distance=(T).6e-3;fiber_length=(T)45e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)15e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));
        fiber.object_frame=FRAME<TV>(TV((T)0,(T)35e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        fiber.use_fiber_length_variation_on_x_axis=false;
        fiber.density=(T)5e4;
        coupled_evolution.apply_force_on_fiber=true;
        band_width_min=(T)2*fiber_radius;band_width_max=(T)2e-3;

        int fiber_number_row=6;int fiber_number_column=20;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas((T)1);
    }break;
    case 85:{   ////folding thin sheet
        T scale=(T)1;//T one_over_scale=(T)1/scale;
        length_scale=scale*(T)1e-3;thickness_scale=(T).25e-3;
        //length_scale_min=(T).05e-3;
        adaptive_length_scale_min=(T).25e-3;
        max_thickness_factor=(T)2;

        ////paint
        T paint_size=(T)10e-3;TV paint_box_size=TV((T).5*paint_size,(T).5*paint_size,paint_size);
        TV box_min=TV((T)-2.5e-3,(T)0,(T)-5e-3);TV box_max=box_min+paint_box_size;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).05;
        tri_surface_tension=surface_tension*(T).2;
        has_viscosity=true;
        use_implicit_viscosity=true;
        kinematic_viscosity*=(T)1e3;
        use_levelset=false;
        has_fiber_fluid_coupling_force=true;
        has_friction=true;
        coupled_evolution.use_squash_force=true;
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_thickness_aware_triangle_remeshing=false;
        example.mesher.use_contact_aware_triangle_remeshing=false;
        example.incompressible.use_target_div=false;
        use_correcting_particle_position_on_solids=false;

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber_radius=(T).2e-3;fiber_distance=(T).6e-3;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)15e-3,(T)30e-3,(T)0),ROTATION<TV>((T)-.3*(T)half_pi,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;band_width_max=(T)2e-3;

        int fiber_number_row=3;int fiber_number_column=20;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 212:{  ////TEST_PAINT_DIP
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T)/*.25*/.1;max_thickness_factor=(T)8;min_thickness_factor=(T).05;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).05;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)4;

        ////paint
        T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).3*paint_size,paint_size);
        TV box_min=TV((T)8e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;

        ////coiling
        TV offset=TV();TV rescale_X=TV((T)1,(T)1,(T)1);//int idx=1;
        {BOX<TV> box(TV::Constant_Vector((T)-.1),TV::Constant_Vector((T).1));box.max_corner.x=(T).01;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"./coiling3",438,1,&box);
        TV trans_X=TV((T)0,(T)0,(T).01);for(int i=1;i<=initial_X.m;i++)initial_X(i)+=trans_X;}

        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //int n=(int)(paint_box_size.x/length_scale);int m=(int)(paint_box_size.z/length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Herring_Bone_Rectangle(initial_mesh,initial_X,m+1,n+1,box_min,length_scale,2);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,"./nov29_14",260,2);  ////alpha

        //////paint
        //T paint_size=(T)40e-3;TV paint_box_size=TV((T).5*paint_size,(T).3*paint_size,(T)1*paint_size);
        //TV box_min=TV((T)-40e-3,(T)0,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        ////brush
        fiber_length_scale=(T)4e-3;
        //fiber_radius=(T).1e-3;fiber_distance=fiber_radius*(T)6;fiber_length=(T)40e-3;
        //fiber.object_frame=FRAME<TV>(TV((T)4e-3,fiber_length*(T).9,(T)0),ROTATION<TV>((T)-.15*(T)half_pi,TV::Axis_Vector(3)));
        fiber_radius=(T).2e-3;fiber_distance=fiber_radius*(T)3;fiber_length=(T)45e-3;
        fiber.object_frame=FRAME<TV>(TV((T)-10e-3,(T)35e-3,(T)10e-3),ROTATION<TV>((T).3*(T)half_pi,TV::Axis_Vector(3)));
        //fiber.object_frame=FRAME<TV>(TV((T)0,(T)0,(T)0),ROTATION<TV>((T)0,TV::Axis_Vector(3)));

        fiber.fiber_deformable_model=FIBER<TV>::MASS_SPRING;
        fiber.implicit_fiber_segment_number=10;
        fiber.fiber_bending_start_frac=(T).6;   
        fiber.fiber_bending_perturb_frac=TV((T).01,(T)0,(T)0.01);
        fiber.use_virtual_springs=true;
        fiber.fiber_evolution.use_ether_damping=true;
        band_width_min=(T)2*fiber_radius;

        //////Predefined motion
        //fiber.fiber_deformable_model=FIBER<TV>::READ_FROM_FILE;
        //std::string dir="../model/Motion_Data/brush_coarse_curve";//T t0=(T).0;T t1=(T).6;T motion_t0=(T).2;
        //fiber.Initialize_Attributes_From_File(dir,/*t0*/(T)0,/*t1*/(T)1,/*motion_t0*/(T)0);

        int fiber_number_row=6;int fiber_number_column=20;
        //int fiber_number_row=1;int fiber_number_column=20;
        //T fiber_width=fiber_number_column*fiber_distance;
        CODIMENSIONAL_MESH_GENERATOR<TV>::Rectangle_Particles_Centered(fiber_positions,fiber_number_row,fiber_number_column,/*TV()*/TV::Axis_Vector(2)*fiber_length_scale*(T)2,fiber_distance,TV::Axis_Vector(1),TV::Axis_Vector(3),1,/*perturb*/fiber_distance*(T).4);
        Add_Fiber_Rigid_Body(fiber_positions,fiber);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();

        //{BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;box.max_corner.x=(T).04;
        ////CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_r",600,1,&box);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,TV(),TV::Constant_Vector(length_scale*(T)3),length_scale);}

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).01;
        example.surface_tension_force.st_on_tris=false;
        example.surface_tension_force.st_on_segs=false;
        has_friction=true;
        friction_magnitude=(T).8;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_paint_trace_boundary=true;
        coupled_evolution.use_squash_force=true;
        use_low_fric_near_fiber=true;

        ////levelset
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_fiber=true;
        adaptive_length_scale_min=length_scale*(T).5; 
        adaptive_length_scale_min2=length_scale*(T).2;
        example.mesher.use_anisotropic_triangle_remeshing=false;

        ////rim
        Use_Rim(true);

        Add_Canvas();
    }break;
    case 213:{     ////TEST_PAINT_SPRAY     
        T s0=(T)5e-2;length_scale=scale*s0;thickness_scale=s0*(T).1;
        max_thickness_factor=(T)1.5;min_thickness_factor=(T).1;
        rim_thickness_scale=thickness_scale*(T)1;max_rim_thickness_factor=(T)8;min_rim_thickness_factor=(T).7;
        adaptive_length_scale_min=length_scale*(T).3;

        ////paint
        T paint_height=s0*(T)2;T radius=s0*(T)2;T tube_height=s0*(T)5;
        TV e1=TV((T)0,(T).2,(T)0);TV dir=TV((T)0,(T)1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)3,(T).6,dir);
        tube->Initialize_Ring_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tube->has_rigid_body=false;
        tube->ring_source_angle=(T)half_pi*(T).4;
        tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)5);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)10,(T).0);
        tube->Initialize_Source_Speed_Bspline();
        tube->ring_source_peak_num=ring_source_peak_number;
        tube->use_ring_source_randomness=use_ring_source_randomness;

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=false;
        has_friction=true;
        use_correcting_particle_position_on_solids=true;
        //use_non_slip_on_solids=true;
        
        ////surface tension
        has_surface_tension=true;
        surface_tension*=(T)1;
        seg_surface_tension=surface_tension*(T)2;
        example.surface_tension_force.viscous=false;

        ////levelset
        use_levelset=false;
        
        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;

        ////shear thinning
        example.viscosity_force.lambda_thinning=(T).05;
        example.viscosity_force.alpha_thinning=(T).8;
        example.viscosity_force.n_thinning=(T).2;

        ////rim
        Use_Rim(true);
        has_rim_segments_for_viscosity=true;
        has_rim_triangles_for_viscosity=false;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_remove_invalid_elements=true;
        //example.mesher.use_continuous_meshing_operations=false;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        adaptive_length_scale_min=length_scale*(T).25;
        use_adaptive_length_scale_on_codim=false;
        use_adaptive_length_scale_on_levelset=false;

        ////mpi
        TV domain_size=TV::Constant_Vector((T)1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas((T)20);
    }break;
    case 214:{  ////TEST_PAINT_SPLASH
        T s0=(T)5e-2;length_scale=scale*s0;
        thickness_scale=length_scale*(T).2;
        max_thickness_factor=(T)8;min_thickness_factor=(T).1;
        rim_thickness_scale=(T).05*s0;max_rim_thickness_factor=(T)10;min_rim_thickness_factor=(T).2;
        
        ////////paint
        T paint_size=(T)1;T paint_h=(T).12*paint_size*paint_height_factor;TV paint_box_size=TV(paint_size,paint_h,paint_size);
        TV box_min=(T)-.5*paint_box_size;box_min.y=(T)0;TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        
        ////paint source
        T paint_height=s0*(T)2;T radius=s0*(T)1.3;T tube_height=s0*(T)5;
        TV e1=TV((T)0,(T).5*paint_box_size.y,(T)0);TV dir=TV((T)0,(T)1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)1,(T).6,dir);
        tube->ring_source_angle=(T)half_pi*(T).75;
        tube->Initialize_Ring_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tube->has_rigid_body=false;
        tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).05,(T)2.5);
        tube->Add_Source_Speed_Bspline_Time_Value((T).15,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).16,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)10,(T).0);
        tube->Initialize_Source_Speed_Bspline();

        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)0,(T)1);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).1,(T).2);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).2,(T).01);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)1,(T).01);
        //tube->Initialize_Source_Thickness_Factor_Bspline();

        T max_c=5;
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)1);
        for(int i=1;i<=max_c;i++){
            T frac_0=(T)(i-1)/(T)max_c;
            T frac_1=(T)i/(T)max_c;
            T a=random.Get_Uniform_Number(frac_0,frac_1);
            T speed=(i%2==1?random.Get_Uniform_Number((T).1,(T).2):random.Get_Uniform_Number((T)3,(T)5));
            tube->Add_Ring_Source_Speed_Bspline_Time_Value(a,speed);
        }
        tube->Initialize_Ring_Source_Speed_Bspline();

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        seg_surface_tension=surface_tension*(T)2;
        example.surface_tension_force.viscous=false;
        has_friction=false;
        friction_magnitude=(T).02;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;

        ////levelset
        use_levelset=false;
        levelset_region_number=3;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_particle_parents=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_rim=false;
        adaptive_length_scale_min=length_scale*(T).3;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;

        ////rim
        Use_Rim(true);
        has_rim_segments_for_viscosity=true;
        has_rim_triangles_for_viscosity=false;

        //////falling rock
        //STREAM_TYPE float_stream_type(1.f);
        //RIGID_BODY<TV>* kinematic_rigid_body;
        //kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("rock3",(T)1.0,(T)0.0,true,true,&float_stream_type);
        //kinematic_rigid_body->Rescale((T)0.01);
        //kinematic_rigid_body->X()=TV();
        //rigid_bodies.Append(kinematic_rigid_body);
        //kinematic_rigid_body->name="rock";
        //kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////T rock_height=paint_box_size.y+(T)0.2;
        //T rock_height=paint_box_size.y+(T)0.065;

        //rock=new ROCK<TV>();
        //rock->motion_type=ROCK<TV>::GRAVITY_WITH_ROTATION;
        ////rock->object_frame.t.y=rock_height;
        //rock->initial_height=rock_height;
        //rock->rigid_body=kinematic_rigid_body;
        //rock->rigid_body->X()=rock->object_frame.t;
        //rock->rigid_body->Rotation()=rock->object_frame.r;

        ////rock->final_height=(T)0.08;
        ////rock->final_height=(T)0.065;
        //rock->final_height=(T)0.054;
        ////rock->initial_velocity=(T)-5;
        //rock->initial_velocity=(T)-5;
        //rock->rigid_body->V()=TV::Axis_Vector(2)*(-rock->initial_velocity);
        ////rock->use_slow_motion=true;
        ////rock->slow_motion_factor=(T)0.5;
        ////rock->slow_motion_threshold=(T)0.218;
        //
        ////rock->motion_type=TURN_SPREAD;
        //rock->Set_Prescribed_Motion_For_Test();

        //TV tank_box_min=box_min;TV tank_box_max=box_max;tank_box_max.y*=(T)1.2;
        //BOX<TV> tank_box=BOX<TV>(tank_box_min,tank_box_max);tank=new WATER_TANK<TV>(tank_box,length_scale*(T)2,true,&example.geometrical_codimensional_mesh_object);

        TV domain_size=TV::Constant_Vector((T)1.1);domain_size.y=(T)2;TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 218:{  ////TEST_PAINT_SPRAY2
        T s0=(T)5e-3;length_scale=scale*s0;
        //thickness_scale=length_scale*(T).1;
        thickness_scale=(T)5e-4;
        max_thickness_factor=(T)4;min_thickness_factor=(T).1;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)10;min_rim_thickness_factor=(T)1;
        
        //////cylinder
        //T cylinder_length=(T)100e-3;
        T cylinder_radius=(T)15e-3;
        //RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        //kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        //rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        //kinematic_rigid_bodies.Append(kinematic_rigid_body);

        ////paint source
        T paint_height=s0*(T)2;T radius=cylinder_radius;T tube_height=s0*(T)5;
        TV e1=TV((T)0,(T)0,(T)0);TV dir=TV((T)0,(T)0,(T)1).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)1,(T).6,TV::Axis_Vector(3));
        tube->ring_source_angle=(T)0;
        tube->Initialize_Ring_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tube->has_rigid_body=false;
        tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).05,(T)2.5);
        tube->Add_Source_Speed_Bspline_Time_Value((T).15,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T).16,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)10,(T).0);
        tube->Initialize_Source_Speed_Bspline();

        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)0,(T)1);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).1,(T).2);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T).2,(T).01);
        //tube->Add_Source_Thickness_Factor_Bspline_Time_Value((T)1,(T).01);
        //tube->Initialize_Source_Thickness_Factor_Bspline();

        //T max_c=30;
        //tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)1);
        //for(int i=1;i<=max_c;i++){
        //    T frac_0=(T)(i-1)/(T)max_c;
        //    T frac_1=(T)i/(T)max_c;
        //    T a=random.Get_Uniform_Number(frac_0,frac_1);
        //    T speed=(i%2==1?random.Get_Uniform_Number((T).1,(T).2):random.Get_Uniform_Number((T)3,(T)5));
        //    tube->Add_Ring_Source_Speed_Bspline_Time_Value(a,speed);
        //}
        //tube->Initialize_Ring_Source_Speed_Bspline();

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        seg_surface_tension=surface_tension*(T)2;
        example.surface_tension_force.viscous=false;
        has_friction=false;
        use_correcting_particle_position_on_solids=false;

        ////levelset
        use_levelset=false;
        levelset_region_number=3;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_half_cylindrical_rim=true;
        example.mesher.use_continuous_meshing_operations=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        use_adaptive_length_scale_on_codim=false;
        use_adaptive_length_scale_on_rim=true;
        adaptive_length_scale_min=length_scale*(T).3;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;

        ////rim
        Use_Rim(true);
        has_rim_segments_for_viscosity=true;
        has_rim_triangles_for_viscosity=false;

        TV domain_size=TV::Constant_Vector((T)1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 221:{  ////TEST_PAINT_SPRAY3
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).5;max_thickness_factor=(T)2;min_thickness_factor=(T).1;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).1;

        ////cylinder
        T cylinder_length=(T)100e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        T tube_height=s0*(T)2;T paint_height=s0*(T)4;T tube_radius=s0*(T)3;
        TV dir=TV((T)0,(T)-1,(T)0).Normalized();//TV e1=TV((T)-.5*cylinder_radius,cylinder_radius+(T)30*s0,(T)0);

        ////tube1
        {TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/spray_motion2.txt")*s0+dir*tube_height;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,tube_radius,length_scale,(T).3,(T).1);
        //tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
        tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/spray_source2.txt");
        }
        ////tube2
        {TV e1=Read_Tube_Initial_Position_From_File("../model/Motion_Data/tube_curve/spray_motion3.txt")*s0+dir*tube_height;
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,tube_radius,length_scale,(T).3,(T).1);
        //tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        if(use_1d_tube)tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        else tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        if(use_tube_rigid)tube->rigid_body_type=PAINT_TUBE<TV>::PLACEHOLDER;Add_Tube_Rigid_Body(*tube);tubes.Append(tube);
        tube->Initialize_Source_Speed_Bspline_From_File("../model/Motion_Data/tube_curve/spray_source3.txt");
        }

        ////paint
        //T paint_size=(T)8e-3;TV paint_box_size=TV(paint_size,(T).5*paint_size,paint_size);
        //TV box_min=TV((T)-paint_box_size.x*(T).5,cylinder_radius-length_scale*(T).2,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //TV e1=TV::Axis_Vector(2)*cylinder_radius;TV e2=e1+TV::Axis_Vector(2)*paint_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Filament(initial_mesh,initial_X,e1,e2,length_scale);

        //T tube_height=(T)12e-3;
        //T tube_radius=(T)4e-3;
        //T tube_height2=(T)80e-3;
        //TV tube_e1=TV(cylinder_radius*(T).5,tube_height2+cylinder_radius,(T)0);
        //TV tube_dir=TV::Axis_Vector(2)*(T)-1;
        //tube=new PAINT_TUBE<TV>(tube_e1,tube_dir,tube_height,tube_radius,length_scale,(T).0,(T).1);
        //tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);
        //Add_Tube_Rigid_Body(*tube);

        //{/*int i=1;*/T paint_radius=(T)5e-3;TV paint_center=TV((T)-.5*cylinder_radius,cylinder_radius+(T)40e-3,(T)0);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        //{/*int i=2;*/T paint_radius=(T)5e-3;TV paint_center=TV((T).2*cylinder_radius,cylinder_radius+(T)80e-3,(T)30e-3);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        //{/*int i=3;*/T paint_radius=(T)5e-3;TV paint_center=TV((T)-.6*cylinder_radius,cylinder_radius+(T)60e-3,(T)-20e-3);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,paint_center,paint_radius,length_scale);}

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        seg_surface_tension=surface_tension*(T)2;
        example.surface_tension_force.viscous=false;

        has_friction=false;
        friction_magnitude=(T).02;
        //use_correcting_particle_position_on_solids=true;
        use_non_slip_on_solids=true;
        has_adhesion=true;

        ////levelset
        use_levelset=false;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_particle_parents=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_remove_invalid_elements=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_rim=false;
        adaptive_length_scale_min=length_scale*(T).3;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;

        ////rim
        example.mesher.use_half_cylindrical_rim=true;
        Use_Rim(true);
        has_rim_segments_for_viscosity=true;
        has_rim_triangles_for_viscosity=false;

        TV domain_size=TV::Constant_Vector((T)1.1);domain_size.y=(T)2;TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 222:{  ////TEST_PAINT_SPRAY4
        T s0=(T)1e-3;length_scale=scale*s0;
        thickness_scale=length_scale*(T).1;max_thickness_factor=(T)10;min_thickness_factor=(T).01;
        rim_thickness_scale=thickness_scale;max_rim_thickness_factor=(T)10;min_rim_thickness_factor=(T).1;

        ////cylinder
        T cylinder_length=(T)100e-3;T cylinder_radius=(T)15e-3;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Cylinder(cylinder_length,cylinder_radius);
        kinematic_rigid_body->X()=TV();kinematic_rigid_body->Rotation()=ROTATION<TV>((T).1,TV::Axis_Vector(3));
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        BOX<TV> box(TV::Constant_Vector((T)-.03),TV::Constant_Vector((T).03));box.max_corner.y=(T).027;
        //if(scale>(T)1)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"./j3_spray_l",1800,1,&box);
        //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"./j3_spray_h",1800,1,&box);
        //if(scale>(T)1)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,&initial_V,0,"./j5_cy3",1080,1,&box);
        //else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,&initial_V,0,"./j5_cy3",1080,1,&box);

        if(scale>(T)1)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,&initial_V,0,spray_filename,spray_init_frame,1,&box);
        else CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,&initial_V,0,spray_filename,spray_init_frame,1,&box);

        ////forces
        has_gravity=true;fiber.use_gravity=true;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        //seg_surface_tension=surface_tension*(T)2;
        example.surface_tension_force.viscous=false;
        has_friction=false;
        friction_magnitude=(T).02;
        use_non_slip_on_solids=true;
        has_adhesion=true;

        ////levelset
        use_levelset=false;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=false;
        example.mesher.use_constrained_thin_segment_break=false;
        example.mesher.use_particle_parents=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_continuous_meshing_operations=false;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)4;
        example.mesher.clamp_segment_thickness_max=thickness_scale*(T)8;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        use_adaptive_length_scale_on_rim=false;
        adaptive_length_scale_min=length_scale*(T).3;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_particle_thickness=true;
        example.viscosity_force.use_staggered_dof=true;

        ////rim
        example.mesher.use_half_cylindrical_rim=true;
        Use_Rim(true);
        has_rim_segments_for_viscosity=true;
        has_rim_triangles_for_viscosity=false;

        TV domain_size=TV::Constant_Vector((T).8);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        //Add_Canvas();
    }break;
    case 215:{  ////TEST_MATERIALS
        T s0=(T)5e-3;length_scale=scale*s0;
        thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).06/*.02*/*min_tri_thickness_factor;
        max_rim_thickness_factor=(T)1.5;min_rim_thickness_factor=(T).07/*.05*/*min_seg_thickness_factor;
        example.mesher.clamp_segment_thickness_max=thickness_scale*(T)1.5;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)2;

        ////paint
        T paint_height=s0*(T)2;T radius=s0*(T)2;T tube_height=s0*(T)5;
        TV e1=TV((T)0,s0*(T)70,(T)1*s0);TV dir=TV((T)0,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).5,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
        //tube->Initialize_Thin_Sheet_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);Add_Tube_Rigid_Body(*tube);

        //tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).3);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).7);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).2);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).6);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).35,(T).45);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).4,(T).2);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).5,(T).6);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).6,(T).4);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).65,(T).4);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).75,(T).4);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).8,(T).4);
        //tube->Add_Source_Speed_Bspline_Time_Value((T).85,(T).25);
        //tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).1);
        //tube->Add_Source_Speed_Bspline_Time_Value((T)2,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T).1,(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).2,(T).6);
        tube->Add_Source_Speed_Bspline_Time_Value((T).25,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).3,(T).5);
        tube->Add_Source_Speed_Bspline_Time_Value((T).35,(T).45);
        tube->Add_Source_Speed_Bspline_Time_Value((T).4,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).5,(T).6);
        tube->Add_Source_Speed_Bspline_Time_Value((T).6,(T).4);
        tube->Add_Source_Speed_Bspline_Time_Value((T).65,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T).75,(T).3);
        tube->Add_Source_Speed_Bspline_Time_Value((T).8,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T).9,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1.1,(T).1);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1.2,(T).4);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1.3,(T).2);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1.4,(T).4);
        tube->Add_Source_Speed_Bspline_Time_Value((T)2,(T).1);
        tube->Initialize_Source_Speed_Bspline();
        tube->object_space_r=tube->radius*(T)5;
        tube->use_texture_coordinates=true;

        ////solids
        {T sphere_radius=s0*(T)8;
        RIGID_BODY<TV>* rigid_body=&rigid_standard_tests.Add_Analytic_Sphere(sphere_radius,(T)1);
        rigid_body->X()=TV();
        rigid_body->X().y=s0*(T)30;
        rigid_body->Rotation()=ROTATION<TV>((T)half_pi,TV::Axis_Vector(1));
        rigid_bodies.Append(rigid_body);}
        Add_Canvas((T)1);

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)2;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_friction=true;
        friction_magnitude=(T).5;

        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        example.viscosity_force.use_variable_viscosity_for_codim=true;

        ////shear thinning
        example.viscosity_force.lambda_thinning=(T).05;
        example.viscosity_force.alpha_thinning=(T).8;
        example.viscosity_force.n_thinning=(T).2;

        ////shear thickening
        //example.viscosity_force.lambda_thickening=(T).05;
        //example.viscosity_force.alpha_thickening=(T).8;
        //example.viscosity_force.n_thickening=(T)6;
        example.viscosity_force.lambda_thickening=(T).05;
        example.viscosity_force.alpha_thickening=(T).8;
        example.viscosity_force.n_thickening=(T)2;
        if(example.viscosity_force.variable_viscosity_mode==(T)4)example.viscosity_force.viscosity_max_clamp=(T).1;    ////for shear thickening example

        ////elasticity
        has_elasticity=true;
        has_plasticity=true;
        use_implicit_elasticity=true;

        ////meshing
        example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_boundary_vertex_snap=true;
        example.mesher.use_split_eulerian_particles=true;
        example.mesher.use_codim_vertex_snap=true;
        example.mesher.use_remove_invalid_elements=true;
        //example.mesher.use_fix_poping=false;
        example.mesher.use_clamp_max_thickness=true;

        ////adaptivity
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).5;        
        
        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).1);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 217:{  ////TEST_MATERIALS_2
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=length_scale*(T).2;

        ////paint
        T paint_size=(T)2*s0;TV paint_box_size=TV(paint_size,paint_size*(T).5,paint_size*(T)2);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);

        has_gravity=true;fiber.use_gravity=has_gravity;gravity*=(T)2;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_friction=true;
        friction_magnitude=(T).02;
        has_viscosity=true;
        use_implicit_viscosity=true;
        has_elasticity=true;
        has_plasticity=true;
        use_implicit_elasticity=true;
        //TODO: use triangle & segment meshes
        example.mesher.use_triangle_mesh=false;//true;
        example.mesher.use_segment_mesh=false;//true;
        example.mesher.use_break_thin_triangles=true;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=true;
        example.mesher.use_adaptive_triangle_remeshing=true;
        example.mesher.use_adaptive_segment_remeshing=true;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_boundary_vertex_snap=true;

        example.viscosity_force.use_thickness_dependent_variable_viscosity=false;
        use_non_slip_on_solids=false;
        use_correcting_particle_position_on_solids=false;
        use_levelset=false;

        ////kinematic solid
        /*
        TV rigid_body_size=TV::Constant_Vector(paint_size);rigid_body_size.x*=(T).25;rigid_body_size.z*=(T).25;
        RIGID_BODY<TV>* kinematic_rigid_body=&rigid_standard_tests.Add_Analytic_Box(rigid_body_size);
        kinematic_rigid_body->X()=TV();
        kinematic_rigid_body->X().x=box_min.x-rigid_body_size.x*(T).52;
        kinematic_rigid_body->X().y=(T).55*rigid_body_size.y;
        kinematic_rigid_body->X().z-=((T).5*rigid_body_size.z+paint_box_size.z*(T).25);
        rigid_bodies.Append(kinematic_rigid_body);kinematic_rigid_body->name="kin_obj1";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);
        */

        ////rigid body
        STREAM_TYPE float_stream_type(1.f);
        RIGID_BODY<TV>* kinematic_rigid_body;
        kinematic_rigid_body=&rigid_standard_tests.Add_Rigid_Body("paint_scraper",(T)1.0,(T)0.0,true,true,&float_stream_type);
        kinematic_rigid_body->Rescale((T)0.01);
        kinematic_rigid_body->X()=TV();
        rigid_bodies.Append(kinematic_rigid_body);
        kinematic_rigid_body->name="scraper";
        kinematic_rigid_bodies.Append(kinematic_rigid_body);

        scraper=new SCRAPER<TV>();
        scraper->initial_X(1)=(T)-0.008;
        scraper->initial_X(2)=(T)0.0005;
        scraper->initial_X(3)=(T)-0.01;
        scraper->initial_V(1)=(T)0.1;
        scraper->initial_V*=rigid_body_speed_factor;
        scraper->motion_type=SCRAPER<TV>::HORIZONTAL_SWEEP;
        scraper->rigid_body=kinematic_rigid_body;
        scraper->Set_Prescribed_Motion_For_Test();

        TV domain_size=TV(paint_box_size.x*(T)10,paint_box_size.y*(T)4,paint_box_size.z*(T)4);TV domain_center=TV(domain_size.x*(T).4,(T)0,(T)0);
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
        Add_Canvas();
    }break;
    case 219:{  ////TEST_EXPOSITION_1
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)10;min_thickness_factor=(T).05/*.02*/;
        max_rim_thickness_factor=(T)10;min_rim_thickness_factor=(T).2/*.05*/;

        ////paint
        T paint_size=(T)8*s0;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV box_min=TV(-paint_box_size.x*(T).5,(T)0,-paint_box_size.z*(T).5);
        TV box_max=TV(paint_box_size.x*(T).5,paint_box_size.y,paint_box_size.z*(T).5);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Randomized_Herring_Bone_Rectangle(initial_mesh,initial_X,box_min,TV2(paint_box_size.z,paint_box_size.x),length_scale,2);

        if(scale>=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh",10,1);
        else if(scale==(T).25)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh_high",10,1);    
        //if(bunny_test==1)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,initial_V,0,"../model/Rigid_Bodies/bunny_mesh_high",268,1);

        initial_mesh.triangle_mesh.elements.Remove_All();
        initial_mesh.segment_mesh.elements.Remove_All();

        ////T y_offset=(T).013;for(int i=1;i<=initial_X.m;i++){initial_X(i).y-=y_offset;}

        ////ground
        TV ground_dir=TV(cos((T)half_pi*(T).5),sin((T)half_pi*(T).5),(T)0);TV ground_center=TV((T)-.4,(T).1,(T)0);
        Add_Canvas((T).6,ground_dir,ground_center);

        ////solids
        Add_Canvas((T).6);  ////slope

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;gravity=TV(gravity.y*sin((T)half_pi*(T).5),gravity.y*cos((T)half_pi*(T).5),(T)0);
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        example.surface_tension_force.viscous=false;
        example.surface_tension_force.contact=false;

        has_friction=true;
        use_correcting_particle_position_on_solids=false;
        use_non_slip_on_solids=false;
        use_levelset=false;
        use_friction_boundary=true;
        use_damped_friction=false;
        use_viscosity_dependent_friction=true;
        
        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        //example.viscosity_force.lambda_thinning=(T)2;
        //example.viscosity_force.alpha_thinning=(T)1;
        //example.viscosity_force.n_thinning=(T).4;

        example.viscosity_force.lambda_thinning=(T).5;
        example.viscosity_force.alpha_thinning=(T)8;
        example.viscosity_force.n_thinning=(T).2;
        if(lambda_cy!=(T)0)example.viscosity_force.lambda_thinning=lambda_cy;
        if(alpha_cy!=(T)0)example.viscosity_force.alpha_thinning=alpha_cy;
        if(n_cy!=(T)0)example.viscosity_force.n_thinning=n_cy;

        ////elasticity
        has_elasticity=true;
        has_plasticity=true;
        use_implicit_elasticity=true;

        ////meshing
        if(bunny_test==3)example.mesher.use_triangle_mesh=false;
        else example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        example.mesher.use_boundary_vertex_snap=true;  ////turn off for exp 1
        example.mesher.use_remove_invalid_elements=true;
        example.mesher.use_clamp_max_thickness=true;
        example.mesher.clamp_triangle_thickness_max=thickness_scale*(T)2;

        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;        

        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV((T).6,(T)1,(T).3);TV domain_center=TV();domain_center.x=-domain_size.x*(T).5+(T).05;
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 220:{  ////TEST_EXPOSITION_2
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).05/*.02*/;
        max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).2/*.05*/;

        ////paint
        T paint_size=(T)16*s0;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV box_min=TV(-paint_box_size.x*(T).5,/*s0*(T)8*/paint_size*(T).01,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,(box_min+box_max)*(T).5,paint_size*(T).5,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Tetrahedralized_Volume_From_File(initial_mesh,initial_X,"../model/Rigid_Bodies/bunny.tet.gz",(T)62.5e-3);
        
        //if(scale>=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh",10,1);
        //else if(scale==(T).25)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh_high",10,1);    

        initial_mesh.triangle_mesh.elements.Remove_All();
        initial_mesh.segment_mesh.elements.Remove_All();

        //T y_offset=s0*(T)8;for(int i=1;i<=initial_X.m;i++){initial_X(i).y+=y_offset;}
        TV init_V=TV::Axis_Vector(2)*(T)-2/*TV()*/;initial_V.Resize(initial_X.m);for(int i=1;i<=initial_V.m;i++){initial_V(i)=init_V;}

        ////solids
        Add_Canvas((T).6);

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_friction=false;
        use_damped_friction=false;
        use_correcting_particle_position_on_solids=false;
        //if(bunny_test!=6)use_non_slip_on_solids=true;   ////false for exp 6: elasticity
        //else use_non_slip_on_solids=false;
        use_non_slip_on_solids=true;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.lambda_thickening=(T).35;
        example.viscosity_force.alpha_thickening=(T)1;
        example.viscosity_force.n_thickening=(T)8;

        ////elasticity
        has_elasticity=true;
        //has_plasticity=true;
        //use_implicit_elasticity=true;

        ////meshing
        //if(bunny_test!=6)remesh=true;   ////for exp 6: elasticity
        //else remesh=false;
        remesh=true;
        if(bunny_test!=4)example.mesher.use_triangle_mesh=false;
        else example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        //if(bunny_test!=6)example.mesher.use_boundary_vertex_snap=true;  ////false for exp 6
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;        

        if(!example.mesher.use_triangle_mesh)initial_mesh.triangle_mesh.elements.Remove_All();
        if(!example.mesher.use_segment_mesh)initial_mesh.segment_mesh.elements.Remove_All();

        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).6);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 228:{
        T s0=(T)5e-3;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).05/*.02*/;
        max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).2/*.05*/;

        ////paint
        T paint_size=(T)16*s0;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV box_min=TV(-paint_box_size.x*(T).5,/*s0*(T)8*/paint_size*(T).01,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        T paint_height=length_scale*(T)1;

        initial_mesh.triangle_mesh.elements.Remove_All();
        initial_mesh.segment_mesh.elements.Remove_All();
        T tube_height=s0*(T)1;
        T radius=paint_size*(T).5;
        TV e1=TV((T)0,s0*(T)96,(T)0);TV dir=TV((T)0,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T)4,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T)4);
        tube->Add_Source_Speed_Bspline_Time_Value((T).03,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();

        //T y_offset=s0*(T)8;for(int i=1;i<=initial_X.m;i++){initial_X(i).y+=y_offset;}
        TV init_V=TV::Axis_Vector(2)*(T)0/*TV()*/;initial_V.Resize(initial_X.m);for(int i=1;i<=initial_V.m;i++){initial_V(i)=init_V;}

        ////solids
        Add_Canvas((T).6);

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_friction=false;
        use_damped_friction=false;
        use_correcting_particle_position_on_solids=false;
        //if(bunny_test!=6)use_non_slip_on_solids=true;   ////false for exp 6: elasticity
        //else use_non_slip_on_solids=false;
        use_non_slip_on_solids=true;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.lambda_thickening=(T).35;
        example.viscosity_force.alpha_thickening=(T)1;
        example.viscosity_force.n_thickening=(T)8;

        ////elasticity
        has_elasticity=true;
        //has_plasticity=true;
        //use_implicit_elasticity=true;

        ////meshing
        //if(bunny_test!=6)remesh=true;   ////for exp 6: elasticity
        //else remesh=false;
        remesh=true;
        if(bunny_test!=4)example.mesher.use_triangle_mesh=false;
        else example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        //if(bunny_test!=6)example.mesher.use_boundary_vertex_snap=true;  ////false for exp 6
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;        

        if(!example.mesher.use_triangle_mesh)initial_mesh.triangle_mesh.elements.Remove_All();
        if(!example.mesher.use_segment_mesh)initial_mesh.segment_mesh.elements.Remove_All();

        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).6);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 229:{
        T s0=(T)5e-4;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).05/*.02*/;
        max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).2/*.05*/;

        ////paint
        T paint_size=(T)16*s0;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV box_min=TV(-paint_box_size.x*(T).5,/*s0*(T)8*/paint_size*(T).01,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        T paint_height=length_scale*(T)1;

        initial_mesh.triangle_mesh.elements.Remove_All();
        initial_mesh.segment_mesh.elements.Remove_All();
        T tube_height=s0*(T)1;
        T radius=paint_size*(T).5;
        TV e1=TV((T)0,s0*(T)96,(T)0);TV dir=TV((T)0,(T)-1,(T)0).Normalized();    ////low
        PAINT_TUBE<TV>* tube=new PAINT_TUBE<TV>(e1,dir,tube_height,paint_height,radius,length_scale,(T).4,(T).1);
        tube->Initialize_Volumetric_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        //tube->Initialize_Filament_Mesh_And_X_And_V(initial_mesh,initial_X,initial_V);tubes.Append(tube);
        Add_Tube_Rigid_Body(*tube);

        tube->Add_Source_Speed_Bspline_Time_Value((T)0,(T).4);
        tube->Add_Source_Speed_Bspline_Time_Value((T).03,(T)0);
        tube->Add_Source_Speed_Bspline_Time_Value((T)1,(T).0);
        tube->Initialize_Source_Speed_Bspline();

        //T y_offset=s0*(T)8;for(int i=1;i<=initial_X.m;i++){initial_X(i).y+=y_offset;}
        TV init_V=TV::Axis_Vector(2)*(T)0/*TV()*/;initial_V.Resize(initial_X.m);for(int i=1;i<=initial_V.m;i++){initial_V(i)=init_V;}

        //////solids
        Add_Canvas((T).6);

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T)1;
        has_friction=false;
        use_damped_friction=false;
        use_correcting_particle_position_on_solids=false;
        //if(bunny_test!=6)use_non_slip_on_solids=true;   ////false for exp 6: elasticity
        //else use_non_slip_on_solids=false;
        use_non_slip_on_solids=true;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.lambda_thickening=(T).35;
        example.viscosity_force.alpha_thickening=(T)1;
        example.viscosity_force.n_thickening=(T)8;

        ////elasticity
        has_elasticity=true;
        //has_plasticity=true;
        //use_implicit_elasticity=true;

        ////meshing
        //if(bunny_test!=6)remesh=true;   ////for exp 6: elasticity
        //else remesh=false;
        remesh=true;
        if(bunny_test!=4)example.mesher.use_triangle_mesh=false;
        else example.mesher.use_triangle_mesh=false;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        //if(bunny_test!=6)example.mesher.use_boundary_vertex_snap=true;  ////false for exp 6
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;        

        if(!example.mesher.use_triangle_mesh)initial_mesh.triangle_mesh.elements.Remove_All();
        if(!example.mesher.use_segment_mesh)initial_mesh.segment_mesh.elements.Remove_All();

        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).6);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    case 230:{  ////TEST_EXPOSITION_2
        T s0=(T)5e-4;length_scale=scale*s0;thickness_scale=s0*(T).5;rim_thickness_scale=thickness_scale;
        max_thickness_factor=(T)2;min_thickness_factor=(T).05/*.02*/;
        max_rim_thickness_factor=(T)4;min_rim_thickness_factor=(T).2/*.05*/;

        ////paint
        T paint_size=(T)16*s0;TV paint_box_size=TV(paint_size,paint_size,paint_size);
        TV box_min=TV(-paint_box_size.x*(T).5,/*s0*(T)8*/paint_size*(T).01,-paint_box_size.z*(T).5);TV box_max=box_min+paint_box_size;
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(initial_mesh,initial_X,box_min,box_max,length_scale);
        CODIMENSIONAL_MESH_GENERATOR<TV>::Sphere(initial_mesh,initial_X,(box_min+box_max)*(T).5,paint_size*(T).5,length_scale);
        //CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Tetrahedralized_Volume_From_File(initial_mesh,initial_X,"../model/Rigid_Bodies/bunny.tet.gz",(T)62.5e-3);

        //if(scale>=(T).5)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh",10,1);
        //else if(scale==(T).25)CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(initial_mesh,initial_X,0,0,0,"../model/Rigid_Bodies/bunny_mesh_high",10,1);    

        initial_mesh.triangle_mesh.elements.Remove_All();
        initial_mesh.segment_mesh.elements.Remove_All();

        //T y_offset=s0*(T)8;for(int i=1;i<=initial_X.m;i++){initial_X(i).y+=y_offset;}
        TV init_V=TV::Axis_Vector(2)*(T)-.2/*TV()*/;initial_V.Resize(initial_X.m);for(int i=1;i<=initial_V.m;i++){initial_V(i)=init_V;}

        ////solids
        Add_Canvas((T).6);

        ////forces
        has_gravity=true;fiber.use_gravity=has_gravity;
        has_incompressibility=true;
        has_surface_tension=true;
        surface_tension*=(T).1;
        has_friction=true;
        use_damped_friction=true;
        use_correcting_particle_position_on_solids=false;
        //if(bunny_test!=6)use_non_slip_on_solids=true;   ////false for exp 6: elasticity
        //else use_non_slip_on_solids=false;
        use_non_slip_on_solids=true;
        use_levelset=false;

        ////viscosity
        has_viscosity=true;
        use_implicit_viscosity=true;
        example.viscosity_force.use_thickness_dependent_variable_viscosity=true;
        example.viscosity_force.lambda_thickening=(T).35;
        example.viscosity_force.alpha_thickening=(T)1;
        example.viscosity_force.n_thickening=(T)8;

        ////elasticity
        has_elasticity=true;
        //has_plasticity=true;
        //use_implicit_elasticity=true;

        ////meshing
        //if(bunny_test!=6)remesh=true;   ////for exp 6: elasticity
        //else remesh=false;
        remesh=true;
        if(bunny_test!=4)example.mesher.use_triangle_mesh=false;
        else example.mesher.use_triangle_mesh=true;
        example.mesher.use_segment_mesh=false;
        example.mesher.use_break_thin_triangles=false;
        example.mesher.use_break_thin_segments=true;
        example.mesher.use_adaptive_tetrahedron_remeshing=false;
        example.mesher.use_adaptive_triangle_remeshing=false;
        example.mesher.use_adaptive_segment_remeshing=false;
        example.mesher.use_continuous_meshing_operations=true;
        //if(bunny_test!=6)example.mesher.use_boundary_vertex_snap=true;  ////false for exp 6
        example.mesher.use_constrained_thin_triangle_break=true;
        example.mesher.use_constrained_thin_segment_break=true;
        use_adaptive_length_scale_on_codim=true;
        adaptive_length_scale_min=length_scale*(T).3;        

        if(!example.mesher.use_triangle_mesh)initial_mesh.triangle_mesh.elements.Remove_All();
        if(!example.mesher.use_segment_mesh)initial_mesh.segment_mesh.elements.Remove_All();

        ////rim
        Use_Rim(true);

        ////mpi
        TV domain_size=TV::Constant_Vector((T).6);TV domain_center=TV();
        global_grid_domain=RANGE<TV>(domain_center-(T).5*domain_size,domain_center+(T).5*domain_size);
    }break;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Prescribed_Object_Motion()
{
    switch(test){
    case 2:{
        fiber.use_prescribed_motion=true;
        FRAME<TV> original_frame=fiber.object_frame;TV x=original_frame.t;TV t;ROTATION<TV> r=original_frame.r;
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T)0,x,r);
        
        t=TV((T).2,(T)0,(T).2);x=original_frame.t+t;r=original_frame.r*ROTATION<TV>((T)half_pi,TV::Axis_Vector(2));
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).1,x,r);

        t=TV((T).4,(T)0,(T).0);x=original_frame.t+t;r=original_frame.r*ROTATION<TV>((T)pi,TV::Axis_Vector(2));
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).2,x,r);

        t=TV((T).6,(T)0,(T)-.2);x=original_frame.t+t;r=original_frame.r*ROTATION<TV>((T)(.5*pi),TV::Axis_Vector(2));
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).3,x,r);

        t=TV((T).8,(T)0,(T).0);x=original_frame.t+t;r=original_frame.r*ROTATION<TV>((T)0,TV::Axis_Vector(2));
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).4,x,r);
    }break;
    case 3:{
        fiber.use_prescribed_motion=true;
        FRAME<TV> original_frame=fiber.object_frame;TV x=original_frame.t;TV t;ROTATION<TV> r=original_frame.r;
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T)0,x,r);

        t=TV((T).15,(T)0,(T).3);x=original_frame.t+t;r=original_frame.r*ROTATION<TV>((T)(.3*pi),TV::Axis_Vector(2));
        fiber.prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).5,x,r);
    }break;
    case 208:{
        PAINT_TUBE<TV>* tube=tubes(1);
        tube->use_prescribed_motion=true;

        FRAME<TV> original_frame=tube->object_frame;TV x=original_frame.t;T y=x.y;ROTATION<TV> r=original_frame.r;TV t=TV();
        tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T)0,x,r);

        x=TV((T)0,y,(T)0);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).1,x,r);
        x=TV((T)23e-3,y,(T)0);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).2,x,r);
        x=TV((T)17e-3,y,(T)-10e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).3,x,r);
        x=TV((T)0e-3,y,(T)-22e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).4,x,r);
        x=TV((T)-14e-3,y,(T)-10e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).5,x,r);
        x=TV((T)0,y,(T)0);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).6,x,r);
        x=TV((T)16e-3,y,(T)13e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).7,x,r);
        x=TV((T)0,y,(T)(T)21e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).8,x,r);
        x=TV((T)-13e-3,y,(T)15e-3);tube->prescribed_motion.Add_Prescribed_Motion_Key_Frame((T).9,x,r);
    }break;
    case 210:{
        T s0=(T)1e-3;
        switch(paint_tube_test){
        case 1001:{     ////S
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/s_motion.txt",r,s0);
        }break;
        case 1003:{     ////G
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/g_motion.txt",r,s0);
        }break;
        case 1004:{     ////G2
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/g2_motion.txt",r,s0);
        }break;
        case 1005:{     ////R
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/r_motion.txt",r,s0);
        }break;
        case 1006:{     ////A
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/a_motion.txt",r,s0);
        }break;
        case 1007:{     ////P
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;            
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/p_motion.txt",r,s0);
        }break;
        case 1008:{     ////H
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;            
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/h_motion.txt",r,s0);
        }break;
        case 1009:{     ////simple example tube 1
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            //tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/sample_motion1.txt",r,s0);
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/exp_motion1.txt",r,s0);
        }break;
        case 1010:{     ////simple example tube 2
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            //tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/sample_motion2.txt",r,s0);
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/exp_motion2.txt",r,s0);
        }break;
        case 1011:{     ////simple example tube 3
            PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
            tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/sample_motion3.txt",r,s0);
        }break;
        }
    }break;
    case 221:{
        //T s0=(T)1e-3;
        //////tube1
        //{PAINT_TUBE<TV>* tube=tubes(1);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
        //tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/spray_motion2.txt",r,s0);}
        //////tube2
        //{PAINT_TUBE<TV>* tube=tubes(2);tube->use_prescribed_motion=true;FRAME<TV> original_frame=tube->object_frame;ROTATION<TV> r=original_frame.r;
        //tube->prescribed_motion.Initialize_Prescribed_Motion_From_File_With_Constant_Rotation("../model/Motion_Data/tube_curve/spray_motion3.txt",r,s0);}
    }break;
    }

    if(fiber.use_prescribed_motion)fiber.prescribed_motion.Initialize_Prescribed_Motion();
    for(int i=1;i<=tubes.m;i++){
        PAINT_TUBE<TV>* tube=tubes(i);
        if(tube&&tube->use_prescribed_motion)tube->prescribed_motion.Initialize_Prescribed_Motion();
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Kinematic_Object_Velocities_And_Positions(const T dt,const T time)
{
    TV v;T_SPIN r;
    switch(test){
    case 1:{r=TV::Axis_Vector(2)*(T)two_pi;}break;
    case 6:{T t=time*(T)10;T s=(T)1;v.x=-s*cos(t);v.z=-s*sin(t);}break;
    case 7:case 41:case 42:{
        T t0=Time_At_Frame(120);
        if(time<t0)v=TV((T).15,(T)-.1,(T)0);
        else{
            T vy=clamp_max((T)-.1+(time-t0)*(T)50,(T)-.01);
            T vx=clamp_min((T).15-(time-t0)*10,(T).1);
            v=TV(vx,vy,(T)0);
        }

        //T vy=clamp_max((T)-.1+time*(T)3,(T)-.01);
        //v=TV((T).1,vy,(T)0);
    }break;
    case 8:{
        if(time<Time_At_Frame(8)){v=(T)-.1*TV::Axis_Vector(2);}
        //else{T t=(time-Time_At_Frame(30))*(T)10;T s=(T).1;v.x=-s*cos(t);v.z=-s*sin(t);}
        else{v=TV::Axis_Vector(1)*(T).1;}
    }break;
    case 11:{v=TV::Axis_Vector(1)*(T).01;}break;
    case 12:{r=TV::Axis_Vector(2)*(T)two_pi;}break;
    case 13:{T t=time*(T)10;T s=(T).4;v.x=-s*cos(t);v.z=-s*sin(t);}break;
    case 26:{v=TV::Axis_Vector(1)*(T).1;}break;
    case 27:{v=TV::Axis_Vector(2)*(T).1;}break;
    case 28:case 29:case 31:{v=TV((T).1,(T)0,(T)0);}break;
    case 30:{v=TV((T)0,(T)-.1,(T)0);}break;
    case 32:case 33:case 34:case 35:case 36:case 37:case 38:case 40:case 44:case 48:{v=TV((T).1,(T)-.01,(T)0);}break;
    case 39:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);
        if(kinematic_rigid_body->X().y>=(T).0041) v=TV((T)0,(T)-.1,(T)0);
        else v=TV((T).1,(T)0,(T)0);
    }break;
    case 43:case 45:case 49:case 81:case 90:{
        //T t0=(T).18;T s=(T)3;
        //if(time<t0){v.x=(T).15;v.y=(T)-.1;}
        //else {v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
        //v.x=(T).15*cos(s*(time-t0));v.z=(T)(T).15*sin(s*(time-t0));
        //r=TV::Axis_Vector(2)*(T)-.0005+TV::Axis_Vector(1)*(T).0005;}

        T t0=(T).12;T s=(T)1;
        if(time<t0){v.x=(T).15;v.y=(T)-.1;}
        else {v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
        v.x=(T).15*cos(s*(time-t0));v.z=(T)(T).15*sin(s*(time-t0));
        r=TV::Axis_Vector(2)*(T)-.2+TV::Axis_Vector(1)*(T).2;}

        if(time<(T).35)fiber.Span_Fibers((T)1+time*(T)2);
    }break;
    case 82:{
        T t0=(T).12;T t1=(T).2;T t2=(T).3;
        //T s=(T)1;
        if(time<t0){
            v.x=(T).15;v.y=(T)-.1;
            r=TV::Axis_Vector(2)*(T)-.2+TV::Axis_Vector(1)*(T).2;
        }
        else if(time<t1){
            v.x=(T).15;
            v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
            r=TV::Axis_Vector(2)*(T)-.2+TV::Axis_Vector(1)*(T).2;
        }
        else if(time<t2){
            v.x=clamp_min((T).15-(time-t1)*(T)2,(T)-.2);
            v.y=clamp_max((T)-.01+(time-t1)*(T)1.2,(T).1);
            r=TV::Axis_Vector(3)*(T)15;
        }
        else{
            v.x=clamp_min((T).15-(time-t1)*(T)2,(T)-.2);
            v.y=clamp_min((T).1-(time-t2)*(T)1,(T)0);
            r=TV();
        }
        //v.x=(T).15*cos(s*(time-t0));v.z=(T)(T).15*sin(s*(time-t0));
    }break;
    case 83:{
        //T t0=(T).12;
        //if(time<t0)v=TV::Axis_Vector(2)*(T).1;
        //else v=TV::Axis_Vector(1)*(T)-.1;

        T t0=(T).15;T t1=(T).48;T t3=(T).65;
        if(time<t0) v=TV((T).1,(T)-.08,(T)0);
        else if(time<t1){
            T v_y=(T).2-time*(T).2;
            v_y=clamp_min((T).04,v_y);
            v=TV((T).08,v_y,(T)0);
        }
        else if(time<t3){
            v=TV((T).2,(T)-.2,(T).04);
            for(int i=1;i<=fiber.fiber_particles.X.m;i++){
                fiber.fiber_particles.V(i)=v;
            }
        }
        else{
            v=TV((T).15,(T)-.05,(T).04);
            static bool stop=false;
            if(!stop){
                for(int i=1;i<=fiber.fiber_particles.X.m;i++){
                    fiber.fiber_particles.V(i)=v;
                }
                stop=true;
            }
        }

        //v=TV::Axis_Vector(2)*(T).1;
        if(time<(T).16)fiber.Span_Fibers((T)1+time*(T)1.5);
    }break;
    case 50:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);
        if(time<(T).015)kinematic_rigid_body->V()=TV((T)0,(T)-.1,(T)0);
        else kinematic_rigid_body->V()=TV((T).1,(T)0,(T)0);
    }break;
    case 46:{v=TV((T).1,(T).1,(T)0);}break;
    case 47:{
        if(time<(T).06)v=TV::Axis_Vector(2)*(T)-.1;
        else v=TV::Axis_Vector(2)*(T).1;
    }break;
    case 57:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);
        kinematic_rigid_body->V()=TV((T).1,(T).2,(T)0);
    }break;
    case 51:{
        if(time>(T).2){RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)500.);}
        else{RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)0);}
    }break;
    case 52:{
        if(time>(T).05){RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)500.);}
        else{RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)0);}
    }break;
    case 53:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)100000.);
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T)0,(T)0);
    }break;
    case 54:{
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)100.);
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T)0,(T)0,(T).1);
    }break;
    case 55:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)100.);
    }break;
    case 133:{
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)100.);
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T)0,(T).1,(T)0);
    }break;
    case 205:{  ////TEST_CHEESE
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).05,(T)0);
        //T t0=(T).05;
        //if(time<=t0){T frac=(T)1-time/t0;
        //    kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)-5*frac,0,(T)-10*frac);
        //}
        //else kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)0);

        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).05,(T)0)/time_scale;
        //T t0=(T).1*time_scale;

        //if(time<(T).2){RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).05,(T)0)/time_scale;}
        //else {RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).005,(T)0)/time_scale;}
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).05,(T)0)/time_scale;
        
        //T t0=(T).1*time_scale;

        //if(time<t0){kinematic_rigid_body->V()=TV((T)0,(T).05,(T)0)/time_scale;}
        //else{kinematic_rigid_body->V()=TV((T).2,(T).03,(T)0)/time_scale;}

        //if(time<=t0){T frac=(T)1-time/t0;T s=(T).5;
        //kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)-5*frac,0,(T)-10*frac)*s/time_scale;
        //}
        //else kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)0);
    }break;
    case 206:{  ////TEST_CHEESE_ON_BREAD
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV((T).1,(T).2,(T)0);
    }break;
    case 211:{  ////TEST_PAINT_TRACE
        ////motion 1
        if(paint_trace_test==1){
            T t0=(T).18;T s=(T)3;
            if(time<t0){v.x=(T).15;v.y=(T)-.1;}
            else{
                v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
                v.x=(T).15*cos(s*(time-t0));v.z=(T).15*sin(s*(time-t0));
                r=TV::Axis_Vector(2)*(T)-.5+TV::Axis_Vector(1)*(T).5;
            }
            if(time<(T).35)fiber.Span_Fibers((T)1+time*(T)1.5);
        }
        ////motion 2
        else if(paint_trace_test==2){
            T t0=(T).18;T s=(T)1.5;
            if(time<t0){v.x=(T).15;v.y=(T)-.1;}
            else{
                v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
                v.x=(T).15*cos(s*(time-t0));v.z=(T)(T)-.15*sin(s*(time-t0));
                r=TV::Axis_Vector(2)*(T)-.01+TV::Axis_Vector(1)*(T).01;
            }
            if(time<(T).3)fiber.Span_Fibers((T)1+time*(T)1.2);
        }

        //else fiber.Use_Virtual_Springs(false);
    }break;
    case 212:{  ////TEST_PAINT_DIP
        T t0=(T).18;T s=(T)3;
        if(time<t0){v.x=(T)-.15;v.y=(T)-.1;}
        else{
            v.y=clamp_max((T)-.1+(time-t0)*(T)5,(T)-.01);
            v.x=(T)-.15*cos(s*(time-t0));v.z=(T)(T).15*sin(s*(time-t0));
            r=TV::Axis_Vector(2)*(T)-.5+TV::Axis_Vector(1)*(T).5;
        }

        if(time<(T).35)fiber.Span_Fibers((T)1+time*(T)2);
    }break;
    case 214:{  ////TEST_PAINT_SPLASH
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->V()=TV::Axis_Vector(2)*((T)-2+(T)-9.8*time);
    }break;
    case 217:{  ////TEST_MATERIALS_2
        v=TV((T).1,(T)-.01,(T)0);
    }break;
    case 218:{  ////TEST_PAINT_SPRAY2
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)500.);
    }break;
    case 221:{  ////TEST_PAINT_SPRAY3
        //if(time>(T).2){RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)500.);}
        //else{RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)0);}
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,(T)20);
    }break;
    case 222:{
        RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);kinematic_rigid_body->Angular_Velocity()=T_SPIN((T)0,(T)0,cylinder_angular_velocity);
    }break;
    }

    //////update fibers
    if(fiber.use_prescribed_motion){
        fiber.object_frame=fiber.prescribed_motion.Prescribed_Frame(time);
        fiber.object_twist=fiber.prescribed_motion.Prescribed_Twist(dt,time);
    }
    else{
        fiber.object_twist.linear=v;fiber.object_twist.angular=r;
        fiber.object_frame.t+=fiber.object_twist.linear*dt;
        fiber.object_frame.r*=ROTATION<TV>::From_Rotation_Vector(fiber.object_twist.angular*dt);
    }

    ////update kinematic rigid body
    if(kinematic_rigid_bodies.m>0){
        for(int i=1;i<=kinematic_rigid_bodies.m;i++){
            RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(i);
            kinematic_rigid_body->X()+=kinematic_rigid_body->V()*dt;
            kinematic_rigid_body->Rotation()*=ROTATION<TV>::From_Rotation_Vector(kinematic_rigid_body->Angular_Velocity()*dt);
        }
    }

    ////correct kinematic body position
    switch(test){
    case 214:{
        //RIGID_BODY<TV>* kinematic_rigid_body=kinematic_rigid_bodies(1);
        ////T sphere_radius=(T).1;T height_min=sphere_radius*(T)1.2;
        //T height_min=(T)0.08;
        //kinematic_rigid_body->X().y=clamp_min(height_min,kinematic_rigid_body->X().y);
    }break;
    }

    ////update specialized rigid bodies
    //if(test==292&&knife){knife->Update_Kinematic_Velocities_And_Positions(dt,time+Time_At_Frame(175));}
    if(test==292&&knife){knife->Update_Kinematic_Velocities_And_Positions(dt,time+Time_At_Frame(350));}
    else if(knife){knife->Update_Kinematic_Velocities_And_Positions(dt,time);}
    //if(knife){knife->Update_Kinematic_Velocities_And_Positions(dt,time);}
    if(ladle){ladle->Update_Kinematic_Velocities_And_Positions(dt,time+Time_At_Frame(70));}
    if(rock){rock->Update_Kinematic_Velocities_And_Positions(dt,time);}
    if(scraper){scraper->Update_Kinematic_Velocities_And_Positions(dt,time);}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Kinematic_Paint_Velocities_And_Positions(const T dt,const T time)
{
    TV v;T_SPIN r;ARRAY<TV> vs;ARRAY<T_SPIN> rs;T scaled_time=time*time_scale;
    switch(test){
#if !defined(_WIN32) || !defined(ACCEL_WINDOWS_COMPILING)
    case 20:{if(time>Time_At_Frame(50)){v=TV::Axis_Vector(2)*(T).1+TV::Axis_Vector(1)*(T)-.02;tubes(1)->source_speed=(T).0;}}break;
    case 21:{if(time>Time_At_Frame(40)){v=TV::Axis_Vector(2)*(T)-.01+TV::Axis_Vector(1)*(T)-.05;tubes(1)->source_speed=(T).01;}}break;
    case 22:{v=TV::Axis_Vector(2)*(T).05+TV::Axis_Vector(1)*(T)-.1;tubes(1)->source_speed=(T).1;}break;
    case 124:{v=TV((T).1,(T).02,(T)0);tubes(1)->source_speed=(T).01;}break;
    case 125:{v=TV((T).1,(T).05,(T)0);tubes(1)->source_speed=(T).0;}break;
    case 127:{v=TV((T).1,(T).05,(T)0);tubes(1)->source_speed=(T).0;tubes(1)->source_speed=(T).1;}break;
    
    case 286:{
        if(time>Time_At_Frame(1450)){
            tubes(1)->Remove_All_Eulerian_Particles();
        }
    }break;
    case 291:{
        if(time>Time_At_Frame(1350)){
            tubes(1)->Remove_All_Eulerian_Particles();
        }
    }break;
    case 685:
    case 670:{
        T phase=(T)10*(T)two_pi*time+(T)half_pi;
        T s=sin(phase+(T)pi)*(T).3;
        v=TV((T)-.1,(T)0,s);
        TV axis=TV::Axis_Vector(1);r=axis*(T).005*sin(phase);
        if(time>(T).28){v=TV((T)-.1,(T)0,(T)0);r=TV();}
    }break;
    case 671:{
        T phase=(T)10*(T)two_pi*time+(T)half_pi;
        T s=sin(phase+(T)pi)*(T).3;
        v=TV((T)-.1,(T)0,s);
        TV axis=TV::Axis_Vector(1);r=axis*(T).005*sin(phase);
        if(time>(T).28){v=TV((T)-.1,(T)0,(T)0);r=TV();}
    }break;
    case 673:{
        if(time<Time_At_Frame(50))
        {
            v=TV((T)-.1,(T)0,(T)0);r=TV();
        }

        if(time>Time_At_Frame(50)&&time<Time_At_Frame(100))
        {
            T new_time=time-Time_At_Frame(50);
            T phase=(T)10*(T)two_pi*new_time+(T)half_pi;
            T s=sin(phase+(T)pi)*(T).15;
            v=TV((T)-.1,(T)0,s);
            //TV axis=TV::Axis_Vector(1);r=axis*(T).005*sin(phase);
            //if(new_time>(T).28){v=TV((T)-.1,(T)0,(T)0);r=TV();}
            r=TV();
        }
    }break;
    case 683:
    case 682:
    case 679:
    case 678:
    case 675:
    case 674:{
        if(time<Time_At_Frame(50))
        {
           v=TV((T)-.1,(T)0,(T)0);r=TV();
        }
        else if(time<Time_At_Frame(100))
        {
            T amp=(T).5*(T).25;
            T disp=1;
            T freq=(time-Time_At_Frame(50))*(T)two_pi/(Time_At_Frame(100)-Time_At_Frame(50));
            T phase=(T)pi;
            v=TV((T)-.1,-(cos(freq+phase)+disp)*amp,(cos(freq+phase)+disp)*amp);
            r=-TV::Axis_Vector(3)*T(10)*(time-Time_At_Frame(50))/(Time_At_Frame(100)-Time_At_Frame(50)); // TODO: This is added
        }
        else
        {
//            v=TV(-.5,0,0);
            T new_time=time-Time_At_Frame(100);
            T phase=(T)10*(T)two_pi*new_time+(T)half_pi;
            T s=sin(phase+(T)pi)*(T).3;
            v=TV(-.5,0,-s);
            if(time<Time_At_Frame(125))
            {
                r=-TV::Axis_Vector(3)*T(10)*(time-Time_At_Frame(50))/(Time_At_Frame(100)-Time_At_Frame(50)); // TODO: This is added
            }
            else
            {
                r=TV();
            }
        }
        /*
        else if(time<Time_At_Frame(60))
        {
            T amp=.5;
            T disp=1;
            T freq=(time-Time_At_Frame(50))*two_pi/(Time_At_Frame(60)-Time_At_Frame(50))/2;
            T phase=pi;
            v=TV(-.1,0,(cos(freq+phase)+disp)*amp);
            r=TV();
        }
        else if(time<Time_At_Frame(70))
        {
            T amp=1;
            T disp=0;
            T freq=(time-Time_At_Frame(60))*two_pi/(Time_At_Frame(70)-Time_At_Frame(60))/2;
            T phase=0;
            v=TV(-.1,0,(cos(freq+phase)+disp)*amp);
            r=TV();
        }
        else if(time<Time_At_Frame(80))
        {
            T amp=.5;
            T disp=-1;
            T freq=(time-Time_At_Frame(70))*two_pi/(Time_At_Frame(80)-Time_At_Frame(70))/2;
            T phase=pi;
            v=TV(-.1,0,(cos(freq+phase)+disp)*amp);
            r=TV();
        }*/
    }break;
    case 672:{
        T speed=80;
        if(time>=Time_At_Frame(10)&&time<Time_At_Frame(40))
        {
            T new_time=time-Time_At_Frame(10);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed;
            v=TV(s,0,0);r=TV();
        }
        else if(time>=Time_At_Frame(40)&&time<Time_At_Frame(80))
        {
            T new_time=time-Time_At_Frame(40);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed;
            v=TV(-s,0,0);r=TV();
        }
        else if(time>=Time_At_Frame(80)&&time<Time_At_Frame(120))
        {
            T new_time=time-Time_At_Frame(80);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed;
            v=TV(s,0,0);r=TV();
        }
        else if(time>=Time_At_Frame(120)&&time<Time_At_Frame(160))
        {
            T new_time=time-Time_At_Frame(120);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed;
            v=TV(-s,0,0);r=TV();
        }
        else if(time>=Time_At_Frame(160)&&time<Time_At_Frame(200))
        {
            T new_time=time-Time_At_Frame(160);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed;
            v=TV(s,0,0);r=TV();
        }
    }break;
    case 681:{
        T speed1=(T)20;
        T speed2=(T)20;

        int phase1=200;
        int phase2=210;
        int phase3=220;
        int phase4=230;
        int phase5=240;
        int phase6=250;

        if(time>=Time_At_Frame(phase1)&&time<Time_At_Frame(phase2))
        {
            T new_time=time-Time_At_Frame(phase1);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed1;s=s;
//            v=TV(0,s,s);r=TV();
            //v=TV(0,s,0);r=TV();
            r=TV::Axis_Vector(1)*T(50)*-new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));
        }
        else if(time>=Time_At_Frame(phase2)&&time<Time_At_Frame(phase3))
        {
            T new_time=time-Time_At_Frame(phase2);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed2;s=s;
//            v=TV(0,-s,-s);r=TV();
            //v=TV(0,-s,0);r=TV();
            r=TV::Axis_Vector(3)*T(40)*-new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));

        }
        else if(time>=Time_At_Frame(phase3)&&time<Time_At_Frame(phase4))
        {
            T new_time=time-Time_At_Frame(phase3);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed1;s=s;
//            v=TV(0,s,s);r=TV();
            //v=TV(0,s,0);r=TV();
            r=TV::Axis_Vector(1)*T(40)*new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));

        }
        else if(time>=Time_At_Frame(phase4)&&time<Time_At_Frame(phase5))
        {
            T new_time=time-Time_At_Frame(phase4);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed2;s=s;
//            v=TV(0,-s,-s);r=TV();
            //v=TV(0,-s,0);r=TV();
            r=TV::Axis_Vector(1)*T(20)*-new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));
            r=TV::Axis_Vector(3)*T(20)*-new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));
        }
        else if(time>=Time_At_Frame(phase5)&&time<Time_At_Frame(phase6))
        {
            T new_time=time-Time_At_Frame(phase5);
            //T s=sin(phase*T(5)+(T)pi)*(T)2;
            T s=new_time*speed1;s=s;
//            v=TV(0,s,s);r=TV();
            //v=TV(0,s,0);r=TV();
            r=TV::Axis_Vector(2)*T(40)*new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));
            r=TV::Axis_Vector(3)*T(40)*new_time/(Time_At_Frame(phase2)-Time_At_Frame(phase1));
        }
    }break;
    case 686:{
        v=TV(0,-.1,0);
        r=TV();
    }break;
#endif
    case 202:{
        if(time<Time_At_Frame(10))
        {
            r=TV::Axis_Vector(3)*T(-.04);
        }
        else
        {
            T new_time=time-Time_At_Frame(10);
            T phase=(T)10*(T)two_pi*new_time+(T)half_pi;
            T s=sin(phase+(T)pi)*(T).15;
            v=TV((T)-.1,(T)0,s);
            TV axis=TV::Axis_Vector(1);r=axis*(T).005*sin(phase);
            if(new_time>(T).28){v=TV((T)-.1,(T)0,(T)0);r=TV();}
        }
    }break;
    case 130:{v=TV((T).1,(T).05,(T)0);}break;
    case 206:{
        T t0=(T).05;T c1=(T)1e4;
        if(time<t0)environment_temperature=c1*time;
        else{
            environment_temperature=c1*t0-(T)3*c1*(time-t0);
            environment_temperature=clamp_min(environment_temperature,(T)0);
        }
        LOG::cout<<"temperature: "<<environment_temperature<<std::endl;
    }break;
    case 207:{  ////TEST_CHEESE_HEATING
        T t0=(T).15;
        if(time<t0)environment_temperature=(T)11000*time;
        else environment_temperature=(T)0;
    }break;
    case 210:{
        switch(paint_tube_test){
        case 1001:{
            //T t0=(T).34;
            //if(time>t0){tubes(1)->eulerian_particles.Remove_All();}
        }break;
        case 1002:{     ////I
            T t0=(T).1;static T phase1=(T)0;

            if(time<t0){
                T phase=(T)10*(T)two_pi*time+(T)pi;
                T s=sin(phase+(T)pi)*(T).01;
                v=TV(s,(T)0,(T).2);
                TV axis=TV::Axis_Vector(1);r=axis*(T).01*(T)two_pi*sin(phase);
                phase1=phase;
            }
            else{
                T phase=(T)5*(T)two_pi*(time-t0)+phase1;
                T s=sin(phase+(T)pi)*(T).1;
                v=TV(s,(T)-.1,(T).2);
                if(time>(T).134)v.y=(T)0;
                //TV axis=TV::Axis_Vector(1);r=axis*(T)2*(T)two_pi*sin(phase);
            }
            
            if(time>(T).374){v.y=(T).05;}   
        }break;
        case 1:{    ////ssss
            T phase=(T)10*(T)two_pi*time+(T)half_pi;
            //T s=sin(phase+(T)pi)>(T)0?(T).15:(T)-.15;
            T s=sin(phase+(T)pi)*(T).15;
            v=TV((T)-.1,(T)0,s);
            TV axis=TV::Axis_Vector(1);r=axis*(T).005*sin(phase);
            if(time>(T).28){v=TV((T)-.1,(T)0,(T)0);r=TV();}
        }break;
        case 2:{    ////c
            T t0=(T).08;T t1=(T).14;T t2=(T).15;
            T f0=(T)8;T f1=(T)4;
            if(scaled_time<t0){
                T phase=f0*(T)two_pi*scaled_time+(T)3*(T)half_pi;T s=(T).4;
                v=TV(sin(phase)*s,(T)0,cos(phase)*s)*time_scale;
                TV axis=TV((T)1,(T).5,(T)0).Normalized();r=axis*f0*(T)two_pi*time_scale;
            }
            else if(scaled_time<t1){
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(scaled_time-t0)+(T)3*(T)half_pi;T s=(T).8;
                v=TV(sin(phase)*s,(T)-.05*s,cos(phase)*s)*time_scale;
                TV axis=TV((T)1,(T).5,(T)0).Normalized();r=axis*f1*(T)two_pi*time_scale;
                //tubes(1)->eulerian_particles.Remove_All();
            }
            else if(scaled_time<t2){
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(scaled_time-t0)+(T)3*(T)half_pi;T s=(T).4;
                T slow=(T)1;//T slow=clamp_min((T)1-(time-t1)*40,(T)0);
                v=TV(sin(phase)*s*slow,(T).0,cos(phase)*s*slow)*time_scale;
                TV axis=TV((T)0,(T)0,(T)1).Normalized();r=axis*(T)-15*time_scale;
            }
            else{
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(t2-t0)+(T)3*(T)half_pi;T s=(T).4;
                T slow=(T)1;//T slow=clamp_min((T)1-(t2-t1)*40,(T)0);
                v=TV(sin(phase)*s*slow,(T).0,cos(phase)*s*slow)*time_scale;
                TV axis=TV((T)0,(T)0,(T)1).Normalized();r=axis*(T)0*time_scale;
            }
        }break;
        case 14:{   ////minor c
            T t0=(T).08;T t1=(T).14;T t2=(T).15;
            T f0=(T)8;T f1=(T)2;
            if(scaled_time<t0){
                T phase=f0*(T)two_pi*scaled_time+(T)3*(T)half_pi;T s=(T).4;
                v=TV(sin(phase)*s,(T)0,cos(phase)*s)*time_scale;
                TV axis=TV((T)1,(T).5,(T)0).Normalized();r=axis*f0*(T)two_pi*time_scale;
            }
            else if(scaled_time<t1){
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(scaled_time-t0)+(T)3*(T)half_pi;T s=(T).4;
                v=TV(sin(phase)*s,(T)0,cos(phase)*s)*time_scale;
                TV axis=TV((T)1,(T).5,(T)0).Normalized();r=axis*f1*(T)two_pi*time_scale;
                //tubes(1)->eulerian_particles.Remove_All();
            }
            else if(scaled_time<t2){
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(scaled_time-t0)+(T)3*(T)half_pi;T s=(T).4;
                T slow=(T)1;//T slow=clamp_min((T)1-(time-t1)*40,(T)0);
                v=TV(sin(phase)*s*slow,(T).0,cos(phase)*s*slow)*time_scale;
                TV axis=TV((T)0,(T)0,(T)1).Normalized();r=axis*(T)-15*time_scale;
            }
            else{
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(t2-t0)+(T)3*(T)half_pi;T s=(T).4;
                T slow=(T)1;//T slow=clamp_min((T)1-(t2-t1)*40,(T)0);
                v=TV(sin(phase)*s*slow,(T).0,cos(phase)*s*slow)*time_scale;
                TV axis=TV((T)0,(T)0,(T)1).Normalized();r=axis*(T)0*time_scale;
            }
        }break;
        case 20:{   ////line
            T t0=(T).1;static T phase1=(T)0;

            if(time<t0){
                T phase=(T)10*(T)two_pi*time+(T)pi;
                T s=sin(phase+(T)pi)*(T).1;
                v=TV(s,(T)0,(T).2);
                TV axis=TV::Axis_Vector(1);r=axis*(T).5*(T)two_pi*sin(phase);
                phase1=phase;
            }
            else{
                T phase=(T)5*(T)two_pi*(time-t0)+phase1;
                T s=sin(phase+(T)pi)*(T).1;
                v=TV(s,(T)-.1,(T).2);
                if(time>(T).134)v.y=(T)0;
                //TV axis=TV::Axis_Vector(1);r=axis*(T)2*(T)two_pi*sin(phase);
            }
        }break;
        case 3:{    ////alpha
            T t0=(T).14;//T t1=(T).2;
            T f0=(T)8;T f1=(T)2;
            if(time<t0){
                T phase=f0*(T)two_pi*time+(T)3*(T)half_pi;T s=(T).4;
                TV axis=TV((T).5,(T)1,(T)0).Normalized();r=axis*f0*(T)two_pi;
                T slow=clamp_min((T)1-time*1,(T)0);
                v=TV(sin(phase)*s*slow,(T)0,cos(phase)*s*slow);
            }
            else{
                T phase=f0*(T)two_pi*t0+f1*(T)two_pi*(time-t0)+(T)3*(T)half_pi;T s=(T).4;
                TV axis=TV((T).5,(T)1,(T)0).Normalized();r=axis*f1*(T)two_pi;
                v=TV(sin(phase)*s,(T)0,cos(phase)*s);
            }
        }break;
        case 10:{   ////s
            T t0=(T).047;T t1=(T).18;static T phase1=(T)0;
            if(time>=t0){tubes(1)->Enable(true,&example.particles);}

            if(time<t1){
                T phase=(T)10*(T)two_pi*time+(T)pi;
                T s=sin(phase+(T)pi)*(T).3;
                v=TV((T)-.2,(T)0,s);
                TV axis=TV::Axis_Vector(1);r=axis*(T)2*(T)two_pi*sin(phase);
                phase1=phase;
            }
            else{
                T phase=(T)5*(T)two_pi*(time-t1)+phase1;
                T s=sin(phase+(T)pi)*(T).1;
                v=TV((T)-.4,(T)0,s);
                TV axis=TV::Axis_Vector(1);r=axis*(T)1*(T)two_pi*sin(phase);
            }
        }break;
        case 11:{   ////line
            T t0=(T).1;T t1=(T).26;T t1_half=(t1-t0)*(T).75+t0;T t2=(T).31;
            if(time<t0){v=TV();r=TV();}
            else if(time<t1){
                if(time<t1_half)v=TV((T)-.1,(T).025,(T)0);
                else v=TV((T)-.1,(T)0,(T)0);
                TV axis=TV::Axis_Vector(3);r=axis*(T)-1*(T)two_pi;
            }
            else if(time<t2){
                v=TV((T)-.1,(T)-.005,(T)0);r=TV();
            }
            else{
                v=TV((T)-.1,(T)0,(T)0);r=TV();
            }
        }break;
        case 12:{   ////sqz tail
            T t0=(T).1;T t1=(T).2;T t2=(T).3;
            if(time<t0){v=TV();r=TV();}
            else if(time<t1){
                v=TV((T)0,(T)-.013,(T)0);
                TV axis=TV((T).2,(T)0,(T)1).Normalized();r=axis*(T)-1.5*(T)two_pi;
                tubes(1)->eulerian_particles.Remove_All();
            }
            else if(time<t2){
                v=TV((T)-.1,(T)0,(T)0);r=TV();
            }
            else{
                v=TV((T)-.1,(T)0,(T)0);r=TV();
            }
        }break;
        case 13:{   ////pinch off
            T t0=(T).11;
            if(time<t0){v=TV();r=TV();}
            else{
                v=TV((T)-.01,(T).01,(T)0);
                TV axis=TV((T).2,(T)0,(T)1).Normalized();r=axis*(T)-.1*(T)two_pi;
                tubes(1)->eulerian_particles.Remove_All();
            }
        }break;
        case 4:{    ////coiling
            T phase=(T)10*(T)two_pi*time+(T)half_pi;T s=(T).001/*-time*(T).12*/;
            v=TV(sin(phase)*s,(T)0,cos(phase)*s);
        }break;
        case 5:{    ////folding
            T phase=(T)2*(T)two_pi*time+(T)half_pi;T s=(T).2/*-time*(T).12*/;
            v=TV(sin(phase)*s,(T)0,(T)0);
               }break;
        case 6:{    ////random coiling
            T phase=(T)10*(T)two_pi*time+(T)half_pi;//T s=(T).2/*-time*(T).12*/;
            //T s=(T).4*random.Get_Uniform_Number((T)1,(T)3);
            T s=(T)1-time*(T).5;s=clamp_min(s,(T).6);
            v=(T).5*v+(T).5*TV(sin(phase)*s,(T)0,cos(phase)*s);
            if(time>(T).3){
                v.y=(T)-.1;v.x=(T).2;v.z=(T)0;
                if(time>(T).375)v.y=(T)0;
            }
            else{
                if(time>(T).1){v.y=(T)-.25*(time-(T).1);}
            }
        }break;
        case 7:{
            T t0=(T).2;
            vs.Resize(2);rs.Resize(2);
            static int count=0;
            if(time<t0){count++;if(count>=2)tubes(2)->Enable(false,&example.particles);}
            else tubes(2)->Enable(true,&example.particles);

            T phase=(T)10*(T)two_pi*time+(T)half_pi;
            T s=sin(phase+(T)pi)*(T).3;
            vs(1)=TV((T)-.1,(T)0,s);
            TV axis=TV::Axis_Vector(1);rs(1)=axis*(T)2*(T)two_pi*sin(phase);

            if(time<t0){
                vs(2)=TV();rs(2)=TV();
            }
            else{
                //T phase=(T)5*(T)two_pi*time+(T)pi;
                //T s=sin(phase+(T)pi)*(T).1;
                //vs(2)=TV((T)s,(T)0,(T).1);
                //rs(2)=TV();
                //TV axis=TV::Axis_Vector(2);rs(2)=axis*(T)5*(T)two_pi*sin(phase);

                T phase=(T)10*(T)two_pi*time+(T)half_pi;T s=(T).2+time*(T).1;
                vs(2)=TV(sin(phase)*s+(T).05,(T)0,cos(phase)*s);
            }
        }break;
        }

        //////pinch off
        //if(time>(T).1)v=TV((T)-.1,(T)0,(T)0);else v=TV();
    }break;
    case 215:{
        //T phase=(T)10*(T)two_pi*time+(T)half_pi;
        //TV axis=TV::Axis_Vector(3);r=axis*(T)two_pi*(T)2*sin(phase);

        T t0=(T).35;T t1=(T).5;T t2=(T).8;T t3=(T)1.05;T t4=(T)2;
        T trans_t1=(T).4;T trans_t2=(T).95;

        ////static T freq_x0=(T)0;static T phase_x0=(T)0;static T s_x0=(T)0;
        ////static T freq_x1=(T)0;static T phase_x1=(T)0;static T s_x1=(T)0;
        ////static T freq_x2=(T)0;static T phase_x2=(T)0;static T s_x2=(T)0;
        ////static T freq_x3=(T)0;static T phase_x3=(T)0;static T s_x3=(T)0;
        ////static T freq_x4=(T)0;static T phase_x4=(T)0;static T s_x4=(T)0;

        T freq_x0=(T)0;T phase_x0=(T)0;T s_x0=(T)0;
        T freq_x1=(T)0;T phase_x1=(T)0;T s_x1=(T)0;
        T freq_x2=(T)0;T phase_x2=(T)0;T s_x2=(T)0;
        T freq_x3=(T)0;T phase_x3=(T)0;T s_x3=(T)0;
        T freq_x4=(T)0;T phase_x4=(T)0;T s_x4=(T)0;
        {    ////speed up for tets
            T time=t0;
            T f1=8;T f2=17;T f3=15;T s1=(T).5;T s2=(T)15;T s3=(T)3.2;
            //T f1=8;T f2=16;T f3=12;T s1=(T).5;T s2=(T)12;T s3=(T)3.2;
            T freq_x=f1+f2*time;
            freq_x=clamp_max(freq_x,f3);
            T phase_x=freq_x*(T)two_pi*time+(T)half_pi;
            T s_x=s1+s2*time;s_x=clamp_max(s_x,s3);
            freq_x0=freq_x;phase_x0=phase_x;s_x0=s_x;
        }
        {   ////slow down
            T time=t1;
            T f1=(T)15;T f2=(T)0;T s1=(T)15;T s2=(T)0;
            T freq_x=freq_x0-f1*(time-t0);freq_x=clamp_min(freq_x,f2);
            T phase_x=phase_x0+(T)freq_x*(T)two_pi*(time-t0);
            T s_x=s_x0-s1*(time-t0);s_x=clamp_min(s_x,s2);
            freq_x1=freq_x;phase_x1=phase_x;s_x1=s_x;
        }
        {   ////speed up for tris
            T time=t2;
            T f1=(T)20;T f2=(T)3;T s1=(T)4;T s2=(T)1.5;
            T freq_x=freq_x1+f1*(time-t1);
            freq_x=clamp_max(freq_x,f2);
            T phase_x=(T)half_pi+/*phase_x1+*/freq_x*(T)two_pi*(time-t1);
            T s_x=s_x1+s1*(time-t1);
            s_x=clamp_max(s_x,s2);
            freq_x2=freq_x;phase_x2=phase_x;s_x2=s_x;
        }
        {   ////slow down a little bit
            T time=t3;
            T f1=(T)2;T f2=(T)0;T s1=(T)2;T s2=(T)0;
            T freq_x=freq_x2-f1*(time-t2);freq_x=clamp_min(freq_x,f2);
            T phase_x=phase_x2+(T)freq_x*(T)two_pi*(time-t2);
            T s_x=s_x2-s1*(time-t2);s_x=clamp_min(s_x,s2);
            freq_x3=freq_x;phase_x3=phase_x;s_x3=s_x;
        }
        {   ////speed up for segs
            T time=t4;
            T f1=(T)8;T f2=(T)10;T s1=(T)2;T s2=(T)4;
            T freq_x=freq_x3+f1*(time-t3);
            freq_x=clamp_max(freq_x,f2);
            T phase_x=phase_x3+freq_x*(T)two_pi*(time-t3);
            T s_x=s_x3+s1*(time-t3);s_x=clamp_max(s_x,s2);
            freq_x4=freq_x;phase_x4=phase_x;s_x4=s_x;
        }

        //static T object_space_r_init=tubes(1)->object_space_r;

        if(time<t0){    ////speed up for tets
            T f1=8;T f2=17;T f3=15;T s1=(T).5;T s2=(T)15;T s3=(T)3.2;
            //T f1=8;T f2=20;T f3=16;T s1=(T).5;T s2=(T)10;T s3=(T)3;
            //T f1=8;T f2=16;T f3=12;T s1=(T).5;T s2=(T)12;T s3=(T)3.2;
            T freq_x=f1+f2*time;
            freq_x=clamp_max(freq_x,f3);
            T phase_x=freq_x*(T)two_pi*time+(T)half_pi;
            T s_x=s1+s2*time;
            s_x=clamp_max(s_x,s3);
            v.x=(T)sin(phase_x)*s_x;
        }
        else if(time<t1){   ////slow down
            T f1=(T)15;T f2=(T)0;T s1=(T)15;T s2=(T)0;
            T freq_x=freq_x0-f1*(time-t0);freq_x=clamp_min(freq_x,f2);
            T phase_x=phase_x0+(T)freq_x*(T)two_pi*(time-t0);
            T s_x=s_x0-s1*(time-t0);s_x=clamp_min(s_x,s2);
            v.x=(T)sin(phase_x)*s_x;
            LOG::cout<<"slow down: "<<phase_x<<", "<<s_x<<": "<<v.x<<std::endl;
        }
        else if(time<t2){   ////speed up for tris
            T f1=(T)20;T f2=(T)3;T s1=(T)4;T s2=(T)1.5;
            T freq_x=freq_x1+f1*(time-t1);
            freq_x=clamp_max(freq_x,f2);
            T phase_x=(T)half_pi+/*phase_x1+*/freq_x*(T)two_pi*(time-t1);
            T s_x=s_x1+s1*(time-t1);
            s_x=clamp_max(s_x,s2);
            v.x=(T)sin(phase_x)*s_x;
            LOG::cout<<"speed up: "<<phase_x<<", "<<s_x<<", "<<v.x<<std::endl;
        }
        else if(time<t3){   ////slow down a little bit
            T f1=(T)2;T f2=(T)0;T s1=(T)2;T s2=(T)0;
            T freq_x=freq_x2-f1*(time-t2);freq_x=clamp_min(freq_x,f2);
            T phase_x=phase_x2+(T)freq_x*(T)two_pi*(time-t2);
            T s_x=s_x2-s1*(time-t2);s_x=clamp_min(s_x,s2);
            v.x=(T)sin(phase_x)*s_x;
        }
        else if(time<t4){   ////speed up for segs
            //T f1=(T)8;T f2=(T)10;T s1=(T)2;T s2=(T)4;
            //T freq_x=freq_x3+f1*(time-t3);
            //freq_x=clamp_max(freq_x,f2);
            //T phase_x=phase_x3+freq_x*(T)two_pi*(time-t3);
            //T s_x=s_x3+s1*(time-t3);
            //s_x=clamp_max(s_x,s2);
            //v.x=(T)sin(phase_x)*s_x;
            T s_x=(T).3+(T)3*(time-t3);
            s_x=clamp_max(s_x,(T).6);
            T phase_x=(T)8*(T)two_pi*(time-t3);
            v.x=cos(phase_x)*s_x+(T).1;
            v.z=sin(phase_x)*s_x;
            LOG::cout<<"phase "<<phase_x<<", "<<s_x<<std::endl;
        }

        T phase_z=(T)2*(T)two_pi*time+(T)half_pi;T s_z=(T).25;
        if(time<t3)v.z=(T)sin(phase_z)*s_z;

        tubes(1)->use_reinitialize_eulerian_particles=true;
        
        //trans_t1=(T).05;trans_t2=(T).2;

        if(time>=trans_t1&&time<trans_t2){
            ////3d->2d
            T rescale_z_max=(T)6;
            T rescale_z=(T)1+(time-trans_t1)*(T)40;rescale_z=clamp_max(rescale_z,rescale_z_max);
            T rescale_x=(T)1-(time-trans_t1)*(T)40/(rescale_z_max-(T)1);rescale_x=clamp_min(rescale_x,(T)0);
            tubes(1)->object_space_z_rescale=rescale_z;
            tubes(1)->object_space_x_rescale=rescale_x;

            if(rescale_x==(T)0){tubes(1)->Set_Constant_Thickness(thickness_scale*(T)2);}

            LOG::cout<<"trans1: "<<tubes(1)->object_space_z_rescale<<", "<<tubes(1)->object_space_x_rescale<<std::endl;
        }
        else if(time>=trans_t2){
            ////2d->1d
            //T frac=(T)1-(T)10*(time-trans_t2);frac=clamp_min(frac,(T).0);
            //T min_r=length_scale*(T)1;
            //if(tubes(1)->object_space_r>min_r)tubes(1)->object_space_r=object_space_r_init*frac;

            T rescale_z=(T)6-(time-trans_t2)*(T)60;rescale_z=clamp_min(rescale_z,(T)0);
            T rescale_x=(T)0;
            tubes(1)->object_space_z_rescale=rescale_z;
            tubes(1)->object_space_x_rescale=rescale_x;

            if(rescale_z==(T)0){tubes(1)->Set_Constant_Thickness(thickness_scale*(T)2);}

            LOG::cout<<"trans2: "<<tubes(1)->object_space_z_rescale<<", "<<tubes(1)->object_space_x_rescale<<std::endl;

            //LOG::cout<<"frac: "<<frac<<", "<<tubes(1)->object_space_r<<", "<<tubes(1)->radius<<std::endl;
        }
    }break;
    case 24:case 216:{
        vs.Resize(2);
        if(time>(T).1)tubes(2)->Enable(true,&example.particles);
        if(time>(T).3){vs(1)=TV((T)-.1,(T).05,(T)0);vs(2)=TV((T).1,(T)0,(T)0);}
        else{vs(1)=TV();vs(2)=TV();}
    }break;
    case 221:{
        vs.Resize(2);
        T f1=(T)10;T f2=(T)10;T s1=(T).6;T s2=(T).8;
        T phase1=f1*(T)two_pi*time;T phase2=(T)pi+f2*(T)two_pi*time;
        TV v1=TV::Axis_Vector(3)*(s1*sin(phase1));TV v2=TV::Axis_Vector(3)*(s2*sin(phase2));
        vs(1)=v1;vs(2)=v2;

        //T src_f=(T)5;T src_phase=src_f*(T)two_pi*time;
        //if(src_phase-(T)two_pi*(T)floor(src_phase/two_pi)<(T)pi){tubes(1)->source_speed=(T).2;tubes(2)->source_speed=(T).2;}
        //else{tubes(1)->source_speed=(T).0;tubes(2)->source_speed=(T).0;}
    }break;
    case 205:{
        T s0=(T)1e-3;
        TV point((T)-20*s0,(T)0,(T)0);
        TV normal1=ROTATION<TV>((T)pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane1(normal1,point);
        TV normal2=ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane2(normal2,point);

        TV point2((T)-12*s0,(T)0,(T)0);PLANE<T> plane3(normal1,point2);PLANE<T> plane4(normal2,point2);

        boundary_particle_hashtable.Remove_All();
        for(int i=1;i<=example.particles.X.m;i++){
            bool is_boundary=(plane1.Lazy_Outside(example.particles.X(i))&&plane2.Lazy_Inside(example.particles.X(i)))&&
                !(plane3.Lazy_Outside(example.particles.X(i))&&plane4.Lazy_Inside(example.particles.X(i)));
            if(is_boundary){boundary_particle_hashtable.Set(example.particles.guid(i));}
        } 

        TV pizza_v=TV((T).1,(T).05,(T)0);

        static bool first=true;
        if(first){
            T s0=(T)1e-3;T piz_radius=s0*(T)40;
            TV point3=TV();PLANE<T> plane5(normal1,point3);PLANE<T> plane6(normal2,point3);T r1=piz_radius-(T)1.6*s0;T r2=piz_radius+s0;T h=length_scale*(T).9;
            for(int i=1;i<=example.particles.X.m;i++){
                T r=(example.particles.X(i)-point3).Magnitude();
                if((plane5.Lazy_Outside(example.particles.X(i))&&plane6.Lazy_Inside(example.particles.X(i)))&&r>r1&&r<r2&&example.particles.X(i).y<h){
                    boundary_particle_hashtable2.Set(example.particles.guid(i),pizza_v);
                }
            }
            first=false;
        }

        ////advect particles in boundary_particle_hashtable2
        if(boundary_particle_hashtable2.Size()>0){
            for(int i=1;i<=example.particles.X.m;i++){
                if(boundary_particle_hashtable2.Contains(example.particles.guid(i))){
                    example.particles.X(i)+=pizza_v*dt;
                }
            }
        }
    }break;
    }
    
    kinematic_particle_velocity_hashtable.Remove_All();
    for(int i=1;i<=tubes.m;i++){
        PAINT_TUBE<TV>* tube=tubes(i);
        ////update tube source speed
        if(tube&&tube->source_speed_bspline){tube->source_speed=tube->source_speed_bspline->Evaluate(scaled_time);}
        ////update tube source thickness factor
        if(tube&&tube->source_thickness_factor_bspline){tube->source_thickness_factor=tube->source_thickness_factor_bspline->Evaluate(scaled_time);}
        ////update kinematic particles from tube
        if(tube){
            if(tube->use_prescribed_motion){
                tube->object_frame=tube->prescribed_motion.Prescribed_Frame(scaled_time);tube->object_twist=tube->prescribed_motion.Prescribed_Twist(dt,scaled_time);
            }
            else{
                TV tube_v=vs.m>0?vs(i):v;T_SPIN tube_r=rs.m>0?rs(i):r;
                tube->object_twist=TWIST<TV>(tube_v,tube_r);
                tube->object_frame.t+=tube->object_twist.linear*dt;
                tube->object_frame.r*=ROTATION<TV>::From_Rotation_Vector(tube->object_twist.angular*dt);
            }

            tube->Update_Position_Based_States();
            tube->Reinitialize_Eulerian_Particles(example.particles);
            tube->Update_Eulerian_Particle_Velocities(kinematic_particle_velocity_hashtable);
            //tube->Update_Eulerian_Particle_Positions_And_Mass(example.geometrical_codimensional_mesh_object,thickness_scale*min_tri_thickness_factor,rim_thickness_scale*min_seg_thickness_factor);  ////TODO: need to reset this for the splash example
            //tube->Update_Eulerian_Particle_Positions_And_Mass(example.geometrical_codimensional_mesh_object,thickness_scale,rim_thickness_scale,time);   ////TODO: comment out this for the twisting splash example
            tube->Update_Eulerian_Particle_Positions_And_Mass(example.geometrical_codimensional_mesh_object,thickness_scale*(T).2,rim_thickness_scale,time);   ////TODO: comment out this for the twisting splash example            
            tube->Update_Eulerian_Particle_Texture_Coordinates(example.particles);
        }
    }

    ////for boundary particles in piz example
    if(boundary_particle_hashtable.Size()>0){
        for(HASHTABLE_ITERATOR<int> iter(boundary_particle_hashtable);iter.Valid();iter.Next()){
            kinematic_particle_velocity_hashtable.Set(iter.Key(),TV());
        }
    }
    if(boundary_particle_hashtable2.Size()>0){
        for(HASHTABLE_ITERATOR<int,TV> iter(boundary_particle_hashtable2);iter.Valid();iter.Next()){
            kinematic_particle_velocity_hashtable.Set(iter.Key(),iter.Data());
            //LOG::cout<<"bd p2: "<<iter.Key()<<": "<<iter.Data()<<std::endl;
        }
    }

    ////update kinematic particle guid in mesher
    example.mesher.Clear_Eulerian_Particles();
    for(HASHTABLE_ITERATOR<int,TV> iter(kinematic_particle_velocity_hashtable);iter.Valid();iter.Next()){
        int guid=iter.Key();
        example.mesher.Add_Eulerian_Particle(guid);
        //LOG::cout<<"ep: "<<guid<<std::endl;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Paint_Color()
{
    switch(test){
    case 6:{
        //for(int i=1;i<=example.particles.X.m;i++){
        //    if(example.particles.X(i).x<(T)-.006)example.particles.color(i)=TV((T)1,0,0);
        //    else example.particles.color(i)=TV(0,(T)1,0);
        //}
    }break;
    case 9:{
        for(int i=1;i<=example.particles.X.m;i++){
            if(example.particles.X(i).x<(T)0)example.particles.color(i)=TV((T)1,0,0);
            else example.particles.color(i)=TV(0,(T)1,0);
        }
    }break;
    case 11:{
        for(int i=1;i<=example.particles.X.m;i++){
            if(example.particles.X(i).x<(T)-.005)example.particles.color(i)=TV((T)1,0,0);
            else example.particles.color(i)=TV(0,(T)1,0);
        }
    }break;
    case 12:{
        for(int i=1;i<=example.particles.X.m;i++){
            if(example.particles.X(i).x<(T)0)example.particles.color(i)=TV((T)1,0,0);
            else example.particles.color(i)=TV(0,(T)1,0);
        }
    }break;
    default:{
        for(int i=1;i<=example.particles.X.m;i++){example.particles.color(i)=TV((T)1,0,0);}
    }break;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Paint_Phis()
{
    if(!use_levelset)return;

    example.particles.Store_Phis(levelset_region_number);
    if(initial_phis.m>0){
        for(int c=1;c<=initial_phis.m;c++){
            for(int i=1;i<=example.particles.X.m;i++){example.particles.phis(c)(i)=initial_phis(c)(i);}
        }
        return;
    }

    switch(test){
    case 6:{
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=(T)-.006-example.particles.X(i).x;
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 10:{
        T s0=(T)5e-3;T paint_radius=s0*(T)2;TV paint_center=TV((T)0,s0*(T)40,(T)0);T h=paint_radius*(T)/*.4*/2;
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        //int axis=2;
        //levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=paint_center.y-h*(T).16;
        //levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=paint_center.y+h*(T).16;
        
        int axis=1;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=paint_center(axis)+h*(T).16;

        for(int c=1;c<=levelset_region_number;c++){for(int i=1;i<=example.particles.X.m;i++){example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));}}
    }break;
    case 12:{
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=(T)-.0003-example.particles.X(i).x;
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 13:{
        ARRAY<BOX<TV> > levelset_regions(4);for(int i=1;i<=4;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T).25),TV::Constant_Vector((T).25));
        levelset_regions(1).max_corner.x=levelset_regions(1).max_corner.z=(T)-.003;
        levelset_regions(2).min_corner.x=levelset_regions(2).max_corner.z=(T)-.003;
        levelset_regions(3).min_corner.x=levelset_regions(3).min_corner.z=(T)-.003;
        levelset_regions(4).min_corner.z=levelset_regions(4).max_corner.x=(T)-.003;
        for(int c=1;c<=4;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 20:{
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=example.particles.X(i).z;
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 24:case 216:{
        for(int c=1;c<=2;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                T sign=(T)0;
                if(c==1)sign=example.particles.X(i).x<(T)0?(T)-1:(T)1;
                else sign=example.particles.X(i).x>(T)0?(T)-1:(T)1;
                T phi=sign*(T)1e10;
                example.particles.phis(c)(i)=phi;
            }
        }
    }break;
    case 51:{
        ARRAY<BOX<TV> > levelset_regions(4);for(int i=1;i<=4;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=3;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)-20e-3;
        levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=(T)0;
        levelset_regions(3).max_corner(axis)=levelset_regions(4).min_corner(axis)=(T)20e-3;

        for(int c=1;c<=4;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 52:{
        ARRAY<BOX<TV> > levelset_regions(3);for(int i=1;i<=3;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=1;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T)-5e-3;
        levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=(T)5e-3;

        for(int c=1;c<=3;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 100:{
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=example.particles.X(i)(1);
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 101:{
        T interface_x=(T)7e-3;
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=example.particles.X(i)(1)-interface_x;
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 210:{
        switch(paint_tube_test){
        case 7:{
            for(int c=1;c<=2;c++){
                for(int i=1;i<=example.particles.X.m;i++){
                    T sign=(T)0;
                    if(c==1)sign=example.particles.X(i).y<(T).01?(T)-1:(T)1;
                    else sign=example.particles.X(i).y>=(T).01?(T)-1:(T)1;
                    T phi=sign*(T)1e10;
                    example.particles.phis(c)(i)=phi;
                }
            }
        }break;
        }
    }break;
    case 213:{
        tubes(1)->Initialize_Ring_Phis(example.particles,levelset_region_number);
    }break;
    case 214:{
        ARRAY<BOX<TV> > levelset_regions(levelset_region_number);for(int i=1;i<=levelset_region_number;i++)levelset_regions(i)=BOX<TV>(TV::Constant_Vector(-(T)1000),TV::Constant_Vector((T)1000));
        int axis=2;
        levelset_regions(1).max_corner(axis)=levelset_regions(2).min_corner(axis)=(T).04;
        levelset_regions(2).max_corner(axis)=levelset_regions(3).min_corner(axis)=(T).08;

        for(int c=1;c<=levelset_region_number;c++){
            for(int i=1;i<=example.particles.X.m;i++){
                example.particles.phis(c)(i)=levelset_regions(c).Signed_Distance(example.particles.X(i));
            }
        }
    }break;
    case 277:{
        T amp=(T)2e-3;
        T scale=3000;
        for(int i=1;i<=example.particles.X.m;i++){
            T x=example.particles.X(i)(1);
            T z=example.particles.X(i)(3);
            T fcn=amp*sin(x*scale);
            T diff=abs(z-fcn);
            if(z>fcn){
                example.particles.phis(1)(i)=diff;
                example.particles.phis(2)(i)=-diff;
            }
            else{
                example.particles.phis(1)(i)=-diff;
                example.particles.phis(2)(i)=diff;
            }
        }
       
    }break;
    case 666:{  //// Aquafresh
        TV middle=TV(0,0,0);
        for(int i=1;i<=example.particles.X.m;i++){
            middle+=example.particles.X(i);
        }
        middle/=T(example.particles.X.m);
        for(int i=1;i<=example.particles.X.m;i++){
            TV v=example.particles.X(i)-middle;
            // 1 2 1   A | B
            // 3 + 3 = --+--
            // 1 2 1   C | D
            v=TV(abs(v.x),abs(v.y),abs(v.z));
            T a_u=3,b_u=-1,c_u=0;
            T a_l=T(1)/a_u,b_l=-1,c_l=0;
            if(v.y>v.z){ // 2 / 1 : Upper side sliced by y=x
                T sign=v.y>a_u*v.z?(T)1:(T)-1; // Inside/Outside of region 2
                example.particles.phis(2)(i)=-sign*abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
                example.particles.phis(1)(i)=-example.particles.phis(2)(i);
                example.particles.phis(3)(i)=abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
            }else{       // 1 / 3 : Lower side sliced by y=x
                T sign=v.y<a_l*v.z?(T)1:(T)-1; // Inside/Outside of region 3
                example.particles.phis(3)(i)=-sign*abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
                example.particles.phis(1)(i)=-example.particles.phis(3)(i);
                example.particles.phis(2)(i)=abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
            }
        }
    }break;
    case 667:{  ////// 2 colors
        TV middle=TV(0,0,0);
        for(int i=1;i<=example.particles.X.m;i++){
            middle+=example.particles.X(i);
        }
        middle/=T(example.particles.X.m);
        
        for(int i=1;i<=example.particles.X.m;i++){
            example.particles.phis(1)(i)=example.particles.X(i).z;
            example.particles.phis(2)(i)=-example.particles.phis(1)(i);
        }
    }break;
    case 668:{  ////// 4 colors
        TV middle=TV(0,0,0);
        for(int i=1;i<=example.particles.X.m;i++){
            middle+=example.particles.X(i);
        }
        middle/=T(example.particles.X.m);
        for(int i=1;i<=example.particles.X.m;i++){
            TV v=example.particles.X(i)-middle;
            // 1 | 2
            // --+--
            // 3 | 4
            if(v.z<=0&&v.y>=0){// 1
                example.particles.phis(1)(i)=max(v.z,-v.y);
                example.particles.phis(2)(i)=-v.z;
                example.particles.phis(3)(i)=v.y;
                example.particles.phis(4)(i)=sqrt(v.z*v.z+v.y*v.y);
            }else if(v.z<=0&&v.y<=0){// 3
                example.particles.phis(3)(i)=max(v.z,v.y);
                example.particles.phis(1)(i)=-v.y;
                example.particles.phis(4)(i)=-v.z;
                example.particles.phis(2)(i)=sqrt(v.z*v.z+v.y*v.y);
            }else if(v.z>=0&&v.y>=0){// 2
                example.particles.phis(2)(i)=max(-v.z,-v.y);
                example.particles.phis(1)(i)=v.z;
                example.particles.phis(4)(i)=v.y;
                example.particles.phis(3)(i)=sqrt(v.z*v.z+v.y*v.y);
            }else if(v.z>=0&&v.y<=0){//4
                example.particles.phis(4)(i)=max(-v.z,v.y);
                example.particles.phis(2)(i)=-v.y;
                example.particles.phis(3)(i)=v.z;
                example.particles.phis(1)(i)=sqrt(v.z*v.z+v.y*v.y);
            }else{
                std::cout<<"Critical error detected in test 668. Abort."<<std::endl;
                exit(668);
            }
        }
    }break;
//    case 686:
//    case 685:
    case 677:
    case 672:
    case 671:
    case 670:{  //// Aquafresh
        TV middle=TV(0,0,0);
        for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=tubes.m;j++){
                PAINT_TUBE<TV>* tube=tubes(j);
                middle+=tube->object_frame.Inverse_Times(example.particles.X(i));
            }
        }
        middle/=T(example.particles.X.m);
        for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=tubes.m;j++){
                PAINT_TUBE<TV>* tube=tubes(j);
                TV v=tube->object_frame.Inverse_Times(example.particles.X(i))-middle;
                // 1 2 1   A | B
                // 3 + 3 = --+--
                // 1 2 1   C | D
                v=TV(abs(v.y),abs(v.x),abs(v.z)); // Notice a change here.
                T a_u=3,b_u=-1,c_u=0;
                T a_l=T(1)/a_u,b_l=-1,c_l=0;
                if(v.y>v.z){ // 2 / 1 : Upper side sliced by y=x
                    T sign=v.y>a_u*v.z?(T)1:(T)-1; // Inside/Outside of region 2
                    example.particles.phis(2)(i)=-sign*abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
                    example.particles.phis(1)(i)=-example.particles.phis(2)(i);
                    example.particles.phis(3)(i)=abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
                }else{       // 1 / 3 : Lower side sliced by y=x
                    T sign=v.y<a_l*v.z?(T)1:(T)-1; // Inside/Outside of region 3
                    example.particles.phis(3)(i)=-sign*abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
                    example.particles.phis(1)(i)=-example.particles.phis(3)(i);
                    example.particles.phis(2)(i)=abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
                }
            }
        }
    }break;
    case 685:
    case 686:
    case 683:
    case 682:
    case 679:
    case 678:
    case 675:
    case 674:{  //// Aquafresh - for this, x and z are flipped!
        TV middle=TV(0,0,0);
        for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=tubes.m;j++){
                PAINT_TUBE<TV>* tube=tubes(j);
                middle+=tube->object_frame.Inverse_Times(example.particles.X(i));
            }
        }
        middle/=T(example.particles.X.m);
        for(int i=1;i<=example.particles.X.m;i++){
            for(int j=1;j<=tubes.m;j++){
                PAINT_TUBE<TV>* tube=tubes(j);
                TV v=tube->object_frame.Inverse_Times(example.particles.X(i))-middle;
            
                // 1 2 1   A | B
                // 3 + 3 = --+--
                // 1 2 1   C | D
                v=TV(abs(v.y),abs(v.x),abs(v.z)); // Notice a change here.
                T a_u=3,b_u=-1,c_u=0;
                T a_l=T(1)/a_u,b_l=-1,c_l=0;
                if(v.y>v.z){ // 2 / 1 : Upper side sliced by y=x
                    T sign=v.y>a_u*v.z?(T)1:(T)-1; // Inside/Outside of region 2
                    example.particles.phis(2)(i)=-sign*abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
                    example.particles.phis(1)(i)=-example.particles.phis(2)(i);
                    example.particles.phis(3)(i)=abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
                }else{       // 1 / 3 : Lower side sliced by y=x
                    T sign=v.y<a_l*v.z?(T)1:(T)-1; // Inside/Outside of region 3
                    example.particles.phis(3)(i)=-sign*abs(a_l*v.z+b_l*v.y+c_l)/sqrt(a_l*a_l+b_l*b_l);
                    example.particles.phis(1)(i)=-example.particles.phis(3)(i);
                    example.particles.phis(2)(i)=abs(a_u*v.z+b_u*v.y+c_u)/sqrt(a_u*a_u+b_u*b_u);
                }
            }
        }
    }break;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Paint_Textures()
{
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Pizza(const ARRAY<TV>& X,ARRAY<T>& thickness,ARRAY<TV>& texture_coordinates,const TV& center,const T radius,const T max_thickness,const T min_thickness)
{
    thickness.Resize(X.m);texture_coordinates.Resize(X.m);
    for(int i=1;i<=X.m;i++){
        T scale=(T)400;TV offset=TV::Constant_Vector((T)10);
        //T frac=NOISE<T>::Turbulence(X(i)*scale,1,(T)1,(T)1);
        T frac=pow(NOISE<T>::Noise1(X(i)*scale+offset,1),(T)1);
        T frac_to_center=clamp_max((X(i)-center).Magnitude()/radius,(T)1);
        T frac2=(T).5*frac_to_center+(T)1;
        T base_thickness=max_thickness*(T).2;
        thickness(i)=(((T)1-frac)*min_thickness+frac*max_thickness)*frac2+base_thickness;
        //LOG::cout<<"pizza thickness "<<i<<": "<<frac<<", "<<frac2<<": "<<thickness(i)<<std::endl;
        
        texture_coordinates(i)=TV((X(i).x-center.x)/radius+(T).5,(X(i).z-center.z)/radius+(T).5,(T)0);
        texture_coordinates(i)=clamp(texture_coordinates(i),(T)0,(T)1);
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Particle_Length_Scale_Based_On_Tests()
{
    switch(test){
    case 205:{
        T s0=(T)1e-3;
        TV point((T)-4*s0,(T)0,(T)0);
        TV normal1=ROTATION<TV>((T)pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane1(normal1,point);
        TV normal2=ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane2(normal2,point);
        //LOG::cout<<"here"<<normal1<<", "<<normal2<<": "<<plane1.Lazy_Inside(TV())<<", "<<plane2.Lazy_Inside(TV())<<std::endl;

        ARRAY<bool> flag_changeable_contact(example.particles.X.m);flag_changeable_contact.Fill(true);
        for(int i=1;i<=example.particles.X.m;i++){
            bool is_adaptive=plane1.Lazy_Outside(example.particles.X(i))&&plane2.Lazy_Inside(example.particles.X(i));
            if(is_adaptive){
                example.particles.length_scale(i)=clamp_max(example.particles.length_scale(i),adaptive_length_scale_min);
                flag_changeable_contact(i)=false;
            }
            //if(!Is_Contact_With_Solid(example.particles.contact(i))){
            //    example.particles.length_scale(i)=clamp_max(example.particles.length_scale(i),length_scale*(T).5);
            //    flag_changeable_contact(i)=false;
            //}
        }
        SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(example.geometrical_codimensional_mesh,example.particles,2,/*refine*/true,&flag_changeable_contact);
    }break;
    case 214:{
        ARRAY<bool> flag_changeable_tank_bottom(example.particles.X.m);flag_changeable_tank_bottom.Fill(false);
        for(int i=1;i<=example.particles.X.m;i++){
            if(example.particles.X(i).y<length_scale){
                example.particles.length_scale(i)=length_scale;
                flag_changeable_tank_bottom(i)=true;
            }
        }
        SOLID_FLUID_COUPLED_EVOLUTION_CODIMENSIONAL<TV>::Smooth_Particle_Length_Scale(example.geometrical_codimensional_mesh,example.particles,2,/*refine*/true,&flag_changeable_tank_bottom);
    }break;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Part_Pizza_Mesh()
{
    switch(test){
    case 205:{
        ROTATION<TV> rot(piz_init_angle,TV::Axis_Vector(2));
        for(int i=1;i<=initial_X.m;i++){initial_X(i)=rot.Rotate(initial_X(i));}

        T s0=(T)1e-3;
        TV point((T)-20*s0,(T)0,(T)0);
        TV normal1=ROTATION<TV>((T)pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane1(normal1,point);
        TV normal2=ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane2(normal2,point);

        ARRAY<bool> rmv_flag(initial_X.m);rmv_flag.Fill(true);
        for(int i=1;i<=initial_X.m;i++){rmv_flag(i)=!(plane1.Lazy_Outside(initial_X(i))&&plane2.Lazy_Inside(initial_X(i)));}

        ARRAY<int> remap(initial_X.m);remap.Fill(0);int c=1;
        ARRAY<TV> initial_X_new;for(int i=1;i<=initial_X.m;i++)if(!rmv_flag(i)){remap(i)=c++;initial_X_new.Append(initial_X(i));}
        initial_X=initial_X_new;

        Remap_Elements(initial_mesh.tetrahedron_mesh.elements,remap,true);
        Remap_Elements(initial_mesh.triangle_mesh.elements,remap,true);
        Remap_Elements(initial_mesh.segment_mesh.elements,remap,true);
    }break;
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_Pizza_Texture_Coordinates(const ARRAY<TV>& X,const TV& center,const T radius,/*rst*/ARRAY<TV>& texture_coordinates)
{
    texture_coordinates.Resize(X.m);
    for(int i=1;i<=X.m;i++){
        texture_coordinates(i)=TV((X(i).x-center.x)/((T)2*radius)+(T).5,(X(i).z-center.z)/((T)2*radius)+(T).5,(T)0);
        texture_coordinates(i)=clamp(texture_coordinates(i),(T)0,(T)1);
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Update_Inactive_Particle_Based_On_Tests()
{
    switch(test){
    case 205:{
        T s0=(T)1e-3;
        TV point((T)-12*s0,(T)0,(T)0);
        TV normal1=ROTATION<TV>((T)pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane1(normal1,point);
        TV normal2=ROTATION<TV>((T)-pi/(T)8,TV::Axis_Vector(2)).Rotate(TV::Axis_Vector(3));PLANE<T> plane2(normal2,point);

        for(int i=1;i<=example.particles.X.m;i++){
            bool is_active=plane1.Lazy_Outside(example.particles.X(i))&&plane2.Lazy_Inside(example.particles.X(i));
            example.particles.inactive(i)=!is_active;
        }
    }break;
    case 210:{
        BOX<TV> box(TV::Constant_Vector((T)-FLT_MAX),TV::Constant_Vector((T)FLT_MAX));
        switch(paint_tube_test){
        case 1002:{ ////I
        }break;
        //case 1003:{
        //    box.min_corner.x=(T)-.015;
        //}break;
        }

        for(int i=1;i<=example.particles.X.m;i++){
            bool is_active=box.Lazy_Inside(example.particles.X(i));
            example.particles.inactive(i)=!is_active;
        }
    }break;
    /*
    case 286:{
        int frame=current_frame;
        PLANE<T> plane1((T)-1*TV::Axis_Vector(3),TV());
        PLANE<T> plane2(TV::Axis_Vector(3),TV());


        if(frame>=0){plane1.x1(3)=(T)-1000;plane2.x1(3)=(T)1000;}
        if(frame>=500){plane1.x1(3)=(T)-.037;}
        if(frame>=700){plane1.x1(3)=(T)-.02;}
        if(frame>=900){plane1.x1(3)=(T).0003;}
        if(frame>=1300){plane1.x1(3)=(T)-.0192;}
        if(frame>=1450){
            plane1.x1(3)=(T)-.04;
            plane2.x1(3)=(T)0.032;
        }
        if(frame>=1600){
            plane1.x1(3)=(T)-1000;
        }

        for(int i=1;i<=example.particles.X.m;i++){
            bool is_active=plane1.Lazy_Inside(example.particles.X(i))&&plane2.Lazy_Inside(example.particles.X(i));
            example.particles.inactive(i)=!is_active;
        }
    }break;
    */
    }
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Initialize_After_Restart()
{
    ////put fiber span update code here for restart
    switch(test){
    case 43:case 90:{
        if(time<(T).35)fiber.Span_Fibers((T)1+time*(T)2);
        LOG::cout<<"restart fiber span "<<time<<std::endl;
    }break;
    case 211:{
        if(paint_trace_test==1){
            T span_time=clamp_max(time,(T).35);
            fiber.Span_Fibers((T)1+span_time*(T)1.5);
        }
        else if(paint_trace_test==2){
            T span_time=clamp_max(time,(T).3);
            fiber.Span_Fibers((T)1+span_time*(T)1.2);
        }

        LOG::cout<<"restart fiber span "<<time<<std::endl;
    }break;
    }

    if(sim_particle_boxes.m>0){Remove_Particles_In_Invalid_Boxes();}
    if(boundary_particle_boxes.m>0){Update_Particles_In_Boundary_Boxes();}
}

template<class TV>
void PhysBAM::PAINT_DRIVER<TV>::Add_Mesh_Objects()
{
    CODIMENSIONAL_MESH added_mesh;
    ARRAY<TV> added_X;
    ARRAY<TV> added_V;
    ARRAY<T> added_thickness;
    ARRAY<ARRAY<T> > added_phis;

    switch(test){
    case 208:{Add_Cheese_Filaments();}break;
    case 211:{
        static bool added=false;
        if(paint_trace_test==1&&use_letter_r&&!added){
            ARRAY<TV> letter_X;CODIMENSIONAL_MESH letter_mesh;ARRAY<T> letter_thickness;
            {BOX<TV> box(TV::Constant_Vector((T)-.028),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;box.max_corner.x=(T).04;
            CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_r",600,1,&box);
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(letter_mesh,letter_X,TV(),TV::Constant_Vector(length_scale*(T)3),length_scale);
            TV trans_X=TV((T).0205,(T).0,(T).005);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
            Add_Particles_And_Mesh(letter_X,letter_mesh,0,0);
            added=true;

            //ARRAY<TV> letter_X;CODIMENSIONAL_MESH letter_mesh;ARRAY<T> letter_thickness;
            //{BOX<TV> box(TV::Constant_Vector((T)-.025),TV::Constant_Vector((T).025));box.min_corner.x=(T)-.05;box.max_corner.x=(T).04;
            ////CODIMENSIONAL_MESH_GENERATOR<TV>::Read_Mesh_And_Attributes_From_File(letter_mesh,letter_X,&letter_thickness,0,0,"./new_r",600,1,&box);
            //CODIMENSIONAL_MESH_GENERATOR<TV>::Box(letter_mesh,letter_X,TV(),TV::Constant_Vector(length_scale*(T)3),length_scale);
            //TV trans_X=TV((T).0205,(T).01,(T).005);for(int i=1;i<=letter_X.m;i++)letter_X(i)+=trans_X;}
            //Add_Particles_And_Mesh(letter_X,letter_mesh,0,/*&letter_thickness*/0);
            //added=true;
        }
    }break;
    }
}